/*!
  -----------------------------------------------------------------------------

  module: gcn00_1.h

  -----------------------------------------------------------------------------

  responsible:  Bernd Vorsprach (D025588 - Bernd.Vorsprach@SAP-AG.de)

  special area: DBM Error Messages

  description:  specification of dbm error messages

  version:      min 7.2.

  -----------------------------------------------------------------------------

                          Copyright by SAP AG, 1998

  -----------------------------------------------------------------------------
*/


#ifndef _GCN00_1_H_
#define _GCN00_1_H_

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/
#include "geo200.h"  // teo200_EventList
#include "geo47.h"   // Unicode _TCHAR
#include "hin100.h"

/*
  -----------------------------------------------------------------------------
  Definitions
  -----------------------------------------------------------------------------
*/

#define TERR_CN00_1               teo200_EventList::Error_eeo200
#define TWRN_CN00_1               teo200_EventList::Warning_eeo200
#define TINF_CN00_1               teo200_EventList::Info_eeo200

#define N_CN00_1(_n)              STATIC_CAST(teo200_EventList::teo200_EventID, (-25000+_n))

#define FUNCTION_DBG_MCN00_1(_s)   const _TCHAR * pName_cn00_1 = _s
#define FUNCTION_NAME_MCN00_1      pName_cn00_1

inline teo200_EventList::teo200_EventID cn00_1_ErrId
    ( const teo200_EventList::teo200_EventID _a, 
      const teo200_EventList::teo200_EventType _b, 
      const _TCHAR * const _c, 
      const _TCHAR * const _d ) 
{
  return _a;
} // end cn00_1_ErrId

inline const _TCHAR * cn00_1_ErrTxt
    ( const teo200_EventList::teo200_EventID _a, 
      const teo200_EventList::teo200_EventType _b, 
      const _TCHAR * const _c, 
      const _TCHAR * const _d ) 
{
  return _d;
} // end cn00_1_ErrTxt

/*
  -----------------------------------------------------------------------------
  Error Codes
  -----------------------------------------------------------------------------
*/
#define  OK_CN00_1                                  0 , TINF_CN00_1, _T("DBM"), _T("OK: everything works fine")

#define  ERR_CN00_1                     N_CN00_1(   1), TERR_CN00_1, _T("DBM"), _T("ERR: common error")
#define  ERR_NOTIMPL_CN00_1             N_CN00_1(   2), TERR_CN00_1, _T("DBM"), _T("ERR_NOTIMPL: sorry, command not yet implemented")
#define  ERR_FILE_CN00_1                N_CN00_1(   3), TERR_CN00_1, _T("DBM"), _T("ERR_FILE: file handling error")
#define  ERR_PARAM_CN00_1               N_CN00_1(   4), TERR_CN00_1, _T("DBM"), _T("ERR_PARAM: wrong parameters")
#define  ERR_MEM_CN00_1                 N_CN00_1(   5), TERR_CN00_1, _T("DBM"), _T("ERR_MEM: memory allocation error")
#define  ERR_RTE_CN00_1                 N_CN00_1(   6), TERR_CN00_1, _T("DBM"), _T("ERR_RTE: runtime environment error")
#define  ERR_RTEEXT_CN00_1              N_CN00_1(   7), TERR_CN00_1, _T("DBM"), _T("ERR_RTEEXT: runtime environment error")
#define  ERR_NOUTILSESSION_CN00_1       N_CN00_1(   8), TERR_CN00_1, _T("DBM"), _T("ERR_NOUTILSESSION: no utility session available")
#define  ERR_NOSQLUSERSESSION_CN00_1    N_CN00_1(   9), TERR_CN00_1, _T("DBM"), _T("ERR_NOSQLUSERSESSION: no sql user session available")
#define  ERR_NOBACKUPSESSION_CN00_1     N_CN00_1(  10), TERR_CN00_1, _T("DBM"), _T("ERR_NOBACKUPSESSION: no backup session available")
#define  ERR_NOHISTORYSESSION_CN00_1    N_CN00_1(  11), TERR_CN00_1, _T("DBM"), _T("ERR_NOHISTORYSESSION: no backup history session available")
#define  ERR_SQL_CN00_1                 N_CN00_1(  12), TERR_CN00_1, _T("DBM"), _T("ERR_SQL: sql error")
#define  ERR_TOOSMALL_CN00_1            N_CN00_1(  13), TERR_CN00_1, _T("DBM"), _T("ERR_TOOSMALL: packet size too small for answer")
#define  ERR_NOMEDIA_CN00_1             N_CN00_1(  14), TERR_CN00_1, _T("DBM"), _T("ERR_NOMEDIA: media file not found")
#define  ERR_MENOTFOUND_CN00_1          N_CN00_1(  15), TERR_CN00_1, _T("DBM"), _T("ERR_MENOTFOUND: medium not found")
#define  ERR_MENOVALID_CN00_1           N_CN00_1(  16), TERR_CN00_1, _T("DBM"), _T("ERR_MENOVALID: invalid medium definition")
#define  ERR_BANOVALIDTYPE_CN00_1       N_CN00_1(  17), TERR_CN00_1, _T("DBM"), _T("ERR_BANOVALIDTYPE: invalid backup type")
#define  ERR_NOHISTORY_CN00_1           N_CN00_1(  18), TERR_CN00_1, _T("DBM"), _T("ERR_NOHISTORY: backup history file not found")
#define  ERR_HIENTRYNOTFOUND_CN00_1     N_CN00_1(  19), TERR_CN00_1, _T("DBM"), _T("ERR_HIENTRYNOTFOUND: backup history entry not found")
#define  ERR_NOHELP_CN00_1              N_CN00_1(  20), TERR_CN00_1, _T("DBM"), _T("ERR_NOHELP: xparam help file not found")
#define  ERR_XPNOTFOUND_CN00_1          N_CN00_1(  21), TERR_CN00_1, _T("DBM"), _T("ERR_XPNOTFOUND: parameter not found")
#define  ERR_XPSYNTAX_CN00_1            N_CN00_1(  22), TERR_CN00_1, _T("DBM"), _T("ERR_XPSYNTAX: xparam syntax error")
#define  ERR_COMMAND_CN00_1             N_CN00_1(  23), TERR_CN00_1, _T("DBM"), _T("ERR_COMMAND: unknown command \"%s\"")
#define  ERR_PARAMEXIST_CN00_1          N_CN00_1(  24), TERR_CN00_1, _T("DBM"), _T("ERR_PARAMEXIST: param file of database already exists")
#define  ERR_NOXPSESSION_CN00_1         N_CN00_1(  25), TERR_CN00_1, _T("DBM"), _T("ERR_NOXPSESSION: param session inactive")
#define  ERR_XPSESSION_CN00_1           N_CN00_1(  26), TERR_CN00_1, _T("DBM"), _T("ERR_NOXPSESSION: param session active")
#define  ERR_XPCHECK_CN00_1             N_CN00_1(  27), TERR_CN00_1, _T("DBM"), _T("ERR_XPCHECK: param check failure/request")
#define  ERR_XPINVCONF_CN00_1           N_CN00_1(  28), TERR_CN00_1, _T("DBM"), _T("ERR_XPINFCONV: invalid value")
#define  ERR_XPNOCHECK_CN00_1           N_CN00_1(  29), TERR_CN00_1, _T("DBM"), _T("ERR_NOCHECK: no running parameter check")
#define  ERR_XPNOMODIFY_CN00_1          N_CN00_1(  30), TERR_CN00_1, _T("DBM"), _T("ERR_XPNOMODIFY: parameter must not be modified")
#define  ERR_XPDIRTY_CN00_1             N_CN00_1(  31), TERR_CN00_1, _T("DBM"), _T("ERR_XPDIRTY: param file is dirty")
#define  ERR_XPREADONLY_CN00_1          N_CN00_1(  32), TERR_CN00_1, _T("DBM"), _T("ERR_XPREADONLY: parameter is read only")
#define  ERR_SYSTEM_CN00_1              N_CN00_1(  33), TERR_CN00_1, _T("DBM"), _T("ERR_SYSTEM: system call error")
#define  ERR_LOGON_CN00_1               N_CN00_1(  34), TERR_CN00_1, _T("DBM"), _T("ERR_LOGON: logon to DBM Server failed")
#define  ERR_SYSLOGON_CN00_1            N_CN00_1(  35), TERR_CN00_1, _T("DBM"), _T("ERR_SYSLOGON: logon to system failed")
#define  ERR_EXECUTE_CN00_1             N_CN00_1(  36), TERR_CN00_1, _T("DBM"), _T("ERR_EXECUTE: error in program execution")
#define  ERR_DEVCHECK_CN00_1            N_CN00_1(  37), TERR_CN00_1, _T("DBM"), _T("ERR_DEVCHECK: error while checking devspace")
#define  ERR_DEVTYPE_CN00_1             N_CN00_1(  38), TERR_CN00_1, _T("DBM"), _T("ERR_DEVTYPE: wrong devspace type")
#define  ERR_STATE_CN00_1               N_CN00_1(  39), TERR_CN00_1, _T("DBM"), _T("ERR_STATE: cannot determine current db state")
#define  ERR_DBRUN_CN00_1               N_CN00_1(  40), TERR_CN00_1, _T("DBM"), _T("ERR_DBRUN: database is runnning")
#define  ERR_XPOVERFLOW_CN00_1          N_CN00_1(  41), TERR_CN00_1, _T("DBM"), _T("ERR_XPOVERFLOW: value out of margin")
#define  ERR_BAOVERWRITE_CN00_1         N_CN00_1(  42), TERR_CN00_1, _T("DBM"), _T("ERR_BAOVERWRITE: file already exists, cannot overwrite")
#define  ERR_NOUSER_CN00_1              N_CN00_1(  43), TERR_CN00_1, _T("DBM"), _T("ERR_NOUSER: required user not set")
#define  ERR_BASAVETYPECON_CN00_1       N_CN00_1(  44), TERR_CN00_1, _T("DBM"), _T("ERR_BASAVETYPECON: backup type conflicts with medium definition")
#define  ERR_BANOSAVETYPE_CN00_1        N_CN00_1(  45), TERR_CN00_1, _T("DBM"), _T("ERR_BANOSAVETYPE: backup type not defined")
#define  ERR_BANOAUTOMED_CN00_1         N_CN00_1(  46), TERR_CN00_1, _T("DBM"), _T("ERR_BANOAUTOMED: no autolog medium found in media list")
#define  ERR_DBEXIST_CN00_1             N_CN00_1(  47), TERR_CN00_1, _T("DBM"), _T("ERR_DBEXIST: database instance already exists")
#define  ERR_USRSAVE_CN00_1             N_CN00_1(  48), TERR_CN00_1, _T("DBM"), _T("ERR_USRSAVE: could not save user data")
#define  ERR_USRREAD_CN00_1             N_CN00_1(  49), TERR_CN00_1, _T("DBM"), _T("ERR_USRREAD: could not read user data")
#define  ERR_USRFAIL_CN00_1             N_CN00_1(  50), TERR_CN00_1, _T("DBM"), _T("ERR_USRFAIL: user authorization failed")
#define  ERR_USREXISTS_CN00_1           N_CN00_1(  51), TERR_CN00_1, _T("DBM"), _T("ERR_USREXISTS: user entry already exists")
#define  ERR_NODBNAME_CN00_1            N_CN00_1(  52), TERR_CN00_1, _T("DBM"), _T("ERR_NODBNAME: database name required")
#define  ERR_DBNAME_CN00_1              N_CN00_1(  53), TERR_CN00_1, _T("DBM"), _T("ERR_DBNAME: do not use dbname")
#define  ERR_TOOMANYPARAM_CN00_1        N_CN00_1(  54), TERR_CN00_1, _T("DBM"), _T("ERR_TOOMANYPARAM: too many parameters")
#define  ERR_TOOFEWPARAM_CN00_1         N_CN00_1(  55), TERR_CN00_1, _T("DBM"), _T("ERR_TOOFEWPARAM: too few parameters")
#define  ERR_USRDELSELF_CN00_1          N_CN00_1(  56), TERR_CN00_1, _T("DBM"), _T("ERR_USRDELSELF: cannot delete current user")
#define  ERR_USRDELCOLD_CN00_1          N_CN00_1(  57), TERR_CN00_1, _T("DBM"), _T("ERR_USRDELCOLD: cannot delete DBM user")
#define  ERR_USRDELSDBA_CN00_1          N_CN00_1(  58), TERR_CN00_1, _T("DBM"), _T("ERR_USRDELSDBA: cannot delete sysdba")
#define  ERR_NEEDLOGON_CN00_1           N_CN00_1(  59), TERR_CN00_1, _T("DBM"), _T("ERR_NEEDLOGON: DBM Server logon required")
#define  ERR_NEEDSYSLOGON_CN00_1        N_CN00_1(  60), TERR_CN00_1, _T("DBM"), _T("ERR_NEEDSYSLOGON: system logon required")
#define  ERR_LASTREG_CN00_1             N_CN00_1(  61), TERR_CN00_1, _T("DBM"), _T("ERR_LASTREG: cannot delete last registration of instance")
#define  ERR_NOENTRY_CN00_1             N_CN00_1(  62), TERR_CN00_1, _T("DBM"), _T("ERR_NOENTRY: no entry found")
#define  ERR_MISSRIGHT_CN00_1           N_CN00_1(  63), TERR_CN00_1, _T("DBM"), _T("ERR_MISSRIGHT: no server rights for this command")
#define  ERR_DBNORUN_CN00_1             N_CN00_1(  64), TERR_CN00_1, _T("DBM"), _T("ERR_DBNORUN: database is not running")
#define  ERR_NOUSERS_CN00_1             N_CN00_1(  65), TERR_CN00_1, _T("DBM"), _T("ERR_NOUSERS: no users found")
#define  ERR_VERSION_CN00_1             N_CN00_1(  66), TERR_CN00_1, _T("DBM"), _T("ERR_VERSION: incompatible versions (param and kernel)")
#define  ERR_NOSERVICESESSION_CN00_1    N_CN00_1(  67), TERR_CN00_1, _T("DBM"), _T("ERR_NOSERVICESESSION: no service kernel session available")
#define  ERR_XPMIN_CN00_1               N_CN00_1(  68), TERR_CN00_1, _T("DBM"), _T("ERR_XPMIN: parameter value smaller than minimum allowed")
#define  ERR_XPMAX_CN00_1               N_CN00_1(  69), TERR_CN00_1, _T("DBM"), _T("ERR_XPMAX: parameter value higher than maximum allowed")
#define  ERR_XPVALUESET_CN00_1          N_CN00_1(  70), TERR_CN00_1, _T("DBM"), _T("ERR_XPVALUESET: parameter value not in value set")
#define  ERR_USRFRM_CN00_1              N_CN00_1(  71), TERR_CN00_1, _T("DBM"), _T("ERR_USRFRM: invalid user record format")
#define  ERR_NOREPLY_CN00_1             N_CN00_1(  72), TERR_CN00_1, _T("DBM"), _T("ERR_NOREPLY: no reply available")
#define  ERR_TOOLCHK_CN00_1             N_CN00_1(  73), TERR_CN00_1, _T("DBM"), _T("ERR_TOOLCHK: external backup tool not found")
#define  ERR_MEDIUMCHK_CN00_1           N_CN00_1(  74), TERR_CN00_1, _T("DBM"), _T("ERR_MEDIUMCHK: medium cannot be used with this external backup tool")
#define  ERR_PREPARE_CN00_1             N_CN00_1(  75), TERR_CN00_1, _T("DBM"), _T("ERR_PREPARE: preparation of backup operation failed")
#define  ERR_DBREQ_CN00_1               N_CN00_1(  76), TERR_CN00_1, _T("DBM"), _T("ERR_DBREQ: cannot start database kernel request")
#define  ERR_TOOLREQ_CN00_1             N_CN00_1(  77), TERR_CN00_1, _T("DBM"), _T("ERR_TOOLREQ: error while starting external backup tool")
#define  ERR_OPCHK_CN00_1               N_CN00_1(  78), TERR_CN00_1, _T("DBM"), _T("ERR_OPCHK: cannot check state of backup operation")
#define  ERR_POSTOP_CN00_1              N_CN00_1(  79), TERR_CN00_1, _T("DBM"), _T("ERR_POSTOP: error while finishing backup operation ")
#define  ERR_BACKUPOP_CN00_1            N_CN00_1(  80), TERR_CN00_1, _T("DBM"), _T("ERR_BACKUPOP: backup operation was unsuccessful")
#define  ERR_CLEANUP_CN00_1             N_CN00_1(  81), TERR_CN00_1, _T("DBM"), _T("ERR_CLEANUP: cannot free all operational ressources after backup operation")
#define  ERR_MISSMIRR_CN00_1            N_CN00_1(  82), TERR_CN00_1, _T("DBM"), _T("ERR_MISSMIRR: missing data for mirror devspace")
#define  ERR_MAXDATA_CN00_1             N_CN00_1(  83), TERR_CN00_1, _T("DBM"), _T("ERR_MAXDATA: max data size exceeded")
#define  ERR_MAXDEVCNT_CN00_1           N_CN00_1(  84), TERR_CN00_1, _T("DBM"), _T("ERR_MAXDEVCNT: max device count exceeded")
#define  ERR_NOADDD_CN00_1              N_CN00_1(  85), TERR_CN00_1, _T("DBM"), _T("ERR_NOADDD: configuration update not allowed (param file is inactive)")
#define  ERR_KEYEXISTS_CN00_1           N_CN00_1(  86), TERR_CN00_1, _T("DBM"), _T("ERR_KEYEXISTS: installation key \"%s\" exists")
#define  ERR_SERVICEDB_CN00_1           N_CN00_1(  87), TERR_CN00_1, _T("DBM"), _T("ERR_SERVICEDB: cannot register service kernel for %s")
#define  ERR_WRONGINST_CN00_1           N_CN00_1(  88), TERR_CN00_1, _T("DBM"), _T("ERR_WRONGINST: DBM Server instroot is %s")
#define  ERR_INVDBNAME_CN00_1           N_CN00_1(  89), TERR_CN00_1, _T("DBM"), _T("ERR_INVDBNAME: database name %s is reserved")
#define  ERR_PARAMVER_CN00_1            N_CN00_1(  90), TERR_CN00_1, _T("DBM"), _T("ERR_PARAMVER: wrong version of param help file (cserv.pcf):\n%s\n%s")
#define  ERR_DBAUNKNOWN_CN00_1          N_CN00_1(  91), TERR_CN00_1, _T("DBM"), _T("ERR_DBAUNKNOWN: SYSDBA unknown")
#define  ERR_DOMPUNKNOWN_CN00_1         N_CN00_1(  92), TERR_CN00_1, _T("DBM"), _T("ERR_DOMPUNKNOWN: DOMAIN password unknown")
#define  ERR_DBAWRONG_CN00_1            N_CN00_1(  93), TERR_CN00_1, _T("DBM"), _T("ERR_DBAWRONG: wrong SYSDBA")
#define  ERR_EBFDIRTY_CN00_1            N_CN00_1(  94), TERR_CN00_1, _T("DBM"), _T("ERR_EBFDIRTY: invalid format of external backup history")
#define  ERR_NOEBIDLIST_CN00_1          N_CN00_1(  95), TERR_CN00_1, _T("DBM"), _T("ERR_NOEBIDLIST: no information on external backup IDs available")
#define  ERR_DOMPWRONG_CN00_1           N_CN00_1(  96), TERR_CN00_1, _T("DBM"), _T("ERR_DOMPWRONG: wrong DOMAIN password")
#define  ERR_IMPERSONATE_CN00_1         N_CN00_1(  97), TERR_CN00_1, _T("DBM"), _T("ERR_IMPERSONATE: cannot impersonate to database owner")
#define  ERR_ASCII_CN00_1               N_CN00_1(  98), TERR_CN00_1, _T("DBM"), _T("ERR_ASCII: 7-Bit ASCII overflow in parameter")
#define  ERR_NOTCOLD_CN00_1             N_CN00_1(  99), TERR_CN00_1, _T("DBM"), _T("ERR_NOTCOLD: database have to be cold")

#define  ERR_UNKNOWN_CN00_1             N_CN00_1(1000), TERR_CN00_1, _T("DBM"), _T("ERR_UNKNOWN: unknown error")

#define  INF_DBMCMD_CN00_1              0,              TINF_CN00_1, _T("DBM"), _T("command %s %.*s")
#define  INF_DBMERR_CN00_1              0,              TINF_CN00_1, _T("DBM"), _T("error   %.*s")
#define  INF_DBMTRC_CN00_1              0,              TINF_CN00_1, _T("DBM"), _T("trace   %.*s")


inline const void cn00_1_Trace
    ( const _TCHAR * const szTrace )
{
  
  teo200_EventList aEvent(_T(""), INF_DBMTRC_CN00_1,
                                  (_tcslen(szTrace) > (MAX_ARGUMENT_LENGTH_EO200 - 20)) ? (MAX_ARGUMENT_LENGTH_EO200 - 20) : _tcslen(szTrace),
                                  szTrace);

  tin100_GlobalLog::writeEntry (aEvent);
} // end cn00_1_Trace

#ifdef DBM_TRACE_FLAG
  #define DBMTRACE(szTrace)    cn00_1_Trace(szTrace)
#else  
  #define DBMTRACE(szTrace)
#endif 

#endif // _GCN00_1_H_
