/*!*****************************************************************************
  module:       hcn35.h
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  connection between dbmsrv and abstract external archiving tool 
  last change:  2000-11-29 10:16
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#if !defined (_HCN35_H_)
#define _HCN35_H_


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include"gcn00.h"
#include"gcn003.h"
#include"hcn90.h"
#include"hin29.h"
#include"hcn36.h"


// -----------------------------------------------------------------------------
// constants
// -----------------------------------------------------------------------------

#define MAX_NUMBER_OF_DATAPIPES_CN35 64
#define MAX_NUMBER_OF_TOOLPROCESSES_CN35 64

extern const char * szOSPathDelimiter;


// -----------------------------------------------------------------------------
// class tcn35_SaveInfo
// -----------------------------------------------------------------------------

class tcn35_SaveInfo
{
  public:
    const char *  pszMachName;      // TODO: should be a string<>
    char       *  pszDBName;        // TODO: should be a string<>
    char       *  DBRunDir;
    const char *  pszActionString;  // TODO: should be a string<>
    long          BlockSize;
    long          PageSize;
    const char *  pszMediumName;    // TODO: should be a string<>
    const char ** ppszFiles;        // TODO: should be a list, not an array
    int           NumFiles;
    
    char              * ReplyData;
    int               * ReplyLen;
    tin01_sql_session * SQLSession;
    const char        * KernelCommand;

    char * ExternalSaveID;
    char * SourceDBName;            // TODO: should be a string<>
    char * SourceDBHost;            // TODO: should be a string<>

    tcn35_SaveInfo(char                    * pcReplyData,
                   int                     * piReplyLen,
                   tin01_sql_session       * ptSQLSession,
                   const char              * pcszKernelCommand,
                   const tcn003_Medium     * pMedium,
                   const tcn003_MediaGroup * pMediaGroup,
                   const char              * pcszBackupType,
                   const int                 WithoutCheckpoint,
                   const tsp00_DbNamec       tDBName,
                   const char *              StartOfExternalBackupID=0,
                   const char *              EndOfExternalBackupID=0);

    tcn35_SaveInfo(char * pcReplyData,
                   int  * piReplyLen,
                   tsp00_DbNamec tDBName,
                   const char * TheSourceDBName,
                   const char * TheSourceDBHost);

    tcn35_SaveInfo(const tcn35_SaveInfo & Original);
    ~tcn35_SaveInfo();

    int IsGood;

    static const char * Action_String_DataMigration;
    static const char * Action_String_DataRecovery;
    static const char * Action_String_PagesMigration;
    static const char * Action_String_PagesRecovery;
    static const char * Action_String_Log;

  private:
    void SetAllMembersToZero();
    void FreeAllDynamicMembers();
    int  InitErrorHandlerAndProtocol(char * pcReplyData,
                                     int * piReplyLen,
                                     tsp00_DbNamec tDBName);
};


// -----------------------------------------------------------------------------
// class tcn35_ArchivingToolConnection
// -----------------------------------------------------------------------------

class tcn35_BackupToolConnector
{
  public:
    enum tcn35_ActionType {ABackup, ARestore, GiveBackupInfo};

    tcn35_BackupToolConnector(tcn35_ActionType Action, const tcn35_SaveInfo & rInfo);
    virtual ~tcn35_BackupToolConnector();

    void BeginBackupOperation();
    int  BackupOperationIsRunning();
    void EndBackupOperation();
    int  ListEBIDsOfAvailableBackups(tcn003_BackupHistoryEBIDList & AvailableEBIDs);

    tcn00_Error GiveLastError();
    int         GiveSQLResult();
    int         IsBad();

    virtual const char * GiveToolName()=0;
    virtual const char * GiveLongToolName()=0;

    const tcn35_ActionType GiveActionType() const;

  protected:
    tcn35_SaveInfo Info;

    enum ARequestState{NoRequestSend, RequestSend, CancelSendByDBMSrv, RequestReplied, CanceledByDBMSrv, TerminatedByDBMSrv};
    ARequestState StatusOfRequest;

    tcn003_BackupResult tTheDBResult;       // containing result, if StatusOfRequest==RequestReplied or ==CanceledByDBMSrv
    int                 SQLCodeOfRequest;   // containing the SQLCode, if StatusOfRequest==ReceivedReply or ==CanceledByDBMSrv
    boolean             SQLErrForRequest;   // true if SQLCodeOfRequest is an error
    tsp00_C256c         SQLErrText;         // containing errmsg if SQLCodeOfRequest is an error
    char *              FirstLogPageNo;     // containing copy of first Log Page number if StatusOfRequest==RequestReplied (maybe just "")
    char *              LastLogPageNo;      // containing copy of last Log Page number if StatusOfRequest==RequestReplied (maybe just "")

    enum AToolState{NotStarted, Started, EndedPartially, Ended, KilledByDBMSrv};
    AToolState StatusOfTool;

    int                        NumberOfCommandLines;
    char *                     CommandLine[MAX_NUMBER_OF_TOOLPROCESSES_CN35];       // TODO: should be an array of strings
    int                        NumberOfRunningToolProcesses;
    tni33_ABackgroundProcess * ToolProcess[MAX_NUMBER_OF_TOOLPROCESSES_CN35];
    int                        ToolProcessWasRunning[MAX_NUMBER_OF_TOOLPROCESSES_CN35];
    tni34_AFile *              ToolProcessOutput;   // TODO: must be a ...[MAX_NUMBER_OF_TOOLPROCESSES_CN35] if we start more than one process!!!!
    tni34_AFile *              ToolProcessError;
    int                        NumberOfExternalSaveIDsFound;
    char *                     ExternalSaveID[MAX_NUMBER_OF_TOOLPROCESSES_CN35];

    char *                     TempDateTimeString;
    tni34_ATextFile *          ToolProcessOutputAsText;

    virtual int CheckToolInstallation(int & IsInstalled)=0;
    virtual int Prepare()=0;
    virtual int StartTool();

    enum AWaitingPhase{NoWaiting, WaitForDBAndBackupTool, WaitForDB, WaitForBackupTool, WaitForKilled};
    AWaitingPhase  CurrentWaitingPhase;

    virtual int MaxWaitingTimeOfCurrentWaitingPhase();

    virtual int  CheckTool();
            int  DBActionWasSuccessful();
    virtual int  ToolIsRunning();
    virtual int  SignalEndToTool();
            void CancelTool();
            void TerminateTool();
    virtual int  ToolWasSuccessful();

    virtual int CollectToolSaveIDs()=0;
    virtual int PostOperationsAfterSuccess();

    virtual int MoveToolOutputToProtocolFile();
    virtual int CleanUpTool()=0;

    virtual int SumOfToolExitCodes();

    virtual int CreatePipes();
    virtual int RemovePipes();

    virtual int CheckMedium();

    virtual int PrepareToolOutputOfAvailableBackups();
    virtual int GiveNextAvailableBackup(tcn003_BackupHistoryEBIDStatus & Status,
                                        const char * & EBID,
                                        size_t       & EBIDLength,
                                        const char * & TypeString,
                                        const char * & DateTimeString,
                                        int          & ReachedEnd);
    virtual int GiveInformationOfLine(const char *   Line,
                                      const char * & EBIDInLine,
                                      size_t       & EBIDLength,
                                      const char * & TypeString,
                                      const char * & DateTimeString,
                                      int          & LineContainsEBID)=0;
    virtual int CloseToolOutputOfAvailableBackups();


    int ExistsToolProcesses();

    const char * ActionStringForDBMSaveType(const char * DBMSaveType, int WithoutCheckpoint);
    const char * ActionStringForActionStringN(const char *ActionString, size_t LengthActionString);

    const char * GivePathOfExternalBackupHistory(void);
    const char * GivePathOfBackupHistory(void);
  
  private:
    int  StartDBAction();
    int  DBIsReady();
    
    int  DBActionIsRunning();
    int  KernelMayBeUsingLogDevspaceForLogRestore();

    void FreeLogAfterLogBackup();
    void SignalEndToDB();
    void TerminateDBAction();

    int  MoveFileToProtocolFile(const char *FileDesignation, tni34_AFile & TheFileToMove);
    void CleanUp();

    int ToolMustNotBeSuccessful();

    int  SendBackupRestoreRequestToKernel(tin01_sql_session & Session,
                                          const char        * KernelCommand,
                                          int               & SQLCode,
                                          int                 BeQuiet=0,
                                          int                 ForBackup=1);
    int  ReplyAvailable(tin01_sql_session & Session);
    void ReceiveReplyFromDBKernel(tin01_sql_session   & Session,
                                  int                 & SQLCode,
                                  tcn003_BackupResult & tResult,
                                  boolean             & IsNoSQLError);
    void ReceiveBackupRestoreRequestReplyFromDBKernel(int ForLogDelete=0);
    int  GetStatusOfUtilityTask(tcn003_BackupResult &StatusInformation);

    const char * GetSaveIDFromDBMKnl(const char * AbsoluteNameOfControlKnl,
                                     const char * Label,
                                     const char * szDBStamp1Date,
                                     const char * szDBStamp1Time);

    void WriteLogFile();
    void FillReplyBuffer();

    tcn35_ActionType ActionType;

    int            NumberOfExistingPipes;
    tni34_APipe   *TheDataPipe[MAX_NUMBER_OF_DATAPIPES_CN35];

    long StartOfCurrentWaitingPhase;
    int  WaitForeverOnDB;

    char * PathOfExternalBackupHistory;
    char * PathOfBackupHistory;
};

class tcn35_BackupToolConnectorEnum
{
    public:
        enum AValue{ADSM=0, NSR=1, BACK=2, None=3}; // =... just to be sure

        explicit tcn35_BackupToolConnectorEnum(const AValue TheValue);
                 tcn35_BackupToolConnectorEnum(const tcn35_BackupToolConnectorEnum & Original);     // not explicit for HP
        explicit tcn35_BackupToolConnectorEnum(const tcn003_BackupHistoryMediumLine * pMediumLine);
        explicit tcn35_BackupToolConnectorEnum(const char * MediumName);

        operator AValue() const;
        int operator == (const tcn35_BackupToolConnectorEnum &Other) const;

        static int NumberOfBackupToolConnectors();

    private:
        AValue Value;
        AValue DetermineToolFromMediumName(const char *MediumName) const;
};

/*!-----------------------------------------------------------------------------
   Function: cn35_DetermineBackupTool

   description: Determines desired external backup tool from the name of the
                medium supplied through rInfo. If a external backup tool is associated
                with the medium the function creates and returns the right
                BackupToolConnector object. If it is not a medium for an external
                backup tool 0 is returned in rpDeterminedTool.

   arguments:   ActionType         [in]  descriebes the type of the action (backup/restore)
                rInfo              [in]  reference to const Info-object 
                rpDeterminedTool   [out] reference to pointer to object representing
                                         the right external backup tool
                Error              [out] contains the error code if a problem occured
  
   return value:
                <UL>
                    <LI> 0, if an error occured
                    <LI> 1, if no error occured
                </UL>
   -----------------------------------------------------------------------------*/

int cn35_DetermineBackupTool(tcn35_BackupToolConnector::tcn35_ActionType    ActionType,
                             const tcn35_SaveInfo                         & rInfo,
                             tcn35_BackupToolConnector                   *& rpDeterminedTool);

int cn35_ConstructBackupToolConnector(tcn35_BackupToolConnector::tcn35_ActionType    ActionType,
                                      const tcn35_BackupToolConnectorEnum          & BackupToolAsEnum,     
                                      const tcn35_SaveInfo                         & rInfo,
                                      tcn35_BackupToolConnector                   *& rpDeterminedTool);

/*!-----------------------------------------------------------------------------
   Function: cn35_IsExternalBackupTool

   description: Determines if pMediumLine is a medium using an external backup tool or not

   arguments:   pMediumLine [in] the medium as found in the backup history (dbm.mdf)
  
   return value:
                <UL>
                    <LI> 0, if no external backup tool was used with pMediumLine
                    <LI> 1, if an external backup tool was used with pMediumLine
                </UL>
   -----------------------------------------------------------------------------*/

int cn35_IsExternalBackupTool(const tcn003_BackupHistoryMediumLine * pMediumLine);


/*!-----------------------------------------------------------------------------
   Function: cn35_IsExternalBackupTool

   description: Determines if a Medium given by pMedium or pMediaGroup is a medium
                using an external backup tool or not

   arguments:   pMedium     [in] pointer to the medium
                pMediaGroup [in] pointer to the media group (used if 0==pMedium)
  
   return value:
                <UL>
                    <LI> 0, if no external backup tool is used with pMedium or pMediaGroup
                    <LI> 1, if an external backup tool is used with pMedium or pMediaGroup
                </UL>
    -----------------------------------------------------------------------------*/
   


int cn35_IsExternalBackupTool(const tcn003_Medium * pMedium, const tcn003_MediaGroup * pMediaGroup);

#endif
