/*******************************************************************************
  module:       hcn352.h
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  connection between dbmsrv and external backup tool NetWorker
  last change:  2000-11-29 10:33
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#if !defined (_HCN352_H_)
#define _HCN352_H_


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include "hni31.h"
#include "hcn35.h"

// -----------------------------------------------------------------------------
// class tcn352_NetWorkerConnection
// -----------------------------------------------------------------------------

class tcn352_NetWorkerConnector:public tcn35_BackupToolConnector
{
  public:
    tcn352_NetWorkerConnector(tcn35_ActionType ActionType, const tcn35_SaveInfo & rInfo);
    ~tcn352_NetWorkerConnector();
    
    virtual const char * GiveToolName();
    virtual const char * GiveLongToolName();

  private:
    int CheckToolInstallation(int & IsInstalled);
    int Prepare();

    int CleanUpTool();
    int CollectToolSaveIDs();
    int CheckMedium();

    int GiveInformationOfLine(const char *   Line,
                              const char * & EBIDInLine,
                              size_t       & EBIDLength,
                              const char * & TypeString,
                              const char * & DateTimeString,
                              int          & LineContainsEBID);

    int BuildEBIDFromNSavetime(char       * & EBID,
                               const char *   StartOfNSavetime,
                               const char *   EndOfNSavetime);
    int ToolWasSuccessful();
    int AnalyzeNSROutputLine(const char * Line, int & FoundCompletionMessage);
    int FoundFileNameFollowedByWarning(const char * FileName,
                                       const char * NSROutputLine);
    int FoundCompletionMessageWithCorrectSize(const char * Line);

    void DeleteListOfErrorsToIgnore();

    char * PathOfNSR_ENVFile;
    char * PathOfNSR_HOMEDirectory;
    char * NameOfNSR_HOST;
    char * NameOfNSR_POOL;
    char * NSR_EXPIREValue;

    char * PathOfNSRSave;
    char * PathOfNSRRecover;
    char * PathOfNSRMminfo;

    int    BuildBackupToolCall();   // TODO: move to base-class 

    long   StartSaveTime;
    long   EndSaveTime;

    char * DirectiveFileName;

    enum LevelFullType {DataMigration, DataRecovery, DataMigrationAndDataRecovery};    
    LevelFullType LevelFullForBackupType;

    enum VerboseFlagForType {None, Save, Recover, SaveAndRecover};
    VerboseFlagForType VerboseFlagFor;

    enum NSREBIDTypeEnum {SSID, NSaveTime};
    NSREBIDTypeEnum     NSREBIDType;
    char *              TempEBID;
    static const char * NSaveTimeEBIDPrefix;
    static const char * NSaveTimeEBIDPartDelimiter;

    enum NSROuputResultType {NotChecked, FoundError, FoundNoError};
    NSROuputResultType NSROuputResult;
    int                UseNSROutput;
    teo00_Uint8        BDBackupSize;
    ni31_List<char *>  ListOfErrorsToIgnore;
};

#endif
