/*!*****************************************************************************
  module:       hcn353.h
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  connection between dbmsrv and Backint for SAP DB conforming
                backup tools
  last change:  2000-11-29 10:37
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 2000


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#if !defined (_HCN353_H_)
#define _HCN353_H_


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include "hni31.h"
#include "hcn35.h"


// -----------------------------------------------------------------------------
// class tcn353_Backint2Connection
// -----------------------------------------------------------------------------

class tcn353_Backint2InputLine;

const int NumberOfBackupHistories_cn353=2;  // ==2, use the backup history file and the external backup history file

class tcn353_Backint2Connector:public tcn35_BackupToolConnector
{
  public:
    tcn353_Backint2Connector(tcn35_ActionType ActionType, const tcn35_SaveInfo & rInfo);
    ~tcn353_Backint2Connector();

    virtual const char * GiveToolName();
    virtual const char * GiveLongToolName();

  private:
    int CheckToolInstallation(int & IsInstalled);

    int Prepare();
    int MaxWaitingTimeOfCurrentWaitingPhase();

    int DetermineAndCountInputEntries(int & NumberOfEntries);
    int RenameABackupHistoryFile(const char * NameOfBackupHistoryFile,
                                 int          NumberOfBackupHistoryFile);

    int BringHistoryFilesToFormerState();

    int AppendToBackintInputList(const char * TheName,
                                 int          IsAPipe,
                                 const char * BID=0,
                                 const char * Destination=0,
                                 const char * EndOfDestination=0,
                                 const char * Type=0,
                                 const char * DateTime=0);
    void DeleteBackintInputList();

    int PostOperationsAfterSuccess();
    int CleanUpTool();
    int CollectToolSaveIDs();
    
    int BuildAEBIDFrom(char * &     EBID,
                       const char * User,
                       const char * BID,
                       const char * Path);

    int PrepareToolOutputOfAvailableBackups();
    int GiveNextAvailableBackup(tcn003_BackupHistoryEBIDStatus & Status,
                            const char * & EBID,
                            size_t       & EBIDLength,
                            const char * & TypeString,
                            const char * & DateTimeString,
                            int          & ReachedEnd);
    int CloseToolOutputOfAvailableBackups();

    int GiveInformationOfLine(const char *   Line,
                              const char * & EBIDInLine,
                              size_t       & EBIDLength,
                              const char * & TypeString,
                              const char * & DateTimeString,
                              int          & LineContainsEBID);

    int BuildBackupToolCall();  // TODO: move to base-class

    int GetBIDAndNameFromLine(const char *   Line,
                              int          & ContainsABID,
                              const char * & BackupID,
                              size_t       & LengthOfBackupID,
                              const char * & Name,
                              size_t       & LengthOfName);

    int ToolWasSuccessful();

    int BackupOrRestoreHistoryWithBackint();

    char * PathToEnvFile;

    char * BackintPath;
    char * BackintInFile;
    char * BackintOutFile;
    char * BackintErrorFile;
    char * BackintParamFile;

    int    NumberOfRequestedNames;


    ni31_List<tcn353_Backint2InputLine *> ListOfFilesProcessedByBackint;

    int BackintTimeOutSuccess;
    int BackintTimeOutFailure;

    enum BackintDataTypeEnum{DataPipes, HistoryFiles};

    BackintDataTypeEnum BackintDataType;
    
    const int NumberOfBackupHistories; // ==2, use the backup history file and the external backup history file
    enum TypeBackupHistoryState{Unknown, NotExist, Exist, Recovered, Renamed, Moved};
    
    const char *           SuffixBackupHistory[NumberOfBackupHistories_cn353];         //TODO: new class containing all information abut one history file
    const char *           SuffixOfCopyOfBackupHistory[NumberOfBackupHistories_cn353];
    char *                 OriginalBackupHistoryName[NumberOfBackupHistories_cn353];
    char *                 OriginalBackupHistoryMovedTo[NumberOfBackupHistories_cn353];
    char *                 PathOfMigrationBackintHistory[NumberOfBackupHistories_cn353];
    TypeBackupHistoryState BackupHistoryState[NumberOfBackupHistories_cn353];

    const ni31_ListElem<tcn353_Backint2InputLine *> *InFileItem;
    char *                                           TempBID;
};

#endif
