/*******************************************************************************
  module:       hcn36.h
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  basic functions used in vcn35.cpp, ..., vcn355.cpp
  last change:  2000-11-29 10:50
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#if !defined(_HCN36_H_)
#define _HCN36_H_


// -----------------------------------------------------------------------------
// includes
// -----------------------------------------------------------------------------

#include "hni00b.h"

#include<string.h>
#if defined (WIN32)
#include<windows.h>
#elif defined (UNIX)
#include<ctype.h>
#else
#error hcn36.h only coded for WIN32 and UNIX (define one of them).
#endif


#include "hni33.h"
#include "hni34.h"
#include "hni35.h"
#include "hcn361.h"
#include "hcn90.h"


// -----------------------------------------------------------------------------
// constants, types and function declarations
// -----------------------------------------------------------------------------

int  cn36_StrAlloc(char * & Dest, size_t Length);
int  cn36_StrAllocCpy(char * & Dest, const char * Src);
void cn36_StrNCpy(char * Dest, const char * Src, size_t Length);
void cn36_StrNCat(char * Dest, const char * Src, size_t Length);
int  cn36_StrNAllocCpy(char * & Dest, const char * Src, size_t Length);
void cn36_StrDealloc(char * & String);

// see hcn361.h: void cn36_StrUpr(char * String);

// see hcn361.h: int cn36_StrToInt(int & TheInt, const char * Buf);
// see hcn361.h: int cn36_StrNToInt(int & TheInt, const char * Buf, size_t Length);

// see hcn361.h: int cn36_StrNToLong(long & TheLong, const char * Buf, size_t Length);

// see hcn361.h: int cn36_StrToInt8(teo00_Int8 & TheInt8, const char * Buf);
// see hcn361.h: int cn36_StrNToInt8(teo00_Int8 & TheInt8, const char * Buf, size_t Length);
// see hcn361.h: int cn36_UInt8ToStrN(char *Buf, size_t Size, teo00_Uint8 i);

// see hcn361.h: int cn36_StrNConsistsOfDigits(const char *String, size_t Length);
// see hcn361.h: int cn36_StrConsistsOfDigits(const char *String);


// see hcn361.h: const teo00_Int8    MAX_INT8_CN36=teo00_Int8(9223372036854775807);      //==2^63-1
// see hcn361.h: const teo00_Uint8   MAX_UINT8_CN36=teo00_Uint8(18446744073709551615);   //==2^64-1

// see hcn361.h:       char * cn36_PositionOfTerminatingZero(      char * s);
// see hcn361.h: const char * cn36_PositionOfTerminatingZero(const char * s);

// see hcn361.h: const char * cn36_FirstNonWhiteSpaceOf(const char * s);
// see hcn361.h:       char * cn36_FirstNonWhiteSpaceOf(      char * s);
// see hcn361.h: const char * cn36_FirstWhiteSpaceOf(const char * s);
// see hcn361.h:       char * cn36_FirstWhiteSpaceOf(      char * s);

// see hcn361.h: void cn36_GetStringBetweenDoubleQuotesOrWhiteSpaces(const char * & Start, const char * & End);

// see hcn361.h: const char * cn36_StrChr(const char *String, char Character);
// see hcn361.h:       char * cn36_StrChr(      char *String, char Character);
// see hcn361.h: const char * cn36_LoopedStrChr(const char *String, char Character, int Number);
// see hcn361.h:       char * cn36_LoopedStrChr(      char *String, char Character, int Number);

// see hcn361.h: int cn36_StrStrNCmp(const char * String1, const char *String2, size_t Length2);
// see hcn361.h: int cn36_StrNUprCmp(const char *String1, const char *String2, size_t Length);
// see hcn361.h: int cn36_StrStrNUprCmp(const char * String1, const char *String2, size_t Length2);

int cn36_DetermineValueOfEnvironmentVariable(const char * const VariableName,
                                                   char * &     VariableValue,
                                             const char *       VariableComment,
                                             const char *       DBName,
                                                   int          ExportDefaultIfNotFound=1);
int cn36_PutEnvironmentVariable(const char * VariableName,
                                const char * VariableValue);


int cn36_ConstructCompletePath(char *& CompletePath, const char * Directory, const char * FileName);

enum tcn36_VerboseTyp{UsePrefix, NoPrefix, AddALevelToPrefix, SubALevelFromPrefix};

void cn36_Verbose(const char * pszString, tcn36_VerboseTyp PrefixModifier=UsePrefix);
void cn36_Verbose(tcn36_VerboseTyp PrefixModifier);

extern char MsgBuf_cn36[1024];


#define MaxLengthOfProtFileName_cn36 255
extern char PathOfProtFile_cn36[MaxLengthOfProtFileName_cn36+1];

int cn36_AppendFileToFile(const char * Src, const char *Dest);
int cn36_MoveFile(const char * Src, const char *Dest);
int cn36_Remove(const char *Name);

const char *cn36_GetHostName();

// see hcn361.h: const char * cn36_BaseNameFrom(const char *Path);
// see hcn361.h:         char * cn36_BaseNameFrom(      char *Path);

// see hcn361.h: const char * cn36_EndOfDirectory(const char *Path);
// see hcn361.h:       char * cn36_EndOfDirectory(      char *Path);

int cn36_CheckExist(const char * FileName);
int cn36_CheckRead(const char * pszFileName);
int cn36_CheckDir(const char * pszDirectoryName);
int cn36_CheckExe(const char * pszProgramName);

void cn36_Sleep(int Seconds);
long cn36_GetCurrentTime();
long cn36_SecondsBetween(long time1, long time2);

class tcn36_ADBMSrvError: public tni35_AError
{
  public:
    enum DBMSrvErrorCase
    {
        CaseOutOfMemory,

        CaseCanNotExportEnvironment,
        CaseCanNotGetComputerName,
        CaseCanNotGetBlockingFactor,
        CaseCanNotGetPageSize,

        CaseADirectoryIsMissing,
        CaseAFileIsMissingOrUnreadable,
        CaseAFileIsNotExecutable,
        CaseAFileAlreadyExists,

        CaseToManyPipes,
        CaseCouldNotCreateDataPipe,
        CaseCanNotCreateADSMInputPipe,  //TODO: ADSM specific errors should be declared in separate error class
        CaseCanNotDetermineNSR_HOME,    //TODO: NetWorker specific errors should be declared in separate error class
        CaseCanNotDetermineNSR_HOST,
        CaseUnknowFormatOfMminfoOutput,
        CasePipeOrFileNameToLong,       //TODO: Backint for SAP DB specific errors should be declared in separate error class
        CaseUnspecifiedBackintPath,
        CaseTwoDifferentBIDs,
        CaseWrongBIDReturned,
        CaseWrongFilesProcessed,
        CaseCouldNotSaveBackupHistoryFiles,
        CaseCouldNotRestoreBackupHistoryFiles,
        CaseUnknownBackintOutputFormat,
        CaseSyntaxErrorEBIDs,
        CaseEBIDsOfDifferentDBs,
        CaseTooManyEBIDs,
        CaseTooFewEBIDs,
        CaseCanNotBuildSaveLabel,
        CaseCanNotSignalEndToTool,

        CaseNoBackupIsRunning,
        CaseBackupToolFailed,
        CaseBackupToolAndDBFailed,
        CaseBackupToolKilled,
        CaseBackupToolFailedAndDBKernelRequestCanceled,
        CaseBackupToolKilledAndDBKernelRequestCanceled,
        CaseDBKernelRequestCanceled,
        CaseCancelDBKernelRequestFailed,
        CaseBackupToolFailedAndCancelDBKernelRequestFailed,
        CaseBackupToolKilledAndCancelDBKernelRequestFailed,
        
        CaseSQLError,
        CaseCN00Error,

        CaseNoMediumFound,
        CaseMediumHasNoName,

        CaseUnknownFormatOfDBMKnlFile,
        CaseUnknownFormatOfDBMEbfFile,
        CaseCouldNotGetNameOfToolLogFile,
        CaseCouldNotGetSaveIDs,
        CaseCouldNotWriteToToolLogFile,
        CaseNoLogPageInformation,
        CaseCouldNotGetNameOfBackupHistory,
        CaseCouldNotGetNameOfToolProtFile,
        CaseCouldNotMoveToolOutputToToolProtFile,

        CaseNeedExternalSaveIDForRestore,
        
        CaseInternalError
    };

    enum DBMSrvErrorSetFlag {DoNotSetErrorNow, SetErrorNow};

    tcn36_ADBMSrvError(DBMSrvErrorCase ErrCase, DBMSrvErrorSetFlag SetFlag=SetErrorNow);
    tcn36_ADBMSrvError(DBMSrvErrorCase ErrCase, const char *ErrInfo );
    tcn36_ADBMSrvError(DBMSrvErrorCase ErrCase, int ErrInfo);

    //using tni35_AError::Set; <- 'using' is not supported by all compilers!
    //tni35_AError & Set();
    //tni35_AError & Set(const char *ErrInfo);

    tcn36_ADBMSrvError & SetI(int ErrInfo);
    tcn36_ADBMSrvError & SetSQL(int TheSQLCode, const char * TheSQLErrText);
    tcn36_ADBMSrvError & SetSQLI(int TheSQLCode, int ErrInfo);

    virtual void FillErrorMsgBuffer() const;
    
    static tcn36_ADBMSrvError OutOfMemory;

    static tcn36_ADBMSrvError CanNotExportEnvironment;
    static tcn36_ADBMSrvError CanNotGetComputerName;
    static tcn36_ADBMSrvError CanNotGetBlockingFactor;
    static tcn36_ADBMSrvError CanNotGetPageSize;
    static tcn36_ADBMSrvError ADirectoryIsMissing;
    static tcn36_ADBMSrvError AFileIsMissingOrUnreadable;
    static tcn36_ADBMSrvError AFileIsNotExecutable;
    static tcn36_ADBMSrvError AFileAlreadyExists;

    static tcn36_ADBMSrvError CouldNotCreateDataPipe;
    static tcn36_ADBMSrvError CanNotCreateADSMInputPipe;
    static tcn36_ADBMSrvError UnknownFormatOfMminfoOutput;
    static tcn36_ADBMSrvError UnspecifiedBackintPath;
    static tcn36_ADBMSrvError TwoDifferentBIDs;
    static tcn36_ADBMSrvError WrongBIDReturned;
    static tcn36_ADBMSrvError WrongFilesProcessed;
    static tcn36_ADBMSrvError CouldNotSaveBackupHistoryFiles;
    static tcn36_ADBMSrvError CouldNotRestoreBackupHistoryFiles;
    static tcn36_ADBMSrvError UnknownBackintOutputFormat;

    static tcn36_ADBMSrvError PipeOrFileNameToLong;
    static tcn36_ADBMSrvError CanNotBuildSaveLabel;
    static tcn36_ADBMSrvError CanNotSignalEndToTool;

    static tcn36_ADBMSrvError NoBackupIsRunning;
    static tcn36_ADBMSrvError BackupToolFailed;
    static tcn36_ADBMSrvError BackupToolAndDBFailed;
    static tcn36_ADBMSrvError BackupToolKilled;
    static tcn36_ADBMSrvError BackupToolFailedAndDBKernelRequestCanceled;
    static tcn36_ADBMSrvError BackupToolKilledAndDBKernelRequestCanceled;
    static tcn36_ADBMSrvError DBKernelRequestCanceled;
    static tcn36_ADBMSrvError CancelDBKernelRequestFailed;
    static tcn36_ADBMSrvError BackupToolFailedAndCancelDBKernelRequestFailed;
    static tcn36_ADBMSrvError BackupToolKilledAndCancelDBKernelRequestFailed;

    static tcn36_ADBMSrvError SQLError;
    static tcn36_ADBMSrvError CN00Error;

    static tcn36_ADBMSrvError NoMediumFound;
    static tcn36_ADBMSrvError MediumHasNoName;
    
    static tcn36_ADBMSrvError UnknownFormatOfDBMKnlFile;
    static tcn36_ADBMSrvError CouldNotGetNameOfToolLogFile;
    static tcn36_ADBMSrvError CouldNotGetSaveIDs;
    static tcn36_ADBMSrvError CouldNotWriteToToolLogFile;
    static tcn36_ADBMSrvError CouldNotGetNameOfBackupHistory;
    static tcn36_ADBMSrvError CouldNotGetNameOfToolProtFile;
    static tcn36_ADBMSrvError CouldNotMoveToolOutputToToolProtFile;

    static tcn36_ADBMSrvError NeedExternalSaveIDForRestore;

    static tcn36_ADBMSrvError InternalError;

    enum DBMSrvErrorInfoType {None, AString, AIntCode, AIntAndASQLCode, ASQLCodeAndAInt, ASQLCodeAndAString};

    struct DBMSrvErrorDescription // must be public for initialization of static member DBMSrvErrorToText
    {
        DBMSrvErrorCase Case;
        DBMSrvErrorInfoType InfoType;
        const char *    PlainErrorText;
        const char *    ExtendedErrorText;
    };

  private:
    static DBMSrvErrorDescription DBMSrvErrorToText[];
    const DBMSrvErrorDescription * ThisError;
    void SetThisError(DBMSrvErrorCase CaseToSet);

    int SQLCode;
    int IntCode;
};

class tcn36_DBMSrvErrorHandler:public tni35_AErrorHandler
{
  public:
    tcn36_DBMSrvErrorHandler();

    void         SetReplyArrayTo(char * pcReplyData, int * piReplyLen);
    virtual void Report(const tni35_AError & Err);
    tcn00_Error  GiveLastError();
    void         SetPhaseTo(tcn00_Error Phase);

  private:
    char      * ReplyData;
    int       * ReplyLen;
    tcn00_Error CurrentErrorPhase;
    tcn00_Error LastError;
    int         ReplyAlreadyConstructed;
}

extern TheDBMSrvErrorHandler_cn36;

#endif
