/*******************************************************************************
  module:       hcn361.h
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  string handling functions used in vcn35.cpp, ..., vcn355.cpp,
                vcn37.cpp
  last change:  2000-11-29 11:22
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 1999


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#if !defined(_HCN361_H_)
#define _HCN361_H_

#include"geo00.h"

void cn36_StrNCpy(char * Dest, const char *Src, size_t Length);
void cn36_StrNCat(char * Dest, const char *Src, size_t Length);

void cn36_StrUpr(char *String);

int cn36_StrToInt(int & TheInt, const char *Buf);
int cn36_StrNToInt(int & TheInt, const char *Buf, size_t Length);

int cn36_StrNToLong(long & TheLong, const char *Buf, size_t Length);

int cn36_StrToInt8(teo00_Int8 & TheInt8, const char *Buf);
int cn36_StrNToInt8(teo00_Int8 & TheInt8, const char *Buf, size_t Length);

int cn36_StrToUInt8(teo00_Uint8 & TheUInt8, const char * Buf);
int cn36_StrNToUInt8(teo00_Uint8 & TheUInt8, const char * Buf, size_t Length);
int cn36_UInt8ToStrN(char *Buf, size_t Size, teo00_Uint8 i);

int cn36_StrNConsistsOfDigits(const char *String, size_t Length);
int cn36_StrConsistsOfDigits(const char *String);


#if defined (WIN32)
const teo00_Int8    MAX_INT8_CN36=teo00_Int8(9223372036854775807);      //==2^63-1
const teo00_Uint8   MAX_UINT8_CN36=teo00_Uint8(18446744073709551615);   //==2^64-1
#elif defined (BIT64)
const teo00_Int8    MAX_INT8_CN36=teo00_Int8(9223372036854775807L);     //==2^63-1
const teo00_Uint8   MAX_UINT8_CN36=teo00_Uint8(18446744073709551615UL); //==2^64-1
#else
const teo00_Int8    MAX_INT8_CN36=teo00_Int8(9223372036854775807LL);    //==2^63-1
const teo00_Uint8   MAX_UINT8_CN36=teo00_Uint8(18446744073709551615ULL);//==2^64-1
#endif

      char * cn36_PositionOfTerminatingZero(      char * s);
const char * cn36_PositionOfTerminatingZero(const char * s);

const char * cn36_FirstNonWhiteSpaceOf(const char * s);
      char * cn36_FirstNonWhiteSpaceOf(      char * s);
const char * cn36_FirstWhiteSpaceOf(const char * s);
      char * cn36_FirstWhiteSpaceOf(      char * s);

void cn36_GetStringBetweenDoubleQuotesOrWhiteSpaces(const char * & Start, const char * & End);

const char * cn36_StrChr(const char *String, char Character);
      char * cn36_StrChr(      char *String, char Character);
const char * cn36_LoopedStrChr(const char *String, char Character, int Number);
      char * cn36_LoopedStrChr(      char *String, char Character, int Number);

int cn36_StrStrNCmp(const char * String1, const char *String2, size_t Length2);
int cn36_StrNUprCmp(const char *String1, const char *String2, size_t Length);
int cn36_StrStrNUprCmp(const char * String1, const char *String2, size_t Length2);

const char * cn36_BaseNameFrom(const char *Path);
      char * cn36_BaseNameFrom(      char *Path);

const char * cn36_EndOfDirectory(const char *Path);
      char * cn36_EndOfDirectory(      char *Path);


void cn361_Sleep(int Seconds);

#endif
