/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#if defined(CN72) && defined(_WIN32)
  #line 3 "d:\\v72\\sys\\src\\cn\\vcn14.c"
#endif 

/*! 
  -----------------------------------------------------------------------------
 
  module: vcn14.c
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV

  special area: Database Manager-API
 
  description:  Implementation of the API for Database Manager Client Applications
                
  version:      7.2.

  -----------------------------------------------------------------------------
 
  copyright:    SAP AG, 1998
 
  -----------------------------------------------------------------------------
*/

/* c-only module!!! */
#if defined (__cplusplus)
  #error "Don't compile this module as C++"
#endif

/*
  -----------------------------------------------------------------------------
  private includes
  -----------------------------------------------------------------------------
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>

#include "heo03.h"
#include "vsp004.h"
#include "gsp09.h"

#include "hcn14.h"

#define MOD__ "vcn14.c: "
#define MF__ "<None>"

/*
  -----------------------------------------------------------------------------
  missing extern prototypes
  -----------------------------------------------------------------------------
*/
#ifdef CN71
  #define tsp00_Bool bool
#endif
  
externC void sqlxucloseuser (
    tsp_errtext         errtext,
    tsp00_Bool*             ok);

externC void sqlxuopenuser (
    tsp_errtext         errtext,
    tsp00_Bool*             ok);

externC void sqlputuser (
    tsp4_xuser_record*      user_params,
    tsp_errtext         errtext,            
    tsp00_Bool*         ok);

externC void sqlgetuser (
    tsp4_xuser_record*      user_params,
    tsp_errtext             errtext,    
    tsp00_Bool*             ok);

externC void sqlclearuser ();


/*
  -----------------------------------------------------------------------------
  private deklarations
  -----------------------------------------------------------------------------
*/
typedef struct ControlSessionT {
    tsp_int4 reference;
    tsp_int4 packetSize;
    void*    packetData;
    char*    packetOutData; /* used for reading */
    char*    packetInData;  /* used for writing */
    tsp_int4 packetPos;
    tsp_int4 packetLen;
} ControlSessionT;

#define C_COMPONENT    ((_TCHAR *) _T("DBM-Api"))
#define XU_CRYPT_BASE  16  
#define XU_MAX_USERS   32
#define XU_DBM_MARK    ((_TCHAR *) _T("1"))
#define CTRL_ERR       ((_TCHAR *) _T("ERR"))
#define RPM_ERR_HEAD   ((_TCHAR *) _T("ERR\n"))
#define XU_SAP_USER    ((_TCHAR *) _T("SAPR3"))
#define XU_SAP_PWD     ((_TCHAR *) _T("SAP"))
#define XU_KEY_DEFAULT ((_TCHAR *) _T("DEFAULT"))

#define BUFFER_LEN  1024
#define DBM_USERLOGON   _T("user_logon")
#define DBM_SYSTEMLOGON _T("user_system")

/*
  -----------------------------------------------------------------------------
  private prototypes
  -----------------------------------------------------------------------------
*/

static tsp_int4 cn14_dbmLogon (
    void                    * sessionParm,
    const tsp_c64c            user,     
    tsp00_ErrTextc            errtext,
    const _TCHAR            * szCommand );

static tsp_int4 cn14_cmdWrite (
    void       * sessionParm,
    const void * data,
    tsp_int4     len);

static tsp_int4 cn14_request (
    void            * sessionParm,
    tsp00_ErrTextc    errtextC );

static tsp_int4 cn14_receive (
    void            * sessionParm,
    tsp00_ErrTextc    errtextC );

static void cn14_errtextToC (
    tsp00_ErrTextc errtextC,
    tsp00_ErrText  errtextP );

static tsp_int4 cn14_setErrtext (
    tsp00_ErrTextc errtextC,
    tsp_int4       nRc );

static tsp_int4 cn14_senduser (
    void                * pSession,
    const tsp00_NodeIdc   szServerNode,
    const tsp00_DbNamec   szDBName);

/*
  -----------------------------------------------------------------------------
  public function cn14connectDBM 
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14connectDBM"
tsp_int4 WINAPI_CN14 
cn14connectDBM (
    const tsp00_NodeIdc       servernode,
    const tsp00_DbNamec       dbname,
    const tsp00_VFilenamec    dbroot,
    void                   ** sessionOut,
    tsp00_ErrTextc            errtext )
{
  tsp_int4         nRc  = DBMAPI_OK_CN14;

  nRc = cn14connect (servernode, dbname, dbroot, DEFAULT_DBMPGM_CN14, sessionOut, errtext);
/* no fallback to cserver anymore 
  if (nRc != DBMAPI_OK_CN14) {
     nRc = cn14connect (servernode, dbname, dbroot, DEFAULT_DBMOLD_CN14, sessionOut, errtext);
  } */ /* end if */
  if (nRc == DBMAPI_OK_CN14 && _tcslen(dbname) > 0) {
    /* connection etablished then search and send user data */
    nRc = cn14_senduser(*sessionOut, servernode, dbname);
    cn14_setErrtext(errtext, nRc);
    if (nRc != DBMAPI_OK_CN14) {
      cn14release (sessionOut);
    } /* end if */
  } /* end if */

  return nRc;
} /* end cn14connectDBM */

/*
  -----------------------------------------------------------------------------
  public function cn14connectDBMUsr 
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14connectDBMUsr"
tsp_int4 WINAPI_CN14 
cn14connectDBMUsr (
    const tsp00_NodeIdc       servernode,
    const tsp00_DbNamec       dbname,
    const tsp00_VFilenamec    dbroot,
    const tsp_c64c            user,     
    void                   ** sessionOut,
    tsp00_ErrTextc            errtext )
{
  tsp_int4         nRc  = DBMAPI_OK_CN14;

  nRc = cn14connect (servernode, dbname, dbroot, DEFAULT_DBMPGM_CN14, sessionOut, errtext);

  if (nRc != DBMAPI_OK_CN14) {
     nRc = cn14connect (servernode, dbname, dbroot, DEFAULT_DBMOLD_CN14, sessionOut, errtext);

  } /* end if */
  if (nRc == DBMAPI_OK_CN14 && _tcslen(user) > 0) {
    nRc = cn14_dbmLogon(*sessionOut, user, errtext, (_tcslen(dbname) == 0) ? DBM_SYSTEMLOGON : DBM_USERLOGON );
    if (nRc != DBMAPI_OK_CN14) {
      cn14release (sessionOut);
    } /* end if */
  } /* end if */

  return nRc;
} /* end cn14connectDBMUsr */

/*
  -----------------------------------------------------------------------------
  public function cn14connectDBMTp
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14connectDBMTp"
tsp_int4 WINAPI_CN14 
cn14connectDBMTp (
    const tsp00_NodeIdc       servernode,
    const tsp00_DbNamec       dbname,
    const tsp00_VFilenamec    dbroot,
    const tsp00_C8c           sysid,     
    const tsp00_C30c          conn,     
    const tsp00_Pathc         profile,     
    void                   ** sessionOut,
    tsp00_ErrTextc            errtext )
{
  tsp_int4         nRc   = DBMAPI_OK_CN14;
  tsp_c64c         user  = "";
  _TCHAR           szCommand[BUFFER_LEN];
  _TCHAR           szUserLine[BUFFER_LEN] = "";
  _TCHAR           szLine[BUFFER_LEN] = "";
  FILE           * hFile = NULL;
  tsp00_Int4       nExitCode;
  bool             bOk;
  int              nLine = 0;
  char             szDummy[BUFFER_LEN];
  char         *   szTmpName   = NULL;
  char         *   szTmpDir    = NULL;
  char             szTmpFile[BUFFER_LEN] = "";

/*  FILE           * hTrace = fopen("dbmrfc1.log", "w"); */

  szTmpName = tmpnam(szDummy);
  if (szTmpName == NULL) {
    strcpy(szUserLine, "can not create temporary file");
  } else {

#ifdef _WIN32
    if ((szTmpDir = getenv("TMP")) == NULL) {
      sprintf(szTmpFile, "%s%s%s", ".", szTmpName, "cst");
    } else {
      char  szShortTmp[BUFFER_LEN] = "";
      GetShortPathName(szTmpDir, szShortTmp, BUFFER_LEN);
      sprintf(szTmpFile, "%s%s.%s", szShortTmp,  szTmpName, "cst");
    } /* end if */
#else
    strcpy(szTmpFile, szTmpName);
#endif
/*    if (hTrace != NULL) fprintf(hTrace, "tmpfile is %s (%s %d)\n", szTmpFile, __FILE__, __LINE__); */

    sprintf(szCommand, "tp pf=%s getlcconnectinfo %s lc-name=%s user-type=DBM 1>%s 2>&1", profile, sysid, conn, szTmpFile);

/*    if (hTrace != NULL) fprintf(hTrace, "command is %s (%s %d)\n", szCommand, __FILE__, __LINE__); */

    if (system(szCommand) != -1) {
      hFile = fopen(szTmpFile, "r");

      if (hFile != NULL) {
        while (fgets ( szLine, BUFFER_LEN - 1, hFile ) != NULL) {
/*          if (hTrace != NULL) fprintf(hTrace, "szLine is %s (%s %d)\n", szLine, __FILE__, __LINE__); */
          nLine++;
          if (nLine == 2) {
            strcpy(szUserLine, szLine);
            bOk = (strncmp(szUserLine, "LC-INFO=", strlen("LC-INFO=")) == 0);
          } /* end if */
        } /* end while */

        fclose(hFile);
        remove(szTmpFile);
      } else {
/*        if (hTrace != NULL) fprintf(hTrace, "errno is %d (%s %d)\n", errno, __FILE__, __LINE__); */
        strcpy(szUserLine, "can not open temporary file");
      } /* end if */
    } else {
      strcpy(szUserLine, "can not execute tp");
    } /* end if */
     remove(szTmpFile);
  } /* end if */

  if (bOk) {
    _TCHAR * pUser;
    pUser = strchr(szUserLine, '=');
    if (pUser != NULL) {
      ++pUser;
      /* delete newline */
      pUser[_tcslen ( pUser ) - 1] = 0; 

      _tcscpy(user, pUser);
    } /* end if */

    nRc = cn14connectDBMUsr (servernode, dbname, dbroot, user, sessionOut, errtext );
  } else {
    nRc = cn14_setErrtext(errtext, DBMAPI_R3TP_CN14);
    if (strlen(szUserLine) > 0) {
      szUserLine[sizeof(tsp00_ErrTextc)  - strlen(errtext) -1] = 0;
      strcat(errtext, szUserLine);
    } else {
      szLine[sizeof(tsp00_ErrTextc)  - strlen(errtext) -1] = 0;
      strcat(errtext, szLine);
    } /* end if */
  } /* end if */
  
/*  if (hTrace != NULL) fclose(hTrace); */

  return nRc;
} /* end cn14connectDBMTp */

  /*
  -----------------------------------------------------------------------------
  public function cn14connectRPM
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14connectRPM"
tsp_int4 WINAPI_CN14 
cn14connectRPM (
    const tsp00_NodeIdc       servernode,
    const tsp00_DbNamec       dbname,
    const tsp00_VFilenamec    dbroot,
    char                   ** ppszData,
    void                   ** sessionOut,
    tsp00_ErrTextc            errtext )
{

    tsp_int4  rc = DBMAPI_OK_CN14;

    rc = cn14connect (servernode, dbname, dbroot, DEFAULT_RPMPGM_CN14, sessionOut, errtext);
    if (DBMAPI_OK_CN14 == rc)
    {
        tsp_int4 rc2 = DBMAPI_OK_CN14;
        tsp_int4 lDataLen;

#ifndef _WIN32 
        const char *myUser;
        char       *cmdStr;
        tsp_int4 len = 0;

        myUser = sqlxusername();       /* hole den usernamen */

        cmdStr = (char *)malloc(strlen("set logowner ") + strlen(myUser) + 1);
        sprintf(cmdStr, "set logowner %s", (char*)myUser);
        len = strlen(cmdStr);
        rc2 = cn14cmdExecute (*sessionOut, cmdStr, len, NULL, &lDataLen, errtext);
        free(cmdStr);
#endif

        /* Wenn NULL mitgegeben wurde, dann nicht auswerten */
        if (NULL != ppszData)
        {
            rc2 = cn14cmdExecute (*sessionOut, "set", 3, NULL, &lDataLen, errtext);
            if (DBMAPI_OK_CN14 == rc2)
            {
                tsp_int4 lErrCode;
                tsp_int4 lSqlCode;
                tsp_int4 lSqlErrLen;
                cn14analyzeRpmAnswer(*sessionOut, &lErrCode, (const char **)ppszData, &lDataLen,
                                                  &lSqlCode, NULL,     &lSqlErrLen);
            }
        }
    }
    
    return rc;
} /* end cn14connectRPM */

/*
  -----------------------------------------------------------------------------
  public function cn14connect
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14connect"
tsp_int4 WINAPI_CN14 
cn14connect (
    const tsp00_NodeIdc       servernode,
    const tsp00_DbNamec       dbname,
    const tsp00_VFilenamec    dbroot,
    const tsp00_VFilenamec    serverpgm,
    void                   ** sessionOut,
    tsp00_ErrTextc            errtextC )
{
  tsp_int4              nRc          = DBMAPI_OK_CN14;
  static tsp_compname   szComponent;
  tsp01_CommErr         commErr;
  ControlSessionT     * session;

  /* init rte with pascal string */
  memset(szComponent, _T(' '), sizeof(szComponent));
  _tcsncpy((_TCHAR *) szComponent, C_COMPONENT, _tcslen(C_COMPONENT));
  sqlinit (szComponent, NULL);

  /* alloc memory for session structure */
  session = (ControlSessionT *) malloc (sizeof (ControlSessionT));
  *sessionOut = session;

  if (session == NULL) {
    nRc = cn14_setErrtext(errtextC, DBMAPI_MEMORY_CN14);
  } else {

    session->packetPos = 0;
    session->packetInData = NULL;
    session->packetOutData = NULL;
    sqlx2connectc (0, 
                   (_TCHAR *) servernode, 
                   (_TCHAR *) dbname, 
                   (_TCHAR *) dbroot, 
                   (_TCHAR *) serverpgm,
                   &session->reference, 
                   &session->packetSize, 
                   &session->packetData,
                   errtextC, 
                   &commErr);

    if (commErr != commErrOk_esp01) {
      nRc = DBMAPI_COMMERR_CN14;
      free (session);
      *sessionOut = NULL;
    } /* end if */

  } /* end if */

  return nRc;
} /* end cn14connect */

/*
  -----------------------------------------------------------------------------
  public function cn14release
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14release"
void WINAPI_CN14 
cn14release (
    void** sessionParm)
{
  ControlSessionT* session = (ControlSessionT*)*sessionParm;
  tsp_int4         nDummy;
  tsp00_ErrTextc   errtext;

  if (session != NULL) {
    cn14cmdExecute(session, _T("release\0"), _tcslen(_T("release")) + 1, NULL, &nDummy, errtext);

    sqlarelease (session->reference);
    free (session);
    *sessionParm = NULL;
  } /* end if */

} /* end cn14release */

/*
  -----------------------------------------------------------------------------
  public function cn14cmdCancel
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14cmdCancel"
void WINAPI_CN14 
cn14cmdCancel (
    void* sessionParm)
{
  ControlSessionT* session = (ControlSessionT*)sessionParm;

  if (session != NULL) {
    sqlacancel (session->reference);
  } /* end if */

} /* cn14cmdCancel */

/*
  -----------------------------------------------------------------------------
  public function cn14replyAvailable
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14replyAvailable"
tsp_int4 WINAPI_CN14 
cn14replyAvailable (
    void           * sessionParm,
    tsp00_ErrTextc   errtextC )
{
  tsp_int4          nRc     = DBMAPI_OK_CN14;
  ControlSessionT * session = (ControlSessionT*)sessionParm;
  tsp01_CommErr     commErr;
  tsp00_ErrText     errtextP;

  if ((session == NULL) || (session->packetInData != NULL)) {
    nRc = cn14_setErrtext(errtextC, DBMAPI_INVSESSION_CN14);
  } else {

    sqlareplyavailable (session->reference, errtextP, &commErr);

    if (commErr != commErrOk_esp01) {
      cn14_errtextToC (errtextC, errtextP);
      nRc = DBMAPI_COMMERR_CN14;
    } /* end if */

  } /* end if */

  return nRc;
} /* end cn14replyAvailable */

/*
  -----------------------------------------------------------------------------
  public function cn14cmdRequest
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14cmdRequest"
tsp_int4 WINAPI_CN14 
cn14cmdRequest (
    void           * sessionParm,
    const void     * pData,
    tsp_int4         nLen,
    tsp00_ErrTextc   errtextC )
{
  tsp_int4          nRc     = DBMAPI_OK_CN14;

  nRc =  cn14_cmdWrite (sessionParm, pData, nLen);
  if (nRc != DBMAPI_OK_CN14) {
    nRc = cn14_setErrtext(errtextC, nRc);
  } else {
    nRc = cn14_request (sessionParm, errtextC);
  } /* end if */

  return nRc;
} /* end cn14cmdRequest */

/*
  -----------------------------------------------------------------------------
  public function cn14replyReceive
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14replyReceive"
tsp_int4 WINAPI_CN14 
cn14replyReceive (
    void           * sessionParm,
    void           * pData,
    tsp_int4       * nLen,
    tsp00_ErrTextc   errtextC )
{
  tsp_int4          nRc     = DBMAPI_OK_CN14;

  nRc = cn14_receive (sessionParm, errtextC);
  if (nRc == DBMAPI_OK_CN14) {
    if (pData != NULL) {
      nRc =  cn14replyRead (sessionParm, pData, nLen);
      if (nRc != DBMAPI_OK_CN14) {
       nRc = cn14_setErrtext(errtextC, nRc);
      } /* end if */
    } /* end if */
  } /* end if */

  return nRc;
} /* end cn14replyReceive */

/*
  -----------------------------------------------------------------------------
  public function cn14cmdExecute
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14cmdExecute"
tsp_int4 WINAPI_CN14 
cn14cmdExecute (
    void           * sessionParm,
    const void     * pInData,
    tsp_int4         nInLen,
    void           * pOutData,
    tsp_int4       * nOutLen,
    tsp00_ErrTextc   errtext )
{
  tsp_int4          nRc     = DBMAPI_OK_CN14;

  nRc = cn14cmdRequest(sessionParm, pInData, nInLen, errtext);
  if (nRc == DBMAPI_OK_CN14) {
    nRc = cn14replyReceive(sessionParm, pOutData, nOutLen, errtext);
  } /* end if */

  return nRc;
} /* end cn14cmdExecute */

/*
  -----------------------------------------------------------------------------
  private function cn14replyRead
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14replyRead"
tsp_int4 WINAPI_CN14
cn14replyRead (
    void     * sessionParm,
    void     * data,
    tsp_int4 * len )
{
  ControlSessionT* session = (ControlSessionT*)sessionParm;
  tsp_int4          nRc     = DBMAPI_OK_CN14;

  if (session == NULL) {
    nRc = DBMAPI_INVSESSION_CN14;
  } else {

    if (*len > 0) {
      if (*len > (session->packetLen - session->packetPos)) {
          *len = session->packetLen - session->packetPos; 
      } /* end if */

      if (*len <= 0) {
          nRc = DBMAPI_NOMOREDATA_CN14;
      } else {
        memcpy (data, session->packetOutData + session->packetPos, *len);
        session->packetPos = session->packetPos + *len;
      } /* end if */
    } /* end if */
  } /* end if */

  return nRc;
} /* end cn14replyRead */

/*
  -----------------------------------------------------------------------------
  public function cn14bytesAvailable
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14bytesAvailable"
int WINAPI_CN14
cn14bytesAvailable (
    void * sessionParm)
{
    ControlSessionT* session = (ControlSessionT*)sessionParm;
    int result;

    if (session->packetInData != NULL) {
        result = session->packetSize - session->packetPos;
    } else if (session->packetOutData != NULL) {
        result = session->packetLen - session->packetPos;
    } else {
        result = 0;
    } /* end if */
    return result;
} /* end cn14bytesAvailable */

/*
  -----------------------------------------------------------------------------
  public function cn14rawReadData
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14rawReadData"
const char * WINAPI_CN14 cn14rawReadData (
    void     * sessionParm,
    tsp_int4 * rc)
{
  const char * pReturn = NULL;

  ControlSessionT* session = (ControlSessionT*)sessionParm;
  if ((session == NULL) || (session->packetOutData == NULL)) {
    *rc = DBMAPI_INVSESSION_CN14;
  } else {
    *rc = DBMAPI_OK_CN14;
    pReturn = session->packetOutData + session->packetPos;
  } /* end if */

  return pReturn;
} /* end cn14rawReadData */

/*
  -----------------------------------------------------------------------------
  private function cn14packetstats
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14packetstats"
void WINAPI_CN14
cn14packetstats (
    void *sessionParm,
    tsp_int4 *packetSize,
    tsp_int4 *packetLen,
    tsp_int4 *packetPos)
{
  ControlSessionT* session = (ControlSessionT*)sessionParm;

  if (session != NULL) {
    *packetSize = session->packetSize;
    *packetLen  = session->packetLen;
    *packetPos  = session->packetPos;
  } else {
    *packetSize = 0;
    *packetLen  = 0;
    *packetPos  = 0;
  } /* end if */

} /* end cn14packetstats */

/*
  -----------------------------------------------------------------------------
  private function cn14_dbmLogon
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14_dbmLogon"
static tsp_int4 
cn14_dbmLogon (
    void                    * sessionParm,
    const tsp_c64c            user,     
    tsp00_ErrTextc            errtext,
    const _TCHAR            * szCommand)
{
  ControlSessionT* session = (ControlSessionT*)sessionParm;
  tsp_int4          nRc     = DBMAPI_OK_CN14;
  tsp_name          pwClear;
  tsp_cryptpw       pwCrypt;
  tsp_cryptname     cryptTemp;
  _TCHAR            pData[csp9_ctrl_packet_size];
  tsp_int4          nCmdLen; 
  tsp_int4          nReplyLen; 
  tsp_c64c          locUser;
  _TCHAR         *  pPassword = NULL;  
  const void     *  pPayLoad;
  tsp_int4          nErrorCode;

  _tcscpy(locUser, user);

  pPassword = _tcschr(locUser, _T(','));
  if (pPassword != NULL) {
    *pPassword = _T('\0');
    pPassword++;

    memset(&pwClear[0], _T(' '), sizeof(pwClear));
    _tcsncpy((_TCHAR *) pwClear, pPassword, _tcslen(pPassword));
    s02applencrypt(pwClear, pwCrypt);

    memcpy(&cryptTemp[0], &pwCrypt[0], sizeof(cryptTemp));
    _stprintf ( pData , _T("%s %s,%08x%08x%08x%08x%08x%08x"), szCommand,
                                                              (_TCHAR *) locUser,
                                                               cryptTemp[0],
                                                               cryptTemp[1],
                                                               cryptTemp[2],
                                                               cryptTemp[3],
                                                               cryptTemp[4],
                                                               cryptTemp[5]);

    nCmdLen = _tcslen(pData);

    nRc = cn14cmdExecute(session, pData, nCmdLen, NULL, NULL, errtext);
    if (nRc == DBMAPI_OK_CN14) {
      if (cn14analyzeDbmAnswer(session, &pPayLoad, &nReplyLen, &nErrorCode, errtext) != DBMAPI_OK_CN14) {
        nRc = DBMAPI_USR_FALSE_CN14;
      } /* end if */ 
    } /* end if */ 

  } else {
    nRc = cn14_setErrtext(errtext, DBMAPI_USR_FALSE_CN14);
  } /* end if */

  return nRc;
} /* end cn14_dbmLogon */

/*
  -----------------------------------------------------------------------------
  private function cn14_cmdWrite
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14_cmdWrite"
static tsp_int4 
cn14_cmdWrite (
    void       * sessionParm,
    const void * data,
    tsp_int4     len)
{
  ControlSessionT * session = (ControlSessionT*)sessionParm;
  tsp_int4          nRc     = DBMAPI_OK_CN14;

  if (session == NULL) {
    nRc = DBMAPI_INVSESSION_CN14;
  } else {
    /* init request packet */
    if (session->packetInData == NULL) {
      session->packetInData = session->packetData;
      session->packetPos = 0;
    } /* end if */
    /* add packet data */
    if (len > (session->packetSize - session->packetPos)) {
       nRc = DBMAPI_TOSMALL_CN14;
    } else {
      memcpy (session->packetInData + session->packetPos, data, len);
      session->packetPos = session->packetPos + len;
    } /* end if */
  } /* end if */
    
  return nRc;
} /* end cn14_cmdWrite */

/*
  -----------------------------------------------------------------------------
  private function cn14_request
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14_request"
static tsp_int4 
cn14_request (
    void            * sessionParm,
    tsp00_ErrTextc    errtextC )
{
  ControlSessionT* session = (ControlSessionT*)sessionParm;
  tsp_int4          nRc     = DBMAPI_OK_CN14;
  tsp00_ErrText     errtextP;
  tsp01_CommErr     commErr;
  int               packetLen;
  int               alignmod;

  if ((session == NULL) || (session->packetInData == NULL)) {
    nRc = cn14_setErrtext(errtextC, DBMAPI_INVSESSION_CN14);
  } else {

    alignmod = session->packetPos % 8;
    if (alignmod != 0) {
      memset (session->packetInData + session->packetPos, _T(' '), 8 - alignmod);
      session->packetPos += 8 - alignmod;
    } /* end if */
 
    sqlarequest (session->reference, 
                 session->packetInData, 
                 session->packetPos,
                 errtextP, 
                 &commErr);
    session->packetInData = NULL;
    session->packetPos = 0;

    if (commErr != commErrOk_esp01) {
      cn14_errtextToC (errtextC, errtextP);
      nRc = DBMAPI_COMMERR_CN14;
    } /* end if */

  } /* end if */

  return nRc;
} /* end cn14_request */

/*
  -----------------------------------------------------------------------------
  private function cn14_receive
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14_receive"
static tsp_int4 
cn14_receive (
    void            * sessionParm,
    tsp00_ErrTextc    errtextC )
{
  ControlSessionT*  session = (ControlSessionT*)sessionParm;
  tsp_int4          nRc     = DBMAPI_OK_CN14;
  tsp01_CommErr     commErr;
  tsp00_ErrText     errtextP;

  if ((session == NULL) || (session->packetInData != NULL)) {
    nRc = cn14_setErrtext(errtextC, DBMAPI_INVSESSION_CN14);
  } else {

    sqlareceive (session->reference, 
                 (void **) &session->packetOutData, 
                 &session->packetLen, 
                 errtextP, 
                 &commErr);

    if (commErr != commErrOk_esp01) {
      cn14_errtextToC (errtextC, errtextP);
      nRc = DBMAPI_COMMERR_CN14;
    } /* end if */

  } /* end if */

  return nRc;
} /* end cn14_receive */

/*
  -----------------------------------------------------------------------------
  private function cn14_errtextToC
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14_errtextToC"
static void 
cn14_errtextToC (
    tsp00_ErrTextc errtextC,
    tsp00_ErrText  errtextP )
{
  int i;

  memset(errtextC, _T('\0'), sizeof(tsp00_ErrTextc));
  memcpy(errtextC, errtextP, sizeof(tsp00_ErrText));

  for (i = sizeof (tsp00_ErrText) - 1; i >= 0; --i) {
    if (errtextC [i] == _T(' ')) {
      errtextC [i] = _T('\0');
    } else {
      break;
    } /* end if */
  } /* end for */

} /* end cn14_errtextToC */

/*
  -----------------------------------------------------------------------------
  private function cn14_setErrtext
  -----------------------------------------------------------------------------
*/
    #undef MF__
    #define MF__ MOD__"cn14_setErrtext"
static tsp_int4 
cn14_setErrtext (
    tsp00_ErrTextc errtextC,
    tsp_int4       nRc )
{
  switch (nRc) {
    case DBMAPI_OK_CN14:
      _tcscpy (errtextC, _T(""));
      break;
    case DBMAPI_USR_UNKNOWN_CN14:
      _tcscpy (errtextC, _T("user unknown"));
      break;
    case DBMAPI_USR_FALSE_CN14:
      _tcscpy (errtextC, _T("wrong user/password"));
      break;  
    case DBMAPI_MEMORY_CN14:
      _tcscpy (errtextC, _T("out of memory"));
      break;     
    case DBMAPI_COMMERR_CN14:
      _tcscpy (errtextC, _T("communication error"));
      break;    
    case DBMAPI_TOSMALL_CN14:
      _tcscpy (errtextC, _T("packet too small"));
      break;    
    case DBMAPI_INVSESSION_CN14:
      _tcscpy (errtextC, _T("invalid session data"));
      break; 
    case DBMAPI_NOMOREDATA_CN14:
      _tcscpy (errtextC, _T("no more data available"));
      break; 
    case DBMAPI_XUSERKEY_CN14:
      _tcscpy (errtextC, _T("generated user key too long"));
      break;     
    case DBMAPI_NOT_OK_CN14:
      _tcscpy (errtextC, _T("some error occured"));
      break;     
    case DBMAPI_R3TP_CN14:
      _tcscpy (errtextC, _T("tp error: "));
      break;     
    default:
      _tcscpy (errtextC, _T("unknown error code"));
      break;
  } /* end switch */

  return nRc;
} /* end cn14_setErrtext */

/*
  -----------------------------------------------------------------------------
  private function cn14_senduser
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14_senduser"
static tsp_int4 
cn14_senduser (
    void                * pSession,
    const tsp00_NodeIdc   szServerNode,
    const tsp00_DbNamec   szDBName)
{
  _TCHAR              pData[csp9_ctrl_packet_size];
  tsp_int4            nRc     = DBMAPI_OK_CN14;
  tsp_int4            nCmdLen; 
  tsp_int4            nReplyLen; 
  tsp00_ErrTextc      szErrC;
  tsp00_ErrText       szErrP;
  tsp4_xuser_record   recUser;
  _TCHAR            * pUserkey;
  tsp00_Bool          bOk;
  tsp_cryptname       cryptTemp;
  int                 nIndex;
  const void       *  pPayLoad;
  tsp_int4            nErrorCode;

  if ((_tcslen(szServerNode) + _tcslen(szDBName) + _tcslen(XU_DBM_MARK)) <= sizeof(recUser.xu_key)) {
    /* prepare userkey */
    memset(&recUser.xu_key[0], _T(' '), sizeof(recUser.xu_key));
    pUserkey = (char *) &recUser.xu_key[0];
    _tcsncpy(pUserkey, XU_DBM_MARK, _tcslen(XU_DBM_MARK));
    pUserkey = pUserkey + _tcslen(XU_DBM_MARK);
    _tcsncpy(pUserkey, szDBName, _tcslen(szDBName));
    pUserkey = pUserkey + _tcslen(szDBName);
    _tcsncpy(pUserkey, szServerNode, _tcslen(szServerNode));

    /* get user */
    sqlgetuser (&recUser, szErrP, &bOk);

    if (bOk) {
      memcpy(cryptTemp, &recUser.xu_password, sizeof(cryptTemp));
      nIndex = sizeof(recUser.xu_user) - 1;
      while (nIndex >= 0 && recUser.xu_user[nIndex] == _T(' ')) {
        recUser.xu_user[nIndex] = _T('\0');
        nIndex--;
      } /* end if */
      _stprintf ( pData , _T("%s %s,%08x%08x%08x%08x%08x%08x"), _T("user_logon"),
                                                                (_TCHAR *) recUser.xu_user,
                                                                 cryptTemp[0],
                                                                 cryptTemp[1],
                                                                 cryptTemp[2],
                                                                 cryptTemp[3],
                                                                 cryptTemp[4],
                                                                 cryptTemp[5]);
      nCmdLen = _tcslen(pData);
      nReplyLen = csp9_ctrl_packet_size - nCmdLen;

      nRc = cn14cmdExecute(pSession, pData, nCmdLen, NULL, NULL, szErrC);
      if (nRc == DBMAPI_OK_CN14) {
        if (cn14analyzeDbmAnswer(pSession, &pPayLoad, &nReplyLen, &nErrorCode, szErrC) != DBMAPI_OK_CN14) {
          nRc = DBMAPI_USR_FALSE_CN14;
        } /* end if */ 
      } /* end if */ 

    } /* end if */

  } /* end if */

  return nRc;

} /* end cn14_senduser */

/*
  -----------------------------------------------------------------------------
  public function cn14saveUser
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14saveUser"
tsp_int4 WINAPI_CN14 
cn14saveUser (
    const tsp00_NodeIdc   szServerNode,
    const tsp00_DbNamec   szDBName,
    const _TCHAR *        szUser)
{
  tsp00_C18c szKey;
  int        nReturn = DBMAPI_USR_FALSE_CN14;

  if ((_tcslen(szServerNode) + _tcslen(szDBName) + _tcslen(XU_DBM_MARK)) <= (sizeof(szKey) - 1)) {
  
    /* prepare userkey */
    sprintf(szKey, "%s%s%s", XU_DBM_MARK, szDBName, szServerNode);

    nReturn = cn14saveUserByKey(szKey, szServerNode, szDBName, szUser, NULL);

  } /* end if */

  return nReturn;
} /* end cn14saveUser */

/*
  -----------------------------------------------------------------------------
  public function cn14saveUserByKey
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14saveUserByKey"
tsp_int4 WINAPI_CN14 
cn14saveUserByKey (
    const _TCHAR *        szKey,
    const tsp00_NodeIdc   szServerNode,
    const tsp00_DbNamec   szDBName,
    const _TCHAR *        szUser,
    const _TCHAR *        szParams)
{
  int                 nReturn = DBMAPI_OK_CN14;
  tsp4_xuser_record   recUser;
  _TCHAR            * pUserkey;
  int                 nIndex;
  _TCHAR              szNumber1[2];
  _TCHAR              szNumber2[2];
  _TCHAR            * pEnd;
  tsp_errtext         szError;
  tsp00_Bool          bOk;
  const _TCHAR      * pPassword;
  int                 nUsrLen;
  tsp_name            szPassword;

  /* userrecord */
  memset(&recUser, _T('\0'), sizeof(recUser));

  /* search for password */
  pPassword = _tcschr (szUser, _T(','));
  if (pPassword != NULL) {
    nUsrLen = pPassword - szUser;
    ++pPassword;
  } else {
    nUsrLen = _tcslen (szUser);
    pPassword = _T("");
  } /* end if */

  if (_tcslen(szKey) <= sizeof(recUser.xu_key)) {
  
    /* prepare userkey */
    memset(&recUser.xu_key[0], _T(' '), sizeof(recUser.xu_key));
    pUserkey = (char *) &recUser.xu_key[0];
    _tcsncpy(pUserkey, szKey, _tcslen(szKey));

    sqlxuopenuser (szError, &bOk);
    if (bOk) {

      /* get user */
      sqlgetuser (&recUser, szError, &bOk);
      bOk = true;

      /* prepare password for save */
      memset(&recUser.xu_password, _T('\0'), sizeof(recUser.xu_password));
      if (_tcslen(pPassword) == (sizeof(recUser.xu_password) * 2)) {
        szNumber1[1] = _T('\0');
        szNumber2[1] = _T('\0');
        for (nIndex = 0; nIndex < sizeof(recUser.xu_password); nIndex++) {
          szNumber1[0] = pPassword[(2 * nIndex)];
          szNumber2[0] = pPassword[(2 * nIndex) + 1];
          recUser.xu_password[nIndex] = strtoul(szNumber1, &pEnd, XU_CRYPT_BASE) * XU_CRYPT_BASE + 
                                        strtoul(szNumber2, &pEnd, XU_CRYPT_BASE);
        } /* end for */
      } else {
        memset(&szPassword[0], _T(' '), sizeof(szPassword));
        _tcsncpy((_TCHAR *) szPassword, pPassword, _tcslen(pPassword));
        s02applencrypt(szPassword, recUser.xu_password);
      } /* end if */

      /* prepare user for save */
      memset(&recUser.xu_user[0], _T(' '), sizeof(recUser.xu_user));
      _tcsncpy((_TCHAR *) recUser.xu_user, szUser, nUsrLen);

      memset(&recUser.xu_servernode[0], _T(' '), sizeof(recUser.xu_servernode));
      _tcsncpy((_TCHAR *) recUser.xu_servernode, szServerNode, _tcslen(szServerNode));

      memset(&recUser.xu_serverdb[0], _T(' '), sizeof(recUser.xu_serverdb));
      _tcsncpy((_TCHAR *) recUser.xu_serverdb, szDBName, _tcslen(szDBName));

      if (szParams != NULL) {
        const _TCHAR * pParam = NULL;
        const _TCHAR * pValue = NULL;
        const _TCHAR * pEnd   = NULL;

        pParam = strstr(szParams, "CACHELIMIT");
        if (pParam != NULL) {
          pValue = strchr(pParam, '=');
          if (pValue != NULL) {
            ++pValue;
            pEnd = strchr(pValue, ';');
            if (pEnd != NULL) {
              recUser.xu_cachelimit = (pEnd == pValue) ? -1 : atoi(pValue);
            } /* end if */
          } /* end if */
        } /* end if */
        pParam = strstr(szParams, "SQLMODE");
        if (pParam != NULL) {
          pValue = strchr(pParam, '=');
          if (pValue != NULL) {
            ++pValue;
            pEnd = strchr(pValue, ';');
            if (pEnd != NULL) {
              memset(&recUser.xu_sqlmode[0], _T(' '), sizeof(recUser.xu_sqlmode));
              _tcsncpy((_TCHAR *) recUser.xu_sqlmode, pValue, pEnd - pValue);
              if (!(_tcsncmp((_TCHAR *) recUser.xu_sqlmode, "INTERNAL", 8) == 0 ||
                    _tcsncmp((_TCHAR *) recUser.xu_sqlmode, "ANSI    ", 8) == 0 ||
                    _tcsncmp((_TCHAR *) recUser.xu_sqlmode, "DB2     ", 8) == 0 ||
                    _tcsncmp((_TCHAR *) recUser.xu_sqlmode, "ORACLE  ", 8) == 0 ||
                    _tcsncmp((_TCHAR *) recUser.xu_sqlmode, "SAPR3   ", 8) == 0    )) {
                bOk = false;
              } /* end if */
            } /* end if */
          } /* end if */
        } /* end if */
        pParam = strstr(szParams, "TIMEOUT");
        if (pParam != NULL) {
          pValue = strchr(pParam, '=');
          if (pValue != NULL) {
            ++pValue;
            pEnd = strchr(pValue, ';');
            if (pEnd != NULL) {
              recUser.xu_timeout = (pEnd == pValue) ? -1 : atoi(pValue);
            } /* end if */
          } /* end if */
        } /* end if */
        pParam = strstr(szParams, "ISOLATION");
        if (pParam != NULL) {
          pValue = strchr(pParam, '=');
          if (pValue != NULL) {
            ++pValue;
            pEnd = strchr(pValue, ';');
            if (pEnd != NULL) {
              recUser.xu_isolation = (pEnd == pValue) ? -1 : atoi(pValue);
              switch (recUser.xu_isolation) {
                case 0: 
                case 1:
                case 2:
                case 3:
                case 10:
                case 15:
                case 20:
                case 30:
                  break;
                default:
                  bOk = false;
                  break;
              } /* end switch */
            } /* end if */
          } /* end if */
        } /* end if */
        pParam = strstr(szParams, "DBLOCALE");
        if (pParam != NULL) {
          pValue = strchr(pParam, '=');
          if (pValue != NULL) {
            ++pValue;
            pEnd = strchr(pValue, ';');
            if (pEnd != NULL) {
              memset(&recUser.xu_dblang[0], _T(' '), sizeof(recUser.xu_dblang));
              _tcsncpy((_TCHAR *) recUser.xu_dblang, pValue, pEnd - pValue);
            } /* end if */
          } /* end if */
        } /* end if */
      } /* end if */

      if (bOk) {
        sqlputuser (&recUser, szError, &bOk);
      } /* end if */
      if (!bOk) {
        nReturn = DBMAPI_XUSER_CN14;
      } /* end if */
      sqlxucloseuser (szError, &bOk);
    } else {
      nReturn = DBMAPI_XUSER_CN14;
    } /* end if */
  } else {
    nReturn = DBMAPI_XUSERKEY_CN14;
  } /* end if */


  return nReturn;
} /* end cn14saveUserByKey */

/*
  -----------------------------------------------------------------------------
  public function cn14checkUser
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14checkUser"
tsp_int4 WINAPI_CN14 
cn14checkUser (
    const tsp00_NodeIdc   szServerNode,
    const tsp00_DbNamec   szDBName,
    const _TCHAR *        szUser)
{
  tsp00_C18c szKey;
  int        nReturn = DBMAPI_USR_FALSE_CN14;

  if ((_tcslen(szServerNode) + _tcslen(szDBName) + _tcslen(XU_DBM_MARK)) <= (sizeof(szKey) - 1)) {
  
    /* prepare userkey */
    sprintf(szKey, "%s%s%s", XU_DBM_MARK, szDBName, szServerNode);

    nReturn = cn14checkUserByKey(szKey, szUser);

  } /* end if */

  return nReturn;
} /* end cn14checkUser */

/*
  -----------------------------------------------------------------------------
  public function cn14checkUserByKey
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14checkUserByKey"
tsp_int4 WINAPI_CN14 
cn14checkUserByKey (
    const _TCHAR *        szKey,
    const _TCHAR *        szUser)
{
  int                 nReturn = DBMAPI_USR_FALSE_CN14;
  tsp4_xuser_record   recUserEnv;
  tsp4_xuser_record   recUserIn;
  _TCHAR            * pUserkey;
  int                 nIndex;
  _TCHAR              szNumber1[2];
  _TCHAR              szNumber2[2];
  _TCHAR            * pEnd;
  tsp_errtext         szError;
  tsp00_Bool          bOk;
  const _TCHAR      * pPassword;
  int                 nUsrLen;
  tsp_name            szPassword;

  /* search for password */
  pPassword = _tcschr (szUser, _T(','));
  if (pPassword != NULL) {
    nUsrLen = pPassword - szUser;
    ++pPassword;
  } else {
    nUsrLen = _tcslen (szUser);
    pPassword = _T("");
  } /* end if */

  /* prepare password for check */
  memset(&recUserIn.xu_password[0], _T('\0'), sizeof(recUserIn.xu_password));
  if (_tcslen(pPassword) == (sizeof(recUserIn.xu_password) * 2)) {
    szNumber1[1] = _T('\0');
    szNumber2[1] = _T('\0');
    for (nIndex = 0; nIndex < sizeof(recUserIn.xu_password); nIndex++) {
      szNumber1[0] = pPassword[(2 * nIndex)];
      szNumber2[0] = pPassword[(2 * nIndex) + 1];
      recUserIn.xu_password[nIndex] = strtoul(szNumber1, &pEnd, XU_CRYPT_BASE) * XU_CRYPT_BASE + 
                                      strtoul(szNumber2, &pEnd, XU_CRYPT_BASE);
    } /* end for */
  } else {
    memset(&szPassword[0], _T(' '), sizeof(szPassword));
    _tcsncpy((_TCHAR *) szPassword, pPassword, _tcslen(pPassword));
    s02applencrypt(szPassword, recUserIn.xu_password);
  } /* end if */

  /* prepare user for check */
  memset(&recUserIn.xu_user[0], _T(' '), sizeof(recUserIn.xu_user));
  _tcsncpy((_TCHAR *) recUserIn.xu_user, szUser, nUsrLen);

  if (_tcslen(szKey) <= sizeof(recUserEnv.xu_key)) {
  
    /* prepare userkey */
    memset(&recUserEnv.xu_key[0], _T(' '), sizeof(recUserEnv.xu_key));
    pUserkey = (char *) &recUserEnv.xu_key[0];
    _tcsncpy(pUserkey, szKey, _tcslen(szKey));

    /* get user */
    sqlgetuser (&recUserEnv, szError, &bOk);

    if (bOk) {
      if (memcmp(&recUserIn.xu_user[0],     &recUserEnv.xu_user[0],     sizeof(recUserIn.xu_user)     ) == 0 && 
          memcmp(&recUserIn.xu_password[0], &recUserEnv.xu_password[0], sizeof(recUserIn.xu_password) ) == 0    ) {
        nReturn = DBMAPI_OK_CN14;
      } /* end if */

    } else {
      nReturn = DBMAPI_USR_UNKNOWN_CN14;
    } /* end if */
    
  } else {
    nReturn = DBMAPI_USR_UNKNOWN_CN14;
  } /* end if */

  return nReturn;
} /* end cn14checkUserByKey */

/*
  -----------------------------------------------------------------------------
  public function cn14checkDefaultUser
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14checkDefaultUser"
tsp_int4 WINAPI_CN14 
cn14checkDefaultUser (
    const _TCHAR *        szUser)
{
  int                 nReturn = DBMAPI_USR_FALSE_CN14;
  tsp4_xuser_record   recUserEnv;
  tsp4_xuser_record   recUserIn;
  _TCHAR            * pUserkey;
  int                 nIndex;
  _TCHAR              szNumber1[2];
  _TCHAR              szNumber2[2];
  _TCHAR            * pEnd;
  tsp_errtext         szError;
  tsp00_Bool          bOk;
  const _TCHAR      * pPassword;
  int                 nUsrLen;
  tsp_name            szPassword;

  /* search for password */
  pPassword = _tcschr (szUser, _T(','));
  if (pPassword != NULL) {
    nUsrLen = pPassword - szUser;
    ++pPassword;
  } else {
    nUsrLen = _tcslen (szUser);
    pPassword = _T("");
  } /* end if */

  /* prepare password for check */
  memset(&recUserIn.xu_password[0], _T('\0'), sizeof(recUserIn.xu_password));
  if (_tcslen(pPassword) == (sizeof(recUserIn.xu_password) * 2)) {
    szNumber1[1] = _T('\0');
    szNumber2[1] = _T('\0');
    for (nIndex = 0; nIndex < sizeof(recUserIn.xu_password); nIndex++) {
      szNumber1[0] = pPassword[(2 * nIndex)];
      szNumber2[0] = pPassword[(2 * nIndex) + 1];
      recUserIn.xu_password[nIndex] = strtoul(szNumber1, &pEnd, XU_CRYPT_BASE) * XU_CRYPT_BASE + 
                                      strtoul(szNumber2, &pEnd, XU_CRYPT_BASE);
    } /* end for */
  } else {
    memset(&szPassword[0], _T(' '), sizeof(szPassword));
    _tcsncpy((_TCHAR *) szPassword, pPassword, _tcslen(pPassword));
    s02applencrypt(szPassword, recUserIn.xu_password);
  } /* end if */

  /* prepare user for check */
  memset(&recUserIn.xu_user[0], _T(' '), sizeof(recUserIn.xu_user));
  _tcsncpy((_TCHAR *) recUserIn.xu_user, szUser, nUsrLen);

  /* prepare userkey */
  memset(&recUserEnv.xu_key[0], _T(' '), sizeof(recUserEnv.xu_key));
  pUserkey = (char *) &recUserEnv.xu_key[0];
  _tcsncpy(pUserkey, XU_KEY_DEFAULT, _tcslen(XU_KEY_DEFAULT));

  /* open XUSer-File */
  sqlxuopenuser (szError, &bOk);

  if (bOk) {

    /* get user */
    sqlgetuser (&recUserEnv, szError, &bOk);

    if (bOk) {

      if ( memcmp(&recUserIn.xu_user[0],     &recUserEnv.xu_user[0],     sizeof(recUserIn.xu_user)     ) == 0 && 
           memcmp(&recUserIn.xu_password[0], &recUserEnv.xu_password[0], sizeof(recUserIn.xu_password) ) == 0 /*&&
           _tcsnicmp(szUser, XU_SAP_USER, nUsrLen)                                                       == 0   */ ) {
        /* match only if SAPR3 */
        nReturn = DBMAPI_OK_CN14;
      } else {
        /* if in the xuser not SAPR3 and the input SAPR3/SAP then overwrite */
        memset(&recUserIn.xu_user[0], _T(' '), sizeof(recUserIn.xu_user));
        _tcsncpy((_TCHAR *) recUserIn.xu_user, XU_SAP_USER, _tcslen(XU_SAP_USER));
        if ( memcmp(&recUserIn.xu_user[0], &recUserEnv.xu_user[0], sizeof(recUserIn.xu_user)) != 0 &&
             _tcsnicmp(szUser, XU_SAP_USER, nUsrLen)                                          == 0 &&  
             _tcsicmp(pPassword, XU_SAP_PWD)                                                  == 0    ) {
          /* write username to structure */
          memset(&recUserEnv.xu_user[0], _T(' '), sizeof(recUserIn.xu_user));
          _tcsncpy((_TCHAR *) &recUserEnv.xu_user[0],    XU_SAP_USER, _tcslen(XU_SAP_USER));
          /* write pwd to structure */
          memset(&szPassword[0], _T(' '), sizeof(szPassword));
          _tcsncpy((_TCHAR *) szPassword, XU_SAP_PWD, _tcslen(XU_SAP_PWD));
          s02applencrypt(szPassword, recUserEnv.xu_password);
          /* save */
          sqlputuser (&recUserEnv, szError, &bOk);
          if (bOk) {
            nReturn = DBMAPI_OK_CN14;
          } /* end if */
        } /* end if */
      } /* end if */

    } else {
      /* default unknown, then save also */
      /* write username to structure */
      memset(&recUserEnv.xu_user[0], _T(' '), sizeof(recUserIn.xu_user));
      _tcsncpy((_TCHAR *) &recUserEnv.xu_user[0],    XU_SAP_USER, _tcslen(XU_SAP_USER));
      /* write pwd to structure */
      memset(&szPassword[0], _T(' '), sizeof(szPassword));
      _tcsncpy((_TCHAR *) szPassword, XU_SAP_PWD, _tcslen(XU_SAP_PWD));
      s02applencrypt(szPassword, recUserEnv.xu_password);
      /* save */
      sqlputuser (&recUserEnv, szError, &bOk);
      if (bOk) {
        nReturn = DBMAPI_OK_CN14;
      } /* end if */
    } /* end if */

     /* close xuser-file */
    sqlxucloseuser (szError, &bOk);

  } /* end if */

  return nReturn;
} /* end cn14checkDefaultUser */

/*
  -----------------------------------------------------------------------------
  public function cn14deleteUser
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14deleteUser"
tsp_int4 WINAPI_CN14 
cn14deleteUser (
    const tsp00_NodeIdc   szServerNode,
    const tsp00_DbNamec   szDBName)
{
  tsp00_C18c szKey;
  int        nReturn = DBMAPI_USR_FALSE_CN14;

  if ((_tcslen(szServerNode) + _tcslen(szDBName) + _tcslen(XU_DBM_MARK)) <= (sizeof(szKey) - 1)) {
  
    /* prepare userkey */
    sprintf(szKey, "%s%s%s", XU_DBM_MARK, szDBName, szServerNode);

    nReturn = cn14deleteUserByKey(szKey);

  } /* end if */

  return nReturn;
} /* end cn14deleteUser */

/*
  -----------------------------------------------------------------------------
  public function cn14deleteUserByKey
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14deleteUserByKey"
tsp_int4 WINAPI_CN14 
cn14deleteUserByKey (
    const _TCHAR *        szKey)
{
  int                 nReturn = DBMAPI_OK_CN14;
  tsp4_xuser_record   recUsers[XU_MAX_USERS];
  tsp4_xuser_record   recUser;
  _TCHAR            * pUserkey;
  int                 nLoadIndex;
  int                 nSaveIndex;


  char                szNumber1[2];
  char                szNumber2[2];
  char              * pEnd;
  tsp_errtext         szError;
  tsp00_Bool          bOk;
  const char        * pPassword;
  int                 nUsrLen;
  tsp_name            szPassword;
  bool                bFound = false;

  /* userrecord */
  memset(&recUser, 0, sizeof(recUser));

  if (_tcslen(szKey) <= sizeof(recUser.xu_key)) {
  
    /* prepare userkey */
    memset(&recUser.xu_key[0], _T(' '), sizeof(recUser.xu_key));
    pUserkey = (char *) &recUser.xu_key[0];
    _tcsncpy(pUserkey, szKey, _tcslen(szKey));

    sqlxuopenuser (szError, &bOk);

    /* read */
    nLoadIndex = 0;
    while (bOk && (nLoadIndex < XU_MAX_USERS)) {
      sqlindexuser(nLoadIndex + 1, &recUsers[nLoadIndex], szError, &bOk);
      nLoadIndex++;
    } /* end while */

    /* clear the file */
    sqlclearuser();

    /* save */
    nLoadIndex = (bOk) ? nLoadIndex : nLoadIndex - 1;
    nSaveIndex = 0;
    bOk = true;
    while (bOk && (nSaveIndex < nLoadIndex)) {
      if (memcmp(&recUser.xu_key, &recUsers[nSaveIndex].xu_key, sizeof(recUser.xu_key)     ) != 0) {
        sqlputuser (&recUsers[nSaveIndex], szError, &bOk);
      } else {
        bFound = true;
      } /* end if */
      nSaveIndex++;
    } /* end while */

    sqlxucloseuser (szError, &bOk);

    if (!bOk) {
      nReturn = DBMAPI_XUSER_CN14;
    } else if (!bFound) {
      nReturn = DBMAPI_USR_UNKNOWN_CN14;
    } /* end if */

  } else {
    nReturn = DBMAPI_XUSERKEY_CN14;
  } /* end if */

  return nReturn;
} /* end cn14deleteUserByKey */

/*
  -----------------------------------------------------------------------------
  public function cn14listUsers
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14listUsers"
long WINAPI_CN14 
cn14listUsers (
    _TCHAR                * pBuffer,
    long                    nBuffer)
{
  tsp4_xuser_record   recUsers[XU_MAX_USERS];
  _TCHAR            * pUserkey;
  int                 nLoadIndex;
  int                 nIndex;
  tsp_errtext         szError;
  tsp00_Bool          bOk;
  long                nUsed = 0;
  _TCHAR            * pCurrent = pBuffer;

  sqlxuopenuser (szError, &bOk);

  /* read */
  nLoadIndex = 0;
  while (bOk && (nLoadIndex < XU_MAX_USERS)) {
    sqlindexuser(nLoadIndex + 1, &recUsers[nLoadIndex], szError, &bOk);
    nLoadIndex++;
  } /* end while */
  nLoadIndex = (bOk) ? nLoadIndex : nLoadIndex - 1;

  /* clear the file */
  sqlxucloseuser (szError, &bOk);
    
  if (nLoadIndex > 0) {
    for (nIndex = 0; nIndex < nLoadIndex; ++nIndex) {
      if ((pCurrent + sizeof(recUsers[nIndex].xu_key) + /*sizeof(recUsers[nIndex].xu_user)*/ 18 + 2) < 
          (pBuffer + nBuffer)  ) {
        _stprintf(pCurrent, _T("%.*s %.*s\n"),
                            sizeof(recUsers[nIndex].xu_key),
                            (_TCHAR *) &recUsers[nIndex].xu_key[0],
                            /* sizeof(recUsers[nIndex].xu_user)*/ 18,
                            
                            (_TCHAR *) &recUsers[nIndex].xu_user[0]);
        pCurrent = pCurrent + _tcslen(pCurrent);
      } /* end if */


    } /* end for */
  } /* end if */

  return nLoadIndex;
} /* end cn14listUsers */

/*
  -----------------------------------------------------------------------------
  public function cn14analyzeDbmAnswer
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14analyzeDbmAnswer"
externC tsp_int4 WINAPI_CN14 
cn14analyzeDbmAnswer (
    void           * sessionParm,
    const void    ** pPayLoad,
    tsp_int4       * nPayLoad,
    tsp_int4       * pErrorCode,
    tsp00_ErrTextc   errtext )
{ 
  tsp_int4          nRc            = DBMAPI_OK_CN14;
  ControlSessionT * session        = (ControlSessionT*)sessionParm;
  const _TCHAR    * pCharData      = NULL;
  const _TCHAR    * pCharPayload   = NULL;
  const _TCHAR    * pCharError     = NULL;
  int               nCopy          = 0;

  if ((session == NULL) || (session->packetOutData == NULL)) {
    nRc = DBMAPI_INVSESSION_CN14;
  } else {

    *nPayLoad = session->packetLen;
    pCharData = (const _TCHAR *) session->packetOutData;

    if (_tcsncmp(pCharData, CTRL_ERR, _tcslen(CTRL_ERR)) == 0) {
      nRc = DBMAPI_NOT_OK_CN14;
      pCharError = _tcschr (pCharData, _T('\n'));
      if (pCharError != NULL) {
        pCharError++;
        *pErrorCode = atol(pCharError);

        pCharPayload = _tcschr (pCharError, _T('\n'));
        if (pCharPayload != NULL) {
          pCharPayload++;
        } /* end if */

        /* copy error text */
        pCharError = _tcschr (pCharError, _T(','));
        if (pCharError != NULL) {
          pCharError++;
          memset(errtext, _T('\0'), sizeof(tsp00_ErrTextc));
          if (pCharPayload != NULL) {
            nCopy = pCharPayload - pCharError - 1;
          } else {
            nCopy = *nPayLoad - (pCharData - pCharError);
          } /* end if */
          nCopy = (nCopy >= sizeof(tsp00_ErrTextc)) ? 
                    sizeof(tsp00_ErrTextc) - 1 : 
                    nCopy;
          _tcsncpy(errtext, pCharError, nCopy);
        } /* end if */

      } else {
        *pErrorCode = 0;
      } /* end if */
    } else {
      pCharPayload = _tcschr (pCharData, _T('\n'));
      if (pCharPayload != NULL) {
        pCharPayload++;
      } /* end if */
    } /* end if */

    if (pCharPayload != NULL) {
      *nPayLoad = *nPayLoad - (pCharPayload - pCharData);
      *pPayLoad = (void *) pCharPayload;
    } else {
      *nPayLoad = 0;
      *pPayLoad = NULL;
    } /* end if */

  } /* end if */

  return nRc;

} /* end cn14analyzeDbmAnswer */

/*
  -----------------------------------------------------------------------------
  public function cn14analyzeDbmAnswer
  -----------------------------------------------------------------------------
 */
    #undef MF__
    #define MF__ MOD__"cn14analyzeRpmAnswer"
externC tsp_int4 WINAPI_CN14 
cn14analyzeRpmAnswer(void         *pRPMSession,
                     tsp_int4     *plErrCode,
                     const _TCHAR **ppszErrText,
                     tsp_int4     *plErrLen,
                     tsp_int4     *plSqlCode,
                     const _TCHAR **ppszSqlErrText,
                     tsp_int4     *plSqlErrLen)
{ 
    tsp_int4 rc = DBMAPI_OK_CN14;
    
    ControlSessionT *pSession = (ControlSessionT*)pRPMSession;
    
    const _TCHAR *pszData   = NULL;
    const _TCHAR *pszErr    = "";
    const _TCHAR *pszSqlErr = "";

    tsp_int4 lMyErrLen = pSession->packetLen;

    *plErrCode = 0;             /* Initialisiere die fehlernummern */
    *plSqlCode = 0;

    *plErrLen = 0;              /* Initialisiere die fehlerlaengen */
    *plSqlErrLen = 0;


    if ( (NULL == pSession) || (NULL == pSession->packetOutData) )
    {
        rc = DBMAPI_INVSESSION_CN14;
    }
    else
    {
        pszData = (const _TCHAR *)pSession->packetOutData;

        if (0 == _tcsncmp(pszData, RPM_ERR_HEAD, _tcslen(RPM_ERR_HEAD)))    /* Fehler zurueckgegeben */
        {
            rc = DBMAPI_NOT_OK_CN14;

            /* Ok, wir haben einen fehler. Das naechste muss die fehlernummer sein, die
               noch genau 5 byte lang ist und hinter dem header liegt */
            *plErrCode = atol(&pszData[_tcslen(RPM_ERR_HEAD)]);
            /* Wird hier 0 zurueckgegeben, dann ist die fehlernummer nicht interpretierbar.
               Ein fehlertext wird dann aber, falls vorhanden, trotzdem eingetragen. */

            /* suche nach der fehlernummer das naechste newline und danach steht der fehlertext */
            pszErr = _tcschr(&pszData[_tcslen(RPM_ERR_HEAD)], _T('\n') );
            if (NULL != pszErr)
            {
                ++pszErr;           /* newline gefunden --> Um eine position weiter bewegen */
            }
            else                    /* kein newline gefunden, fehler aber offensichtlich gesetzt */
            {
                pszErr = &pszData[_tcslen(RPM_ERR_HEAD)];
            }

            if (0 == _tcsnicmp(pszErr, "sql error", 9) )    /* sql fehler */
            {
                pszSqlErr = pszErr + 10;
                *plSqlCode = atol(pszSqlErr);
                if (0 != *plSqlCode)
                {
                    pszSqlErr = _tcschr(pszSqlErr, _T('=') );
                    if (NULL != pszSqlErr)
                    {
                        ++pszSqlErr;    /* Ueberspringe das gleichheitszeichen */
                        while (0 != isspace(*pszSqlErr))
                        {
                            ++pszSqlErr;
                        }
                        *plSqlErrLen = lMyErrLen - (pszSqlErr - pszData);
                    }
                    else
                    {
                        pszSqlErr = "";     /* fehlerstringlaenge ist bereits mit init gesetzt*/
                    }
                }
                /* SQL fehler, kein allgemeiner fehler, fehlerstringlaenge ist bereits mit init gesetzt*/
                pszErr = "";
            }                            
            else
            {/* kein sql fehler und pszErr ist bereits gesetzt.*/

                *plErrLen = lMyErrLen - (pszErr - pszData);
            }
        }
        else    /* kein fehler, sondern ok gefunden, setze ptr hinter das new line nach ok, */
        {       /* denn dort faengt der text an, wenn einer vorhanden ist */
            pszErr = _tcschr(pszData, _T('\n'));
            
            if (pszErr != NULL)
            {
                pszErr++;
            }
            *plErrLen = lMyErrLen - (pszErr - pszData);
        }

        /* Setze zu guter letzt die rueckgabeptr richtig */
        if (NULL != ppszErrText)
        {
            *ppszErrText    = pszErr;
        }
        if (NULL != ppszSqlErrText)
        {
            *ppszSqlErrText = pszSqlErr;
        }
    }

    return rc;
}
/* end cn14analyzeRpmAnswer */
