/*******************************************************************************
  module:       vcn382.cpp
  ------------------------------------------------------------------------------
  responsible:  TiloH
  special area: cn layer
  description:  implementation of functions for handling Backint for Oracle
  last change:  2000-11-29 16:07
  version:      7.2.5
  see also:     
  ------------------------------------------------------------------------------
                         Copyright by SAP AG, 2000


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*******************************************************************************/

#include "hcn361.h"
#include "hcn382.h"
#include "hcn385.h"

#include <stdio.h>

static int cn382_WriteALineToBackintInput(tni34_ATextFile & BackintInput,
                                          const char *      TheLine,
                                          int &             FilesWrittenToInput)
{
    long BytesWritten;
    int rc=BackintInput.WriteLine(TheLine, BytesWritten);

    if(rc)
        FilesWrittenToInput++;
    else
        TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotWriteBackintInput));

    return rc;
}

static int cn382_CreateBackintInput(ni31_List<tcn381_InputEntry *> & TheFileList,
                                    tni34_ATextFile &                BackintInput,
                                    const char *                     Function,
                                    int                              ForPipes,
                                    int                              MaxNumberOfFilesPerBackintCall,
                                    int &                            FilesWrittenToInput)
{
    int rc=1;

    FilesWrittenToInput=0;
 
    ni31_ListElem<tcn381_InputEntry *> *hBackup;
    ni31_ListElem<char *> *hPart;
    const char * StageName;
    int          StageSuffix;
    const char * BID;
    char         Buf[1024];

    tcn381_InputEntry *IE; //a shortcut for hBackup->GiveInfo();

    for(hBackup=TheFileList.First(); rc && 0!=hBackup; hBackup=hBackup->Next()) //loop through file list
    {
        IE=hBackup->GiveInfo();
        StageName=IE->StageName;
        StageSuffix=0;

        if(!strcmp(Function, FunctionStringInquire_cn38)) // if it is for inquire
        {
            for(hPart=IE->ListOfPartialBIDs.First(); rc && 0!=hPart; hPart=hPart->Next(), StageSuffix++)
            {
                BID=hPart->GiveInfo();

                if(IE->IsPipe)
                    sprintf(Buf, "%s %s.%d", BID, StageName, StageSuffix);
                else
                    sprintf(Buf, "%s %s", BID, StageName);

                rc=cn382_WriteALineToBackintInput(BackintInput, Buf, FilesWrittenToInput);
            }

            //for handling all partial BIDs of a pipe 
            IE->FoundPartInThisBackintOutput=new int[IE->NumberOfParts];

            if(0==IE->FoundPartInThisBackintOutput)
            {
                rc=0;
                TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::OutOfMemory));
            }
            else
            {
                int i;
                for(i=0; i<IE->NumberOfParts; i++)
                    IE->FoundPartInThisBackintOutput[i]=0;
            }
        }
        else
            if(!strcmp(Function, FunctionStringBackup_cn38)) // if we are doing a backup
            {
                if(ForPipes)
                {
                    if(IE->IsPipe &&
                       tcn381_APipeToFileConverter::ConverterFinishedStage==IE->Converter->State &&
                       FilesWrittenToInput<MaxNumberOfFilesPerBackintCall)
                    {
                        sprintf(IE->Converter->EndOfStageName, "%d", int(IE->NumberOfParts));

                        if(!cn382_WriteALineToBackintInput(BackintInput, IE->Converter->CompleteStageName, FilesWrittenToInput))
                            rc=0; // error message is given by function
                        else
                            IE->Converter->State=tcn381_APipeToFileConverter::BackintIsWorking;
                    }
                }
                else
                {
                    // handle normal files
                    if(!IE->IsPipe)
                    {
                        rc=cn382_WriteALineToBackintInput(BackintInput,
                                                          IE->Name,
                                                          FilesWrittenToInput);
                    }
                }
            }
            else
                if(!strcmp(Function, FunctionStringRestore_cn38)) // if we are doing a restore
                {
                    if(ForPipes)
                    {
                        if(IE->IsPipe &&
                           tcn381_APipeToFileConverter::StageIsFree==IE->Converter->State &&
                           FilesWrittenToInput<MaxNumberOfFilesPerBackintCall &&
                           IE->RestorePart<IE->NumberOfParts)
                        {
                            sprintf(IE->Converter->EndOfStageName, "%d", int(IE->RestorePart));
                            
                            sprintf(Buf, "%s %s", IE->ListOfPartialBIDs[IE->RestorePart]->GiveInfo(),
                                                      IE->Converter->CompleteStageName);

                            if(0!=IE->Converter->EndOfCompleteDestinationStage)
                            {
                                sprintf(IE->Converter->EndOfCompleteDestinationStage, "%d", int(IE->RestorePart)); //we need this name for removal of the file
                                sprintf(cn36_PositionOfTerminatingZero(Buf), " %s", IE->Converter->DestinationStageDirectory); //use the destination feature of backint if needed
                            } 
                            
                            if(!cn382_WriteALineToBackintInput(BackintInput, Buf, FilesWrittenToInput))
                                rc=0; // error message is given by function
                            else
                                IE->Converter->State=tcn381_APipeToFileConverter::BackintIsWorking;
                        }
                    }
                    else
                    {
                        if(!IE->IsPipe)
                        {
                            if(0==IE->DestinationName)
                                sprintf(Buf, "%s %s", IE->BID, IE->Name);
                            else
                                sprintf(Buf, "%s %s %s", IE->BID, IE->Name, IE->DestinationName);
                            
                            rc=cn382_WriteALineToBackintInput(BackintInput, Buf, FilesWrittenToInput);
                        }
                    }
                }
                else
                {
                    rc=0;
                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::InternalError));
                }

        IE->FoundInThisBackintOutput=0; // entry not found in output so far
    }

    return rc;
}

static int cn382_CallBackint(const char *        PathOfBackint,
                             const char *        UID,
                             const char *        Function,
                             const char *        PathOfBackintParamFile,
                             const tni34_AFile & InFile,
                             const tni34_AFile & OutFile,
                             const tni34_AFile & ErrFile)
{
    int rc=1;

    char BackintCmdLine[2048];

    sprintf(BackintCmdLine, "%s -u %s -f %s -t file%s%s%s -i %s -c", PathOfBackint,
                                                                   UID,
                                                                   Function,
															       '\0'==PathOfBackintParamFile[0]?"":" -p \"",
                                                                   PathOfBackintParamFile,
                                                                   '\0'==PathOfBackintParamFile[0]?"":"\"",
                                                                   InFile.GiveName());

    tni33_AProcess BackIntProc(BackintCmdLine, 0, &OutFile, 0, &ErrFile, 0);

    if(!BackIntProc.WasStarted())
    {
        rc=0;
        TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotStartBackint));
    }
    else
    {   
	    if(0!=BackIntProc.GiveReturnCode() && 1!=BackIntProc.GiveReturnCode())
        {
            rc=0;
            TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::BackintWasUnsuccessful));
        }
    }
    
    return rc;
}

static int cn382_ReadBIDs(ni31_List<tcn381_InputEntry *> & TheFileList,
                          int                              ForPipes,
                          const char *                     FunctionString,
                          tni34_ATextFile &                BackintOutput)
{
    int rc=1;

    const char *Line;
    ni31_ListElem<tcn381_InputEntry *> *h;
    tcn381_InputEntry *IE;

    const char *KeyWordForFunction;

    // determine the right keyword for the function
    if(!strcmp(FunctionString, FunctionStringBackup_cn38))
        KeyWordForFunction="#SAVED";
    else
        if(!strcmp(FunctionString, FunctionStringInquire_cn38))
            KeyWordForFunction="#BACKUP";
        else
            if(!strcmp(FunctionString, FunctionStringRestore_cn38))
                KeyWordForFunction="#RESTORED";

    //for every line
    while(BackintOutput.ReadLine(Line) && Line!=0)
    {
        const char *KeyWord=cn36_FirstNonWhiteSpaceOf(Line);
        const char *EndOfKeyWord=cn36_FirstWhiteSpaceOf(KeyWord);

        if(!cn36_StrStrNCmp(KeyWordForFunction, KeyWord, size_t(EndOfKeyWord-KeyWord)))
        {
            const char *BID=cn36_FirstNonWhiteSpaceOf(EndOfKeyWord);
            const char *EndOfBID=cn36_FirstWhiteSpaceOf(BID);
            const char *FileName=cn36_FirstNonWhiteSpaceOf(EndOfBID);
            const char *EndOfFileName=cn36_FirstWhiteSpaceOf(FileName);

            if(EndOfBID[0]!='\0' && FileName[0]!='\0') //found BID and Name behind the keyword
            {
                if(EndOfBID-BID>16) //is the BID a BID? this is needed for the StrNCpy calls
                {
                    rc=0;
                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::SyntaxErrorInBackintOutput));
                }
                else
                {
                    const char *CurrentName;

                    for(h=TheFileList.First(); rc && h!=0; h=h->Next()) // for all files from the list
                    {
                        IE=h->GiveInfo(); // just a shortcut to the current tcn381_InputEntry

                        if(!strcmp(FunctionString, FunctionStringBackup_cn38) ||
                           !strcmp(FunctionString, FunctionStringRestore_cn38)) // if we do a backup or restore
                        {
                            if(IE->IsPipe)
                                CurrentName=IE->Converter->CompleteStageName; //if it's a pipe, we saved only a part with backint
                            else
                                CurrentName=IE->Name; //if it's a file, we saved it at once with backint

                            if(ForPipes==IE->IsPipe &&                                            //if IE is from the rigth type
                               !cn36_StrStrNCmp(CurrentName, FileName, EndOfFileName-FileName))//the name we have to look for is equal the found name
                            {
                                if(0!=IE->FoundInThisBackintOutput ||         //was this file already found?
                                   (IE->IsPipe &&
                                    tcn381_APipeToFileConverter::BackintIsWorking!=IE->Converter->State)) //the file should not be saved by backint
                                {
                                    rc=0;
                                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::BackintProcessedWrongFile));
                                }
                                else
                                {
                                    if(!strcmp(FunctionString, FunctionStringBackup_cn38))
                                        rc=IE->SetBID(BID, EndOfBID-BID);

                                    IE->FoundInThisBackintOutput=1;
                                }
                            }
                        }
                        else
                            if(!strcmp(FunctionString, FunctionStringInquire_cn38))
                            {
                                if(IE->IsPipe)
                                {
                                    int SuffixNumber;

                                    if(!strncmp(FileName, IE->StageName, strlen(IE->StageName)) &&
                                       '.'==FileName[strlen(IE->StageName)] &&
                                       cn36_StrNToInt(SuffixNumber, FileName+strlen(IE->StageName)+1, EndOfFileName-FileName-strlen(IE->StageName)-1) &&
                                       0<=SuffixNumber &&
                                       IE->NumberOfParts>SuffixNumber)
                                    {
                                        char SuffixAsString[21];
                                        sprintf(SuffixAsString, "%d", SuffixNumber);

                                        if(!cn36_StrStrNCmp(SuffixAsString, FileName+strlen(IE->StageName)+1, EndOfFileName-FileName-strlen(IE->StageName)-1) &&
                                           !cn36_StrStrNCmp(IE->ListOfPartialBIDs[SuffixNumber]->GiveInfo(), BID, EndOfBID-BID))
                                        {
                                            IE->FoundPartInThisBackintOutput[SuffixNumber]=1;//test BID !!!!
                                        }
                                    }
                                }
                                else
                                {
                                    if(!cn36_StrStrNCmp(IE->StageName, FileName, EndOfFileName-FileName) &&
                                       !cn36_StrStrNCmp(IE->BID, BID, EndOfBID-BID))
                                    {
                                        IE->FoundInThisBackintOutput=1;
                                    }
                                }
                            }
                            else
                            {
                                rc=0;
                                TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::InternalError));
                            }
                    }
                }
            }
        }
    }

    if(!strcmp(FunctionString, FunctionStringBackup_cn38) ||
       !strcmp(FunctionString, FunctionStringRestore_cn38))
    {
        //check that every file was processed successfully
        for(h=TheFileList.First(); h!=0; h=h->Next())
        {
            IE=h->GiveInfo();

            if((0==ForPipes && !IE->IsPipe) ||
               (1==ForPipes && IE->IsPipe && tcn381_APipeToFileConverter::BackintIsWorking==IE->Converter->State))
            {
                if(!IE->FoundInThisBackintOutput)
                {
                    IE->ProcessedSuccessful=0;
                    rc=0;
                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::BackintProcessedUnsuccessful));
                }
            }
        }
    }
    else
        if(!strcmp(FunctionString, FunctionStringInquire_cn38))
        {
            int FoundAllParts,
                i;

            for(h=TheFileList.First(); rc && h!=0; h=h->Next()) // for all files from the list
                if(h->GiveInfo()->IsPipe)
                {
                    for(FoundAllParts=1, i=0; FoundAllParts && i<h->GiveInfo()->NumberOfParts; i++)
                        if(!h->GiveInfo()->FoundPartInThisBackintOutput[i])
                            FoundAllParts=0;

                    if(FoundAllParts)
                        h->GiveInfo()->FoundInThisBackintOutput=1;
                }
        }

    return rc;
}

int cn382_UseBackint(ni31_List<tcn381_InputEntry *> & TheInputFileList,
                     const char *                     PathOfBackint,
                     const char *                     UserID,
                     const char *                     FunctionString,
                     int                              ForPipes,
                     int                              MaxNumberOfFilesPerBackintCall,
                     const char *                     PathOfBackintParamFile,
                     const char *                     InFileName,
                     const char *                     OutFileName,
                     const char *                     ErrFileName)
{
    int rc=1;

    //create/truncate the input file
    tni34_ATextFile * BackintInput;
    
    if('\0'==InFileName[0])
        BackintInput=new tni34_ATextFile(tni34_AFile::ToDiskTruncate);
    else
        BackintInput=new tni34_ATextFile(InFileName, tni34_AFile::ToDiskTruncate);

    if(0==BackintInput)
    {
        rc=0;
        TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::OutOfMemory));
    }
    else
    {
        if(BackintInput->IsGood())
        {
            //write all necessary data to the input file
            int NumberOfFilesWrittenToInput;

            if(cn382_CreateBackintInput(TheInputFileList,
                                        *BackintInput,
                                        FunctionString,
                                        ForPipes,
                                        MaxNumberOfFilesPerBackintCall,
                                        NumberOfFilesWrittenToInput))
            {
                if(!BackintInput->Close())
                {
                    rc=0;
                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotCloseBackintInput));
                }
                else
                    if(0<NumberOfFilesWrittenToInput) //there is something to do for backint
                    {
                        //create/truncate the output file
                        tni34_ATextFile * BackintOutput=0;
                        tni34_ATextFile * BackintError=0;
                
                        if('\0'==OutFileName[0])
                            BackintOutput=new tni34_ATextFile(tni34_AFile::FromDisk);
                        else
                            BackintOutput=new tni34_ATextFile(OutFileName, tni34_AFile::FromDisk);

                        if('\0'==ErrFileName[0])
                            BackintError=new tni34_ATextFile(tni34_AFile::FromDisk);
                        else
                            BackintError=new tni34_ATextFile(ErrFileName, tni34_AFile::FromDisk);

                        if(0==BackintOutput ||
                           !BackintOutput->IsGood() ||
                           0==BackintError ||
                           !BackintError->IsGood())
                        {
                            rc=0;
                            TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::OutOfMemory));
                        }
                        else
                        {
                            if(!BackintOutput->Close())
                            {
                                rc=0;
                                TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotCreateBackintOutput));
                            }

                            if(!BackintError->Close())
                            {
                                rc=0;
                                TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotCreateBackintErrorOutput));
                            }

                            if(rc)
                            {
                                // now execute backint
                                if(!cn382_CallBackint(PathOfBackint,
                                                      UserID,
                                                      FunctionString,
                                                      PathOfBackintParamFile,
                                                      *BackintInput,//BackintInput->GiveName(),
                                                      *BackintOutput,//BackintOutput->GiveName(),
                                                      *BackintError))
                                {
                                    rc=0; // error message is given by function
                                }
                                
                                //analyse the results of the backint call (if there are any)
                                if(!BackintOutput->Exists())
                                {
                                    rc=0;
                                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::BackintOutputVanished));
                                }
                                else
                                    if(!BackintOutput->Open())
                                    {
                                        rc=0;
                                        TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotOpenBackintOutput));
                                    }

                                if(rc)
                                {
                                    rc=cn382_ReadBIDs(TheInputFileList,
                                                      ForPipes,
                                                      FunctionString,
                                                      *BackintOutput);

                                    if(!BackintOutput->Close())
                                    {
                                        rc=0;
                                        TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotCloseBackintOutput));
                                    }
                                }

                                if(BackintOutput->Exists() && !BackintOutput->Remove()) //clean up the output
                                {
                                    rc=0;
                                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotRemoveBackintOutput));
                                }

                                if(BackintError->Exists() && !BackintError->Remove()) //clean up the error output
                                {
                                    rc=0;
                                    TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotRemoveBackintErrorOutput));
                                }
                            }
                        }

                        if(0!=BackintOutput)
                            delete BackintOutput;

                        if(0!=BackintError)
                            delete BackintError;
                    }
            }
            else
                rc=0;
  
            if(BackintInput->Exists() && !BackintInput->Remove()) //clean up the input
            {
                rc=0;
                TheBackint2ErrorHandler_cn385->Report(tcn385_Backint2Error(tcn385_Backint2Error::CouldNotRemoveBackintInput));
            }
        }
        else
            rc=0;

        delete BackintInput;
    }

    return rc;
}
