/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/

/*
  -----------------------------------------------------------------------------

  module: vcn40.cpp

  -----------------------------------------------------------------------------

  responsible:  BerndV

  special area: DBMServer

  description:  DBMServer start-stop / system / registry access

  -----------------------------------------------------------------------------
*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#include "vsp001.h"
#include "gsp09.h"
#include "geo28.h"
#include "geo43.h"

#include "heo02.h"
#include "heo03.h"
#include "heo06.h"
#include "heo11.h"
#include "hsp100.h"

#include "hin101.h"
#include "hin110.h"

#include "hcn20.h"
#include "hcn30.h"
#include "hcn31.h"
#include "hcn42.h"
#include "hcn50.h"
#include "hcn51.h"
#include "hcn80.h"
#include "hcn82.h"
#include "hcn90.h"

#include "hcn40.h"

extern "C" int chdir(const char * szDir);

/*
  -----------------------------------------------------------------------------
  specification private macros
  -----------------------------------------------------------------------------
*/
#define SYS_SYSTEM_ERR        -1
#define SYS_MKDIR_ERR         -1
#define SYS_MKDIR_OK          0
#define CHAR_SLASH            _T('/')
#define CHAR_BACKSLASH        _T('\\')
#define CHAR_DRIVE            _T(':')
#define TMP_ENV               _T("TMP")
#define TMP_LOCDIR            _T(".")
#define TMP_EXT               _T("cst")
#define XLOAD_PRT             _T("load.prt")
#define XLOAD_PROT            _T("load.prot")
#define REPMAN_PRT            _T("repserver.log")
#define UTIL_PRT              _T("util.prt")

#define OPT_OPT_AUTO          _T("-auto")
#define OPTS_OPT_AUTO         _T("-a")

#define OPT_SPEED_FAST        _T("-fast")
#define OPTS_SPEED_FAST       _T("-f")

#define OPT_SPEED_QUICK       _T("-quick")
#define OPTS_SPEED_QUICK      _T("-q")

#define OPT_SPEED_SLOW        _T("-slow")
#define OPTS_SPEED_SLOW       _T("-s")

#define OPT_STOP_IMMEDIATE    _T("-immediate")
#define OPTS_STOP_IMMEDIATE   _T("-i")

#define OPTS_STOP_DUMP        _T("-d")
#define OPT_STOP_DUMP         _T("-dump")

#define OPT_RESTART_UNTIL     _T("-until")
#define OPTS_RESTART_UNTIL    _T("-u")

#define TXT_RESTART           _T("RESTART")
#define TXT_RESTART_UNTIL     _T("RESTART UNTIL '%s' '%s'")

#define OPT_LOAD_DBA          _T("-u")
#define OPT_LOAD_DOM          _T("-ud")
#define LOAD_FILE             _T("lsystab.ins")
#define PYTHON_FILE           _T("lsystab.py")
#define LOAD_OPT              _T("en ENG")
#define R3_FILE               _T("CNR3DATA.ins")
#define PYR3_FILE             _T("CNR3DATA.py")

#define LOAD_INIT             _T("INIT AFTER LOAD")

#define TXT_CMD_SHOW          _T("SHOW")
#ifdef _WIN32
  #define TXT_CMD_XCONS         _T("pgm\\cons")
#else
  #define TXT_CMD_XCONS         _T("pgm/console")
#endif
#define TXT_CMD_HELP          _T("help")

#define TXT_CMD_STATE         _T("State")
/*
#define TXTU_STATE_WARM       _T("WARM")
#define TXTL_STATE_WARM       _T("warm")

#define TXTU_STATE_SHUTDOWN   _T("SHUTDOWN")
#define TXTL_STATE_SHUTDOWN   _T("shutdown")

#define TXTU_STATE_COLD       _T("COLD")
#define TXTL_STATE_COLD       _T("cold")

#define TXTU_STATE_OFFLINE    _T("OFFLINE")
#define TXTL_STATE_OFFLINE    _T("offline")

#define TXTU_STATE_RUIN       _T("STOPPED INCORRECTLY")
#define TXTL_STATE_RUIN       _T("stopped incorrectly")

#define TXTU_STATE_UNKNOWN    _T("UNKNOWN")
#define TXTL_STATE_UNKNOWN    _T("unknown")

#define TXTU_STATE_RUNNING    _T("RUNNING")
#define TXTL_STATE_RUNNING    _T("running")
*/
#define TXT_CMD_SPEED         _T("Speed")

#define TXTU_SPEED_FAST       _T("FAST")
#define TXTL_SPEED_FAST       _T("fast")

#define TXTU_SPEED_QUICK      _T("QUICK")
#define TXTL_SPEED_QUICK      _T("quick")

#define TXTU_SPEED_SLOW       _T("SLOW")
#define TXTL_SPEED_SLOW       _T("slow")

#define TXTU_SPEED_UNKNOWN    _T("UNKNOWN")
#define TXTL_SPEED_UNKNOWN    _T("unknown")

#define KEYVER_VERSION        _T("VERSION")
#define KEYVER_BUILD          _T("BUILD")
#define KEYVER_OS             _T("OS")
#define KEYVER_DBROOT         _T("INSTROOT")
#define KEYVER_LOGON          _T("LOGON")
#define KEYVER_CODE           _T("CODE")
#define KEYVER_SWAP           _T("SWAP")

#define SWAP_NORMAL           _T("normal")
#define SWAP_FULL             _T("full")
#define SWAP_PART             _T("part")

#define CODE_ASCII            _T("ASCII")
#define CODE_EBCDIC           _T("EBCDIC")

#define TXT_CMD_LCINIT        _T("sap%clcinit %s %s")

#define KEYDROP_WITHOUTFILES  _T("WITHOUTFILES")

#define RTE_INDPROG           _T("IndepProgPath")
#define RTE_INDDATA           _T("IndepDataPath")

#define OPTS_INST_VERSION     _T("-v")
#define OPTS_INST_KEY         _T("-k")
#define OPTS_INST_CLIENT      _T("-c")
#define OPTS_ENUM_SRV         _T("-s")

#define TEST_OPT_WAIT         _T("-w")

#define SQL_CHECKSYSDBA       _T("SELECT USERNAME FROM SYSDD.USERS WHERE SYSDBA = USER")
/*
  -----------------------------------------------------------------------------
  specification private functions
  -----------------------------------------------------------------------------
*/

static tcn00_Error cn40_DBDrop
      (       tsp00_DbNamec   szDbName,
        const tsp00_Pathc     szDbRoot,
        const bool            bWithFiles,
        char                * replyData,
        int                 * replyLen,
        const int             replyLenMax);

static tcn00_Error cn40_ExecCommand
      ( char          * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const char    * szProgram,
        const char    * szOutput);

static tcn00_Error cn40_ExecXLoad
      ( VControlDataT * vcontrol,
        const char    * szLoadFile,
        const char    * szUser,
        const char    * szPassword,
        const char    * szOptions,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax);

static tcn00_Error cn40_ExecPythonLoad
      ( VControlDataT * vcontrol,
        const char    * szLoadFile,
        const char    * szUser,
        const char    * szPassword,
        const char    * szOptions,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax);

static tcn00_Error cn40_OfflineToCold
      ( const tsp00_DbNamec   szDBName,
        const tsp00_Pathc     szDBRoot,
        const tsp9_pgm_kind   pgmKind,
        char                * replyData,
        int                 * replyLen);

static tcn00_Error cn40_ColdToWarm
      ( const tsp00_DbNamec & szDBName,
        const char          * pRestart,
        char                * replyData,
        int                 * replyLen);

static tcn00_Error cn40_WarmToCold
      ( const tsp00_DbNamec   szDBName,
        const bool            bImmediate,
        char                * replyData,
        int                 * replyLen);

static tcn00_Error cn40_ColdToOffline
      ( const tsp00_DbNamec   szDBName,
        const tsp00_Pathc     szDBRoot,
        const bool            bDump,
        char                * replyData,
        int                 * replyLen);

static tcn00_Error cn40_CheckVersion
      ( const tsp00_DbNamec   szDBName);

static tcn00_Error cn40_StartStopOptions
      ( const _TCHAR  * szParameters,
        tsp9_pgm_kind & pgmKind,
        bool          & bDump,
        bool          & bImmediate,
        _TCHAR        * pRestart);

static tcn00_Error cn40_ChangeDBState
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax,
        tcn00_DBState   targetState,
        bool            bStop = false);

static bool cn40_CheckInstKey
      ( const tsp09_RteInstallationInfoNew  & relInfoIn );

static bool cn40_RegisterServiceDB
      ( const tsp09_RteInstallationInfoNew  & relInfoIn); 

/*
  -----------------------------------------------------------------------------
  implementation of public functions
  -----------------------------------------------------------------------------
*/

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40ExecCommand
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ExecCommand
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  char            szCommand [PARAMETER_MAXLEN_CN90];

  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if (_tcslen(command->args) > 0) {
      _tcscpy(szCommand, command->args);
      nFuncReturn = cn40_ExecCommand(replyData, replyLen, replyLenMax, szCommand, szCommand);
    } else {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  return nFuncReturn;
} /* end cn40ExecCommand */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40ExecLcInit
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ExecLcInit
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  char            szCommand      [PARAMETER_MAXLEN_CN90] = "";

  _stprintf(szCommand, TXT_CMD_LCINIT, PATH_DELIMITER_CN90, vcontrol->dbname.asCharp(), command->args);

  nFuncReturn = cn40_ExecCommand(replyData, replyLen, replyLenMax, szCommand, szCommand);

  if (nFuncReturn == OK_CN00) {
    // check the lcinit output
    char * pTemp  = new char[*replyLen + 10];
    char * pData  = NULL;

    if (pTemp != NULL) {
      strcpy(pTemp, replyData);

      // find Nutzlast
      pData = strchr (pTemp, '\n');
      if (pData != NULL) {
        ++pData;
        pData = strchr (pData, '\n');
      } // end if
      if (pData != NULL) {
        ++pData;
        if (strstr(pData, "\nERROR") != NULL) {
          // fake answer
          nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_EXECUTE_CN00);
          pTemp = replyData + strlen(replyData);
          strcpy(pTemp, pData);
          *replyLen = (int) strlen(replyData);
        } // end if

      } // end if

      delete [] pTemp;
    } // end if

  } // end if

  return nFuncReturn;

} /* end cn40ExecLcInit */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40ExecLoad
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ExecLoad
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  _TCHAR          szUsrPwd    [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR          szFile      [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR          szOptions   [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR          szPassword  [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR        * pPassword = NULL;

  cn90GetToken(command->args, szUsrPwd,  1);
  cn90GetToken(command->args, szFile,    2);
  cn90GetToken(command->args, szOptions, 3);
  cn90StripQuotes(szOptions);
  
  pPassword = _tcschr(szUsrPwd, PWD_SEPARATOR_CN00);
  if (pPassword != NULL) {
    *pPassword = CHAR_STRINGTERM_CN90;
    pPassword++;
    cn90Uncrypt(pPassword, false);
  } else {
    pPassword = &szPassword[0];
    _tcscpy(szPassword, szFile);
    _tcscpy(szFile, szOptions);
    cn90GetToken(command->args, szOptions, 4);
    cn90StripQuotes(szOptions);
    cn90Uncrypt(pPassword, false);
  } // end if

  if (_tcslen(szUsrPwd)  == 0 ||
      _tcslen(szFile)    == 0 ||
      pPassword          == NULL ) {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  } else {
    nFuncReturn = cn40_ExecPythonLoad(vcontrol,
                                      szFile,
                                      szUsrPwd,
                                      pPassword,
                                      szOptions,
                                      replyData,
                                      replyLen,
                                      replyLenMax);
  } // end if

  return nFuncReturn;
} /* end cn40ExecLoad */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40ExecXLoad
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ExecXLoad
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  _TCHAR          szUsrPwd    [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR          szFile      [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR          szOptions   [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR          szPassword  [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR        * pPassword = NULL;

  cn90GetToken(command->args, szUsrPwd,  1);
  cn90GetToken(command->args, szFile,    2);
  cn90GetToken(command->args, szOptions, 3);
  cn90StripQuotes(szOptions);
  
  pPassword = _tcschr(szUsrPwd, PWD_SEPARATOR_CN00);
  if (pPassword != NULL) {
    *pPassword = CHAR_STRINGTERM_CN90;
    pPassword++;
    cn90Uncrypt(pPassword, false);
  } else {
    pPassword = &szPassword[0];
    _tcscpy(szPassword, szFile);
    _tcscpy(szFile, szOptions);
    cn90GetToken(command->args, szOptions, 4);
    cn90StripQuotes(szOptions);
    cn90Uncrypt(pPassword, false);
  } // end if

  if (_tcslen(szUsrPwd)  == 0 ||
      _tcslen(szFile)    == 0 ||
      pPassword          == NULL ) {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  } else {
    nFuncReturn = cn40_ExecXLoad(vcontrol,
                                 szFile,
                                 szUsrPwd,
                                 pPassword,
                                 szOptions,
                                 replyData,
                                 replyLen,
                                 replyLenMax);
  } // end if

  return nFuncReturn;
} /* end cn40ExecXLoad */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40LoadSysTab
 * ------------------------------------------------------------------
 */
tcn00_Error cn40LoadSysTab
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  FUNCTION_DBG_MCN00_1(_T("cn40LoadSysTab"));

  tcn00_Error          nFuncReturn = OK_CN00;
  _TCHAR               szToken   [PARAMETER_MAXLEN_CN90];
  _TCHAR               szOptions [PARAMETER_MAXLEN_CN90];
  int                  nToken = 1;
  tsp00_UserNamec      szUsername;
  tsp00_Namec          szPassword;
  tsp00_Namec          szDomPwd;
  _TCHAR             * pPassword;
  tin101_XDBOperation  aXDB(vcontrol->dbname,
                            vcontrol->pCurrentUser->getUserName(),
                            vcontrol->pCurrentUser->getClearMasterPwd());
  tsp00_Pathc          aLogFile;

  aXDB.start(XDBOP_LOADSYSTAB_IN101, XDBCO_DBM_IN101, aLogFile);

  if (cn90DBState(vcontrol->dbname) != STATE_WARM_CN00) {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_DBNORUN_CN00);
  } // end if

  szUsername.Init();
  szPassword.Init();
  szDomPwd.Init();

  while (cn90GetToken(command->args, szToken, nToken) && nFuncReturn == OK_CN00) {

    if (_tcscmp(szToken, OPT_LOAD_DBA) == 0) {
      // DBA found
      nToken++;
      if (cn90GetToken(command->args, szToken, nToken)) {
        /* search for password */
        pPassword = strchr (szToken, PWD_SEPARATOR_CN00);
        if (pPassword != NULL) {
          *pPassword = CHAR_STRINGTERM_CN90;
          ++pPassword;
          cn90Uncrypt(pPassword, false);
          szPassword.rawAssign(pPassword);
        } /* end if */
        szUsername.rawAssign(szToken);
      } else {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } // end if

    } else if (_tcscmp(szToken, OPT_LOAD_DOM) == 0) {

      // DOMAIN pwd found
      nToken++;
      if (cn90GetToken(command->args, szToken, nToken)) {
        cn90Uncrypt(szToken, false);
        szDomPwd.rawAssign(szToken);
      } else {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } // end if

    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if

    nToken++;
  } // end while

  // dba management
  if (nFuncReturn == OK_CN00) {
    if (szUsername.length() == 0) {
      // load dba user
      cn51DBMConfigValue  cfgDBAUsr (vcontrol->dbname, CFG_DBAUSR_CN51);
      cn50DBMUser         usrDBA    (vcontrol->dbname, (tsp00_Namec) cfgDBAUsr );

      szUsername = usrDBA.getUserName();
      szPassword = usrDBA.getClearMasterPwd();

      // check sysdba property
      nFuncReturn = ERR_DBAUNKNOWN_CN00;
      tin01_sql_session aSession;
      if (cn80ConnectSQL  (usrDBA, &aSession, replyData, replyLen) == OK_CN00) {
        tin110_SQLResultTable aTable(&aSession, SQL_CHECKSYSDBA);
        if (aTable.sqlRc() == cin01_db_ok) {
          nFuncReturn = OK_CN00;
        } // end if
        cn80ReleaseSQL  ( &aSession, false);
      } // end if
      nFuncReturn = cn90AnswerIError(replyData, replyLen, nFuncReturn);

    } else {
      cn50DBMUser  usrTempDBA;
      
      usrTempDBA.setDBName   (vcontrol->dbname)
                .setUserName (szUsername)
                .setMasterPwd(szPassword);

      // check sysdba property
      nFuncReturn = ERR_DBAWRONG_CN00;
      tin01_sql_session aSession;
      if (cn80ConnectSQL  (usrTempDBA, &aSession, replyData, replyLen) == OK_CN00) {
        tin110_SQLResultTable aTable(&aSession, SQL_CHECKSYSDBA);
        if (aTable.sqlRc() == cin01_db_ok) {
          nFuncReturn = OK_CN00;
        } // end if
        cn80ReleaseSQL  ( &aSession, false);
      } // end if
      nFuncReturn = cn90AnswerIError(replyData, replyLen, nFuncReturn);

      if (nFuncReturn == OK_CN00) {
        // create dba user
        cn50DBMUser         usrOldDBA( vcontrol->dbname, szUsername);
        cn50DBMUser         usrNewDBA;

        usrNewDBA.setDBName(vcontrol->dbname)
                 .setUserName ( szUsername)
                 .setSDBAUser ( )
                 .setSQLUser  ( )
                 .setMasterPwd( szPassword)
                 .setDBMUser      ( true )
                 .setServerRights ( DBMMaskAll_CN50, DBMUserDefault_CN50 )
                 .setGUIRights    ( DBMMaskAll_CN50, DBMUserDefault_CN50 );

        if (usrOldDBA.existsOnDisk()) {
         usrNewDBA.setDBMUser      ( usrOldDBA.isDBMUser())
                  .setServerRights ( DBMMaskAll_CN50, usrOldDBA.getServerRights() )
                  .setGUIRights    ( DBMMaskAll_CN50, usrOldDBA.getGUIRights() );
        } // end if

        // delete old dba user
        usrOldDBA.setExistsInDB(false);
        usrOldDBA.deleteUsr();

        // save new dba user
        nFuncReturn = usrNewDBA.save();

        // save sysdba-configvalue
        if (nFuncReturn == OK_CN00) {
          cn51DBMConfigValue  cfgDBAUsr(vcontrol->dbname, CFG_DBAUSR_CN51);
          cfgDBAUsr = usrNewDBA.getUserName();
        } else {
          cn90AnswerIError(replyData, replyLen, nFuncReturn);
        } // end if
      } // end if
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00 && szUsername.length() == 0) {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_DBAUNKNOWN_CN00);
  } // end if

  // domain management
  cn50DBMUser       usrDom( vcontrol->dbname, USERNAME_DOMAIN_CN50);
  tin01_sql_session aSession;

  if (nFuncReturn == OK_CN00) {

    if (szDomPwd.length() == 0) {
      szDomPwd = usrDom.getClearMasterPwd();
    } else {
      if (usrDom.existsOnDisk()) {

        usrDom.setMasterPwd( szDomPwd);

        // check new password
        if (cn80ConnectSQL  (usrDom, &aSession, replyData, replyLen) != OK_CN00) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_DOMPWRONG_CN00_1);
          nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
        } else {
          cn80ReleaseSQL  ( &aSession, false);
          // save new password
          nFuncReturn = usrDom.save();
          if (nFuncReturn != OK_CN00) {
            cn90AnswerIError(replyData, replyLen, nFuncReturn);
          } // end if
        } // end if

      } else {
        usrDom.setSQLUser()
              .setMasterPwd( szDomPwd)
              .setSQLUsermode(cn50DBMUser::SQLUsrModeDBA);

        nFuncReturn = usrDom.save();
        if (nFuncReturn != OK_CN00) {
          cn90AnswerIError(replyData, replyLen, nFuncReturn);
        } // end if
      } // end if

    } // end if
  } // end if

  if (nFuncReturn == OK_CN00 && szDomPwd.length() == 0) {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_DOMPUNKNOWN_CN00);
  } else {
    // check domain user
    if (cn80ConnectSQL  (usrDom, &aSession, replyData, replyLen) != OK_CN00) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_DOMPUNKNOWN_CN00);
    } else {
      cn80ReleaseSQL  ( &aSession, false);
    } // end if
  } // end if

  aXDB.run(XDB_RC_CN00(nFuncReturn), "");

  if (nFuncReturn == OK_CN00) {
// *PY*   cn51DBMConfigValue  cfgXLoadUse(vcontrol->dbname, CFG_XLOAD_CN51);
    cn51DBMConfigValue  cfgPythonUse(vcontrol->dbname, CFG_PYTHON_CN51);

// *PY*   if (cfgXLoadUse == 1) {
    if (cfgPythonUse != 1) {
      _stprintf(szOptions, _T("%s %s"), LOAD_OPT, (_TCHAR *) szDomPwd);
      nFuncReturn = cn40_ExecXLoad(vcontrol,
                                   LOAD_FILE,
                                   (_TCHAR *) szUsername,
                                   (_TCHAR *) szPassword,
                                   (_TCHAR *) szOptions,
                                   replyData,
                                   replyLen,
                                   replyLenMax);
    } else {
      _stprintf(szOptions, _T("-dompwd %s %s"), (_TCHAR *) szDomPwd, LOAD_OPT);
      nFuncReturn = cn40_ExecPythonLoad(vcontrol,
                                        PYTHON_FILE,
                                        (_TCHAR *) szUsername,
                                        (_TCHAR *) szPassword,
                                        (_TCHAR *) szOptions,
                                        replyData,
                                        replyLen,
                                        replyLenMax);
    } // end if

    if (nFuncReturn == OK_CN00 || nFuncReturn == ERR_EXECUTE_CN00) {
// *PY*      if (cfgXLoadUse == 1) {
      if (cfgPythonUse != 1) {
       cn42BackupFile(vcontrol->dbname, FGET_LOADPRT_CN42);
      } else  {
        cn42BackupFile(vcontrol->dbname, FGET_REPMAN_CN42);
      } // end if

      cn51DBMConfigValue  cfgColdUsr (vcontrol->dbname, CFG_COLDUSR_CN51);
      cn50DBMUser         usrCold    ( vcontrol->dbname,
                                     (tsp00_Namec) cfgColdUsr );
      cn80ExecuteUtil(usrCold, LOAD_INIT);
      cn20SaveParamsInfoToDB(vcontrol->dbname);
    } // end if

  } // end if

  aXDB.complete(XDB_RC_CN00(nFuncReturn));

  return nFuncReturn;
} // end cn40LoadSysTab

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40LoadR3Tab
 * ------------------------------------------------------------------
 */
tcn00_Error cn40LoadR3Tab
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error          nFuncReturn = OK_CN00;
  _TCHAR               szUsername [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR               szPassword [PARAMETER_MAXLEN_CN90] = "";
  tsp00_C64c           szValue;
  _TCHAR               szOptions [PARAMETER_MAXLEN_CN90];

  // dba management
  cn51DBMConfigValue  cfgDBAUsr (vcontrol->dbname, CFG_DBAUSR_CN51);
  cn50DBMUser         usrDBA    ( vcontrol->dbname, (tsp00_Namec) cfgDBAUsr );

  if (!usrDBA.existsOnDisk()) {
    nFuncReturn = ERR_USRREAD_CN00;
  } else {
    _tcscpy(szUsername, usrDBA.getUserName());
    cn90CalculateUncrypt(usrDBA.getMasterPwd(), szPassword, false);
  } // end if

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn20XParamGetValue(vcontrol->dbname,
                                     XP_ID_RUNDIRECTORY_CN00,
                                     szValue);
  } // end if

  if (nFuncReturn == OK_CN00) {
// *PY*    cn51DBMConfigValue  cfgXLoadUse(vcontrol->dbname, CFG_XLOAD_CN51);
    cn51DBMConfigValue  cfgPythonUse(vcontrol->dbname, CFG_PYTHON_CN51);

// *PY*    if (cfgXLoadUse == 1) {
    if (cfgPythonUse != 1) {
      strcpy(szOptions, szValue);
      strcat(szOptions, "/");
      cn90ChangePathDelimiter(szOptions);
    
      nFuncReturn = cn40_ExecXLoad(vcontrol,
                                   R3_FILE,
                                   (_TCHAR *) szUsername,
                                   (_TCHAR *) szPassword,
                                   (_TCHAR *) szOptions,
                                   replyData,
                                   replyLen,
                                   replyLenMax);
    } else {
      _stprintf(szOptions, _T("-fdir %s"), szValue.asCharp());
      nFuncReturn = cn40_ExecPythonLoad(vcontrol,
                                        PYR3_FILE,
                                        (_TCHAR *) szUsername,
                                        (_TCHAR *) szPassword,
                                        (_TCHAR *) szOptions,
                                        replyData,
                                        replyLen,
                                        replyLenMax);
    } // end if



    cn20SaveParamsInfoToDB(vcontrol->dbname);

  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  return nFuncReturn;
} // end cn40LoadR3Tab

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40ExecXKernprot
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ExecXKernprot
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  tsp00_C256      szDir;
  char            szOptions   [PARAMETER_MAXLEN_CN90] = "";
  char            szCommand   [PARAMETER_MAXLEN_CN90] = "";

  if (!cn90GetToken(command->args, szOptions, 1)) {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  }

  if (nFuncReturn == OK_CN00) {

    cn90GetVersionBinPath(szDir);

#ifdef _WIN32
    sprintf(szCommand, "\"\"%sxkernprot\" -d %s %s\"", (char *) szDir, &vcontrol->dbname[0], szOptions);
#else
    sprintf(szCommand, "%skernprot -d %s %s", (char *) szDir, &vcontrol->dbname[0], szOptions);
#endif

    nFuncReturn = cn40_ExecCommand(replyData, replyLen, replyLenMax, szCommand, szCommand);
  } // end if

  return nFuncReturn;
} /* end cn40ExecXKernprot */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40GetVersion
 * ------------------------------------------------------------------
 */
tcn00_Error cn40GetVersion
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp_os              osNumber;
  char                osString [100];
  char                vsString [100];
  int                 nSwap;
  union
  {
    tsp00_Int4                      i4 ;
    tsp00_C4                        c4 ;
  } uMap;

  uMap.i4 = 65536;
  for ( nSwap = 0 ; nSwap < 4 ; nSwap ++ ) {
    if ( uMap.c4[nSwap] == 1 ) {
      break;
    } /* end if */
  } /* end for */

  tsp100_VersionID0   VersionID0;

  sp100_GetVersionID ( VersionIDType0_esp100, s100buildnumberEx, &VersionID0 );

  sprintf(vsString, "%d.%d.%d",
                    (int) VersionID0.MajorVersion_sp100,
                    (int) VersionID0.MinorVersion_sp100,
                    (int) VersionID0.CorrLevel_sp100);

  sqlos (&osNumber);

  switch (osNumber) {
    case os_unix:
      strcpy (osString, "UNIX");
      break;
    case os_vms:
      strcpy (osString, "VMS");
      break;
    case os_windows:
      strcpy (osString, "WINDOWS");
      break;
    case os_win32:
      strcpy (osString, "WIN32");
      break;
    case os_os2:
      strcpy (osString, "OS2");
      break;
    default:
      sprintf (osString, "<Unknown OS %d>", (int) osNumber);
      break;
  } /* end switch */

  sprintf(replyData, "%s%s%s%s\"%s\",\"%s\",\"%s\",%s,%s,%d%s",
                     ANSWER_OK_CN00,
                     LINE_SEPSTRING_CN00,
                     "version,os,dbroot,logon,code,swap",
                     LINE_SEPSTRING_CN00,
                     vsString,
                     osString,
                     (char *) vcontrol->dbroot,
                     BOOLSTR_CN00(vcontrol->bSysLogon),
                     (' ' == 32) ? "ASCII" : "EBCDIC",
                     nSwap,
                     LINE_SEPSTRING_CN00);

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn40GetVersion */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBMVersion
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBMVersion
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szOption[PARAMETER_MAXLEN_CN90];
  _TCHAR            * pCurrent = replyData;
  bool                bAll     = false;

  bAll = !cn90GetToken(command->args, szOption, 1);

  _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
  pCurrent = pCurrent + _tcslen(pCurrent);

  if (_tcsicmp(szOption, KEYVER_VERSION) == 0 || bAll) {
    tsp100_VersionID0   VersionID0;

    if (bAll) {
      _stprintf(pCurrent, "%-10s = ", KEYVER_VERSION);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if

    sp100_GetVersionID ( VersionIDType0_esp100, s100buildnumberEx, &VersionID0 );
    sprintf(pCurrent, "%d.%d.%d%s",
                      (int) VersionID0.MajorVersion_sp100,
                      (int) VersionID0.MinorVersion_sp100,
                      (int) VersionID0.CorrLevel_sp100,
                      LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);
  } // end if

  if (_tcsicmp(szOption, KEYVER_BUILD) == 0 || bAll) {
    tsp00_Versionc VersionString;

    if (bAll) {
      _stprintf(pCurrent, "%-10s = ", KEYVER_BUILD);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if

    sp100_GetVersionString  ( COMP_NAME_DBMSERVER_SP100, s100buildnumber, VersionString );

    _stprintf(pCurrent, "%s%s", (char *) VersionString, LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);
  } // end if

  if (_tcsicmp(szOption, KEYVER_OS) == 0 || bAll) {
    tsp_os              osNumber;

    if (bAll) {
      _stprintf(pCurrent, "%-10s = ", KEYVER_OS);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if

    sqlos (&osNumber);
    switch (osNumber) {
      case os_unix:
        _stprintf(pCurrent, "%s%s", "UNIX", LINE_SEPSTRING_CN00);
        break;
      case os_vms:
        _stprintf(pCurrent, "%s%s", "VMS", LINE_SEPSTRING_CN00);
        break;
      case os_windows:
        _stprintf(pCurrent, "%s%s", "WINDOWS", LINE_SEPSTRING_CN00);
        break;
      case os_win32:
        _stprintf(pCurrent, "%s%s", "WIN32", LINE_SEPSTRING_CN00);
        break;
      case os_os2:
        _stprintf(pCurrent, "%s%s", "OS2", LINE_SEPSTRING_CN00);
        break;
      default:
        _stprintf(pCurrent, "%d%s", (int) osNumber, LINE_SEPSTRING_CN00);
        break;
    } /* end switch */
    pCurrent = pCurrent + _tcslen(pCurrent);

  } // end if

  if (_tcsicmp(szOption, KEYVER_DBROOT) == 0 || bAll) {
    if (bAll) {
      _stprintf(pCurrent, "%-10s = ", KEYVER_DBROOT);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if
    _stprintf(pCurrent, "%s%s", (char *) vcontrol->dbroot, LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);
  } // end if

  if (_tcsicmp(szOption, KEYVER_LOGON) == 0 || bAll) {
    if (bAll) {
      _stprintf(pCurrent, "%-10s = ", KEYVER_LOGON);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if
    _stprintf(pCurrent, "%s%s", BOOLSTR_CN00(vcontrol->bSysLogon), LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);
  } // end if

  if (_tcsicmp(szOption, KEYVER_CODE) == 0 || bAll) {
    if (bAll) {
      _stprintf(pCurrent, "%-10s = ", KEYVER_CODE);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if
    _stprintf(pCurrent, "%s%s", (' ' == 32) ? CODE_ASCII : CODE_EBCDIC, LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);
  } // end if

  if (_tcsicmp(szOption, KEYVER_SWAP) == 0 || bAll) {
    tsp00_Int4     i234   = 1;
    char         * c234   = (char *) &i234;;
    const _TCHAR * szSwap = SWAP_PART;

    if (c234[3] == 1) {
        szSwap = SWAP_NORMAL;
    } else if (c234[0] == 1) {
        szSwap = SWAP_FULL;
    } // end if

    if (bAll) {
      _stprintf(pCurrent, "%-10s = ", KEYVER_SWAP);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if

    _stprintf(pCurrent, "%s%s", szSwap, LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);
  } // end if

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn40DBMVersion */


/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBMGetPath
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBMGetPath
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szMode[PARAMETER_MAXLEN_CN90];
  tsp00_Pathc         szPath;
  tsp01_RteError      RteError;
  tsp00_Bool          bDBRootWarning = false;

  cn90GetToken(command->args, szMode, 1);

  if (_tcsicmp(szMode, RTE_INDPROG) == 0) {
    if (!sqlGetIndependentProgramsPath ( szPath, 0, &RteError)) {
      nFuncReturn = cn90AnswerNewRTEError(replyData, replyLen, ERR_RTEEXT_CN00, &RteError);
    } // end if
    bDBRootWarning = !sqlIsIndependentProgramsPathInRegistry();
  } else if (_tcsicmp(szMode, RTE_INDDATA) == 0) {
    if (!sqlGetIndependentDataPath ( szPath, 0, &RteError)) {
      nFuncReturn = cn90AnswerNewRTEError(replyData, replyLen, ERR_RTEEXT_CN00, &RteError);
    } // end if
    bDBRootWarning = !sqlIsIndependentDataPathInRegistry();
  } else {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  } // end if

  if (nFuncReturn == OK_CN00) {
    _stprintf(replyData, "%s%s%s%s%s%s", 
                        ANSWER_OK_CN00, 
                        LINE_SEPSTRING_CN00,
                        szPath.asCharp(),
                        LINE_SEPSTRING_CN00,
                        bDBRootWarning ? "INFO: Path extracted from $DBROOT" : "",
                        bDBRootWarning ? LINE_SEPSTRING_CN00 : "");

    *replyLen = (int)strlen(replyData);
  } /* end if */

  return nFuncReturn;
} /* end cn40DBMGetPath */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBMSetPath
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBMSetPath
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szMode[PARAMETER_MAXLEN_CN90];
  tsp00_Pathc         szPath;
  tsp01_RteError      RteError;

  cn90GetToken(command->args, szMode, 1);
  cn90GetToken(command->args, szPath, 2);
  cn90StripQuotes(szPath);

  if (_tcsicmp(szMode, RTE_INDPROG) == 0) {
    if (!sqlSetIndependentProgramsPath ( szPath, &RteError)) {
      nFuncReturn = cn90AnswerNewRTEError(replyData, replyLen, ERR_RTEEXT_CN00, &RteError);
    } // end if
  } else if (_tcsicmp(szMode, RTE_INDDATA) == 0) {
    if (!sqlSetIndependentDataPath ( szPath, &RteError)) {
      nFuncReturn = cn90AnswerNewRTEError(replyData, replyLen, ERR_RTEEXT_CN00, &RteError);
    } // end if
  } else {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} /* end cn40DBMSetPath */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBRegister
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBRegister
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  char                szOption[PARAMETER_MAXLEN_CN90];
  tsp9_pgm_kind       pgmKind     = csp9_any_pgm;
  tsp9_rte_xerror     xError;

  if (cn90GetToken(command->args, szOption, 1)) {
    if (strcmp(szOption, OPT_SPEED_FAST) == 0 || strcmp(szOption, OPTS_SPEED_FAST) == 0) {
      pgmKind = csp9_fast_pgm;
    } else if (strcmp(szOption, OPT_SPEED_QUICK) == 0 || strcmp(szOption, OPTS_SPEED_QUICK) == 0) {
      pgmKind = csp9_quick_pgm;
    } else if (strcmp(szOption, OPT_SPEED_SLOW) == 0 || strcmp(szOption, OPTS_SPEED_SLOW) == 0) {
      pgmKind = csp9_slow_pgm;
    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } /* end if */
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    sqlxregisterdb (vcontrol->dbname, vcontrol->dbroot, pgmKind, csp9_no_options, &xError);

    if (xError.xe_result != csp9_xrte_ok) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);

      // in case of error try quick/fast (only if "any" selected)
      if (pgmKind == csp9_any_pgm) {
        pgmKind = csp9_quick_pgm;
        sqlxregisterdb (vcontrol->dbname, vcontrol->dbroot, pgmKind, csp9_no_options, &xError);
        if (xError.xe_result != csp9_xrte_ok) {
          pgmKind = csp9_slow_pgm;
          sqlxregisterdb (vcontrol->dbname, vcontrol->dbroot, pgmKind, csp9_no_options, &xError);
          if (xError.xe_result == csp9_xrte_ok) {
            nFuncReturn = OK_CN00;
          } // end if
        } else {
          nFuncReturn = OK_CN00;
        } // end if
      } // end if

    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} /* end cn40DBRegister */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBCreate
 * ------------------------------------------------------------------
 */
#ifndef SERVICE_PARAMFILE_ID_EO01
  #define SERVICE_PARAMFILE_ID_EO01       "_"
#endif
tcn00_Error cn40DBCreate
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  FUNCTION_DBG_MCN00_1(_T("cn40DBCreate"));

  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szToken    [PARAMETER_MAXLEN_CN90];
  _TCHAR            * pPassword   = NULL;
  tsp00_DbNamec       szDBName;
  tsp00_UserNamec     szUser;
  tsp00_Namec         szPassword;
  tsp00_Namec         szSuperPwd;
  tsp00_UserNamec     szOsUser;
  tsp00_Namec         szOsPassword;
  tsp9_rte_xerror     xError;
  bool                bRegistered = false;
  int                 nToken      = 1;
  bool                inOptions   = true;
  bool                bOsUser     = false;
  tsp9_pgm_kind       pgmKind     = csp9_any_pgm;
  tsp9_server_options pgmOption   = csp9_no_options;
  int                 nIndex      = 0;
  int                 nLength     = 0;

  while (inOptions) {
    cn90GetToken(command->args, szToken, nToken);
    if (_tcscmp(szToken, OPT_SPEED_FAST) == 0 || strcmp(szToken, OPTS_SPEED_FAST) == 0) {
      pgmKind = csp9_fast_pgm;
    } else if (_tcscmp(szToken, OPT_SPEED_QUICK) == 0 || _tcscmp(szToken, OPTS_SPEED_QUICK) == 0) {
      pgmKind = csp9_quick_pgm;
    } else if (_tcscmp(szToken, OPT_SPEED_SLOW) == 0 || _tcscmp(szToken, OPTS_SPEED_SLOW) == 0) {
      pgmKind = csp9_slow_pgm;
    } else if (_tcscmp(szToken, OPT_OPT_AUTO) == 0 || _tcscmp(szToken, OPTS_OPT_AUTO) == 0) {
      pgmOption  = csp9_start_automatic;
    } else {
      inOptions = false;
      szDBName.rawAssign(szToken);
      cn90StrUpperCopy(szDBName, szDBName, false);
    } // end if
    nToken++;
  } // end while

  // check dbname
  if (_tcsncmp(szDBName, SERVICE_PARAMFILE_ID_EO01, _tcslen(SERVICE_PARAMFILE_ID_EO01)) == 0 ) {
    teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_INVDBNAME_CN00_1, szDBName.asCharp());
    nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
  } // end if

  if (nFuncReturn == OK_CN00) {
    // read User
    cn90GetToken(command->args, szToken, nToken);
    pPassword = _tcschr(szToken, PWD_SEPARATOR_CN00);
    if (pPassword != NULL) {
      *pPassword = CHAR_STRINGTERM_CN90;
      pPassword++;
      szUser.rawAssign(szToken);
      szPassword.rawAssign(pPassword);

      bOsUser = cn90GetToken(command->args, szToken, nToken + 1);
      if (bOsUser) {
        pPassword = _tcschr(szToken, PWD_SEPARATOR_CN00);
        if (pPassword != NULL) {
          *pPassword = CHAR_STRINGTERM_CN90;
          pPassword++;
          szOsUser.rawAssign(szToken);
          szOsPassword.rawAssign(pPassword);
        } else {
          nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
        } // end if
      } // end if
    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  // os logon
  if (nFuncReturn == OK_CN00 && bOsUser) {
    cn90Uncrypt(szOsPassword, false);
    sqlxlogon (szOsUser, szOsPassword, &xError);

    if (xError.xe_result != 0) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_SYSLOGON_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } else {
      vcontrol->bSysLogon = true;
    } // end if
  } // end if

  // check logon state
  if (nFuncReturn == OK_CN00 && !vcontrol->bSysLogon) {
    nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_NEEDSYSLOGON_CN00);
  } // end if

  // check existence of database
  if (nFuncReturn == OK_CN00) {
    if (cn90DBExists(szDBName)) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_DBEXIST_CN00);
    } // end if
  } // end if

  // register database
  if (nFuncReturn == OK_CN00) {
    sqlxregisterdb (szDBName, vcontrol->dbroot, pgmKind, pgmOption, &xError);
    if (xError.xe_result != csp9_xrte_ok) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);

      // in case of error try quick/fast (only if "any" selected)
      if (pgmKind == csp9_any_pgm) {
        pgmKind = csp9_quick_pgm;
        sqlxregisterdb (szDBName, vcontrol->dbroot, pgmKind, pgmOption, &xError);
        if (xError.xe_result != csp9_xrte_ok) {
          pgmKind = csp9_slow_pgm;
          sqlxregisterdb (szDBName, vcontrol->dbroot, pgmKind, pgmOption, &xError);
          if (xError.xe_result == csp9_xrte_ok) {
            nFuncReturn = OK_CN00;
          } // end if
        } else {
          nFuncReturn = OK_CN00;
        } // end if
      } // end if

    } // end if

    if (nFuncReturn == OK_CN00) {
      bRegistered = true;
    } // end if
  } // end if

  // create first DBM-User
  if (nFuncReturn == OK_CN00) {
    cn90Uncrypt(pPassword, false);

    vcontrol->pCurrentUser->clear();
    vcontrol->pCurrentUser->setDBName       ( szDBName    )
                           .setUserName     ( szUser      )
                           .setMasterPwd    ( szPassword  )
                           .setDBMUser      ( true )
                           .setServerRights ( DBMMaskAll_CN50, DBMUserDefault_CN50 )
                           .setGUIRights    ( DBMMaskAll_CN50, DBMUserDefaultGUI_CN50 );

    nFuncReturn = vcontrol->pCurrentUser->save();
    if (nFuncReturn == OK_CN00) {
      vcontrol->pCurrentUser->checkMasterPwd(szPassword);
    } else {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } // end if
  } // end if

  // create super user and operator
  if (nFuncReturn == OK_CN00) {
    cn50DBMUser usrOperator;
    cn50DBMUser usrSuperusr;

    usrOperator.setDBName       ( szDBName )
               .setUserName     ( USERNAME_OPERATOR_CN50        )
               .setMasterPwd    ( USERNAME_OPERATOR_CN50        )
               .setDBMUser      ( true )
               .setServerRights ( DBMMaskAll_CN50, OperatorUserDefault_CN50 )
               .setGUIRights    ( DBMMaskAll_CN50, OperatorUserDefaultGUI_CN50 );

    usrOperator.save();

    memset(szSuperPwd, CHAR_STRINGTERM_CN90, sizeof(szSuperPwd));
    nLength = (int)_tcslen(szPassword);
    for (nIndex = 0; nIndex < nLength; nIndex++) {
      szSuperPwd[nLength - nIndex - 1] = szPassword[nIndex];
    } // end for
    usrSuperusr.setDBName       ( szDBName   )
               .setUserName     ( USERNAME_SUPERUSER_CN50       )
               .setMasterPwd    ( szSuperPwd )
               .setDBMUser      ( true )
               .setServerRights ( DBMMaskAll_CN50, SuperUserDefault_CN50 )
               .setGUIRights    ( DBMMaskAll_CN50, SuperUserDefaultGUI_CN50 );

    usrSuperusr.save();
  } // end if

  // generate answer
  if (nFuncReturn == OK_CN00) {
    vcontrol->dbname = szDBName;
    cn90AnswerOK (replyData, replyLen, NULL);
  } else {
    if (bRegistered) {
      sqlxunregisterdb (szDBName, csp9_any_pgm, &xError);
    } // end if
    vcontrol->pCurrentUser->clear();
  } // end if

  return nFuncReturn;
} /* end cn40DBCreate */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBDrop
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBDrop
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error          nFuncReturn = OK_CN00;
  _TCHAR               szToken    [PARAMETER_MAXLEN_CN90];
  bool                 bWithFiles = true;

  if (cn90GetToken(command->args, szToken, 1)) {
    if (_tcsicmp(szToken, KEYDROP_WITHOUTFILES) == 0) {
      bWithFiles = false;
    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn40_DBDrop(vcontrol->dbname, vcontrol->dbroot, bWithFiles, replyData, replyLen, replyLenMax);
  } // end if

  return nFuncReturn;
} // end cn40DBDrop

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_DBDrop
 * ------------------------------------------------------------------
 */
static tcn00_Error cn40_DBDrop
      (       tsp00_DbNamec   szDbName,
        const tsp00_Pathc     szDbRoot,
        const bool            bWithFiles,
        char                * replyData,
        int                 * replyLen,
        const int             replyLenMax)
{
  tcn00_Error          nFuncReturn = OK_CN00;
  tcn00_Error          nTmpState = OK_CN00;
  tsp9_rte_xerror      xError;
  tcn00_DevspaceParams aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  int                  nParam     = 0;
  int                  nNumber    = 0;
  tsp00_XpKeyTypec     szParameter;
  tsp00_C64c           szDevFile;
  tsp00_C64c           szDevType;
  tsp00_C64c           szRunDir;
  tsp5_rte_fileerror   fileErr;
  tsp01_RteError       rteError;
  tcn00_DBState        aState;

  if (nFuncReturn == OK_CN00) {

    aState = cn90DBState(szDbName);
    if (aState == STATE_ERROR_CN00) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_STATE_CN00);
    } else if (aState != STATE_OFFLINE_CN00 && aState != STATE_RUIN_CN00) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_DBRUN_CN00);
    } else {
      if ( aState == STATE_RUIN_CN00 ) {
        sqlxclear( szDbName, szDbRoot, &xError );
      } // end if

      sqlxunregisterdb (szDbName, csp9_any_pgm, &xError);
      if (xError.xe_result != csp9_xrte_ok) {
        nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
      } else if (bWithFiles) {

        // -- devspaces
        // change to rundir
        nTmpState = cn20XParamGetValue(szDbName, XP_ID_RUNDIRECTORY_CN00, szRunDir);
        if (sqlchdirc((char *) szRunDir, &rteError)) {

          // iterate trough devspacetypes
          nParam = 0;
          while (aParamTable[nParam].nDevSpace != XPDevSpaceUnknown_ecn00) {
            nNumber = 0;

            // iterate to devspaces of one type
            do {
              nNumber++;

              // look for devicename
              _stprintf(szParameter, aParamTable[nParam].szName, nNumber);
              nTmpState = cn20XParamGetValue(szDbName, szParameter, szDevFile);

              if (nTmpState == OK_CN00) {
                // look for devicetype
                _stprintf(szParameter, aParamTable[nParam].szType, nNumber);
                nTmpState = cn20XParamGetValue(szDbName, szParameter, szDevType);

                // delete device
                if (nTmpState == OK_CN00) {
                  if (szDevType[0] == _T('F')) {
                    sqlferasec(szDevFile, &fileErr);
                  } // end if
                } // end if
              } // end if

            } while (nTmpState == OK_CN00);

            nParam++;
          } // end while

          sqlchdirc(szDbRoot.asCharp(), &rteError);
        } // end if

        // -- all other file
        cn42RemoveFiles(szDbName);

        // -- xparam
        sqlXParamRemoveAll(szDbName, &rteError);

        // -- user profile
        teo200_EventList        EventList;
        teo28_ProfileContainer  Profile(szDbName, &EventList);
        if (!EventList) {
          Profile.eo28_DeleteProfileContainer(&EventList);
        } // end if

        // -- rundir
        if (sqlchdirc((char *) szRunDir, &rteError)) {
          sqlremove_empty_dirc( "rtedump_dir", &rteError);
          sqlchdirc(szDbRoot.asCharp(), &rteError);
        } // end if
        sqlremove_empty_dirc(szRunDir, &rteError);

      } // end if
    } // end if
  } // end if

  // generate answer
  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} /* end cn40_DBDrop */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBUnregister
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBUnregister
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  char                szOption[PARAMETER_MAXLEN_CN90];
  tsp9_pgm_kind       pgmKind = csp9_any_pgm;
  tsp9_rte_xerror     xError;
  tsp_int4            nCounter = 0;

  if (cn90GetToken(command->args, szOption, 1)) {
    if (_tcsicmp(szOption, OPT_SPEED_FAST) == 0 || _tcsicmp(szOption, OPTS_SPEED_FAST) == 0) {
      pgmKind = csp9_fast_pgm;
    } else if (_tcsicmp(szOption, OPT_SPEED_QUICK) == 0 || _tcsicmp(szOption, OPTS_SPEED_QUICK) == 0) {
      pgmKind = csp9_quick_pgm;
    } else if (_tcsicmp(szOption, OPT_SPEED_SLOW) == 0 || _tcsicmp(szOption, OPTS_SPEED_SLOW) == 0) {
      pgmKind = csp9_slow_pgm;
    } else {
      nFuncReturn = ERR_PARAM_CN00;
    } /* end if */
  } /* end if */

  // check existence of database
  if (nFuncReturn == OK_CN00) {
    if (!cn90DBExists(vcontrol->dbname, nCounter)) {
      nFuncReturn = ERR_NOENTRY_CN00;
    } else {
      nFuncReturn = (nCounter == 1) ? ERR_LASTREG_CN00 : nFuncReturn;
      if (nFuncReturn == OK_CN00) {
        pgmKind = csp9_any_pgm;
        nFuncReturn = (cn90DBSpeed(vcontrol->dbname) == pgmKind) ? ERR_DBRUN_CN00 : nFuncReturn;
      } // end if
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    sqlxunregisterdb (vcontrol->dbname, pgmKind, &xError);

    if (xError.xe_result != csp9_xrte_ok) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } else {
      cn90AnswerOK (replyData, replyLen, NULL);
    } /* end if */
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  return nFuncReturn;
} /* end cn40DBUnregister */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBEnum
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBEnum
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp9_rte_xerror     xError;
  void              * hEnum;
  tsp9_rte_dbinfo     dbInfo;
  char              * pCurrPos;
  const char        * szMode;
  char                szState[20];
  char                szOption[PARAMETER_MAXLEN_CN90];
  bool                bService = false;

  if (cn90GetToken(command->args, szOption, 1)) {
    if (_tcsicmp(szOption, OPTS_ENUM_SRV) == 0) {
      bService = true;
    } else {
      nFuncReturn = cn90AnswerIError (replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  sqlxopen_db_enum (&hEnum, true, &xError);

  if (xError.xe_result == csp9_xrte_ok) {
    /* Init Output */
    sprintf(replyData, "%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00);
    pCurrPos = replyData + strlen (replyData);

    while (xError.xe_result == csp9_xrte_ok) {
      sqlxnext_db (hEnum, &dbInfo, &xError);
      if (xError.xe_result == csp9_xrte_ok) {

        switch (dbInfo.pgmKind) {
          case csp9_fast_pgm:
            szMode = TXTL_SPEED_FAST;
            break;
          case csp9_slow_pgm:
            szMode = TXTL_SPEED_SLOW;
            break;
          case csp9_quick_pgm:
            szMode = TXTL_SPEED_QUICK;
            break;
          default:
            szMode = "";
            break;
        } /* end switch */

        switch (dbInfo.state) {
          case csp9_state_off:
            _tcscpy(szState, TXT_STATE_OFFLINE_CN00);
            break;
          case csp9_state_running:
            _tcscpy(szState, TXT_STATE_RUNNING_CN00);
            break;
          case csp9_state_ruin:
            _tcscpy(szState, TXT_STATE_RUIN_CN00);
            break;
          default:
            _tcscpy(szState, TXT_STATE_UNKNOWN_CN00);
            break;
        } /* end switch */

        char * pState = &szState[0];
        while (*pState != 0) {
          *pState = tolower(*pState);
          ++pState;
        } // end if

        if ( (_tcsncmp(dbInfo.dbname, SERVICE_PARAMFILE_ID_EO01, _tcslen(SERVICE_PARAMFILE_ID_EO01)) == 0 && bService) ||
             (_tcsncmp(dbInfo.dbname, SERVICE_PARAMFILE_ID_EO01, _tcslen(SERVICE_PARAMFILE_ID_EO01)) != 0            )     ) {
          sprintf (pCurrPos, "%s%s%s%s%d.%d.%d.%d%s%s%s%s%s",
                             (char *) dbInfo.dbname,
                             VALUE_SEPSTRING_CN00,
                             (char *) dbInfo.dbroot,
                             VALUE_SEPSTRING_CN00,
                             (dbInfo.version.no[0] != csp9_invalid_version) ? dbInfo.version.no[0] : 0,
                             (dbInfo.version.no[1] != csp9_invalid_version) ? dbInfo.version.no[1] : 0,
                             (dbInfo.version.no[2] != csp9_invalid_version) ? dbInfo.version.no[2] : 0,
                             (dbInfo.version.no[3] != csp9_invalid_version) ? dbInfo.version.no[3] : 0,
                             VALUE_SEPSTRING_CN00,
                             szMode,
                             VALUE_SEPSTRING_CN00,
                             szState,
                             LINE_SEPSTRING_CN00);
        } // end if

        pCurrPos = pCurrPos + strlen (pCurrPos);
      } /* end if */
    } /* end while */

    if (xError.xe_result != csp9_xrte_atend) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } /* end if */

    sqlxclose_db_enum (hEnum);
  } else {
    nFuncReturn = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn40DBEnum */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBClear
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBClear
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp9_rte_xerror     xError;

  sqlxclear( vcontrol->dbname, vcontrol->dbroot, &xError );

  if (xError.xe_result != csp9_xrte_ok) {
    nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
  } else {
    cn90AnswerOK (replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;

} /* end cn40DBClear */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBStart
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBStart
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn40_ChangeDBState ( vcontrol,
                              command,
                              replyData,
                              replyLen,
                              replyLenMax,
                              STATE_COLD_CN00);
} /* end cn40DBStart */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBStop
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBStop
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn40_ChangeDBState ( vcontrol,
                              command,
                              replyData,
                              replyLen,
                              replyLenMax,
                              STATE_OFFLINE_CN00,
                              true);

} /* end cn40DBStop */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBWarm
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBWarm
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn40_ChangeDBState ( vcontrol,
                              command,
                              replyData,
                              replyLen,
                              replyLenMax,
                              STATE_WARM_CN00);

} /* end cn40DBWarm */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBOffline
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBOffline
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn40_ChangeDBState ( vcontrol,
                              command,
                              replyData,
                              replyLen,
                              replyLenMax,
                              STATE_OFFLINE_CN00);

} /* end cn40DBOffline */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBCold
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBCold
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn40_ChangeDBState ( vcontrol,
                              command,
                              replyData,
                              replyLen,
                              replyLenMax,
                              STATE_COLD_CN00);
} /* end cn40DBCold */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBRestart
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBRestart
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  tsp9_pgm_kind   pgmKind;
  bool            bDump;
  bool            bImmediate;
  _TCHAR          szNewArgs [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR          szRestart [PARAMETER_MAXLEN_CN90] = "";

  // check current and requested speed
  if (cn40_StartStopOptions(command->args, pgmKind, bDump, bImmediate, szRestart) == OK_CN00) {
    if (pgmKind == csp9_any_pgm) {
      pgmKind = cn90DBSpeed(vcontrol->dbname);

      switch (pgmKind) {
        case csp9_slow_pgm:
          _stprintf (szNewArgs, "%s %s", (_TCHAR *) command->args, (_TCHAR *) OPT_SPEED_SLOW);
          break;
        case csp9_quick_pgm:
          _stprintf (szNewArgs, "%s %s", (_TCHAR *) command->args, (_TCHAR *) OPT_SPEED_QUICK);
          break;
        case csp9_fast_pgm:
          _stprintf (szNewArgs, "%s %s", (_TCHAR *) command->args, (_TCHAR *) OPT_SPEED_FAST);
          break;
        default:
          _stprintf (szNewArgs, "%s", (_TCHAR *) command->args);
          break;
      } // end switch

      command->args = &szNewArgs[0];
    } // end if
  }

  nFuncReturn = cn40_ChangeDBState ( vcontrol,
                                     command,
                                     replyData,
                                     replyLen,
                                     replyLenMax,
                                     STATE_OFFLINE_CN00);
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn40_ChangeDBState ( vcontrol,
                                       command,
                                       replyData,
                                       replyLen,
                                       replyLenMax,
                                       STATE_WARM_CN00);
  } // end if

  return nFuncReturn;
} /* end cn40DBRestart */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40ShowSomething
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ShowSomething
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;

  _TCHAR       * pCurrent  = replyData;
  _TCHAR       * pContinue = NULL;
  bool           bNoNewLine;
  tsp00_Bool     bOk;
  tsp00_Int4     nExitCode;
  tsp00_Int4     nSize;
  ceo43_xshow  * pObjShow = NULL;
  ceo43_xshow  * pObjTmp  = NULL;

  pObjShow = new ceo43_xshow(command->args, vcontrol->dbname, bOk);

  // Init OK Message
  if (bOk && pObjShow != NULL) {

    _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00,  LINE_SEPSTRING_CN00);
    pCurrent = pCurrent    + _tcslen(pCurrent);
    _stprintf(pCurrent, "%*s%s", (int) _tcslen(KEY_DIRCONTINUE_CN00), _T(" "), LINE_SEPSTRING_CN00);
    pContinue = pCurrent;
    pCurrent = pCurrent    + _tcslen(pCurrent);

    nSize    = (tsp00_Int4)(replyLenMax - _tcslen(replyData));

    while (pObjShow->eo43NextLine(pCurrent, nSize, bNoNewLine, nExitCode) > 0 && nSize > 200) {
      pCurrent = pCurrent + _tcslen(pCurrent);
      nSize = (tsp00_Int4)(replyLenMax - _tcslen(replyData));
      if (bNoNewLine && nSize > 0) {
        _stprintf(pCurrent, LINE_SEPSTRING_CN00);
        pCurrent = pCurrent + _tcslen(pCurrent);
      } // end if
      nSize = (tsp00_Int4)(replyLenMax - _tcslen(replyData));
    } // end while

    if (nSize <= 200) {
      cn40FreeCommandObj(vcontrol);
      vcontrol->pCommandObj = (void *) pObjShow;
      _tcsncpy(pContinue, KEY_DIRCONTINUE_CN00, _tcslen(KEY_DIRCONTINUE_CN00));
    } else {
      if (bNoNewLine) {
        _stprintf(pCurrent, LINE_SEPSTRING_CN00);
        pCurrent = pCurrent + _tcslen(pCurrent);
      } // end if
      delete pObjShow;
    } // end if

    *replyLen = (int)_tcslen(replyData);

  } else {
    nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_EXECUTE_CN00);
  } // end if

  return nFuncReturn;
} /* end cn40ShowSomething */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40ShowList
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ShowList
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR       * pCurrent = replyData;
  _TCHAR       * pShowCmd = NULL;
  _TCHAR         szLine[PARAMETER_MAXLEN_CN90];
  _TCHAR         szToken[PARAMETER_MAXLEN_CN90];
  bool           bNoNewLine;
  tsp00_Bool     bOk;
  tsp00_Int4     nExitCode;
  tsp00_Int4     nSize;
  ceo43_CmdPipe  pObjXCons;

  _stprintf(szLine, _T("%s %s %s"), TXT_CMD_XCONS, (_TCHAR *) vcontrol->dbname, TXT_CMD_HELP);
  pObjXCons.eo43OpenCommand(szLine, bOk);

  // Init OK Message
  if (bOk) {

    _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00,  LINE_SEPSTRING_CN00);
    pCurrent = pCurrent    + _tcslen(pCurrent);
    nSize    = (tsp00_Int4)(replyLenMax - _tcslen(replyData));

    while (pObjXCons.eo43NextLine(szLine, PARAMETER_MAXLEN_CN90, bNoNewLine, nExitCode) > 0) {
      pShowCmd = strstr(szLine, TXT_CMD_SHOW);
      if (pShowCmd != NULL) {
        cn90GetToken(pShowCmd, szToken, 2);
        _stprintf(pCurrent, "%s%s", szToken,  LINE_SEPSTRING_CN00);
        pCurrent = pCurrent + _tcslen(pCurrent);
      } // end if
    } // end while

      *replyLen = (int)_tcslen(replyData);

  } else {
    nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_EXECUTE_CN00);
  } // end if

  return nFuncReturn;
} /* end cn40ShowList */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40ShowNext
 * ------------------------------------------------------------------
 */
tcn00_Error cn40ShowNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR       * pCurrent  = replyData;
  _TCHAR       * pContinue = NULL;
  bool           bNoNewLine;
  tsp00_Int4     nExitCode;
  tsp00_Int4     nSize;
  ceo43_xshow  * pObjShow = NULL;
  ceo43_xshow  * pObjTmp  = NULL;

  pObjShow = (ceo43_xshow *) vcontrol->pCommandObj;

  // Init OK Message
  if (pObjShow != NULL) {

  _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00,  LINE_SEPSTRING_CN00);
  pCurrent = pCurrent    + _tcslen(pCurrent);
  _stprintf(pCurrent, "%.*s%s", (int) _tcslen(KEY_DIRCONTINUE_CN00), _T(""), LINE_SEPSTRING_CN00);
  pContinue = pCurrent;
  pCurrent = pCurrent    + _tcslen(pCurrent);

  nSize    = (tsp00_Int4)(replyLenMax - _tcslen(replyData));

  while (pObjShow->eo43NextLine(pCurrent, nSize, bNoNewLine, nExitCode) > 0 && nSize > 200) {
    pCurrent = pCurrent + _tcslen(pCurrent);
    nSize = (tsp00_Int4)(replyLenMax - _tcslen(replyData));
    if (bNoNewLine && nSize > 0) {
      _stprintf(pCurrent, LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if
    nSize = (tsp00_Int4)(replyLenMax - _tcslen(replyData));
  } // end while

  if (nSize <= 200) {
    vcontrol->pCommandObj = (void *) pObjShow;
    _tcsncpy(pContinue, KEY_DIRCONTINUE_CN00, _tcslen(KEY_DIRCONTINUE_CN00));
  } else {
    if (bNoNewLine) {
      _stprintf(pCurrent, LINE_SEPSTRING_CN00);
      pCurrent = pCurrent + _tcslen(pCurrent);
    } // end if
    cn40FreeCommandObj(vcontrol);
  } // end if

  *replyLen = (int)_tcslen(replyData);

  } else {
    nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_EXECUTE_CN00);
  } // end if

  return nFuncReturn;
} /* end cn40ShowNext */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBConsole
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBConsole
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR       * pCurrent = replyData;
  _TCHAR         szLine[PARAMETER_MAXLEN_CN90];
  bool           bNoNewLine;
  tsp00_Bool     bOk;
  tsp00_Int4     nExitCode;
  tsp00_Int4     nSize;
  ceo43_CmdPipe  pObjXCons;

  cn90GetToken(command->args, szLine, 1);
  if (_tcsicmp(szLine, TXT_CMD_SHOW) == 0) {
    nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
  } else {
    _stprintf(szLine, _T("%s %s %s"), TXT_CMD_XCONS, (_TCHAR *) vcontrol->dbname, command->args);

    pObjXCons.eo43OpenCommand(szLine, bOk);

    if (bOk) {

      // init message
      _stprintf(pCurrent, "%70s%s", " ",  LINE_SEPSTRING_CN00);
      pCurrent = pCurrent    + _tcslen(pCurrent);

      nSize    = (tsp00_Int4)(replyLenMax - _tcslen(replyData));

      while (pObjXCons.eo43NextLine(pCurrent, nSize, bNoNewLine, nExitCode) > 0 && nSize > 100) {
        pCurrent = pCurrent + _tcslen(pCurrent);
        nSize = (tsp00_Int4)(replyLenMax - _tcslen(replyData));
        if (bNoNewLine && nSize > 0) {
          _stprintf(pCurrent, LINE_SEPSTRING_CN00);
          pCurrent = pCurrent + _tcslen(pCurrent);
        } // end if
        nSize = (tsp00_Int4)(replyLenMax - _tcslen(replyData));
      } // end while

      if (bNoNewLine) {
        _stprintf(pCurrent, LINE_SEPSTRING_CN00);
        pCurrent = pCurrent + _tcslen(pCurrent);
      } // end if

      if (nExitCode != 0) {
        nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_EXECUTE_CN00);
      } else {
        cn90AnswerOK (replyData, replyLen, NULL);
      } // end if
      replyData[_tcslen(replyData) - 1] = _T(' ');
      replyData[_tcslen(replyData)] = _T(' ');

      *replyLen = (int)_tcslen(replyData);

    } else {
      nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_EXECUTE_CN00);
    } // end if

  } // end if

  return nFuncReturn;
} /* end cn40DBConsole */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40FreeCommandObj
 * ------------------------------------------------------------------
 */
void cn40FreeCommandObj
      ( VControlDataT * vcontrol )
{
  ceo43_xshow  * pObjShow = NULL;

  if (vcontrol != NULL) {
    pObjShow = (ceo43_xshow *) vcontrol->pCommandObj;
    if (pObjShow != NULL) {
      delete pObjShow;
    } // end if
    vcontrol->pCommandObj = NULL;
  } // end if

} // end cn40FreeCommandObj

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBState
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBState
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  tcn00_DBState  DBState = STATE_UNKNOWN_CN00;

  DBState = cn90DBState(vcontrol->dbname);

  switch (DBState) {
    case STATE_OFFLINE_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_OFFLINE_CN00,      LINE_SEPSTRING_CN00);
      break;
    case STATE_STARTING_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_STARTING_CN00,     LINE_SEPSTRING_CN00);
      break;
    case STATE_COLD_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_COLD_CN00,         LINE_SEPSTRING_CN00);
      break;
    case STATE_WARM_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_WARM_CN00,         LINE_SEPSTRING_CN00);
      break;
    case STATE_SHTDWN_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_SHTDWN_CN00,       LINE_SEPSTRING_CN00);
      break;
    case STATE_SHTDWNREI_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_SHTDWNREI_CN00,    LINE_SEPSTRING_CN00);
      break;
    case STATE_SHTDWNKILL_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_SHTDWNKILL_CN00,   LINE_SEPSTRING_CN00);
      break;
    case STATE_STOP_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_STOP_CN00,         LINE_SEPSTRING_CN00);
      break;
    case STATE_KILL_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_KILL_CN00,         LINE_SEPSTRING_CN00);
      break;
    case STATE_ABORT_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_ABORT_CN00,        LINE_SEPSTRING_CN00);
      break;
    case STATE_RUIN_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_RUIN_CN00,         LINE_SEPSTRING_CN00);
      break;
    case STATE_RUNNING_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_RUNNING_CN00,      LINE_SEPSTRING_CN00);
      break;
    case STATE_ERROR_CN00:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_ERROR_CN00,        LINE_SEPSTRING_CN00);
      break;
    default:
      _stprintf(replyData, "%s%s%s%s%s%s",
                           ANSWER_OK_CN00,              LINE_SEPSTRING_CN00,
                           TXT_CMD_STATE,               LINE_SEPSTRING_CN00,
                           TXT_STATE_UNKNOWN_CN00,      LINE_SEPSTRING_CN00);
      break;
  } // end if

  *replyLen = (int)_tcslen(replyData);

  return nFuncReturn;
} /* end cn40DBState */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DBSpeed
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DBSpeed
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  tsp9_pgm_kind  pgmKind = csp9_any_pgm;
  tcn00_DBState  DBState = STATE_UNKNOWN_CN00;

  pgmKind = cn90DBSpeed(vcontrol->dbname);
  DBState = cn90DBState(vcontrol->dbname);

  if (DBState == STATE_WARM_CN00 || DBState == STATE_COLD_CN00) {
    switch (pgmKind) {
      case csp9_fast_pgm:
        _stprintf(replyData, "%s%s%s%s%s%s",
                             ANSWER_OK_CN00,    LINE_SEPSTRING_CN00,
                             TXT_CMD_SPEED,     LINE_SEPSTRING_CN00,
                             TXTU_SPEED_FAST,   LINE_SEPSTRING_CN00);
        break;
      case csp9_quick_pgm:
        _stprintf(replyData, "%s%s%s%s%s%s",
                             ANSWER_OK_CN00,    LINE_SEPSTRING_CN00,
                             TXT_CMD_SPEED,     LINE_SEPSTRING_CN00,
                             TXTU_SPEED_QUICK,  LINE_SEPSTRING_CN00);
        break;
      case csp9_slow_pgm:
        _stprintf(replyData, "%s%s%s%s%s%s",
                             ANSWER_OK_CN00,    LINE_SEPSTRING_CN00,
                             TXT_CMD_SPEED,     LINE_SEPSTRING_CN00,
                             TXTU_SPEED_SLOW,   LINE_SEPSTRING_CN00);
        break;
      default:
        _stprintf(replyData, "%s%s%s%s%s%s",
                             ANSWER_OK_CN00,    LINE_SEPSTRING_CN00,
                             TXT_CMD_SPEED,     LINE_SEPSTRING_CN00,
                             TXTU_SPEED_UNKNOWN,LINE_SEPSTRING_CN00);
        break;
    } // end if

    *replyLen = (int)_tcslen(replyData);
  } else {
    nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_DBNORUN_CN00);
  } // end if

  return nFuncReturn;
} /* end cn40DBSpeed */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40_RegisterServiceDB
 * ------------------------------------------------------------------
 */
#define DBNAME_LENGTH 8
// 65 bis 90

static bool cn40_RegisterServiceDB
      ( const tsp09_RteInstallationInfoNew  & relInfoIn)
{
  FUNCTION_DBG_MCN00_1(_T("cn40_RegisterServiceDB"));

  bool             bReturn = false;
  _TCHAR           szTemp[PARAMETER_MAXLEN_CN90];
  _TCHAR         * pTemp = NULL;
  tsp00_DbNamec    szDBName;
  int              nPostfix = 1;
  tsp9_rte_xerror  xError;

  cn90ServiceDB(szDBName, relInfoIn.dbroot);

  if (szDBName.length() == 0) {

    if (relInfoIn.key.length() == 0) {
      _stprintf(szTemp, "%sS%d%02d%02d", SERVICE_PARAMFILE_ID_EO01, 
                                         relInfoIn.version.no[0],
                                         relInfoIn.version.no[1],
                                         relInfoIn.version.no[2]);
    } else {
      int i,j;

      _tcscpy(szTemp, SERVICE_PARAMFILE_ID_EO01);
      i = (int)_tcslen(szTemp);
      j = 0;
      while (i < DBNAME_LENGTH && relInfoIn.key[j] != CHAR_STRINGTERM_CN90) {
        if (isalnum(relInfoIn.key[j])) {
          szTemp[i] = _totupper(relInfoIn.key[j]);
          ++i;
        } // end if
        ++j;
      } // end while
      szTemp[i] = CHAR_STRINGTERM_CN90;
    } // end if

    pTemp     = &szTemp[(_tcslen(szTemp) > (DBNAME_LENGTH - 2)) ? (DBNAME_LENGTH - 2) : _tcslen(szTemp)];

    szDBName.rawAssign(szTemp);

    while (cn90DBExists(szDBName) && nPostfix < 100) {
      _stprintf(pTemp, "%02d", nPostfix);
      ++nPostfix;
      szDBName.rawAssign(szTemp);
    } // end while

  } // end if  

  if (nPostfix < 100) {
    sqlxregisterdb (szDBName, relInfoIn.dbroot, csp9_any_pgm, csp9_no_options, &xError);
    if (xError.xe_result == csp9_xrte_ok) {
      if (cn20CreateSrvParam( szDBName ) == OK_CN00) {
        bReturn = true;
      } else {
        sqlxunregisterdb (szDBName, csp9_any_pgm, &xError);
      } // end if
    } else {
      teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, xError.xe_result,  TERR_CN00_1, _T("DBM"), xError.xe_text);
      tin100_GlobalLog::writeEntry (aEvent);
    } // end if
  } else {
    teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_SERVICEDB_CN00_1, szDBName.asCharp());
    tin100_GlobalLog::writeEntry (aEvent);
  } // end if
  
   return bReturn;
} // end cn40_RegisterServiceDB

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40_CheckInstKey
 * ------------------------------------------------------------------
 */
static bool cn40_CheckInstKey
      ( const tsp09_RteInstallationInfoNew  & relInfoIn )
{
  bool                           bFound = false;
  tsp9_rte_xerror                xError;
  void                         * pHandle;
  tsp09_RteInstallationInfoNew   relInfo;

  // open enum
  sqlxopen_installation_enum (&pHandle, &xError);

  if (xError.xe_result == csp9_xrte_ok) {
    // iterate trough installations
    while (xError.xe_result == csp9_xrte_ok && !bFound) {
      relInfo.RecordVersion = INSTALLATION_RECORD_VERSION_V721;
      relInfo.RecordLength  = sizeof(relInfo);
      sqlxnext_installationNew (pHandle, relInfo, &xError);
      if (xError.xe_result == csp9_xrte_ok) {
        bFound = ( _tcscmp ( relInfo.key,     relInfoIn.key                               ) == 0 &&
                   memcmp  ( &relInfo.version, &relInfoIn.version, sizeof(relInfo.version)) == 0 &&
#ifdef _WIN32
                   _tcsicmp( relInfo.dbroot,  relInfoIn.dbroot                            ) != 0    );
#else
                   _tcscmp ( relInfo.dbroot,  relInfoIn.dbroot                            ) != 0    );
#endif
      } // end if
    } // end while

    // close enum
    sqlxclose_installation_enum(pHandle);
  } // end if

  return bFound;
} // end cn40_CheckInstKey

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40InstEnum
 * ------------------------------------------------------------------
 */
tcn00_Error cn40InstEnum
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error                    nFuncReturn = OK_CN00;
  tsp9_rte_xerror                xError;
  void                      *    pHandle;
  tsp09_RteInstallationInfoNew   relInfo;
  _TCHAR                       * pCurrent    = replyData;
  _TCHAR                         szMode[PARAMETER_MAXLEN_CN90];
  bool                           bKey = false;
  int                            nLength;

  if (cn90GetToken(command->args, szMode, 1)) {
    if (_tcsicmp(szMode, OPTS_INST_KEY) == 0) {
      bKey = true;
    } else {
      nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    // init message
    _stprintf(pCurrent, "%s%s", ANSWER_OK_CN00,  LINE_SEPSTRING_CN00);
    pCurrent = pCurrent + _tcslen(pCurrent);

    // open enum
    sqlxopen_installation_enum (&pHandle, &xError);
    
    if (xError.xe_result == csp9_xrte_ok) {
      // iterate trough installations
      while (xError.xe_result == csp9_xrte_ok) {
        relInfo.RecordVersion = INSTALLATION_RECORD_VERSION_V721;
        relInfo.RecordLength  = sizeof(relInfo);
        sqlxnext_installationNew (pHandle, relInfo, &xError);
        if (xError.xe_result == csp9_xrte_ok) {
          _stprintf (pCurrent, _T("%d.%d.%d.%d"),
                     relInfo.version.no [0],
                     relInfo.version.no [1],
                     relInfo.version.no [2],
                     relInfo.version.no [3]);
          nLength  = (int)_tcslen(pCurrent);
          pCurrent = pCurrent + nLength;
          _stprintf (pCurrent, _T("%*s"), 13 - ((nLength >= 13) ? 12 : nLength), _T(" "));
          pCurrent = pCurrent + _tcslen(pCurrent);

          if (bKey) {
            _stprintf (pCurrent, _T("%-*s "),
                       18,
                       relInfo.key.asCharp());
            pCurrent = pCurrent + _tcslen(pCurrent);
          } // end if
          _stprintf (pCurrent, _T("%s%s"),
                     relInfo.dbroot.asCharp(),
                     LINE_SEPSTRING_CN00);
          pCurrent = pCurrent + _tcslen(pCurrent);
        } // end if

      } // end while

      sqlxclose_installation_enum(pHandle);

    } // end if

    if (xError.xe_result != csp9_xrte_ok && xError.xe_result != csp9_xrte_atend) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } else {
      *replyLen = (int)_tcslen(replyData);
    } // end if

  } // end if


  return nFuncReturn;
} /* end cn40InstEnum */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40InstReg
 * ------------------------------------------------------------------
 */
tcn00_Error cn40InstReg
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  FUNCTION_DBG_MCN00_1(_T("cn40InstReg"));

  tcn00_Error                    nFuncReturn                      = OK_CN00;
  tsp09_RteInstallationInfoNew   relInfo;
  _TCHAR                         szToken[PARAMETER_MAXLEN_CN90]   = "";
  _TCHAR                         szVersion[PARAMETER_MAXLEN_CN90] = "";
  _TCHAR                       * pToken;
  int                            nIndex;
  int                            nToken                           = 1;
  tsp9_rte_xerror                xError;
  tsp100_VersionID0              VersionID0;
  bool                           bClient = false;
  
  relInfo.key.Init();
  relInfo.dbroot.Init();

  // read parameters is Version or switch
  while (cn90GetToken(command->args, szToken, nToken)) {
    if (_tcsicmp(szToken, OPTS_INST_VERSION) == 0) {
      ++nToken;
      if (!cn90GetToken(command->args, szVersion, nToken)) {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } // end if
    } else if (_tcsicmp(szToken, OPTS_INST_KEY) == 0) {
      ++nToken;
      if (!cn90GetToken(command->args, szToken, nToken)) {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } else {
        relInfo.key.rawAssign(szToken);
      } // end if
    } else if (_tcsicmp(szToken, OPTS_INST_CLIENT) == 0) {
      bClient = true;
    } else {
      if (nToken == 1) {
        relInfo.dbroot.rawAssign(szToken);
      } else if (nToken == 2) {
        _tcscpy(szVersion , szToken);
      } else {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } // end if
    } // end if
 
    ++nToken;    
  } // end while

  // check dbroot
  if (nFuncReturn == OK_CN00) {

    if (relInfo.dbroot.length() == 0) {
      relInfo.dbroot = vcontrol->dbroot;
    } else {
      if (
#ifndef _WIN32
        _tcscmp( relInfo.dbroot, vcontrol->dbroot )  != 0
#else
        _tcsicmp(relInfo.dbroot, vcontrol->dbroot ) != 0
#endif
         ) {
        teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_WRONGINST_CN00_1, relInfo.dbroot.asCharp(), vcontrol->dbroot.asCharp());
        nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);

      } // end if
    } // end if

  } // end if

  // checking parameters
  if (nFuncReturn == OK_CN00) {
    for (nIndex = 0; nIndex < csp9_version_digits; nIndex++) {
      relInfo.version.no [nIndex] = STATIC_CAST(signed char, 0);
    } // end for
    // reading verison
    if (_tcslen(szVersion) == 0) {
      sp100_GetVersionID ( VersionIDType0_esp100, s100buildnumberEx, &VersionID0 );

      relInfo.version.no [0] = STATIC_CAST(signed char, VersionID0.MajorVersion_sp100);
      relInfo.version.no [1] = STATIC_CAST(signed char, VersionID0.MinorVersion_sp100);
      relInfo.version.no [2] = STATIC_CAST(signed char, VersionID0.CorrLevel_sp100);
    } else {
      // Read Version from arguments
      pToken = strtok(szVersion, ".");
      nIndex = 0;
      while (pToken && nIndex < csp9_version_digits) {
        relInfo.version.no [nIndex] = STATIC_CAST(signed char, atoi(pToken));
        nIndex++;
        pToken = strtok(NULL, ".");
      } // end while
    } // end if

    // checking key
    if (cn40_CheckInstKey(relInfo)) {
      teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, 
                              ERR_KEYEXISTS_CN00_1,
                              (relInfo.key.length() == 0) ? _T("(empty)") : relInfo.key.asCharp()); 
      nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
    } // end if
  } // end if

  // register version
  if (nFuncReturn == OK_CN00) {
    relInfo.RecordVersion = INSTALLATION_RECORD_VERSION_V721;
    relInfo.RecordLength  = sizeof(relInfo);

    sqlxregisterinstNew (relInfo, &xError);

    if (xError.xe_result != csp9_xrte_ok) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } else {
      if (!bClient) {
        if (!cn40_RegisterServiceDB(relInfo)) {
          teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, 
                                  ERR_SERVICEDB_CN00_1,
                                  relInfo.dbroot.asCharp());
          nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
        } else {
          cn90AnswerOK (replyData, replyLen, NULL);
        } // end if
      } else {
        cn90AnswerOK (replyData, replyLen, NULL);
      } // end if
    } // end if

  } // end if

  return nFuncReturn;

} /* end cn40InstReg */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40InstUnreg
 * ------------------------------------------------------------------
 */
tcn00_Error cn40InstUnreg
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  FUNCTION_DBG_MCN00_1(_T("cn40InstUnreg"));

  tcn00_Error          nFuncReturn = OK_CN00;
  _TCHAR               szToken[PARAMETER_MAXLEN_CN90];
  tsp00_Pathc          szDbRoot;
  tsp00_DbNamec        szDbName;
  tsp9_rte_xerror      xError;

  cn90GetToken(command->args, szToken, 1);
  cn90StripQuotes(szToken);
  szDbRoot.rawAssign(szToken);

  // check dbroot

  if (szDbRoot.length() == 0) {
    szDbRoot = vcontrol->dbroot;
//  } else {
//    if (
//#ifndef _WIN32
//      _tcscmp( szDbRoot, vcontrol->dbroot )  != 0
//#else
//      _tcsicmp(szDbRoot, vcontrol->dbroot ) != 0
//#endif
//       ) {
//      teo200_EventList aEvent(FUNCTION_NAME_MCN00_1, ERR_WRONGINST_CN00_1, szDbRoot.asCharp(), vcontrol->dbroot.asCharp());
//      nFuncReturn = cn90AnswerEvent(replyData, replyLen, aEvent);
//
//    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90ServiceDB(szDbName, szDbRoot);
    if (szDbName.length() > 0) {
      cn40_DBDrop(szDbName, szDbRoot, true, replyData, replyLen, replyLenMax);
    } // end if
  
    sqlxunregisterinst (szDbRoot, &xError);

    if (xError.xe_result != csp9_xrte_ok) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } else {
      cn90AnswerOK (replyData, replyLen, NULL);
    } /* end if */

  } // end if

  return nFuncReturn;

} /* end cn40InstUnreg */

/*
 * IMPLEMENTATION PRIVATE FUNCTION cn40_ComputeEsc
 * ------------------------------------------------------------------
 */
#define ESC_MARK _T('#')

static void cn40_ComputeEsc
      ( tsp00_Pathc & szFile )
{
  tsp00_Int4 nLength = 0;
  tsp00_Int4 nSource = 0;
  tsp00_Int4 nTarget = 0;
  unsigned char nDig1   = 0;
  unsigned char nDig2   = 0;
  unsigned char nChar   = 0;

  nLength = szFile.length();
  while (nSource < nLength) {

    nDig1   = 0;
    nDig2   = 0;
    nChar   = 0;

    if (szFile[nSource] == ESC_MARK) {
      ++nSource;
      nDig1 = szFile[nSource];
      if (nDig1 == ESC_MARK) {
        nChar = nDig1;
      } else {
        ++nSource;
        nDig2 = szFile[nSource];

        // Konverting to upper
        nDig1 = _totupper(nDig1);
        nDig2 = _totupper(nDig2);
        // checking hexdigit
        nDig1 = (nDig1 >= _T('0') && nDig1 <= _T('9') ||
                 nDig1 >= _T('A') && nDig1 <= _T('F')    ) ? nDig1 : 0;
        nDig2 = (nDig2 >= _T('0') && nDig2 <= _T('9') ||
                 nDig2 >= _T('A') && nDig2 <= _T('F')    ) ? nDig2 : 0;
        // checking second digit
        nDig1 = (nDig2 == 0) ? 0 : nDig1;
        nDig2 = (nDig1 == 0) ? 0 : nDig2;

        if (nDig1 > 0 && nDig2 > 0) {
          nDig1 = (nDig1 <= _T('9')) ? nDig1 - 0x30 : nDig1 - 0x41;
          nDig2 = (nDig2 <= _T('9')) ? nDig2 - 0x30 : nDig2 - 0x41;

          nChar = (nDig1 * 0x10) + nDig2;
        } else {
          --nSource;
          --nSource;
          nChar = ESC_MARK;
        } // end if

      } // end if
    } else {
      nChar = szFile[nSource];
    } // end if

    if (nChar != 0) {
      szFile[nTarget] = nChar;
      ++nTarget;
    } // end if

    ++nSource;
  } // end while

   szFile[nTarget] = 0;

} // end cn40_ComputeEsc

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40DbAddDevice
 * ------------------------------------------------------------------
 */
tcn00_Error cn40DbAddDevice
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  FUNCTION_DBG_MCN00_1("cn40DbAddDevice");

  tcn00_Error            nFuncReturn    = OK_CN00;
  bool                   bMirror     = false;
  char                   szMode  [PARAMETER_MAXLEN_CN90];
  char                   szType  [PARAMETER_MAXLEN_CN90];
  char                   szMType [PARAMETER_MAXLEN_CN90];
  char                   szSize  [PARAMETER_MAXLEN_CN90];
  tsp00_Pathc            szName;
  tsp00_Pathc            szMName;
  tcn00_DBState          aState;
  char                   szCommand [PARAMETER_MAXLEN_CN90];
  int                    nNumber;
  tcn00_DevspaceKeywords aKeyTable  [] = XP_DEVSPACE_KEYWORDS_CN00;
  tcn00_DevspaceParams   aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  tsp00_Int4             nKey          = 0;
  tsp00_Int4             nParam        = 0;
  tsp00_Int4             nMaxSize      = 0;
  tsp00_Int4             nSize         = 0;
  tsp00_XpKeyTypec       szParam;
  tsp00_C64c             szValue;
  tsp00_Int4             nValue;
  bool                   bParamWritten = false;

  cn90GetToken(command->args, szMode,   1);
  cn90GetToken(command->args, szName,   2);
  cn90GetToken(command->args, szType,   3);
  cn90GetToken(command->args, szSize,   4);
  cn90GetToken(command->args, szMName,  5);
  cn90GetToken(command->args, szMType,  6);

  // checking mode
  if (nFuncReturn == OK_CN00) {
    if (_tcsicmp(szMode, XP_DEV_DAT_CN00) != 0 &&
        _tcsicmp(szMode, XP_DEV_LOG_CN00) != 0    ) {
      nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  // checking dbstate
  if (nFuncReturn == OK_CN00) {
    aState = cn90DBState(vcontrol->dbname);
    if (aState != STATE_COLD_CN00 && aState != STATE_WARM_CN00) {
      nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_DBNORUN_CN00);
    } // end if
  } // end if

  // computing number and size
  if (nFuncReturn == OK_CN00) {
    /* search in keywordtable */
    for (nKey = 0; (aKeyTable[nKey].nDevSpace != XPDevSpaceUnknown_ecn00) &&
                   (_tcsicmp(aKeyTable[nKey].szText, szMode) != 0  );    nKey++);

    /* search in paramtable */
    for (nParam = 0; aKeyTable[nKey].nDevSpace != aParamTable[nParam].nDevSpace; nParam++);

    if (aKeyTable[nKey].nDevSpace == XPDevSpaceUnknown_ecn00) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } else {
      nNumber  = 0;
      nMaxSize = 0;
      do {
        nNumber++;
        _stprintf(szParam, aParamTable[nParam].szSize, nNumber);
        cn20XParamGetValue( vcontrol->dbname, szParam, nValue);
        nMaxSize = nMaxSize + nValue;
        _stprintf(szParam, aParamTable[nParam].szName, nNumber);
        cn20XParamGetValue( vcontrol->dbname, szParam, szValue);
      } while (_tcslen(szValue) > 0);
    } // end if
  } /* end if */

  // checking number
  if (nFuncReturn == OK_CN00) {
    nValue = 0;
    szParam.rawAssign(aParamTable[nParam].szMaxCount);
    cn20XParamGetValue( vcontrol->dbname, szParam, nValue);
    if (nValue < nNumber) {
      nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_MAXDEVCNT_CN00);
    } // end if
  } // end if

  // checking mirrordata
  if (nFuncReturn == OK_CN00) {
    szValue.Init();
    if (_tcsicmp(szMode, XP_DEV_DAT_CN00) == 0) {
      szParam.rawAssign(XP_ID_MIRROR_CN00);
    } else {
      szParam.rawAssign(XP_ID_LOG_MODE_CN00);
    } // end if

    cn20XParamGetValue( vcontrol->dbname, szParam, szValue);

    if (_tcsicmp(szValue, XP_VAL_DUAL_CN00) == 0 || _tcsicmp(szValue, XP_VAL_YES_CN00) == 0) {
      bMirror = true;
      if (szMName.length() == 0) {
        nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_MISSMIRR_CN00);
      } // end if
    } // end if
  } // end if

  // checking size
  if (nFuncReturn == OK_CN00) {
    nSize = atol(szSize);
    if (_tcsicmp(szMode, XP_DEV_DAT_CN00) == 0) {
      szParam.rawAssign(aParamTable[nParam].szMaxSize);
      cn20XParamGetValue( vcontrol->dbname, szParam, nValue);
      if ( nMaxSize + nSize > nValue) {
        nFuncReturn =  cn90AnswerIError(replyData, replyLen, ERR_MAXDATA_CN00);
      } // end if
    } // end if
  } // end if

  // all tests passed

  // scanning names for escape sequences
  if (nFuncReturn == OK_CN00) {
    cn40_ComputeEsc(szName);
    if (bMirror) {
      cn40_ComputeEsc(szMName);
    } // end if

  } // end if

  // adding params
  if (nFuncReturn == OK_CN00) {
    _stprintf(szCommand, "%d %s %s %s %s",
                         nNumber,
                         szMode,
                         szName.asCharp(),
                         szType,
                         szSize);
    command->args = szCommand;
    nFuncReturn = cn20ParamAddDevSpace(vcontrol,
                                       command,
                                       replyData,
                                       replyLen,
                                       replyLenMax);
    bParamWritten = (nFuncReturn == OK_CN00);
    if (bMirror && nFuncReturn == OK_CN00) {
      _stprintf(szCommand, "%d M%s %s %s",
                           nNumber,
                           szMode,
                           szMName.asCharp(),
                           szMType);
      command->args = szCommand;
      nFuncReturn = cn20ParamAddDevSpace(vcontrol,
                                         command,
                                         replyData,
                                         replyLen,
                                         replyLenMax);
    } // end if
  } // end if

  // utility
  if (nFuncReturn == OK_CN00) {
    if (!bMirror) {
      _stprintf(szCommand, "ADD %s DEVSPACE '%s' PAGES %s",
                           szMode,
                           szName.asCharp(),
                           szSize);
    } else {
      _stprintf(szCommand, "ADD %s DEVSPACE '%s' '%s' PAGES %s",
                           szMode,
                           szName.asCharp(),
                           szMName.asCharp(),
                           szSize);
    } // end if
    nFuncReturn = cn80ExecuteUtil (vcontrol->dbname, szCommand, replyData, replyLen);
  } // end if

  // restore param
  if (bParamWritten && nFuncReturn != OK_CN00) {
    char               szSource[PARAMETER_MAXLEN_CN90];
    char               szTarget[PARAMETER_MAXLEN_CN90];
    tsp00_C256         szConfig;
    tsp01_RteError     aRteError;

    cn90GetCfgPath(szConfig);
    sprintf(szTarget, "%s%s",  (char *) szConfig, vcontrol->dbname.asCharp());
    sprintf(szSource, "%s.01", (char *) szTarget);
    sqlfilecopyc( szSource, szTarget, &aRteError );
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} /* end cn40DbAddDevice */

/*
 * IMPLEMENTATION PUBLIC FUNCTION cn40Test
 * ------------------------------------------------------------------
 */
tcn00_Error cn40Test
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  FUNCTION_DBG_MCN00_1("cn40Test");

  tcn00_Error   nFuncReturn                    = OK_CN00;
  _TCHAR        szToken[PARAMETER_MAXLEN_CN90] = "";
  int           nToken                         = 1;
  long          lMilliseconds                  = 0;

  if (cn90GetToken(command->args, szToken, nToken)) {
    nToken++;
    if (_tcsicmp(szToken, TEST_OPT_WAIT) == 0) {
      if (cn90GetToken(command->args, szToken, nToken)) {
        nToken++;
        lMilliseconds = atol(szToken);
        sqlsleep((tsp00_Int2)(lMilliseconds/1000));
      } else {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      }
    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    }
  }

/*
  tin101_XDBOperation   aXDB(vcontrol->dbname,
                             vcontrol->pCurrentUser->getUserName(),
                             vcontrol->pCurrentUser->getClearMasterPwd());

  tsp00_Pathc                 szLogFile;

  aXDB.start(XDBOP_ENAUTO_IN101, XDBCO_DBM_IN101, szLogFile);

  sqlsleep(2);

  aXDB.run(XDBST_OK_IN101, "");
 
  sqlsleep(2);
  
  aXDB.complete(XDBST_OK_IN101);
*/
  if (nFuncReturn == OK_CN00) {
      cn90AnswerOK (replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} /* end cn40Test */

/* ------------------------------------------------------------------
 * START OF IMPLEMENTATION PRIVATE
 * ------------------------------------------------------------------
 */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_ExecCommand
 * ------------------------------------------------------------------
 */
static tcn00_Error cn40_ExecCommand
      ( char          * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const char    * szProgram,
        const char    * szOutput)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  int             nSysReturn  = 0;
  char        *   szTmpName   = NULL;
  char        *   szTmpDir    = NULL;
  char            szDummy[PARAMETER_MAXLEN_CN90];
  char            szTmpFile[PARAMETER_MAXLEN_CN90] = "";
  char            szCommand[PARAMETER_MAXLEN_CN90] = "";

  /* generate temp file name */
  if (nFuncReturn == OK_CN00) {
    szTmpName = tmpnam(szDummy);
    if (szTmpName == NULL) {
      nFuncReturn = ERR_SYSTEM_CN00;

    } else {

#ifdef _WIN32
      if ((szTmpDir = getenv(TMP_ENV)) == NULL) {
        sprintf(szTmpFile, "%s%s%s", TMP_LOCDIR, szTmpName, TMP_EXT);
      } else {
        char  szShortTmp[PARAMETER_MAXLEN_CN90] = "";
        GetShortPathName(szTmpDir, szShortTmp, PARAMETER_MAXLEN_CN90);
        sprintf(szTmpFile, "%s%s.%s", szShortTmp,   szTmpName, TMP_EXT);
      } /* end if */
#else
      strcpy(szTmpFile, szTmpName);
#endif

    } /* end if */
  } /* end if */

  /* execute command */
  if (nFuncReturn == OK_CN00) {
    if (strlen(szProgram) + strlen(szTmpFile) + 10 < PARAMETER_MAXLEN_CN90) {

#ifdef _WIN32
      sprintf(szCommand, "%s 1>%s 2>&1", szProgram, szTmpFile);
#else
      sprintf(szCommand, "%s 1>%s 2>&1", szProgram, szTmpFile);
#endif
      nSysReturn = system(cn90Strip(szCommand));

      if (nSysReturn == SYS_SYSTEM_ERR) {
        nFuncReturn = ERR_SYSTEM_CN00;
      } else if (nSysReturn != 0) {
        nFuncReturn = ERR_EXECUTE_CN00;
      } /* end if */

    } else {
      nFuncReturn = ERR_MEM_CN00;
    } /* end if */
  } /* end if */

  /* check result */
  if (nFuncReturn == OK_CN00 || nFuncReturn == ERR_EXECUTE_CN00) {
    cn90AnswerExecute(replyData, replyLen, replyLenMax, nFuncReturn, nSysReturn, szOutput, szTmpFile);
  } else if (nFuncReturn == ERR_SYSTEM_CN00) {
    cn90AnswerErrnoError(replyData, replyLen, nFuncReturn, errno, szCommand);
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  /* remove temporary files */
  remove(szTmpFile);

  return nFuncReturn;
} /* end cn40_ExecCommand */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_ExecXLoad
 * ------------------------------------------------------------------
 */
static tcn00_Error cn40_ExecXLoad
      ( VControlDataT * vcontrol,
        const char    * szLoadFile,
        const char    * szUser,
        const char    * szPassword,
        const char    * szOptions,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  bool            bAbsolute   = false;
  unsigned int    nIndex      = 0;
  FILE        *   hFile       = NULL;
  tsp00_C256      szDir;
  char            szCommand   [PARAMETER_MAXLEN_CN90] = "";
  char            szFile      [PARAMETER_MAXLEN_CN90] = "";
  char            szTemp      [PARAMETER_MAXLEN_CN90] = "";
  char            szLine      [PARAMETER_MAXLEN_CN90] = "";

  /* create command filename */
  strcpy(szFile, szLoadFile);

  if (nFuncReturn == OK_CN00) {
    for (nIndex = 0; nIndex < strlen(szFile); nIndex++) {
      szFile[nIndex] = (szFile[nIndex] == CHAR_BACKSLASH) ? CHAR_SLASH : szFile[nIndex];
    } /* end for */

    if ( !(isalpha(szFile[0]) && szFile[1] == CHAR_DRIVE && szFile[2] == CHAR_SLASH) &&
         !(szFile[0] == CHAR_SLASH) ) {
       strcpy(szTemp, cn90GetVersionEnvPath(szDir));
       strcat(szTemp, szFile);
       strcpy(szFile, szTemp);
    } /* end if */

    cn90GetVersionBinPath(szDir);

    tsp00_NodeId     hostP;      
    tsp00_NodeIdc    hostC;      
    tsp00_ErrText    errtext;
    tsp01_CommErr    rc;
    
    sqlhostname (hostP, errtext, &rc);
    if (rc != commErrOk_esp01) {
      hostC.Init();
    } else {
      hostC.p2c(hostP);
    } // end if


#ifdef _WIN32
    sprintf(szCommand, "\"\"%sxload\" -S INTERNAL %s %s -d %s -u %s,%s -b \\\"%s\\\" %s\"", (char *) szDir,
                                                                                    hostC.length() > 0 ? "-n" : "",
                                                                                    hostC.length() > 0 ? hostC.asCharp() : "",
                                                                                    &vcontrol->dbname[0],
                                                                                    szUser,
                                                                                    szPassword,
                                                                                    szFile,
                                                                                    szOptions);
    sprintf(szTemp,    "\"\"%sxload\" -S INTERNAL %s %s -d %s -u %s,%s -b \\\"%s\\\" %s\"", (char *) szDir,
                                                                                    hostC.length() > 0 ? "-n" : "",
                                                                                    hostC.length() > 0 ? hostC.asCharp() : "",
                                                                                    &vcontrol->dbname[0],
                                                                                    szUser,
                                                                                    "*",
                                                                                    szFile,
                                                                                    szOptions);
#else
    sprintf(szCommand, "%sxload -S INTERNAL %s %s -d %s -u %s,%s -b %s %s", (char *) szDir,
                                                                    hostC.length() > 0 ? "-n" : "",
                                                                    hostC.length() > 0 ? hostC.asCharp() : "",
                                                                    &vcontrol->dbname[0],
                                                                    szUser,
                                                                    szPassword,
                                                                    szFile,
                                                                    szOptions);
    sprintf(szTemp,    "%sxload -S INTERNAL %s %s -d %s -u %s,%s -b %s %s", (char *) szDir,
                                                                    hostC.length() > 0 ? "-n" : "",
                                                                    hostC.length() > 0 ? hostC.asCharp() : "",
                                                                    &vcontrol->dbname[0],
                                                                    szUser,
                                                                    "*",
                                                                    szFile,
                                                                    szOptions);
#endif
    /* remove old protocol */
    remove(XLOAD_PRT);
    remove(XLOAD_PROT);

    nFuncReturn = cn40_ExecCommand(replyData, replyLen, replyLenMax, szCommand, szTemp);

  } /* end if */

  /* add protocol file */
  if (nFuncReturn == OK_CN00 || nFuncReturn == ERR_EXECUTE_CN00) {
    hFile = NULL;
    hFile = fopen(XLOAD_PRT, "r");
    if (hFile == NULL) {
      hFile = fopen(XLOAD_PROT, "r");
    } /* end if */

    if (hFile) {
      bool bSummary = false;
      while (fgets(szLine, PARAMETER_MAXLEN_CN90, hFile) != NULL && ((int) (strlen(replyData) + strlen(szLine) + 5) < replyLenMax)) {
        if (cn90CheckKeyword(szLine, "/ S") ||
            bSummary                           ) {
          bSummary = true;
          strcat(replyData, cn90StripTrailingChar(szLine, '\n'));
          strcat(replyData, LINE_SEPSTRING_CN00);
        } /* end if */
      } /* end while */
      fclose(hFile);
      *replyLen = (int)strlen(replyData);
    } /* end if */
  } /* end if */

  /* remove current protocol */
  // remove(XLOAD_PRT);
  // remove(XLOAD_PROT);

  return nFuncReturn;
} /* end cn40_ExecXLoad */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_ExecPythonLoad
 * ------------------------------------------------------------------
 */
static tcn00_Error cn40_ExecPythonLoad
      ( VControlDataT * vcontrol,
        const char    * szLoadFile,
        const char    * szUser,
        const char    * szPassword,
        const char    * szOptions,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error     nFuncReturn = OK_CN00;
  int             nIndex      = 0;
  FILE        *   hFile       = NULL;
  char            szCommand   [PARAMETER_MAXLEN_CN90] = "";
  char            szProtocol  [PARAMETER_MAXLEN_CN90] = "";
  char            szLine      [PARAMETER_MAXLEN_CN90] = "";
  tsp01_RteError  aRteError;
  tsp00_Pathc     szBinDir;
  tsp00_Pathc     szEnvDir;
  tsp00_Pathc     szScript;
  tsp00_C264c     szProt;

  /* create command filename */
  szScript.rawAssign(szLoadFile);

  if (nFuncReturn == OK_CN00) {
    // remove backslashes
    for (nIndex = 0; nIndex < szScript.length(); nIndex++) {
      szScript[nIndex] = (szScript[nIndex] == CHAR_BACKSLASH) ? CHAR_SLASH : szScript[nIndex];
    } /* end for */

    // if no absolute path for scriptfile use "env" directory
    szEnvDir.Init();
    if ( !(isalpha(szScript[0]) && szScript[1] == CHAR_DRIVE && szScript[2] == CHAR_SLASH) &&
         !(szScript[0] == CHAR_SLASH) ) {
      sqlGetDbrootEnvPath(szEnvDir, TERM_WITH_DELIMITER_EO01, &aRteError);
    } /* end if */

    // look for "bin" directory
    szBinDir.Init();
    sqlGetDbrootBinPath(szBinDir, TERM_WITH_DELIMITER_EO01, &aRteError);
    
#ifdef _WIN32
    char Tueddelchen[2] = "\"";
#else
    char Tueddelchen[2] = "";
#endif

    _stprintf(szCommand,  "%s%s%sx_python%s %s%s%s%s -d %s -u %s,%s %s%s", Tueddelchen,
                                                                           Tueddelchen,
                                                                           szBinDir.asCharp(),
                                                                           Tueddelchen,
                                                                           Tueddelchen,
                                                                           szEnvDir.asCharp(),
                                                                           szScript.asCharp(),
                                                                           Tueddelchen,
                                                                           vcontrol->dbname.asCharp(),
                                                                           szUser,
                                                                           szPassword,
                                                                           szOptions,
                                                                           Tueddelchen);
    _stprintf(szProtocol, "%s%s%sx_python%s %s%s%s%s -d %s -u %s,%s %s%s", Tueddelchen,
                                                                           Tueddelchen,
                                                                           szBinDir.asCharp(),
                                                                           Tueddelchen,
                                                                           Tueddelchen,
                                                                           szEnvDir.asCharp(),
                                                                           szScript.asCharp(),
                                                                           Tueddelchen,
                                                                           vcontrol->dbname.asCharp(),
                                                                           szUser,
                                                                           "*",
                                                                           szOptions,
                                                                           Tueddelchen);
    // remove old protocol
    cn42GetFileName(vcontrol->dbname, FGET_REPMAN_CN42, szProt);
    remove(szProt);

    // ... and GO!
    nFuncReturn = cn40_ExecCommand(replyData, replyLen, replyLenMax, szCommand, szProtocol);

  } //

  // read protocol file
#ifdef ELCH
  if (nFuncReturn == OK_CN00 || nFuncReturn == ERR_EXECUTE_CN00) {
    hFile = NULL;
    hFile = fopen(szProt, "r");

    if (hFile) {
      while (fgets(szLine, PARAMETER_MAXLEN_CN90, hFile) != NULL && ((int) (strlen(replyData) + strlen(szLine) + 5) < replyLenMax)) {
        if (cn90CheckKeyword(szLine, "/ M") ||
            cn90CheckKeyword(szLine, "/ E")    ) {
          strcat(replyData, cn90StripTrailingChar(szLine, '\n'));
          strcat(replyData, LINE_SEPSTRING_CN00);
        } /* end if */
      } /* end while */
      fclose(hFile);
      *replyLen = strlen(replyData);
    } /* end if */
  } /* end if */
#endif

  return nFuncReturn;
} /* end cn40_ExecPythonLoad */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_OfflineToCold
 * ------------------------------------------------------------------
 */
static tcn00_Error cn40_OfflineToCold
      ( const tsp00_DbNamec   szDBName,
        const tsp00_Pathc     szDBRoot,
        const tsp9_pgm_kind   pgmKind,
        char                * replyData,
        int                 * replyLen)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp9_rte_xerror     xError;
  tsp00_C64c          szValue;
  tsp00_DbNamec       szDBNameL = szDBName;

  // always try to create rundir
  if (cn20XParamGetValue(szDBNameL, XP_ID_RUNDIRECTORY_CN00, szValue) == OK_CN00) {
    cn90DirCreate(szValue, NULL);
  } // end if

  // Silently ignore 'crash state' in dboffline command
  if ( cn90DBState(szDBName) == STATE_RUIN_CN00 ) {
    tsp9_rte_xerror   xError;
    // clear database
    sqlxclear( szDBName, szDBRoot, &xError );
    // create rundir
  } // end if

  sqlxstart (szDBName, pgmKind, szDBRoot, false, 0, NULL, &xError);

  if (xError.xe_result != csp9_xrte_ok) {

    // check Versions
    nFuncReturn = cn90AnswerIError(replyData, replyLen, cn40_CheckVersion(szDBName));

    if (nFuncReturn == OK_CN00) {

      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);

      // try quick or slow
      if (pgmKind == csp9_any_pgm) {
        // try quick
        sqlxstart (szDBName, csp9_quick_pgm, szDBRoot, false, 0, NULL, &xError);
        if (xError.xe_result != csp9_xrte_ok) {
          // try slow
          sqlxstart (szDBName, csp9_slow_pgm, szDBRoot, false, 0, NULL, &xError);
          if (xError.xe_result == csp9_xrte_ok) {
            // reset to OK
            nFuncReturn = cn90AnswerOK (replyData, replyLen, NULL);
          } // end if
        } else {
          // reset to OK
          nFuncReturn = cn90AnswerOK (replyData, replyLen, NULL);
        } // end if
      } // end if
    } /* end if */

  } else {
    nFuncReturn = cn90AnswerOK (replyData, replyLen, NULL);
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn82RestoreEvents(szDBName);
  } // end if

  return nFuncReturn;
} // cn40_OfflineToCold

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_ColdToWarm
 * ------------------------------------------------------------------
 */
#undef  CN_FUNCTION
#define CN_FUNCTION "cn40_ColdToWarm"
static tcn00_Error cn40_ColdToWarm
      ( const tsp00_DbNamec & szDBName,
        const char          * pRestart,
        char                * replyData,
        int                 * replyLen)
{
  CONSOLE_TRACE_ENTER_CN00;
  
  tcn00_Error nFuncReturn = OK_CN00;

  cn90AnswerOK (replyData, replyLen, NULL);

  nFuncReturn = cn80ExecuteUtil(szDBName, pRestart, replyData, replyLen);

  if (nFuncReturn == OK_CN00) {
    cn30RefreshMediaInDB(szDBName);
  } // end if

  return nFuncReturn;

} // cn40_ColdToWarm

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_WarmToCold
 * ------------------------------------------------------------------
 */
#undef  CN_FUNCTION
#define CN_FUNCTION "cn40_WarmToCold"
static tcn00_Error cn40_WarmToCold
      ( const tsp00_DbNamec   szDBName,
        const bool            bImmediate,
        char                * replyData,
        int                 * replyLen)
{
  CONSOLE_TRACE_ENTER_CN00;

  tcn00_Error     nFuncReturn = cn90AnswerOK (replyData, replyLen, NULL);;

  if (bImmediate) {
    nFuncReturn = cn80ExecuteUtil(szDBName, "SHUTDOWN QUICK", replyData, replyLen);
  } else {
    nFuncReturn = cn80ExecuteUtil(szDBName, "SHUTDOWN", replyData, replyLen);
  } // end if

  return nFuncReturn;
} // cn40_WarmToCold

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_ColdToOffline
 * ------------------------------------------------------------------
 */
#undef  CN_FUNCTION
#define CN_FUNCTION "cn40_ColdToOffline"
static tcn00_Error cn40_ColdToOffline
      ( const tsp00_DbNamec   szDBName,
        const tsp00_Pathc     szDBRoot,
        const bool            bDump,
        char                * replyData,
        int                 * replyLen)
{
  CONSOLE_TRACE_ENTER_CN00;

  tcn00_Error         nFuncReturn = OK_CN00;
  tsp9_rte_xerror     xError;

  sqlxstop (szDBName, szDBRoot, bDump, &xError);

  if ((xError.xe_result != csp9_xrte_ok) &&  (cn90DBState(szDBName) != STATE_OFFLINE_CN00)) {
    nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
  } else {
    cn90AnswerOK (replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} // cn40_ColdToOffline

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_CheckVersion
 * ------------------------------------------------------------------
 */
#undef  CN_FUNCTION
#define CN_FUNCTION "cn40_CheckVersion"
static tcn00_Error cn40_CheckVersion
      ( const tsp00_DbNamec   szDBName)
{
  CONSOLE_TRACE_ENTER_CN00;

  tcn00_Error     nFuncReturn = OK_CN00;
  tsp00_Versionc  szMyVersion;
  tsp00_C64c      szParamVersion;
  tsp00_DbNamec   szLocDBName = szDBName;

  if (cn20XParamGetValue(szLocDBName, XP_ID_KERNELVERSION_CN00, szParamVersion) == OK_CN00) {
    sp100_GetVersionString(COMP_NAME_KERNEL_SP100, s100buildnumber, szMyVersion);

    if ((_tcsncmp ( szParamVersion + BEG_OF_1ST_VERS_SUBSTR_SP100,
                    szMyVersion    + BEG_OF_1ST_VERS_SUBSTR_SP100,
                    END_OF_1ST_VERS_SUBSTR_SP100  - BEG_OF_1ST_VERS_SUBSTR_SP100   + 1 ) != 0 ) ||
        (_tcsncmp ( szParamVersion + BEG_OF_2CND_VERS_SUBSTR_SP100,
                    szMyVersion    + BEG_OF_2CND_VERS_SUBSTR_SP100,
                    END_OF_2CND_VERS_SUBSTR_SP100 - BEG_OF_2CND_VERS_SUBSTR_SP100  + 1 ) != 0 )     ) {
      nFuncReturn = ERR_VERSION_CN00;
    } // end if
  } // end if

  return nFuncReturn;
} // cn40_CheckVersion

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_StartStopOptions
 * ------------------------------------------------------------------
 */
#undef  CN_FUNCTION
#define CN_FUNCTION "cn40_StartStopOptions"
static tcn00_Error cn40_StartStopOptions
      ( const _TCHAR  * szParameters,
        tsp9_pgm_kind & pgmKind,
        bool          & bDump,
        bool          & bImmediate,
        _TCHAR        * pRestart)
{
  CONSOLE_TRACE_ENTER_CN00;

  tcn00_Error     nFuncReturn = OK_CN00;
  _TCHAR          szToken   [PARAMETER_MAXLEN_CN90];
  _TCHAR          szToken1  [PARAMETER_MAXLEN_CN90];
  int             nToken = 1;

  // init Options
  pgmKind    = csp9_any_pgm;
  bDump      = false;
  bImmediate = false;
  _tcscpy(pRestart, TXT_RESTART);

  while (cn90GetToken(szParameters, szToken, nToken) && nFuncReturn == OK_CN00) {

    if (_tcsicmp(szToken, OPT_SPEED_FAST) == 0 || _tcsicmp(szToken, OPTS_SPEED_FAST) == 0) {
      pgmKind = csp9_fast_pgm;
    } else if (_tcsicmp(szToken, OPT_SPEED_QUICK) == 0 || _tcsicmp(szToken, OPTS_SPEED_QUICK) == 0) {
      pgmKind = csp9_quick_pgm;
    } else if (_tcsicmp(szToken, OPT_SPEED_SLOW) == 0 || _tcsicmp(szToken, OPTS_SPEED_SLOW) == 0) {
      pgmKind = csp9_slow_pgm;
    } else if (_tcsicmp(szToken, OPT_STOP_IMMEDIATE) == 0 || _tcsicmp(szToken, OPTS_STOP_IMMEDIATE) == 0) {
      bImmediate = true;
    } else if (_tcsicmp(szToken, OPT_STOP_DUMP) == 0 || _tcsicmp(szToken, OPTS_STOP_DUMP) == 0) {
      bDump = true;
    } else if (_tcsicmp(szToken, OPT_RESTART_UNTIL) == 0 || _tcsicmp(szToken, OPTS_RESTART_UNTIL) == 0) {
      ++nToken;
      cn90GetToken(szParameters, szToken,  nToken);
      ++nToken;
      cn90GetToken(szParameters, szToken1, nToken);
      _stprintf(pRestart, TXT_RESTART_UNTIL, szToken, szToken1);
    } else {
      nFuncReturn = ERR_PARAM_CN00;
    } // end if

    ++nToken;
  } // end while

  return nFuncReturn;
} // cn40_StartStopOptions

  /* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn40_ChangeDBState
 * ------------------------------------------------------------------
 */
#undef  CN_FUNCTION
#define CN_FUNCTION "cn40_ChangeDBState"
static tcn00_Error cn40_ChangeDBState
      ( VControlDataT * vcontrol,
        CommandT      * command,
        char          * replyData,
        int           * replyLen,
        int             replyLenMax,
        tcn00_DBState   targetState,
        bool            bStop)
{
  CONSOLE_TRACE_ENTER_CN00;

  tcn00_Error     nFuncReturn = OK_CN00;
  tsp9_pgm_kind   pgmKind;
  bool            bDump;
  bool            bImmediate;
  _TCHAR          szRestart [PARAMETER_MAXLEN_CN90] = "";
  tcn00_DBState   currentState;

  cn51DBMConfigValue  cfgAuto (vcontrol->dbname, CFG_AUTOSSAVE_CN51);

  // read Options
  nFuncReturn = cn90AnswerIError(replyData, replyLen, cn40_StartStopOptions(command->args, pgmKind, bDump, bImmediate, szRestart));

  if (nFuncReturn == OK_CN00) {
    currentState = cn90DBState(vcontrol->dbname);

    // cancel utility session
    if ( ( currentState == STATE_WARM_CN00 ||
           currentState == STATE_COLD_CN00    ) &&
         ( currentState != targetState        ) &&
         ( vcontrol->pUtilSession != NULL     )    ) {
         cn80UtilRelease(vcontrol, command, replyData, replyLen, replyLenMax);
    } // end if

    currentState = bStop                              &&
//                   currentState != STATE_OFFLINE_CN00 &&
//                   currentState != STATE_RUIN_CN00    &&
                   targetState == STATE_OFFLINE_CN00     ? STATE_COLD_CN00 :
                                                           currentState;

    switch (currentState) {

      case STATE_OFFLINE_CN00:  // Ist
      case STATE_RUIN_CN00:     // Ist

        switch (targetState) {
          case STATE_COLD_CN00:  // Soll
            cn20CalcDevices( vcontrol, command, replyData, replyLen, replyLenMax);
            nFuncReturn = cn40_OfflineToCold(vcontrol->dbname, vcontrol->dbroot, pgmKind, replyData, replyLen);
            break;

          case STATE_WARM_CN00:  // Soll
            cn20CalcDevices( vcontrol, command, replyData, replyLen, replyLenMax);
            nFuncReturn = cn40_OfflineToCold(vcontrol->dbname, vcontrol->dbroot, pgmKind, replyData, replyLen);
            if (nFuncReturn == OK_CN00) {
              nFuncReturn = cn40_ColdToWarm(vcontrol->dbname, szRestart, replyData, replyLen);
            } // end if
            if (cfgAuto == 1 && nFuncReturn == OK_CN00) {
              cn31AutosaveOn(vcontrol);
            } // end if
            break;

        } // end switch
        break;

      case STATE_COLD_CN00: // Ist

        switch (targetState) {
          case STATE_OFFLINE_CN00: // Soll
            nFuncReturn = cn40_ColdToOffline(vcontrol->dbname, vcontrol->dbroot, bDump, replyData, replyLen);
            break;

          case STATE_WARM_CN00:    // Soll
            nFuncReturn = cn40_ColdToWarm(vcontrol->dbname, szRestart, replyData, replyLen);
            if (cfgAuto == 1 && nFuncReturn == OK_CN00) {
              cn31AutosaveOn(vcontrol);
            } // end if
            break;

        } // end switch
        break;

      case STATE_WARM_CN00: // Ist

        switch (targetState) {
          case STATE_OFFLINE_CN00: // Soll
            if (cfgAuto == 1) {
              cn31AutosaveEnd(vcontrol);
            } // end if
            nFuncReturn = cn40_WarmToCold(vcontrol->dbname, bImmediate, replyData, replyLen);
            if (nFuncReturn == OK_CN00) {
              nFuncReturn = cn40_ColdToOffline(vcontrol->dbname, vcontrol->dbroot, bDump, replyData, replyLen);
            } // end if
            break;

          case STATE_COLD_CN00:    // Soll
            if (cfgAuto == 1) {
              cn31AutosaveEnd(vcontrol);
            } // end if
            nFuncReturn = cn40_WarmToCold(vcontrol->dbname, bImmediate, replyData, replyLen);
            break;

        } // end switch
        break;

      default:
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_STATE_CN00);
        break;
    } // end switch

  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } // end if

  return nFuncReturn;
} // end cn40_ChangeDBState
