/* 
  -----------------------------------------------------------------------------
 
  module: vcn42.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: DBMServer File Access
   
  description:  DBMServer File Access - Implementation

  -----------------------------------------------------------------------------


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*/

/*
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
 */

#include <stdio.h>
#include <stdlib.h>

#if  defined(CN71) || defined(CN72)

  #include "heo06.h"
  #include "hcn20.h"
  #include "hcn90.h"
  #include "hcn42.h"

#endif

/*
  -----------------------------------------------------------------------------
  specification private macros
  -----------------------------------------------------------------------------
 */
#define SEEK_MODE_CUR         _T("CUR")
#define OPEN_SWITCH           _T('-')
#define OPEN_MODE_BIN         _T('b')
#define OPEN_MODE_ASCII       _T('a')
#define OPEN_MODE_READ        _T('r')
#define OPEN_MODE_WRITE       _T('w')
#define OPEN_MODE_APPEND      _T('x')

#define FILE_OPT_NOCLOSE      _T("NOCLOSE")

#define FILEINFO_MULTI        _T("-m")
#define FILEINFO_TITLE        _T("filename,exists,readable,writable,islink,size,date_modified,time_modified,media_kind\n")
#define FILELIST_TITLE        _T("keyname,mode,size,date,time,comment,filename\n")
#define FILEINFO_MEDIATYPE    { _T("Unknown"),          \
                                _T("File"),             \
                                _T("Pipe"),             \
                                _T("Raw"),              \
                                _T("Tape (Norewind)"),  \
                                _T("Tape (Rewind)"),    \
                                _T("Tape"),             \
                                _T("Directory") }

#define CONT_FLAG_END         _T("     END")
#define CONT_FLAG_CONTINUE    _T("CONTINUE")

#define FGET_TOKENDEL         _T("#")
#define FGET_WILDCARD         _T("*")


#define FGET_LOC_DBROOT       1
#define FGET_LOC_WRK          2
#define FGET_LOC_RUNDIR       3
#define FGET_LOC_CONFIG       4
#define FGET_LOC_SAP          5

#define FGET_NAME_DB          ((const _TCHAR *) 1)
#define FGET_ORG_EXT          ((const _TCHAR *) 1)

#ifdef WIN32
  #define LO_EXT      _T(".prt")
  #define LCICMD_EXT  _T(".bat")
#else 
  #define LO_EXT _T(".prot")
  #define LCICMD_EXT  _T("")
#endif

//  File                   publ    bin    remove   XParam                   name                     ext          location          backupname     backupext          backuplocation   
#define FGET_FILES \
  {{FGET_KNLDIAG_CN42    , true  , false , true  , _T("_KERNELDIAGFILE")   , NULL                  , NULL       , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Database Messages")                },  \
   {FGET_KNLDIAGERR_CN42 , true  , false , true  , _T("_KERNELDIAGFILE")   , NULL                  , _T(".err") , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Database Errors")                  },  \
   {FGET_KNLDIAGOLD_CN42 , true  , false , true  , _T("_KERNELDIAGFILE")   , NULL                  , _T(".old") , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Database Messages (OLD)")          },  \
   {FGET_KNLTRC_CN42     , true  , true  , true  , _T("_KERNELTRACEFILE")  , NULL                  , NULL       , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Database Trace")                   },  \
   {FGET_KNLDUMP_CN42    , true  , true  , true  , _T("_KERNELDUMPFILE")   , NULL                  , NULL       , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Database Dump")                    },  \
   {FGET_KNLEVT_CN42     , true  , false , true  , _T("_EVENTFILE")        , NULL                  , NULL       , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Database Events")                  },  \
   {FGET_RTEDUMP_CN42    , true  , true  , true  , _T("_RTEDUMPFILE")      , NULL                  , NULL       , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Runtime Environment Dump")         },  \
   {FGET_UTLPRT_CN42     , true  , false , true  , _T("_UTILITY_PROTFILE") , NULL                  , NULL       , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_CONFIG  , _T("Utility Statements")               },  \
   {FGET_BACKHIST_CN42   , true  , false , true  , _T("_BACKUP_HISTFILE")  , NULL                  , NULL       , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_WRK     , _T("Backup History")                   },  \
   {FGET_BACKMDF_CN42    , true  , false , true  , _T("_BACKUP_MED_DEF")   , NULL                  , NULL       , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_WRK     , _T("Backup Media History")             },  \
   {FGET_BACKEBF_CN42    , true  , false , true  , NULL                    , _T("dbm")             , _T(".ebf") , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_WRK     , _T("External Backup History")          },  \
   {FGET_BACKEBP_CN42    , true  , false , true  , NULL                    , _T("dbm")             , _T(".ebp") , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_WRK     , _T("External Backup Protocol")         },  \
   {FGET_LOADPRT_CN42    , false , false , false , NULL                    , _T("load")            , LO_EXT     , FGET_LOC_DBROOT , _T("dbm")    , _T(".ins")     , FGET_LOC_RUNDIR  , _T("XLoad Protocol")                   },  \
   {FGET_XSRVPRT_CN42    , false , false , false , NULL                    , _T("xserver")         , _T(".prt") , FGET_LOC_WRK    , NULL         , NULL           , 0                , _T("XServer Messages")                 },  \
   {FGET_XSRVPRTOLD_CN42 , false , false , false , NULL                    , _T("xserver")         , _T(".old") , FGET_LOC_WRK    , NULL         , NULL           , 0                , _T("XServer Messages(OLD)")            },  \
   {FGET_REPMAN_CN42     , false , false , false , NULL                    , _T("repserver")       , _T(".log") , FGET_LOC_WRK    , _T("dbm")    , _T(".ins")     , FGET_LOC_RUNDIR  , _T("Replication Manager Protocol")     },  \
   {FGET_DBMPRT_CN42     , true  , false , true  , NULL                    , _T("dbm")             , _T(".prt") , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Database Manager Protocol")        },  \
   {FGET_DBMMDF_CN42     , true  , false , true  , NULL                    , _T("dbm")             , _T(".mmm") , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_CONFIG  , _T("Database Manager Media")           },  \
   {FGET_DBMPAHI_CN42    , true  , false , true  , NULL                    , FGET_NAME_DB          , _T(".pah") , FGET_LOC_CONFIG , NULL         , NULL           , 0                , _T("Database Parameter History")       },  \
   {FGET_CTRLMDF_CN42    , true  , false , true  , NULL                    , _T("control")         , _T(".med") , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_CONFIG  , _T("XControl Media")                   },  \
   {FGET_CTRLLOG_CN42    , true  , false , true  , NULL                    , _T("control")         , _T(".log") , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("XControl Protocol")                },  \
   {FGET_DBMCFG_CN42     , false , true  , true  , NULL                    , _T("dbm")             , _T(".cfg") , FGET_LOC_RUNDIR , FGET_NAME_DB , FGET_ORG_EXT   , FGET_LOC_CONFIG  , _T("Database Manager Configuration")   },  \
   {FGET_PARAM_CN42      , false , true  , false , NULL                    , FGET_NAME_DB          , NULL       , FGET_LOC_CONFIG , _T("param")  , FGET_ORG_EXT   , FGET_LOC_RUNDIR  , _T("Database Parameters")              },  \
   {FGET_LCINIT_CN42     , true  , false , false , NULL                    , _T("lcinit")          , _T(".log") , FGET_LOC_DBROOT , NULL         , NULL           , 0                , _T("LiveCache Initialisation")         },  \
   {FGET_LCINITCMD_CN42  , true  , false , false , NULL                    , _T("lcinit")          , LCICMD_EXT , FGET_LOC_SAP    , NULL         , NULL           , 0                , _T("LiveCache Initialisation Script")  },  \
   {FGET_LCINITHIS_CN42  , true  , false , false , NULL                    , _T("lcinit")          , _T(".his") , FGET_LOC_DBROOT , NULL         , NULL           , 0                , _T("LiveCache Initialisation History") },  \
   {FGET_APOCOMTRC_CN42  , true  , false , true  , NULL                    , _T("apo_com_trace*")  , _T(".txt") , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("APO Trace")                        },  \
   {FGET_LCTRC_CN42      , true  , false , true  , NULL                    , _T("lc*")             , NULL       , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("LiveCache Trace (ASCII)")          },  \
   {FGET_LCBIN_CN42      , true  , true  , true  , NULL                    , _T("lc*")             , NULL       , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("LiveCache Trace (binary)")         },  \
   {FGET_INSTPRT_CN42    , true  , false , true  , NULL                    , _T("dbm")             , _T(".ins") , FGET_LOC_RUNDIR , NULL         , NULL           , 0                , _T("Installation Protocol")            },  \
   {FGET_USER_CN42       , false , true  , false , NULL                    , FGET_NAME_DB          , _T(".upc") , FGET_LOC_CONFIG , NULL         , NULL           , 0                , _T("User Profiles")                    },  \
   {FGET_KNLTRCPRT_CN42  , true  , false , true  , NULL                    , FGET_NAME_DB          , _T(".prt") , FGET_LOC_DBROOT , NULL         , NULL           , 0                , _T("Kernel Trace Protocol")            },  \
   {NULL                 , false , false , false , NULL                    , NULL                  , NULL       , 0               , NULL         , NULL           , 0                , _T("")                                 }}


#define ASCII_OR_BINARY(a)  (a ? _T("BINARY") : _T("ASCII"))

typedef struct tcn42FGetFiles {
    const _TCHAR * szType;
    bool           bPublic;
    bool           bBinary;
    bool           bRemove;
    const _TCHAR * szXParam;
    const _TCHAR * szName;
    const _TCHAR * szExt;
    long           nLocation;
    const _TCHAR * szBaName;
    const _TCHAR * szBaExt;
    long           nBaLocation;
    const _TCHAR * szComment;
} tcn42FGetFiles;

/*  
  -----------------------------------------------------------------------------
  specification private functions  
  -----------------------------------------------------------------------------
 */  
static void cn42_RemoveFile
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType);

static void cn42_PrintFileInfo  
      ( _TCHAR     *szName,  
        _TCHAR     *szBuffer );  
  
static tcn00_Error cn42_FileRead  
      ( const _TCHAR  * szArguments,
        _TCHAR        * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const bool      bNext);

static tcn00_Error cn42_FileReadNext
      ( _TCHAR           * replyData,
        int              * replyLen,
        const int          replyLenMax,
        const tsp_int4     hFile,
        const tsp_longint  nSize,
        const bool         bNoClose,
        const bool         bAscii );

static tcn00_Error cn42_FileWrite
      ( const _TCHAR  * szArguments,
        _TCHAR        * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const bool      bNext);

static tcn00_Error cn42_FileWriteNext
      ( _TCHAR        * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const tsp_int4  hFile,
        const bool      bAscii,
        const long      nData,
        const _TCHAR  * pData );

static bool cn42_FileMatch
      ( tsp00_C264c        szFileName, 
        tsp00_C256c        szRealName, 
        tsp00_C264c      & szToken);

static tcn00_Error cn42_ReadDir
      ( void               * hDirectory, 
        _TCHAR             * szFirst, 
        _TCHAR             * replyData, 
        int                  replyLenMax,
        tsp5_rte_fileerror * pRteFileErr);

/*
  -----------------------------------------------------------------------------
  implementation public functions
  -----------------------------------------------------------------------------
 */

/* 
  -----------------------------------------------------------------------------
  public function cn42DirCreate
  -----------------------------------------------------------------------------
*/
tcn00_Error cn42DirCreate
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error    nFuncReturn = OK_CN00;
  _TCHAR         szDirectory [PARAMETER_MAXLEN_CN90];
  tsp01_RteError aRteError;

  if (nFuncReturn == OK_CN00) {
    if (!cn90GetToken(command->args, szDirectory, 1) || cn90GetToken(command->args, NULL, 2)) {
      nFuncReturn = ERR_PARAM_CN00;
    } /* end if */
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn90DirCreate(szDirectory, &aRteError);
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } else if (nFuncReturn == ERR_RTEEXT_CN00) {
    cn90AnswerNewRTEError (replyData, replyLen, nFuncReturn, &aRteError );
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  return nFuncReturn;
} /* end cn42DirCreate */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn42DirOpen
 * ------------------------------------------------------------------
 */
tcn00_Error cn42DirOpen
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  void              * hDirectory  = NULL;  
  tsp_c256            szDirectory;
  tsp_c256            szFirst;
  tsp5_rte_fileerror  rteFileErr;
  bool                bOpen       = false;
  unsigned int        nIndex      = 0;

  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if (!cn90GetToken(command->args, szDirectory, 1) || cn90GetToken(command->args, NULL, 2)) {
      nFuncReturn = ERR_PARAM_CN00;
    } /* end if */
  } /* end if */

  /* open dir */
  if (nFuncReturn == OK_CN00) {
    
    cn90ChangePathDelimiter(szDirectory);
    sqlfopendirc ( &szDirectory, &hDirectory,  &szFirst, &rteFileErr);
    if  ( rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
    } else {
      bOpen = true;
    } /* end if */
  } /* end if */

  /* dir is open */
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn42_ReadDir(hDirectory, szFirst, replyData, replyLenMax, &rteFileErr);
  } /* end if */

  /* check result */
  if (nFuncReturn == OK_CN00) {
    *replyLen = (int)strlen(replyData);
  } else if (nFuncReturn == ERR_RTE_CN00) {
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  /* close on error */
  if (nFuncReturn != OK_CN00 && bOpen) {
    sqlfclosedirc (hDirectory, &rteFileErr);
  } /* end if */

  return nFuncReturn;
} /* end cn42DirOpen */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn42DirReadNext
 * ------------------------------------------------------------------
 */
tcn00_Error cn42DirReadNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szDirectory [PARAMETER_MAXLEN_CN90];
  void              * hDirectory  = NULL;  
  tsp5_rte_fileerror  rteFileErr;

  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if (!cn90GetToken(command->args, szDirectory, 1) || cn90GetToken(command->args, NULL, 2)) {
      nFuncReturn = ERR_PARAM_CN00;
    } /* end if */
  } /* end if */

  /* read dir */
  if (nFuncReturn == OK_CN00) {
    sscanf(szDirectory, "%x", &hDirectory);
#if  defined(CN72)
    rteFileErr.sp5fe_result.becomes(vf_ok);
#else
    rteFileErr.sp5fe_result = vf_ok;
#endif
    nFuncReturn = cn42_ReadDir(hDirectory, NULL, replyData, replyLenMax, &rteFileErr);
  } /* end if */

  /* check result */
  if (nFuncReturn == OK_CN00) {
    *replyLen = (int)strlen(replyData);
  } else if (nFuncReturn == ERR_RTE_CN00) {
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  /* close on error */
  if (nFuncReturn != OK_CN00 && hDirectory != NULL) {
    sqlfclosedirc (hDirectory, &rteFileErr);
  } /* end if */

  return nFuncReturn;
} /* end cn42DirReadNext */

/* ------------------------------------------------------------------
 * IMPLEMENTATION PUBLIC FUNCTION cn42DirClose
 * ------------------------------------------------------------------
 */
tcn00_Error cn42DirClose
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szDirectory [PARAMETER_MAXLEN_CN90];
  void              * hDirectory  = NULL;  
  tsp5_rte_fileerror  rteFileErr;

  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if (!cn90GetToken(command->args, szDirectory, 1) || cn90GetToken(command->args, NULL, 2)) {
      nFuncReturn = ERR_PARAM_CN00;
    } /* end if */
  } /* end if */

  /* close dir */
  if (nFuncReturn == OK_CN00) {
    sscanf(szDirectory, "%x", &hDirectory);
    sqlfclosedirc (hDirectory, &rteFileErr);
    if  ( rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
    } /* end if */
  } /* end if */

  /* check result */
  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } else if (nFuncReturn == ERR_RTE_CN00) {
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  return nFuncReturn;
} /* end cn42DirClose */

/*
  -----------------------------------------------------------------------------
  function:     cn42FileSeek
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileSeek
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp_int4            hFile;
  tsp_longint         nPosition;
  _TCHAR              szHandle [PARAMETER_MAXLEN_CN90];
  _TCHAR              szPosition [PARAMETER_MAXLEN_CN90];
  _TCHAR              szMode [PARAMETER_MAXLEN_CN90];
  void              * hDirectory  = NULL;  
  tsp5_rte_fileerror  rteFileErr;
  tsp5_rte_seekkind   rteSeekMode;

  /* read the parameters */
  if (nFuncReturn == OK_CN00) {
    if ( !cn90GetToken(command->args, szHandle,   1) ||
         !cn90GetToken(command->args, szPosition, 2)    ) {
      nFuncReturn = ERR_PARAM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */

  /* read parameters and execute seek*/
  if (nFuncReturn == OK_CN00) {
    hFile     = atol ( szHandle );
    nPosition = atol ( szPosition );
    if (strcmp(szMode, SEEK_MODE_CUR) == 0) {
      rteSeekMode = sp5vf_seek_cur; 
    } else {
      rteSeekMode = sp5vf_seek_begin; 
    } /* end if */

    sqlfseekc(hFile, nPosition, rteSeekMode, &rteFileErr);

    if (rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
    } /* end if */

  } /* end if */

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK (replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;
} /* end cn42FileSeek */

/*
  -----------------------------------------------------------------------------
  function:     cn42FileOpen
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileOpen
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  bool                bBinary = false;
  bool                bAscii  = false;
  bool                bRead   = false;
  bool                bWrite  = false;
  bool                bAppend = false;
  _TCHAR              szName  [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR              szToken [PARAMETER_MAXLEN_CN90] = "";
  long                nIndex;
  long                nSwitch;
  tsp_int4            hFile;

  tsp5_rte_fileerror       rteFileErr;
  tsp5_rte_fileinfo        rteFInfo;
  tsp5_rte_datakind        rteDatakind;
  tsp5_rte_filemode        rteFilemode;
  tsp5_rte_buffering_kind  rteBuffering;

  /* read the parameters */
  nIndex = 1;
  while ( cn90GetToken(command->args, szToken, nIndex) ) {
    nSwitch = 0;
    if (szToken[nSwitch] == OPEN_SWITCH) {
      nSwitch++;
      while (szToken[nSwitch] != CHAR_STRINGTERM_CN90) {
        switch (szToken[nSwitch]) {
          case OPEN_MODE_BIN:
            bBinary = true;
            break;
          case OPEN_MODE_ASCII:
            bAscii = true;
            break;
          case OPEN_MODE_READ:
            bRead = true;
            break;
          case OPEN_MODE_WRITE:
            bWrite = true;
            break;
          case OPEN_MODE_APPEND:
            bAppend = true;
            break;
          default:
            nFuncReturn = ERR_PARAM_CN00;
            break;
        } /* end switch */
        nSwitch++;
      } /* end while */
    } else {
      strcpy(szName, szToken);
    } /* end if */
    nIndex++;
  } /* end while */
  if (strlen(szName) == 0) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  /* execute fileopen */
  if (nFuncReturn == OK_CN00) {
    
    rteDatakind  = (bBinary) ? sp5vf_binary : sp5vf_text;

    if (bAppend) {
      rteFilemode  = sp5vf_append;
    } else if (bRead && bWrite) {
      rteFilemode  = sp5vf_readwrite;
    } else if (bWrite) {
      rteFilemode  = sp5vf_write;
    } else if (bRead) {
      rteFilemode  = sp5vf_read;
    } else {
      rteFilemode  = sp5vf_read;
    } /* end if */

    rteBuffering = sp5bk_unbuffered;

    sqlfinfoc (szName, &rteFInfo, &rteFileErr );
    if (rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
    } else {
      sqlfopenc (szName, rteDatakind, rteFilemode, rteBuffering, &hFile, &rteFileErr );

      if (rteFileErr.sp5fe_result != vf_ok) {
        nFuncReturn = ERR_RTE_CN00;
        cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
      } /* end if */
    } /* end if */

  } /* end if */

  if (nFuncReturn == OK_CN00) {
    /* print OK and handle of file */
    sprintf (replyData, "%s%s%5d%s%lu%s", ANSWER_OK_CN00,      LINE_SEPSTRING_CN00,
                                          hFile,               LINE_SEPSTRING_CN00,
                                          (unsigned long) rteFInfo.sp5fi_size, LINE_SEPSTRING_CN00);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn42FileOpen */

/*
  -----------------------------------------------------------------------------
  function:     cn42FileErase
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileErase
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szName  [PARAMETER_MAXLEN_CN90] = "";
  tsp5_rte_fileerror  rteFileErr;

  // read the parameters
  if ( !cn90GetToken(command->args, szName,   1) ||
       cn90GetToken(command->args, NULL,     2)    ) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  // execute delete
  if (nFuncReturn == OK_CN00) {
    sqlferasec (szName, &rteFileErr);
    if (rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
    } else {
      cn90AnswerOK (replyData, replyLen, NULL);
    } // end if 
  } // end if

  return nFuncReturn;

} // end cn42FileErase

/*
  -----------------------------------------------------------------------------
  function:     cn42FileClose
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileClose
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp_int4            hFile;
  _TCHAR              szHandle [PARAMETER_MAXLEN_CN90];
  void              * hDirectory  = NULL;  
  tsp5_rte_fileerror  rteFileErr;

  // read the parameters
  if ( !cn90GetToken(command->args, szHandle,   1)) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  // execute close
  if (nFuncReturn == OK_CN00) {
    hFile     = atol ( szHandle );
    sqlfclosec ( hFile, sp5vf_close_normal, &rteFileErr );

    if (rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
    } else {
      cn90AnswerOK (replyData, replyLen, NULL);
    } // end if 
  } // end if

  return nFuncReturn;

} // cn42FileClose

/*
  -----------------------------------------------------------------------------
  function:     cn42FileInfo
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileInfo
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szName  [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR            * pCurrPos;
  long                nToken = 1;
 
  // read the parameters
  if ( !cn90GetToken(command->args, szName, nToken)) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } else {
    // skip multi option
    if (strcmp (szName, FILEINFO_MULTI) == 0) {
      nToken++;
    } // end if
    // print OK and FileInfoTitle
    sprintf (replyData, "%s%s%s%s", ANSWER_OK_CN00,      LINE_SEPSTRING_CN00,
                                    FILEINFO_TITLE,      LINE_SEPSTRING_CN00);
    pCurrPos = replyData + strlen(replyData);
  } // end if

  if (nFuncReturn == OK_CN00) {
    while (cn90GetToken(command->args, szName, nToken)) {
      cn42_PrintFileInfo(szName, pCurrPos);
      pCurrPos = pCurrPos + strlen(pCurrPos);
      nToken++;
    } // end while
  } // end if

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;

} // end cn42FileInfo

/*
  -----------------------------------------------------------------------------
  function:     cn42FileRead
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileRead
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn42_FileRead ( command->args, replyData, replyLen, replyLenMax, false);
} // end cn42FileRead

/*
  -----------------------------------------------------------------------------
  function:     cn42FileReadNext
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileReadNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn42_FileRead ( command->args, replyData, replyLen, replyLenMax, true);
} // end cn42FileReadNext

/*
  -----------------------------------------------------------------------------
  function:     cn42FileWrite
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileWrite
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn42_FileWrite ( command->args, replyData, replyLen, replyLenMax, false);
} // end cn42FileWrite

/*
  -----------------------------------------------------------------------------
  function:     cn42FileWriteNext
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileWriteNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  return cn42_FileWrite ( command->args, replyData, replyLen, replyLenMax, true);
} // end cn42FileWriteNext

/*
  -----------------------------------------------------------------------------
  function:     cn42FileGet
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileGet
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_C264c         szType;
  tsp00_C264c         szFileName;
  bool                bBinary;
  tsp_int4            hFile;


  tsp5_rte_fileerror       rteFileErr;
  tsp5_rte_fileinfo        rteFInfo;

  // read the parameters
  if (!cn90GetToken(command->args, szType, 1)        ||
       cn90GetToken(command->args, NULL,   2)        ||
       vcontrol->dbname[0] == CHAR_STRINGTERM_CN90    ) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  // look for file
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn42GetFileName(vcontrol->dbname, szType, szFileName, bBinary);
    if (nFuncReturn != OK_CN00) {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } // end if
  } // end if

  if (nFuncReturn == OK_CN00) {
    // execute fileopen
    sqlfinfoc (szFileName, &rteFInfo, &rteFileErr );
    if (rteFileErr.sp5fe_result != vf_ok) {
      nFuncReturn = ERR_RTE_CN00;
      cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
    } else {
      sqlfopenc (szFileName, ((bBinary) ? sp5vf_binary : sp5vf_text), 
                 sp5vf_read, sp5bk_unbuffered, &hFile, &rteFileErr );

      if (rteFileErr.sp5fe_result != vf_ok) {
        nFuncReturn = ERR_RTE_CN00;
        cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
      } // end if
    } // end if

  } // end if

  // read contents
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn42_FileReadNext( replyData,
                                     replyLen,
                                     replyLenMax,
                                     hFile,
                                     rteFInfo.sp5fi_size,
                                     false,
                                     !bBinary);
  } // end if

  return nFuncReturn;

} // end cn42FileGet

/*
  -----------------------------------------------------------------------------
  function:     cn42FileGetNext
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileGetNext
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_C264c         szType;
  tsp00_C264c         szFileName;
  _TCHAR              szHandle   [PARAMETER_MAXLEN_CN90] = "";
  bool                bBinary;

  // read the parameters
  if (!cn90GetToken(command->args, szType,   1)    ||
      !cn90GetToken(command->args, szHandle, 2)    ||
       cn90GetToken(command->args, NULL,     3)    ||
       vcontrol->dbname[0] == CHAR_STRINGTERM_CN90    ) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  // look for file
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn42GetFileName(vcontrol->dbname, szType, szFileName, bBinary);
  } // end if

  // read contents
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn42_FileReadNext( replyData,
                                     replyLen,
                                     replyLenMax,
                                     atol(szHandle),
                                     0,
                                     false,
                                     !bBinary);
  } // end if

  return nFuncReturn;

} // end cn42FileGetNext

/*
  -----------------------------------------------------------------------------
  function:     cn42FileGetList
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileGetList
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  long                nIndex       = 0;
  tcn42FGetFiles      sFGetFiles[] = FGET_FILES;
  _TCHAR            * pCurrent     = replyData;
  tsp5_rte_fileinfo   rteFileInfo;
  tsp5_rte_fileerror  rteFileErr;
  tsp00_C264c         szFileName;
  tsp00_C264c         szToken;
  tsp00_C256          szRealName;
  tsp00_C256c         szFullReal;
  tsp00_C256          szFileDir;
  tsp00_C256c         szTemp;
  bool                bBinary;
  tsp00_Namec         szType;
  void              * hDirectory  = NULL;  

  // init OK Answer
  _stprintf (pCurrent, "%s%s%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00, (_TCHAR *) FILELIST_TITLE);
  pCurrent = pCurrent + _tcslen(pCurrent);

  // search for Type in file array
  while (sFGetFiles[nIndex].szType != NULL) {
    _tcscpy(szType, sFGetFiles[nIndex].szType);
    if (sFGetFiles[nIndex].bPublic && 
        cn42GetFileName(vcontrol->dbname, szType, szFileName, bBinary, szTemp) == OK_CN00) {
     
      _tcscpy(szFileDir, szTemp);

      if (strstr(szFileName, FGET_WILDCARD) != NULL) {
        sqlfopendirc (&szFileDir, &hDirectory, &szRealName, &rteFileErr);

        while (rteFileErr.sp5fe_result == vf_ok) {

          strcpy(szFullReal, szFileDir);
          strcat(szFullReal, szRealName);
          if (cn42_FileMatch(szFileName, szFullReal, szToken)) {
            sqlfinfoc (szFullReal, &rteFileInfo, &rteFileErr );

            if (rteFileInfo.sp5fi_exists) {
              _stprintf (pCurrent, "%s%s%s%s%s%s%ld%s%.*s%s%.*s%s%s%s%s%s", 
                                   (_TCHAR *) sFGetFiles[nIndex].szType,
                                   (_TCHAR *) FGET_TOKENDEL,
                                   (_TCHAR *) szToken,
                                   VALUE_SEPSTRING_CN00,
                                   (_TCHAR *) ASCII_OR_BINARY(bBinary),
                                   VALUE_SEPSTRING_CN00,
                                   (long) rteFileInfo.sp5fi_size,
                                   VALUE_SEPSTRING_CN00,
                                   sizeof (rteFileInfo.sp5fi_date_modified), (_TCHAR *) rteFileInfo.sp5fi_date_modified,
                                   VALUE_SEPSTRING_CN00,
                                   sizeof (rteFileInfo.sp5fi_time_modified), (_TCHAR *) rteFileInfo.sp5fi_time_modified,
                                   VALUE_SEPSTRING_CN00,
                                   (_TCHAR *) sFGetFiles[nIndex].szComment,
                                   VALUE_SEPSTRING_CN00,
                                   (_TCHAR *) szFullReal  + _tcslen(szFileDir),
                                   LINE_SEPSTRING_CN00);
              pCurrent = pCurrent + _tcslen(pCurrent);
            } // end if
          } // end if

          sqlfreaddirc(hDirectory, &szRealName, &rteFileErr);
        } /* end while */

        sqlfclosedirc (hDirectory, &rteFileErr);

      } else {
        sqlfinfoc (szFileName, &rteFileInfo, &rteFileErr );

        if (rteFileInfo.sp5fi_exists) {
          _stprintf (pCurrent, "%s%s%s%s%ld%s%.*s%s%.*s%s%s%s%s%s", 
                               (_TCHAR *) sFGetFiles[nIndex].szType,
                               VALUE_SEPSTRING_CN00,
                               (_TCHAR *) ASCII_OR_BINARY(bBinary),
                               VALUE_SEPSTRING_CN00,
                               (long) rteFileInfo.sp5fi_size,
                               VALUE_SEPSTRING_CN00,
                               sizeof (rteFileInfo.sp5fi_date_modified), (_TCHAR *) rteFileInfo.sp5fi_date_modified,
                               VALUE_SEPSTRING_CN00,
                               sizeof (rteFileInfo.sp5fi_time_modified), (_TCHAR *) rteFileInfo.sp5fi_time_modified,
                               VALUE_SEPSTRING_CN00,
                               (_TCHAR *) sFGetFiles[nIndex].szComment,
                               VALUE_SEPSTRING_CN00,
                               (_TCHAR *) szFileName + _tcslen(szFileDir),
                               LINE_SEPSTRING_CN00);
          pCurrent = pCurrent + _tcslen(pCurrent);
        } // end if
      } // end if

    } // end if

    nIndex++;
  } // end while

  *replyLen = (int)_tcslen(replyData);

  return OK_CN00;
} // end cn42FileGetList

/*
  -----------------------------------------------------------------------------
  function:     cn42FileBackup
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileBackup
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szType     [PARAMETER_MAXLEN_CN90] = "";

  // read the parameters
  if (!cn90GetToken(command->args, szType,   1)    ||
       cn90GetToken(command->args, NULL,     2)    ||
       vcontrol->dbname[0] == CHAR_STRINGTERM_CN90    ) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } else {
    nFuncReturn = cn42BackupFile(vcontrol->dbname, *(tsp00_Namec *) szType, replyData, *replyLen);
  } // end if

  return nFuncReturn;
} // end cn42FileBackup

/*
  -----------------------------------------------------------------------------
  function:     cn42FileRestore
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42FileRestore
      ( VControlDataT * vcontrol,
        CommandT      * command,
        _TCHAR        * replyData,
        int           * replyLen,
        int             replyLenMax)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szType     [PARAMETER_MAXLEN_CN90] = "";

  // read the parameters
  if (!cn90GetToken(command->args, szType,   1)    ||
       cn90GetToken(command->args, NULL,     2)    ||
       vcontrol->dbname[0] == CHAR_STRINGTERM_CN90    ) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } else {
    nFuncReturn = cn42RestoreFile(vcontrol->dbname, *(tsp00_Namec *) szType, replyData, *replyLen);
  } // end if

  return nFuncReturn;
} // end cn42FileRestore

/*
  -----------------------------------------------------------------------------
  function:     cn42BackupFile
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42BackupFile
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType)
{
  tsp00_Int4 replyLen;

  return cn42BackupFile(szDbName, szType, NULL, replyLen);
} // end cn42BackupFile
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42BackupFile
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        _TCHAR            * replyData,
        tsp00_Int4        & replyLen )
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_C264c         szFileName;
  tsp00_C264c         szBackupName;
  tsp01_RteError      aNewRteError;
  tsp5_rte_fileerror  aRteFileError;
  tsp5_rte_fileinfo   aRteFileInfo;

  // backup one file
  cn42GetFileName(szDbName, szType, szFileName, szBackupName);

  if (strlen(szBackupName) == 0) {

    // can't compute backupname (invalid type or backup impossible)
    nFuncReturn = ERR_PARAM_CN00;
    if (replyData != NULL) {
      cn90AnswerIError(replyData, &replyLen, nFuncReturn);
    } // end if

  } else {

    sqlfinfoc (szFileName, &aRteFileInfo, &aRteFileError);

    if (aRteFileError.sp5fe_result == vf_ok) {

      if (aRteFileInfo.sp5fi_exists) {
        // copy the file
        if (!sqlfilecopyc( szFileName, szBackupName, &aNewRteError )) {
          nFuncReturn = ERR_RTEEXT_CN00;
          if (replyData != NULL) {
            cn90AnswerNewRTEError (replyData, &replyLen, nFuncReturn, &aNewRteError );
          } // end if
        } // end if
      } // end if
        
    } else {

      nFuncReturn = ERR_RTE_CN00;
      if (replyData != NULL) {
        cn90AnswerRTEError (replyData, &replyLen, nFuncReturn, aRteFileError.sp5fe_text, aRteFileError.sp5fe_text.length(), aRteFileError.sp5fe_result);
      } // end if

    } // end if

  } // end if

  return nFuncReturn;

} // end cn42BackupFile

/*
  -----------------------------------------------------------------------------
  function:     cn42RestoreFile
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42RestoreFile
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType)
{
  tsp00_Int4 replyLen;

  return cn42RestoreFile(szDbName, szType, NULL, replyLen);
} // end cn42RestoreFile
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42RestoreFile
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        _TCHAR            * replyData,
        tsp00_Int4        & replyLen )
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_C264c         szFileName;
  tsp00_C264c         szBackupName;
  tsp01_RteError      aNewRteError;
  tsp5_rte_fileerror  aRteFileError;
  tsp5_rte_fileinfo   aRteFileInfo;

  // backup one file
  cn42GetFileName(szDbName, szType, szFileName, szBackupName);

  if (strlen(szBackupName) == 0) {

    // can't compute backupname (invalid type or backup impossible)
    nFuncReturn = ERR_PARAM_CN00;
    if (replyData != NULL) {
      cn90AnswerIError(replyData, &replyLen, nFuncReturn);
    } // end if

  } else {

    sqlfinfoc (szBackupName, &aRteFileInfo, &aRteFileError);

    if (aRteFileError.sp5fe_result == vf_ok) {

      if (aRteFileInfo.sp5fi_exists) {
        // copy the file
        if (!sqlfilecopyc( szBackupName, szFileName, &aNewRteError )) {
          nFuncReturn = ERR_RTEEXT_CN00;
          if (replyData != NULL) {
            cn90AnswerNewRTEError (replyData, &replyLen, nFuncReturn, &aNewRteError );
          } // end if
        } // end if
      } // end if
        
    } else {

      nFuncReturn = ERR_RTE_CN00;
      if (replyData != NULL) {
        cn90AnswerRTEError (replyData, &replyLen, nFuncReturn, aRteFileError.sp5fe_text, aRteFileError.sp5fe_text.length(), aRteFileError.sp5fe_result);
      } // end if

    } // end if

  } // end if

  return nFuncReturn;

} // end cn42RestoreFile

/*
  -----------------------------------------------------------------------------
  function:     cn42GetFileName
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        tsp00_C264c       & szFileName)
{
  tsp00_C264c       szBackupName;
  bool              bBinary;
  tsp00_C256c       szFileDir;
  tsp00_C256c       szBackupDir;
  tsp00_C264c       lokType;

  return cn42GetFileName ( szDbName,
                           lokType.rawAssign(szType),  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        tsp00_C264c       & szFileName,
        bool              & bBinary,
        tsp00_C256c       & szFileDir)
{
  tsp00_C264c       szBackupName;
  tsp00_C256c       szBackupDir;
  tsp00_C264c       lokType;

  return cn42GetFileName ( szDbName,
                           lokType.rawAssign(szType),  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        tsp00_C264c       & szFileName,
        tsp00_C264c       & szBackupName)
{
  bool              bBinary;
  tsp00_C256c       szFileDir;
  tsp00_C256c       szBackupDir;
  tsp00_C264c       lokType;

  return cn42GetFileName ( szDbName,
                           lokType.rawAssign(szType),  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        tsp00_C264c       & szFileName,
        bool              & bBinary)
{
  tsp00_C264c       szBackupName;
  tsp00_C256c       szFileDir;
  tsp00_C256c       szBackupDir;
  tsp00_C264c       lokType;

  return cn42GetFileName ( szDbName,
                           lokType.rawAssign(szType),  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_C264c   szType,  
        tsp00_C264c       & szFileName,
        bool              & bBinary)
{
  tsp00_C264c       szBackupName;
  tsp00_C256c       szFileDir;
  tsp00_C256c       szBackupDir;

  return cn42GetFileName ( szDbName,
                           szType,  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        bool              & bBinary)
{
  tsp00_C264c       szBackupName;
  tsp00_C264c       szFileName;
  tsp00_C256c       szFileDir;
  tsp00_C256c       szBackupDir;
  tsp00_C264c       lokType;

  return cn42GetFileName ( szDbName,
                           lokType.rawAssign(szType),  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        tsp00_C264c       & szFileName,
        tsp00_C264c       & szBackupName,
        bool              & bBinary)
{
  tsp00_C256c       szFileDir;
  tsp00_C256c       szBackupDir;
  tsp00_C264c       lokType;

  return cn42GetFileName ( szDbName,
                           lokType.rawAssign(szType),  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType,  
        tsp00_C264c       & szFileName,
        tsp00_C264c       & szBackupName,
        bool              & bBinary,
        tsp00_C256c       & szFileDir,
        tsp00_C256c       & szBackupDir)
{
  tsp00_C264c lokType;

  return cn42GetFileName ( szDbName,
                           lokType.rawAssign(szType),  
                           szFileName,
                           szBackupName,
                           bBinary,
                           szFileDir,
                           szBackupDir);
} // end cn42GetFileName
/*
  -----------------------------------------------------------------------------
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42GetFileName
      ( tsp00_DbNamec       szDbName,
        const tsp00_C264c   szType,  
        tsp00_C264c       & szFileName,
        tsp00_C264c       & szBackupName,
        bool              & bBinary,
        tsp00_C256c       & szFileDir,
        tsp00_C256c       & szBackupDir)
{
  tcn00_Error         nFuncReturn  = OK_CN00;
  long                nIndex       = 0;
  tcn42FGetFiles      sFGetFiles[] = FGET_FILES;
  _TCHAR            * pExtension   = NULL;
  tsp00_C256          szPath;
  tsp00_C64c          szValue;
  _TCHAR              szTemp  [PARAMETER_MAXLEN_CN90] = "";
  const _TCHAR      * pToken  = 0;
  _TCHAR            * pToken2 = 0;
  int                 nLength = 0;
  bool                bOValid  = true;
  bool                bBValid  = true;
   
  strcpy(szFileName, "");
  strcpy(szFileDir,  "");
  strcpy(szBackupName, "");
  strcpy(szBackupDir,  "");

  pToken = strstr(szType, FGET_TOKENDEL);
  if (pToken == NULL) {
    nLength = (int)strlen(szType);
  } else {
    nLength = (int)(pToken - &szType[0]);
    pToken++;
  } // end if

  // search for Type in file array
  while (sFGetFiles[nIndex].szType != NULL) {

    // match?
    if (strncmp(sFGetFiles[nIndex].szType, szType, nLength) == 0) {
  
      // init out values
      strcpy(szFileName,   "");

      // compute source directory
      switch (sFGetFiles[nIndex].nLocation) {
        case FGET_LOC_DBROOT:
          // dir of dbmserver
          strcpy(szFileName, "");
          break;
        case FGET_LOC_WRK:
          // work dir
          cn90GetWrkPath(szPath);
          strcpy(szFileName, (_TCHAR *) szPath);
          break;
        case FGET_LOC_SAP:
          // sap dir
          strcpy(szFileName, (_TCHAR *) _T("sap"));
          strcat(szFileName, "/");
          break;
        case FGET_LOC_RUNDIR:
          // rundirectory
          if (cn20XParamGetValue(szDbName, XP_ID_RUNDIRECTORY_CN00, szValue) == OK_CN00) {
            strcpy(szFileName, szValue);
            strcat(szFileName, "/");
          } else {
            bOValid = false;
          } // end if
          break;
        case FGET_LOC_CONFIG:
          // config dir
          cn90GetCfgPath(szPath);
          strcpy(szFileName, (_TCHAR *) szPath);
          break;
       default:
          bOValid = false;
          break;
      } // end switch

      strcpy(szFileDir, szFileName);
      cn90ChangePathDelimiter(szFileDir);
      
      // compute source filename  
      if (sFGetFiles[nIndex].szXParam != NULL) {
        // look for xparam
        if (cn20XParamGetValue(szDbName, *(tsp00_XpKeyTypec *) sFGetFiles[nIndex].szXParam, szValue) == OK_CN00) {
          strcat(szFileName, szValue);
        } else {
          bOValid = false;
        } // end if
      } else if (sFGetFiles[nIndex].szName == FGET_NAME_DB) {
        // use dbname
        strcat(szFileName, szDbName);
      } else if (sFGetFiles[nIndex].szName != NULL) {
        // use szName
        strcat(szFileName, sFGetFiles[nIndex].szName);
      } // end if

      // compute source extension
      if (sFGetFiles[nIndex].szExt != NULL) {
        // use szExt-Field
        strcat(szFileName, sFGetFiles[nIndex].szExt);
      } // end if

      pToken2 = strstr(szFileName, FGET_WILDCARD);
      if (pToken != NULL && pToken2 != NULL) {
        *pToken2 = 0;
        pToken2++;
        strcpy(szTemp, szFileName);
        strcat(szTemp, pToken);
        strcat(szTemp, pToken2);
        strcpy(szFileName, szTemp);
      } // end if

      cn90ChangePathDelimiter(szFileName);

      // backupname required and possible
      if (sFGetFiles[nIndex].szBaName != NULL) {

        // init out values
        strcpy(szBackupName, "");

        // compute backup directory
        switch (sFGetFiles[nIndex].nBaLocation) {
          case FGET_LOC_DBROOT:
            // dir of dbmserver
            strcpy(szBackupName, "");
            break;
          case FGET_LOC_WRK:
            // work dir
            cn90GetWrkPath(szPath);
            strcpy(szBackupName, (_TCHAR *) szPath);
            break;
          case FGET_LOC_RUNDIR:
            // rundirectory
            if (cn20XParamGetValue(szDbName, XP_ID_RUNDIRECTORY_CN00, szValue) == OK_CN00) {
              strcpy(szBackupName, szValue);
              strcat(szBackupName, "/");
            } else {
              bBValid = false;
            } // end if
            break;
          case FGET_LOC_CONFIG:
            // config dir
            cn90GetCfgPath(szPath);
            strcpy(szBackupName, (_TCHAR *) szPath);
            break;
         default:
            bBValid = false;
            break;
        } // end switch

        strcpy(szBackupDir, szBackupName);
        cn90ChangePathDelimiter(szBackupDir);

        // compute backup filename  
        if (sFGetFiles[nIndex].szBaName == FGET_NAME_DB) {
          // use dbname
          strcat(szBackupName, szDbName);
        } else if (sFGetFiles[nIndex].szBaName != NULL) {
          // use szName
          strcat(szBackupName, sFGetFiles[nIndex].szBaName);
        } // end if

        // compute backup extension
        if (sFGetFiles[nIndex].szBaExt == FGET_ORG_EXT) {
          // use extension form original file
          pExtension = strrchr(szFileName, '.');
          if (pExtension != NULL) {
            strcat(szBackupName, pExtension);
          } // end if
        } else if (sFGetFiles[nIndex].szBaExt != NULL) {
          // use szExt
          strcat(szBackupName, sFGetFiles[nIndex].szBaExt);
        } // end if

        pToken2 = strstr(szBackupName, FGET_WILDCARD);
        if (pToken != NULL && pToken2 != NULL) {
          *pToken2 = 0;
          pToken2++;
          strcpy(szTemp, szFileName);
          strcat(szTemp, pToken);
          strcat(szTemp, pToken2);
          strcpy(szBackupName, szTemp);
        } // end if

        cn90ChangePathDelimiter(szBackupName);

      } // end if
 
      if (!bOValid) {
        strcpy(szFileName, "");
        strcpy(szFileDir,  "");
      } // end if
      if (!bBValid) {
        strcpy(szBackupName, "");
        strcpy(szBackupDir,  "");
      } // end if

      // binary flag
      bBinary = sFGetFiles[nIndex].bBinary;

      break;
    } // end if

    nIndex++;
  } // end while

  if (sFGetFiles[nIndex].szType == NULL) {
    nFuncReturn = ERR_PARAM_CN00;
  } // end if

  return nFuncReturn;
} // end cn42GetFileName

/*
  -----------------------------------------------------------------------------
  function:     cn42RemoveFiles
  -----------------------------------------------------------------------------
 */
void cn42RemoveFiles
      ( tsp00_DbNamec       szDbName)
{
  long                nIndex       = 0;
  tcn42FGetFiles      sFGetFiles[] = FGET_FILES;
  tsp00_Namec         szType;

  while (sFGetFiles[nIndex].szType != NULL) {
    if (sFGetFiles[nIndex].bRemove) {
      _tcscpy(szType, sFGetFiles[nIndex].szType);
      cn42_RemoveFile(szDbName, szType);
    } // end if
    nIndex++;
  } // end while

} // end cn42RemoveFiles

/*
  -----------------------------------------------------------------------------
  public function:     cn42OpenFile
  -----------------------------------------------------------------------------
 */
tcn00_Error cn42OpenFile
      ( tsp00_DbNamec        szDbName,
        const tsp00_Namec  & szType, 
        tsp5_rte_filemode    fileMode,
        tsp00_Int4         & hFile,
        tsp5_rte_fileerror & errFile)
{
  tcn00_Error         nReturn = OK_CN00;
  tsp00_C264c         szConfigFile;
  tsp00_C264c         szBackupFile;
  bool                bBinary;

  // prepare file name
  nReturn = cn42GetFileName(szDbName, szType, szConfigFile, szBackupFile, bBinary);

  if (nReturn == OK_CN00) {
    // open file for write
    sqlfopenc(szConfigFile, bBinary ? SP5VF_BINARY : SP5VF_TEXT, fileMode, SP5BK_BUFFERED, &hFile, &errFile );
    if ((errFile.sp5fe_result != vf_ok) && (strlen(szBackupFile) > 0)) {
      cn42RestoreFile(szDbName, FGET_DBMCFG_CN42);
      sqlfopenc(szConfigFile, bBinary ? SP5VF_BINARY : SP5VF_TEXT, fileMode, SP5BK_BUFFERED, &hFile, &errFile );
      if (errFile.sp5fe_result != vf_ok) {
        sqlfopenc(szBackupFile, bBinary ? SP5VF_BINARY : SP5VF_TEXT, fileMode, SP5BK_BUFFERED, &hFile, &errFile ); 
      }  // end if
    }  // end if
    if (errFile.sp5fe_result != vf_ok) {
      nReturn = ERR_FILE_CN00;
    }  // end if
  }  // end if

  return nReturn;
} /* end cn42OpenFile */

/*
  -----------------------------------------------------------------------------
  implementation private functions
  -----------------------------------------------------------------------------
 */

/*
  -----------------------------------------------------------------------------
  function:     cn42_RemoveFile
  -----------------------------------------------------------------------------
 */
static void cn42_RemoveFile
      ( tsp00_DbNamec       szDbName,
        const tsp00_Namec   szType) 
{
  tsp00_C264c         szFileName;
  tsp00_C264c         szBackupName;
  tsp5_rte_fileerror  rteFileErr;

  if (cn42GetFileName(szDbName, szType, szFileName, szBackupName) == OK_CN00) {
    if (_tcslen(szFileName) > 0) {
      sqlferasec(szFileName, &rteFileErr);
    } // end if
    if (_tcslen(szBackupName) > 0) {
      sqlferasec(szBackupName, &rteFileErr);
    } // end if
  } // end if

} // end cn42_RemoveFile

/*
  -----------------------------------------------------------------------------
  function:     cn42_PrintFileInfo
  -----------------------------------------------------------------------------
 */
static void cn42_PrintFileInfo
      ( _TCHAR       *szName,
        _TCHAR       *szBuffer)
{
    tsp5_rte_fileinfo     rteFileInfo;
    tsp5_rte_fileerror    rteFileErr;
    static const _TCHAR   * mediastr [] = FILEINFO_MEDIATYPE;

    sqlfinfoc (szName, &rteFileInfo, &rteFileErr );

    if (rteFileInfo.sp5fi_exists) {
      sprintf (szBuffer,
               "\"%s\",true,%s,%s,%s,%ld,\"%.*s\",\"%.*s\",\"%s\"\n",
               szName,
               BOOLSTR_CN00 (rteFileInfo.sp5fi_readable), 
               BOOLSTR_CN00 (rteFileInfo.sp5fi_writeable),
               BOOLSTR_CN00 (rteFileInfo.sp5fi_is_link), 
               rteFileInfo.sp5fi_size,
               sizeof (rteFileInfo.sp5fi_date_modified), (_TCHAR *) rteFileInfo.sp5fi_date_modified,
               sizeof (rteFileInfo.sp5fi_time_modified), (_TCHAR *) rteFileInfo.sp5fi_time_modified,
               mediastr [rteFileInfo.sp5fi_media_kind]);
    } else {
      sprintf  (szBuffer, "\"%s\",false,,,,,,,\n", szName);
    } // end if

} // end cn42_PrintFileInfo

/*
  -----------------------------------------------------------------------------
  function:     cn42_FileRead
  -----------------------------------------------------------------------------
 */
static tcn00_Error cn42_FileRead
      ( const _TCHAR  * szArguments,
        _TCHAR        * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const bool      bNext)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szName     [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR              szToken    [PARAMETER_MAXLEN_CN90] = "";
  bool                bNoClose   = false;
  long                nIndex;
  long                nSwitch;
  tsp_int4            hFile;

  tsp5_rte_fileerror       rteFileErr;
  tsp5_rte_fileinfo        rteFInfo;
  tsp5_rte_datakind        rteDatakind   = sp5vf_text;
  tsp5_rte_filemode        rteFilemode   = sp5vf_read;
  tsp5_rte_buffering_kind  rteBuffering  = sp5bk_unbuffered;

  /* read the parameters */
  nIndex = 1;
  while ( cn90GetToken(szArguments, szToken, nIndex) ) {
    nSwitch = 0;
    if (szToken[nSwitch] == OPEN_SWITCH) {
      nSwitch++;
      while (szToken[nSwitch] != CHAR_STRINGTERM_CN90) {
        switch (szToken[nSwitch]) {
          case OPEN_MODE_BIN:
            rteDatakind  = sp5vf_binary;
            break;
          case OPEN_MODE_ASCII:
            rteDatakind  = sp5vf_text;
            break;
          default:
            nFuncReturn = ERR_PARAM_CN00;
            break;
        } /* end switch */
        nSwitch++;
      } /* end while */
    } else {
      if (strlen(szName) == 0) {
        strcpy(szName, szToken);
      } else {
        if (strcmp(szToken, FILE_OPT_NOCLOSE) == 0) {
          bNoClose = true;
        } else {
          nFuncReturn = ERR_PARAM_CN00;
        } // end if
      } /* end if */
    } /* end if */
    nIndex++;
  } /* end while */

  if (strlen(szName) == 0 || nFuncReturn == ERR_PARAM_CN00) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  if (nFuncReturn == OK_CN00) {

    if (bNext) {
      // read handle
      hFile = atol(szName);
    } else {

      // execute fileopen
      sqlfinfoc (szName, &rteFInfo, &rteFileErr );
      if (rteFileErr.sp5fe_result != vf_ok) {
        nFuncReturn = ERR_RTE_CN00;
        cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
      } else {
        sqlfopenc (szName, rteDatakind, rteFilemode, rteBuffering, &hFile, &rteFileErr );

        if (rteFileErr.sp5fe_result != vf_ok) {
          nFuncReturn = ERR_RTE_CN00;
          cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
        } // end if
      } // end if

    } // end if

    // read contents
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn42_FileReadNext( replyData,
                                       replyLen,
                                       replyLenMax,
                                       hFile,
                                       rteFInfo.sp5fi_size,
                                       bNoClose,
                                       (rteDatakind == sp5vf_text));
    } /* end if */
  } /* end if */

  return nFuncReturn;

} // end cn42_FileRead

/*
  -----------------------------------------------------------------------------
  function:     cn42_FileReadNext
  -----------------------------------------------------------------------------
 */
static tcn00_Error cn42_FileReadNext
      ( _TCHAR           * replyData,
        int              * replyLen,
        const int          replyLenMax,
        const tsp_int4     hFile,
        const tsp_longint  nSize,
        const bool         bNoClose,
        const bool         bAscii)
{
  tcn00_Error            nFuncReturn = OK_CN00;
  _TCHAR               * pCurrPos;
  _TCHAR               * pFlagPos;
  _TCHAR               * pLenPos;
  tsp_longint            nLength = 0;
  tsp_longint            nOutLen = 0;
  tsp5_rte_fileerror     rteFileErr;
  _TCHAR                 szTmp[PARAMETER_MAXLEN_CN90];

  /* print OK and handle of file */
  sprintf (replyData, "%s%s%5d%s", ANSWER_OK_CN00,      LINE_SEPSTRING_CN00,
                                   hFile,               LINE_SEPSTRING_CN00);

  pFlagPos = replyData + strlen(replyData);
  sprintf ( pFlagPos , "%s%s" , CONT_FLAG_CONTINUE, LINE_SEPSTRING_CN00 );

  pLenPos = replyData + strlen(replyData);
  sprintf ( pLenPos, "%20ld%20ld%s" , (long) 0, (long) 0, LINE_SEPSTRING_CN00 );

  pCurrPos  = replyData + strlen(replyData);
  *replyLen = (int)strlen(replyData);

  // for test only
  // replyLenMax = 150;

  // fill buffer
  do {
    sqlfreadc(hFile, pCurrPos, replyLenMax - *replyLen - ((bAscii) ? 2 : 0), &nOutLen, &rteFileErr);
    *replyLen = *replyLen + (int)nOutLen;
    nLength   = nLength   + nOutLen;

    if  ( rteFileErr.sp5fe_warning == sp5vfw_no_warning && 
          rteFileErr.sp5fe_result  == vf_ok             &&
          bAscii) {
      *replyLen = *replyLen + 2;
      nLength   = nLength   + 2;
      replyData[*replyLen-2] = '\r';
      replyData[*replyLen-1] = '\n';
    } // end if

    pCurrPos = replyData + *replyLen;

  } while ( *replyLen + ((bAscii) ? 2 : 0) <  replyLenMax       && 
            rteFileErr.sp5fe_warning       == sp5vfw_no_warning && 
            rteFileErr.sp5fe_result        == vf_ok                );

  // end handling
  replyData[*replyLen] = '\0';
  sprintf ( szTmp, "%20ld%20ld", (long) nSize, (long) nLength);
  strncpy(pLenPos, szTmp, strlen(szTmp));

  if ( rteFileErr.sp5fe_result == vf_eof ) {
    sprintf ( szTmp, "%s", CONT_FLAG_END );
    strncpy(pFlagPos, szTmp, strlen(szTmp));

    if (!bNoClose) {
      sqlfclosec ( hFile, sp5vf_close_normal, &rteFileErr );
    } // end if
  } // end if

  if (rteFileErr.sp5fe_result == vf_notok) {
    nFuncReturn = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
    sqlfclosec ( hFile, sp5vf_close_normal, &rteFileErr );
  } // end if 

  return nFuncReturn;

} // end cn42_FileReadNext

/*
  -----------------------------------------------------------------------------
  function:     cn42_FileWrite
  -----------------------------------------------------------------------------
 */
static tcn00_Error cn42_FileWrite
      ( const _TCHAR  * szArguments,
        _TCHAR        * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const bool      bNext)
{
  tcn00_Error         nFuncReturn = OK_CN00;
  _TCHAR              szName     [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR              szToken    [PARAMETER_MAXLEN_CN90] = "";
  _TCHAR              szLocalArgs [PARAMETER_MAXLEN_CN90]  = "";
  bool                bNoClose   = false;
  long                nIndex;
  long                nSwitch;
  tsp_int4            hFile;
  long                nData   = 0;
  const _TCHAR      * pData   = NULL;      

  tsp5_rte_fileerror       rteFileErr;
  tsp5_rte_datakind        rteDatakind   = sp5vf_text;
  tsp5_rte_filemode        rteFilemode   = sp5vf_write;
  tsp5_rte_buffering_kind  rteBuffering  = sp5bk_buffered;

  if (strstr(szArguments, LINE_SEPSTRING_CN00) != NULL) {
    memset(szLocalArgs, CHAR_STRINGTERM_CN90, sizeof(szLocalArgs));
    strncpy(szLocalArgs, szArguments, strstr(szArguments, LINE_SEPSTRING_CN00) - &szArguments[0]);
  } else {
    strcpy(szLocalArgs, szArguments);
  } // end if

  /* read the parameters */
  nIndex = 1;
  while ( cn90GetToken(szLocalArgs, szToken, nIndex) ) {
    nSwitch = 0;
    if (szToken[nSwitch] == OPEN_SWITCH) {
      nSwitch++;
      while (szToken[nSwitch] != CHAR_STRINGTERM_CN90) {
        switch (szToken[nSwitch]) {
          case OPEN_MODE_BIN:
            rteDatakind  = sp5vf_binary;
            break;
          case OPEN_MODE_ASCII:
            rteDatakind  = sp5vf_text;
            break;
          default:
            nFuncReturn = ERR_PARAM_CN00;
            break;
        } /* end switch */
        nSwitch++;
      } /* end while */
    } else {
      if (strlen(szName) == 0) {
        strcpy(szName, szToken);
      } else {
        if (strcmp(szToken, FILE_OPT_NOCLOSE) == 0) {
          bNoClose = true;
        } else {
          nFuncReturn = ERR_PARAM_CN00;
        } // end if
      } /* end if */
    } /* end if */
    nIndex++;
  } /* end while */

  // look for data size and data
  pData = strstr(szArguments, LINE_SEPSTRING_CN00) + strlen(LINE_SEPSTRING_CN00);
  nData = (pData == NULL) ? 0 : atol(pData);
  pData = strstr(pData, LINE_SEPSTRING_CN00) + strlen(LINE_SEPSTRING_CN00);
 
  if (strlen(szName) == 0 || nFuncReturn == ERR_PARAM_CN00 || pData == NULL || nData == 0) {
    nFuncReturn = ERR_PARAM_CN00;
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } /* end if */

  if (nFuncReturn == OK_CN00) {

    if (bNext) {
      // read handle
      hFile = atol(szName);
    } else {

      sqlfopenc (szName, rteDatakind, rteFilemode, rteBuffering, &hFile, &rteFileErr );

      if (rteFileErr.sp5fe_result != vf_ok) {
        nFuncReturn = ERR_RTE_CN00;
        cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
      } // end if

    } // end if

    // read contents
    if (nFuncReturn == OK_CN00) {
      nFuncReturn = cn42_FileWriteNext( replyData,
                                        replyLen,
                                        replyLenMax,
                                        hFile,
                                        (rteDatakind == sp5vf_text),
                                        nData,
                                        pData);
    } /* end if */
  } /* end if */

  return nFuncReturn;

} // end cn42_FileWrite

/*
  -----------------------------------------------------------------------------
  function:     cn42_FileWriteNext
  -----------------------------------------------------------------------------
 */
static tcn00_Error cn42_FileWriteNext
      ( _TCHAR        * replyData,
        int           * replyLen,
        const int       replyLenMax,
        const tsp_int4  hFile,
        const bool      bAscii,
        const long      nData,
        const _TCHAR  * pData )
{
  tcn00_Error  nFuncReturn = OK_CN00;
  _TCHAR     * pNewline    = NULL;
  const _TCHAR * pCurrent    = pData;
  long         nOutput     = 0;
  long         nWrite      = 0;

  tsp5_rte_fileerror       rteFileErr;

  do {

    if (bAscii) {

      // skip newline
      pNewline = (_TCHAR *) memchr (pCurrent, '\n', nData - nOutput);
      if (pNewline == NULL) {
        nWrite = nData - nOutput;
      } else if (pNewline == pCurrent) {
        nWrite = 0;
        nOutput++;
      } else {
        nWrite = pNewline - pCurrent - (pNewline[-1] == '\r') ? 1 : 0;
        nOutput++;
      } // end if

      sqlfwritec (hFile, pCurrent, nWrite, &rteFileErr );
      nOutput = nOutput + nWrite;
      pCurrent = (pNewline != NULL) ? pNewline + 1 : pCurrent + nWrite;

    } else {
      nWrite = nData;
      sqlfwritec (hFile, pCurrent, nWrite, &rteFileErr );
      nOutput = nOutput + nWrite;
    } // end if

  } while ((nOutput < nData) && (rteFileErr.sp5fe_result == vf_ok));

  if (rteFileErr.sp5fe_result != vf_ok) {
    nFuncReturn = ERR_RTE_CN00;
    cn90AnswerRTEError (replyData, replyLen, nFuncReturn, rteFileErr.sp5fe_text, rteFileErr.sp5fe_text.length(), rteFileErr.sp5fe_result );
  } else {
    /* print OK and handle of file */
    sprintf (replyData, "%s%s%5d%s", ANSWER_OK_CN00,      LINE_SEPSTRING_CN00,
                                     hFile,               LINE_SEPSTRING_CN00);
    *replyLen = (int)strlen(replyData);
  } // end if

  return nFuncReturn;

} // end cn42_FileWriteNext

/*
  -----------------------------------------------------------------------------
  function:     cn42_FileMatch
  -----------------------------------------------------------------------------
 */
static bool cn42_FileMatch
      ( tsp00_C264c        szFileName, 
        tsp00_C256c        szRealName, 
        tsp00_C264c      & szToken)
{

  _TCHAR * pPostfix   = strstr(szFileName, FGET_WILDCARD);
  _TCHAR * pPrefix    = szFileName;
  _TCHAR * pReal      = szRealName;
  bool     bReturn    = false;

  if (pPostfix != NULL) {
    pPostfix++;

    if (_tcsncmp(pPrefix,  pReal,  pPostfix - pPrefix -1 )                                 == 0  &&
        _tcsncmp(pPostfix, pReal + _tcslen(pReal) - _tcslen(pPostfix) , _tcslen(pPostfix)) == 0     ) {
      bReturn = true;
      memset(szToken, CHAR_STRINGTERM_CN90, sizeof(tsp00_C264c     ));
      _tcsncpy(szToken, pReal + (pPostfix - pPrefix - 1), _tcslen(pReal) - _tcslen(pPrefix) + 1);
    } // end if
  } // end if

  return bReturn;
} // end cn42_FileMatch

/* ------------------------------------------------------------------
 * IMPLEMENTATION PRIVATE FUNCTION cn42_ReadDir
 * ------------------------------------------------------------------
 */
static tcn00_Error cn42_ReadDir
      ( void               * hDirectory, 
        _TCHAR             * szFirst, 
        _TCHAR             * replyData, 
        int                  replyLenMax,
        tsp5_rte_fileerror * pRteFileErr)
{
  tcn00_Error nFuncReturn = OK_CN00;
  _TCHAR  *   pCurrent    = replyData;
  _TCHAR  *   pContinue   = NULL;
  tsp_c256    szFile;

  /* print OK and handle of dir */
  sprintf (pCurrent, "%s%s%p%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00,
                                 hDirectory,     LINE_SEPSTRING_CN00);

  /* print CONTINUE */
  pContinue = replyData + strlen(replyData);
  sprintf (pContinue, "%s%s", KEY_DIRCONTINUE_CN00, LINE_SEPSTRING_CN00);
  
  /* print first file if exist */
  if (szFirst != NULL) {
    pCurrent = replyData + strlen(replyData);
    sprintf(pCurrent, "%s%s", szFirst, LINE_SEPSTRING_CN00);
  } /* end if */

  /* print files in loop */
  while ((pRteFileErr->sp5fe_result == vf_ok) && (replyLenMax > (int)(strlen(replyData) + sizeof(szFile) + 2))) {

    sqlfreaddirc(hDirectory,  &szFile, pRteFileErr);

    if (pRteFileErr->sp5fe_result == vf_ok) {
      pCurrent = replyData + strlen(replyData);
      sprintf(pCurrent, "%s%s", (_TCHAR *) szFile, LINE_SEPSTRING_CN00);
    } /* end if */

  } /* end while */

  /* if EOF replace continue with end */
  if (pRteFileErr->sp5fe_result != vf_ok) {
    strncpy(pContinue, KEY_DIREND_CN00, strlen(KEY_DIREND_CN00));
  } /* end if */

  return nFuncReturn;

} /* end cn42_ReadDir */
