/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
/* @lastChanged: "1998-06-24  10:45"
 *
 * @filename:    vcn80.cpp
 * @purpose:     "utility and sql access"
 * @release:     7.1.0.0
 * @see:         "-.-"
 *
 * @copyright:   "Copyright by SAP AG, 1998"
 *
 * ==================================================================
 *
 * responsible:   Martin Reinecke (D024853 - Martin.Reinecke@SAP-AG.de)
 *                Bernd Vorsprach (D025588 - Bernd.Vorsprach@SAP-AG.de)
 *
 * created:       1998-03-19 by Martin Reinecke
 *
 * purpose:       utility connect/execute/release
 *
 * ==================================================================
 */

/* ------------------------------------------------------------------
 * includes
 * ------------------------------------------------------------------
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>

#if  defined(CN71) || defined(CN72)

  #include "vsp001.h"
  #include "vsp009c.h"
  #include "heo02.h"
  #include "hsp02.h"

  #include "hcn20.h"
  #include "hcn31.h"
  #include "hcn51.h"
  #include "hcn80.h"
  #include "hcn90.h"
#endif

/* ------------------------------------------------------------------
 * SPECIFICATION PRIVATE TYPES/MACROS
 * ------------------------------------------------------------------
 */
#define SESSION_SQLUSER       1
#define SESSION_UTIL          2
#define SESSION_SERVICE       3

#define SQL_SELECT            "SELECT"
#define SQL_CHAR_FIELDSEP     ';'
#define SQL_CHAR_RECSEP       '\n'
#define SQL_CHAR_STRINGDEL    '\''

#define PAGE_MODE_NAME     "NAME"
#define PAGE_PUT_STMT      "RESTORE"
#define PAGE_GET_STMT      "TO 'X'"

#if defined (CN72)

/* PRIVATE FUNCTION cn80_DiagnoseStatement
 * ------------------------------------------------------------------
 * purpose: analyses
 *
 */
static tcn00_Error  cn80_DiagnoseStatement (
      VControlDataT    * vcontrol,
      const char       * szArguments,
      const char       * szGetOrPut,
            char       * szStatement,
            long       * nPageSize)
{
  tcn00_Error	        nFuncReturn                          = OK_CN00;
  bool                bDevRequest                          = false;
  char                szLocalArgs [PARAMETER_MAXLEN_CN90]  = "";
  char                szMode      [PARAMETER_MAXLEN_CN90]  = "";
  char                szDevNum    [PARAMETER_MAXLEN_CN90]  = "";
  char                szPageNum   [PARAMETER_MAXLEN_CN90]  = "";
  char                szPageSize  [PARAMETER_MAXLEN_CN90]  = "";
  char                szDevSpace  [XP_VALUE_LENGTH_MXCN00] = "";
  char                szParameter [XP_VALUE_LENGTH_MXCN00] = "";
  long                nKey                                 = 0;   
  long                nParam                               = 0;   
  long                nLength                              = 0;   
  unsigned long       nArgs                                = 0;   
  tcn00_DevspaceKeywords aKeyTable[] = XP_DEVSPACE_KEYWORDS_CN00;
  tcn00_DevspaceParams   aParamTable[] = XP_DEVSPACE_PARAMS_CN00;
  
  if (strstr(szArguments, LINE_SEPSTRING_CN00) != NULL) {
    memset(szLocalArgs, CHAR_STRINGTERM_CN90, sizeof(szLocalArgs));
    strncpy(szLocalArgs, szArguments, strstr(szArguments, LINE_SEPSTRING_CN00) - &szArguments[0]);
  } else {
    strcpy(szLocalArgs, szArguments);
  } // end if

  /* analyze parameters */
  nArgs = nArgs + (cn90GetToken(szLocalArgs, szMode,     1) ? 1 : 0);
  nArgs = nArgs + (cn90GetToken(szLocalArgs, szDevNum,   2) ? 1 : 0);
  nArgs = nArgs + (cn90GetToken(szLocalArgs, szPageNum,  3) ? 1 : 0);
  nArgs = nArgs + (cn90GetToken(szLocalArgs, szPageSize, 4) ? 1 : 0);
  nArgs = nArgs + (cn90GetToken(szLocalArgs, NULL,       5) ? 1 : 0);

  if (strcmp(szGetOrPut, PAGE_GET_STMT) == 0 && (nArgs == 1 || nArgs == 3)) {

    if (nArgs == 1) {
      strcpy(szPageNum,  szMode);
    } else {
      bDevRequest = true;
    } // end if
      
  } else if (strcmp(szGetOrPut, PAGE_PUT_STMT) == 0 && (nArgs == 1 || nArgs == 4)) {

    if (nArgs == 1) {
      strcpy(szPageSize, szMode);
    } else {
      bDevRequest = true;
    } // end if

    // compute pagesize
    if (nPageSize != NULL) {
      *nPageSize = atol(szPageSize);
    } // end if

  } else {
    nFuncReturn = ERR_PARAM_CN00;
  } // end if

  if (nFuncReturn == OK_CN00 && bDevRequest) {

    /* search in keywordtable */
    for (nKey = 0; (aKeyTable[nKey].nDevSpace != XPDevSpaceUnknown_ecn00) &&
                   (strcmp(aKeyTable[nKey].szText, szMode) != 0  ); nKey++);

    /* search in paramtable */
    for (nParam = 0; aKeyTable[nKey].nDevSpace != aParamTable[nParam].nDevSpace; nParam++);

    if (aKeyTable[nKey].nDevSpace == XPDevSpaceUnknown_ecn00) {
      if (strcmp(szMode, PAGE_MODE_NAME) == 0) {
        strcpy(szDevSpace, szDevNum);
      } else {
        nFuncReturn = ERR_PARAM_CN00;
      } /* end if */
    } else {
      sprintf(szParameter, aParamTable[nParam].szName, atoi(szDevNum));
      nFuncReturn = cn20XParamGetValue(vcontrol, szParameter, szDevSpace, NULL);
      if (strlen(szDevSpace) == 0) {
        nFuncReturn = ERR_XPNOTFOUND_CN00;
      } /* end if */
    } /* end if */

  } // end if

  if (nFuncReturn == OK_CN00) {
    /* built statement */
    if (bDevRequest) {
     sprintf(szStatement, "DIAGNOSE %s DEVSPACE '%s' FROM %s TO %s",
                           szGetOrPut,
                           szDevSpace,
                           szPageNum,
                           szPageNum);
    } else {
      sprintf(szStatement, "DIAGNOSE %s DATA %s", szGetOrPut, szPageNum);
    } /* end if */
  } /* end if */

  return nFuncReturn;

} /* end cn80_DiagnoseStatement */

#endif

/* PRIVATE FUNCTION cn80_SetSession
 * ------------------------------------------------------------------
 * purpose: set the session to global struct
 *
 */
static void cn80_SetSession (
      VControlDataT    * vcontrol,
      tin01_sql_session * pSession,
      int                nMode )
{
  switch (nMode) {
    case SESSION_SQLUSER:
      vcontrol->pSQLUserSession = (void *) pSession;
      break;
    case SESSION_UTIL:
      vcontrol->pUtilSession = (void *) pSession;
      break;
    case SESSION_SERVICE:
      vcontrol->pServiceSession = (void *) pSession;
      break;
  } /* end switch */

} /* end cn80_SetSession */

/* PRIVATE FUNCTION cn80_Session
 * ------------------------------------------------------------------
 * purpose: get the session from global struct and check
 *
 */
static tcn00_Error cn80_Session (
      VControlDataT    *  vcontrol,
      char             *  replyData,
      int              *  replyLen,
      tin01_sql_session ** pSession,
      int                 nMode )
{
  tcn00_Error nFuncReturn    = OK_CN00;
  tcn00_Error nPossibleError = OK_CN00;

  /**/

  switch (nMode) {
    case SESSION_SQLUSER:
      *pSession =  (tin01_sql_session *) vcontrol->pSQLUserSession;
      nPossibleError = ERR_NOSQLUSERSESSION_CN00;
      break;
    case SESSION_UTIL:
      *pSession =  (tin01_sql_session *) vcontrol->pUtilSession;
      nPossibleError = ERR_NOUTILSESSION_CN00;
      break;
    case SESSION_SERVICE:
      *pSession =  (tin01_sql_session *) vcontrol->pServiceSession;
      nPossibleError = ERR_NOSERVICESESSION_CN00;
      break;
  } /* end switch */

  /* is there an active  session */
  if (nFuncReturn == OK_CN00) {
    if ( (*pSession == NULL)          ||
         (!(*pSession)->is_connected)    ) {
      nFuncReturn = nPossibleError;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } /* end if */
  } /* end if */
 
  /**/

  return nFuncReturn;
} /* end cn80_Session */

/* PRIVATE FUNCTION cn80_FreeSession
 * ------------------------------------------------------------------
 * purpose: frees the session from the VControlDataT structure
 *
 */
static void cn80_FreeSession(
      VControlDataT     * vcontrol,
      int                 nMode)
{
  tin01_sql_session  * pSession;

  /**/

  if (vcontrol != NULL) {
    switch (nMode) {
      case SESSION_SQLUSER:
        pSession =  (tin01_sql_session *) vcontrol->pSQLUserSession;
        vcontrol->pSQLUserSession = NULL;
        break;
      case SESSION_UTIL:
        cn31FreeBackupSession(vcontrol);
        pSession =  (tin01_sql_session *) vcontrol->pUtilSession;
        vcontrol->pUtilSession = NULL;
        break;
      case SESSION_SERVICE:
        pSession =  (tin01_sql_session *) vcontrol->pServiceSession;
        vcontrol->pServiceSession = NULL;
        break;
    } /* end switch */
  } /* end if */

  if (pSession != NULL) {
    if (pSession->is_connected ) {
      i29release(pSession, true);
    } /* end if */
    cn00MyFree(pSession);
  } /* end if */

} /* end cn80_FreeSession */

/* PRIVATE FUNCTION cn80_Connect
 * ------------------------------------------------------------------
 * purpose: open a session
 *
 */
static tcn00_Error cn80_Connect (
      char              * replyData,
      int               * replyLen,
      tin01_sql_session * pSession,
      char              * szDatabase,
      char              * szUsername,
      char              * szPassword,
      int                 nMode)
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  tsp4_xuser_record * pXUser      = NULL;
  int                 rc;
  tsp_name            pwd;
  int                 i;
  tsp_errtextc        errtext;
  tsp_knl_identifierc errname;
  tsp_int2            errcode;
  tsp_int2            errpos;

  /**/


  /* create session */
  if (nFuncReturn == OK_CN00) {
    i29initsession(pSession, NULL);

    pXUser = &pSession->xuser;

    /* database */
    i29pascalstring(pXUser->xu_serverdb, sizeof (pXUser->xu_serverdb), szDatabase, -1);

    /* username */
    if (szUsername != NULL) {
      i29pascalstring (pXUser->xu_user, sizeof (pXUser->xu_user), szUsername, -1);
      for (i = 0; i < sizeof (pXUser->xu_user); ++i)
      {
        pXUser->xu_user [i] = (char) toupper (pXUser->xu_user [i]);
      }
    } /* end if */

    /* password */
    if (szPassword != NULL) {
      i29pascalstring (pwd, sizeof (pwd), szPassword, -1);
      for (i = 0; i < sizeof (pwd); ++i)
      {
        pwd [i] = (char) toupper (pwd [i]);
      }
    s02applencrypt (pwd, pXUser->xu_password);
    } /* end if */

    /* utility flag */
    if (nMode == SESSION_UTIL || nMode == SESSION_SERVICE) {
      pSession->as_utility = true;
    } else {
      pSession->as_utility = false;
    } /* end if */

    /* and connect */
    i29lasterr_on(pSession);
    rc = i29sqlconnect(pSession);
    if (rc == DB_ERR_LOGFULL_CN00) {
      i29lasterr_on(pSession);
      pSession->as_utility = true;
      rc = i29sqlconnect(pSession);
    } // end if

    /* something wrong? */
    if (rc != 0) {
      nFuncReturn = ERR_SQL_CN00;
      i29sqllasterr(pSession, errtext, errname, &errcode, &errpos);
      cn90AnswerSQLError(replyData, replyLen, nFuncReturn, errtext, rc);
    }
  }

  /**/

  return nFuncReturn;
} /* end cn80_Connect */

/* PUBLIC FUNCTION cn80_InternConnect
 * ------------------------------------------------------------------
 * purpose: open a session
 *
 */
static tcn00_Error cn80_InternConnect (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax,
      int             nMode )
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  _TCHAR              szUsername[PARAMETER_MAXLEN_CN90];
  _TCHAR              szPassword[PARAMETER_MAXLEN_CN90];
  _TCHAR            * pPassword   = &szPassword[0];
  tin01_sql_session * pSession    = NULL;
  bool                bMallocOK   = true;

  /**/

  /* read the parameters */
  if (cn90GetToken(command->args, szUsername, 1)) {
    pPassword = _tcschr(szUsername, PWD_SEPARATOR_CN00);
    if (pPassword != NULL) {
      *pPassword = CHAR_STRINGTERM_CN90;
      pPassword++;
      cn90Uncrypt(pPassword, false);
    } else {
#ifdef CN71
      pPassword   = &szPassword[0];
      if (!cn90GetToken(command->args, szPassword, 2)) {
        nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
      } else {
        cn90Uncrypt(pPassword, false);
      } // end if
#else  /* CN71 */
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
#endif /* CN71 */
    } // end if
  } else {
    // look for cold user
    cn51DBMConfigValue  cfgColdUsr (vcontrol->dbname, CFG_COLDUSR_CN51);
    cn50DBMUser         usrCold    ( vcontrol->dbname,
                                     (tsp00_Namec) cfgColdUsr );

    _tcscpy(szUsername, usrCold.getUserName());
    cn90CalculateUncrypt(usrCold.getMasterPwd(), pPassword, false);
  } // end if

  /* Get session structure */
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80_Session(vcontrol, replyData, replyLen, &pSession, nMode);
    /* disconnect existing session */
    if (nFuncReturn == OK_CN00) {
      if (pSession->is_connected ) {
        i29release(pSession, true);
      } 
    } else {
      nFuncReturn = OK_CN00;
    } /* end if */
  } /* end if */

  /* allocate new memory for sql session */
  if (nFuncReturn == OK_CN00) {
    if (pSession == NULL) {
      cn00MyMalloc (sizeof (tin01_sql_session), &pSession, &bMallocOK);
    } /* end if */

    if ((pSession == NULL) || (!bMallocOK)) {
      nFuncReturn = ERR_MEM_CN00;
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } else {
      cn80_SetSession(vcontrol, pSession, nMode);
    } /* end if */
  } /* end if */

  /* execute connect */
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80_Connect(replyData, 
                               replyLen, 
                               pSession,
                               vcontrol->dbname,
                               szUsername, 
                               pPassword, 
                               nMode);
  } /* end if */

  /**/

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn80_InternConnect */

/* PUBLIC FUNCTION cn80_Release
 * ------------------------------------------------------------------
 * purpose: release the session
 *
 */
tcn00_Error cn80_Release (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax,
      int             nMode )
{
  /* free memory from sql session */
  cn80_FreeSession(vcontrol, nMode);
  
  return cn90AnswerOK(replyData, replyLen, NULL);
} /* end cn80_Release */

/* SPECIFICATION PUBLIC FUNCTIONS
 * ------------------------------------------------------------------
 */

/* PUBLIC FUNCTION cn80UtilConnect
 * ------------------------------------------------------------------
 * purpose: open a utility session
 *
 */
tcn00_Error cn80UtilConnect (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  return cn80_InternConnect(vcontrol, command, replyData, replyLen, replyLenMax, SESSION_UTIL);
} /* end cn80UtilConnect */

/* PUBLIC FUNCTION cn80UtilRelease
 * ------------------------------------------------------------------
 * purpose: release the utility session
 *
 */
tcn00_Error cn80UtilRelease (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  return cn80_Release(vcontrol, command, replyData, replyLen, replyLenMax, SESSION_UTIL);
} /* end cn80UtilRelease */

/* PRIVATE FUNCTION cn80_UtilActivate
 * ------------------------------------------------------------------
 * purpose: execute actvtivate serverdb
 *
 */
static tcn00_Error cn80_UtilActivate (
      VControlDataT         * vcontrol,
      const tsp00_UserNamec & szUser,
      const tsp00_Namec     & szPassword,
      char                  * replyData,
      int                   * replyLen,
      int                     replyLenMax )
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  tin01_sql_session * pSession;
  char                szCommand[PARAMETER_MAXLEN_CN90];

  /**/

  nFuncReturn = cn80UtilSession(vcontrol, replyData, replyLen, &pSession);

  // execute util command
  if (nFuncReturn == OK_CN00) {
    sprintf(szCommand, "ACTIVATE SERVERDB SYSDBA %s PASSWORD %s", szUser.asCharp(), szPassword.asCharp());
    nFuncReturn = cn80UtilExecSQL(replyData, replyLen, pSession, szCommand); 
  } /* end if */

  // usermanagement
  if (nFuncReturn == OK_CN00) {

    // read sysdba-configvalue
    cn51DBMConfigValue  cfgDBAUsr(vcontrol->dbname, CFG_DBAUSR_CN51);

    // read old sysdba-user
    cn50DBMUser         usrOldDBA(vcontrol->dbname, (tsp00_Namec) cfgDBAUsr );
    
    // delete old sysdba-user
    usrOldDBA.setExistsInDB(false);
    usrOldDBA.deleteUsr();
    
    // create new sysdba-user
    cn50DBMUser         usrDBA;

    // save data of new sysdba-user
    usrDBA.setDBName(vcontrol->dbname)
          .setUserName ( szUser)
          .setSDBAUser ( )
          .setSQLUser  ( )
          .setMasterPwd( szPassword)
          .setDBMUser      ( true )
          .setServerRights ( DBMMaskAll_CN50, DBMUserDefault_CN50 )
          .setGUIRights    ( DBMMaskAll_CN50, DBMUserDefault_CN50 );

    nFuncReturn = usrDBA.save();
    if (nFuncReturn != OK_CN00) {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } // end if 

    // save sysdba-configvalue
    cfgDBAUsr = szUser;

    // delete Domain-User
    cn50DBMUser         usrDom( vcontrol->dbname, USERNAME_DOMAIN_CN50);
    usrDom.setExistsInDB(false);
    usrDom.deleteUsr();
  } // end if

  /**/

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn80_UtilActivate */

/* PRIVATE FUNCTION cn80_CheckActivate
 * ------------------------------------------------------------------
 * purpose: execute actvtivate serverdb
 *
 */
static bool cn80_CheckActivate (
      const _TCHAR    * szStatement,
      tsp00_UserNamec & szUser,
      tsp00_Namec     & szPassword) 
{
  bool      bActivate = false;
  _TCHAR    szToken    [PARAMETER_MAXLEN_CN90];

  cn90GetToken(szStatement, szToken, 1);
  if (_tcsicmp(szToken, "ACTIVATE") == 0) {
    cn90GetToken(szStatement, szToken, 2);
    if (_tcsicmp(szToken, "SERVERDB") == 0) {
      cn90GetToken(szStatement, szToken, 3);
      if (_tcsicmp(szToken, "SYSDBA") == 0) {
      } // end if
        cn90GetToken(szStatement, szToken, 4);
        szUser.rawAssign(szToken);
        cn90GetToken(szStatement, szToken, 5);
        if (_tcsicmp(szToken, "PASSWORD") == 0) {
          cn90GetToken(szStatement, szToken, 6);
          szPassword.rawAssign(szToken);
          bActivate = true;
        } // end if
    } // end if
  } // end if

  return bActivate;
} // end cn80_CheckActivate

/* PUBLIC FUNCTION cn80UtilActivate
 * ------------------------------------------------------------------
 * purpose: execute actvtivate serverdb
 *
 */
tcn00_Error cn80UtilActivate (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  tin01_sql_session * pSession;
  _TCHAR              szToken    [PARAMETER_MAXLEN_CN90];
  _TCHAR            * pPassword   = NULL; 
  tsp00_UserNamec     szUser;
  tsp00_Namec         szPassword;

  /**/

  nFuncReturn = cn80UtilSession(vcontrol, replyData, replyLen, &pSession);

  if (nFuncReturn == OK_CN00) {

    /* read the parameters */
    cn90GetToken(command->args, szToken, 1);
    pPassword = _tcschr(szToken, PWD_SEPARATOR_CN00);
    if (pPassword != NULL) {
      *pPassword = CHAR_STRINGTERM_CN90;
      pPassword++;
      cn90Uncrypt(pPassword, false);
      szUser.rawAssign(szToken);
      szPassword.rawAssign(pPassword);
    } else {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_PARAM_CN00);
    } // end if
  } // end if

  // ... and go
  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80_UtilActivate(vcontrol, szUser, szPassword, replyData, replyLen, replyLenMax);
  } // end if

  return nFuncReturn;
} /* end cn80UtilActivate */

/* PUBLIC FUNCTION cn80UtilExecute
 * ------------------------------------------------------------------
 * purpose: execute a utility statement
 *
 */
tcn00_Error cn80UtilExecute (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	         nFuncReturn = OK_CN00;
  tin01_sql_session  * pSession;
  tin01_sql_session    aSQLSession;
  bool                 bConnect    = false;
  
  /**/

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80UtilSession(vcontrol, replyData, replyLen, &pSession);
    if (nFuncReturn == ERR_NOUTILSESSION_CN00) {
      pSession = &aSQLSession;
      nFuncReturn = cn80ConnectUtil ( vcontrol->dbname, pSession, replyData,replyLen);
      bConnect = (nFuncReturn == OK_CN00);
    } // end if
  } // end if
  // nFuncReturn = cn80UtilSession(vcontrol, replyData, replyLen, &pSession);

  if (nFuncReturn == OK_CN00) {
    tsp00_UserNamec     szUser;
    tsp00_Namec         szPassword;

    if (cn80_CheckActivate(command->args, szUser, szPassword)) {
      nFuncReturn = cn80_UtilActivate(vcontrol, szUser, szPassword, replyData, replyLen, replyLenMax);
    } else { 
      nFuncReturn = cn80UtilExecSQL(replyData, replyLen, pSession, command->args); 
    } /* end if */
  } /* end if */

  if (bConnect) {
    cn80ReleaseUtil ( pSession );
  } // end if

  /**/

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn80UtilExecute */

/* PUBLIC FUNCTION cn80UtilGetPage
 * ------------------------------------------------------------------
 * purpose: reads a page from kernel
 *
 */
tcn00_Error cn80UtilGetPage (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
#if defined (CN72)

  tcn00_Error	        nFuncReturn = OK_CN00;
  char                szStatement[PARAMETER_MAXLEN_CN90];
  tin01_sql_session * pSession;
  int                 rc;
  long                nPageSize;
  void              * pPageBuf;
  tsp00_C256c         errtext;

  /**/

  nFuncReturn = cn80_DiagnoseStatement (vcontrol,
                                        command->args,
                                        PAGE_GET_STMT,
                                        &szStatement[0],
                                        NULL);

  if (nFuncReturn != OK_CN00) {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  if (nFuncReturn == OK_CN00) {
    /* get util session */
    nFuncReturn = cn80UtilSession(vcontrol, replyData, replyLen, &pSession);
  } /* end if */

  if (nFuncReturn == OK_CN00) {
    rc = i29utildiagtocmd(pSession, szStatement, &nPageSize, &pPageBuf);

    if (rc != 0) {
      nFuncReturn = ERR_SQL_CN00;

      i29errmsg(pSession, errtext);

      cn90AnswerSQLError(replyData, replyLen, nFuncReturn, (char *)errtext, rc);
    } else {
      sprintf (replyData, "%s%s%ld%s", ANSWER_OK_CN00, LINE_SEPSTRING_CN00,
                                       nPageSize,      LINE_SEPSTRING_CN00);
      
      *replyLen = (int)strlen(replyData);
      memcpy(&replyData[*replyLen], pPageBuf, nPageSize);
      *replyLen = *replyLen + nPageSize;
    } /* end if */
  } /* end if */

  if (nFuncReturn != OK_CN00) {
    *replyLen = (int)strlen(replyData);
  } /* end if */
 
  /**/

  return nFuncReturn;

#else

  cn90AnswerIError(replyData, replyLen, ERR_NOTIMPL_CN00);
  return ERR_NOTIMPL_CN00;

#endif 
} /* end cn80UtilGetPage */

/* PUBLIC FUNCTION cn80UtilPutPage
 * ------------------------------------------------------------------
 * purpose: writes a page to kernel
 *
 */
tcn00_Error cn80UtilPutPage (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
#if defined (CN72)

  tcn00_Error	        nFuncReturn = OK_CN00;
  char                szStatement[PARAMETER_MAXLEN_CN90];
  tin01_sql_session * pSession;
  int                 rc;
  long                nPageSize;
  const void        * pPageBuf;
  tsp00_C256c         errtext;

  nFuncReturn = cn80_DiagnoseStatement (vcontrol,
                                        command->args,
                                        PAGE_PUT_STMT,
                                        &szStatement[0],
                                        &nPageSize);

  if (nFuncReturn != OK_CN00) {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  if (nFuncReturn == OK_CN00) {
    /* get util session */
    nFuncReturn = cn80UtilSession(vcontrol, replyData, replyLen, &pSession);
  } /* end if */

  if (nFuncReturn == OK_CN00) {

    // look for page data
    pPageBuf = strstr(command->args, LINE_SEPSTRING_CN00) + strlen(LINE_SEPSTRING_CN00);

    rc = i29utildiagrestcmd(pSession, szStatement, nPageSize, pPageBuf);

    if (rc != 0) {
      nFuncReturn = ERR_SQL_CN00;
      i29errmsg(pSession, errtext);
      cn90AnswerSQLError(replyData, replyLen, nFuncReturn, (char *)errtext, rc);
    } else {
      cn90AnswerOK(replyData, replyLen, NULL);
    } /* end if */
  } /* end if */

  return nFuncReturn;

#else

  cn90AnswerIError(replyData, replyLen, ERR_NOTIMPL_CN00);
  return ERR_NOTIMPL_CN00;

#endif 
} /* end cn80UtilPutPage */

/* PUBLIC FUNCTION cn80ServiceConnect
 * ------------------------------------------------------------------
 * purpose: open a utility session with the service kernel
 *
 */
tcn00_Error cn80ServiceConnect (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_DbNamec       szDBName;
  tsp00_UserNamec     szUserName;
  tsp00_Namec         szPassword;
  tsp9_rte_xerror     xError;
  bool                bStarted = false; 
  tin01_sql_session * pSession    = NULL;
  bool                bMallocOK   = true;

  //  get parameter of service kernel
  nFuncReturn =  cn20XParamGetServiceValues ( szDBName,
                                              vcontrol->dbroot,
                                              szUserName,
                                              szPassword );

  //  start service kernel
  if (nFuncReturn == OK_CN00) {
    cn20CreateSrvParam(szDBName, true);

    sqlxclear( szDBName, vcontrol->dbroot, &xError );

    if (xError.xe_result == csp9_xrte_ok) {
      sqlxstart (szDBName, csp9_fast_pgm, vcontrol->dbroot, false, 0, NULL, &xError);
      if (xError.xe_result != csp9_xrte_ok) {
        // try quick
        sqlxstart (szDBName, csp9_quick_pgm, vcontrol->dbroot, false, 0, NULL, &xError);
        if (xError.xe_result != csp9_xrte_ok) {
          // try slow
          sqlxstart (szDBName, csp9_slow_pgm, vcontrol->dbroot, false, 0, NULL, &xError);
        } // end if
      } // end if
    } // end if

    if (xError.xe_result != csp9_xrte_ok) {
      nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
    } else {
      bStarted = true;
    } // end if
  } else {
    cn90AnswerIError(replyData, replyLen, nFuncReturn);
  } // end if

  // connect
  if (nFuncReturn == OK_CN00) {
    /* allocate new memory for sql session */
    cn00MyMalloc (sizeof (tin01_sql_session), &pSession, &bMallocOK);

    if ((pSession == NULL) || (!bMallocOK)) {
      nFuncReturn = cn90AnswerIError(replyData, replyLen, ERR_MEM_CN00);
    } else {
      cn80_SetSession(vcontrol, pSession, SESSION_SERVICE);
      nFuncReturn = cn80_Connect(replyData, 
                                 replyLen, 
                                 pSession,
                                 szDBName,
                                 szUserName, 
                                 szPassword, 
                                 SESSION_SERVICE);
    } /* end if */
  } // end if

  // stop kernel in case of errors
  if (nFuncReturn != OK_CN00 && bStarted) {
    sqlxstop (szDBName, vcontrol->dbroot, false, &xError);
  } // end if

  if (nFuncReturn == OK_CN00) {
    cn90AnswerOK(replyData, replyLen, NULL);
  } /* end if */

  return nFuncReturn;

} /* end cn80ServiceConnect */

/* PUBLIC FUNCTION cn80ServiceRelease
 * ------------------------------------------------------------------
 * purpose: release the utility session with the service kernel
 *
 */
tcn00_Error cn80ServiceRelease (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error         nFuncReturn = OK_CN00;
  tsp00_DbNamec       szDBName;
  tsp9_rte_xerror     xError;
  tin01_sql_session * pSession = NULL;

  if (cn80_Session (vcontrol, replyData, replyLen, &pSession, SESSION_SERVICE ) == OK_CN00) {
    nFuncReturn = cn80_Release(vcontrol, command, replyData, replyLen, replyLenMax, SESSION_SERVICE);

    //  get parameter of service kernel
    nFuncReturn =  cn20XParamGetServiceValues ( szDBName, vcontrol->dbroot);
    if (nFuncReturn == OK_CN00) {
      sqlxstop (szDBName, vcontrol->dbroot, false, &xError);
      if (xError.xe_result != csp9_xrte_ok) {
        nFuncReturn = cn90AnswerRTEError (replyData, replyLen, ERR_RTE_CN00, xError.xe_text, sizeof(xError.xe_text), xError.xe_result);
      } else {
        cn90AnswerOK(replyData, replyLen, NULL);
      } // end if
    } else {
      cn90AnswerIError(replyData, replyLen, nFuncReturn);
    } // end if
  } else {
    nFuncReturn = cn80_Release(vcontrol, command, replyData, replyLen, replyLenMax, SESSION_SERVICE);
  } // end if

  return nFuncReturn;

} /* end cn80ServiceRelease */

/* PUBLIC FUNCTION cn80SQLUserConnect
 * ------------------------------------------------------------------
 * purpose: open a SQLUser session
 *
 */
tcn00_Error cn80SQLUserConnect (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  return cn80_InternConnect(vcontrol, command, replyData, replyLen, replyLenMax, SESSION_SQLUSER);
} /* end cn80SQLUserConnect */

/* PUBLIC FUNCTION cn80SQLUserRelease
 * ------------------------------------------------------------------
 * purpose: release the SQLUser session
 *
 */
tcn00_Error cn80SQLUserRelease (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  return cn80_Release(vcontrol, command, replyData, replyLen, replyLenMax, SESSION_SQLUSER);
} /* end cn80SQLUserRelease */

/* PUBLIC FUNCTION cn80SQLUserExecute
 * ------------------------------------------------------------------
 * purpose: execute a statement with the SQLUser session
 *
 */
tcn00_Error cn80SQLUserExecute (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  tin01_sql_session  * pSession;
  tin01_sql_session    aSQLSession;
  bool                 bConnect    = false;
  
  /**/

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80SQLUserSession(vcontrol, replyData, replyLen, &pSession);
    if (nFuncReturn == ERR_NOSQLUSERSESSION_CN00) {
      pSession = &aSQLSession;
      nFuncReturn = cn80ConnectSQL ( vcontrol->dbname, pSession, replyData,replyLen);
      bConnect = (nFuncReturn == OK_CN00);
    } // end if
  } // end if
  // nFuncReturn = cn80SQLUserSession(vcontrol, replyData, replyLen, &pSession);

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80SQLExecSQL(replyData, replyLen,  replyLenMax, pSession, command->args); 
  } /* end if */

  if (bConnect) {
    cn80ReleaseSQL ( pSession );
  } // end if

  /**/

  *replyLen = (int)strlen(replyData);

  return nFuncReturn;
} /* end cn80SQLUserExecute */

/* PUBLIC FUNCTION cn80SQLUserInfo
 * ------------------------------------------------------------------
 * purpose: provide info about sql result
 *
 */
tcn00_Error cn80SQLUserInfo (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  tin01_sql_session  * pSession;
  int                 rc;
  tsp00_C256c         errtext;
  
  /**/

  nFuncReturn = cn80SQLUserSession(vcontrol, replyData, replyLen, &pSession);

  if (nFuncReturn == OK_CN00) {

    cn90AnswerOK(replyData, replyLen, NULL);

    /* execute select */
    rc = i29selectinfo(pSession, 
                       command->args, 
                       replyData + strlen(replyData), 
                       SQL_CHAR_FIELDSEP);

    strcpy(&replyData[strlen(replyData)], LINE_SEPSTRING_CN00);

    if (rc != 0) {
      nFuncReturn = ERR_SQL_CN00;
      i29errmsg(pSession, errtext);
      cn90AnswerSQLError(replyData, replyLen, nFuncReturn, (char *)errtext, rc);
    } /* end if */
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return OK_CN00;
} /* end cn80SQLUserInfo */

/* PUBLIC FUNCTION cn80SQLUserFetch
 * ------------------------------------------------------------------
 * purpose: fetch next rows
 *
 */
tcn00_Error cn80SQLUserFetch (
      VControlDataT * vcontrol,
      CommandT      * command,
      char          * replyData,
      int           * replyLen,
      int             replyLenMax )
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  tin01_sql_session  * pSession;
  int                 rc;
  tsp00_C256c         errtext;
   
  /**/

  nFuncReturn = cn80SQLUserSession(vcontrol, replyData, replyLen, &pSession);

  if (nFuncReturn == OK_CN00) {

    cn90AnswerOK(replyData, replyLen, NULL);

    /* execute select */
    rc = i29fetch(pSession, 
                  replyData + strlen(replyData), 
                  replyLenMax - (long)strlen(replyData),
                  SQL_CHAR_FIELDSEP,SQL_CHAR_STRINGDEL,SQL_CHAR_RECSEP);

    strcpy(&replyData[strlen(replyData)], LINE_SEPSTRING_CN00);

    if (rc != 0) {
      nFuncReturn = ERR_SQL_CN00;
      i29errmsg(pSession, errtext);
      cn90AnswerSQLError(replyData, replyLen, nFuncReturn, (char *)errtext, rc);
    } /* end if */
  } /* end if */

  *replyLen = (int)strlen(replyData);

  return OK_CN00;
} /* end cn80SQLUserFetch */

/*
 * ------------------------------------------------------------------
 * FOR INTERNAL USE ONLY
 * ------------------------------------------------------------------
 */

/* INTERN FUNCTION cn80UtilExecSQL
 * ------------------------------------------------------------------
 * purpose: execute a statement
 *
 */
tcn00_Error cn80UtilExecSQL (
      char              * replyData,
      int               * replyLen,
      tin01_sql_session * pSession,
      const char        * szStatement)
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  int                 rc;
  tsp00_C256c         errtext;

  /**/

  rc = i29utility(pSession, szStatement);
  if (rc == -8888) {
    pSession->is_connected = false;
    rc = i29sqlconnect(pSession);
    if (rc == 0) {
      rc = i29utility(pSession, szStatement);
    } /* end if */
  } /* end if */

  if (rc != 0) {
    nFuncReturn = ERR_SQL_CN00;

    i29errmsg(pSession, errtext);

    cn90AnswerSQLError(replyData, replyLen, nFuncReturn, (char *)errtext, rc);
  }
 
  /**/

  return nFuncReturn;
} /* end cn80UtilExecSQL */

/* INTERN FUNCTION cn80SQLExecSQL
 * ------------------------------------------------------------------
 * purpose: execute a statement
 *
 */
tcn00_Error cn80SQLExecSQL (
      char               * replyData,
      int                * replyLen,
      int                  replyLenMax,
      tin01_sql_session  * pSession,
      const char         * szStatement)
{
  tcn00_Error	        nFuncReturn = OK_CN00;
  int                 rc;
  tsp00_C256c         errtext;
  char                szCommand[PARAMETER_MAXLEN_CN90];
  bool                bSelect = false;
  /**/

  /* check the command */
  if (cn90GetToken(szStatement, szCommand, 1)) {
    bSelect = (cn90Stricmp(szCommand, SQL_SELECT) == 0);
  } /* end if */

  /* execute statement */
  if (!bSelect) {
    /* execute non select */
    rc = i29adbs(pSession, szStatement);

    if (rc != 0) {
      nFuncReturn = ERR_SQL_CN00;
      i29errmsg(pSession, errtext);
      cn90AnswerSQLError(replyData, replyLen, nFuncReturn, errtext, rc);
    } else {
      cn90AnswerOK(replyData, replyLen, NULL);
    } /* end if */

  } else {
    cn90AnswerOK(replyData, replyLen, NULL);

    /* execute select */
    rc = i29select(pSession, 
                   szStatement, 
                   replyData + strlen(replyData), 
                   replyLenMax - (long)strlen(replyData),
                   SQL_CHAR_FIELDSEP,SQL_CHAR_STRINGDEL,SQL_CHAR_RECSEP);

    strcpy(&replyData[strlen(replyData)], LINE_SEPSTRING_CN00);

    if (rc != 0) {
      nFuncReturn = ERR_SQL_CN00;
      i29errmsg(pSession, errtext);
      cn90AnswerSQLError(replyData, replyLen, nFuncReturn, (char *)errtext, rc);
    } else {
      *replyLen = (int)strlen(replyData);
    } /* end if */
  } /* end if */
 
  /**/

  return nFuncReturn;
} /* end cn80SQLExecSQL */

/* INTERN FUNCTION cn80UtilSession
 * ------------------------------------------------------------------
 * purpose: get the utility session from global struct and check
 *
 */
tcn00_Error cn80UtilSession (
      VControlDataT    *  vcontrol,
      char             *  replyData,
      int              *  replyLen,
      tin01_sql_session ** pUtilSession )
{
  return cn80_Session(vcontrol, replyData, replyLen, pUtilSession, SESSION_UTIL);
} /* end cn80UtilSession */

/* INTERN FUNCTION cn80ServiceSession
 * ------------------------------------------------------------------
 * purpose: get the service kernel utility session from global struct and check
 *
 */
tcn00_Error cn80ServiceSession (
      VControlDataT    *  vcontrol,
      char             *  replyData,
      int              *  replyLen,
      tin01_sql_session ** pUtilSession )
{
  return cn80_Session(vcontrol, replyData, replyLen, pUtilSession, SESSION_SERVICE);
} /* end cn80ServiceSession */

/* INTERN FUNCTION cn80SQLUserSession
 * ------------------------------------------------------------------
 * purpose: get the sql user session from global struct and check
 *
 */
tcn00_Error cn80SQLUserSession (
      VControlDataT    *  vcontrol,
      char             *  replyData,
      int              *  replyLen,
      tin01_sql_session ** pSQLUserSession )
{
  return cn80_Session(vcontrol, replyData, replyLen, pSQLUserSession, SESSION_SQLUSER);
} /* end cn80SQLUserSession */

/* INTERN FUNCTION cn80FreeUtilSession
 * ------------------------------------------------------------------
 */
void cn80FreeUtilSession(
      VControlDataT     * vcontrol)
{
  cn80_FreeSession(vcontrol, SESSION_UTIL);
} /* end cn80FreeUtilSession */
 
/* INTERN FUNCTION cn80FreeSQLUserSession
 * ------------------------------------------------------------------
 */
void cn80FreeSQLUserSession(
      VControlDataT     * vcontrol)
{
  cn80_FreeSession(vcontrol, SESSION_SQLUSER);
} /* end cn80FreeSQLUserSession */

/* INTERN FUNCTION cn80FreeServiceSession
 * ------------------------------------------------------------------
 */
void cn80FreeServiceSession(
      VControlDataT     * vcontrol)
{
  tcn00_Error      nFuncReturn = OK_CN00;
  tsp00_DbNamec    szDBName;
  tsp9_rte_xerror  xError;

  if (vcontrol->pServiceSession != NULL) {
    cn80_FreeSession(vcontrol, SESSION_SERVICE);

    //  get parameter of service kernel
    nFuncReturn =  cn20XParamGetServiceValues ( szDBName ,vcontrol->dbroot);
    // stop service kernel
    if (nFuncReturn == OK_CN00) {
      sqlxstop (szDBName, vcontrol->dbroot, false, &xError);
    } // end if

  } // end if

} /* end cn80FreeServiceSession */

/*
  -----------------------------------------------------------------------------
  public function cn80ExecuteSQL
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteSQL 
      ( const tsp00_DbNamec   szDBName,
        const _TCHAR      * szStatement)
{
  _TCHAR              szReply[PARAMETER_MAXLEN_CN90];
  int                 replyLen;

  return cn80ExecuteSQL (szDBName,
                         szStatement,
                         &szReply[0],
                         PARAMETER_MAXLEN_CN90,
                         replyLen);
} // end cn80ExecuteSQL
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteSQL 
      ( const tsp00_DbNamec   szDBName,
        const _TCHAR        * szStatement,
        _TCHAR              * replyData,
        int                   replyLenMax,
        int                 & replyLen )

{
  cn51DBMConfigValue  cfgColdUsr (szDBName,  CFG_COLDUSR_CN51);
  cn50DBMUser         usrCold    (szDBName, (tsp00_Namec) cfgColdUsr );

  return cn80ExecuteSQL ( usrCold,
                          szStatement,
                          replyData,
                          replyLenMax,
                          replyLen );
} // end cn80ExecuteSQL
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteSQL  
      ( const cn50DBMUser   aUser,
        const _TCHAR      * szStatement)
{
  _TCHAR              szReply[PARAMETER_MAXLEN_CN90];
  int                 replyLen;

  return cn80ExecuteSQL (aUser,
                         szStatement,
                         &szReply[0],
                         PARAMETER_MAXLEN_CN90,
                         replyLen);
} // end cn80ExecuteSQL
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteSQL 
      ( const cn50DBMUser     aUser,
        const _TCHAR        * szStatement,
        _TCHAR              * replyData,
        int                   replyLenMax,
        int                 & replyLen )

{
  tcn00_Error         nFuncReturn  = OK_CN00;
  tin01_sql_session   aSession;

  nFuncReturn = cn80ConnectSQL(aUser,
                               &aSession,
                               replyData,
                               &replyLen);

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80ExecuteSQL(&aSession, 
                                 szStatement,
                                 replyData,  
                                 replyLenMax, 
                                 replyLen);

    cn80ReleaseSQL(&aSession);
  } // end if

  return nFuncReturn;
} // end cn80ExecuteSQL
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteSQL  
      ( tin01_sql_session   * pSQLSession,
        const _TCHAR        * szStatement,
        _TCHAR              * replyData,
        int                   replyLenMax,
        int                 & replyLen )
{
    return cn80SQLExecSQL(replyData,  
                          &replyLen,  
                          replyLenMax, 
                          pSQLSession, 
                          szStatement);
} // end cn80ExecuteSQL


/*
  -----------------------------------------------------------------------------
  public function cn80ConnectSQL
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ConnectSQL
      ( const tsp00_DbNamec   szDBName,
        tin01_sql_session   * pSQLSession,
        char                * replyData,
        int                 * replyLen )
{
  cn51DBMConfigValue  cfgColdUsr (szDBName,  CFG_COLDUSR_CN51);
  cn50DBMUser         usrCold    (szDBName, (tsp00_Namec) cfgColdUsr );

  return cn80ConnectSQL(usrCold, pSQLSession, replyData, replyLen);

} // end cn80ConnectSQL
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ConnectSQL 
      ( const cn50DBMUser     aUser,
        tin01_sql_session   * pSQLSession,
        char                * replyData,
        int                 * replyLen )
{
  _TCHAR              szReply[PARAMETER_MAXLEN_CN90];
  int                 nReplyLen;

  if (replyData == NULL) {
    replyData = &szReply[0];
    replyLen  = &nReplyLen;
  } // end if

  return  cn80_Connect(replyData, 
                       replyLen, 
                       pSQLSession,
                       aUser.getDBName(),
                       aUser.getUserName(), 
                       aUser.getMasterPwd(true), 
                       SESSION_SQLUSER);

} // end cn80ConnectSQL

/*
  -----------------------------------------------------------------------------
  public function cn80ReleaseSQL
  -----------------------------------------------------------------------------
 */
void cn80ReleaseSQL 
     ( tin01_sql_session * pSQLSession,
       bool                bWithCommit )
{
  if (pSQLSession->is_connected ) {
    i29release(pSQLSession, bWithCommit);
  } // end if

} // end cn80ReleaseSQL

/*
  -----------------------------------------------------------------------------
  public function cn80ExecuteUtil
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteUtil (const tsp00_DbNamec   szDBName,
                             const _TCHAR        * szStatement,
                             char                * replyData,
                             int                 * replyLen)
{
  cn51DBMConfigValue  cfgColdUsr (szDBName,  CFG_COLDUSR_CN51);
  cn50DBMUser         usrCold    (szDBName, (tsp00_Namec) cfgColdUsr );

  return cn80ExecuteUtil(usrCold, szStatement, replyData, replyLen);

} // end cn80ExecuteUtil
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteUtil (const cn50DBMUser     aUser,
                             const _TCHAR        * szStatement,
                             char                * replyData,
                             int                 * replyLen)
{
  tcn00_Error         nFuncReturn  = OK_CN00;
  tin01_sql_session   aSession;

  nFuncReturn = cn80ConnectUtil(aUser,
                                &aSession,
                                replyData, 
                                replyLen);

  if (nFuncReturn == OK_CN00) {
    nFuncReturn = cn80ExecuteUtil(&aSession, 
                                  szStatement,
                                  replyData,  
                                  replyLen);

    cn80ReleaseUtil(&aSession);
  } // end if


  return nFuncReturn;
} // end cn80ExecuteUtil
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ExecuteUtil ( tin01_sql_session   * pUtilSession,
                              const _TCHAR        * szStatement,
                              char                * replyData,
                              int                 * replyLen )
{
  _TCHAR              szReply[PARAMETER_MAXLEN_CN90];
  int                 nReplyLen;

  if (replyData == NULL) {
    replyData = &szReply[0];
    replyLen  = &nReplyLen;
  } // end if

  return cn80UtilExecSQL(replyData,  
                         replyLen,  
                         pUtilSession, 
                         szStatement);
} // end cn80ExecuteUtil

/*
  -----------------------------------------------------------------------------
  public function cn80ConnectUtil
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ConnectUtil 
      ( const tsp00_DbNamec   szDBName,
        tin01_sql_session   * pUtilSession,
        char                * replyData,
        int                 * replyLen )
{
  cn51DBMConfigValue  cfgColdUsr (szDBName,  CFG_COLDUSR_CN51);
  cn50DBMUser         usrCold    (szDBName, (tsp00_Namec) cfgColdUsr );

  return cn80ConnectUtil(usrCold, pUtilSession, replyData, replyLen);

} // end cn80ConnectUtil
/*
  -----------------------------------------------------------------------------
 */
tcn00_Error cn80ConnectUtil 
      ( const cn50DBMUser     aUser,
        tin01_sql_session   * pUtilSession,
        char                * replyData,
        int                 * replyLen )
{
  _TCHAR              szReply[PARAMETER_MAXLEN_CN90];
  int                 nReplyLen;

  if (replyData == NULL) {
    replyData = &szReply[0];
    replyLen  = &nReplyLen;
  } // end if

  return  cn80_Connect(replyData, 
                       replyLen, 
                       pUtilSession,
                       aUser.getDBName(),
                       aUser.getUserName(), 
                       aUser.getMasterPwd(true), 
                       SESSION_UTIL);

} // end cn80ConnectUtil

/*
  -----------------------------------------------------------------------------
  public function cn80ReleaseUtil
  -----------------------------------------------------------------------------
 */
void cn80ReleaseUtil 
     ( tin01_sql_session * pUtilSession,
       bool                bWithCommit )
{
  if (pUtilSession->is_connected ) {
    i29release(pUtilSession, bWithCommit);
  } // end if

} // end cn80ReleaseUtil

