/****************************************************************************

  module      : geo52.h

  -------------------------------------------------------------------------

  responsible : FrankH

  special area: LZU Statistics
  description : description ...


  last changed: 2000-06-19  11:03
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/


#ifndef GEO52_H
#define GEO52_H



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#include "geo00.h"

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#define MAX_DEVQUEUELEN_STAT_EO52 3

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/

#define EO52_TIME_DIFF(_start_sec, _start_ms, _stop_sec, _stop_ms, _sec, _ms ) \
  { int _time_diff_sec = _stop_sec - _start_sec ,      \
        _time_diff_ms  = _stop_ms  - _start_ms  ;      \
    if ( _time_diff_ms < 0 )                           \
      { _time_diff_sec--; _time_diff_ms += 1000000 ; } \
    _sec += _time_diff_sec ;                           \
    _ms  += _time_diff_ms  ;                           \
    if ( _ms >= 1000000 )                              \
      { _sec++; _ms -= 1000000 ; }                     \
  }

/*===========================================================================*
 *  CLASSES, STRUCTURES, TYPES, UNIONS ...                                   *
 *===========================================================================*/

/*
// +---------------------------------------------------------------+
// |   STATE STATISTIC                                             |
// +---------------------------------------------------------------+
*/

 typedef struct teo52_ActionTimeRec
   {
   tsp00_Uint                              ulCount;
   tsp00_Uint                              ulSec;
   tsp00_Uint                              ulMicroSec;
   } teo52_ActionTime ;


 typedef struct teo52_StateStatisticRec
   {
   tsp00_Uint                              ulCount;
   tsp00_Uint                              ulRelSec;
   tsp00_Uint                              ulRelMicroSec;
   tsp00_Uint                              ulSec;
   tsp00_Uint                              ulMicroSec;
   } teo52_StateStatistic;

 typedef struct teo52_IOStatisticRec
   {
   tsp00_Uint                              ulCount;
   tsp00_Uint                              ulPageCount;
   tsp00_Uint                              ulSec;
   tsp00_Uint                              ulMicroSec;
   } teo52_IOStatisticRec ;


 typedef struct teo52_TaskDevIOStatisticRec
   {
   tsp00_Uint                              ulCount;
   tsp00_Uint                              ulPageCount;
   tsp00_Uint                              ulRelSec;
   tsp00_Uint                              ulRelMicroSec;
   tsp00_Uint                              ulSec;
   tsp00_Uint                              ulMicroSec;
   } teo52_TaskDevIOStatistic;

 /* PTS 1106969 */
 typedef struct teo52_TaskOMSStatisticRec
   {
   tsp00_KnlIdentifierc                    ProcedurName;
   tsp00_Uint                              CallbackCount;
   tsp00_Int4                              MethodIdx;
   tsp00_Int4                              OMSCallActive;
   } teo52_TaskOMSStatisticRec ;
/*
// +---------------------------------------------------------------+
// |   TASK STATE STATISTIC                                        |
// +---------------------------------------------------------------+
*/

typedef struct teo52_TaskStateStatisticRec
  {
  teo52_IOStatisticRec            SelfIOReadStat, SelfIOWriteStat ;
  teo52_TaskDevIOStatistic        DevIOReadStat , DevIOWriteStat ;
  teo52_TaskDevIOStatistic        AsynIOReadStat, AsynIOWriteStat ;
  tsp00_Uint                      ulDevQueueLen[MAX_DEVQUEUELEN_STAT_EO52] ;
  teo52_StateStatistic            VwaitStat , VsuspStat, VsleepStat ;
  teo52_StateStatistic            RcvRplStat, RplRcvStat ;
  tsp00_Int4                      iRcvRplStartSec, iRcvRplStartMicroSec;
  tsp00_Int4	                  iRelStopSec, iRelStopMicroSec;
  tsp00_Uint                      ulRcvRplToLong ;
  tsp00_Uint                      ulReschedMsecCount;
  } teo52_TaskStateStatisticRec;


typedef struct teo52_TaskQueueStatisticRec
  {
  tsp00_Uint                               ulUKTQueCount;
  tsp00_Uint                               ulUToUQueCount;
  tsp00_Uint                               ulCOMQueCount;
  } teo52_TaskQueueStatistic;

/*
// +---------------------------------------------------------------+
// |   TASK PRIO STATISTIC                                         |
// +---------------------------------------------------------------+
*/
typedef struct teo52_TaskPrioStatisticRec
  {
  tsp00_Uint                               ulTotalCount;
  tsp00_Uint                               ulOtherUkt;
  } teo52_TaskPrioStatistic;

/*
// +---------------------------------------------------------------+
// |   DEV / ASDEV I/O STATISTIC                                         |
// +---------------------------------------------------------------+
*/
typedef struct teo52_DevIOStatisticRec
  {
  teo52_IOStatisticRec            DevRead;
  teo52_IOStatisticRec            DevWrite;
  tsp00_Uint                      ulIOQueLen;
  tsp00_Uint                      ulMaxIOQueLen;
  } teo52_DevIOStatistic ;

/*===========================================================================*
 *  FUNCTION PROTOTYPES                                                      *
 *===========================================================================*/


/*===========================================================================*
 *  EXTERNAL VARIABLES                                                       *
 *===========================================================================*/


#endif  /* GEO52_H */
