/****************************************************************************

  module      : veo43.cpp

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: c_eo43xshow
  description : class interface to access X_CONS show output

  last changed: 2000-08-10  16:18
  see also    : example.html ...

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

#if defined(UNICODE)
#undef UNICODE
#endif /* UNICODE */
/*
 * Sorry no UNICODE variant possible yet...
 */

#include "heo01.h"
#include "geo43.h"
#include <string.h>

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#if defined(_WIN32)
#if defined(UNICODE)
#define POPEN_CALL _wpopen
#define PCLOSE_CALL _wpclose
#define FGETS_CALL  fgetws
#else
#define POPEN_CALL _popen
#define PCLOSE_CALL _pclose
#define FGETS_CALL  fgets
#endif /* UNICODE */
#else
#define POPEN_CALL  popen
#define PCLOSE_CALL pclose
#define FGETS_CALL  fgets
#endif

/* space is needed to separate from database name */
#ifdef _WIN32
  #define X_CONS_EXE         _T("cons ")
#else
  #define X_CONS_EXE         _T("console ")
#endif
/* spaces are needed to separate from database name and argument */
#define SHOW_CMD _T(" show ")

/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  GLOBAL VARIABLES                                                         *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL VARIABLES                                                          *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  PUBLIC FUNCTIONS (CODE)                                                  *
 *===========================================================================*/

ceo43_CmdPipe::ceo43_CmdPipe()
:m_CommandPipe_eo43((FILE *)0) 
{}

void ceo43_CmdPipe::eo43OpenCommand(const TCHAR *command, tsp00_Bool & openOk )
{
    tsp00_Int4 lastExitCode = 0;

    eo43CloseCommand( lastExitCode );

    size_t cmdLen = _tcslen(command);

    if ( cmdLen > 0 )
    {
        m_CommandPipe_eo43 = POPEN_CALL(command, _T("r") );
    }

    openOk = (m_CommandPipe_eo43 != NULL);
}

void ceo43_CmdPipe::eo43CloseCommand( tsp00_Int4 & ExitCode )
{
    if ( m_CommandPipe_eo43 )
    {
        m_LastExitCode_eo43 = PCLOSE_CALL(m_CommandPipe_eo43);
        m_CommandPipe_eo43  = (FILE *)0;
    }
    ExitCode = m_LastExitCode_eo43;
}

ceo43_CmdPipe::~ceo43_CmdPipe()
{
    tsp00_Int4 lastExitCode = 0;

    eo43CloseCommand( lastExitCode );
}

tsp00_Int4
ceo43_CmdPipe::eo43NextLine(
TCHAR           *buffer,
const tsp00_Int4 sizeInBytes,
bool           & NoNewline,
tsp00_Int4     & ExitCode)
{
    int returnedSize = 0;
    
    if ( m_CommandPipe_eo43
        && sizeInBytes > 0 )
    {
        if ( FGETS_CALL ( buffer, sizeInBytes - 1, m_CommandPipe_eo43 ) != NULL )
        {
            returnedSize = (int)strlen ( buffer ); /* point to newline */
            NoNewline = ( buffer[returnedSize - 1] != '\n' );
        }
    }
    buffer[returnedSize] = 0;
    if ( returnedSize == 0 )
    {
        eo43CloseCommand( ExitCode );
    }
    return returnedSize;
}

ceo43_xshow::ceo43_xshow( const TCHAR *showCommand, tsp00_DbNamec dbName, tsp00_Bool & openOk )
{
    size_t binpathLen;
    size_t execLen;
    size_t cmdLen;
    size_t dbnameLen;
    size_t argLen;
    TCHAR *localCmd;
    tsp00_Pathc    binPath;
    tsp01_RteError rteError;

    openOk = 0;

    if (!sqlGetDbrootPgmPath(binPath, true, &rteError)) {
      return;
    } // end if

    binpathLen = _tcslen(binPath);
    execLen = _tcslen(X_CONS_EXE);
    dbnameLen  = _tcslen(dbName);
    cmdLen  = _tcslen(SHOW_CMD);

    if ( showCommand )
    {
        argLen = _tcslen(showCommand);
    }
    
    localCmd = new TCHAR[ binpathLen + execLen + dbnameLen + cmdLen + argLen + 1];
    
    if ( localCmd )
    {
        TCHAR *pCmd = localCmd;
        memcpy(pCmd, (const char *) binPath, binpathLen * sizeof(TCHAR));
        pCmd += binpathLen;
        memcpy(pCmd, X_CONS_EXE, execLen * sizeof(TCHAR));
        pCmd += execLen;
        memcpy(pCmd, (const char *) dbName, dbnameLen * sizeof(TCHAR));
        pCmd += dbnameLen;
        memcpy(pCmd, SHOW_CMD, cmdLen * sizeof(TCHAR));
        pCmd += cmdLen;
        memcpy(pCmd, showCommand, argLen * sizeof(TCHAR) );
        pCmd += argLen;
        *pCmd = 0;

        eo43OpenCommand(localCmd, openOk);

        delete [] localCmd;
    }
}

/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

#ifdef TEST

#include <stdlib.h>
/*#include <unistd.h>*/
#include <iostream.h>

extern "C" void sp36_BuildSet (
                 unsigned char * data,
                 int             maxVal,
                 int             argc,
                 va_list         argptr)
{
    return;
}


void Usage(char *msg)
{
    if ( msg )
    {
        cerr << msg << endl;
    }
    cerr << "Syntax: show_api dbname xcons_show_cmd" << endl;
    cerr << "Usage: Execute an x_cons show command and display result" << endl;
    cerr << "Options: none" << endl;
    exit(1);
}

int
main(int argc, char *argv[])
{
#define BUFFER_SIZE_IN_BYTES 80
    char buffer[BUFFER_SIZE_IN_BYTES];
    int bufferSize = BUFFER_SIZE_IN_BYTES;

    bool       NoNewline;
    int        lineCount;
    tsp00_Int4 bytes;
    tsp00_DbNamec dbName;

    if ( argc < 3 )
    {
        Usage("Missing command");
    }

    if ( argv[1][0] == '-' )
    {
        switch ( argv[1][1] )
        {
        case '?':
        case 'h':
        case 'H':
            Usage(NULL);
            break;
        default:
            break;
        }
    }

    dbName.rawAssign(argv[1]);

    tsp00_Bool openOk;
    tsp00_Int4 ExitCode;

    ceo43_xshow showCommand(argv[2], dbName, openOk);

    if ( !openOk )
    {
        cerr << "showCommand " << argv[2] << " failed" << endl;
    }

    NoNewline = 0;
    lineCount = 0;

    do
    {
        if ( !NoNewline )
        {
            lineCount++;
            bytes = showCommand.eo43NextLine(buffer, bufferSize, NoNewline, ExitCode);
            if ( bytes > 0 )
            {
                cout << "Line " << lineCount << ": " << buffer;
            }
        }
        else
        {
            bytes = showCommand.eo43NextLine(buffer, bufferSize, NoNewline, ExitCode);
            if ( bytes > 0 )
            {
                cout << buffer;
            }
        }
    } while ( bytes != 0 );

    cout << "x_cons terminated. Exitcode :" << (int)ExitCode << endl;

    return 0;
}
#endif /* TEST */

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/
