/****************************************************************************

  module      : veo670.c

  -------------------------------------------------------------------------

  responsible : JoergM

  special area: Stack Back Trace
  description : Interface that allows to produce a stack backtrace


  last changed: 2001-01-18  14:57

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 2000



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

*****************************************************************************/



/*===========================================================================*
 *  INCLUDES                                                                 *
 *===========================================================================*/

/* PTS 1108958 */
#undef _FILE_OFFSET_BITS 
#include "heo670.h"
#include "gsp00.h"

#if defined(USE_MSG_PRIO7)
#include "heo01.h" /* nocheck */
#endif

/*===========================================================================*
 *  DEFINES                                                                  *
 *===========================================================================*/

#include <stdio.h>

#ifdef HPUX
extern void U_STACK_TRACE(void);
#endif /* HPUX */

#ifdef SUN
static void SunTraceStack(void);
#endif /* SUN */

#ifdef WIN32
static void Win32TraceStack(void);
#endif /* WIN32 */

#ifdef AIX
static void AixTraceStack(void);
#endif /*AIX */

#ifdef OSF1
static void OsfTraceStack(void);
#endif /* OSF1 */

#ifdef LINUX
static void LinuxTraceStack(void);
#endif


/*===========================================================================*
 *  MACROS                                                                   *
 *===========================================================================*/



/*===========================================================================*
 *  LOCAL CLASSES, STRUCTURES, TYPES, UNIONS ...                             *
 *===========================================================================*/



/*===========================================================================*
 *  STATIC/INLINE FUNCTIONS (PROTOTYPES)                                     *
 *===========================================================================*/



/*===========================================================================*
 *  METHODS                                                                  *
 *===========================================================================*/


static volatile int inTraceStack = 0;

/*
 * Produce a Stackbacktrace on Stderr
 */
void eo670_CTraceStack(void)
{
    /* not very secure, but reduces chance of crash... */
    if ( inTraceStack )
        return;

    inTraceStack = 1;

#ifdef HPUX
    fflush(stderr);
    U_STACK_TRACE();
#endif /* HPUX */

#ifdef SUN
    SunTraceStack();
#endif /* SUN */

#ifdef WIN32
    Win32TraceStack();
#endif /* WIN32 */

#ifdef AIX
    AixTraceStack();
#endif /* AIX */

#ifdef OSF1
    OsfTraceStack();
#endif /* OSF1 */

#ifdef LINUX
    LinuxTraceStack();
#endif

    inTraceStack = 0;
}

/*===========================================================================*
 *  LOCAL/PRIVATE/PROTECTED FUNCTIONS (CODE)                                 *
 *===========================================================================*/

/*---------------------------------------------------------------------------*/

#if defined(SUN) || defined(LINUX)
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <string.h>
#ifdef LINUX
#include <elf.h>
#else
#include <libelf.h>
#include <sys/elf.h>
#include <sys/frame.h>
#include <ucontext.h>
#include <sys/frame.h>
#ifdef BIT64
#include <sys/stack.h>
#endif /* BIT64 */
#endif /* !LINUX */
#include <sys/file.h>
#ifdef LINUX
/* PTS 1106738 */
/* With RedHat6.2 the mmap() prototype was fixed. This prevented the kernel from */
/* starting. The message given was 'invalid argument'. This Message was given due */
/* to the last argument of mmap() which was defined as 'off_t' and later 'off64_t' */
/* So on RedHat6.2 the last argument was expected to be a 64bit value, therefore */
/* the extra 32bits are collected from the stack. Such an offset is not PAGESIZE */
/* aligned, so it leads to 'invalid argument'... */
#undef __USE_FILE_OFFSET64
#define __USE_LARGEFILE64
#endif /* LINUX */
#include <sys/mman.h>
#include <sys/signal.h>
#include <sys/procfs.h>
#include <sys/param.h>
#include <sys/stat.h>
/*LINUX+SUN*/
#define NOHEAP_MALLOC noheap_malloc
#define NOHEAP_CALLOC noheap_calloc
#define NOHEAP_FREE noheap_free
#define USRTEXT  0x2000
#ifdef BIT64
/*LINUX+SUN*/typedef Elf64_Sym  Elf_Symbol;
/*LINUX+SUN*/typedef Elf64_Ehdr Elf_ElfHeader;
/*LINUX+SUN*/typedef Elf64_Shdr Elf_SectionHeader;
/*LINUX+SUN*/typedef Elf64_Word Elf_Word;
/*LINUX+SUN*/typedef Elf64_Addr Elf_Addr;
#define ELF_ST_TYPE(_i) ELF64_ST_TYPE(_i)
#else
/*LINUX+SUN*/typedef Elf32_Sym  Elf_Symbol;
/*LINUX+SUN*/typedef Elf32_Ehdr Elf_ElfHeader;
/*LINUX+SUN*/typedef Elf32_Shdr Elf_SectionHeader;
/*LINUX+SUN*/typedef Elf32_Word Elf_Word;
/*LINUX+SUN*/typedef Elf32_Addr Elf_Addr;
#define ELF_ST_TYPE(_i) ELF32_ST_TYPE(_i)
#endif
/*LINUX+SUN*/
/*LINUX+SUN*/typedef struct t_Symbol {
/*LINUX+SUN*/    char *        st_name;
/*LINUX+SUN*/    Elf_Addr      st_value;
/*LINUX+SUN*/    unsigned char st_info;
/*LINUX+SUN*/    unsigned char padding[sizeof(void*)-1];
/*LINUX+SUN*/} Symbol_t;
/*LINUX+SUN*/
/*LINUX+SUN*/typedef struct t_map {
/*LINUX+SUN*/    size_t          nsyms;    /* Number of symbols                */
/*LINUX+SUN*/    char    	   *strtab;   /* String Table                     */
/*LINUX+SUN*/    Symbol_t       *symtab;   /* Symbol Table                     */
/*LINUX+SUN*/    Elf_SectionHeader     *secthdr;
/*LINUX+SUN*/    Elf_SectionHeader     *sym_sh;
/*LINUX+SUN*/    Elf_SectionHeader     *str_sh;
/*LINUX+SUN*/    Elf_SectionHeader     *shent;
/*LINUX+SUN*/    Elf_Word      symsect;
/*LINUX+SUN*/    Elf_Word      strsect;
/*LINUX+SUN*/    char         *filename; /* ELF file name                        */
/*LINUX+SUN*/    int		    fd_proc;
/*LINUX+SUN*/    caddr_t         vaddr;
/*LINUX+SUN*/    struct t_map   *next;    /* Ptr to other shared libraries       */
/*LINUX+SUN*/} map_t;
/*LINUX+SUN*/
/*LINUX+SUN*/
/*LINUX+SUN*//* name to return if symbol name is not found                       */
/*LINUX+SUN*/static char no_symbol_name[9] = "????????";
/*LINUX+SUN*/
#define USRTEXT  0x2000
#define PROC_MMAP_FILE_OFFSET 49
#define PROC_MMAP_ACCESS_OFFSET 18
#define PROC_MMAP_VADDR_OFFSET 0
/*LINUX+SUN*/
/*LINUX+SUN*//*** Local functions ***/
/*LINUX+SUN*/
/*LINUX+SUN*/static map_t     *tb_proc_init(pid_t);
/*LINUX+SUN*/static Symbol_t  *tb_get_func(unsigned long addr, map_t *);
/*LINUX+SUN*/static char      *tb_get_filename(map_t *);
/*LINUX+SUN*/
/*LINUX+SUN*/static Symbol_t  *process_syms(Elf_Symbol *, size_t, 
/*LINUX+SUN*/                               size_t *, caddr_t, char *, char **);
/*LINUX+SUN*/static map_t     *get_symtab(int, caddr_t);
/*LINUX+SUN*/static int        compare(Symbol_t *, Symbol_t *);
/*LINUX+SUN*/static void       freemap(map_t *);
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/
/*LINUX+SUN*/static void write_string(const char *str)
/*LINUX+SUN*/{
/*LINUX+SUN*/    write(2, str, strlen(str));
/*LINUX+SUN*/}
/*LINUX+SUN+SUN*/
/*LINUX+SUN+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/
/*LINUX+SUN*/static void *
/*LINUX+SUN*/noheap_malloc(size_t size)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static int fd_dev_zero;
/*LINUX+SUN*/  static caddr_t ra;
/*LINUX+SUN*/
/*LINUX+SUN*/    if ((fd_dev_zero = open("/dev/zero", O_RDWR, 0)) == -1) {
/*LINUX+SUN*/        perror("opening /dev/zero");
/*LINUX+SUN*/        return (NULL);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (size == 0) size = sizeof (int);
/*LINUX+SUN*/
/*LINUX+SUN*/    if ((ra = (caddr_t)mmap((caddr_t) 0, size + sizeof(caddr_t), 
/*LINUX+SUN*/                   (PROT_READ | PROT_WRITE), MAP_PRIVATE,
/*LINUX+SUN*/                   fd_dev_zero, 0)) == (caddr_t)-1) {
/*LINUX+SUN*/        perror("mmap'ing /dev/zero");
/*LINUX+SUN*/        ra = NULL;
/*LINUX+SUN*/    } else {
/*LINUX+SUN*/        *((size_t *)ra) = size;
/*LINUX+SUN*/        ra += sizeof(size_t);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    close(fd_dev_zero);
/*LINUX+SUN*/    return((void *)ra);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*/static void *
/*LINUX+SUN*/noheap_calloc(size_t nelem, size_t elsize)
/*LINUX+SUN*/{
/*LINUX+SUN*/    /* While mmap of /dev/zero is being used, the block               */
/*LINUX+SUN*/    /* should come back already zero filled.                          */
/*LINUX+SUN*/    return(noheap_malloc((nelem * elsize)));
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/
/*LINUX+SUN*/static void
/*LINUX+SUN*/noheap_free(void *ptr)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static size_t len;
/*LINUX+SUN*/  static caddr_t real_ptr;
/*LINUX+SUN*/
/*LINUX+SUN*/    if (ptr == NULL) return;
/*LINUX+SUN*/
/*LINUX+SUN*/    real_ptr = (caddr_t)(((size_t *)ptr) - 1);
/*LINUX+SUN*/    len = *((size_t *)real_ptr) + sizeof(caddr_t);
/*LINUX+SUN*/    if (munmap(real_ptr, len))
/*LINUX+SUN*/        perror("munmap");
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
#ifdef LINUX
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/static void PrepareProcMap(char *procMap, char *endProcMap)
/*LINUX*/{
/*LINUX*/  static char *p;
/*LINUX*/
/*LINUX*/    p = procMap;
/*LINUX*/    while ( p < endProcMap ) {
/*LINUX*/        if ( *p == '\n' ) *p = 0;
/*LINUX*/        ++p;
/*LINUX*/    }
/*LINUX*/}
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/static char *NextProcMap(char *procMap, size_t *pBufferedChars, int proc_fd)
/*LINUX*/{
/*LINUX*/  static char *p;
/*LINUX*/  static size_t bytesStored;
/*LINUX*/  static int    bytesRead;
/*LINUX*/
/*LINUX*/    bytesStored = *pBufferedChars;
/*LINUX*/
/*LINUX*/    p = procMap + strlen(procMap);
/*LINUX*/    while ( *p == 0 && p < (procMap + bytesStored) ) {
/*LINUX*/        ++p;
/*LINUX*/    }
/*LINUX*/    bytesStored -= (p-procMap);
/*LINUX*/
/*LINUX*/    memcpy(procMap, p, bytesStored);
/*LINUX*/    bytesRead = read(proc_fd, procMap + bytesStored, 1024 - bytesStored );
/*LINUX*/    if ( bytesRead < 0 )
/*LINUX*/    {
/*LINUX*/        write_string("cant read from /proc file\n");
/*LINUX*/        return (NULL);
/*LINUX*/    }
/*LINUX*/    PrepareProcMap(procMap + bytesStored, procMap + bytesStored + bytesRead);
/*LINUX*/
/*LINUX*/    *pBufferedChars = bytesStored + bytesRead;
/*LINUX*/
/*LINUX*/    return ( *pBufferedChars > 0 ? procMap : NULL );
/*LINUX*/}
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/static map_t *
/*LINUX*/tb_proc_init(pid_t pid)
/*LINUX*/{
/*LINUX*/  static int       fd_proc;	/* /proc fd                           */
/*LINUX*/  static char      buff[64]; /* buffer for /proc file name */
/*LINUX*/
/*LINUX*/  static map_t    *map;      /* Main executable map                */
/*LINUX*/  static map_t    *pmap;	/* Executable map ptr                 */
/*LINUX*/
/*LINUX*/  static char     *procMap;  /* Pointer to procMap buffer */
/*LINUX*/  static size_t    bufferedChars; /* Number of currently buffered character */
/*LINUX*/  static caddr_t   vaddr;
/*LINUX*/
/*LINUX*/    sprintf(buff, "/proc/%05d/maps", getpid());
/*LINUX*/
/*LINUX*/    if ((fd_proc = open(buff, O_RDONLY)) == -1){
/*LINUX*/        write_string("cant open /proc file: ");
/*LINUX*/        write_string(buff);
/*LINUX*/        return (NULL);
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    procMap = (char *)alloca(1024);
/*LINUX*/
/*LINUX*/    if ( (bufferedChars = read(fd_proc, procMap, 1024)) <= 0 )
/*LINUX*/    {
/*LINUX*/        write_string("cant read /proc file: ");
/*LINUX*/        write_string(buff);
/*LINUX*/        return (NULL);
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    PrepareProcMap(procMap, procMap + bufferedChars);
/*LINUX*/    pmap  = map = NULL;
/*LINUX*/    do {
/*LINUX*/        int  duplicate;
/*LINUX*/        map_t  *currmap;
/*LINUX*/        map_t  *pmp;
/*LINUX*/        int  fd_obj;
/*LINUX*/
/*LINUX*/
/*LINUX*/        if ( strlen(procMap) < PROC_MMAP_FILE_OFFSET )
/*LINUX*/            continue;
/*LINUX*/
/*LINUX*/        if ( strncmp(procMap+PROC_MMAP_ACCESS_OFFSET, "r-xp", 4) != 0 )
/*LINUX*/            continue;
/*LINUX*/
/*LINUX*/        vaddr = (caddr_t)
/*LINUX*/                strtoul(procMap+PROC_MMAP_VADDR_OFFSET, NULL, 16);
/*LINUX*/
/*LINUX*/        if ((fd_obj = open(procMap+PROC_MMAP_FILE_OFFSET, O_RDONLY)) == -1)
/*LINUX*/            continue;
/*LINUX*/
/*LINUX*/        if((currmap = get_symtab(fd_obj, vaddr)) == NULL) {
/*LINUX*/            close(fd_obj);
/*LINUX*/            continue;
/*LINUX*/        }
/*LINUX*/
/*LINUX*/        close(fd_obj);
/*LINUX*/
/*LINUX*/        if (map == NULL) {
/*LINUX*/            map = pmap = currmap;
/*LINUX*/            pmap->vaddr = vaddr;
/*LINUX*/        } else {
/*LINUX*/            pmap->next = currmap;
/*LINUX*/            pmap = currmap;
/*LINUX*/            pmap->vaddr = vaddr;
/*LINUX*/        }
/*LINUX*/    } while ( NextProcMap(procMap, &bufferedChars, fd_proc) != NULL );
/*LINUX*/
/*LINUX*/    close(fd_proc);
/*LINUX*/
/*LINUX*/    return(map);
/*LINUX*/}
/*LINUX*/
/*LINUX*//*---------------------------------------------------------------------------*/
/*LINUX*/
/*LINUX*/struct frame {
/*LINUX*/    struct frame *fr_savfp;
/*LINUX*/    unsigned long fr_savpc;
/*LINUX*/    unsigned long fr_arg[4];
/*LINUX*/};
/*LINUX*/
/*LINUX*/static void LinuxTraceStack()
/*LINUX*/{
/*LINUX*/  static Symbol_t       *funcsym;
/*LINUX*/  static map_t          *map;
/*LINUX*/  static map_t          *newmap;
/*LINUX*/  static char           *funcname;
/*LINUX*/  static int             depth;
/*LINUX*/  static unsigned long   pc;
/*LINUX*/  static struct frame   *fp;
/*LINUX*/  static char           *MessageLine;
/*LINUX*/
/*LINUX*/    depth = 0;
/*LINUX*/    MessageLine = (char *)NOHEAP_MALLOC(2048);
/*LINUX*/    if ( !MessageLine ) return;
/*LINUX*/
/*LINUX*/    pc = (unsigned long)LinuxTraceStack;
/*LINUX*/    asm( "movl %%ebp,%0" : "=r" (fp) : /* no input operand */ );
/*LINUX*/
/*LINUX*/    if (!(map = tb_proc_init(getuid()))) {
/*LINUX*/        return;
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    while ( fp 
/*LINUX*/        && (fp != fp->fr_savfp)
/*LINUX*/	    && ((unsigned long)fp->fr_savfp >= USRTEXT) )
/*LINUX*/    {
/*LINUX*/        newmap = map;
/*LINUX*/        /* newmap is not NULL on the first pass                       */
/*LINUX*/        while(newmap && (funcsym = tb_get_func(pc, newmap)) == NULL)
/*LINUX*/            newmap = newmap->next;
/*LINUX*/
/*LINUX*/        funcname = (funcsym == NULL) ? no_symbol_name :
/*LINUX*/            (char *)funcsym->st_name;
/*LINUX*/
/*LINUX*/        sprintf (MessageLine,
/*LINUX*/                 "[%d] %s+0x%lx (0x%lx,0x%lx,0x%lx,0x%lx), at 0x%lx\n", 
/*LINUX*/                 depth, 
/*LINUX*/                 funcname, 
/*LINUX*/                 pc - funcsym->st_value,
/*LINUX*/                 fp->fr_arg[0],
/*LINUX*/                 fp->fr_arg[1],
/*LINUX*/                 fp->fr_arg[2],
/*LINUX*/                 fp->fr_arg[3],
/*LINUX*/                 pc);
/*LINUX*/        write_string(MessageLine);
/*LINUX*/        depth++;
/*LINUX*/
/*LINUX*/        /* filter out tail recursion                                 */
/*LINUX*/        /*   while (sp->fr_savpc == (sp->fr_savfp)->fr_savpc)        */
/*LINUX*/        /*   sp = fp->fr_savfp;                                      */
/*LINUX*/		
/*LINUX*/        pc = fp->fr_savpc;
/*LINUX*/        fp = fp->fr_savfp;
/*LINUX*/    }
/*LINUX*/
/*LINUX*/    NOHEAP_FREE(MessageLine);
/*LINUX*/
/*LINUX*/    while(map)
/*LINUX*/    {
/*LINUX*/        newmap=map->next;
/*LINUX*/        freemap(map);
/*LINUX*/        map=newmap;
/*LINUX*/    }
/*LINUX*/}
/*LINUX*/
#endif /*LINUX*/
#ifdef SUN
/*SUN*/
/*SUN*/static map_t *
/*SUN*/tb_proc_init(pid_t pid)
/*SUN*/{
/*SUN*/  static int       fd_proc;	/* /proc fd                           */
/*SUN*/  static map_t    *map;      /* Main executable map                */
/*SUN*/  static map_t    *pmap;	/* Executable map ptr                 */
/*SUN*/  static int       nmaps;	/* # of mappings.                     */
/*SUN*/  static prmap_t  *prmaps;
/*SUN*/  static char      buff[32];
/*SUN*/  static int       i;
/*SUN*/
/*SUN*/    sprintf(buff, "/proc/%05d", getpid());
/*SUN*/
/*SUN*/    if ((fd_proc = open(buff, O_RDONLY)) == -1){
/*SUN*/        write_string("cant open /proc file: ");
/*SUN*/        write_string(buff);
/*SUN*/        return (NULL);
/*SUN*/    }
/*SUN*/
/*SUN*/    /* Get address space (shared library) mappings                    */
/*SUN*/    if (ioctl(fd_proc, PIOCNMAP, &nmaps) == -1) {
/*SUN*/        write_string("Unable to get number of memory mappings.\n");
/*SUN*/        goto cleanup;
/*SUN*/    }
/*SUN*/
/*SUN*/    prmaps = (prmap_t *)NOHEAP_CALLOC(sizeof(prmap_t), nmaps+1);
/*SUN*/    if(prmaps == NULL) {
/*SUN*/        write_string("out of memory\n");
/*SUN*/        goto cleanup;
/*SUN*/    }
/*SUN*/
/*SUN*/    if (ioctl(fd_proc, PIOCMAP, prmaps) == -1) {
/*SUN*/        write_string("Unable to get memory mappings.\n");
/*SUN*/        goto cleanup;
/*SUN*/    }
/*SUN*/
/*SUN*/    pmap  = map = NULL;
/*SUN*/    for(i=0;i<nmaps;i++) {
/*SUN*/        static int  duplicate;
/*SUN*/        static prmap_t  *pm;
/*SUN*/        static map_t  *currmap;
/*SUN*/        static map_t  *pmp;
/*SUN*/        static int  fd_obj;
/*SUN*/
/*SUN*/        pm = prmaps+i;
/*SUN*/        if (pm->pr_mflags != (MA_READ | MA_EXEC))
/*SUN*/            continue;
/*SUN*/
/*SUN*/        fd_obj = ioctl(fd_proc, PIOCOPENM, &pm->pr_vaddr);
/*SUN*/        if(fd_obj == -1) {
/*SUN*/            continue;
/*SUN*/        }
/*SUN*/        if((currmap = get_symtab(fd_obj, pm->pr_vaddr)) == NULL) {
/*SUN*/            close(fd_obj);
/*SUN*/            continue;
/*SUN*/        }
/*SUN*/        close(fd_obj);
/*SUN*/
/*SUN*/        if (map == NULL) {
/*SUN*/            map = pmap = currmap;
/*SUN*/            pmap->vaddr = pm->pr_vaddr;
/*SUN*/        } else {
/*SUN*/            pmap->next = currmap;
/*SUN*/            pmap = currmap;
/*SUN*/            pmap->vaddr = pm->pr_vaddr;
/*SUN*/        }
/*SUN*/    }
/*SUN*/    (void) close(fd_proc);
/*SUN*/    NOHEAP_FREE(prmaps);
/*SUN*/    return(map);
/*SUN*/ cleanup:
/*SUN*/    (void) close(fd_proc);
/*SUN*/    NOHEAP_FREE(map);
/*SUN*/    NOHEAP_FREE(prmaps);
/*SUN*/    return(NULL);
/*SUN*/}
/*SUN*/
#ifdef BIT64
#define SP_BIAS V9BIAS64
#else
#define SP_BIAS 0
#endif
/*SUN*/
/*SUN*//*---------------------------------------------------------------------------*/
/*SUN*/static struct frame *BiasedFramePtr(void *fp)
/*SUN*/{
/*SUN*/    return (struct frame *)(((unsigned long)fp)+SP_BIAS);
/*SUN*/}
/*SUN*/
/*SUN*//*---------------------------------------------------------------------------*/
/*SUN*/
/*SUN*/static void SunTraceStack()
/*SUN*/{
/*SUN*/  static Symbol_t       *funcsym;
/*SUN*/  static map_t          *map;
/*SUN*/  static map_t          *newmap;
/*SUN*/  static char           *funcname;
/*SUN*/  static int             depth;
/*SUN*/  static unsigned long   pc;
/*SUN*/  static struct frame   *fp;
/*SUN*/  static ucontext_t     *pContext;
/*SUN*/  static char           *MessageLine;
/*SUN*/
/*SUN*/    depth = 0;
/*SUN*/    MessageLine = (char *)NOHEAP_MALLOC(2048);
/*SUN*/    if ( !MessageLine ) return;
/*SUN*/    pContext = (ucontext_t *)NOHEAP_MALLOC(sizeof(ucontext_t));
/*SUN*/    if ( !pContext ) { NOHEAP_FREE(MessageLine); return; }
/*SUN*/    getcontext (pContext);
/*SUN*/
/*SUN*/    pc = pContext->uc_mcontext.gregs[REG_PC]; /* program counter           */
/*SUN*/    fp = BiasedFramePtr((void *)pContext->uc_mcontext.gregs[REG_SP]); /* stack pointer*/
/*SUN*/
/*SUN*/    if (!(map = tb_proc_init(getuid()))) {
/*SUN*/        return;
/*SUN*/    }
/*SUN*/
/*SUN*/    while ( (((long)pc)&3) == 0
/*SUN*/        &&  fp 
/*SUN*/        && (fp != BiasedFramePtr(fp->fr_savfp))
/*SUN*/	    && ((((unsigned long) BiasedFramePtr(fp->fr_savfp)) + SP_BIAS) >= USRTEXT) )
/*SUN*//*	 && ((unsigned long) (fp->fr_savfp) <= USRSTACK)) */
/*SUN*/    {
/*SUN*/        newmap = map;
/*SUN*/        /* newmap is not NULL on the first pass                       */
/*SUN*/        while(newmap && (funcsym = tb_get_func(pc, newmap)) == NULL)
/*SUN*/            newmap = newmap->next;
/*SUN*/
/*SUN*/        funcname = (funcsym == NULL) ? no_symbol_name :
/*SUN*/            (char *)funcsym->st_name;
/*SUN*/
/*SUN*/        sprintf (MessageLine,
/*SUN*/                 "[%d] %s+0x%lx ( 0x%lx, 0x%lx, 0x%lx, 0x%lx, 0x%lx, 0x%lx ), at 0x%lx\n", 
/*SUN*/                 depth, 
/*SUN*/                 funcname, 
/*SUN*/                 pc - funcsym->st_value,
/*SUN*/                 fp->fr_arg[0],
/*SUN*/                 fp->fr_arg[1],
/*SUN*/                 fp->fr_arg[2],
/*SUN*/                 fp->fr_arg[3],
/*SUN*/                 fp->fr_arg[4],
/*SUN*/                 fp->fr_arg[5],
/*SUN*/                 pc);
/*SUN*/        write_string(MessageLine);
/*SUN*/        depth++;
/*SUN*/
/*SUN*/        /* filter out tail recursion                         */                           
/*SUN*/        /* while (sp->fr_savpc == (sp->fr_savfp)->fr_savpc)  */
/*SUN*/        /* sp = fp->fr_savfp; */ 
/*SUN*/		
/*SUN*/        pc = fp->fr_savpc;
/*SUN*/        fp = BiasedFramePtr(fp->fr_savfp);
/*SUN*/
/*SUN*/    }
/*SUN*/
/*SUN*/    NOHEAP_FREE(MessageLine);
/*SUN*/
/*SUN*/    while(map)
/*SUN*/    {
/*SUN*/        newmap=map->next;
/*SUN*/        freemap(map);
/*SUN*/        map=newmap;
/*SUN*/    }
/*SUN*/}
#endif /* SUN */
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*//*                                                                           */
/*LINUX+SUN*//* tb_get_func() takes an address and dereferences it to give the            */
/*LINUX+SUN*//* function name which starts at the closest lower address. If no            */
/*LINUX+SUN*//* such function exists, it passes an empty string.                          */
/*LINUX+SUN*//* Relies on the qsorted symtab list described above.                        */
/*LINUX+SUN*//* Basically a binary search algorithm.                                      */
/*LINUX+SUN*/
/*LINUX+SUN*/static Symbol_t    *
/*LINUX+SUN*/tb_get_func(unsigned long addr, map_t *map)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static  int         left;
/*LINUX+SUN*/  static  int         right;
/*LINUX+SUN*/  static  int         middle, sclass;
/*LINUX+SUN*/  static  Symbol_t   *symtab;
/*LINUX+SUN*/  static  int i;
/*LINUX+SUN*/
/*LINUX+SUN*/    left = 0;
/*LINUX+SUN*/    right = map->nsyms - 1;
/*LINUX+SUN*/    symtab = map->symtab;
/*LINUX+SUN*/    i = 0;
/*LINUX+SUN*/    if ((symtab->st_value > addr) ||
/*LINUX+SUN*/        ((symtab + right)->st_value < addr)) {
/*LINUX+SUN*/        return (NULL);  /* should be invalid name                     */
/*LINUX+SUN*/    }
/*LINUX+SUN*/    while (left <= right) {  /* Binary Search (sort of)               */
/*LINUX+SUN*/        if ((right-left) <= 1)
/*LINUX+SUN*/            return(symtab + left);
/*LINUX+SUN*/        middle = (left + right) / 2;
/*LINUX+SUN*/        if ((symtab + middle)->st_value <= addr) {
/*LINUX+SUN*/            left = middle;
/*LINUX+SUN*/        } else {
/*LINUX+SUN*/            right = middle;
/*LINUX+SUN*/        }
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* should never get here                                          */
/*LINUX+SUN*/    return (NULL);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/typedef int (*CompareFunc) (const void *, const void *);
/*LINUX+SUN*/
/*LINUX+SUN*/static map_t *
/*LINUX+SUN*/get_symtab(int fd, caddr_t voffset)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static map_t         *map;
/*LINUX+SUN*/  static  Elf_ElfHeader    elfhdr;    /* Elf header                            */
/*LINUX+SUN*/  static  char        *sname = NULL;
/*LINUX+SUN*/  static  Elf_Symbol     *sp;
/*LINUX+SUN*/  static  int        i;
/*LINUX+SUN*/  static  size_t        syms;
/*LINUX+SUN*/  static  Elf_Symbol     *symtab;    /* Unsorted file image symbol table     */
/*LINUX+SUN*/
/*LINUX+SUN*/
/*LINUX+SUN*/    /* Allocate a map structure                                       */
/*LINUX+SUN*/
/*LINUX+SUN*/    map = (map_t *)NOHEAP_CALLOC(sizeof(map_t), 1);
/*LINUX+SUN*/    if(map == 0) {
/*LINUX+SUN*/        write_string("Out of memory\n");
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* Get the ELF header                                             */
/*LINUX+SUN*/
/*LINUX+SUN*/    if (read (fd, &elfhdr, sizeof(elfhdr)) != sizeof(elfhdr)){
/*LINUX+SUN*/        write_string("Unable to read ELF header \n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* Get the section header table                                   */
/*LINUX+SUN*/
/*LINUX+SUN*/    if (elfhdr.e_shnum == 0) {
/*LINUX+SUN*/        write_string("No section header entries.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (lseek(fd, (long) elfhdr.e_shoff, L_SET) != elfhdr.e_shoff) {
/*LINUX+SUN*/        write_string("Unable to seek to section header.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    if ((map->secthdr = (Elf_SectionHeader *)
/*LINUX+SUN*/         NOHEAP_MALLOC(elfhdr.e_shentsize*elfhdr.e_shnum)) == NULL) {
/*LINUX+SUN*/        write_string("Unable to allocate section header.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (read(fd, map->secthdr, elfhdr.e_shentsize * elfhdr.e_shnum)
/*LINUX+SUN*/        !=  (elfhdr.e_shentsize * elfhdr.e_shnum)) {
/*LINUX+SUN*/        write_string("Unable to read ELF header section.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* If there is a section for section names, allocate a  */
/*LINUX+SUN*/    /* buffer to read it and seek to its offset in the file.*/
/*LINUX+SUN*/          
/*LINUX+SUN*/          if ((elfhdr.e_shstrndx != SHN_UNDEF) && 
/*LINUX+SUN*/              ((sname = (char *)
/*LINUX+SUN*/                NOHEAP_MALLOC(map->secthdr[elfhdr.e_shstrndx].sh_size))
/*LINUX+SUN*/               != NULL) &&
/*LINUX+SUN*/              (lseek(fd, (long) map->secthdr[elfhdr.e_shstrndx].sh_offset,
/*LINUX+SUN*/                     L_SET) == map->secthdr[elfhdr.e_shstrndx].sh_offset)) {
/*LINUX+SUN*/
/*LINUX+SUN*/              /* Read the section names.  If the read fails, make      */
/*LINUX+SUN*/              /* it look as though none of this happened               */
/*LINUX+SUN*/                    if (read(fd, sname, 
/*LINUX+SUN*/                             map->secthdr[elfhdr.e_shstrndx].sh_size) !=
/*LINUX+SUN*/                        map->secthdr[elfhdr.e_shstrndx].sh_size) {
/*LINUX+SUN*/			(void) NOHEAP_FREE(sname);
/*LINUX+SUN*/			sname = NULL;
/*LINUX+SUN*/                    }
/*LINUX+SUN*/          }
/*LINUX+SUN*/
/*LINUX+SUN*/
/*LINUX+SUN*/    map->shent  = map->secthdr;    /* set to first entry              */
/*LINUX+SUN*/    for (i = 0; i < (int) elfhdr.e_shnum; i++, map->shent++ ){
/*LINUX+SUN*/        /* go thru each section header                                */
/*LINUX+SUN*/        if (map->shent->sh_type ==  SHT_SYMTAB) {
/*LINUX+SUN*/            map->symsect = i;
/*LINUX+SUN*/            map->strsect = map->shent->sh_link;
/*LINUX+SUN*/        }
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    if (sname) NOHEAP_FREE(sname);
/*LINUX+SUN*/
/*LINUX+SUN*/    if (map->symsect == 0) {
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    map->sym_sh = &map->secthdr[map->symsect];
/*LINUX+SUN*/    syms = map->sym_sh->sh_size/map->sym_sh->sh_entsize;
/*LINUX+SUN*/    if(map->sym_sh->sh_link == elfhdr.e_shnum ||
/*LINUX+SUN*/       map->secthdr[map->sym_sh->sh_link].sh_size == 0) {
/*LINUX+SUN*/        write_string("Warning - empty string table; no symbols.\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    map->str_sh = &map->secthdr[map->sym_sh->sh_link];
/*LINUX+SUN*/    if ((map->strtab = (char *)
/*LINUX+SUN*/         NOHEAP_MALLOC(map->str_sh->sh_size +
/*LINUX+SUN*/                       sizeof(map->str_sh->sh_size)+8)) == NULL) {
/*LINUX+SUN*/        write_string("Warning: Couldnt allocate space for strtab\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    *(Elf_Word *)map->strtab = map->str_sh->sh_size + 
/*LINUX+SUN*/        sizeof(map->str_sh->sh_size);
/*LINUX+SUN*/    lseek(fd, (long)map->str_sh->sh_offset, L_SET);
/*LINUX+SUN*/    if(read(fd, map->strtab + sizeof(map->str_sh->sh_size), 
/*LINUX+SUN*/	    map->str_sh->sh_size) != map->str_sh->sh_size) {
/*LINUX+SUN*/        write_string("Warning - error reading string table\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    lseek(fd, (long)map->sym_sh->sh_offset, L_SET);
/*LINUX+SUN*/    if ((symtab = (Elf_Symbol *) 
/*LINUX+SUN*/         NOHEAP_MALLOC(syms*sizeof(Elf_Symbol))) == NULL) {
/*LINUX+SUN*/        write_string("Warning: Couldnt allocate space for symtab\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/    if (read(fd, symtab, syms*sizeof(Elf_Symbol)) != 
/*LINUX+SUN*/        syms*sizeof(Elf_Symbol)){
/*LINUX+SUN*/        write_string("Warning - error reading symbl table\n");
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* process_syms() returns an array of 'interesting symbols'       */
/*LINUX+SUN*/    /* i.e. those with a non-NULL st_name field.                      */
/*LINUX+SUN*/
/*LINUX+SUN*/    if (elfhdr.e_type == ET_EXEC)
/*LINUX+SUN*/        voffset = 0;
/*LINUX+SUN*/    map->symtab = process_syms(symtab, syms, &map->nsyms, voffset, 
/*LINUX+SUN*/                                map->strtab + sizeof(map->str_sh->sh_size), &map->filename);
/*LINUX+SUN*/    NOHEAP_FREE(symtab);
/*LINUX+SUN*/
/*LINUX+SUN*/    if(map->symtab == NULL) {
/*LINUX+SUN*/        freemap(map);
/*LINUX+SUN*/        return(0);
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    qsort(map->symtab, map->nsyms, sizeof(Symbol_t), (CompareFunc)compare);
/*LINUX+SUN*/
/*LINUX+SUN*/    return(map);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*/
/*LINUX+SUN*/static void
/*LINUX+SUN*/freemap(map_t *map)
/*LINUX+SUN*/{
/*LINUX+SUN*/    if(!map)
/*LINUX+SUN*/        return;
/*LINUX+SUN*/    if(map->secthdr) NOHEAP_FREE(map->secthdr);
/*LINUX+SUN*/    if(map->strtab)  NOHEAP_FREE(map->strtab);
/*LINUX+SUN*/    if(map->symtab)  NOHEAP_FREE(map->symtab);
/*LINUX+SUN*/    NOHEAP_FREE(map);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*//*                                                                           */
/*LINUX+SUN*//* Translate symbol table data particularly for sorting.                     */
/*LINUX+SUN*//* Input is the symbol table data structure, number of symbols,              */
/*LINUX+SUN*//* opened ELF file, and the string table link offset.                        */
/*LINUX+SUN*/
/*LINUX+SUN*/static Symbol_t *
/*LINUX+SUN*/process_syms(Elf_Symbol *data, size_t num, size_t *nsyms,
/*LINUX+SUN*/             caddr_t offset, char* strtab, char **pFilename)
/*LINUX+SUN*/{
/*LINUX+SUN*/  static Symbol_t  *s;
/*LINUX+SUN*/  static Symbol_t  *buf;
/*LINUX+SUN*/  static int i;
/*LINUX+SUN*/  static int useful_sym;   /* Not all symbols read are useful          */
/*LINUX+SUN*/
/*LINUX+SUN*/    useful_sym = 0;
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/    /* This loop exists only to detect number of function names */
/*LINUX+SUN*/    /* so that space is not overallocated for it                */
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/    for (i = 0; i < num; i ++) {
/*LINUX+SUN*/		if (data[i].st_name && 
/*LINUX+SUN*/                    (ELF_ST_TYPE(data[i].st_info) == STT_FUNC))
/*LINUX+SUN*/        {
/*LINUX+SUN*/            useful_sym++;
/*LINUX+SUN*/        }
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    s = (Symbol_t *)NOHEAP_MALLOC(useful_sym*sizeof(Symbol_t));
/*LINUX+SUN*/    if(s == NULL) {
/*LINUX+SUN*/        write_string("Out of memory\n");
/*LINUX+SUN*/        return NULL;
/*LINUX+SUN*/    }
/*LINUX+SUN*/
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/    /* Skip the first entry. It contains a special STN_UNDEF */
/*LINUX+SUN*/    /* symbol                                                */
/*LINUX+SUN*/    /* */
/*LINUX+SUN*/
/*LINUX+SUN*/      for(i=1, buf = s; i<= num; i++, data++) {
/*LINUX+SUN*/          if ( data->st_name != (Elf_Word)0 ) {
/*LINUX+SUN*/              if ( ELF_ST_TYPE(data->st_info) == STT_FUNC ) {
/*LINUX+SUN*/                  buf->st_name  = strtab + data->st_name;
/*LINUX+SUN*/                  buf->st_value = data->st_value + (unsigned long) offset;
/*LINUX+SUN*/                  buf->st_info  = data->st_info;
/*LINUX+SUN*/                  buf ++ ;
/*LINUX+SUN*/              } else if ( ELF_ST_TYPE(data->st_info) == STT_FILE ) {
/*LINUX+SUN*/                 *pFilename = strtab + data->st_name;
/*LINUX+SUN*/              }
/*LINUX+SUN*/          }
/*LINUX+SUN*/      }
/*LINUX+SUN*/
/*LINUX+SUN*/    *nsyms = useful_sym;
/*LINUX+SUN*/    return(s);
/*LINUX+SUN*/}
/*LINUX+SUN*/
/*LINUX+SUN*/
/*LINUX+SUN*//*---------------------------------------------------------------------------*/
/*LINUX+SUN*//*                                                                           */
/*LINUX+SUN*//* Compare either by name or by value for sorting.                           */
/*LINUX+SUN*//* This is the comparison function called by qsort to                        */
/*LINUX+SUN*//* sort the symbols either by name or value when requested.                  */
/*LINUX+SUN*/
/*LINUX+SUN*/static int 
/*LINUX+SUN*/compare(Symbol_t *a, Symbol_t *b)
/*LINUX+SUN*/{
/*LINUX+SUN*/    if(a->st_value > b->st_value)
/*LINUX+SUN*/        return 1;
/*LINUX+SUN*/    else if (a->st_value < b->st_value)
/*LINUX+SUN*/        return -1;
/*LINUX+SUN*/    else 
/*LINUX+SUN*/        return (((unsigned char)ELF_ST_TYPE(a->st_info) <
/*LINUX+SUN*/                 (unsigned char)ELF_ST_TYPE(b->st_info)) 
/*LINUX+SUN*/                ? 1 : 0);
/*LINUX+SUN*/
/*LINUX+SUN*/}
/*LINUX+SUN*/
#endif /* LINUX+SUN */

#ifdef OSF1
/*OSF1*/
#include <loader.h>
#include <stdlib.h>
#include <excpt.h>
#include <ldfcn.h>
#include <setjmp.h>
#include <unistd.h>
#include <errno.h>
/*OSF1*/
#define MAX_STACK     (256)
#define SKIP_TOP	(2)
#define SKIP_BOT	(1)
#define STACK_IM_ARSCH ((unsigned long)(-1))
/*OSF1*/
/*OSF1*/typedef	struct modul_chain_elem	*mce_p;
/*OSF1*/struct modul_chain_elem
/*OSF1*/{
/*OSF1*/    mce_p		next;
/*OSF1*/    ldr_module_info_t	info;
/*OSF1*/    elf32_LDFILE	*ldptr;
/*OSF1*/    int			opened;		/* offen oder zumindest versucht */
/*OSF1*/    int			symtab_ok;	
/*OSF1*/
/*OSF1*/};
/*OSF1*/typedef struct modul_chain_elem mce;
/*OSF1*/
/*OSF1*/typedef	struct region_chain_elem	*rce_p;
/*OSF1*/struct region_chain_elem
/*OSF1*/{
/*OSF1*/    rce_p		next;
/*OSF1*/    mce_p		modul;
/*OSF1*/    rce_p		next_pc;
/*OSF1*/    int			used;
/*OSF1*/    ldr_region_info_t	info;
/*OSF1*/};
/*OSF1*/typedef struct region_chain_elem rce;
/*OSF1*/
/*OSF1*/static mce_p	mc_anchor = NULL;
/*OSF1*/static mce_p	*mc_next  = NULL;
/*OSF1*/static rce_p	rc_anchor = NULL;
/*OSF1*/static rce_p	*rc_next  = NULL;
/*OSF1*/static rce_p	pc_anchor = NULL;
/*OSF1*/static rce_p	*pc_next  = NULL;
/*OSF1*/
/*OSF1*/static char		*routine;
/*OSF1*/static jmp_buf  emergency;
/*OSF1*/static unsigned long	stack[MAX_STACK];
/*OSF1*/static long		stack_depth;
/*OSF1*/
/*OSF1*/static void SigbusHndlr   (int i);
/*OSF1*/static void SigsegvHndlr  (int i);
/*OSF1*/static void BuildRegionChain(void);
/*OSF1*/static void WalkStack     (void);
/*OSF1*/static void StackTrace    ();
/*OSF1*/static void GetSyms	  ();
/*OSF1*/static long ipd_adr	  (LDFILE *ldptr, unsigned long adr);
/*OSF1*/
/*OSF1*/static void write_string(const char *buffer)
/*OSF1*/{
/*OSF1*/    write(2, buffer, strlen(buffer) );
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void OsfTraceStack(void)
/*OSF1*/{
/*OSF1*/  static void    (*old_sigbus) (int);
/*OSF1*/  static void    (*old_sigsegv) (int);
/*OSF1*/  static int	    rc;
/*OSF1*/
/*OSF1*/    rc = sigsetjmp ( emergency, 1 );
/*OSF1*/    if ( rc == 0 )
/*OSF1*/    {
/*OSF1*/    	routine = "<none>";
/*OSF1*/    	old_sigbus  = sigset ( SIGBUS,  SigbusHndlr );
/*OSF1*/	    old_sigsegv = sigset ( SIGSEGV, SigsegvHndlr);
/*OSF1*/	    StackTrace();
/*OSF1*/    }
/*OSF1*/    else
/*OSF1*/    {
/*OSF1*/    	write_string ( "*** Error in OSF_stacktrace: Signal in ");
/*OSF1*/        write_string ( routine );
/*OSF1*/        write_string ( "\n" );
/*OSF1*/    }
/*OSF1*/    sigset ( SIGBUS,  old_sigbus );
/*OSF1*/    sigset ( SIGSEGV, old_sigsegv);
/*OSF1*/    return;
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void SigbusHndlr  (int i)
/*OSF1*/{
/*OSF1*/    siglongjmp ( emergency, SIGBUS );
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void SigsegvHndlr (int i)
/*OSF1*/{
/*OSF1*/    siglongjmp ( emergency, SIGSEGV );
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void StackTrace()
/*OSF1*/{
/*OSF1*/    routine = "WalkStack";
/*OSF1*/    WalkStack();
/*OSF1*/    routine = "BuildRegionChain";
/*OSF1*/    BuildRegionChain();
/*OSF1*/    routine = "GetSyms";
/*OSF1*/    GetSyms();
/*OSF1*/}
/*OSF1*/
/*OSF1*/static void BuildRegionChain(void)
/*OSF1*/{
/*OSF1*/
/*OSF1*/  static int	    i;
/*OSF1*/  static int	    j;
/*OSF1*/  static int	    rc;
/*OSF1*/  static int	    no_regions;
/*OSF1*/  static mce_p   me_p;
/*OSF1*/  static rce_p   re_p;
/*OSF1*/  static rce_p   re_tmp;
/*OSF1*/  static void    *ptr;
/*OSF1*/
/*OSF1*/  static ldr_process_t	my_proc;
/*OSF1*/  static ldr_module_t	mod_id;
/*OSF1*/  static size_t		ret_size;
/*OSF1*/
/*OSF1*/    while ( mc_anchor )
/*OSF1*/    {
/*OSF1*/	    if ( mc_anchor->ldptr )
/*OSF1*/	    {
/*OSF1*/	        ldclose ( mc_anchor->ldptr );
/*OSF1*/	    }
/*OSF1*/	    me_p = mc_anchor->next;
/*OSF1*/	    free (mc_anchor);
/*OSF1*/	    mc_anchor = me_p;
/*OSF1*/    }
/*OSF1*/    mc_next  = &mc_anchor;
/*OSF1*/
/*OSF1*/    while ( rc_anchor )
/*OSF1*/    {
/*OSF1*/    	re_p = rc_anchor->next;
/*OSF1*/	    free (rc_anchor);
/*OSF1*/    	rc_anchor = re_p;
/*OSF1*/    }
/*OSF1*/    rc_next   = &rc_anchor;
/*OSF1*/    pc_anchor = NULL;
/*OSF1*/    pc_next   = &pc_anchor;
/*OSF1*/
/*OSF1*/    no_regions = 0;
/*OSF1*/    my_proc = ldr_my_process();
/*OSF1*/
/*OSF1*/    for ( mod_id = LDR_NULL_MODULE
/*OSF1*/	, i = 0
/*OSF1*/	, rc = 0
/*OSF1*/	; rc == 0
/*OSF1*/	; i++ )
/*OSF1*/    {
/*OSF1*/    	rc = ldr_next_module ( my_proc, &mod_id );
/*OSF1*/
/*OSF1*/        if ( rc )
/*OSF1*/    	{
/*OSF1*/	        perror("ldr_next_module");
/*OSF1*/    	    break;
/*OSF1*/	    }
/*OSF1*/    	if ( mod_id == LDR_NULL_MODULE)
/*OSF1*/	    {
/*OSF1*/    	    break;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/    	me_p = (mce_p)calloc ( 1, sizeof(mce) );
/*OSF1*/	    if ( me_p == NULL ) break;
/*OSF1*/
/*OSF1*/    	rc = ldr_inq_module ( my_proc, mod_id
/*OSF1*/			    , &me_p->info, sizeof(me_p->info)
/*OSF1*/			    , &ret_size );
/*OSF1*/
/*OSF1*/	    if ( rc )
/*OSF1*/	    {
/*OSF1*/    	    perror("ldr_inq_module");
/*OSF1*/    	    break;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/    	*mc_next = me_p;
/*OSF1*/	    mc_next  = &me_p->next;
/*OSF1*/
/*OSF1*/    	for ( j = 0
/*OSF1*/	    ; j < me_p->info.lmi_nregion
/*OSF1*/	    ; j++ )
/*OSF1*/	    {
/*OSF1*/	        re_p = (rce_p)calloc ( 1, sizeof(rce) );
/*OSF1*/	        if ( re_p == NULL ) break;
/*OSF1*/	        re_p->modul = me_p;
/*OSF1*/
/*OSF1*/	        rc = ldr_inq_region ( my_proc, mod_id, j
/*OSF1*/				    , &re_p->info, sizeof(re_p->info)
/*OSF1*/				    , &ret_size );
/*OSF1*/	        if ( rc )
/*OSF1*/	        {
/*OSF1*/    		    perror("ldr_inq_region");
/*OSF1*/	    	    break;
/*OSF1*/	        }
/*OSF1*/
/*OSF1*/	        *rc_next = re_p;
/*OSF1*/	        rc_next  = &re_p->next;
/*OSF1*/	        no_regions++;
/*OSF1*/	    }
/*OSF1*/    }
/*OSF1*/
/*OSF1*/    for ( i = 0; i < no_regions; i++ )
/*OSF1*/    {
/*OSF1*/	    for ( re_p = rc_anchor
/*OSF1*/	        , ptr = (void *)(-1)
/*OSF1*/	        , re_tmp = NULL
/*OSF1*/	        ; re_p
/*OSF1*/	        ; re_p = re_p->next )
/*OSF1*/
/*OSF1*/	    {
/*OSF1*/	        if ( re_p->used ) continue;
/*OSF1*/	        if ( re_p->info.lri_vaddr > ptr ) continue;
/*OSF1*/	        re_tmp = re_p;
/*OSF1*/	        ptr = re_p->info.lri_vaddr;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/	    if ( re_tmp )
/*OSF1*/	    {
/*OSF1*/	        re_tmp->used = 1;
/*OSF1*/	        *pc_next = re_tmp;
/*OSF1*/	        pc_next  = &re_tmp->next_pc;
/*OSF1*/	    }
/*OSF1*/    }
/*OSF1*/    return;
/*OSF1*/}
/*OSF1*/
/*OSF1*/extern int ld_ifd_symnum(LDFILE *ldptr, int);
/*OSF1*/extern char * st_str_ifd_iss(int, int);
/*OSF1*/
/*OSF1*/static void WalkStack (void)
/*OSF1*/{
/*OSF1*/
/*OSF1*/  static PRUNTIME_FUNCTION	prf;
/*OSF1*/  static CONTEXT		context;
/*OSF1*/  static int			rc;
/*OSF1*/
/*OSF1*/    exc_capture_context(&context);
/*OSF1*/    stack_depth = 0;
/*OSF1*/
/*OSF1*/    do
/*OSF1*/    {
/*OSF1*/	    stack[stack_depth++] = context.sc_pc - 4;
/*OSF1*/	    prf = exc_lookup_function_entry (context.sc_pc);
/*OSF1*/	    if ( prf == NULL )
/*OSF1*/	    {
/*OSF1*/	        /* damit diese frame auf jeden fall ausgewertet wird */
/*OSF1*/	        do
/*OSF1*/	        {
/*OSF1*/    		    stack[stack_depth++] = STACK_IM_ARSCH;
/*OSF1*/	        } while (stack_depth <= SKIP_TOP );	
/*OSF1*/	        stack_depth += SKIP_BOT;
/*OSF1*/	        break;
/*OSF1*/	    }
/*OSF1*/	    rc = exc_virtual_unwind (prf, &context);
/*OSF1*/    } while ((context.sc_pc) && (stack_depth < MAX_STACK));
/*OSF1*/}
/*OSF1*/
/*OSF1*/
/*OSF1*/static void GetSyms ()
/*OSF1*/{
/*OSF1*/  static char          MessageBuffer[4096];
/*OSF1*/  static int           i;
/*OSF1*/  static unsigned long pc;
/*OSF1*/  static rce_p	       re_p;
/*OSF1*/  static mce_p	       me_p;
/*OSF1*/  static PDR	       apd;
/*OSF1*/  static SYMR	       asym;
/*OSF1*/  static char	      *pname;
/*OSF1*/  static int	       isym;
/*OSF1*/  static int	       ifd;
/*OSF1*/  static unsigned long tmp;
/*OSF1*/
/*OSF1*/    write_string(" \n");
/*OSF1*/
/*OSF1*/    for ( i = SKIP_TOP; i < (stack_depth - SKIP_BOT); i++ )
/*OSF1*/    {
/*OSF1*/    	pc = stack[i];
/*OSF1*/    	if ( pc == STACK_IM_ARSCH )
/*OSF1*/	    {
/*OSF1*/    	    write_string("      *** Stack corrupted ***\n \n");
/*OSF1*/    	    break;
/*OSF1*/	    }
/*OSF1*/
/*OSF1*/    	for ( re_p = rc_anchor
/*OSF1*/	        ; re_p
/*OSF1*/	        ; re_p = re_p->next )
/*OSF1*/    	{
/*OSF1*/	        tmp = (unsigned long) re_p->info.lri_vaddr;
/*OSF1*/	        if ( ( pc >= tmp)
/*OSF1*/	           &&( pc <= (tmp + re_p->info.lri_size)))
/*OSF1*/	        {
/*OSF1*/		        break;
/*OSF1*/    	    }
/*OSF1*/	    }
/*OSF1*/    	if ( re_p )
/*OSF1*/	    {
/*OSF1*/    	    me_p = re_p->modul;
/*OSF1*/	        if ( me_p->opened == 0 )
/*OSF1*/	        {
/*OSF1*/    		    me_p->opened = 1;
/*OSF1*/	        	me_p->ldptr = ldopen(me_p->info.lmi_name, NULL);
/*OSF1*/        		if ( me_p->ldptr == NULL )
/*OSF1*/		        {
/*OSF1*/        		    sprintf ( MessageBuffer,
/*OSF1*/        			        "*** GetSyms: Cannot ldopen %s (errno: %d)\n",
/*OSF1*/			                me_p->info.lmi_name, errno );
/*OSF1*/                    write_string(MessageBuffer);
/*OSF1*/                }
/*OSF1*/	        	else
/*OSF1*/		        {
/*OSF1*/        		    if (PSYMTAB( me_p->ldptr ) == NULL)
/*OSF1*/	        	    {
/*OSF1*/		            	write_string("*** GetSyms: No symbol table for module ");
/*OSF1*/                        write_string( me_p->info.lmi_name );
/*OSF1*/                        write_string("\n");
/*OSF1*/                    }
/*OSF1*/        		    else
/*OSF1*/		            {
/*OSF1*/                		me_p->symtab_ok = 1;
/*OSF1*/	            	}
/*OSF1*/	    	    }
/*OSF1*/	        } /* if me_p->opened == 0 */
/*OSF1*/
/*OSF1*/	        if ( me_p->symtab_ok )
/*OSF1*/	        {
/*OSF1*/    		    ldgetpd(me_p->ldptr, ipd_adr(me_p->ldptr, pc), &apd);
/*OSF1*/	    	    isym = apd.isym;
/*OSF1*/		        if (isym != isymNil)
/*OSF1*/    		    {
/*OSF1*/	    	        if (ldtbread(me_p->ldptr, isym, &asym) == FAILURE)
/*OSF1*/		            {
/*OSF1*/			            sprintf( MessageBuffer,
/*OSF1*/			                     "%4d: 0x%lx\n"
/*OSF1*/                                 "      <symbol not available>\n"
/*OSF1*/                                 "      %s \\ %s\n"
/*OSF1*/                                 " \n",
/*OSF1*/			                     (i - SKIP_TOP),
/*OSF1*/			                     pc,
/*OSF1*/			                     me_p->info.lmi_name,
/*OSF1*/			                     re_p->info.lri_name
/*OSF1*/        			           );
/*OSF1*/                        write_string(MessageBuffer);
/*OSF1*/		        	    continue;
/*OSF1*/		            }
/*OSF1*/	    	        pname = ldgetname(me_p->ldptr, &asym);
/*OSF1*/    		        ifd = ld_ifd_symnum ( me_p->ldptr, apd.isym );
/*OSF1*/	    	        if ( ifd == -1 )
/*OSF1*/		            {
/*OSF1*/			            sprintf( MessageBuffer,
/*OSF1*/        			             "%4d: 0x%lx\n"
/*OSF1*/                                 "      %s\n"
/*OSF1*/                                 "      <source info not available>\n"
/*OSF1*/                                 "%s \\ %s\n \n",
/*OSF1*/			                     (i - SKIP_TOP),
/*OSF1*/			                     pc,
/*OSF1*/			                     pname,
/*OSF1*/			                     me_p->info.lmi_name,
/*OSF1*/			                     re_p->info.lri_name
/*OSF1*/			                    );
/*OSF1*/                        write_string(MessageBuffer);
/*OSF1*/    		        }
/*OSF1*/	    	        else
/*OSF1*/		            {
/*OSF1*/			            sprintf( MessageBuffer,
/*OSF1*/			                    "%4d: 0x%lx\n"
/*OSF1*/			                    "      %s\n"
/*OSF1*/			                    "      %s : %5d\n"
/*OSF1*/			                    "      %s \\ %s\n"
/*OSF1*/			                    " \n",
/*OSF1*/			                    (i - SKIP_TOP),
/*OSF1*/        			            pc,
/*OSF1*/		        	            pname,
/*OSF1*/			                    st_str_ifd_iss(ifd, 1),
/*OSF1*/			                    SYMTAB(me_p->ldptr)->pline[apd.iline
/*OSF1*/						    	    + ((pc - apd.adr)/4)],
/*OSF1*/		        	            me_p->info.lmi_name,
/*OSF1*/        			            re_p->info.lri_name
/*OSF1*/        			           );
/*OSF1*/                        write_string(MessageBuffer);
/*OSF1*/		            }
/*OSF1*/		        }
/*OSF1*/    		    else
/*OSF1*/	    	    {
/*OSF1*/		            sprintf( MessageBuffer,
/*OSF1*/                            "%4d: 0x%lx\n"
/*OSF1*/			                "      <symbol not available>\n"
/*OSF1*/			                "      %s \\ %s\n"
/*OSF1*/			                " \n",
/*OSF1*/                            (i - SKIP_TOP),
/*OSF1*/			                pc,
/*OSF1*/			                me_p->info.lmi_name,
/*OSF1*/			                re_p->info.lri_name
/*OSF1*/    			            );
/*OSF1*/                     write_string(MessageBuffer);
/*OSF1*/    		    }
/*OSF1*/
/*OSF1*/	        } /* me_p->symtab_ok == TRUE */		
/*OSF1*/	        else
/*OSF1*/	        {
/*OSF1*/	    	    sprintf( MessageBuffer,
/*OSF1*/		                "%4d: 0x%lx\n"
/*OSF1*/			            "      <no symbols available for this region>\n"
/*OSF1*/			            "      %s \\ %s\n"
/*OSF1*/			            " \n",
/*OSF1*/		                (i - SKIP_TOP),
/*OSF1*/		                pc,
/*OSF1*/		                me_p->info.lmi_name,
/*OSF1*/		                re_p->info.lri_name
/*OSF1*/		           );
/*OSF1*/                 write_string(MessageBuffer);
/*OSF1*/	        } /* me_p->symtab_ok == FALSE */		
/*OSF1*/	    } /* if re_p */
/*OSF1*/	    else
/*OSF1*/	    {
/*OSF1*/	        sprintf( MessageBuffer,
/*OSF1*/                    "%4d: 0x%lx\n",
/*OSF1*/		            "      <not in any loaded region>\n",
/*OSF1*/		            " \n",
/*OSF1*/    		        (i - SKIP_TOP),
/*OSF1*/		            pc
/*OSF1*/		       );
/*OSF1*/            write_string(MessageBuffer);
/*OSF1*/	    }
/*OSF1*/    } /* end for loop ueber stack-array */
/*OSF1*/    return;
/*OSF1*/}
/*OSF1*/
/*OSF1*/static long ipd_adr (LDFILE *ldptr, unsigned long adr)
/*OSF1*/{
/*OSF1*/  static int ilow;
/*OSF1*/  static int ihigh;
/*OSF1*/  static int ihalf;
/*OSF1*/  static int ilowold;
/*OSF1*/  static int ihighold;
/*OSF1*/  static int ipd;
/*OSF1*/  static PDR apd;
/*OSF1*/
/*OSF1*/    ilow  = 0;
/*OSF1*/    ihigh = SYMHEADER(ldptr).ipdMax;
/*OSF1*/
/*OSF1*/    /* binary search proc table */
/*OSF1*/
/*OSF1*/    while (ilow < ihigh)
/*OSF1*/    {
/*OSF1*/    	ihalf    = (ilow + ihigh) / 2;
/*OSF1*/	    ilowold  = ilow;
/*OSF1*/    	ihighold = ihigh;
/*OSF1*/	    ldgetpd(ldptr, ihalf, &apd);
/*OSF1*/    	if (adr < apd.adr)
/*OSF1*/	    {
/*OSF1*/    	    ihigh = ihalf;
/*OSF1*/	    }
/*OSF1*/    	else if (adr > apd.adr)
/*OSF1*/	    {
/*OSF1*/    	    ilow = ihalf;
/*OSF1*/	    }
/*OSF1*/    	else
/*OSF1*/	    {
/*OSF1*/    	    ilow = ihigh = ihalf;
/*OSF1*/	        break;
/*OSF1*/    	}
/*OSF1*/	    if (ilow == ilowold && ihigh == ihighold)
/*OSF1*/    	{
/*OSF1*/	        break;
/*OSF1*/    	}
/*OSF1*/    } /* while */
/*OSF1*/
/*OSF1*/    ipd =  ((ilow < ihigh || ihigh < 0) ? ilow : ihigh);
/*OSF1*/
/*OSF1*/    return (ipd);
/*OSF1*/}
#endif /* OSF1 */

#ifdef AIX
/*AIX*/
#include <sys/debug.h>
#include <unistd.h>
#include <setjmp.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
/**************************************************************************/
/* All the following code is required for using the StackTrace() function */
/* Each function has a word of 0000 after it's code followed by a pointer */
/* to a traceback table.  Contained within the traceback table is the name*/
/* of that function.  Due to the possiblity of very large functions or    */
/* functions without this information, there is a literal defined below   */
/* called MAX_FUNC_SEARCH_LEN.  This is the maximum number of words that  */
/* will be searched (following any function return address located on the */
/* stack) before the getname function decides to just return NONAME.      */
/* IMPORTANT NOTE:  This function also sets up a signal handler for       */
/* SIGSEGV.  This is to handle having this function run off the end of    */
/* the code segment when attempting to locate the tracetable and function */
/* name.  Please NOTE that YOU MUST call sigset() after this function to  */
/* re-install/set your signal handler for the signal SIGSEGV!.            */
/**************************************************************************/
/* ADDITIONAL NOTE:  When AIX prepares to call into a new function it will*/
/* have a small section of the stack set aside for parameters/etc.  It    */
/* will then execute a BL instruction which will save the iar (Instruction*/
/* pointer after execution of the BL instruction) into the link regiser.  */
/*  The function that receives control will then normally move the link   */
/* register into the r0 register and from there it will move the r0 info  */
/* into the stack area set aside from the calling process.  It will do    */
/* this at an offset of 2 32 bit words higher than the current stack ptr. */
/*   Unfortunately, some functions (OS specific) will not save off the    */
/* link register information when entered.  This is because they are not  */
/* going to be passing control off to any lower levels using this stack.  */
/* So...You may sometimes notice 1 line of invalid function information   */
/* from this routine if there was a signal involved and the currently     */
/* executing code was somewhere inbetween the creation of the stack area  */
/* and the moving of the link register into that area after function      */
/* transfer via the bl instruction.  Due to this inaccuracy it is very    */
/* important to take all pre signal stack information carefully.  This    */
/* is especially true if the last instruction was a bctr - Russ Blaisdell */
/**************************************************************************/
/*AIX*/
/*AIX*/static void write_string(const char *buffer)
/*AIX*/{
/*AIX*/    write(2, buffer, strlen(buffer) );
/*AIX*/}
/*AIX*/
/*AIX*/static void GetFuncName(unsigned int *lr, char *funcname);
#define MAX_FUNC_SEARCH_LEN 5000
/*AIX*/
/*AIX*/static jmp_buf ReturnLocation;
/*AIX*/
/*AIX*//*	This is the stack trace back SEGV signal handler. */
/*AIX*/static void StackTraceSegvFunc(int Signal)
/*AIX*/{
/*AIX*/    siglongjmp(ReturnLocation,1);
/*AIX*/}
/*AIX*/
/*AIX*/
/*AIX*/static void GetFuncName(unsigned int *lr, char *funcname)
/*AIX*/{
/*AIX*/static struct tbtable *tb;
/*AIX*/static unsigned int   *ip;
/*AIX*/static short          name_len;
/*AIX*/static unsigned int  searchcount;
/*AIX*/ searchcount = 0;
/*AIX*//*	First, search from the input LR address looking for a 32 bit work */
/*AIX*//*	of zeros. */
/*AIX*/	ip = lr;
/*AIX*/	while((*ip != NULL) && (searchcount++ < MAX_FUNC_SEARCH_LEN)) ip++;
/*AIX*/
/*AIX*/	if(*ip != 0) return;
/*AIX*/
/*AIX*//* set up addressability to the traceback table. */
/*AIX*/	tb = (struct tbtable *) (ip+1);
/*AIX*/
/*AIX*//* Are we in a global linkage routine? */
/*AIX*/	if ( tb->tb.globallink != TRUE )
/*AIX*/	{
/*AIX*/
/*AIX*//* 	Existence of fields in the tbtable extension are       */
/*AIX*//* 	contingent upon specific fields in the base table.     */
/*AIX*//* 	Check for their existence so that we can address the   */
/*AIX*//* 	function name if it exists.                            */
/*AIX*/
/*AIX*/
/*AIX*/        ip = (unsigned int *) tb +
/*AIX*/         		sizeof(struct tbtable_short)/sizeof(int);
/*AIX*/
/*AIX*/		if ( tb->tb.fixedparms != 0 || tb->tb.floatparms != 0 )
/*AIX*/      			ip++;
/*AIX*/
/*AIX*/		if ( tb->tb.has_tboff == TRUE )
/*AIX*/			ip++;
/*AIX*/
/*AIX*/		if ( tb->tb.int_hndl == TRUE )
/*AIX*/			*ip++;
/*AIX*/
/*AIX*/		if ( tb->tb.has_ctl == TRUE ) 
/*AIX*/			ip += (*ip) + 1; /* don't care */
/*AIX*/
/*AIX*//*		Get function name if it exists.       */
/*AIX*/		if ( tb->tb.name_present == TRUE )
/*AIX*/		{
/*AIX*/			name_len = * (short *) ip;
/*AIX*/            memcpy (funcname,(char *)ip+2,name_len);
/*AIX*/			funcname[name_len] = 0;
/*AIX*/		}
/*AIX*/	}
/*AIX*/}
/*AIX*/
/*AIX*/
/*AIX*//*	In AIX the stack pointer is in GPR 1.  The following pragma is used   */
/*AIX*//*	by the stack traceback to locate the stack.  The function moves       */
/*AIX*//*	GPR 1 to GPR 3 which is the register used for function return info.   */
/*AIX*/
/*AIX*/ulong getstkptr();
/*AIX*/
#pragma mc_func getstkptr   { "7c230b78" }          /* mr r1,r3 */
#pragma  reg_killed_by   getstkptr	gr3
/*AIX*/
/*AIX*/
/*AIX*/static void AixTraceStack (void)
/*AIX*/{
/*AIX*/static unsigned long *sp;		/* stack pointer		*/
/*AIX*/static unsigned int  *lr;		/* contains the link reg from stack */
/*AIX*/static unsigned long *ip;		/* working pointers		*/
/*AIX*/static unsigned long *oldsp;
/*AIX*/static char           funcname[4096];
/*AIX*/static void *         oldSEGVFunc;
/*AIX*/static int            WeFailed;
/*AIX*/static char           MessageBuffer[4096];
/*AIX*/
/*AIX*/static struct sigaction   dp_sigorig[NSIG];
/*AIX*/static int                sig;
/*AIX*/
/*AIX*//*	save signal settings, restore at the end of AixStack (hq, 20.11.95)  */
/*AIX*/	for (sig=1; sig<NSIG; sig++)
/*AIX*/  	{
/*AIX*/    	if (sigaction (sig, (struct sigaction *) 0,
/*AIX*/		   &dp_sigorig[sig]) < 0)
/*AIX*/        {
/*AIX*/            sprintf(MessageBuffer, "DpSigReaction: sigaction sig %d (%d)\n", sig, errno); 
/*AIX*/            write_string(MessageBuffer);
/*AIX*/        }
/*AIX*/  	}
/*AIX*/
/*AIX*//* 	First insure that we do not cause a segment violation when we look for  */
/*AIX*//* 	the TraceBack table in the code segment.  So, we add a signal handle    */
/*AIX*//* 	for SIGSEGV.                                                            */
/*AIX*/
/*AIX*/  	oldSEGVFunc = (void *) sigset(SIGSEGV,StackTraceSegvFunc);
/*AIX*/
/*AIX*/  	sp = (unsigned long *)getstkptr();	/* get GPR 1 stackptr	*/
/*AIX*/
/*AIX*//*	Start off with our caller */
/*AIX*/  	oldsp = sp-10;
/*AIX*/
/*AIX*//*	Loop while we still have a valid stack pointer */
/*AIX*/
/*AIX*/  	while(1)
/*AIX*/  	{
/*AIX*/
/*AIX*//*	If this is the end of the stack then just get out. */
/*AIX*/
#ifdef BIT64
/*AIX*/		if((((long)sp & 0xf000000000000000) != 0) ||
#else
/*AIX*/		if((((long)sp & 0xf0000000) != 0x20000000) ||
#endif
/*AIX*/		    (sp <= oldsp))
/*AIX*/      			break;
/*AIX*/
/*AIX*/    		oldsp = sp;
/*AIX*/
/*AIX*//*	See if this is a signal.  This is done by looking at the LR field  */
/*AIX*//*	on the stack.  The LR is at offset long 2 (byte 8 in 32 bit mode   */
/*AIX*//*	or byte 16 in 64 bit mode).  If the address is from segment 0,     */
/*AIX*//*	then this must be a signal.                                        */
/*AIX*/
/*AIX*/    		ip = sp;
/*AIX*/    		if (*(ip+2) < 0x0fffffff) 	/* check LR field on stack */
/*AIX*/    		{
/*AIX*/                write_string("Cannot traceback signal stack (YET)\n");
/*AIX*//*	GPR1 in the saved context is the stack pointer from the function */
/*AIX*//*	that was interrupted. */
/*AIX*/        
/*AIX*/        /*      sp = (unsigned long *)scp->sc_jmpbuf.jmp_context.gpr[1]; */
/*AIX*/        /*      lr = (unsigned int *)scp->sc_jmpbuf.jmp_context.iar;     */
/*AIX*/    		}
/*AIX*/    		else
/*AIX*/    		{
/*AIX*//*	If here, the stack frame is not the result of a signal.  The LR is */
/*AIX*//*	accessed at long + 2, but the pointer to the next stack frame is   */
/*AIX*//*	pointed to by the contents of where the current pointer points.    */
/*AIX*/
/*AIX*/      		    lr = (unsigned int *)  *(sp+2);
/*AIX*/      		    sp = (unsigned long *)  *sp;
/*AIX*/    		}
/*AIX*/
/*AIX*/
/*AIX*//*	Handle potential local stack allocations within a function  */
/*AIX*/
/*AIX*/    		if(lr == NULL) continue;
/*AIX*/
/*AIX*//*	This is to take care of SEGV. */
/*AIX*/    		WeFailed = sigsetjmp(ReturnLocation,1);
/*AIX*/
/*AIX*//*	Plan on not finding a name */
/*AIX*/    		sprintf(funcname,"NONAME");
/*AIX*/
/*AIX*//*	Attempt to locate name in tracebacktable (if we don't trap looking) */
/*AIX*/
/*AIX*/    		if (WeFailed == 0) GetFuncName(lr,funcname);
/*AIX*/
/*AIX*//*	Print location where call was made from */
/*AIX*/
/*AIX*/    		sprintf(MessageBuffer, "%s at 0x%lx\n",funcname, (char *)lr-sizeof(int));
/*AIX*/            write_string(MessageBuffer);
/*AIX*/
/*AIX*/            /* Main stack ends here... */
/*AIX*/    		if (strcmp (funcname, "main") == 0)
/*AIX*/    			break;
/*AIX*/
/*AIX*/            /* UKT Stack ends here... */
/*AIX*/            if (strcmp (funcname, "en88_CallA91Main") == 0)
/*AIX*/    			break;
/*AIX*/    }
/*AIX*/
/*AIX*//*	restore signal settings (hq, 20.11.95) */
/*AIX*/
/*AIX*/  	for (sig=1; sig<NSIG; sig++)
/*AIX*/  	{
/*AIX*/    		switch (sig)
/*AIX*/    		{
/*AIX*/      			case SIGKILL:
/*AIX*/      			case SIGSTOP:
/*AIX*/
/*AIX*/			continue;
/*AIX*/    		}
/*AIX*/
/*AIX*/    		if (sigaction (sig, &dp_sigorig[sig],
/*AIX*/    		   	(struct sigaction *) 0) < 0)
/*AIX*/            {
/*AIX*/                sprintf(MessageBuffer, "DpSigReset: sigaction sig %d (%d)\n", sig, errno); 
/*AIX*/                write_string(MessageBuffer);
/*AIX*/            }
/*AIX*/    }
/*AIX*/}
#endif /* AIX */

#ifdef WIN32

#include <windows.h>
#include <imagehlp.h>

#define MAXNAMELEN 1024 // max name length for found symbols
#define IMGSYMLEN ( sizeof(IMAGEHLP_SYMBOL) )
#define TTBUFLEN 65536 // for a temp buffer

// SymCleanup()
typedef BOOL (__stdcall *tSC)( IN HANDLE hProcess );

// SymGetOptions()
typedef DWORD (__stdcall *tSGO)( VOID );

 // SymInitialize()
 typedef BOOL (__stdcall *tSI)( IN HANDLE hProcess, IN PSTR UserSearchPath, IN BOOL fInvadeProcess );

 // SymSetOptions()
 typedef DWORD (__stdcall *tSSO)( IN DWORD SymOptions );

 // UnDecorateSymbolName()
 typedef DWORD (__stdcall WINAPI *tUDSN)( PCSTR DecoratedName, PSTR UnDecoratedName,
	 DWORD UndecoratedLength, DWORD Flags );

#if defined (_WIN64)
 // SymFunctionTableAccess()
 typedef PVOID (__stdcall *tSFTA)( HANDLE hProcess, DWORD64 AddrBase );

 // SymGetLineFromAddr()
 typedef BOOL (__stdcall *tSGLFA)( IN HANDLE hProcess, IN DWORD64 dwAddr,
	 OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_LINE64 Line );

 // SymGetModuleBase()
 typedef DWORD64 (__stdcall *tSGMB)( IN HANDLE hProcess, IN DWORD64 dwAddr );

 // SymGetModuleInfo()
 typedef BOOL (__stdcall *tSGMI)( IN HANDLE hProcess, IN DWORD64 dwAddr, OUT PIMAGEHLP_MODULE64 ModuleInfo );

 // SymGetSymFromAddr()
 typedef BOOL (__stdcall *tSGSFA)( IN HANDLE hProcess, IN DWORD64 dwAddr,
 	 OUT PDWORD64 pdwDisplacement, OUT PIMAGEHLP_SYMBOL64 Symbol );

 // SymLoadModule()
 typedef DWORD64 (__stdcall *tSLM)( IN HANDLE hProcess, IN HANDLE hFile,
 	 IN PSTR ImageName, IN PSTR ModuleName, IN DWORD64 BaseOfDll, IN DWORD SizeOfDll );

 // StackWalk()
 typedef BOOL (__stdcall *tSW)( DWORD MachineType, HANDLE hProcess,
 	 HANDLE hThread, LPSTACKFRAME64 StackFrame, PVOID ContextRecord,
	 PREAD_PROCESS_MEMORY_ROUTINE64 ReadMemoryRoutine,
	 PFUNCTION_TABLE_ACCESS_ROUTINE64 FunctionTableAccessRoutine,
	 PGET_MODULE_BASE_ROUTINE64 GetModuleBaseRoutine,
	 PTRANSLATE_ADDRESS_ROUTINE64 TranslateAddress );

#else /* WIN32 only */

 // SymFunctionTableAccess()
 typedef PVOID (__stdcall *tSFTA)( HANDLE hProcess, DWORD AddrBase );

 // SymGetLineFromAddr()
 typedef BOOL (__stdcall *tSGLFA)( IN HANDLE hProcess, IN DWORD dwAddr,
	 OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_LINE Line );

 // SymGetModuleBase()
 typedef DWORD (__stdcall *tSGMB)( IN HANDLE hProcess, IN DWORD dwAddr );

 // SymGetModuleInfo()
 typedef BOOL (__stdcall *tSGMI)( IN HANDLE hProcess, IN DWORD dwAddr, OUT PIMAGEHLP_MODULE ModuleInfo );

 // SymGetSymFromAddr()
 typedef BOOL (__stdcall *tSGSFA)( IN HANDLE hProcess, IN DWORD dwAddr,
 	 OUT PDWORD pdwDisplacement, OUT PIMAGEHLP_SYMBOL Symbol );

 // SymLoadModule()
 typedef BOOL (__stdcall *tSLM)( IN HANDLE hProcess, IN HANDLE hFile,
 	 IN PSTR ImageName, IN PSTR ModuleName, IN DWORD BaseOfDll, IN DWORD SizeOfDll );

 // StackWalk()
 typedef BOOL (__stdcall *tSW)( DWORD MachineType, HANDLE hProcess,
 	 HANDLE hThread, LPSTACKFRAME StackFrame, PVOID ContextRecord,
	 PREAD_PROCESS_MEMORY_ROUTINE ReadMemoryRoutine,
	 PFUNCTION_TABLE_ACCESS_ROUTINE FunctionTableAccessRoutine,
	 PGET_MODULE_BASE_ROUTINE GetModuleBaseRoutine,
	 PTRANSLATE_ADDRESS_ROUTINE TranslateAddress );
#endif

/*---------------------------------------------------------------------------*/

static void ShowStack( HANDLE hThread, CONTEXT * c );
static void enumAndLoadModuleSymbols( HANDLE hProcess, DWORD pid );
static bool fillModuleListPSAPI( DWORD pid, HANDLE hProcess );
static bool fillModuleListTH32( DWORD pid, HANDLE hProcess );

#if defined(USE_MSG_PRIO7)
extern void sql60_int_err_msg_prio7( char *pszFormatString, ... );
#endif

static void write_string(const char *str)
{
#if defined(USE_MSG_PRIO7)
  static char *pc;
  pc = strchr(str,'\n');
  while ( pc ) 
  {
    *pc = ' ';
    pc = strchr(pc,'\n');
  }
  sql60_int_err_msg_prio7( "%s", str );
#else
  fprintf(stderr, "%s", str);
#endif
}

void eo670_CTraceContextStack(void* context) {
  ShowStack(GetCurrentThread(), (CONTEXT*)context);
};

static void Win32TraceStack()
{
  static CONTEXT c;
  static HANDLE  thisThread;

  memset( &c, '\0', sizeof(c) );
  c.ContextFlags = CONTEXT_FULL;

  // init CONTEXT record so we know where to start the stackwalk
  thisThread = GetCurrentThread();
  if ( ! GetThreadContext( thisThread, &c ) )
  {
    write_string("Failed to get context\n");
    return;
  }

  ShowStack(thisThread, &c);
}

static tSLM pSymLoadModule = NULL;

static void ShowStack( HANDLE hThread, CONTEXT *c )
{
  // normally, call ImageNtHeader() and use machine info from PE header
# if defined (_WIN64)
   static IMAGEHLP_LINE64   Line;
   static STACKFRAME64      s; // in/out stackframe
   static DWORD64           offsetFromSymbol; // tells us how far from the symbol we were
# else
   static IMAGEHLP_LINE     Line;
   static STACKFRAME        s; // in/out stackframe
   static DWORD             offsetFromSymbol; // tells us how far from the symbol we were
# endif

  static DWORD  offsetFromLine; 
  static DWORD imageType = IMAGE_FILE_MACHINE_I386;
  static HANDLE hProcess;
  static int frameNum; // counts walked frames
  static DWORD symOptions; // symbol handler settings
  static double pSymSpace[(IMGSYMLEN + MAXNAMELEN)/sizeof(double)+1];
  static IMAGEHLP_SYMBOL *pSym = (IMAGEHLP_SYMBOL *) pSymSpace;
  static char undName[MAXNAMELEN]; // undecorated name
  static char undFullName[MAXNAMELEN]; // undecorated name with all shenanigans
  static IMAGEHLP_MODULE Module;
  static char symSearchPath[4096];
  static char *p;
  static char MessageLine[2048];
  static char tt[TTBUFLEN];
  
  static tSC pSymCleanup = NULL;
  static tSFTA pSymFunctionTableAccess = NULL;
  static tSGLFA pSymGetLineFromAddr = NULL;
  static tSGMB pSymGetModuleBase = NULL;
  static tSGMI pSymGetModuleInfo = NULL;
  static tSGO pSymGetOptions = NULL;
  static tSGSFA pSymGetSymFromAddr = NULL;
  static tSI pSymInitialize = NULL;
  static tSSO pSymSetOptions = NULL;
  static tSW pStackWalk = NULL;
  static tUDSN pUnDecorateSymbolName = NULL;
  static HINSTANCE hImagehlpDll = NULL;
  
  hProcess = GetCurrentProcess(); // hProcess normally comes from outside
  // we load imagehlp.dll dynamically because the NT4-version does not
  // offer all the functions that are in the NT5 lib
  hImagehlpDll = LoadLibrary( "imagehlp.dll" );
  if ( hImagehlpDll != NULL )
  {
    pSymCleanup = (tSC) GetProcAddress( hImagehlpDll, "SymCleanup" );
    pSymFunctionTableAccess = (tSFTA) GetProcAddress( hImagehlpDll, "SymFunctionTableAccess" );
    pSymGetLineFromAddr = (tSGLFA) GetProcAddress( hImagehlpDll, "SymGetLineFromAddr" );
    pSymGetModuleBase = (tSGMB) GetProcAddress( hImagehlpDll, "SymGetModuleBase" );
    pSymGetModuleInfo = (tSGMI) GetProcAddress( hImagehlpDll, "SymGetModuleInfo" );
    pSymGetOptions = (tSGO) GetProcAddress( hImagehlpDll, "SymGetOptions" );
    pSymGetSymFromAddr = (tSGSFA) GetProcAddress( hImagehlpDll, "SymGetSymFromAddr" );
    pSymInitialize = (tSI) GetProcAddress( hImagehlpDll, "SymInitialize" );
    pSymLoadModule = (tSLM) GetProcAddress( hImagehlpDll, "SymLoadModule" );
    pSymSetOptions = (tSSO) GetProcAddress( hImagehlpDll, "SymSetOptions" );
    pStackWalk = (tSW) GetProcAddress( hImagehlpDll, "StackWalk" );
    pUnDecorateSymbolName = (tUDSN) GetProcAddress( hImagehlpDll, "UnDecorateSymbolName" );
  } 
#if !defined(USE_MSG_PRIO7)
  else {
    pSymCleanup = SymCleanup;
    pSymFunctionTableAccess = SymFunctionTableAccess;
    pSymGetLineFromAddr = SymGetLineFromAddr;
    pSymGetModuleBase = SymGetModuleBase;
    pSymGetModuleInfo = SymGetModuleInfo;
    pSymGetOptions = SymGetOptions;
    pSymGetSymFromAddr = SymGetSymFromAddr;
    pSymInitialize = SymInitialize;
    pSymLoadModule = SymLoadModule;
    pSymSetOptions = SymSetOptions;
    pStackWalk = StackWalk;
    pUnDecorateSymbolName = UnDecorateSymbolName;
  }
#endif

  if (
    pSymCleanup == NULL
    || pSymFunctionTableAccess == NULL
//    || pSymGetLineFromAddr == NULL
    || pSymGetModuleBase == NULL
    || pSymGetModuleInfo == NULL
    || pSymGetOptions == NULL
    || pSymGetSymFromAddr == NULL
    || pSymInitialize == NULL
    || pSymLoadModule == NULL
    || pSymSetOptions == NULL
    || pStackWalk == NULL
    || pUnDecorateSymbolName == NULL )
  {
    write_string( "GetProcAddress(): some required function not found.\n" );
    if ( pSymCleanup == NULL )
      write_string( "GetProcAddress(): 'SymCleanup' not found.\n" );
    if ( pSymFunctionTableAccess == NULL )
      write_string( "GetProcAddress(): 'SymFunctionTableAccess' not found.\n" );
    if ( pSymGetLineFromAddr == NULL )
      write_string( "GetProcAddress(): 'SymGetLineFromAddr' not found.\n" );
    if ( pSymGetModuleBase == NULL )
      write_string( "GetProcAddress(): 'SymGetModuleBase' not found.\n" );
    if ( pSymGetModuleInfo == NULL )
      write_string( "GetProcAddress(): 'SymGetModuleInfo' not found.\n" );
    if ( pSymGetOptions == NULL )
      write_string( "GetProcAddress(): 'SymGetOptions' not found.\n" );
    if ( pSymGetSymFromAddr == NULL )
      write_string( "GetProcAddress(): 'SymGetSymFromAddr' not found.\n" );
    if ( pSymInitialize == NULL )
      write_string( "GetProcAddress(): 'SymInitialize' not found.\n" );
    if ( pSymLoadModule == NULL )
      write_string( "GetProcAddress(): 'SymLoadModule' not found.\n" );
    if ( pSymSetOptions == NULL )
      write_string( "GetProcAddress(): 'SymSetOptions' not found.\n" );
    if ( pStackWalk == NULL )
      write_string( "GetProcAddress(): 'StackWalk' not found.\n" );
    if ( pUnDecorateSymbolName == NULL )
      write_string( "GetProcAddress(): 'UnDecorateSymbolName' not found.\n" );

    return;
  }
  
  memset( &s, '\0', sizeof(s) );
  
  // NOTE: normally, the exe directory and the current directory should be taken
  // from the target process. The current dir would be gotten through injection
  // of a remote thread; the exe dir through either ToolHelp32 or PSAPI.
  
  // build symbol search path from:
  *symSearchPath = 0;
  // current directory
  if ( GetCurrentDirectory( TTBUFLEN, tt ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // dir with executable
  if ( GetModuleFileName( 0, tt, TTBUFLEN ) )
  {
    for ( p = tt + strlen( tt ) - 1; p >= tt; -- p )
    {
      // locate the rightmost path separator
      if ( *p == '\\' || *p == '/' || *p == ':' )
        break;
    }
    // if we found one, p is pointing at it; if not, tt only contains
    // an exe name (no path), and p points before its first byte
    if ( p != tt ) // path sep found?
    {
      if ( *p == ':' ) // we leave colons in place
        ++ p;
      *p = '\0'; // eliminate the exe name and last path sep
      strcat(symSearchPath, tt);
      strcat(symSearchPath, ";");
    }
  }
  // environment variable _NT_SYMBOL_PATH
  if ( GetEnvironmentVariable( "_NT_SYMBOL_PATH", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // environment variable _NT_ALTERNATE_SYMBOL_PATH
  if ( GetEnvironmentVariable( "_NT_ALTERNATE_SYMBOL_PATH", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // environment variable _NT_ALT_SYMBOL_PATH
  if ( GetEnvironmentVariable( "_NT_ALT_SYMBOL_PATH", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  // environment variable SYSTEMROOT
  if ( GetEnvironmentVariable( "SYSTEMROOT", tt, TTBUFLEN ) ) {
    strcat(symSearchPath, tt);
    strcat(symSearchPath, ";");
  }
  
  if ( strlen(symSearchPath) > 0 ) // if we added anything, we have a trailing semicolon
    symSearchPath[strlen(symSearchPath)-1] = 0;
  
#if defined(USE_MSG_PRIO7)
  {
  static tsp00_Pathc       szInstPath;
  static tsp01_RteError    rteError;
  if ( sqlGetDbrootSapPath ( szInstPath, TERM_WITH_DELIMITER_EO01, &rteError )) {
    strcat(symSearchPath, ";");
    strcat(symSearchPath, &szInstPath[0]); 
  }
  }
#endif
  
  write_string("symSearchPath:");
  write_string(symSearchPath);
  write_string("\n");
  
  if ( ! pSymInitialize( hProcess, symSearchPath, false ) )
  {
    write_string("SymInitialize() failed\n");
    goto cleanup;
  }
  
  symOptions = pSymGetOptions();
  symOptions |= SYMOPT_LOAD_LINES;
  symOptions &= ~SYMOPT_UNDNAME;
  pSymSetOptions( symOptions );
  
  // Enumerate modules and tell imagehlp.dll about them.
  // On NT, this is not necessary, but it won't hurt.
  enumAndLoadModuleSymbols( hProcess, GetCurrentProcessId() );
  
  // init STACKFRAME for first call
  // Notes: AddrModeFlat is just an assumption. I hate VDM debugging.
  // Notes: will have to be #ifdef-ed for Alphas; MIPSes are dead anyway,
  // and good riddance.
  s.AddrPC.Offset = c->Eip;
  s.AddrPC.Mode = AddrModeFlat;
  s.AddrStack.Offset = c->Esp;
  s.AddrStack.Mode = AddrModeFlat;
  s.AddrFrame.Offset = c->Ebp;
  s.AddrFrame.Mode = AddrModeFlat;
  
  memset( pSym, '\0', IMGSYMLEN + MAXNAMELEN );
  pSym->SizeOfStruct = IMGSYMLEN;
  pSym->MaxNameLength = MAXNAMELEN;
  
  memset( &Line, '\0', sizeof(Line) );
  Line.SizeOfStruct = sizeof(Line);
  
  memset( &Module, '\0', sizeof(Module) );
  Module.SizeOfStruct = sizeof(Module);
  
  offsetFromSymbol = 0;
  offsetFromLine   = 0;
  
  write_string( "\n--# FV EIP----- RetAddr- FramePtr StackPtr Symbol\n" );
  for ( frameNum = 0; ; ++ frameNum )
  {
    // get next stack frame (StackWalk(), SymFunctionTableAccess(), SymGetModuleBase())
    // if this returns ERROR_INVALID_ADDRESS (487) or ERROR_NOACCESS (998), you can
    // assume that either you are done, or that the stack is so hosed that the next
    // deeper frame could not be found.
    if ( ! pStackWalk( imageType, hProcess, hThread, &s, &c, NULL,
      pSymFunctionTableAccess, pSymGetModuleBase, NULL ) )
      break;
    
    // display its contents
#if defined (_WIN64)
    sprintf( MessageLine, "\n%3d %c%c %16I64x %16I64x %16I64x %16I64x ",
             frameNum, s.Far? 'F': '.', s.Virtual? 'V': '.',
             s.AddrPC.Offset, 
             s.AddrReturn.Offset,
             s.AddrFrame.Offset, 
             s.AddrStack.Offset );
#else
    sprintf( MessageLine, "\n%3d %c%c %08lx %08lx %08lx %08lx ",
             frameNum, s.Far? 'F': '.', s.Virtual? 'V': '.',
             s.AddrPC.Offset, 
             s.AddrReturn.Offset,
             s.AddrFrame.Offset, 
             s.AddrStack.Offset );
#endif
    write_string(MessageLine);
    
    if ( s.AddrPC.Offset == 0 )
    {
      write_string("(-nosymbols- PC == 0)\n" );
    }
    else
    { // we seem to have a valid PC
      // show procedure info (SymGetSymFromAddr())
      if ( ! pSymGetSymFromAddr( hProcess, s.AddrPC.Offset, &offsetFromSymbol, pSym ) )
      {
        if ( GetLastError() != 487 )
          write_string( "SymGetSymFromAddr() failed\n");
      }
      else
      {
        // UnDecorateSymbolName()
        pUnDecorateSymbolName( pSym->Name, undName, MAXNAMELEN, UNDNAME_NAME_ONLY );
        pUnDecorateSymbolName( pSym->Name, undFullName, MAXNAMELEN, UNDNAME_COMPLETE );
        if ( offsetFromSymbol != 0 ) {
#if defined (_WIN64)
          sprintf(MessageLine, "%s   %+I64d bytes", undName, offsetFromSymbol );
#else
          sprintf(MessageLine, "%s   %+ld bytes", undName, offsetFromSymbol );
#endif
          write_string(MessageLine);
        }
        else
          write_string(undName);
#if !defined(USE_MSG_PRIO7)
        write_string( "\n" );
#endif
        if ( strcmp(pSym->Name, undName) != 0 )
        {
          sprintf(MessageLine, " Symbol:  %s\n", pSym->Name );
          write_string(MessageLine);
        }
        if ( strcmp(undFullName, undName) != 0 )
        {
          sprintf(MessageLine, " Fully undecorated: %s\n", undFullName );
          write_string(MessageLine);
        }
      }
      
      // show line number info, NT5.0-method (SymGetLineFromAddr())
      if ( pSymGetLineFromAddr != NULL )
      { // yes, we have SymGetLineFromAddr()

        if ( ! pSymGetLineFromAddr( hProcess, s.AddrPC.Offset, &offsetFromLine, &Line ) )
        {
          if ( GetLastError() != 487 )
            write_string( "SymGetLineFromAddr() failed\n");
        }
        else
        {
          sprintf(MessageLine, "    Line: %s(%lu) %+ld bytes\n",
                  Line.FileName, Line.LineNumber, offsetFromLine );
          write_string(MessageLine);
        }
      } // yes, we have SymGetLineFromAddr()
      
      // show module info (SymGetModuleInfo())
      if ( ! pSymGetModuleInfo( hProcess, s.AddrPC.Offset, &Module ) )
      {
        write_string( "SymGetModuleInfo() failed\n");
      }
      else
      { // got module info OK
        static char ty[80];
        switch ( Module.SymType )
        {
        case SymNone:
          strcpy( ty, "-nosymbols-" );
          break;
        case SymCoff:
          strcpy( ty, "COFF" );
          break;
        case SymCv:
          strcpy( ty, "CV" );
          break;
        case SymPdb:
          strcpy( ty, "PDB" );
          break;
        case SymExport:
          strcpy( ty, "-exported-" );
          break;
        case SymDeferred:
          strcpy( ty, "-deferred-" );
          break;
        case SymSym:
          strcpy( ty, "SYM" );
          break;
        default:
          sprintf( ty, "symtype=%ld", (long) Module.SymType );
          break;
        }
        
#if defined (_WIN64)
        sprintf(MessageLine, "    Mod:  %s[%s], base: %16I64xh\n",
                Module.ModuleName, Module.ImageName, Module.BaseOfImage );
#else
        sprintf(MessageLine, "    Mod:  %s[%s], base: %08lxh\n",
                Module.ModuleName, Module.ImageName, Module.BaseOfImage );
#endif
        write_string(MessageLine);
#if !defined(USE_MSG_PRIO7)
        sprintf(MessageLine, "    Sym:  type: %s, file: %s\n",
          ty, Module.LoadedImageName );
        write_string(MessageLine);
#endif
      } // got module info OK
    } // we seem to have a valid PC
    
    // no return address means no deeper stackframe
    if ( s.AddrReturn.Offset == 0 )
    {
      // avoid misunderstandings in the printf() following the loop
      SetLastError( 0 );
      break;
    }
    
  } // for ( frameNum )
  
  if ( GetLastError() != 0 ) {
    sprintf(MessageLine, "\nStackWalk(): gle = %lu\n", GetLastError() );
    write_string(MessageLine);
  }
  
cleanup:
  pSymCleanup( hProcess );
}

static void enumAndLoadModuleSymbols( HANDLE hProcess, DWORD pid )
{
  // try toolhelp32 first
  if ( fillModuleListTH32( pid, hProcess ) )
    return;
  // nope? try psapi, then
  fillModuleListPSAPI( pid, hProcess );
}

// miscellaneous toolhelp32 declarations; we cannot #include the header
// because not all systems may have it
#define MAX_MODULE_NAME32 255
#define TH32CS_SNAPMODULE   0x00000008
#pragma pack( push, 8 )
typedef struct tagMODULEENTRY32
{
  DWORD   dwSize;
  DWORD   th32ModuleID;       // This module
  DWORD   th32ProcessID;      // owning process
  DWORD   GlblcntUsage;       // Global usage count on the module
  DWORD   ProccntUsage;       // Module usage count in th32ProcessID's context
  BYTE  * modBaseAddr;        // Base address of module in th32ProcessID's context
  DWORD   modBaseSize;        // Size in bytes of module starting at modBaseAddr
  HMODULE hModule;            // The hModule of this module in th32ProcessID's context
  char    szModule[MAX_MODULE_NAME32 + 1];
  char    szExePath[MAX_PATH];
} MODULEENTRY32;
typedef MODULEENTRY32 *  PMODULEENTRY32;
typedef MODULEENTRY32 *  LPMODULEENTRY32;
#pragma pack( pop )



static bool fillModuleListTH32( DWORD pid, HANDLE hProcess )
{
  // CreateToolhelp32Snapshot()
  typedef HANDLE (__stdcall *tCT32S)( DWORD dwFlags, DWORD th32ProcessID );
  // Module32First()
  typedef BOOL (__stdcall *tM32F)( HANDLE hSnapshot, LPMODULEENTRY32 lpme );
  // Module32Next()
  typedef BOOL (__stdcall *tM32N)( HANDLE hSnapshot, LPMODULEENTRY32 lpme );
  
  // I think the DLL is called tlhelp32.dll on Win9X, so we try both
  static const char *dllname[] = { "kernel32.dll", "tlhelp32.dll" };
  static HINSTANCE hToolhelp;
  static tCT32S pCT32S;
  static tM32F pM32F;
  static tM32N pM32N;
  
  static HANDLE hSnap;
  static MODULEENTRY32 me = { sizeof(me) };
  static bool keepGoing;
  static int i;
  
  for ( i = 0; i < (sizeof(dllname)/sizeof(dllname[0])); ++ i )
  {
    hToolhelp = LoadLibrary( dllname[i] );
    if ( hToolhelp == 0 )
      continue;
    pCT32S = (tCT32S) GetProcAddress( hToolhelp, "CreateToolhelp32Snapshot" );
    pM32F = (tM32F) GetProcAddress( hToolhelp, "Module32First" );
    pM32N = (tM32N) GetProcAddress( hToolhelp, "Module32Next" );
    if ( pCT32S != 0 && pM32F != 0 && pM32N != 0 )
      break; // found the functions!
    FreeLibrary( hToolhelp );
    hToolhelp = 0;
  }
  
  if ( hToolhelp == 0 ) // nothing found?
    return false;
  
  hSnap = pCT32S( TH32CS_SNAPMODULE, pid );
  if ( hSnap == (HANDLE) -1 )
    return false;
  
  keepGoing = !!pM32F( hSnap, &me );
  if ( !keepGoing ) 
  {
    CloseHandle( hSnap );
    FreeLibrary( hToolhelp );
    return false;
  }
  
  while ( keepGoing )
  {
    static char MessageBuffer[4096];
    // here, we have a filled-in MODULEENTRY32
#if defined (_WIN64)
    sprintf( MessageBuffer, "%16I64Xh %6lu %-15.15s %s\n", me.modBaseAddr, me.modBaseSize, me.szModule, me.szExePath );
#else
    sprintf( MessageBuffer, "%08lXh %6lu %-15.15s %s\n", me.modBaseAddr, me.modBaseSize, me.szModule, me.szExePath );
#endif
    write_string(MessageBuffer);
    if ( !pSymLoadModule( hProcess, 0, me.szExePath, me.szModule, (tsp00_Longuint)me.modBaseAddr, me.modBaseSize ) ) {
      sprintf( MessageBuffer, "Error %lu loading symbols for \"%s\"\n",
        GetLastError(), me.szModule );
      write_string(MessageBuffer);
    } 
#if !defined(USE_MSG_PRIO7)
    else {
      sprintf( MessageBuffer, "Symbols loaded: \"%s\"\n", me.szModule );
      write_string(MessageBuffer);
    }
#endif
    keepGoing = !!pM32N( hSnap, &me );
  }
  
  CloseHandle( hSnap );
  
  FreeLibrary( hToolhelp );
  
  return true;
}



// miscellaneous psapi declarations; we cannot #include the header
// because not all systems may have it
typedef struct _MODULEINFO {
  LPVOID lpBaseOfDll;
  DWORD SizeOfImage;
  LPVOID EntryPoint;
} MODULEINFO, *LPMODULEINFO;



static bool fillModuleListPSAPI( DWORD pid, HANDLE hProcess )
{
  static char MessageBuffer[4096];
  // EnumProcessModules()
  typedef BOOL (__stdcall *tEPM)( HANDLE hProcess, HMODULE *lphModule, DWORD cb, LPDWORD lpcbNeeded );
  // GetModuleFileNameEx()
  typedef DWORD (__stdcall *tGMFNE)( HANDLE hProcess, HMODULE hModule, LPSTR lpFilename, DWORD nSize );
  // GetModuleBaseName() -- redundant, as GMFNE() has the same prototype, but who cares?
  typedef DWORD (__stdcall *tGMBN)( HANDLE hProcess, HMODULE hModule, LPSTR lpFilename, DWORD nSize );
  // GetModuleInformation()
  typedef BOOL (__stdcall *tGMI)( HANDLE hProcess, HMODULE hModule, LPMODULEINFO pmi, DWORD nSize );
  
  static HINSTANCE hPsapi;
  static tEPM pEPM;
  static tGMFNE pGMFNE;
  static tGMBN pGMBN;
  static tGMI pGMI;
  
  static int i;
  static DWORD cbNeeded;
  static MODULEINFO mi;
  static HMODULE hMods[sizeof(HMODULE) * (TTBUFLEN / sizeof(HMODULE))];
  static char tt[TTBUFLEN];
  static char ttImage[TTBUFLEN];

  hPsapi = LoadLibrary( "psapi.dll" );
  if ( hPsapi == 0 )
    return false;
  
  pEPM = (tEPM) GetProcAddress( hPsapi, "EnumProcessModules" );
  pGMFNE = (tGMFNE) GetProcAddress( hPsapi, "GetModuleFileNameExA" );
  pGMBN = (tGMFNE) GetProcAddress( hPsapi, "GetModuleBaseNameA" );
  pGMI = (tGMI) GetProcAddress( hPsapi, "GetModuleInformation" );
  if ( pEPM == 0 || pGMFNE == 0 || pGMBN == 0 || pGMI == 0 )
  {
    // yuck. Some API is missing.
    FreeLibrary( hPsapi );
    return false;
  }
  
  if ( ! pEPM( hProcess, hMods, TTBUFLEN, &cbNeeded ) )
  {
    sprintf( MessageBuffer, "EPM failed, gle = %lu\n", GetLastError() );
    write_string(MessageBuffer);
    goto cleanup;
  }
  
  if ( cbNeeded > TTBUFLEN )
  {
    sprintf( MessageBuffer, "More than %lu module handles. Huh?\n", (TTBUFLEN/sizeof(HMODULE)) );
    write_string(MessageBuffer);
    goto cleanup;
  }
  
  for ( i = 0; i < (int)(cbNeeded / sizeof(HMODULE)); ++ i )
  {
    // for each module, get:
    // base address, size
    pGMI( hProcess, hMods[i], &mi, sizeof(mi) );
    // image file name
    ttImage[0] = '\0';
    pGMFNE( hProcess, hMods[i], ttImage, TTBUFLEN );
    // module name
    tt[0] = '\0';
    pGMBN( hProcess, hMods[i], tt, TTBUFLEN );
    
    if ( !pSymLoadModule( hProcess, 0, ttImage, tt, (tsp00_Longuint)mi.lpBaseOfDll, mi.SizeOfImage ) ) {
      sprintf( MessageBuffer, "Error %lu loading symbols for \"%s\"\n", GetLastError(), tt );
      write_string(MessageBuffer);
    } 
#if !defined(USE_MSG_PRIO7)
    else {
      sprintf( MessageBuffer, "Symbols loaded: \"%s\"\n", tt );
      write_string(MessageBuffer);
    }
#endif
  }
  
cleanup:
  if ( hPsapi )
    FreeLibrary( hPsapi );
  
  return true;
}

#endif /* WIN32 */

/*===========================================================================*
 *  END OF CODE                                                              *
 *===========================================================================*/