/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#if defined(_WIN32)
  #line 3 "d:\\v72\\sys\\src\\in\\vin105.cpp"
#endif 

/*! 
  -----------------------------------------------------------------------------
 
  module: vin105.cpp
 
  -----------------------------------------------------------------------------
 
  responsible:  BerndV
 
  special area: protocol file class
 
  description:  implementation of a protocol file class

  version:      min 7.2.

  -----------------------------------------------------------------------------
 
                          Copyright by SAP AG, 1998
 
  -----------------------------------------------------------------------------
*/
 

/* 
  -----------------------------------------------------------------------------
  includes
  -----------------------------------------------------------------------------
*/

#include "heo02.h"   // virtual file api
#include "heo06.h"   // virtual file api

#include "gin100.h"  // error messages
#include "hin105.h"  // class specification

/* 
  -----------------------------------------------------------------------------
  implementation class tin105_LogFile
  -----------------------------------------------------------------------------
*/

/* 
  -----------------------------------------------------------------------------
  public function tin105_LogFile :: enable
  -----------------------------------------------------------------------------
*/
void tin105_LogFile :: enable(const bool bEnabled)
{
  m_bEnabled = bEnabled;
} // end tin105_LogFile :: enable

/* 
  -----------------------------------------------------------------------------
  public function tin105_LogFile :: disable
  -----------------------------------------------------------------------------
*/
void tin105_LogFile :: disable()
{
  m_bEnabled = false;
} // end tin105_LogFile :: disable

/* 
  -----------------------------------------------------------------------------
  public function tin105_LogFile :: isEnabled
  -----------------------------------------------------------------------------
*/
bool tin105_LogFile :: isEnabled() const
{
  return m_bEnabled;
} // end tin105_LogFile :: isEnabled

  /* 
  -----------------------------------------------------------------------------
  protected constructor tin105_LogFile :: tin105_LogFile
  -----------------------------------------------------------------------------
*/
tin105_LogFile :: tin105_LogFile(const int nEventsToLog)
                : m_pNextFile(NULL),
                  m_nEventsToLog(nEventsToLog),
                  m_bEnabled(true),
                  m_bDestruct(false)
{

} // end tin105_LogFile :: tin105_LogFile

/* 
  -----------------------------------------------------------------------------
  protected function tin105_LogFile :: setNext
  -----------------------------------------------------------------------------
*/
void tin105_LogFile :: setNext(tin105_LogFile * pLogFile) 
{ 
  m_pNextFile = pLogFile;
} // end tin105_LogFile :: setNext

/* 
  -----------------------------------------------------------------------------
  protected function tin105_LogFile :: next
  -----------------------------------------------------------------------------
*/
tin105_LogFile * tin105_LogFile :: next() const
{ 
  return m_pNextFile;
} // end tin105_LogFile :: next

/* 
  -----------------------------------------------------------------------------
  protected function tin105_LogFile :: setToDestroy
  -----------------------------------------------------------------------------
*/
void tin105_LogFile :: setToDestroy(const bool bDestroy) 
{ 
  m_bDestruct = bDestroy;
} // end tin105_LogFile :: setToDestroy

/* 
  -----------------------------------------------------------------------------
  protected function tin105_LogFile :: isToDestroy
  -----------------------------------------------------------------------------
*/
bool tin105_LogFile :: isToDestroy() const
{ 
  return m_bDestruct;
} // end tin105_LogFile :: isToDestroy

/* 
  -----------------------------------------------------------------------------
  protected function tin105_LogFile :: writeEvents
  -----------------------------------------------------------------------------
*/
bool tin105_LogFile :: writeEvents(const teo200_EventList & EvtLst)
{

  bool                     bReturn = true;

  if ( isEnabled () && 
       ( ( (EvtLst.eo200_EventType() == teo200_EventList::Error_eeo200  ) && 
           ((m_nEventsToLog & LOG_ERR_MIN105) == LOG_ERR_MIN105          )    ) ||
         ( (EvtLst.eo200_EventType() == teo200_EventList::Warning_eeo200) && 
           ((m_nEventsToLog & LOG_WRN_MIN105) == LOG_WRN_MIN105          )    ) ||
         ( (EvtLst.eo200_EventType() == teo200_EventList::Info_eeo200   ) && 
           ((m_nEventsToLog & LOG_INF_MIN105) == LOG_INF_MIN105          )    )    ) ) {
    bReturn = writeOneEventList(EvtLst);
  } // end if

  return bReturn;
} // tin105_LogFile :: writeEvents

/* 
  -----------------------------------------------------------------------------
  implementation class tin105_SequentialLogFile
  -----------------------------------------------------------------------------
*/

#define LOG_TITLE1 _T("-------------------------------------------------------------------------------")
#define LOG_TITLE2 _T("Date  Time       TID(hex) Typ      MsgID Label    Message-Text                 ")
#define LINE_LEN   1024
#define MAX_FILE   (800 * 1024)
#define COPY_LINES (MAX_FILE / 80 / 2)
/* 
  -----------------------------------------------------------------------------
  public constructor tin105_SequentialLogFile :: tin105_SequentialLogFile
  -----------------------------------------------------------------------------
*/
tin105_SequentialLogFile :: tin105_SequentialLogFile
      (const tsp00_Pathc  szFilename, 
       teo200_EventList * pEvtList)
       : tin105_LogFile(LOG_WRN_MIN105 + LOG_ERR_MIN105)
{
  m_szFile = szFilename;
} // end tin105_SequentialLogFile :: tin105_SequentialLogFile
/* 
  -----------------------------------------------------------------------------
*/
tin105_SequentialLogFile :: tin105_SequentialLogFile
      (const tsp00_Pathc  szFilename, 
       const int          nEventsToLog,
       teo200_EventList * pEvtList)
       : tin105_LogFile(nEventsToLog)
{
  m_szFile = szFilename;
} // end tin105_SequentialLogFile :: tin105_SequentialLogFile

/* 
  -----------------------------------------------------------------------------
  protected function tin105_SequentialLogFile :: openFile
  -----------------------------------------------------------------------------
*/
bool tin105_SequentialLogFile :: openFile
      (const tsp00_Pathc  szFilename,
       tsp00_Int4       & hFile) const
{
  FUNCTION_DBG_MIN100("tin105_SequentialLogFile :: openFile");

  tsp5_rte_fileinfo   fileInfo;
  tsp5_rte_fileerror  fileError;

  // check existence of file
  sqlfinfoc (szFilename.asCharp(), &fileInfo, &fileError);
  if (fileError.sp5fe_result != vf_ok) {
    return false;
  } // end if

  // garbage collection
  if (fileInfo.sp5fi_exists && fileInfo.sp5fi_size > MAX_FILE) {
    shrinkFile(szFilename);
  } // end if

  // open
  sqlfopenc (szFilename.asCharp(), SP5VF_TEXT, SP5VF_APPEND, SP5BK_DISKSYNCED, &hFile, &fileError);
  if (fileError.sp5fe_result != vf_ok) {
    return false;
  } // end if

  // write header (if new)
  if (!fileInfo.sp5fi_exists) {
    sqlfwritec (hFile, LOG_TITLE1, -1, &fileError);
    sqlfwritec (hFile, LOG_TITLE2, -1, &fileError);
    sqlfwritec (hFile, LOG_TITLE1, -1, &fileError);
  } // end if

  return true;
} // end tin105_SequentialLogFile :: openFile

/* 
  -----------------------------------------------------------------------------
  protected function tin105_SequentialLogFile :: closeFile
  -----------------------------------------------------------------------------
*/
void tin105_SequentialLogFile :: closeFile
      (const tsp00_Int4 hFile) const
{
  FUNCTION_DBG_MIN100("tin105_SequentialLogFile :: closeFile");

  tsp5_rte_fileerror   fileError;

  sqlfclosec (hFile, SP5VF_CLOSE_NORMAL, &fileError);

} // end tin105_SequentialLogFile :: closeFile

/* 
  -----------------------------------------------------------------------------
  protected function tin105_SequentialLogFile :: shrinkFile
  -----------------------------------------------------------------------------
*/
void tin105_SequentialLogFile :: shrinkFile
      (const tsp00_Pathc  szFilename) const
{
  FUNCTION_DBG_MIN100("tin105_SequentialLogFile :: shrinkFile");

  tsp_int4            hSource;
  tsp_int4            hTarget;
  tsp00_Pathc         szTemp = szFilename;
  tsp00_Pathc         szOrg = szFilename;
  tsp01_RteError      rteError;
  tsp5_rte_fileerror  fileError;
  bool                bState = true;
  tsp_int4            nLine = 0;
  _TCHAR              szBuffer[LINE_LEN];
  tsp_longint         nOutSize;


  strcat(szTemp.asCharp(), ".1");

    if (sqlfilecopyc(szOrg.asCharp(), szTemp.asCharp(), &rteError)) {

    // open source
    sqlfopenc (szTemp.asCharp(), SP5VF_TEXT, SP5VF_READ, SP5BK_BUFFERED, &hSource, &fileError);
    bState = (fileError.sp5fe_result == vf_ok);

    // open target
    if (bState) {
      sqlfopenc (szOrg.asCharp(), SP5VF_TEXT, SP5VF_WRITE, SP5BK_BUFFERED, &hTarget, &fileError);
      bState = (fileError.sp5fe_result == vf_ok);
    } // end if

    // copy file
    while (bState) {
      sqlfreadc (hSource, szBuffer, LINE_LEN, &nOutSize, &fileError);
      bState = (fileError.sp5fe_result == vf_ok);
      ++nLine;
      if (bState && (nLine < 4 || nLine > COPY_LINES)) {
        sqlfwritec (hTarget, szBuffer, -1, &fileError);
        bState = (fileError.sp5fe_result == vf_ok);
      } // end if
    } // end while

    sqlfclosec (hTarget, SP5VF_CLOSE_NORMAL, &fileError);
    bState = (fileError.sp5fe_result == vf_ok);
    sqlfclosec (hSource, SP5VF_CLOSE_NORMAL, &fileError);
    
    // restore source
    if (!bState) {
      sqlfilecopyc(szTemp.asCharp(), szOrg.asCharp(), &rteError);
      sqlferasec(szTemp, &fileError);
    } // end if
    sqlferasec(szTemp, &fileError);

  } // end if
  

} // end tin105_SequentialLogFile :: shrinkFile

/* 
  -----------------------------------------------------------------------------
  protected function tin105_SequentialLogFile :: writeOneEventList
  -----------------------------------------------------------------------------
*/
bool tin105_SequentialLogFile :: writeOneEventList
      ( const teo200_EventList & EvtLst )
{
  const teo200_EventList * pNextEvt = &EvtLst;
  _TCHAR                   szBuffer[LINE_LEN];
  _TCHAR                 * pBuffer;
  bool                     bStatus = true;
  tsp5_rte_fileerror       fileError;
  const _TCHAR           * pDateTime;
  teo200_EventList         fileEvent;
  tsp00_Int4               hFile;
  teo200_EventList         evtDateTime(_T("tin105_SequentialLogFile :: writeOneEventList"), OK_IN100);

  // open file
  if (openFile(m_szFile, hFile)) {

    // local event - holding only log date and time
    pDateTime = evtDateTime.eo200_EventDateTime(false);
 
    while (pNextEvt != NULL) {
      tsp00_Int4 nPid;
      sqlgetpid(&nPid);
      _stprintf(szBuffer, "%-14s %#010x %-3s %10d %-*s ",
                          pDateTime,
                          (int) nPid,
                          pNextEvt->eo200_EventType(false),
                          (int) pNextEvt->eo200_EventID(),
                          MAX_EVENTLABEL_LEN_EO200, pNextEvt->eo200_EventLabel());

      pBuffer = &szBuffer[_tcslen(szBuffer)];
      _tcsncpy(pBuffer, pNextEvt->eo200_EventMsg(), LINE_LEN -_tcslen(szBuffer) - 1);
      szBuffer[LINE_LEN - 1] = '\0';

      pBuffer = strstr(szBuffer, "\n");
      if (pBuffer != NULL) {
        *pBuffer = '\0';
      } // end if

      sqlfwritec (hFile, &szBuffer[0], -1, &fileError);

      bStatus = bStatus && (fileError.sp5fe_result == vf_ok);
    
      pNextEvt  = pNextEvt->eo200_NextEvent();
      pDateTime = _T("");
    } // end while

    closeFile(hFile);
  } // end if
    
  return bStatus;
} // end tin105_SequentialLogFile :: writeOneEventList