/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#if defined(_WIN32)
  #line 3 "d:\\v72\\sys\\src\\in\\vin29.cpp"
#endif
/* @lastChanged: "1998-05-19  7:51"
 * @filename:    vin29.cpp
 * @filename:   vin29.cpp
 * @purpose:    "Interface to SQL packet"
 * @release:    7.1.0.0
 * @see:        ""
 *
 * @copyright:  "Copyright by   SAP AG, 1998"
 */

#include <stdlib.h>
#include <stdarg.h>
#include <ctype.h>
#include <stdio.h>

#include "heo02.h"
#include "heo03.h"
#include "hsp26.h"
#include "hsp100.h"
#include "hin29.h"

#if COMPILEMODE_MSP00 >= QUICK_MSP00
// quick fix for PTS 1104718
#define ASSERT_DBG(cond)        ((cond) ? (void) 0 : \
            sqlk_assert (assertionFailed_esp00, __LINE__, __FILE__, #cond))
#else

#define ASSERT_DBG(cond)        0

#endif

//static const char i29_applicationName [4] = "CON";
//static const char i29_applicationVersion [6] = "70200";

#define SQLTRACE (session->tracer)

externPascal int s30klen (
    const void    *buf,
    char           ch,
    int            len);

externPascal void s40glint (
    tsp00_Number    *num,
    tsp00_Int4       pos,
    int            len,
    tsp00_Int4      &dest,
    tsp00_NumError &res);

externPascal void s41plint (
    tsp00_Number    *num,
    tsp00_Int4       pos,
    int            len,
    int            frac,
    tsp00_Int4       source,
    tsp00_NumError &nrc);

externPascal void s43pstr (
    tsp00_Number    *num,
    tsp00_Int4       pos,
    int            len,
    int            frac,
    char          *source,
    tsp00_Int4       spos,
    int            slen,
    tsp00_NumError &res);

externPascal void s44egchr (
    void          *buf,
    tsp00_Int4       pos,
    integer        len,
    integer        frac,
    void          *dest,
    tsp00_Int4       dpos,
    integer       &dlen,
    tsp00_DecimalPresentation &decimal,
    tsp00_NumError &res);

externPascal void s45stoi4 (
    tsp00_Int4      *val,
    void          *buf,
    int            pos,
    int            len,
    tsp00_NumError &res);

/* ------------------------------- */
/* specification private functions */
/* ------------------------------- */
static void     i29_fetchcmd
    ( tin01_sql_session * session,
      tin01_cstr          fetchCmd,
      int                fetchVars );
/* ------------------------------- */
static void     i29_reset
    ( tin01_sql_session * session,
      tin01_bool          initClientInfo );
/* ------------------------------- */
static void     i29_machinfo
    ( tsp00_CodeType    * code_type,
      tsp00_SwapKind    * swap_kind,
      char             * termid);
/* ------------------------------- */
void            i29pnull
    ( tin01_sql_session * session,
      int                sqllen,
      tsp00_Int4           bufpos);
/* ------------------------------- */
tsp00_Int4        i29resultcount
    ( tin01_sql_session * session);
/* ------------------------------- */
tsp01_CommErr i29receive
    ( tin01_sql_session      * pSession );
/* ------------------------------- */
tsp01_CommErr i29request
    ( tin01_sql_session * session);
/* ------------------------------- */
void            i29callsql
    ( tin01_sql_session * session,
      tsp01_CommErr  * comm_error);
/* ------------------------------- */
void i29pascalstring
      ( char               * target,
        int                  targetlen,
        const char         * source,
        int                  sourcelen );
/* ------------------------------- */
void i29cleanup_session
      ( tin01_sql_session   * session);
/* ------------------------------- */
int i29connect (
    tin01_sql_session * session);
/* ------------------------------- */


#define UNICODE_CLIENT(session) (M_IS_UNICODE((session)->code_type))
/*
 *
 *#define UNICODE_CLIENT(session) (((session)->code_type == CSP_UNICODE)||((session)->code_type == CSP_UNICODE_SWAP))
 */
/**\
--------------------------------------MF__ MOD__ chapter packet handling
\**/

#define EMPTYSET_MIN01                     0
#define M_WARNINGSET(set)               (*(tsp00_Int2*)&set)

tsp1_segment   *
i29firstsegment (
    tsp1_packet * packet)
{
    ROUTINE_DBG_MSP00 ("i29firstsegment");
    return &packet->sp1_segm ();
}

/*----------------------------------------*/

tsp1_segment   *
i29nextsegment (
    tsp1_segment * segment)
{
    ROUTINE_DBG_MSP00 ("i29nextsegment");
    s26next_segment (segment);
    return segment;
}

/*----------------------------------------*/

tsp1_segment   *
i29_lastsegment (
    tsp1_packet * packet)
{
    ROUTINE_DBG_MSP00 ("i29_lastsegment");
    int             segmentCount;
    int             i;
    tsp1_segment   *segment;

    segmentCount = packet->sp1_header.sp1h_no_of_segm;
    segment = i29firstsegment (packet);
    for (i = 0; i < segmentCount - 1; ++i) {
        segment = i29nextsegment (segment);
    }
    return segment;
}

/*----------------------------------------*/

tsp1_segment   *
i29advsegment (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29advsegment");
    session->segment = i29nextsegment (session->segment);
    return session->segment;
}

/*----------------------------------------*/

tsp1_segment   *
i29newsegment (
    tin01_sql_session * session,
    tsp1_cmd_mess_type_Enum messType)
{
    ROUTINE_DBG_MSP00 ("i29newsegment");
    s26first_segment_init (session->send_packet, sp1sk_cmd,
        session->segment);
    session->segment->sp1c_mess_type ().becomes(messType);
    session->segment->sp1c_sqlmode () = tsp1_sqlmode::fromInt (session->sqlMode);
    return session->segment;
}

/*----------------------------------------*/

tsp1_part      *
i29firstpart (
    tsp1_segment * segment)
{
    ROUTINE_DBG_MSP00 ("i29firstpart");
    return &segment->sp1p_part ();
}

/*----------------------------------------*/

tsp1_part      *
i29nextpart (
    tsp1_part * part)
{
    ROUTINE_DBG_MSP00 ("i29nextpart");
    s26nextpart (part);
    return part;
}

/*----------------------------------------*/

tsp1_part      *
i29advpart (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29advpart");
    session->part = i29nextpart (session->part);
    return session->part;
}

/*----------------------------------------*/

tin01_bool
i29findpart (
    tin01_sql_session * session,
    tsp1_part_kind_Enum requiredKind)
{
    ROUTINE_DBG_MSP00 ("i29findpart");
    /* Find Part return it in session->part */
    if (session->part != NULL) {
        if (session->part->sp1p_part_kind () == requiredKind)
            return true;
    }
    s26find_part (*session->segment, requiredKind, session->part);
    return (session->part != NULL);
} /* i29findpart */

/*----------------------------------------*/

tin01_bool
i29_findpart (
    tin01_sql_session * session,
    tsp1_part_kind_Enum requiredKind,
    tsp1_part        ** found_part)
{
    ROUTINE_DBG_MSP00 ("i29_findpart");
    /* Find Part without changing session->part */
    if (session->part != NULL) {
        if (session->part->sp1p_part_kind () == requiredKind) {
            *found_part = session->part;
            return true;
        }
    }
    s26find_part (*session->segment, requiredKind, *found_part);
    return (*found_part != NULL);
} /* i29_findpart */

/*----------------------------------------*/

tsp1_part      *
i29newpart (
    tin01_sql_session * session,
    tsp1_part_kind_Enum partKind)
{
    ROUTINE_DBG_MSP00 ("i29newpart");
    if (session->segment->sp1s_no_of_parts () > 0)
        s26finish_part (session->send_packet, *session->part);
    s26new_part_init (session->send_packet, *session->segment, session->part);
    ASSERT_DBG (session->part != NULL);
    session->part->sp1p_part_kind ().becomes(partKind);
    return session->part;
}

/*----------------------------------------*/

tsp00_Int4
i29packetlen (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29packetlen");
    return s26packet_len (session->send_packet);
}

/*----------------------------------------*/

void
i29closesend (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29closesend");
    ASSERT_DBG (session->buildingCommand);
    s26finish_part (session->send_packet, *session->part);
    session->send_packet->sp1_header.sp1h_varpart_len = i29packetlen (session) -  
                                                        sizeof (session->send_packet->sp1_header);
}

/*----------------------------------------*/

void
i29resetpackets (
    tin01_sql_session * session,
    tin01_bool forCommand)
{
    ROUTINE_DBG_MSP00 ("i29resetpackets");
    tsp1_packet    *packet;

    session->buildingCommand = forCommand;
    if (forCommand)
        packet = session->send_packet;
    else
        packet = session->rec_packet;
    session->segment = &packet->sp1_segm ();
    if (forCommand) {
        session->part = &session->segment->sp1p_part ();
    }
    else {
        session->part = NULL;
    }
}

/*----------------------------------------*/

void
i29clearpacketref (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29clearpacketref");
    session->send_packet = NULL;
    session->rec_packet = NULL;
    session->segment = NULL;
    session->part = NULL;
}

/*----------------------------------------*/

void
i29initpacketref (
    tin01_sql_session * session,
    tsp1_packet * packet,
    tin01_bool forCommand)
{
    ROUTINE_DBG_MSP00 ("i29initpacketref");
    session->buildingCommand = forCommand;
    if (packet == NULL)
        i29clearpacketref (session);
    else {
        if (forCommand) {
            session->send_packet = packet;
            session->rec_packet = NULL;
        }
        else
            session->rec_packet = packet;
        session->segment = i29firstsegment (packet);
        session->part = i29firstpart (session->segment);
    }
}

/*----------------------------------------*/

void
i29initsession (
    tin01_sql_session * session,
    tsp1_packet * packet)
{
    ROUTINE_DBG_MSP00 ("i29initsession");

    memset (&session->xuser, '\0', sizeof (session->xuser));
    session->xuser.xu_cachelimit = -1;
    session->xuser.xu_timeout = -1;
    session->xuser.xu_isolation = -1;
    session->reference = 0;
    session->is_connected = false;
    session->inUse = false;
    session->buildingCommand = false;
    i29_machinfo (&session->code_type, &session->swap_kind, session->senderid);
    i29initpacketref (session, packet, true);
    session->sqlMode = sp1sm_session_sqlmode;
    session->more_data = NULL;
    session->switch_possible = true;
    session->new_session = true;
    session->as_utility = false;
}

/*----------------------------------------*/

void
i29unicodeclient (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29unicodeclient");
    ASSERT_DBG (!session->is_connected);
    session->code_type = csp_unicode;
}

/*----------------------------------------*/

int
i29adabasmode (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29adabasmode");
    int             result;

    result = session->sqlMode;
    session->sqlMode = sp1sm_adabas;
    return result;
}

/*----------------------------------------*/

int
i29sessionmode (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29sessionmode");
    int             result;

    result = session->sqlMode;
    session->sqlMode = sp1sm_session_sqlmode;
    return result;
}

/*----------------------------------------*/

int
i29oldmode (
    tin01_sql_session * session,
    int oldMode)
{
    ROUTINE_DBG_MSP00 ("i29oldmode");
    int             result;

    result = session->sqlMode;
    session->sqlMode = oldMode;
    return result;
}

/**\
--------------------------------------MF__ MOD__ chapter high level sql
\**/

void
i29initadbs (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29initadbs");
    i29_reset (session, true);
    i29newsegment (session, sp1m_dbs);
    i29newpart (session, sp1pk_command);
}

/*----------------------------------------*/

void
i29initparse (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29initparse");
    i29_reset (session, true);
    i29newsegment (session, sp1m_parse);
    i29newpart (session, sp1pk_command);
}

/*----------------------------------------*/

void
i29initexecute (
    tin01_sql_session * session,
    tin01_parseid pid)
{
    ROUTINE_DBG_MSP00 ("i29initexecute");
    i29_reset (session, true);
    i29newsegment (session, sp1m_execute);
    i29newpart (session, sp1pk_parsid);
    memcpy (session->part->sp1p_buf ().asCharp () + session->part->sp1p_buf_len (),
        pid, sizeof (tin01_parseid));
    session->part->sp1p_buf_len () += sizeof (tin01_parseid);
    /* i29newpart (session, sp1pk_data); */
}

/*----------------------------------------*/

void
i29initspecial (
    tin01_sql_session * session,
    tsp1_cmd_mess_type_Param messType)
{
    ROUTINE_DBG_MSP00 ("i29initspecial");
    i29_reset (session, true);
    i29newsegment (session, messType);
    i29newpart (session, sp1pk_command);
}

/*----------------------------------------*/

void
i29initoutargs (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29initoutargs");
    i29_reset (session, false);
    s26first_segment_init (session->send_packet, sp1sk_return,
        session->segment);
    i29newpart (session, sp1pk_data);
}

/*----------------------------------------*/
static int
i29sqlrequest (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29sqlrequest");
  tsp00_Int2         result       = cin01_db_ok;

  if (!session->is_connected) {
    TRACE_MIN01 (4, SQLTRACE, "No connection: command skipped");
    result = cin01_db_not_accessible;
    return result;
  } /* end if */

  ASSERT_DBG (session->is_connected);
  session->rec_packet = NULL;
  i29closesend (session);
  IFTRACE_MIN01 (4, j96dumppacket (session, "REQUEST", SQLTRACE));

  if (i29request (session) != commErrOk_esp01) {
    result = cin01_db_not_accessible;
  } /* end if */

  return result;
} /* end i29sqlrequest */

/*----------------------------------------*/
static int
i29sqlreceive (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29sqlreceive");
  tsp01_CommErr  comm_error;
  tsp00_Int2         result       = cin01_db_ok;

  comm_error = i29receive (session);
  i29lasterr_on (session);

  if ((int) comm_error != commErrOk_esp01) {
    result = cin01_db_not_accessible;
  } else {
    i29resetpackets (session, false);
    IFTRACE_MIN01 (4, j96dumppacket (session, "RECEIVE", SQLTRACE));
    result = i29_lastsegment (session->rec_packet)->sp1r_returncode ();

//    result = RECINFO (i29_lastsegment (session->rec_packet), returncode);
  } /* end if */

  return result;

} /* end i29sqlreceive */

/*----------------------------------------*/

int
i29sql (
    tin01_sql_session * session,
    tin01_sqlresult * sqlresult)
{
    ROUTINE_DBG_MSP00 ("i29sql");
    tsp01_CommErr  comm_error;
    tsp00_Int2        result;
    tsp1_packet    *send_packet = session->send_packet;
    tsp1_segment   *segment;
    int             msglen;

    if (sqlresult == NULL)
        sqlresult = &(session->lasterr.sqlresult);
    if (!session->is_connected) {
        TRACE_MIN01 (4, SQLTRACE, "No connection: command skipped");
        result = cin01_db_not_accessible;
        if (sqlresult != NULL) {
            sqlresult->returnCode = result;
            M_WARNINGSET (sqlresult->warnings) = EMPTYSET_MIN01;
            sqlresult->errorPos = 0;
            sqlresult->rowCount = 0;
            MEMCPY_MIN01 (sqlresult->sqlstate, "I8888");
            sqlresult->sqlmsg[0] = '\0';
        }
        return result;
    }
    ASSERT_DBG (session->is_connected);
    session->rec_packet = NULL;
    i29closesend (session);
    IFTRACE_MIN01 (4, j96dumppacket (session, "REQUEST", SQLTRACE));
    i29callsql (session, &comm_error);
    if ((int) comm_error != commErrOk_esp01) {
        result = cin01_db_not_accessible;
        /* session->is_connected = false; lower level routines try reconnect */
        if (sqlresult != NULL) {
            sqlresult->returnCode = result;
            M_WARNINGSET (sqlresult->warnings) = EMPTYSET_MIN01;
            sqlresult->errorPos = 0;
            sqlresult->rowCount = 0;
            MEMCPY_MIN01 (sqlresult->sqlstate, "I8888");
            sqlresult->sqlmsg[0] = '\0';
        }
    }
    else {
        i29resetpackets (session, false);
        IFTRACE_MIN01 (4, j96dumppacket (session, "RECEIVE", SQLTRACE));
        segment = i29_lastsegment (session->rec_packet);
        result = segment->sp1r_returncode ();
        /* to lasterr, only store errors: */
        if ((sqlresult != NULL) &&
            ((result != cin01_db_ok) ||
                (sqlresult != &(session->lasterr.sqlresult)))) {
            sqlresult->returnCode = result;
            M_WARNINGSET(sqlresult->warnings) = M_WARNINGSET
                (segment->sp1r_extern_warning ());
            sqlresult->errorPos = segment->sp1r_errorpos ();
            if (result == cin01_db_row_not_found) {
                sqlresult->rowCount = 0;
            }
            else {
                sqlresult->rowCount = i29resultcount (session);
            }
            MEMCPY_MIN01 (sqlresult->sqlstate, segment->sp1r_sqlstate ());
            if (result != cin01_db_ok) {
                if (i29findpart (session, sp1pk_errortext)) {
                    memcpy (sqlresult->sqlmsg, session->part->sp1p_buf ().asCharp (),
                        sizeof (sqlresult->sqlmsg) - 1);
                    msglen = session->part->sp1p_buf_len ();
                    sqlresult->sqlmsg[msglen] = '\0';
                }
            }
        }
    }
    return result;
}

/*----------------------------------------*/

static void
i29_reset (
    tin01_sql_session * session,
    tin01_bool initClientInfo)
{
    ROUTINE_DBG_MSP00 ("i29_reset");
    tsp1_packet_header *header;
    /*
     * init packet
     */
    if (!session->buildingCommand)
        i29resetpackets (session, true);
    header = &session->send_packet->sp1_header;
    if (initClientInfo) {
        header->sp1h_mess_code = session->code_type;
        header->sp1h_mess_swap = session->swap_kind;
        header->sp1h_filler2 = 0;

        memcpy (header->sp1h_appl_version, session->senderid,
            sizeof (session->senderid));
    } // end if
    header->sp1h_filler1 = 0;
    header->sp1h_no_of_segm = 1;
}

/*----------------------------------------*/

void
i29pcmd (
    tin01_sql_session * session,
    tin01_cstr cmd,
    int len)
{
    ROUTINE_DBG_MSP00 ("i29pcmd");
    tsp1_packet    *send_packet = session->send_packet;
    char           *target;

    ASSERT_DBG (send_packet != NULL);
    if (len == UNDEF_SP00)
        len = (int)strlen (cmd);
    target = session->part->sp1p_buf () + session->part->sp1p_buf_len ();
    memcpy (target, cmd, len);
    session->part->sp1p_buf_len () += len;
}

/*----------------------------------------*/

void
i29pcmdf (
    tin01_sql_session * session,
    tin01_cstr cmd,
    ...)
{
    ROUTINE_DBG_MSP00 ("i29pcmdf");
    va_list         args;
    int             bytecnt;
    char           *cmdBuf;
    tsp1_packet    *send_packet = session->send_packet;

    ASSERT_DBG (send_packet != NULL);
    va_start (args, cmd);
    cmdBuf = session->part->sp1p_buf () + session->part->sp1p_buf_len ();
    bytecnt = vsprintf (cmdBuf, cmd, args);
    va_end (args);
    if (bytecnt < 0)
        return;
    session->part->sp1p_buf_len () += bytecnt;
}

/*----------------------------------------*/

void
i29pusercmd (
    tin01_sql_session * session,
    tin01_cstr cmd,
    int len)
{
    ROUTINE_DBG_MSP00 ("i29pusercmd");
    tsp1_packet    *send_packet = session->send_packet;

    ASSERT_DBG (send_packet != NULL);
    if (len == UNDEF_SP00)
        len = (int)strlen (cmd);
    memcpy (session->part->sp1p_buf () + session->part->sp1p_buf_len (),
        cmd, len);
    session->part->sp1p_buf_len () += len;
}

/*----------------------------------------*/

static void
i29_inc_buflen (
    tin01_sql_session * session,
    tsp00_Int4 bufpos,
    int sqllen)
{
    ROUTINE_DBG_MSP00 ("i29_inc_buflen");
    int newlen = bufpos + sqllen;

    /* old: session->part->sp1p_buf_len () += sqllen + 1; */
    if (session->part->sp1p_buf_len () < newlen)
        session->part->sp1p_buf_len () = newlen;
}

/*----------------------------------------*/

int
i29parg (
    tin01_sql_session * session,
    const tin01_byte * arg,
    int varlen,  /* in bytes */
    int sqllen,  /* in bytes */
    tsp00_Int4 bufpos,
    char defByte)
{
    ROUTINE_DBG_MSP00 ("i29parg");
    int             mvlen;
    tin01_byte      *bufaddr;
    int             maxlen;
    tsp1_packet    *send_packet = session->send_packet;

    if (arg == NULL) {
        i29pnull (session, sqllen, bufpos);
        return cin01_db_ok;
    }
    --sqllen;                   /* because of defined byte */
    ASSERT_DBG (send_packet != NULL);
    /*
     * IFTRACE_MIN01(4, j99tprintf (SQLTRACE, "
     * arg: "PFMT"; varlen: %d; sqllen: %d; bufpos: %ld; %s\n",
     * arg, varlen, sqllen, bufpos, ischar?"is char":"is tin01_byte" ));
     */
    bufaddr = (tin01_byte *) (session->part->sp1p_buf () + bufpos - 1);
    *bufaddr++ = defByte;
    maxlen = sqllen;
    if (varlen < sqllen)
        mvlen = varlen;
    else
        mvlen = sqllen;
    memcpy (bufaddr, arg, mvlen);

    if (maxlen > mvlen) {
        memset (bufaddr + mvlen, defByte, maxlen - mvlen);
    }
    i29_inc_buflen (session, bufpos, sqllen);
    return cin01_db_ok;
}

/*----------------------------------------*/

void
i29pnull (
    tin01_sql_session * session,
    int sqllen,
    tsp00_Int4 bufpos)
{
    ROUTINE_DBG_MSP00 ("i29pnull");
    tin01_byte           *bufaddr;
    tsp1_packet    *send_packet = session->send_packet;

    ASSERT_DBG (send_packet != NULL);
    bufaddr = (tin01_byte *) (session->part->sp1p_buf () + bufpos - 1);
    *bufaddr = (tin01_byte) csp_undef_byte;
#if !defined (FAST)
    memset (bufaddr + 1, '\0', sqllen - 1);
#endif
    i29_inc_buflen (session, bufpos, sqllen);
}

/*----------------------------------------*/

void
i29ppw (
    tin01_sql_session * session,
    tsp00_CryptPw pw,
    char defbyte)
{
    ROUTINE_DBG_MSP00 ("i29ppw");
    tin01_byte           *bufaddr;
    tsp1_packet    *send_packet = session->send_packet;

    ASSERT_DBG (send_packet != NULL);
    bufaddr = session->part->sp1p_buf () + session->part->sp1p_buf_len ();
    *bufaddr = defbyte;
    memcpy (bufaddr + 1, pw, sizeof (tsp00_CryptPw));
    session->part->sp1p_buf_len () += sizeof (tsp00_CryptPw) + 1;
}

/*----------------------------------------*/

void
i29gparseid (
    tin01_sql_session * session,
    tin01_parseid pid)
{
    ROUTINE_DBG_MSP00 ("i29gparseid");
    tsp1_packet    *rec_packet = session->rec_packet;
    tin01_bool           found;

    ASSERT_DBG (rec_packet != NULL);
    found = i29findpart (session, sp1pk_parsid);
    if (found) {
        memcpy (pid, session->part->sp1p_buf (), sizeof (tin01_parseid));
    }
    else {
        memset (pid, '\0', sizeof (tin01_parseid));
        pid[mxin1_parseid - 2] = csp1_p_use_adbs;
    }
}

/*----------------------------------------*/

int
i29paramcount (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29paramcount");
    if (i29findpart (session, sp1pk_shortinfo))
        return session->part->sp1p_arg_count ();
    else {
        /*
         * currently, no hostvars -> no shortinfo part
         */
        return 0;
    }
}

/*----------------------------------------*/

tsp1_param_info *
i29gparaminfo (
    tin01_sql_session * session,
    int i)
{
    ROUTINE_DBG_MSP00 ("i29gparaminfo");
    tsp1_param_info *paramInfo;

    ASSERT_DBG (session->rec_packet != NULL);
    // ASSERT_DBG (j02between (i, 0, i29paramcount (session) - 1));
    i29findpart (session, sp1pk_shortinfo);
    paramInfo = (tsp1_param_info *) (void*) session->part->sp1p_buf ();
    return paramInfo + i;
}

/*----------------------------------------*/

tsp00_Int4
i29nameslen (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29nameslen");
    i29findpart (session, sp1pk_columnnames);
    return session->part->sp1p_buf_len ();
}

/*----------------------------------------*/

tin01_cstr
i29colnames (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29colnames");
    i29findpart (session, sp1pk_columnnames);
    return session->part->sp1p_buf ().asCharp ();
}

/*----------------------------------------*/

tin01_bool
i29garg (
    tin01_sql_session * session,
    tin01_byte * arg,
    int varlen,
    int sqllen,
    tsp00_Int4 bufpos)
{
    ROUTINE_DBG_MSP00 ("i29garg");
    int             mvlen;
    unsigned char  *bufaddr;
    tsp1_packet    *rec_packet = session->rec_packet;

    ASSERT_DBG (rec_packet != NULL);
    if (varlen < sqllen)
        mvlen = varlen;
    else
        mvlen = sqllen;
    bufaddr = (unsigned char *) (session->part->sp1p_buf () + bufpos - 1);
    if (*bufaddr == (unsigned char) csp_undef_byte)
        return false;
    memcpy (arg, bufaddr + 1, mvlen);
    return true;
}

/*----------------------------------------*/

tin01_byte           *
i29argaddr (
    tin01_sql_session * session,
    tsp00_Int4 bufpos)
{
    ROUTINE_DBG_MSP00 ("i29argaddr");
    unsigned char  *bufaddr;
    tin01_byte           *byteaddr;
    tin01_byte           *result;
    tsp1_packet    *rec_packet = session->rec_packet;

    ASSERT_DBG (rec_packet != NULL);
    ASSERT_DBG (session->part->sp1p_part_kind () == sp1pk_data);
    byteaddr = session->part->sp1p_buf () + bufpos - 1;
    bufaddr = (unsigned char *) byteaddr;
    if (*bufaddr == (unsigned char) csp_undef_byte)
        result = NULL;
    else
        result = byteaddr + 1;
    return result;
}

/*----------------------------------------*/

tsp00_Int4
i29resultcount (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29resultcount");
    tsp00_Int4        result = UNDEF_SP00;
    tin01_byte      *number;
    tsp00_NumError   res;
    tsp1_part      *part_ptr;

    if (i29_findpart (session, sp1pk_resultcount, &part_ptr)) {
        number = part_ptr->sp1p_buf () + 1;
        s40glint ((tsp00_Number*)number, 1,
                (part_ptr->sp1p_buf_len () - 2) * 2, result, res);
    }
    return result;
} /* i29resultcount */

/*----------------------------------------*/

void
i29_machinfo (
    tsp00_CodeType *code_type,
    tsp00_SwapKind *swap_kind,
    char          *termid)
{
    ROUTINE_DBG_MSP00 ("i29_machinfo");
    tsp00_Int4        i234;
    char             *c234;
    tsp100_VersionID0 VersionID0;

    *code_type = csp_ascii;
    i234 = 1;
    c234 = (char *) &i234;
    if (c234[3] == 1)
        (*swap_kind).becomes(sw_normal);
    else if (c234[0] == 1)
        (*swap_kind).becomes(sw_full_swapped);
    else
        (*swap_kind).becomes(sw_part_swapped);

    sp100_GetVersionID ( VersionIDType0_esp100, s100buildnumberEx, &VersionID0 );

    sprintf(termid, "%01d%02d%02d",
                    (int) VersionID0.MajorVersion_sp100,
                    (int) VersionID0.MinorVersion_sp100,
                    (int) VersionID0.CorrLevel_sp100);
    memcpy (termid + sizeof (tsp00_C5), csp_comp_db_manager, sizeof (tsp00_C3));
}

/*----------------------------------------*/

int
i29adbs (
    tin01_sql_session * session,
    tin01_cstr cmd)
{
    ROUTINE_DBG_MSP00 ("i29adbs");
    i29initadbs (session);
    session->segment->sp1c_with_info () = true;
    i29pcmd (session, cmd, UNDEF_SP00);
    return i29sql (session, NULL);
}

/*----------------------------------------*/

int
i29adbs_noinfo (
    tin01_sql_session * session,
    tin01_cstr cmd)
{
    ROUTINE_DBG_MSP00 ("i29adbs_noinfo");
    i29initadbs (session);
    i29pcmd (session, cmd, UNDEF_SP00);
    return i29sql (session, NULL);
}

/*----------------------------------------*/

int
i29adbsf (
    tin01_sql_session * session,
    tin01_cstr cmd,
    ...)
{
    ROUTINE_DBG_MSP00 ("i29adbsf");
    va_list         args;
    int             bytecnt;
    tsp1_packet    *send_packet = session->send_packet;

    ASSERT_DBG (send_packet != NULL);
    va_start (args, cmd);
    i29initadbs (session);
    bytecnt = vsprintf (session->part->sp1p_buf ().asCharp (), cmd, args);
    va_end (args);
    if (bytecnt < 0)
        return UNDEF_SP00;
    session->part->sp1p_buf_len () = bytecnt;
    return i29sql (session, NULL);
}

/*----------------------------------------*/

int
i29adbsinfof (
    tin01_sql_session * session,
    tin01_cstr cmd,
    ...)
{
    ROUTINE_DBG_MSP00 ("i29adbsinfof");
    va_list         args;
    int             bytecnt;
    tsp1_packet    *send_packet = session->send_packet;

    ASSERT_DBG (send_packet != NULL);
    va_start (args, cmd);
    i29initadbs (session);
    session->segment->sp1c_with_info () = true;
    bytecnt = vsprintf (session->part->sp1p_buf ().asCharp (), cmd, args);
    va_end (args);
    if (bytecnt < 0)
        return UNDEF_SP00;
    session->part->sp1p_buf_len () = bytecnt;
    return i29sql (session, NULL);
}

/*----------------------------------------*/

int
i29adbs_describe (
    tin01_sql_session * session,
    tin01_parseid       pid)
{
    ROUTINE_DBG_MSP00 ("i29adbs_describe");
    int             bytecnt;
    tsp1_packet    *send_packet = session->send_packet;

    TRACEF_MIN01 (7, (SQLTRACE, "i29adbs_describe\n"));
    ASSERT_DBG (send_packet != NULL);
    i29initadbs (session);
    bytecnt = sprintf (session->part->sp1p_buf ().asCharp (), "%s", "DESCRIBE ");

    if (bytecnt < 0)
        return UNDEF_SP00;
    session->part->sp1p_buf_len () = bytecnt;

    i29newpart (session, sp1pk_data);
    memcpy (session->part->sp1p_buf () + session->part->sp1p_buf_len (),
        pid, sizeof (tin01_parseid));
    session->part->sp1p_buf_len () += sizeof (tin01_parseid);
    TRACEF_MIN01 (7, (SQLTRACE, "<-i29adbs_describe rc : i29sql\n"));
    return i29sql (session, NULL);
}   /* i29adbs_describe */

/*----------------------------------------*/


#define c_max_try  5

static int
i29_parse (
    tin01_sql_session * session,
    tin01_cstr cmd,
    tin01_c_sqlca * sqlca,
    tin01_bool internal)
{
    ROUTINE_DBG_MSP00 ("i29_parse");
    int             rc;
    int             paramCount;

    if (sqlca->inUse)
        return cin01_db_ok;
    i29initparse (session);
    if (internal)
        session->segment->sp1c_producer ().becomes(sp1pr_internal_cmd);
    i29pcmd (session, cmd, UNDEF_SP00);
    rc = i29sql (session, NULL);
    if (rc == cin01_db_ok) {
        i29gparseid (session, sqlca->pid);
        paramCount = i29paramcount (session);
        if (paramCount > sqlca->varCount)
            paramCount = sqlca->varCount;
/*            rc = -1200; */        /* too few values */
/*        else { */
            sqlca->paramCount = paramCount;
            if (paramCount > 0) {
                tsp1_param_info *pinfo = i29gparaminfo (session, 0);
                int              i;

                for (i = 0; i < paramCount; ++i, ++pinfo) {
                    memcpy (&sqlca->hostVar [i].pinfo, pinfo,
                            sizeof (tsp1_param_info));
                }
            }
            sqlca->inUse = true;
/*        } */
    }
    return rc;
}

/*----------------------------------------*/

static          int
i29_put_num_string (
    tin01_sql_session * session,
    char *arg,
    tsp1_param_info * paramInfo)
{
    ROUTINE_DBG_MSP00 ("i29_put_num_string");
    tsp00_NumError   res;
    int             frac;

    frac = paramInfo->sp1i_frac;
    switch (paramInfo->sp1i_data_type) {
        case dfloat:
        case dvfloat:
            frac = -1;
            /* fall through */
        case dinteger:
        case dsmallint:
        case dfixed:
            {
                tsp00_Number num;

                s43pstr (&num, 1, paramInfo->sp1i_length, frac,
                    arg, 1, (int)strlen(arg), res);
                if (res != num_ok)
                    return cin01_db_invalid_number;
                return i29parg (session, num,
                    paramInfo->sp1i_in_out_len,
                    paramInfo->sp1i_in_out_len,
                    paramInfo->sp1i_bufpos, csp_defined_byte);
            }
     default:
            SWITCH_ERROR_DBG_MIN00 ("data type %d not yet implemented",
                paramInfo->sp1i_data_type );
            break;
    }
    return cin01_db_ok;
}

/*----------------------------------------*/

static          int
i29_put_bool_string (
    tin01_sql_session * session,
    char             * arg,
    int                datalen,
    tsp1_param_info  * paramInfo)
{
    ROUTINE_DBG_MSP00 ("i29_put_bool_string");
   char            BoolVal;
   tsp00_C20         Temp;
   tsp00_Int4        NumVal;
   tsp00_NumError   res;
   int             i;

   if (datalen == UNDEF_SP00)
      datalen = (int)strlen(arg);
   strncpy (Temp, arg, datalen);
   Temp [datalen] = '\0';
   for (i = 0; i < datalen; ++i) {
       Temp [i] = (char) toupper (Temp [i]);
   }

   if (strcmp (Temp, "true") == 0)
      BoolVal = '\1';
   else if (strcmp (Temp, "false") == 0)
      BoolVal = '\0';
   else {
       s45stoi4 (&NumVal, arg, 1, (int)strlen (arg), res);
       if (res != num_ok)
          return cin01_db_invalid_number;
       if (NumVal == 0)
          BoolVal = '\0';
      else
          BoolVal = '\1';
   }
   return i29parg (session, &BoolVal, 1, paramInfo->sp1i_in_out_len,
      paramInfo->sp1i_bufpos, csp_defined_byte);
}

/*----------------------------------------*/

static          int
i29_inarg (
    tin01_sql_session * session,
    tin01_c_hostvar * hostVar)
{
    ROUTINE_DBG_MSP00 ("i29_inarg");
    int             datalen;
    void           *dataptr;
    unsigned char   defByte = csp_undef_byte;
    tsp1_param_info *paramInfo = &hostVar->pinfo;

    if (hostVar->addr == NULL) {
        i29pnull (session, paramInfo->sp1i_length, paramInfo->sp1i_bufpos);
        return cin01_db_ok;
    }
    switch (paramInfo->sp1i_data_type) {
        case dfixed:
        case dfloat:
        case dvfloat:
        case dsmallint:
        case dinteger:
            if (hostVar->c_type != cin01_c_charp) {
                SWITCH_ERROR_DBG_MIN00 ("Numeric sql type %d not implemented "
                    "for nonstring C type ", paramInfo->sp1i_data_type);
                break;
            }
            return i29_put_num_string (session, (char*)hostVar->addr, paramInfo);
            break;
       case dboolean:
           /* put boolean value */
            if (hostVar->c_type != cin01_c_charp) {
                SWITCH_ERROR_DBG_MIN00 ("%s not implemented "
                    "for nonstring C type ", "boolean");
                break;
           }
           return i29_put_bool_string
                      (session, (char*)hostVar->addr, (int) hostVar->len, paramInfo);
           break;
        case dchb:
        case dvarcharb:
           defByte = csp_defined_byte;
           /* fall through */
        case dcha:
        case dche:
        case dvarchara:
        case dvarchare:
        case ddate:
        case dtime:
        case dtimestamp:
        case dunicode:
            /* put string value */
            if (hostVar->c_type == cin01_c_charpp)
                dataptr = *(char **) hostVar->addr;
            else
                dataptr = hostVar->addr;
            if (hostVar->len >= 0)
                datalen = (int) hostVar->len;
            else {
                datalen = (int)strlen ((char*)dataptr);
            }
            if (defByte == csp_undef_byte) {
            /* sp1_data_type is not a BYTE type  */
                defByte = csp_defined_byte;
            }
             return i29parg (session, (char*)dataptr, datalen,
                paramInfo->sp1i_in_out_len,
                paramInfo->sp1i_bufpos, defByte);
             break;
        default:
            SWITCH_ERROR_DBG_MIN00 ("sql type not supported %d",
                (int) paramInfo->sp1i_data_type);
            break;
    }
    return cin01_db_ok;
}

/*----------------------------------------*/

static int
i29_inargs (
    tin01_sql_session * session,
    tin01_c_sqlca * sqlca)
{
    ROUTINE_DBG_MSP00 ("i29_inargs");
    int             i;
    tin01_bool           hasDataPart = false;
    int             rc;

    i29initexecute (session, sqlca->pid);
    for (i = 0; i < sqlca->paramCount; ++i) {
        if (sqlca->hostVar [i].pinfo.sp1i_io_type != sp1io_output) {
            if (!hasDataPart) {
                i29newpart (session, sp1pk_data);
                hasDataPart = true;
            }
            rc = i29_inarg (session, &sqlca->hostVar[i]);
            if (rc != cin01_db_ok)
                return rc;
        }
    }
    return cin01_db_ok;
}

/*----------------------------------------*/

static          tsp00_Int4
i29_get_num_param (
    void *addr,
    tsp1_param_info * paramInfo)
{
    ROUTINE_DBG_MSP00 ("i29_get_num_param");
    tsp00_NumError   res;
    tsp00_Int4        resval;

    s40glint ((tsp00_Number*)addr, 1,
        paramInfo->sp1i_in_out_len,
        resval, res);
    return resval;
}

/*----------------------------------------*/

static void
i29_num_param_to_string (
    char *DestStr,
    void *addr,
    tsp1_param_info * paramInfo)
{
    ROUTINE_DBG_MSP00 ("i29_num_param_to_string");
    int len;
    tsp00_DecimalPresentation decimal;
    tsp00_NumError res;

    decimal.thousand_token = 'N';
    decimal.zero_point = '.';
    s44egchr (addr, 1,
        paramInfo->sp1i_in_out_len,
        paramInfo->sp1i_frac, DestStr, 1, len,
        decimal, res);
    DestStr [len] = '\0';
}

/*----------------------------------------*/

tin01_byte           *
i29_longargaddr (
    tin01_sql_session * session,
    tsp1_param_info * paramInfo)
{
    ROUTINE_DBG_MSP00 ("i29_longargaddr");
    unsigned char  *bufaddr;
    tsp00_LongDescriptor *descriptor;
    tin01_byte           *byteaddr;
    tin01_byte           *result;
    tsp1_packet    *rec_packet = session->rec_packet;

    ASSERT_DBG (rec_packet != NULL);
    ASSERT_DBG (paramInfo->sp1i_length == sizeof (tsp00_LongDescriptor));
    ASSERT_DBG (session->part->sp1p_part_kind () == sp1pk_data);
    byteaddr = session->part->sp1p_buf () + paramInfo->sp1i_bufpos - 1;
    bufaddr = (unsigned char *) byteaddr;
    if (*bufaddr == (unsigned char) csp_undef_byte)
        return NULL;
    descriptor = (tsp00_LongDescriptor *)(byteaddr + 1);
    result = session->part->sp1p_buf () + descriptor->ld_valpos () - 1;
    paramInfo->sp1i_length = descriptor->ld_vallen();
    return result;
}

/*----------------------------------------*/

static int
i29_get_stringvalue (
    tin01_sql_session    * session,
    tsp1_param_info     * paramInfo,
    tin01_c_hostvar      * hostVar,
    void                * addr)
{
    ROUTINE_DBG_MSP00 ("i29_get_stringvalue");
    int       datalen; /* bytewise */
    char    * hostbuf;

    TRACEF_MIN01 (5, (session->tracer, "i29_get_stringvalue; input value:\n"));
    IFTRACE_MIN01 (5, j99hexbuf (addr, 0,
        paramInfo->sp1i_length * sizeof (tin01_unichar),
        10, session->tracer));
    datalen = s30klen (addr, ' ', paramInfo->sp1i_length);
    if (hostVar->c_type == cin01_c_charpp) {
        hostbuf = (char *) malloc (datalen + 1);
        ASSERT_DBG (NOT_NULL_MIN01 (hostbuf));
        *((void **) hostVar->addr) = hostbuf;
    }
    else
        hostbuf = (char*) hostVar->addr;
    memcpy (hostbuf, addr, datalen);
    hostbuf[datalen] = '\0';
    return cin01_db_ok;
}

/*----------------------------------------*/

static int
i29_outarg (
    tin01_sql_session * session,
    tin01_c_hostvar * hostVar)
{
    ROUTINE_DBG_MSP00 ("i29_outarg");
    void           *addr;
    char           *hostbuf;
    tsp1_param_info * paramInfo = &hostVar->pinfo;
    int             rc = cin01_db_ok;

    switch (paramInfo->sp1i_data_type) {
        case dstra:
        case dstre:
        case dstrb:
        case dstrdb:
            addr = i29_longargaddr (session, paramInfo);
            break;
        default:
            addr = i29argaddr (session, paramInfo->sp1i_bufpos);
            break;
    }
    if (addr == NULL) {
        hostVar->is_null = true;
        return rc;
    }
    else
        hostVar->is_null = false;
    switch (paramInfo->sp1i_data_type) {
        case dfixed:
        case dfloat:
        case dvfloat:
        case dsmallint:
        case dinteger:
            switch (hostVar->c_type) {
                case cin01_c_short:
                    *((short *) hostVar->addr) =
                        (short) i29_get_num_param (addr, paramInfo);
                    break;
                case cin01_c_ushort:
                    *((unsigned short *) hostVar->addr) =
                        (unsigned short) i29_get_num_param (addr, paramInfo);
                    break;
                case cin01_c_int:
                    *((int *) hostVar->addr) =
                        (int) i29_get_num_param (addr, paramInfo);
                    break;
                case cin01_c_uint:
                    *((unsigned int *) hostVar->addr) =
                        (unsigned int) i29_get_num_param (addr, paramInfo);
                    break;
                case cin01_c_long:
                    *((long *) hostVar->addr) =
                        i29_get_num_param (addr, paramInfo);
                    break;
                case cin01_c_ulong:
                    *((unsigned long *) hostVar->addr) =
                        (unsigned long) i29_get_num_param (addr, paramInfo);
                    break;
                case cin01_c_charp:
                    i29_num_param_to_string ((char*) hostVar->addr, addr, paramInfo);
                    break;
                default:
                    SWITCH_ERROR_DBG_MIN00 ("FIXED to hostvar type %d not supported",
                        (int) hostVar->c_type);
                    break;
            }
            break;
        case dcha:
        case dche:
        case dchb:
        case dunicode:
        case dvarchara:
        case dvarchare:
        case dvarcharb:
        case ddate:
        case dtime:
        case dtimestamp:
        case dstra:
        case dstre:
        case dstrb:
        case dstrdb:
            rc = i29_get_stringvalue (session, paramInfo, hostVar, addr);
            break;
        case dboolean:
           {
               const char *result;
               hostbuf = (char*) hostVar->addr;
               if (hostVar->c_type == cin01_c_charp) {
                   if (*((char *)addr))
                       result = "true";
                   else
                       result = "false";

                   strcpy (hostbuf, result);
               }
               else {
                   SWITCH_ERROR_DBG_MIN00 ("boolean not supported for C type %d",
                       (int) hostVar->c_type);
               }
               break;
           }
        default:
            SWITCH_ERROR_DBG_MIN00 ("sql type not supported %d",
                (int) paramInfo->sp1i_data_type);
            break;
    }
    return rc;
}

/*----------------------------------------*/

static int
i29_outargs (
    tin01_sql_session * session,
    tin01_c_sqlca * sqlca)
{
    ROUTINE_DBG_MSP00 ("i29_outargs");
    int             i;
    int             rc = cin01_db_ok;

    i29findpart (session, sp1pk_data);
    for (i = 0; i < sqlca->paramCount; ++i) {
        if ((int) sqlca->hostVar [i].pinfo.sp1i_io_type != sp1io_input) {
            rc = i29_outarg (session, &sqlca->hostVar[i]);
            if (rc != cin01_db_ok)
                return rc;
        }
    }
    return rc;
}

/*----------------------------------------*/

static          tin01_bool
i29_try_again (
    int rc)
{
    ROUTINE_DBG_MSP00 ("i29_try_again");
    tin01_bool           result;

    switch (rc) {
        case -8:               /* parse again */
        case 500:              /* wait for lock release */
        case 700:              /* timeout */
            result = true;
            break;
        default:
            result = false;
            break;
    }
    return result;
}

/*----------------------------------------*/

int
i29sqlcmd (
    tin01_sql_session * session,
    tin01_c_sqlca * sqlca,
    tin01_cstr cmd,
    ...)
{
    ROUTINE_DBG_MSP00 ("i29sqlcmd");
    va_list         args;
    int             rc;
    int             trycnt = 0;
    tin01_bool      try_again = true;
    int             i;

    do {
        rc = i29_parse (session, cmd, sqlca, false);
        if (rc == cin01_db_ok) {
            va_start (args, cmd);
            for (i = 0; i < sqlca->varCount; ++i) {
                sqlca->hostVar[i].addr = va_arg (args, void *);
            }
            va_end (args);
            rc = i29_inargs (session, sqlca);
            if (rc == cin01_db_ok)
                rc = i29sql (session, NULL);
            if (rc == cin01_db_ok) {
                rc = i29_outargs (session, sqlca);
            }
        }
        if (i29_try_again (rc)) {
            try_again = true;
            sqlca->inUse = false;
        }
        else
            try_again = false;
        ++trycnt;
    }
    while (try_again && (trycnt < c_max_try));
    return rc;
}

/*----------------------------------------*/

int
i29sqlcmdArr (
    tin01_sql_session * session,
    tin01_c_sqlca * sqlca,
    tin01_cstr cmd,
    char *args[])
{
    ROUTINE_DBG_MSP00 ("i29sqlcmdArr");
    int             rc;
    int             trycnt = 0;
    tin01_bool           try_again = true;
    int             i;

    do {
        rc = i29_parse (session, cmd, sqlca, false);
        if (rc == cin01_db_ok) {
            for (i = 0; i < sqlca->varCount; ++i) {
                sqlca->hostVar[i].addr = args[i];
            }
            rc = i29_inargs (session, sqlca);
            if (rc == cin01_db_ok)
                rc = i29sql (session, NULL);
            if (rc == cin01_db_ok) {
                rc = i29_outargs (session, sqlca);
            }
        }
        if (i29_try_again (rc)) {
            try_again = true;
            sqlca->inUse = false;
        }
        else
            try_again = false;
        ++trycnt;
    }
    while (try_again && (trycnt < c_max_try));
    return rc;
}

/*----------------------------------------*/

int
i29internalcmd (
    tin01_sql_session * session,
    tin01_c_sqlca * sqlca,
    tin01_cstr cmd,
    ...)
{
    ROUTINE_DBG_MSP00 ("i29internalcmd");
    va_list         args;
    int             rc;
    int             trycnt = 0;
    tin01_bool       try_again = true;
    int             i;

    do {
        rc = i29_parse (session, cmd, sqlca, true);
        if (rc == cin01_db_ok) {
            va_start (args, cmd);
            for (i = 0; i < sqlca->varCount; ++i) {
                sqlca->hostVar[i].addr = va_arg (args, void *);
            }
            va_end (args);
            rc = i29_inargs (session, sqlca);
            if (rc == cin01_db_ok)
                rc = i29sql (session, NULL);
            if (rc == cin01_db_ok) {
                rc = i29_outargs (session, sqlca);
            }
        }
        if (i29_try_again (rc)) {
            try_again = true;
            sqlca->inUse = false;
        }
        else
            try_again = false;
        ++trycnt;
    }
    while (try_again && (trycnt < c_max_try));
    return rc;
}

/*----------------------------------------*/

int
i29longiolen (
    int d_type,
    int len)
{
    ROUTINE_DBG_MSP00 ("i29longiolen");
    int             result;

    switch (d_type) {
        case dfixed:
        case dfloat:
        case dvfloat:
        case dsmallint:
        case dinteger:
            result = ((len + 1) / 2) + 1;
            break;
        case dboolean:
            result = 1;
            break;
        default:
            result = len;
            break;
    }
    return result + 1;
}

/*----------------------------------------*/

void
i29errmsg (
    tin01_sql_session * session,
    tsp00_C256c       & errbuf)
{
    ROUTINE_DBG_MSP00 ("i29errmsg");

    errbuf[0] = '\0';

    if (session->rec_packet != NULL) {
      if (session->segment->sp1r_returncode () != cin01_db_ok) {
        int     msglen;
        char    *errbuffer;

        if (i29findpart(session, sp1pk_errortext)) {
          errbuffer = session->part->sp1p_buf ().asCharp ();
          msglen = s30klen (errbuffer, ' ', (int) session->part->sp1p_buf_len ());

        if(msglen > errbuf.size())
            msglen=errbuf.size();

          memcpy ((char *)errbuf, errbuffer, msglen);
          errbuf[msglen] = '\0';
        } /* end if */
      } /* end if */
    } else {
      if (session->lasterr.lasterr_on) {
          size_t minlen=errbuf.size();

          if(minlen>sizeof(tsp00_ErrText)-1)
              minlen=sizeof(tsp00_ErrText)-1;

        memcpy((char *)errbuf, session->lasterr.errtext, minlen);
        errbuf[minlen] = '\0';
      } /* end if */
    } /* end if */

} /* end i29errmsg */

/*----------------------------------------*/

int
i29droppid (
    tin01_sql_session * session,
    tin01_parseid pid)
{
    ROUTINE_DBG_MSP00 ("i29droppid");
    tsp1_packet    *send_packet = session->send_packet;

    ASSERT_DBG (send_packet != NULL);
    i29initadbs (session);
    i29pcmd (session, "DROP PARSEID", UNDEF_SP00);
    i29newpart (session, sp1pk_parsid);
    memcpy (session->part->sp1p_buf () + session->part->sp1p_buf_len (),
        pid, sizeof (tin01_parseid));
    session->part->sp1p_buf_len () += sizeof (tin01_parseid);
    return i29sql (session, NULL);
}

/*----------------------------------------*/

#define c_layerstr_len   20
#define c_routinename_len 16

int
i29switch (
    tin01_sql_session * session,
    tin01_cstr trace,
    tin01_cstr test)
{
    ROUTINE_DBG_MSP00 ("i29switch");
    tsp00_Int2        result;

    if (!session->switch_possible) {
        return cin01_db_ok;
    }
    i29_reset (session, true);
    i29newsegment (session, sp1m_switch);
    i29newpart (session, sp1pk_command);
    i29pascalstring (session->part->sp1p_buf ().asCharp (), c_layerstr_len, trace, UNDEF_SP00);
    i29pascalstring (session->part->sp1p_buf ().asCharp () + c_layerstr_len,
        c_layerstr_len, test, UNDEF_SP00);
    session->part->sp1p_buf_len () = 2 * c_layerstr_len;
    result = i29sql (session, NULL);
    if (result != cin01_db_ok) {
        /* next time be quiet */
        TRACE_MIN01 (4, SQLTRACE, "SWITCH disabled after failure\n");
        session->switch_possible = false;
    }
    return result;
}

/*----------------------------------------*/

int
i29switchlimit (
    tin01_sql_session * session,
    tin01_cstr trace,
    tin01_cstr test,
    tin01_cstr start,
    tin01_cstr stop,
    int limit)
{
    ROUTINE_DBG_MSP00 ("i29switchlimit");
    tsp00_Int2 result;
    int pos = 0;
    unsigned char buf [2];

    if (!session->switch_possible) {
        return cin01_db_ok;
    }
    i29_reset (session, true);
    i29newsegment (session, sp1m_switchlimit);
    i29newpart (session, sp1pk_command);
    i29pascalstring (session->part->sp1p_buf ().asCharp (), c_layerstr_len, trace, UNDEF_SP00);
    pos += c_layerstr_len;
    i29pascalstring (session->part->sp1p_buf ().asCharp () + pos,
        c_layerstr_len, test, UNDEF_SP00);
    pos += c_layerstr_len;
    i29pascalstring (session->part->sp1p_buf ().asCharp () + pos,
        c_routinename_len, start, UNDEF_SP00);
    pos += c_routinename_len;
    i29pascalstring (session->part->sp1p_buf ().asCharp () + pos,
        c_routinename_len, stop, UNDEF_SP00);
    pos += c_routinename_len;
    session->part->sp1p_buf_len () = pos;
    buf [0] = limit / 256;
    buf [1] = limit % 256;
    i29pcmd (session, (const char *) buf, sizeof (buf));
    result = i29sql (session, NULL);
    if (result != cin01_db_ok) {
        /* next time be quiet */
        TRACE_MIN01 (4, SQLTRACE, "SWITCH disabled after failure\n");
        session->switch_possible = false;
    }
    return result;
}

/*----------------------------------------*/

int
i29buflimit (
    tin01_sql_session * session,
    int limit)
{
    ROUTINE_DBG_MSP00 ("i29buflimit");
    tsp00_Int2        result;
    unsigned char buf [2];

    if (!session->switch_possible) {
        return cin01_db_ok;
    }
    i29_reset (session, true);
    i29newsegment (session, sp1m_buflength);
    i29newpart (session, sp1pk_command);
    buf [0] = limit / 256;
    buf [1] = limit % 256;
    i29pcmd (session, (const char *) buf, sizeof (buf));
    result = i29sql (session, NULL);
    if (result != cin01_db_ok) {
        /* next time be quiet */
        TRACE_MIN01 (4, SQLTRACE, "SWITCH disabled after failure\n");
        session->switch_possible = false;
    }
    return result;
}

/*----------------------------------------*/

static int
i29_minmaxbuf (
    tin01_sql_session * session,
    tsp1_cmd_mess_type_Enum messKind)
{
    ROUTINE_DBG_MSP00 ("i29_minmaxbuf");
    tsp00_Int2        result;

    if (!session->switch_possible) {
        return cin01_db_ok;
    }
    i29_reset (session, true);
    i29newsegment (session, messKind);
    i29newpart (session, sp1pk_command);
    result = i29sql (session, NULL);
    if (result != cin01_db_ok) {
        /* next time be quiet */
        TRACE_MIN01 (4, SQLTRACE, "SWITCH disabled after failure\n");
        session->switch_possible = false;
    }
    return result;
}

/*----------------------------------------*/

int
i29minbuf (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29minbuf");

    return i29_minmaxbuf (session, sp1m_minbuf);
}

/*----------------------------------------*/

int
i29maxbuf (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29maxbuf");

    return i29_minmaxbuf (session, sp1m_maxbuf);
}

/*----------------------------------------*/

/**\
--- ----------------------------------MF__ MOD__ high level select
    Here some functions providing a interface for execute select
    statements and retrieve the results
\**/

/*----------------------------------------*/

#define cin29_cont_cont "CONTINUE\n"
#define cin29_cont_end  "END     \n"
#define cin29_cont_len  (strlen(cin29_cont_end))
#define cin29_min_len   10
#define cin29_nulltext  "(null)"

/*----------------------------------------*/

/* checks if a data type of the db is a string type
   (for seting of string delimiter sin the result) */
static tin01_bool
i29_isstring (tsp00_DataType aDataType)
{
    ROUTINE_DBG_MSP00 ("i29_isstring");

  tin01_bool bRc = false;

  switch(aDataType) {
    /* Had I hit all data types??? */
    case dcha:
    case dche:
    case dchb:
    case dunicode:
    case dvarchara:
    case dvarchare:
    case dvarcharb:
    case ddate:
    case dtime:
    case dtimestamp:
    case dstra:
    case dstre:
    case dstrb:
    case dstrdb:
      bRc = true;
  } /* end switch */

  return bRc;

} /* end i29_isstring */

/*----------------------------------------*/

/* calculates a approximate length of one result record */
static int
i29_rowlen (
    tin01_sql_session  * session)
{
    ROUTINE_DBG_MSP00 ("i29_rowlen");

    int               nParamCount = 0,
                      i           = 0,
                      nRowLen     = 0;
    tsp1_param_info * pParamInfo  = NULL;

    /**/

    /* get some infos */
    pParamInfo = i29gparaminfo(session, 0);
    nParamCount = i29paramcount(session);

    /* calculate approximate rowlen */
    for (i = 0; i < nParamCount; i++) {
      nRowLen = nRowLen + ((pParamInfo[i].sp1i_length < cin29_min_len) ? pParamInfo[i].sp1i_length : cin29_min_len);
    } /* end for */

    /**/

    return nRowLen;

} /* end i29_rowlen */

/*----------------------------------------*/

/* gets the selected records in the result buffer until
   nResultLen is reached
   (at first in the result buffer a keyword (END or CONTINUE)
   will indicate whether the end of result is reached or not) */
static int
i29_fetchnext (
    tin01_sql_session  * session,
    char              * pResult,
    long                nResultLen,
    char                cFieldSep,
    char                cStringDel,
    char                cRecSep)
{
    ROUTINE_DBG_MSP00 ("i29_fetchnext");

    int               rc          = cin01_db_ok,
                      i           = 0,
                      nParamCount = 0,
                      nRowLen     = 0;
    tsp1_param_info * pParamInfo  = NULL;
    char            * pCurrent    = pResult;
    char            * pContinue   = NULL;
    tin01_c_hostvar    aHostVar;

    /**/

    *pResult = 0;

    /* place for end/continue mark */
    pContinue = pCurrent;
    pCurrent  = pCurrent + cin29_cont_len;

    /* get some infos */
    nParamCount = i29paramcount(session);
    pParamInfo = i29gparaminfo(session, 0);
    nRowLen = i29_rowlen(session) + (nParamCount * 3);

    while ((rc == cin01_db_ok) && (pCurrent - pResult + nRowLen < nResultLen) ) {

      i29findpart (session, sp1pk_data);
      for (i = 0; i < nParamCount; i++) {

        /* set string delimiter */
        if (i29_isstring(pParamInfo[i].sp1i_data_type) && cStringDel != 0) {
          *pCurrent++ = cStringDel;
        } /* end if */

        /* fill hostvar struct */
        memset(&aHostVar, 0, sizeof(aHostVar));
        aHostVar.len = -1;
        aHostVar.c_type.becomes(cin01_c_charp);
        memcpy(&(aHostVar.pinfo), &pParamInfo[i], sizeof(tsp1_param_info));
        aHostVar.addr = pCurrent;
        *((char *) (aHostVar.addr)) = 0;

        /* read value */
        rc = i29_outarg (session, &aHostVar);

        /* check for <NULL> */
        if (aHostVar.is_null) {
          /* skip string delimiter */
          if (i29_isstring(pParamInfo[i].sp1i_data_type) && cStringDel != 0) {
            pCurrent--;
          } /* end if */
          strcpy(pCurrent, cin29_nulltext);
        } /* end if */

        pCurrent = pCurrent + strlen(pCurrent);

        /* set string delimiter */
        if (i29_isstring(pParamInfo[i].sp1i_data_type) && !aHostVar.is_null  && cStringDel != 0) {
          *pCurrent++ = cStringDel;
        } /* end if */

        if (cFieldSep != 0) {
          *pCurrent++ = cFieldSep;
        } /*end if */
      } /* end for */

      /* skip last field separator */
      if (cFieldSep != 0) {
        pCurrent--;
      } /*end if */
      if (cRecSep != 0) {
        *pCurrent++ = cRecSep;
      } /*end if */

      /* next row */
      i29_fetchcmd (session, "FETCH NEXT", nParamCount);
      session->segment->sp1c_with_info () = true;
      rc =  i29sql (session, NULL);
      
      if (rc == cin01_db_ok) {
        nParamCount = i29paramcount(session);
        pParamInfo = i29gparaminfo(session, 0);
        nRowLen = i29_rowlen(session) + (nParamCount * 3);
      } // end if

    } /* end while */

    /* handle end/continue mark */
    if (rc == cin01_db_ok) {
      strncpy(pContinue, cin29_cont_cont, cin29_cont_len);
    } else if (rc == cin01_db_row_not_found) {
      strncpy(pContinue, cin29_cont_end, cin29_cont_len);
      rc = cin01_db_ok;
    } /* end if */

    /* skip last record separator */
    if (cRecSep != 0) {
      pCurrent--;
    } /* end if */

    *pCurrent = 0;

    return rc;

} /* end i29_fetchnext */

/*----------------------------------------*/

static int
i29_fetchnice (
    tin01_sql_session  * session,
    char               * pResult,
    char               * pCurrent,
    char               * pContinue,
    long               nResultLen,
    bool               bOneRowResult,
    long               nMaxNameLen,
    int                nRowLen,
	bool               bDataOnly)
{
    ROUTINE_DBG_MSP00 ("i29_fetchnice");

    int               rc          = cin01_db_ok;
    int               i           = 0;
    int               nParamCount = 0;
//    int               nRowLen     = i29_rowlen(session);
    tsp1_param_info * pParamInfo  = NULL;
    tin01_c_hostvar   aHostVar;

    char              nLen        = 0;
//    char              nMaxLen     = 0;
    short              nMaxLen     = 0;

    /**/

    while ((rc == cin01_db_ok) &&
           (pCurrent - pResult + nRowLen < nResultLen)) {

      /* get some infos */
      tin01_cstr pNames = i29colnames(session);

      pParamInfo  = i29gparaminfo(session, 0);
      nParamCount = i29paramcount(session);

      i29findpart (session, sp1pk_data);
      for (i = 0; i < nParamCount; i++) {

        nLen = *pNames;
        pNames++;

        if (bOneRowResult) {
          memcpy(pCurrent, pNames, nLen);
          pCurrent += nLen;
          memset(pCurrent, 32, nMaxNameLen - nLen);
          pCurrent += (nMaxNameLen - nLen);
          strcpy(pCurrent, " = ");
          pCurrent = pCurrent + strlen(pCurrent);
         }

        pNames += nLen;

        nMaxLen = nLen;
        if (pParamInfo[i].sp1i_length > nLen) {
          nMaxLen = pParamInfo[i].sp1i_length;
        }

        /* fill hostvar struct */
        memset(&aHostVar, 0, sizeof(aHostVar));
        aHostVar.len = -1;
        aHostVar.c_type.becomes(cin01_c_charp);
        memcpy(&(aHostVar.pinfo), &pParamInfo[i], sizeof(tsp1_param_info));
        aHostVar.addr = pCurrent;
        *((char *) (aHostVar.addr)) = 0;

        /* read value */
        rc = i29_outarg (session, &aHostVar);

        /* check for <NULL> */
        if (aHostVar.is_null) {
          strcpy(pCurrent, " "); /* cin29_nulltext */
        }

        nLen = (int)strlen(pCurrent);

        pCurrent = pCurrent + nLen;

        memset(pCurrent, 32, nMaxLen - nLen);
        pCurrent += (nMaxLen - nLen);

        if (bOneRowResult) {
          *pCurrent++ = '\n';
        } else {
          if (i < (nParamCount -1 )) {
            *pCurrent++ = ' ';
            *pCurrent++ = '|';
            *pCurrent++ = ' ';
          }
        }
      }

      if (!bOneRowResult) {
        *pCurrent++ = '\n';
      }

      /* next row */
      i29_fetchcmd (session, "FETCH NEXT", nParamCount);
      session->segment->sp1c_with_info () = true;
      rc =  i29sql (session, NULL);

    } /* end while */

    /* handle end/continue mark */
	if (!bDataOnly) {
      if (rc == cin01_db_ok) {
        strncpy(pContinue, cin29_cont_cont, cin29_cont_len);
	  } else if (rc == cin01_db_row_not_found) {
        strncpy(pContinue, cin29_cont_end, cin29_cont_len);
        rc = cin01_db_ok;
	  } /* end if */
	}

    *pCurrent = 0;

    return rc;
} /* end i29_fetchnice */

/*----------------------------------------*/

int
i29selectnice (
    tin01_sql_session  * session,
    tin01_cstr           cmd,
    char               * pResult,
    long                 nResultLen,
    bool                 bOneRowResult,
    bool                 bDesc,
	bool                 bDataOnly)
{
    ROUTINE_DBG_MSP00 ("i29selectnice");

    int               rc           = cin01_db_ok;
    int               i            = 0;
    int               j            = 0;
    int               nParamCount  = 0;
    tin01_cstr        pNames       = NULL;
    char            * pCurrent     = pResult;
    char            * pContinue    = NULL;
    char            * pHeader      = NULL;
    char              nLen         = 0;
    char              nMaxNameLen  = 0;
//    char              nMaxParamLen = 0;
    short             nMaxParamLen = 0;
    tsp1_param_info * pParamInfo   = NULL;
//  char              nMaxLen      = 0;
    short             nMaxLen      = 0;
    int               nRowLen      = 0;

    char              szDesc[255];

    /**/

    *pResult = 0;

    pContinue = pCurrent;
	if (!bDataOnly) {
		pCurrent  = pCurrent + cin29_cont_len;
	}
    pHeader   = pCurrent;

    /* prepare statement execution */
    i29initadbs (session);
    session->segment->sp1c_with_info () = true;
    i29pcmd (session, cmd, UNDEF_SP00);

    /* execute statement */
    rc = i29sql (session, NULL);

    if (rc == cin01_db_ok) {

      pParamInfo  = i29gparaminfo(session, 0);
      nParamCount = i29paramcount(session);
      pNames      = i29colnames(session);

      if (bDesc) {
        for (i = 0; i < nParamCount; i++) {

          szDesc[0] = '\0';

          if (bOneRowResult) {
            strcat(szDesc, "SS");
          }

          for (i = 0; i < nParamCount; i++) {
            switch (pParamInfo[i].sp1i_data_type) {
              case dfixed:
              case dfloat:
              case dvfloat:
              case dsmallint:
              case dinteger:
                strcat(szDesc, "N");
                break;
              case dcha:
              case dche:
              case dchb:
              case dunicode:
              case dvarchara:
              case dvarchare:
              case dvarcharb:
              case dstra:
              case dstre:
              case dstrb:
              case dstrdb:
                strcat(szDesc, "S");
                break;
              case ddate:
              case dtime:
              case dtimestamp:
                strcat(szDesc, "D");
                break;
            case dboolean:
                strcat(szDesc, "B");
                break;
            default:
                strcat(szDesc, "-");
                break;
            }
          }
          strcat(szDesc, "\n");
          strcpy(pCurrent, szDesc);
          pCurrent += strlen(pCurrent);
        }
      }

      for (i = 0; i < nParamCount; i++) {

        nLen = *pNames;

        if (pParamInfo[i].sp1i_length > nMaxParamLen) {
          nMaxParamLen = pParamInfo[i].sp1i_length;
        }

        if (bOneRowResult) {
          if (nLen > nMaxNameLen) {
            nMaxNameLen = nLen;
          }
          pNames += (nLen + 1);
        } else {
          nMaxLen = nLen;
          if (pParamInfo[i].sp1i_length > nLen) {
            nMaxLen = pParamInfo[i].sp1i_length;
          }
          pNames++;
          memcpy(pCurrent, pNames, nLen);
          pCurrent += nLen;
          memset(pCurrent, 32, nMaxLen - nLen);
          pCurrent += (nMaxLen - nLen);
          if (i < (nParamCount - 1)) {
            strcpy(pCurrent, " | ");
          } else {
            strcpy(pCurrent, "\n\n");
          }
          pCurrent += strlen(pCurrent);
          pNames += nLen;
        }
      }

      if (bOneRowResult) {
        strcpy(pCurrent, "Name");
        pCurrent += strlen(pCurrent);
        memset(pCurrent, 32, nMaxNameLen - 4);
        pCurrent += (nMaxNameLen - 4);
        strcpy(pCurrent, " | Value");
        pCurrent += strlen(pCurrent);
        memset(pCurrent, 32, nMaxParamLen - 5);
        pCurrent += (nMaxParamLen - 5);
        strcpy(pCurrent, "\n\n");
        pCurrent += strlen(pCurrent);
      }

      if (!bOneRowResult) {
        nRowLen = (int)strlen(pHeader) - 1;
      }

	  if (bDataOnly) {
        pCurrent = pHeader;
	  }

    }

    /* get first row */
    if (rc == cin01_db_ok) {
      i29_fetchcmd (session, "FETCH FIRST", nParamCount);
      session->segment->sp1c_with_info () = true;
      rc =  i29sql (session, NULL);
    }

    /* analyze rows */
    if (rc == cin01_db_ok) {
      rc = i29_fetchnice(session, pResult, pCurrent, pContinue, nResultLen, bOneRowResult, nMaxNameLen, nRowLen, bDataOnly);
    }

    if (rc == cin01_db_row_not_found) {
      rc = cin01_db_ok;
    }

    return rc;
} /* end i29selectnice */

int
i29fetchnice (
    tin01_sql_session  * session,
    char               * pResult,
    long               nResultLen)
{
    ROUTINE_DBG_MSP00 ("i29fetchnice");

    int               rc          = cin01_db_ok;
    int               i           = 0;
    int               j           = 0;
    int               nParamCount = 0;
    tin01_cstr        pNames      = NULL;
    char            * pCurrent    = pResult;
    char            * pContinue   = NULL;
    char            * pHeader     = NULL;
    char              nLen        = 0;
    char              nMaxNameLen = 0;
    tsp1_param_info * pParamInfo  = NULL;
//  char              nMaxLen     = 0;
    short             nMaxLen     = 0;
    int               nRowLen     = 0;

    /**/

    *pResult = 0;

    pContinue = pCurrent;
    pCurrent  = pCurrent + cin29_cont_len;
    pHeader   = pCurrent;

    pParamInfo  = i29gparaminfo(session, 0);
    nParamCount = i29paramcount(session);
    pNames      = i29colnames(session);

    for (i = 0; i < nParamCount; i++) {
      nLen = *pNames;

      nMaxLen = nLen;
      if (pParamInfo[i].sp1i_length > nLen) {
        nMaxLen = pParamInfo[i].sp1i_length;
      }
      pNames++;
      memcpy(pCurrent, pNames, nLen);
      pCurrent += nLen;
      memset(pCurrent, 32, nMaxLen - nLen);
      pCurrent += (nMaxLen - nLen);
      if (i < (nParamCount - 1)) {
        strcpy(pCurrent, " | ");
      } else {
        strcpy(pCurrent, "\n\n");
      }
      pCurrent += strlen(pCurrent);
      pNames += nLen;
    }

    nRowLen = (int)strlen(pHeader) - 1;

    /* get next row */
    i29_fetchcmd (session, "FETCH NEXT", nParamCount);
    session->segment->sp1c_with_info () = true;
    rc = i29sql (session, NULL);

    /* analyze rows */
    if (rc == cin01_db_ok) {
      rc = i29_fetchnice(session, pResult, pCurrent, pContinue, nResultLen, false, nMaxNameLen, nRowLen, false);
    }

    /**/

    return rc;
} /* end i29fetchnice */


/*----------------------------------------*/

/* puts infos about the result of the select statement
   into the result buffer
   at the moment (1998-04-20 - BV)  only the field names */
int
i29selectinfo (
    tin01_sql_session  * session,
    tin01_cstr           cmd,
    char              * pResult,
    char                cFieldSep)
{
    ROUTINE_DBG_MSP00 ("i29selectinfo");

    int               rc          = cin01_db_ok,
                      i           = 0,
                      j           = 0,
                      nParamCount = 0;
    tin01_cstr         pNames      = NULL;
    char            * pCurrent    = pResult;
    char              nLen        = 0;

    /**/

    *pResult = 0;

    /* prepare statement execution */
    i29initadbs (session);
    session->segment->sp1c_with_info () = true;
    session->segment->sp1c_producer ().becomes(sp1pr_internal_cmd);
    i29pcmd (session, cmd, UNDEF_SP00);

    /* execute statement */
    rc = i29sql (session, NULL);

    /* analyze result */
    if (rc == cin01_db_ok) {

      /* get some infos */
      nParamCount = i29paramcount(session);
      pNames      = i29colnames(session);

      /* get columns */
      for (i = 0; i < nParamCount; i++) {
        nLen = *pNames++;
        for (j = 0; j < nLen; j++) {
          *pCurrent++ = *pNames++;
        } /* end for */
        if (cFieldSep != 0) {
          *pCurrent++ = cFieldSep;
        } /* end if */
      } /* end for */
      /* skip last field separator */
      if (cFieldSep != 0) {
        pCurrent--;
      } /* end if */
    } /* end if */

    *pCurrent = 0;

    return rc;
} /* end i29selectinfo */

/*----------------------------------------*/

/* execute a select statement and call i29_fetchnext
   for retrieving (the first part) of the result
   (try to execute a non select statement will produce an error
   in the kernel) */
int
i29select (
    tin01_sql_session  * session,
    tin01_cstr           cmd,
    char              * pResult,
    long                nResultLen,
    char                cFieldSep,
    char                cStringDel,
    char                cRecSep)
{
    ROUTINE_DBG_MSP00 ("i29select");

    int rc = cin01_db_ok;

    /**/

    *pResult = 0;

    /* prepare statement execution */
    i29initadbs (session);
    session->segment->sp1c_with_info () = true;
    session->segment->sp1c_producer ().becomes(sp1pr_internal_cmd);
    i29pcmd (session, cmd, UNDEF_SP00);

    /* execute statement */
    rc = i29sql (session, NULL);

    /* get first row */
    if (rc == cin01_db_ok) {
      i29_fetchcmd (session, "FETCH FIRST", i29paramcount(session));
      session->segment->sp1c_with_info () = true;
      rc =  i29sql (session, NULL);
    } /* end if */

    /* analyze rows */
    if (rc == cin01_db_ok) {
      rc = i29_fetchnext(session, pResult, nResultLen, cFieldSep, cStringDel, cRecSep);
    } /* end if */

   return rc;
} /* end i29select */

/*----------------------------------------*/

/* retrieve the next part of a result. no open select
   in the session will produce an error */
int
i29fetch (
    tin01_sql_session  * session,
    char              * pResult,
    long                nResultLen,
    char                cFieldSep,
    char                cStringDel,
    char                cRecSep)
{
    ROUTINE_DBG_MSP00 ("i29fetch");

    int rc = cin01_db_ok;

    /**/

    *pResult = 0;

    /* get next row */
    i29_fetchcmd (session, "FETCH NEXT", i29paramcount(session));
    session->segment->sp1c_with_info () = true;
    rc =  i29sql (session, NULL);

    /* analyze rows */
    if (rc == cin01_db_ok) {
      rc = i29_fetchnext(session, pResult, nResultLen, cFieldSep, cStringDel, cRecSep);
    } /* end if */

    /**/

    return rc;
} /* end i29fetch */

/*----------------------------------------*/

int
i29record (
    tin01_sql_session  * session,
    tin01_cstr           cmd)
{
  ROUTINE_DBG_MSP00 ("i29record");

  int rc = cin01_db_ok;

  /**/

  /* prepare statement execution */
  i29initadbs (session);
  session->segment->sp1c_with_info () = true;
  session->segment->sp1c_producer ().becomes(sp1pr_internal_cmd);
  i29pcmd (session, cmd, UNDEF_SP00);

  /* execute statement */
  rc = i29sql (session, NULL);

  /* get first row */
  if (rc == cin01_db_ok) {
    i29_fetchcmd (session, "FETCH FIRST", i29paramcount(session));
    session->segment->sp1c_with_info () = true;
    rc =  i29sql (session, NULL);
  } /* end if */

  /**/

  return rc;
} /* end i29record */

/*----------------------------------------*/

int
i29nextrecord (
    tin01_sql_session  * session)
{
  ROUTINE_DBG_MSP00 ("i29nextrecord");

  int rc = cin01_db_ok;

  /**/

  /* get next row */
  i29_fetchcmd (session, "FETCH NEXT", i29paramcount(session));
  session->segment->sp1c_with_info () = true;
  rc =  i29sql (session, NULL);

  /**/

  return rc;
} /* end i29nextrecord */

/*----------------------------------------*/

int
i29fieldvalue (
    tin01_sql_session  * session,
    const tsp00_Int4     nField,
    tin01_c_hostvar    * pHostvar,
    const char           cStringQuote)

{
  ROUTINE_DBG_MSP00 ("i29fieldvalue");

  int               rc          = cin01_db_ok;
  tsp1_param_info * pParamInfo  = NULL;
  int               nParamCount = 0;
  char            * pChar       = (char *) pHostvar->addr;
  /**/

  pParamInfo  = i29gparaminfo(session, 0);
  nParamCount = i29paramcount(session);

  if (nParamCount >= nField) {
    memcpy(&(pHostvar->pinfo), &pParamInfo[nField - 1], sizeof(tsp1_param_info));
    i29findpart (session, sp1pk_data);

    // handle strings
    if (i29_isstring(pHostvar->pinfo.sp1i_data_type) && cStringQuote != 0) {
      *pChar = cStringQuote;
      ++pChar;
      pHostvar->addr = pChar;
    } // end if

    /* read value */
    rc = i29_outarg (session, pHostvar);

    // handle strings
    if (i29_isstring(pHostvar->pinfo.sp1i_data_type) && cStringQuote != 0) {
      if (rc == cin01_db_ok) {
        pChar[strlen(pChar)    ] = cStringQuote;
        pChar[strlen(pChar) + 1] = 0;
      } // end if

      --pChar;
      pHostvar->addr = pChar;
    } // end if
  } else {
    rc = cin01_db_invalid_number;
  } // end if

  /**/

  return rc;
} /* end i29fieldvalue */

/*----------------------------------------*/

/**\
--- ----------------------------------MF__ MOD__ chapter special cases
    The following routines serve very specific needs.
    They are included so that clients don't need the
    internal structure of segments and parts
\**/

char           *
i29outargsbuf (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29outargsbuf");
    i29findpart (session, sp1pk_data);
    return session->part->sp1p_buf ().asCharp ();
}

/*----------------------------------------*/

char           *
i29argsbuf (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29argsbuf");
    return session->part->sp1p_buf ().asCharp ();
}

/*----------------------------------------*/

void
i29argsbuflen (
    tin01_sql_session * session,
    int buflen)
{
    ROUTINE_DBG_MSP00 ("i29argsbuflen");
    session->part->sp1p_buf_len () = buflen;
}

/*----------------------------------------*/

static void
i29_puttermid (
    tin01_sql_session * session,
    tsp00_TermId termid)
{
    ROUTINE_DBG_MSP00 ("i29_puttermid");
    char           *target;
    tsp1_packet    *packet = session->send_packet;

    target = session->part->sp1p_buf () + session->part->sp1p_buf_len ();
    *target = csp_defined_byte;
    ++target;
    memcpy (target, termid, sizeof (tsp00_TermId));
    session->part->sp1p_buf_len () += 1 + sizeof (tsp00_TermId);
}

/*----------------------------------------*/

static void
i29_connectOptions (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29_sqlmode");
    if (isalpha (session->xuser.xu_sqlmode[0]) != '\0') {
        i29pcmdf (session, " SQLMODE %.*s",
            sizeof (session->xuser.xu_sqlmode), (char *) session->xuser.xu_sqlmode);
    }
    if (session->xuser.xu_cachelimit != -1) {
        i29pcmdf (session, " CACHELIMIT %d ", session->xuser.xu_cachelimit);
    }
    if (session->xuser.xu_timeout != -1) {
        i29pcmdf (session, " TIMEOUT %d ", session->xuser.xu_timeout);
    }
    if (session->xuser.xu_isolation != -1) {
        i29pcmdf (session, " ISOLATION LEVEL %d ", session->xuser.xu_isolation);
    }
}

/*----------------------------------------*/

void
i29connectcommand (
    tin01_sql_session * session,
    tsp00_TermId termid)
{
    ROUTINE_DBG_MSP00 ("i29connectcommand");
    session->buildingCommand = false;
    i29initadbs (session);
    session->segment->sp1c_producer ().becomes(sp1pr_internal_cmd);
    i29pcmdf (session, "CONNECT \"%.*s\" IDENTIFIED by :PW",
        sizeof (session->xuser.xu_user), (char *) session->xuser.xu_user);
    i29_connectOptions (session);
    i29newpart (session, sp1pk_data);
    i29ppw (session, session->xuser.xu_password, csp_defined_byte);
    i29_puttermid (session, termid);
}

/*----------------------------------------*/

void
i29tablebuffer (
    tin01_sql_session * session,
    char **tablebuffer,
    tsp00_Int4 * buflen)
{
    ROUTINE_DBG_MSP00 ("i29tablebuffer");
    tsp1_packet    *rec_packet = session->rec_packet;

    tin01_bool           partFound;

    ASSERT_DBG (rec_packet != NULL);
    partFound = i29findpart (session, sp1pk_conv_tables_returned);
    ASSERT_DBG (partFound);
    *tablebuffer = session->part->sp1p_buf ().asCharp ();
    *buflen = session->part->sp1p_buf_len ();
}

/*----------------------------------------*/

char           *
i29cmdbuf (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29cmdbuf");
    return session->part->sp1p_buf ().asCharp ();
}

/*----------------------------------------*/

char           *
i29getpart (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29getpart");
    return session->part->sp1p_buf ().asCharp ();
}

/*----------------------------------------*/

void
i29extendpart (
    tin01_sql_session * session,
    int lendiff)
{
    ROUTINE_DBG_MSP00 ("i29extendpart");
    session->part->sp1p_buf_len () += lendiff;
}

/*----------------------------------------*/

tsp00_Int4
i29partlen (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29partlen");
    return session->part->sp1p_buf_len ();
}

/*----------------------------------------*/

tsp00_Int4
i29bytesfree (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29bytesfree");
    return session->part->sp1p_buf_size ()
        - session->part->sp1p_buf_len () - 4;
}

/*----------------------------------------*/

void
i29addpartdata (
    tin01_sql_session * session,
    const void *data,
    int len)
{
    ROUTINE_DBG_MSP00 ("i29addpartdata");
    memcpy (session->part->sp1p_buf () + session->part->sp1p_buf_len (),
        data, len);
    session->part->sp1p_buf_len () += len;
    ASSERT_DBG (session->part->sp1p_buf_len ()
        <= session->part->sp1p_buf_size ());
}

/*----------------------------------------*/

tsp1_param_info *
i29allocparaminfos (
    tin01_sql_session * session,
    int paramCount)
{
    ROUTINE_DBG_MSP00 ("i29allocparaminfos");
    i29_reset (session, true);
    i29newsegment (session, sp1m_parse);
    i29newpart (session, sp1pk_shortinfo);
    session->part->sp1p_arg_count () = (tsp00_Int2) paramCount;
    session->part->sp1p_buf_len () = paramCount * sizeof (tsp1_param_info);
    return (tsp1_param_info *) session->part->sp1p_buf ().asCharp ();
}

/*----------------------------------------*/

void
i29newnamepart (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29newnamepart");
    i29newpart (session, sp1pk_columnnames);
    session->part->sp1p_arg_count () = 0;
}

/*----------------------------------------*/

void
i29addname (
    tin01_sql_session * session,
    tin01_cstr name)
{
    ROUTINE_DBG_MSP00 ("i29addname");
    int             namelen;
    char           *buf;

    ASSERT_DBG (session->part->sp1p_part_kind () == sp1pk_columnnames);
    namelen = (int)strlen (name); /* physical, not logical */
    buf = session->part->sp1p_buf () + session->part->sp1p_buf_len ();
    buf[0] = (char) namelen;
    memcpy (buf + 1, name, namelen);
    session->part->sp1p_buf_len () += namelen + 1;
    ++session->part->sp1p_arg_count ();
}

/*----------------------------------------*/

tin01_bool
i29hello (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29hello");
    i29_reset (session, true);
    i29newsegment (session, sp1m_hello);
    i29sql (session, NULL);
    return session->is_connected;
}

/*----------------------------------------*/

#define c_intostr " Into ?"
#define c_intolen 7
#define c_varstr ", ?"
#define c_varlen 3

static void
i29_fetchcmd (
    tin01_sql_session * session,
    tin01_cstr fetchCmd,
    int fetchVars)
{
    ROUTINE_DBG_MSP00 ("i29_fetchcmd");
    char           *cmdBuf;
    int             fetchLen;
    int             additional;

    i29initadbs (session);
    fetchLen = (int)strlen (fetchCmd);
    cmdBuf = session->part->sp1p_buf () + session->part->sp1p_buf_len ();
    memcpy (cmdBuf, fetchCmd, fetchLen);
    cmdBuf += fetchLen;
    memcpy (cmdBuf, c_intostr, c_intolen);
    --fetchVars;
    cmdBuf += c_intolen;
    additional = fetchLen + c_intolen + (fetchVars * c_varlen);
    while (fetchVars > 0) {
        memcpy (cmdBuf, c_varstr, c_varlen);
        cmdBuf += c_varlen;
        --fetchVars;
    }
    session->part->sp1p_buf_len () += additional;
}

/*----------------------------------------*/

int
i29fetchcmd (
    tin01_sql_session * session,
    tin01_cstr fetchCmd,
    int fetchVars)
{
    ROUTINE_DBG_MSP00 ("i29fetchcmd");
    i29_fetchcmd (session, fetchCmd, fetchVars);
    return i29sql (session, NULL);
}

/*----------------------------------------*/

int
i29mfetchcmd (
    tin01_sql_session * session,
    tin01_cstr fetchCmd,
    int fetchVars,
    int rowsToFetch)
{
    ROUTINE_DBG_MSP00 ("i29mfetchcmd");
    tsp00_Number      num;
    tsp00_NumError   res;
    int             rc = cin01_db_ok;

    i29_fetchcmd (session, fetchCmd, fetchVars);
    i29newpart (session, sp1pk_resultcount);
    s41plint (&num, 1, 18, 0, rowsToFetch, res);
    rc = i29parg (session, num, sizeof (num), sizeof (num) + 1,
        1, csp_defined_byte);
    if (rc != cin01_db_ok)
        return rc;
    session->segment->sp1c_mass_cmd () = true;
    return i29sql (session, NULL);
}

/*----------------------------------------*/

int
i29functioncode (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29functioncode");
    return session->segment->sp1r_function_code ();
}

/*----------------------------------------*/

tin01_bool
i29gresulttablename (
    tin01_sql_session * session,
    int              * len,
    char             * resulttablename)
{
    ROUTINE_DBG_MSP00 ("i29gresulttablename");
    int                r_len = 0;

    if (i29findpart (session, sp1pk_resulttablename)) {
        r_len = session->part->sp1p_buf_len ();
        if (r_len > 0) {
            memcpy (resulttablename, session->part->sp1p_buf (), r_len);
        }
    }
    *len = r_len;
    return (r_len > 0);
} /* i29gresulttablename */

/*----------------------------------------*/

void
i29sresulttablename (
    tin01_sql_session * session,
    int                len,
    char             * resulttablename)
{
    ROUTINE_DBG_MSP00 ("i29sresulttablename");

    i29newpart (session, sp1pk_resulttablename);
    memcpy (session->part->sp1p_buf () + session->part->sp1p_buf_len (),
        resulttablename, len );
    session->part->sp1p_buf_len () += len;
} /* i29sresulttablename */

/*----------------------------------------*/

tin01_bool
i29forreuse (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29forreuse");
    tin01_parseid       pid;

    if (!i29findpart (session, sp1pk_parsid))
        return false;
    MEMCPY_MIN01 (pid, session->part->sp1p_buf ());
    return ((pid [10] == csp1_p_reuse_mass_select_found)
         || (pid [10] == csp1_p_reuse_update_sel_found)
         || (pid [10] == csp1_p_reuse_mselect_found)
         || (pid [10] == csp1_p_reuse_upd_mselect_found)
           );
} /* i29forreuse */

/*----------------------------------------*/

tin01_bool
i29forupdate (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29forupdate");
    tin01_parseid pid;
    if (!i29findpart (session, sp1pk_parsid))
        return false;
    MEMCPY_MIN01 (pid, session->part->sp1p_buf () + session->part->sp1p_buf_len ());
    return ((pid [10] == csp1_p_select_for_update_found)
         || (pid [10] == csp1_p_for_upd_mselect_found)
         || (pid [10] == csp1_p_reuse_update_sel_found)
         || (pid [10] == csp1_p_reuse_upd_mselect_found)
           );
} /* i29forupdate */

/*----------------------------------------*/

int
i29rowsfetched (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29rowsfetched");
    i29findpart (session, sp1pk_data);
    return session->part->sp1p_arg_count ();
}

/*----------------------------------------*/

int
i29rowlen (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29rowlen");
    int             result;
    int             recordCount;

    i29findpart (session, sp1pk_data);
    recordCount = session->part->sp1p_arg_count ();
    result = session->part->sp1p_buf_len () / recordCount;
    return result;
}

/*----------------------------------------*/

tin01_bool
i29endoffetch (
    tin01_sql_session * session,
    int rowsExpected)
{
    ROUTINE_DBG_MSP00 ("i29endoffetch");
    int             endOfFetch;
    int             recordCount;
    int             recordLen;
    int             bytesRemaining;

    switch (session->segment->sp1r_returncode ()) {
        case cin01_db_row_not_found:
            endOfFetch = true;
            break;
        case cin01_db_ok:
            i29findpart (session, sp1pk_data);
            recordCount = session->part->sp1p_arg_count ();
            if (recordCount == rowsExpected)
                endOfFetch = false;
            else {
                recordLen = session->part->sp1p_buf_len () / recordCount;
                bytesRemaining = session->part->sp1p_buf_size ()
                    - session->part->sp1p_buf_len ();
                if (bytesRemaining >= recordLen)
                    endOfFetch = true;
                else
                    endOfFetch = false;
            }
            break;
        default:
            return true;
            break;
    }
    return endOfFetch;
}

/*----------------------------------------*/
/**\
--- ----------------------------------MF__ MOD__ chapter utility
\**/
void
i29initutility (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29initutility");
    i29_reset (session, true);
    i29newsegment (session, sp1m_utility);
    i29newpart (session, sp1pk_command);
}

/*----------------------------------------*/

void
i29initdiagoutcopy (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MEO00 ("i29initdiagoutcopy");
    i29_reset (session, true);
    i29newsegment (session, sp1m_diag_outcopy);
    i29newpart (session, sp1pk_command);
} /* end i29initdiagoutcopy */

/*----------------------------------------*/

int
i29utility (
    tin01_sql_session * session,
    tin01_cstr cmd)
{
    ROUTINE_DBG_MSP00 ("i29utility");
    i29initutility (session);
    i29pcmd (session, cmd, UNDEF_SP00);
    return i29sql (session, NULL);
}

/*----------------------------------------*/

int
i29utilcmd (
    tin01_sql_session * session,
    tin01_c_sqlca * sqlca,
    tin01_cstr cmd,
    ...)
{
    ROUTINE_DBG_MSP00 ("i29utilcmd");
    va_list         args;
    int             i;
    int             rc;
    int             paramCount;

    /* execute command */
    rc = i29utility (session, cmd);
    if (rc == cin01_db_ok) {
        /* get c variables addresses */
        va_start (args, cmd);
        for (i = 0; i < sqlca->varCount; ++i) {
            sqlca->hostVar[i].addr = va_arg (args, void *);
        }
        va_end (args);
        /* get sql parameter info */
        paramCount = i29paramcount (session);
        if (paramCount > sqlca->varCount)
            paramCount = sqlca->varCount;
/*            rc = -1200; */         /* too few values */
/*        else { */
            sqlca->paramCount = paramCount;
            if (paramCount > 0) {
                tsp1_param_info *pinfo = i29gparaminfo (session, 0);
                int              i;

                for (i = 0; i < paramCount; ++i, ++pinfo) {
                    memcpy (&sqlca->hostVar [i].pinfo, pinfo,
                            sizeof (tsp1_param_info));
                }
            }
            sqlca->inUse = true;
/*        } */
        /* get sql parameters into c variables */
        if (rc == 0) {
            rc = i29_outargs (session, sqlca);
        }
    }
    return rc;
}

/*----------------------------------------*/
int
i29utilbackuprequest (
    tin01_sql_session  * session,
    boolean              bBackupState,
    tin01_cstr           cmd)
{
    ROUTINE_DBG_MSP00 ("i29utilbackuprequest");

    if (!bBackupState) {
        i29initutility (session);
    } else {
        i29initadbs (session);
    } /* end if */
    i29pcmd (session, cmd, UNDEF_SP00);
    return i29sqlrequest(session);
} /* end i29utilbackuprequest */

/*----------------------------------------*/
int
i29utilbackupreceive (
    tin01_sql_session  * session,
    tin01_c_sqlca      * sqlca,
    boolean            * bBackupRC,
    ...)
{
  ROUTINE_DBG_MSP00 ("i29utilbackupreceive");
  va_list         args;
  int             i;
  int             rc;
  int             paramCount;

  rc = i29sqlreceive(session);

  switch (rc) {
    case -7075:                       /* current save skipped           */
    case -7076:                       /* no more data to read from tape */
    case -8020:                       /* next volumn required           */
      *bBackupRC = true;
      break;
    default:
      *bBackupRC = false;
      break;
  } // end switch

  if (rc != -8888) {
    /* get c variables addresses */
    va_start (args, bBackupRC);
    for (i = 0; i < sqlca->varCount; ++i) {
      sqlca->hostVar[i].addr = va_arg (args, void *);
    } // end for
    va_end (args);

    /* get sql parameter info */
    paramCount = i29paramcount (session);
    if (paramCount > sqlca->varCount) {
      paramCount =  sqlca->varCount;
    } // end if

    sqlca->paramCount = paramCount;
    if (paramCount > 0) {
      tsp1_param_info *pinfo = i29gparaminfo (session, 0);
      int              i;

      for (i = 0; i < paramCount; ++i, ++pinfo) {
        memcpy (&sqlca->hostVar [i].pinfo, pinfo,
        sizeof (tsp1_param_info));
      } // end for
    } // end if

    sqlca->inUse = true;

    /* get sql parameters into c variables */
    if (rc == cin01_db_ok) {
      rc = i29_outargs (session, sqlca);
    } else {
      i29_outargs (session, sqlca);
    } // end if

  } // end if

  return rc;
} /* end i29utilbackupreceive */

/*----------------------------------------*/

int
i29utilbackupcmd (
    tin01_sql_session  * session,
    tin01_c_sqlca      * sqlca,
    boolean           * bBackupRC,
    boolean           bBackupState,
    tin01_cstr         cmd,
    ...)
{
    ROUTINE_DBG_MSP00 ("i29utilbackupcmd");
    va_list         args;
    int             i;
    int             rc;
    int             paramCount;
    int             nBackupRC;

    /* execute command */
    if (!bBackupState) {
      rc = i29utility (session, cmd);
    } else {
      rc = i29adbs (session, cmd);
    }

    switch (rc) {
      case -7075:                       /* current save skipped           */
      case -7076:                       /* no more data to read from tape */
      case -8020:                       /* next volumn required           */
        *bBackupRC = true;
        break;
      default:
        *bBackupRC = false;
        break;
    }

    if (*bBackupRC) {
      nBackupRC = rc;
    } else {
      nBackupRC = 0;
    }

    if ( (rc == cin01_db_ok)   ||
         (*bBackupRC)    ) {
        /* get c variables addresses */
        va_start (args, cmd);
        for (i = 0; i < sqlca->varCount; ++i) {
            sqlca->hostVar[i].addr = va_arg (args, void *);
        }
        va_end (args);
        /* get sql parameter info */
        paramCount = i29paramcount (session);
        if (paramCount > sqlca->varCount)
            paramCount = sqlca->varCount;
/*            rc = -1200; */        /* too few values */
/*        else { */
            sqlca->paramCount = paramCount;
            if (paramCount > 0) {
                tsp1_param_info *pinfo = i29gparaminfo (session, 0);
                int              i;

                for (i = 0; i < paramCount; ++i, ++pinfo) {
                    memcpy (&sqlca->hostVar [i].pinfo, pinfo,
                            sizeof (tsp1_param_info));
                }
            }
            sqlca->inUse = true;
/*        } */
        /* get sql parameters into c variables */
        if ( (rc == 0)            ||
             (*bBackupRC)    ){
            rc = i29_outargs (session, sqlca);
        }
    }

    if ( (rc == 0)            &&
         (*bBackupRC)    ) {
      return nBackupRC;
    } else {
      return rc;
    }
} // end i29utilbackupcmd

/*----------------------------------------*/

int i29utildiagtocmd
      ( tin01_sql_session  * session,
        tin01_cstr           cmd,
        long               * nPageSize,
        void              ** pPageBuf)
{
    ROUTINE_DBG_MEO00 ("i29utildiagtocmd");

  int             rc;

  /* init out params */
  *nPageSize = 0;
  *pPageBuf  = NULL;

  /* execute command */
  rc = i29utility (session, cmd);

  if (rc == cin01_db_ok) {
    /* if there no page part get the page(s) with a diag_outcopy message */
    if (i29findpart (session, sp1pk_data) && !i29findpart (session, sp1pk_page)) {
      i29initdiagoutcopy (session);
      rc = i29sql (session, NULL);
    } /* end if */
    if (rc == cin01_db_ok) {
      /* page part */
      if (i29findpart (session, sp1pk_page)) {
        /* get page */
        *nPageSize = session->part->sp1p_buf_len ();
        *pPageBuf  = (void *) session->part->sp1p_buf ().asCharp ();
      } /* end if */
    } /* end if */

  } /* end if */

  return rc;

} /* end i29utildiagtocmd  */

/*----------------------------------------*/

int i29utildiagrestcmd
      ( tin01_sql_session  * session,
        tin01_cstr           cmd,
        long                 nPageSize,
        const void         * pPageBuf)
{
    ROUTINE_DBG_MEO00 ("i29utildiagrestcmd");

  // init segment and commad part
  i29initutility (session);
  i29pcmd (session, cmd, UNDEF_SP00);

  // init pagepart and commad part
  i29newpart (session, sp1pk_page);
  i29findpart (session, sp1pk_page);
  session->part->sp1p_buf_len() = nPageSize;
  memcpy(session->part->sp1p_buf ().asCharp (), pPageBuf, nPageSize);

  return i29sql (session, NULL);

} /* end i29utildiagrestcmd  */

/*----------------------------------------*/
/**\
--- ----------------------------------MF__ MOD__ chapter communication
\**/
void
i29pascalstring (
    char       *target,
    int         targetlen,
    const char *source,
    int         sourcelen)
{
    ROUTINE_DBG_MSP00 ("i29pascalstring");
    if (source == NULL) {
        /* one blank so memcpy is ok */
        source = " ";
        sourcelen = 1;
    }
    else {
        if (sourcelen == -1)
            sourcelen = (int)strlen (source);
    }
    if (sourcelen > targetlen)
        memcpy (target, source, (int) targetlen);
    else {
        memcpy (target, source, (int) sourcelen);
        memset (target + sourcelen, ' ', (int) (targetlen - sourcelen));
    }
}

/*----------------------------------------*/

tin01_sql_session *
i29lasterr_off (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29lasterr_off");
    session->lasterr.lasterr_on = false;
    return session;
} /* i29lasterr_off */

/*----------------------------------------*/

tin01_sql_session *
i29lasterr_on (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29lasterr_on");
    session->lasterr.lasterr_on = true;
    return session;
} /* i29lasterr_on */

/*----------------------------------------*/

void
i29setlasterr_rte (
    tin01_sql_session * session,
    tsp00_ErrText     errtext,
    tsp01_CommErr sp_rc)
{
    ROUTINE_DBG_MSP00 ("i29setlasterr_rte");
    ASSERT_DBG (session != NULL);
    if (!session->lasterr.lasterr_on)
        return;
    session->lasterr.sp_rc = sp_rc;
    if (sp_rc == commErrOk_esp01)
        return;
    MEMCPY_MIN01 (session->lasterr.errtext, errtext);
    memset (session->lasterr.errname, ' ',
            sizeof (session->lasterr.errname));
}

/*----------------------------------------*/

void
i29resetlasterr (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29resetlasterr");
    ASSERT_DBG (session != NULL);
    session->lasterr.sqlresult.returnCode = cin01_db_ok;
    session->lasterr.sqlresult.errorPos = 0;
    memset (session->lasterr.sqlresult.sqlmsg, ' ',
            sizeof (session->lasterr.sqlresult.sqlmsg));
    memset (session->lasterr.errname, ' ',
            sizeof (session->lasterr.errname));
}

/*----------------------------------------*/

void
i29whenever_errpos (
    tin01_sql_session    *session,
    int                 whenever_pos)
{
    ROUTINE_DBG_MSP00 ("i29whenever_errpos");
    tin01_sqlresult  *sqlresult;
    if (whenever_pos == UNDEF_SP00)
        return;
    sqlresult = &session->lasterr.sqlresult;
    if (sqlresult->errorPos >= whenever_pos)  {
        sqlresult->errorPos = -(sqlresult->errorPos - whenever_pos);
    }
}

/*----------------------------------------*/
int
i29replyavailable (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29replyavailable");
    tsp00_ErrText     errtext;
    tsp01_CommErr  sp_rc;

    sqlareplyavailable (session->reference, errtext, &sp_rc);

    i29setlasterr_rte (session, errtext, sp_rc) ;
    return sp_rc;
} /* end i29replyavailable */

/*----------------------------------------*/
void
i29cancel (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29cancel");
    sqlacancel(session->reference);

} /* end i29cancel */

/*----------------------------------------*/

tsp01_CommErr
i29request (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29request");
    tsp00_ErrText     errtext;
    tsp01_CommErr  sp_rc;

    ASSERT_DBG (session->is_connected);
    sqlarequest (session->reference, session->send_packet,
        i29packetlen (session), errtext, &sp_rc);
    if ((sp_rc == commErrTimeout_esp01) || (sp_rc == commErrCrash_esp01)) {
        i29cleanup_session (session);
        /*
        sqlarelease (session->reference);
        session->is_connected = false;
        */
    }
    if (sp_rc != commErrOk_esp01) {
        TRACEF_MIN01 (5, (SQLTRACE, "Request error: %40.40s\n",
                errtext));
        if (sp_rc == commErrTimeout_esp01)
            i29connect (session);
    }
    i29setlasterr_rte (session, errtext, sp_rc) ;
    return sp_rc;
}

/*----------------------------------------*/

tsp01_CommErr
i29receive (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29receive");
    tsp00_Int4        length;
    tsp00_ErrText     errtext;
    tsp01_CommErr  sp_rc;

    ASSERT_DBG (session->is_connected);
    sqlareceive (session->reference, (void**)&session->rec_packet,
        &length, errtext, &sp_rc);
    if ((sp_rc == commErrTimeout_esp01) || (sp_rc == commErrCrash_esp01)) {
        sqlarelease (session->reference);
        session->is_connected = false;
    }
    if (sp_rc != commErrOk_esp01) {
        TRACEF_MIN01 (5, (SQLTRACE, "Receive error: %40.40s\n",
                errtext));
        session->rec_packet = NULL;
    }
    i29setlasterr_rte (session, errtext, sp_rc) ;
    return sp_rc;
}

/*----------------------------------------*/

void
i29callsql (
    tin01_sql_session * session,
    tsp01_CommErr * comm_error)
{
    ROUTINE_DBG_MSP00 ("i29callsql");
    *comm_error = i29request (session);
    if (*comm_error == commErrOk_esp01)
        *comm_error = i29receive (session);
    i29lasterr_on (session);
}

/*----------------------------------------*/

void
i29setlasterr_name (
    tin01_sql_session    * session,
    tsp00_KnlIdentifier  & errname)
{
    ROUTINE_DBG_MSP00 ("i29setlasterr_name");
    ASSERT_DBG (session != NULL);
    MEMCPY_MIN01 (session->lasterr.errname, errname);
}

/*----------------------------------------*/

void
i29sqllasterr (
    tin01_sql_session     * session,
    tsp00_ErrTextc        & errtext,
    tsp00_KnlIdentifierc  & errname,
    tsp00_Int2            * errcode,
    tsp00_Int2            * errpos)
{
    ROUTINE_DBG_MSP00 ("i29sqllasterr");
    #define RTEERR_OFFSET   10000
    * errpos = 0;
    *errcode = 0;
    memset(errtext, 0, sizeof(tsp00_ErrTextc));
    memset(errname, 0, sizeof(tsp00_KnlIdentifierc));
    if (session->lasterr.sp_rc != commErrOk_esp01) {
        memcpy (errtext, session->lasterr.errtext, sizeof (tsp00_ErrText));
        memcpy (errname, session->lasterr.errname, sizeof (tsp00_KnlIdentifier));
        *errcode = (tsp00_Int2) (session->lasterr.sp_rc + RTEERR_OFFSET);
        session->lasterr.sp_rc.becomes(commErrOk_esp01);
    }
    else if (session->lasterr.sqlresult.returnCode != cin01_db_ok) {
        *errcode = session->lasterr.sqlresult.returnCode;
        *errpos = session->lasterr.sqlresult.errorPos;
        memcpy (errtext, session->lasterr.sqlresult.sqlmsg, sizeof (tsp00_ErrText));
        memcpy (errname, session->lasterr.errname, sizeof (tsp00_KnlIdentifier));
        session->lasterr.sqlresult.returnCode = cin01_db_ok;
    }
    else {
        *errcode = 0;
        memset (errtext, ' ', sizeof (tsp00_ErrText));
        memset (errname, ' ', sizeof (tsp00_KnlIdentifier));
    }
}

/*----------------------------------------*/

int
i29sqlconnect (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29sqlconnect");

    tsp00_TaskId      aid;
    tsp4_xuser_record   *xuser;
    int                 connect_rc;
    tsp00_Int4          packetSize;
    tsp00_ErrText       errtext;
    tsp01_CommErr       sp_rc;
    tsp01_Service       connectKind;
    void               *packetList[1];

    sqluid (&aid);
    xuser = &session->xuser;
    if (session->as_utility) {
        connectKind.becomes(srvUtility_esp01);
    }
    else {
        connectKind.becomes(srvUser_esp01);
    }
    sqlaconnect (aid, xuser->xu_servernode, xuser->xu_serverdb,
        connectKind, 1, &session->reference, &packetSize,
        packetList, errtext, &sp_rc);
    if (sp_rc != commErrOk_esp01) {
        TRACEF_MIN01 (5, (SQLTRACE, "Connect error: %40.40s\n",
                errtext));
       i29setlasterr_rte (session, errtext, sp_rc) ;
    }
    else {
        tsp00_TermId      termid;

        ASSERT_DBG (!session->is_connected);
        session->send_packet = (tsp1_packet_ptr ) packetList [0];
        session->send_packet->sp1_header.sp1h_varpart_size =
            packetSize - sizeof (tsp1_packet_header);
        session->is_connected = true;
        sqltermid (termid);
        i29connectcommand (session, termid);
        connect_rc = i29sql (session, NULL);
        if (connect_rc != cin01_db_ok) {
            sqlarelease (session->reference);
            session->is_connected = false;
            TRACEF_MIN01 (5, (SQLTRACE, "Connect command failed with %d\n",
                (int) connect_rc));
        }
        return connect_rc;
    }
    return commErrNotOk_esp01;
}

/*----------------------------------------*/

int
i29_connect (
    tin01_sql_session * session,
    tsp01_Service       connectKind)
{
    ROUTINE_DBG_MSP00 ("i29_connect");

    tsp00_TaskId      aid;
    tsp4_xuser_record   *xuser;
    tsp00_Int4          packetSize;
    tsp00_ErrText       errtext;
    tsp01_CommErr       sp_rc;
    void              *packetList [1];

    sqluid (&aid);
    xuser = &session->xuser;

    sqlaconnect (aid, xuser->xu_servernode, xuser->xu_serverdb,
        connectKind, 1, &session->reference, &packetSize,
        packetList, errtext, &sp_rc);

    if (sp_rc != commErrOk_esp01) {

      TRACEF_MIN01 (5, (SQLTRACE, "Connect error: %40.40s\n", errtext));
      i29setlasterr_rte (session, errtext, sp_rc) ;

    } else {

      ASSERT_DBG (!session->is_connected);
      session->send_packet = (tsp1_packet_ptr ) packetList [0];
      session->send_packet->sp1_header.sp1h_varpart_size = packetSize - sizeof (tsp1_packet_header);
      session->is_connected = true;

    }
    return sp_rc;
} /* end i29_connect */

/*----------------------------------------*/

int
i29specialconnect (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29specialconnect");

    tsp01_Service       connectKind;

    connectKind.becomes(srvUser_esp01);

    return i29_connect( session, connectKind);
} /* end i29specialconnect */

/*----------------------------------------*/

int
i29eventconnect (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29eventconnect");

    tsp01_Service       connectKind;

    connectKind.becomes(srvEvent_esp01);

    return i29_connect( session, connectKind);
} /* end i29eventconnect */

/*----------------------------------------*/

int
i29eventwait (
    tin01_sql_session       * session,
    tsp31_event_description * event)
{
  ROUTINE_DBG_MSP00 ("i29eventwait");

  int rc;

  // packet initialisieren
  i29initspecial(session, sp1m_wait_for_event);

  // request absetzen und ins receive gehen
  rc = i29sql (session, NULL);

  // struktur auswerten
  if (rc == cin01_db_ok) {
    if (i29findpart (session, sp1pk_data)) {
      memcpy(event, session->part->sp1p_buf ().asCharp () + 1, sizeof(tsp31_event_description));

    } // end if
  } // end if

  return rc;
} /* end i29eventwait */

/*----------------------------------------*/

int
i29connect (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29connect");
    return i29sqlconnect ( session );
}

/*----------------------------------------*/

void
i29cleanup_session (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29cleanup_session");
    sqlarelease (session->reference);
    session->is_connected = false;
    i29clearpacketref (session);
}

/*----------------------------------------*/

void
i29release (
    tin01_sql_session * session,
    tin01_bool with_commit)
{
    ROUTINE_DBG_MSP00 ("i29release");
    int oldMode;

    oldMode = i29adabasmode (session);
    IFDEBUG_MIN01 (i29lasterr_off (session));
    IFDEBUG_MIN01 (i29adbs_noinfo (session, "VTRACE"));
    i29oldmode (session, oldMode);
    i29lasterr_off (session);
    if (with_commit)
        i29adbs_noinfo (session, "COMMIT WORK RELEASE");
    else
        i29adbs_noinfo (session, "ROLLBACK WORK RELEASE");
    i29cleanup_session (session);
}

/*----------------------------------------*/

void
i29specialrelease (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("i29specialrelease");

    i29cleanup_session(session);
}

/*----------------------------------------*/

void
i29getFreePartInfo (
    tin01_sql_session * session,
    void ** dataPtr,
    int   * bufPos,
    int   * freeLen)
{
    ROUTINE_DBG_MSP00 ("i29getFreePartInfo");
    *dataPtr = session->part->sp1p_buf () + session->part->sp1p_buf_len ();
    *bufPos = session->part->sp1p_buf_len () + 1 /* is pascal index */;
    *freeLen = i29bytesfree (session);
}

/*----------------------------------------*/

static int
i29_handleCommError (
    tin01_sql_session * session,
    tin01_sqlresult * sqlresult,
    tsp01_CommErr  comm_error)
{
    ROUTINE_DBG_MSP00 ("i29_handleCommError");
    int result = cin01_db_ok;

    if ((int) comm_error != commErrOk_esp01) {
        result = cin01_db_not_accessible;
        /* session->is_connected = false; lower level routines try reconnect */
        if (sqlresult != NULL) {
            sqlresult->returnCode = result;
            M_WARNINGSET (sqlresult->warnings) = EMPTYSET_MIN01;
            sqlresult->errorPos = 0;
            sqlresult->rowCount = 0;
            MEMCPY_MIN01 (sqlresult->sqlstate, "I8888");
            sqlresult->sqlmsg[0] = '\0';
        }
    }
    return result;
}

/*----------------------------------------*/

static int
i29_handleSQLError (
    tin01_sql_session * session,
    tin01_sqlresult * sqlresult)
{
    ROUTINE_DBG_MSP00 ("i29_handleSQLError");
    int             result;
    tsp1_segment   *segment;
    int             msglen;

    i29resetpackets (session, false);
    IFTRACE_MIN01 (4, j96dumppacket (session, "RECEIVE", SQLTRACE));
    segment = i29_lastsegment (session->rec_packet);
    result = segment->sp1r_returncode ();
    /* to lasterr, only store errors: */
    if ((sqlresult != NULL) &&
        ((result != cin01_db_ok) ||
            (sqlresult != &(session->lasterr.sqlresult)))) {
        sqlresult->returnCode = result;
        M_WARNINGSET(sqlresult->warnings) = M_WARNINGSET
            (segment->sp1r_extern_warning ());
        sqlresult->errorPos = segment->sp1r_errorpos ();
        if (result == cin01_db_row_not_found) {
            sqlresult->rowCount = 0;
        }
        else {
            sqlresult->rowCount = i29resultcount (session);
        }
        MEMCPY_MIN01 (sqlresult->sqlstate, segment->sp1r_sqlstate ());
        if (result != cin01_db_ok) {
            if (i29findpart (session, sp1pk_errortext)) {
                memcpy (sqlresult->sqlmsg, session->part->sp1p_buf ().asCharp (),
                    sizeof (sqlresult->sqlmsg) - 1);
                msglen = session->part->sp1p_buf_len ();
                sqlresult->sqlmsg[msglen] = '\0';
            }
        }
    }
    return result;
}

/*----------------------------------------*/

int
i29requestForPython (
    tin01_sql_session * session,
    tin01_sqlresult * sqlresult)
{
    ROUTINE_DBG_MSP00 ("i29requestForPython");
    tsp01_CommErr  comm_error;
    tsp00_Int2        result;

    if (sqlresult == NULL)
        sqlresult = &(session->lasterr.sqlresult);
    if (!session->is_connected) {
        TRACE_MIN01 (4, SQLTRACE, "No connection: command skipped");
        result = cin01_db_not_accessible;
        if (sqlresult != NULL) {
            sqlresult->returnCode = result;
            M_WARNINGSET (sqlresult->warnings) = EMPTYSET_MIN01;
            sqlresult->errorPos = 0;
            sqlresult->rowCount = 0;
            MEMCPY_MIN01 (sqlresult->sqlstate, "I8888");
            sqlresult->sqlmsg[0] = '\0';
        }
        return result;
    }
    session->rec_packet = NULL;
    i29closesend (session);
    IFTRACE_MIN01 (4, j96dumppacket (session, "REQUEST", SQLTRACE));
    comm_error = i29request (session);
    result = i29_handleCommError (session, sqlresult, comm_error);
    return result;
}

/*----------------------------------------*/

int
i29receiveForPython (
    tin01_sql_session * session,
    tin01_sqlresult * sqlresult)
{
    ROUTINE_DBG_MSP00 ("i29receiveForPython");
    tsp01_CommErr  comm_error;
    tsp00_Int2        result;

    if (sqlresult == NULL)
        sqlresult = &(session->lasterr.sqlresult);
    comm_error = i29receive (session);
    i29lasterr_on (session);
    result = i29_handleCommError (session, sqlresult, comm_error);
    if (result == cin01_db_ok) {
        result = i29_handleSQLError (session, sqlresult);
    }
    return result;
}

/*----------------------------------------*/

