/*
===========================================
   responsible:    DanielD
   description:    Interface to SAP DB
   target:         Python
   last changed:   27.11.2000
   copyright:      2000 SAP AG


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

===========================================*/


#include <stddef.h>

#if defined (HPUX) || defined (AIX) || defined (SUN) || defined (OSF1) || defined (LINUX)
#define BAD_STATIC_FORWARD
#endif

#include "Python.h"/* no check */

#if defined (HPUX) || defined (AIX) || defined (SUN) || defined (OSF1)
#undef statichere
#define statichere
#endif

#define MF__
#if defined (__cplusplus)
#define externC extern "C"
#define internC extern "C" static
#else
#define externC
#define internC static
#endif
#define PYTHON_GLUE 1

#if defined (__cplusplus)
#define LIT(c) CONST_CAST(char *, c)
#else
#define LIT(c) (c)
#endif

#if defined (_WIN32)
#define GLUEEXPORT __declspec(dllexport)
#else
#define GLUEEXPORT
#endif

static PyObject *
createExceptionKind (
    const char * exceptionName,
    const char * exceptionCode)
{
    PyObject * result;
    PyObject * runResult = NULL;
    PyObject * globals = NULL;
    PyObject * locals = NULL;

    if (exceptionCode != NULL) {
        globals = PyDict_New ();
        locals = PyDict_New ();
        runResult = PyRun_String ((char *) exceptionCode, Py_file_input,
            globals, locals);
        if (PyErr_Occurred ()) {
            PyErr_Print ();
        }
    }
    result = PyErr_NewException ((char *) exceptionName, NULL, locals);
    Py_XDECREF (locals);
    Py_XDECREF (globals);
    Py_XDECREF (runResult);
    return result;
}

static PyObject *CommunicationErrorType;
static const char * CommunicationErrorCodeC = 
    "\n"
    "def __str__ (self):\n"
    "    return \"sapdb.CommunicationError: [%d] %s\" % (self.errorCode, self.message)\n"
    "\n";
static PyObject *SQLErrorType;
static const char * SQLErrorCodeC = 
    "\n"
    "def __str__ (self):\n"
    "    return \"sapdb.SQLError: [%d] %s\" % (self.errorCode, self.message)\n"
    "\n";
/* snippet Generic_precode */
#include <ctype.h>

#if defined (PERL_GLUE)
    /* conflict with gsp00 enum */
    #undef na
    /* conflict with malloc.h realloc included later */
    #undef realloc
#endif
#if defined (__cplusplus)
#include "heo02.h"
#else
#include "heo102.h"
#endif
#include "hsp02.h"
#include "hsp30.h"
#include "hsp40.h"
#include "hsp41.h"
#include "hsp42.h"
#include "hsp77.h"
#include "hsp100.h"
#include "hin28.h"

#if defined (PERL_GLUE)
    #include "patchlevel.h"/* no check */
    #if PATCHLEVEL >= 5
        #define na PL_na
    #else
        #define na Perl_na
    #endif
    #undef read
#endif

#if defined (PYTHON_GLUE)
#include <compile.h>
#include <frameobject.h>
typedef PyObject SL_ValueT;
#elif defined (PERL_GLUE)
typedef SV SL_ValueT;
#else
    #error Invalid Glue Environment
#endif

/*----------------------------------------*/

static void c2p (
    void * instanceData,
    int    instanceSize,
    const char * str)
{
    int len = strlen (str);

    if (len > instanceSize)
        len = instanceSize;
    memcpy (instanceData, str, len);
    if (len < instanceSize)
        memset ((char *) instanceData + len, ' ', instanceSize - len);
}

static void * tracingMalloc (size_t len)
{
    void * result = malloc (len);
    printf("malloc: 0x%p\n", result);
    return result;
}

static void tracingFree (void * ptr)
{
    printf("free: 0x%p\n", ptr);
    free (ptr);
}

/*
#define malloc(len) tracingMalloc(len)
#define free(ptr)   tracingFree(ptr)
*/

#if defined (__cplusplus) && (COMPILEMODE_MSP00 >= SLOW_MSP00)
    /*!
      Class: ScriptingTracer
     */
    class ScriptingTracer
    {
    public:
        ScriptingTracer (const char * name);
        ~ScriptingTracer ();
        void trace (int layer, const char *format, ...);
        static void switchTrace (bool on);
        static void initTrace ();
    private:
        const char * procName;
        static bool traceIsOn;
    };

    ScriptingTracer::ScriptingTracer (const char * name)
    {
        if (traceIsOn) {
            printf("<%s>\n", name);
        }
        this->procName = name;
    }

    ScriptingTracer::~ScriptingTracer ()
    {
        if (traceIsOn) {
            printf("</%s>\n", this->procName);
        }
    }

    void ScriptingTracer::trace (int layer, const char *format, ...)
    {
        if (traceIsOn) {
            va_list         ap;

            va_start (ap, format);
            vprintf(format, ap);
            va_end (ap);
        }
    }

    bool ScriptingTracer::traceIsOn = false;

    void ScriptingTracer::switchTrace (bool on)
    {
        traceIsOn = on;
    }

    void ScriptingTracer::initTrace ()
    {
        const char * sapdbTrace = getenv ("SAPDBTRACE");
        if (sapdbTrace == NULL) {
            traceIsOn = false;
        }
        else {
            traceIsOn = true;
        }
    }
    /*! EndClass: ScriptingTracer */

    #undef TRACER_CLASS_MSP00
    #define TRACER_CLASS_MSP00 ScriptingTracer

#endif

#if defined (__cplusplus)
#define REFPARM(name)   name
#define INLINE          inline
#else
#define REFPARM(name)   &name
#define INLINE

#undef ROUTINE_DBG_MSP00
#define ROUTINE_DBG_MSP00(lit) static const char * funcname_glue_ = lit

#define REINTERPRET_CAST(type, expr) ((type) expr)
#define STATIC_CAST(type, expr) ((type) expr)
#define CONST_CAST(type, expr) ((type) expr)
#endif

char * buildInfo ()
{
    static tsp00_Versionc versionString;
    static bool isInitialized;
#if defined (PYTHON_GLUE)
    tsp100_CompName compName = "sapdbpy  ";
#elif defined (PERL_GLUE)
    tsp100_CompName compName = "sapdbperl";
#else
    #error must be either Python or Perl
#endif

    if (!isInitialized) {
        sp100_GetVersionString (compName, s100buildnumber, versionString);
    }
    return versionString;
}


static const int internalErrorC = -16506;

typedef unsigned char byte_t;

typedef struct ParseInfoT {
    tin01_parseid     pid;
    int               paramCount;
    int               outputParamCount;
    int               inLongCount;
    int               outLongCount;
    tsp1_param_info * paramInfo;
    char            * sqlString;
} ParseInfoT;

typedef struct LongReaderT LongReaderT;

static SL_ValueT *
SL_fromString (
    const char * value,
    int len);

static int
SL_readCall (
    SL_ValueT * value,
    byte_t * streamBuf,
    int requestedLen);

static SL_ValueT *
readLong (
    LongReaderT * nself,
    int requested);

static bool
fillParamInfos (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    bool initial);

static void
increaseRef (
    void * obj);

static void
decreaseRef (
    void * obj);

static void
setAutocommit (
    tin01_sql_session * session,
    bool autocommit);

static bool
getAutocommit (
    tin01_sql_session * session);

static bool
lastWasEOT (
    tin01_sql_session * session);

static void
enableCursorNames (
    tin01_sql_session * session,
    bool enable);

static int
nextCursorCount (
    tin01_sql_session * session);

static void
addCursorName (
    tin01_sql_session * session);

static void
clearEOTFlag (
    tin01_sql_session * session);

static void
setEOTFlag (
    tin01_sql_session * session);



/*----------------------------------------*/

static void
clearParseInfo (
    tin01_sql_session * session,
    ParseInfoT * parseInfo)
{
    ROUTINE_DBG_MSP00 ("clearParseInfo");
    if ((session != NULL) && session->is_connected) {
        if (parseInfo->paramCount != -1) {
            i28droppid (session, parseInfo->pid);
        }
    }
    if (parseInfo->paramInfo != NULL) {
        free (parseInfo->paramInfo);
        parseInfo->paramInfo = NULL;
    }
    parseInfo->paramCount = -1;
    if (parseInfo->sqlString != NULL) {
        free (parseInfo->sqlString);
    }
}

/*----------------------------------------*/

typedef struct PreparedT {
    tin01_sql_session * session;
    ParseInfoT        parseInfo;
} PreparedT;

typedef enum fetchKindsT {
    nextFetchE, relativeFetchE, absoluteFetchE,
    firstFetchE, lastFetchE,
    fetchKindsE
} fetchKindsT;

static const char * fetchCommandsC [] = {
    " next ", " relative ? ", " pos (?) ", " first ", " last "
};

typedef struct ResultSetT {
    tin01_sql_session * session;
    tsp00_KnlIdentifierc cursorName;
    ParseInfoT        fetchInfo [fetchKindsE];
    byte_t            * dataBuf;
    int               dataSize;
    int               dataLen;
    byte_t            * currentRow;
    byte_t            * eoRecords;
    int               rowSize;
    int               maxRecordsPerFetch;
    const char      * colNames;
    tsp1_param_info * colInfo;
    int               colCount;
    int               currentPos;
} ResultSetT;

/*----------------------------------------*/

static int dbgSizes [] = {36, 256, 1111};

static void
checkMem ()
{
    ROUTINE_DBG_MSP00 ("checkMem");
    int i;
    for (i = 0; i < 3; ++i) {
        void * p  = malloc (dbgSizes [i]);
        free (p);
    }
}

/*----------------------------------------*/

static const int atEndStreamC = -2;
static const int errorStreamC = -1;

struct InputStreamVMT;

typedef struct InputStreamVMT* InputStreamT;

typedef int readFunctionT (InputStreamT * instream, byte_t * buf, int requestedLen);

typedef struct InputStreamVMT {
    readFunctionT * read;
} InputStreamVMT;

/*----------------------------------------*/

typedef struct StringInputStreamT {
    const InputStreamVMT * vmt;
    const char * data;
    const char * eoData;
    const char * pos;
} StringInputStreamT;

/*----------------------------------------*/

static int
stringInputRead (
    StringInputStreamT * self,
    byte_t * buf,
    int requestedLen)
{
    ROUTINE_DBG_MSP00 ("stringInputRead");
    int copyLen = requestedLen;
    int remaining = self->eoData - self->pos;

    if (requestedLen <= 0) {
        return errorStreamC;
    }
    if (remaining <= 0) {
        return atEndStreamC;
    }
    if (copyLen > remaining) {
        copyLen = remaining;
    }
    memcpy (buf, self->pos, copyLen);
    self->pos += copyLen;
    return copyLen;
}

/*----------------------------------------*/

static const InputStreamVMT StringInputStreamVMT_Imp = {
    (readFunctionT*) stringInputRead,
};

/*----------------------------------------*/

static InputStreamT *
initStringInputStream (
    StringInputStreamT * self,
    const char * data,
    int len)
{
    ROUTINE_DBG_MSP00 ("initStringInputStream");
    self->vmt = &StringInputStreamVMT_Imp;
    self->data = data;
    self->pos = data;
    self->eoData = data + len;
    return(InputStreamT*) self;
}

typedef struct CallableInputStreamT {
    const InputStreamVMT * vmt;
    SL_ValueT * callable;
    bool atEnd;
} CallableInputStreamT;

/*----------------------------------------*/

static int
callableInputRead (
    CallableInputStreamT * self,
    byte_t * buf,
    int requestedLen)
{
    ROUTINE_DBG_MSP00 ("callableInputRead");
    if (self->atEnd) {
        return atEndStreamC;
    }
    return SL_readCall (self->callable, buf, requestedLen);
}

/*----------------------------------------*/

static const InputStreamVMT CallableInputStreamVMT_Imp = {
    (readFunctionT*) callableInputRead,
};

/*----------------------------------------*/

static InputStreamT *
initCallableInputStream (
    CallableInputStreamT * self,
    SL_ValueT * callable)
{
    ROUTINE_DBG_MSP00 ("initStringInputStream");
    self->vmt = &CallableInputStreamVMT_Imp;
    self->callable = callable;
    self->atEnd = false;
    return(InputStreamT*) self;
}

/*----------------------------------------*/

struct LongReaderT {
    tin01_sql_session  * session;
    tsp00_LongDescriptor descriptor;
    byte_t               * dataBuf;
    int                  dataSize;
    int                  dataPos;
    int                  dataLen;
    bool                 lastGetvalFound;
};

/*----------------------------------------*/

bool keywordMatch (
    const char * kwArg,
    int kwArgLen,
    const char * token,
    int tokenLen)
{
    ROUTINE_DBG_MSP00 ("keywordMatch");

    if (kwArgLen != tokenLen) {
        return false;
    }
    if (memcmp (kwArg, token, kwArgLen) == 0) {
        return true;
    }
    else {
        return false;
    }
}

/*----------------------------------------*/

void setStringConfig (
    char * configField,
    int fieldLen,
    const char * value,
    int valueLen)
{
    ROUTINE_DBG_MSP00 ("setStringConfig");

    if (fieldLen <= valueLen) {
        memcpy (configField, value, fieldLen);
    }
    else {
        memcpy (configField, value, valueLen);
        memset (configField + valueLen, ' ', fieldLen - valueLen);
    }
}

/*----------------------------------------*/

int parseInt (
    const char * token,
    int tokenLen)
{
    ROUTINE_DBG_MSP00 ("parseInt");
    int result = 0;
    int i;

    for (i = 0; i < tokenLen; ++i) {
        result *= 10;
        result += token [i] - '0';
    }
    return result;
}

/*----------------------------------------*/

bool parseBool (
    const char * token,
    int tokenLen,
    bool defaultValue)
{
    ROUTINE_DBG_MSP00 ("parseBool");
    bool result = defaultValue;

    if (memcmp (token, "true", tokenLen) == 0) {
        result = true;
    }
    else if (memcmp (token, "on", tokenLen) == 0) {
        result = true;
    }
    else if (memcmp (token, "1", tokenLen) == 0) {
        result = true;
    }
    else if (memcmp (token, "false", tokenLen) == 0) {
        result = false;
    }
    else if (memcmp (token, "off", tokenLen) == 0) {
        result = false;
    }
    else if (memcmp (token, "0", tokenLen) == 0) {
        result = false;
    }
    return result;
}

/*----------------------------------------*/

static void
handleConfigArgument (
    tin01_sql_session * session,
    const char * token,
    int tokenLen,
    const char * value,
    int valueLen)
{
    ROUTINE_DBG_MSP00 ("handleConfigArgument");

    if (keywordMatch ("version", 7, token, tokenLen)) {
        setStringConfig ((char *)session->senderid, sizeof (tsp00_C5),
            value, valueLen);
    }
    else if (keywordMatch ("component", 9, token, tokenLen)) {
        setStringConfig ((char *) session->senderid + sizeof (tsp00_C5), sizeof (tsp00_C3),
            value, valueLen);
    }
    else if (keywordMatch ("sqlmode", 7, token, tokenLen)) {
        setStringConfig ((char*)session->xuser.xu_sqlmode, sizeof (session->xuser.xu_sqlmode),
            value, valueLen);
    }
    else if (keywordMatch ("cachelimit", 10, token, tokenLen)) {
        session->xuser.xu_cachelimit = parseInt (value, valueLen);
    }
    else if (keywordMatch ("timeout", 7, token, tokenLen)) {
        session->xuser.xu_timeout = parseInt (value, valueLen);
    }
    else if (keywordMatch ("isolation", 9, token, tokenLen)) {
        session->xuser.xu_isolation = parseInt (value, valueLen);
    }
    else if (keywordMatch ("autocommit", 10, token, tokenLen)) {
        setAutocommit (session, parseBool (value, valueLen, false));
    }
    else if (keywordMatch ("cursornames", 11, token, tokenLen)) {
        enableCursorNames (session, parseBool (value, valueLen, true));
    }
}

/*----------------------------------------*/

static void
parseSessionConfig (
    tin01_sql_session * session,
    const char * configString)
{
    ROUTINE_DBG_MSP00 ("parseSessionConfig");
    const char * argStart;
    const char * nextConfig;
    bool atEnd = (configString [0] == '\0');
    int tokenLen;
    int argLen;

    while (!atEnd) {
        argStart = strchr (configString, '=');
        if (argStart == NULL) {
            return;
        }
        tokenLen = argStart - configString;
        ++argStart;
        nextConfig = strchr (argStart, '&');
        if (nextConfig == NULL) {
            atEnd = true;
            nextConfig = argStart + strlen (argStart);
        }
        argLen = nextConfig - argStart;
        handleConfigArgument (session, configString, tokenLen, argStart, argLen);
        configString = nextConfig + 1;
    }

}

/*----------------------------------------*/

/* endsnippet Generic_precode */
/* snippet Python_precode */

static void
dumpObjectInfo (
    void * obj,
    const char * comment)
{
    ROUTINE_DBG_MSP00 ("dumpObjectInfo");
    PyObject * pyObj = REINTERPRET_CAST (PyObject*, obj);
    printf("%s0x%0x (%s, %d)\n", comment, pyObj, pyObj->ob_type->tp_name, pyObj->ob_refcnt);
}

/*----------------------------------------*/

static void
increaseRef (
    void * obj)
{
    ROUTINE_DBG_MSP00 ("increaseRef");
    Py_INCREF (REINTERPRET_CAST (PyObject*, obj));
}

/*----------------------------------------*/

static void
decreaseRef (
    void * obj)
{
    ROUTINE_DBG_MSP00 ("decreaseRef");
    Py_DECREF(REINTERPRET_CAST (PyObject*, obj));
}

/*----------------------------------------*/

#define SL_None Py_None
#define SL_isNone(val) (val == SL_None)
#define SL_isTrue(val) PyObject_IsTrue (val)

/*----------------------------------------*/

static bool
    SL_isString (
    PyObject* value)
{
    ROUTINE_DBG_MSP00 ("SL_isString");
    return PyString_Check (value);
}

/*----------------------------------------*/

static bool
SL_isCallable (
    PyObject* value)
{
    ROUTINE_DBG_MSP00 ("SL_isCallable");
    bool result;

    if (PyCallable_Check (value)) {
        result = true;
    }
    else {
        result = false;
    }
    return result;
}

/*----------------------------------------*/

static void
SL_getRawString (
    PyObject* value,
    const char ** chars,
    int * len)
{
    if (PyString_Check (value)) {
        *chars = PyString_AsString (value);
        *len = PyString_Size (value);
    }
    else {
        PyObject * str = PyObject_Str (value);
        *chars = PyString_AsString (str);
        *len = PyString_Size (str);
        Py_DECREF (str);
    }
}

/*----------------------------------------*/

static int
SL_getNumber (
    PyObject * value,
    tsp1_param_info * paramInfo,
    tsp00_Number * number)
{
    ROUTINE_DBG_MSP00 ("SL_getNumber");
    enum {
        isInt, isFloat
    };
    long intValue;
    double floatValue;
    int valueKind;
    tsp00_NumError numrc;

    if (PyInt_Check (value)) {
        intValue = PyInt_AS_LONG (value);
        valueKind = isInt;
    }
    else if (PyLong_Check (value)) {
        floatValue = PyLong_AsDouble (value);
        valueKind = isFloat;
    }
    else if (PyFloat_Check (value)) {
        floatValue = PyFloat_AsDouble (value);
        valueKind = isFloat;
    }
    else {
        PyObject * floatObject = PyNumber_Float (value);
        if (floatObject == NULL) {
            return 0;
        }
        floatValue = PyFloat_AsDouble (value);
        valueKind = isFloat;
        Py_DECREF (floatObject);
    }
    if (valueKind == isInt) {
        s41plint (number, 1, csp_fixed, paramInfo->sp1i_frac, intValue, &numrc);
    }
    else {
        s41plrel (number, 1, csp_fixed, paramInfo->sp1i_frac, floatValue, &numrc);
    }
    return 1;
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromNone ()
{
    ROUTINE_DBG_MSP00 ("SL_fromNone");

    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromInt (
    long value)
{
    ROUTINE_DBG_MSP00 ("SL_fromInt");

    return PyInt_FromLong (value);
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromDouble (
    double value)
{
    ROUTINE_DBG_MSP00 ("SL_fromDouble");

    return PyFloat_FromDouble (value);
}

/*----------------------------------------*/

static SL_ValueT *
SL_fromString (
    const char * value,
    int len)
{
    ROUTINE_DBG_MSP00 ("SL_fromString");

    if (len == -1) {
        len = strlen (value);
    }
    return PyString_FromStringAndSize (value, len);
}

/*----------------------------------------*/

static int
SL_readCall (
    PyObject * value,
    byte_t * streamBuf,
    int requestedLen)
{
    ROUTINE_DBG_MSP00 ("SL_readCall");
    PyObject   * callResult;
    const char * resultData;
    int          resultLen;
    int          readResult;

    /* argTuple = Py_BuildValue ("i", requestedLen); */
    callResult = PyObject_CallFunction (value, LIT("(i)"), requestedLen);
    if ((callResult != NULL) && PyString_Check (callResult)) {
        resultData = PyString_AsString (callResult);
        resultLen = PyString_Size (callResult);
        if (resultLen == 0) {
            readResult = atEndStreamC;
        }
        else if (resultLen < 0) {
            readResult = errorStreamC;
        }
        else {
            if (resultLen > requestedLen) {
                readResult = requestedLen;
            }
            else {
                readResult = resultLen;
            }
            memcpy (streamBuf, resultData, readResult);
        }
    }
    else {
        readResult = errorStreamC;
    }
    Py_XDECREF (callResult);
    return readResult;
}

/*----------------------------------------*/

static void
    SL_forget (
    PyObject * value)
{
    Py_DECREF (value);
}

/*----------------------------------------*/

typedef struct SequenceT {
    PyObject * pySeq;
} SequenceT;

static SequenceT EmptySequenceC = {NULL};

static int
sequencePy2C (
    void * dummy,
    PyObject * pySeq,
    SequenceT * cSeq)
{
    ROUTINE_DBG_MSP00 ("sequencePy2C");
    cSeq->pySeq = pySeq;
    return 1;
}

/*----------------------------------------*/

static int
sequenceLength (
    SequenceT * cSeq)
{
    ROUTINE_DBG_MSP00 ("sequenceLength");
    int result;

    if (cSeq->pySeq == NULL) {
        result = 0;
    }
    else {
        result = PyObject_Length (cSeq->pySeq);
    }
    return result;
}

/*----------------------------------------*/

static PyObject *
getSequenceElement (
    SequenceT * cSeq,
    int index)
{
    ROUTINE_DBG_MSP00 ("getSequenceElement");
    PyObject * result;

    PyObject * key = PyInt_FromLong (index);
    result = PyObject_GetItem (cSeq->pySeq, key);
    Py_DECREF (key);
    return result;
}

/*----------------------------------------*/

typedef struct SQLResultT {
    PyObject * resultObj;
    int isTuple;
    int isOK;
    int additionalValues;
} SQLResultT;

static int
sqlResultC2Py (
    void * dummy,
    SQLResultT result,
    PyObject ** pyResult)
{
    ROUTINE_DBG_MSP00 ("sqlResultC2Py");
    *pyResult = result.resultObj;
    return 1;
}

/*----------------------------------------*/

static void
initResult (
    SQLResultT * result,
    int additional)
{
    ROUTINE_DBG_MSP00 ("initResult");
    result->resultObj = NULL;
    result->isTuple = 0;
    result->isOK = true;
    result->additionalValues = additional;
}

/*----------------------------------------*/

static bool
allocResult (
    SQLResultT * result,
    int elements)
{
    ROUTINE_DBG_MSP00 ("allocResult");
    bool allocOK = true;

    /* ASSERT_DBG_MSP00 (result->resultObj == NULL); */
    result->resultObj = PyTuple_New (elements + result->additionalValues);
    result->isTuple = 1;
    allocOK = (result->resultObj != 0);
    return allocOK;
}

/*----------------------------------------*/

static bool
setResult (
    SQLResultT * result,
    PyObject * newElement)
{
    ROUTINE_DBG_MSP00 ("setResult");

    /* ASSERT_DBG_MSP00 (result->resultObj == NULL); */
    result->isTuple = false;
    if (newElement == NULL) {
        Py_INCREF (Py_None);
        newElement = Py_None;
    }
    result->resultObj = newElement;
    return true;
}

/*----------------------------------------*/

static bool
setResultElement (
    SQLResultT * result,
    int index,
    PyObject * newElement)
{
    ROUTINE_DBG_MSP00 ("setElement");
    bool assignOK = true;

    if (result->isTuple) {
        if (index < 0) {
            index = PyObject_Length (result->resultObj) - index;
        }
        assignOK = !PyTuple_SetItem (result->resultObj, index, newElement);
    }
    else {
        if (index == 0) {
            result->resultObj = newElement;
        }
        else {
            Py_XDECREF(newElement);
            PyErr_SetString(PyExc_IndexError,
                "tuple assignment index out of range");
            assignOK = false;
        }
    }
    return assignOK;
}

/*----------------------------------------*/

static int
sqlResultOK (
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("sqlResultOK");
    return(result->isOK);
}

/*----------------------------------------*/

static void
raiseCommunicationError (
    int code,
    const char * msg)
{
    ROUTINE_DBG_MSP00 ("raiseCommunicationError");
    PyObject * exception = PyInstance_New (CommunicationErrorType, NULL, NULL);
    PyObject * pycode = NULL;
    PyObject * pymsg = NULL;

    pycode = PyInt_FromLong  (code);
    pymsg = PyString_FromString  (msg);
    if (exception != NULL) {
        PyObject_SetAttrString (exception, LIT("errorCode"), pycode) ;
        PyObject_SetAttrString (exception, LIT("message"), pymsg);
        Py_XDECREF (pycode);
        Py_XDECREF (pymsg);
    }
    else {
        exception = Py_BuildValue (LIT("NN"), pycode, pymsg);
    }
    PyErr_SetObject (CommunicationErrorType, exception);
}

/*----------------------------------------*/

static void
raiseSQLError (
    int code,
    int pos,
    tsp00_C5  sqlStateStr,
    const char * msg)
{
    ROUTINE_DBG_MSP00 ("raiseSQLError");
    PyObject * exception = PyInstance_New (SQLErrorType, NULL, NULL);
    PyObject * errorCode = NULL;
    PyObject * message = NULL;
    PyObject * errorPos = NULL;
    PyObject * sqlState = NULL;

    errorCode = PyInt_FromLong  (code);
    message = PyString_FromString  (msg);
    errorPos = PyInt_FromLong  (pos);
    sqlState = PyString_FromStringAndSize  ((char *)sqlStateStr, sizeof (sqlState));
    if (exception != NULL) {
        PyObject_SetAttrString (exception, LIT("errorCode"), errorCode) ;
        PyObject_SetAttrString (exception, LIT("message"), message);
        PyObject_SetAttrString (exception, LIT("errorPos"), errorPos);
        PyObject_SetAttrString (exception, LIT("sqlState"), sqlState);
    }
    else {
        exception = Py_BuildValue (LIT("NNNN"), errorCode, message,
            errorPos, sqlState);
    }
    PyErr_SetObject (SQLErrorType, exception);
}

/*----------------------------------------*/

static boolean
sqlErrOccured (tin01_sql_session * nself, char * croakBuf)
{
    ROUTINE_DBG_MSP00 ("sqlErrOccurred");
    boolean errOccurred = false;
    tin01_sqlresult * sqlresult;

    if (nself->lasterr.sqlresult.returnCode != 0) {
        sqlresult = &nself->lasterr.sqlresult;
        raiseSQLError (sqlresult->returnCode, sqlresult->errorPos,
            sqlresult->sqlstate, sqlresult->sqlmsg);
        errOccurred = true;
    }
    else if (nself->lasterr.sp_rc != 0) {
        raiseCommunicationError (nself->lasterr.sp_rc, (const char *) nself->lasterr.errtext);
        errOccurred = true;
    }
    return errOccurred;
}

/*----------------------------------------*/

static boolean
isSessionConnected (tin01_sql_session * session, char * croakBuf)
{
    ROUTINE_DBG_MSP00 ("isSessionConnected");

    if ((session != NULL) && session->is_connected) {
        return true;
    }
    raiseCommunicationError (1, "Invalid Session");
    return false;
}

/*----------------------------------------*/

static void
addSourceInfo (tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("addSourceInfo");
    PyThreadState * threadState = PyThreadState_Get ();
    PyObject * filenameObject = threadState->frame->f_code->co_filename;
    const char * filename;
    int lineno = threadState->frame->f_lineno;
    char * dataBuf;
    int dataLen;

    if (!PyString_Check (filenameObject)) {
        return;
    }
    i28newpart (session, sp1pk_abap_info);
    dataBuf = i28argsbuf (session);

    filename = PyString_AS_STRING(filenameObject);
    dataLen = PyString_GET_SIZE(filenameObject);
    dataBuf [0] = (unsigned char) dataLen;
    memcpy (dataBuf + 1, filename, dataLen);
    i28extendpart (session, dataLen + 1);
    dataBuf += dataLen + 1;

    dataLen = sprintf (dataBuf + 1, "%d", lineno);
    dataBuf [0] = (unsigned char) dataLen;
    i28extendpart (session, dataLen + 1);
}

/*----------------------------------------*/

static int
requestReceive (
    tin01_sql_session * session,
    bool withInfo,
    bool allowRetry)
{
    ROUTINE_DBG_MSP00 ("requestReceive");
    int rc;

#if 0
    /* this cannot run against 6.2, so don't do it */
    if (withInfo) {
        addSourceInfo (session);
    }
#endif
#if defined (SAPDB_GARBAGE)
    Py_BEGIN_ALLOW_THREADS
    rc = i28sql (session, NULL);
    Py_END_ALLOW_THREADS
#else
    rc = i28requestForPython (session, NULL);
    /* printf("<session %p>\n", session); */
    if (rc == cin01_db_ok) {
        Py_BEGIN_ALLOW_THREADS
        rc = i28receiveForPython (session, NULL);
        /* printf("</session %p>\n", session); */
        Py_END_ALLOW_THREADS
    }
#endif
    if (rc == cin01_db_not_accessible) {
        Py_BEGIN_ALLOW_THREADS
        rc = i28connect (session);
        Py_END_ALLOW_THREADS
        if (rc == cin01_db_ok) {
            tin01_sqlresult * sqlresult = &(session->lasterr.sqlresult);
            sqlresult->sqlmsg [0] = '\0';
            if (lastWasEOT (session) && allowRetry) {
                rc = cin01_db_parse_again;
            }
            else {
                rc = cin01_db_cmd_timeout;
                strcpy (sqlresult->sqlmsg,
                    "Session inactivity timeout (work rolled back)");
            }
            sqlresult->returnCode = rc;
        }
    }
    clearEOTFlag (session);
    return rc;
}

/*----------------------------------------*/

typedef struct CommErrorT {
    int             code;
    tsp00_ErrTextc  msg;
} CommErrorT;

/*----------------------------------------*/

typedef struct SQLErrorT {
    int             code;
    int             pos;
    tsp00_C5c       sqlState;
    tsp00_ErrTextc msg;
} SQLErrorT;


#define croakBuf NULL

/* endsnippet Python_precode */
typedef struct SapDB_LongReaderObjectT {
    PyObject_HEAD
    LongReaderT nself;
/* no code for key SapDB_LongReader Python_/Generic_ in ['Generic_precode', 'Python_precode']*/
} SapDB_LongReaderObjectT;

staticforward PyTypeObject SapDB_LongReaderType;
#define isSapDB_LongReaderObject(op) ((op)->ob_type == &SapDB_LongReaderType)


static SapDB_LongReaderObjectT*
newSapDB_LongReader ()
{
    SapDB_LongReaderObjectT* result;

    result = PyObject_NEW (SapDB_LongReaderObjectT, &SapDB_LongReaderType);
    if (result != NULL) {
        memset (&result->nself, '\0', sizeof (SapDB_LongReaderObjectT) - offsetof (SapDB_LongReaderObjectT, nself));
    }
    return result;
}

static void
freeSapDB_LongReader (
    SapDB_LongReaderObjectT * value)
{
    Py_DECREF (value);
}
/* snippet SapDB_LongReader Generic_precode */

#define DESCFIELD(descriptor, name) (descriptor.variant.C_false.name##_F)
#if defined (__cplusplus)
#define BECOMES(lvalue, expr)       lvalue.becomes (expr)
#else
#define BECOMES(lvalue, expr)       lvalue = expr
#endif
/*----------------------------------------*/

static void
setReaderState (
    LongReaderT * nself,
    char * longBuf)
{
    ROUTINE_DBG_MSP00 ("setReaderState");
    tsp00_ValMode_Enum valMode = DESCFIELD (nself->descriptor, ld_valmode);
    int valLen = DESCFIELD (nself->descriptor, ld_vallen);
    const char * longData;

    if ((valMode == vm_datapart)
            || (valMode == vm_alldata)
            || (valMode == vm_lastdata)) {
        if (valLen > nself->dataSize) {
            if (nself->dataBuf == NULL) {
                nself->dataBuf = REINTERPRET_CAST (byte_t*, malloc (valLen));
            }
            else {
                nself->dataBuf = REINTERPRET_CAST (byte_t*, realloc (nself->dataBuf, valLen));
            }
            memset (nself->dataBuf, '%', valLen);
            nself->dataSize = valLen;
        }
        longData = longBuf + DESCFIELD (nself->descriptor, ld_valpos) - 1;
        memcpy (nself->dataBuf, longData, valLen);
        nself->dataPos = 0;
        nself->dataLen = valLen;
    }
    if ((valMode == vm_alldata) || (valMode == vm_lastdata)) {
        nself->lastGetvalFound = true;
    }
}

/*----------------------------------------*/

static void
initLongReader (
    LongReaderT * nself,
    tin01_sql_session * session,
    tsp00_LongDescriptor * descriptor)
{
    ROUTINE_DBG_MSP00 ("initLongReader");

    nself->session = session;
    increaseRef (session->more_data);
    memcpy (&nself->descriptor, descriptor, sizeof (*descriptor));
    setReaderState (nself, i28argsbuf (session));
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static PyObject *
createLongReader (
    tin01_sql_session * session,
    tsp00_LongDescriptor * descriptor)
{
    ROUTINE_DBG_MSP00 ("createLongReader");
    SapDB_LongReaderObjectT * reader = newSapDB_LongReader ();

    if (reader == NULL) {
        return NULL;
    }
    initLongReader (&reader->nself, session, descriptor);
    return REINTERPRET_CAST (PyObject *, reader);
}

/*----------------------------------------*/

#elif defined (PERL_GLUE)
static SV *
createLongReader (
    tin01_sql_session * session,
    tsp00_LongDescriptor * descriptor)
{
    ROUTINE_DBG_MSP00 ("createLongReader");
    char croakBuf [200];
    int resultCount;
    SV * perlObject = NULL;
    SapDB_LongReaderObjectT * reader = NULL;
    dSP;

    /*
     * create resultset object
     */
    /* ENTER; */
    /* SAVETMPS; */
    PUSHMARK (sp);
    PUTBACK;
    resultCount = perl_call_pv (LIT("sapdb::createReaderZombie"),
                      G_SCALAR | G_EVAL);
    SPAGAIN;
    if (SvTRUE (GvSV (errgv))) {
        POPs;
    }
    else {
        perlObject = POPs;
        SvREFCNT_inc (perlObject);  /* increase refcount to prevent delete */
    }
    PUTBACK;
    /* FREETMPS; */
    /* LEAVE; */
    /*
     * retrieve actual pointer from this mess
     */
    if (resultCount != 1) {
        return NULL;
    }
    reader = REINTERPRET_CAST (SapDB_LongReaderObjectT*,
         getGluePointer (perlObject, SapDB_LongReaderClassID, croakBuf));
    if ((reader == NULL) || !isSapDB_LongReaderObject (reader)) {
        sv_2mortal (perlObject);
        return NULL;
    }
    /*
     * fill reader object with soul
     */
    initLongReader (&reader->nself, session, descriptor);
    return perlObject;
}
#endif
/*----------------------------------------*/

static void
closeReader (
    LongReaderT * nself)
{
    ROUTINE_DBG_MSP00 ("closeReader");
    if (nself->dataBuf != NULL) {
        free (nself->dataBuf);
    }
    decreaseRef (nself->session->more_data);
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static void
destructor_SapDB_LongReader (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_LongReader");
    SapDB_LongReaderObjectT* self = REINTERPRET_CAST (SapDB_LongReaderObjectT*, pyself);
    LongReaderT * nself = &self->nself;
    closeReader (nself);
    free (pyself);
}
#endif

/*----------------------------------------*/

static int
doGetval (
    LongReaderT * nself)
{
    ROUTINE_DBG_MSP00 ("doGetval");
    tin01_sql_session * session = nself->session;
    int rc;
    char * resultBuf;

    i28initspecial (session, sp1m_getval);
    i28newpart (session, sp1pk_longdata);
    i28parg (session, REINTERPRET_CAST (const char *, &nself->descriptor),
        sizeof (tsp00_LongDescriptor), sizeof (tsp00_LongDescriptor),
        1, '\0');
    rc = requestReceive (session, false, false);
    if (rc != cin01_db_ok) {
        nself->lastGetvalFound = true;
        nself->dataLen = 0;
        nself->dataPos = 0;
        return rc;
    }
    i28findpart (session, sp1pk_longdata);
    resultBuf = i28argsbuf (session);
    memcpy (&nself->descriptor, resultBuf + 1, sizeof (tsp00_LongDescriptor));
    setReaderState (nself, resultBuf);
    return 0;
}

/*----------------------------------------*/

static SL_ValueT *
readLong (
    LongReaderT * nself,
    int requested)
{
    ROUTINE_DBG_MSP00 ("readLong");
    SL_ValueT * result = NULL;
    char * resultBuf;
    long   resultLen = 0;
    int remaining = nself->dataLen - nself->dataPos;

    resultBuf = REINTERPRET_CAST (char*, malloc (requested));
    while (requested > 0) {
        if (remaining >= requested) {
            /* copy requested */
            memcpy (resultBuf + resultLen, nself->dataBuf + nself->dataPos, requested);
            nself->dataPos += requested;
            resultLen += requested;
            requested = 0;
        }
        else {
            /* copy remaining */
            if (remaining > 0) {
                memcpy (resultBuf + resultLen, nself->dataBuf + nself->dataPos, remaining);
                requested -= remaining;
                resultLen += remaining;
                nself->dataPos += remaining;
            }
            if (nself->lastGetvalFound) {
                break;
            }
            doGetval (nself);
            remaining = nself->dataLen - nself->dataPos;
        }
    }
    result = SL_fromString (resultBuf, resultLen);
    free (resultBuf);
    return result;
}

/* endsnippet SapDB_LongReader Generic_precode */
/* snippet SapDB_LongReader Python_precode */

static PyObject *
read_SapDB_LongReader (PyObject *pyself, PyObject* args, PyObject* keywArgs);

static PyObject *
callableRead (PyObject *pyself, PyObject* args, PyObject* keywArgs)
{
    ROUTINE_DBG_MSP00 ("callableRead");
    return read_SapDB_LongReader (pyself, args, keywArgs);
}

/* endsnippet SapDB_LongReader Python_precode */
/*----------------------------------------*/

static PyObject *
read_SapDB_LongReader (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("read_SapDB_LongReader");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_LongReaderObjectT* self = REINTERPRET_CAST (SapDB_LongReaderObjectT*, pyself);
    LongReaderT* nself = &self->nself;
    static const char * kwlist [] = {
        LIT("count"), NULL};
    int count;
    SL_ValueT * result;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("i"), (char **) kwlist, &count)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = readLong (nself, count);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return result;
}

static PyMethodDef SapDB_LongReaderClassMethods [] = {
    {LIT("read"), (PyCFunction)read_SapDB_LongReader, METH_VARARGS|METH_KEYWORDS, 
        "returns the subsequent <i>count</i> characters.\n\n                    Returns an empty String when the end of the LONG\n                    is reached.\n                    "},
    {NULL, NULL}
};


static PyObject*
SapDB_LongReaderGetattr (PyObject* self, char* name)
{
    return Py_FindMethod (SapDB_LongReaderClassMethods, self, name);
}

statichere PyTypeObject SapDB_LongReaderType = {
    PyObject_HEAD_INIT (NULL)
    0,
    LIT("SapDB_LongReader"),    /* tp_name */
    sizeof (SapDB_LongReaderObjectT),    /* tp_basicsize */
    0,    /* tp_itemsize */
    REINTERPRET_CAST (destructor, destructor_SapDB_LongReader),    /* tp_dealloc */
    0,    /* tp_print */
    REINTERPRET_CAST (getattrfunc, SapDB_LongReaderGetattr),    /* tp_getattr */
    0,    /* tp_setattr */
    0,    /* tp_compare */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_repr */
    0,    /* tp_as_number */
    0,    /* tp_as_sequence */
    0,    /* tp_as_mapping */
    REINTERPRET_CAST (hashfunc, 0),    /* tp_hash */
    REINTERPRET_CAST (ternaryfunc, callableRead),    /* tp_call */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_str */
    0,    /* tp_getattro */
    0,    /* tp_setattro */
    0,    /* *tp_as_buffer */
    0,    /* tp_xxx4 */
    "allows retrieval of LONGs.\n\n        Python: This is a 'callable' object. Calling it with an\n        integer argument will return the next bytes of the LONG\n        (similar to calling the <i>read</i> method).",    /* tp_doc */
};

typedef struct SapDB_ResultSetObjectT {
    PyObject_HEAD
    ResultSetT nself;
/* no code for key SapDB_ResultSet Python_/Generic_ in ['Generic_precode']*/
} SapDB_ResultSetObjectT;

staticforward PyTypeObject SapDB_ResultSetType;
#define isSapDB_ResultSetObject(op) ((op)->ob_type == &SapDB_ResultSetType)


static SapDB_ResultSetObjectT*
newSapDB_ResultSet ()
{
    SapDB_ResultSetObjectT* result;

    result = PyObject_NEW (SapDB_ResultSetObjectT, &SapDB_ResultSetType);
    if (result != NULL) {
        memset (&result->nself, '\0', sizeof (SapDB_ResultSetObjectT) - offsetof (SapDB_ResultSetObjectT, nself));
    }
    return result;
}

static void
freeSapDB_ResultSet (
    SapDB_ResultSetObjectT * value)
{
    Py_DECREF (value);
}
/* snippet SapDB_ResultSet Generic_precode */
static int
openCursor (
    tin01_sql_session * session,
    ResultSetT * resultSet);

/*----------------------------------------*/
#if defined (PYTHON_GLUE)
static PyObject *
createResultSetFromReply (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("createResultSetFromReply");
    SapDB_ResultSetObjectT * newObj = newSapDB_ResultSet ();
    int rc;

    if (newObj == NULL) {
        return NULL;
    }
    rc = openCursor (session, &newObj->nself);
    if (rc != 0) {
        sqlErrOccured (session, NULL);
        Py_DECREF (newObj);
        newObj = NULL;
    }
    return REINTERPRET_CAST(PyObject *, newObj);
}
#elif defined (PERL_GLUE)
static SV *
createResultSetFromReply (
    tin01_sql_session * session)
{
    ROUTINE_DBG_MSP00 ("createResultSetFromReply");
    char croakBuf [200];
    int resultCount;
    SV * perlObject = NULL;
    SapDB_ResultSetObjectT * resultSet = NULL;
    int rc;
    dSP;

    /*
     * create resultset object
     */
    /* ENTER; */
    /* SAVETMPS; */
    PUSHMARK (sp);
    PUTBACK;
    resultCount = perl_call_pv (LIT("sapdb::createCursorZombie"),
                      G_SCALAR | G_EVAL);
    SPAGAIN;
    if (SvTRUE (GvSV (errgv))) {
        POPs;
    }
    else {
        perlObject = POPs;
    }
    PUTBACK;
    /* FREETMPS; */
    /* LEAVE; */
    /*
     * retrieve actual pointer from this mess
     */
    if (resultCount != 1) {
        return NULL;
    }
    resultSet = REINTERPRET_CAST (SapDB_ResultSetObjectT*,
         getGluePointer (perlObject, SapDB_ResultSetClassID, croakBuf));
    if ((resultSet == NULL) || !isSapDB_ResultSetObject (resultSet)) {
        sv_2mortal (perlObject);
        return NULL;
    }
    /*
     * fill resultset object with soul
     */
    rc = openCursor (session, &resultSet->nself);
    if (rc != 0) {
        sv_2mortal (perlObject);
        return NULL;
    }
    return perlObject;
}
#endif
/*----------------------------------------*/

static void
getOutputParameter (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("getOutputParameter");
    int         outputIndex = -1;
    int         i;
    tsp1_param_info * paramInfo = parseInfo->paramInfo;
    SL_ValueT * value;
    void      * addr;
    int         intVal;
    double      floatVal;
    tsp00_NumError  numErr;
    int         valLen;

    allocResult (result, parseInfo->outputParamCount);
    for (i = 0; i < parseInfo->paramCount; ++i, ++paramInfo) {
        if (paramInfo->sp1i_io_type != sp1io_input) {
            ++outputIndex;
            addr = i28argaddr (session, paramInfo->sp1i_bufpos);
            if (addr == NULL) {
                value = SL_fromNone ();
            }
            else {
                switch (paramInfo->sp1i_data_type) {
                    case dfixed:
                        if (paramInfo->sp1i_frac == 0) {
                            s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (intVal), REFPARM (numErr));
                            value = SL_fromInt (intVal);
                        }
                        else {
                            s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (floatVal), REFPARM (numErr));
                            value = SL_fromDouble (floatVal);
                        }
                        break;
                    case dfloat:
                    case dvfloat:
                        s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (floatVal), REFPARM (numErr));
                        value = SL_fromDouble (floatVal);
                        break;
                    case dsmallint:
                    case dinteger:
                        s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (intVal), REFPARM (numErr));
                        value = SL_fromInt (intVal);
                        break;
                    case dcha:
                    case dche:
                    case dvarchara:
                    case dvarchare:
                    case ddate:
                    case dtime:
                    case dtimestamp:
                        valLen = s30klen (addr, ' ', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST(const char *, addr),
                            valLen);
                        break;
                    case dchb:
                    case dvarcharb:
                        valLen = s30klen (addr, '\0', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST (const char *, addr),
                            valLen);
                        break;
                    case dstra:
                    case dstre:
                    case dstrb:
                    case dlonga:
                    case dlonge:
                    case dlongb:
                        value = createLongReader (session,
                            REINTERPRET_CAST (tsp00_LongDescriptor*, addr));
                        break;
                    default:
                        value = SL_fromNone ();
                        break;
                }
            }
            setResultElement (result, outputIndex, value);
        }
    }
}

/*----------------------------------------*/

static void *
argAddr (
    byte_t * dataBuf,
    int bufPos)
{
    ROUTINE_DBG_MSP00 ("argAddr");
    byte_t    *bufaddr;
    byte_t        *result;

    bufaddr = dataBuf + bufPos;
    if (bufaddr [-1] == (unsigned char) csp_undef_byte)
        result = NULL;
    else
        result = bufaddr;
    return result;
}

/*----------------------------------------*/

static void
getOutputParameter2 (
    tin01_sql_session * session,
    byte_t * dataBuf,
    ParseInfoT * parseInfo,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("getOutputParameter");
    int         outputIndex = -1;
    int         i;
    tsp1_param_info * paramInfo = parseInfo->paramInfo;
    SL_ValueT * value;
    void      * addr;
    int         intVal;
    double      floatVal;
    tsp00_NumError  numErr;
    int         valLen;

    allocResult (result, parseInfo->outputParamCount);
    for (i = 0; i < parseInfo->paramCount; ++i, ++paramInfo) {
        if (paramInfo->sp1i_io_type != sp1io_input) {
            ++outputIndex;
            addr = argAddr (dataBuf, paramInfo->sp1i_bufpos);
            if (addr == NULL) {
                value = SL_fromNone ();
            }
            else {
                switch (paramInfo->sp1i_data_type) {
                    case dfixed:
                        if (paramInfo->sp1i_frac == 0) {
                            s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (intVal), REFPARM (numErr));
                            if (numErr == num_ok) {
                                value = SL_fromInt (intVal);
                            }
                        }
                        else {
                            s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                                1, paramInfo->sp1i_length,
                                REFPARM (floatVal), REFPARM (numErr));
                            if (numErr == num_ok) {
                                value = SL_fromDouble (floatVal);
                            }
                        }
                        if (numErr != num_ok) {
                            char strbuf [100];
                            int outlen;
                            s42gchr (addr, 1, paramInfo->sp1i_length,
                                paramInfo->sp1i_frac, paramInfo->sp1i_length,
                                strbuf, 1, REFPARM (outlen), REFPARM (numErr));
                            value = SL_fromString (strbuf, outlen);
                        }
                        break;
                    case dfloat:
                    case dvfloat:
                        s40glrel (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (floatVal), REFPARM (numErr));
                        value = SL_fromDouble (floatVal);
                        break;
                    case dsmallint:
                    case dinteger:
                        s40glint (REINTERPRET_CAST (tsp00_Number*, addr),
                            1, paramInfo->sp1i_length,
                            REFPARM (intVal), REFPARM (numErr));
                        value = SL_fromInt (intVal);
                        break;
                    case dcha:
                    case dche:
                    case dvarchara:
                    case dvarchare:
                    case ddate:
                    case dtime:
                    case dtimestamp:
                        valLen = s30klen (addr, ' ', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST (const char *, addr),
                            valLen);
                        break;
                    case dchb:
                    case dvarcharb:
                        valLen = s30klen (addr, '\0', paramInfo->sp1i_length);
                        value = SL_fromString (REINTERPRET_CAST (const char *, addr),
                            valLen);
                        break;
                    case dboolean:
                        if (((byte_t*)addr)[0] == '\0') {
                            value = SL_fromString ("", 0);
                        }
                        else {
                            value = SL_fromInt (1);
                        }
                        break;
                    case dstra:
                    case dstre:
                    case dstrb:
                    case dlonga:
                    case dlonge:
                    case dlongb:
                        value = createLongReader (session, REINTERPRET_CAST (tsp00_LongDescriptor*, addr));
                        break;
                    default:
                        value = SL_fromNone ();
                        break;
                }
            }
            setResultElement (result, outputIndex, value);
        }
    }
}

/*----------------------------------------*/

static int
doDescribe (
    tin01_sql_session * session,
    const char * cursorName)
{
    ROUTINE_DBG_MSP00 ("doDescribe");
    int rc;

    if (cursorName [0] != '\0') {
        rc = i28adbsf (session, "DESCRIBE \"%s\"", cursorName);
    }
    else {
        rc = i28adbs_noinfo  (session, "DESCRIBE", false);
    }
    return rc;
}

/*----------------------------------------*/

static int
openCursor (
    tin01_sql_session * session,
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("openCursor");
    int nameLen = -1;
    int rc = 0;
    int infoSize;
    int i;

    resultSet->session = session;
    increaseRef (session->more_data);
    for (i = 0; i < fetchKindsE; ++i) {
        resultSet->fetchInfo [i].paramCount = -1;
    }
    memset (resultSet->cursorName, '\0', sizeof (resultSet->cursorName));
    i28gresulttablename (session, &nameLen, resultSet->cursorName);
    if (i28findpart (session, sp1pk_columnnames)) {
        nameLen = i28nameslen (session);
    }
    else {
        nameLen = -1;
    }
    if (nameLen == -1) {
        rc = doDescribe (session, resultSet->cursorName);
        if (rc == 0) {
            nameLen = i28nameslen (session);
        }
    }
    if (rc == 0) {
        resultSet->colNames = REINTERPRET_CAST (const char *, malloc (nameLen));
        memcpy ((void *) resultSet->colNames , i28colnames (session), nameLen);
        resultSet->colCount = i28paramcount (session);
        infoSize = sizeof (tsp1_param_info) * resultSet->colCount;
        resultSet->colInfo = REINTERPRET_CAST (tsp1_param_info *, malloc (infoSize));
        memcpy (resultSet->colInfo, i28gparaminfo (session, 0), infoSize);
        resultSet->currentPos = 0;
    }
    return rc;
}

/*----------------------------------------*/

static bool
row2result (
    tin01_sql_session * session,
    ResultSetT * resultSet,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("row2result");

    if (i28findpart (session, sp1pk_data)) {
        getOutputParameter (session, &resultSet->fetchInfo [nextFetchE], result);
    }
    else {
        setResult (result, NULL);
    }
    return true;
}

/*----------------------------------------*/

static int
getFetchBufPos (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("getFetchBufPos");
    return(resultSet->currentRow - resultSet->dataBuf)
    / resultSet->rowSize;
}

/*----------------------------------------*/

static int
getFetchBufRecords (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("getFetchBufRecords");
    return(resultSet->currentRow - resultSet->eoRecords)
    / resultSet->rowSize;
}

/*----------------------------------------*/

static boolean
moveFetchBufPos (
    ResultSetT * resultSet,
    int relative)
{
    ROUTINE_DBG_MSP00 ("moveFetchBufPos");
    boolean result;
    byte_t * newPos = resultSet->currentRow + (resultSet->rowSize * relative);
    if ((resultSet->dataBuf <= newPos) && (newPos < resultSet->eoRecords)) {
        result = true;
        resultSet->currentRow = newPos;
    }
    else {
        result = false;
    }
    return result;
}

/*----------------------------------------*/

static int
calculateFetchRecord (
    ParseInfoT * parseInfo)
{
    int result = 0;
    int i;

    for (i = 0; i < parseInfo->paramCount; ++i) {
        tsp1_param_info * paramInfo = &parseInfo->paramInfo [i];
        if (paramInfo->sp1i_io_type == sp1io_output) {
            result += paramInfo->sp1i_in_out_len;
        }
    }
    return result;
}

/*----------------------------------------*/

static int
parseFetch (
    ResultSetT * resultSet,
    fetchKindsT fetchKind)
{
    ROUTINE_DBG_MSP00 ("parseFetch");
    tin01_sql_session * session = resultSet->session;
    ParseInfoT * parseInfo;
    int rc;
    const char * fmt;
    int i;

    i28initparse (session);
    if (strlen (resultSet->cursorName) > 0) {
        fmt = "Fetch %s \"%s\"";
    }
    else {
        fmt = "Fetch %s ";
    }
    i28pcmdf (session, fmt,
        fetchCommandsC [fetchKind], STATIC_CAST (char *, resultSet->cursorName));
    i28pcmd (session, "into ?", -1);
    for (i = 1; i < resultSet->colCount; ++i) {
        i28pcmd (session, ", ?", -1);
    }
#if defined (__cplusplus)
    session->segment->sp1c_sqlmode ()
        = tsp1_sqlmode::fromInt (sp1sm_adabas);
#else
    session->segment->variant.C_3.sp1c_sqlmode_F = sp1sm_adabas;
#endif
    session->segment->variant.C_3.sp1c_mass_cmd_F = true;
    rc = requestReceive (session, false, false);
    if (rc != cin01_db_ok) {
        return rc;
    }
    parseInfo = &resultSet->fetchInfo [fetchKind];
    if (!fillParamInfos (session, parseInfo, true)) {
        return internalErrorC;
    }
    if (resultSet->rowSize == 0) {
        resultSet->rowSize = calculateFetchRecord (parseInfo);
    }
    return cin01_db_ok;
}

/*----------------------------------------*/

static int
copyFetchBuffer (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("copyFetchBuffer");
    int partLen;
    byte_t * newBuf;
    int recordsInFetch;

    if (!i28findpart (resultSet->session, sp1pk_data)) {
        return cin01_db_row_not_found;
    }
    if (resultSet->session->part->variant.C_2.sp1p_arg_count_F == 0) {
        /* workaround: current bug in fetch relative */
        return cin01_db_row_not_found;
    }
    /*
     * make sure buffer is large enough
     */
    partLen = i28partlen (resultSet->session);
    if (resultSet->dataSize < partLen) {
        newBuf = REINTERPRET_CAST (byte_t*, realloc (resultSet->dataBuf, partLen));
        if (newBuf == NULL) {
            return internalErrorC;
        }
        else {
            resultSet->dataBuf = newBuf;
            resultSet->dataSize = partLen;
        }
    }
    /*
     * copy data
     */
    resultSet->dataLen = partLen;
    memcpy (resultSet->dataBuf, i28argsbuf (resultSet->session), partLen);
    /*
     * set currentRow and eoRecords
     */
    recordsInFetch = resultSet->session->part->variant.C_2.sp1p_arg_count_F;
    if (recordsInFetch > resultSet->maxRecordsPerFetch) {
        resultSet->maxRecordsPerFetch = recordsInFetch;
    }
    resultSet->currentRow = resultSet->dataBuf;
    resultSet->eoRecords = resultSet->dataBuf
                           + (recordsInFetch * resultSet->rowSize);
    return cin01_db_ok;
}

/*----------------------------------------*/

static int
doFetch (
    ResultSetT * resultSet,
    fetchKindsT fetchKind,
    int parameter)
{
    ROUTINE_DBG_MSP00 ("doFetch");
    tin01_sql_session * session = resultSet->session;
    ParseInfoT     * parseInfo;
    int          rc;
    tsp00_Number     num;
    tsp00_NumError   res;
    tsp1_param_info * paramInfo;
    int          fetchCount = 30000;

    parseInfo = &resultSet->fetchInfo [fetchKind];
    if (parseInfo->paramCount == -1) {
        rc = parseFetch (resultSet, fetchKind);
        if (rc != cin01_db_ok) {
            return rc;
        }
    }
    i28initexecute (session, parseInfo->pid);
    switch (fetchKind) {
        case relativeFetchE:
            if (parameter < 0) {
                if (resultSet->maxRecordsPerFetch > 0) {
                    fetchCount = resultSet->maxRecordsPerFetch;
                }
                else {
                    fetchCount = 1;
                }
            }
            /*FALLTHROUGH*/
        case absoluteFetchE:
            /*
             * add parameter
             */
            i28newpart (session, sp1pk_data);
            s41plint (&num, 1, csp_fixed, 0, parameter, REFPARM (res));
            paramInfo = &parseInfo->paramInfo [0];
            i28parg (session, (const tin01_byte *) num, sizeof (num),
                paramInfo->sp1i_in_out_len, paramInfo->sp1i_bufpos,
                csp_defined_byte);
            break;
    }
    /*
     * add result count
     */
    i28newpart (session, sp1pk_resultcount);
    s41plint (&num, 1, csp_fixed, 0, fetchCount, REFPARM (res));
    i28parg (session, (const tin01_byte *) num, sizeof (num),
        sizeof (num) + 1, 1, csp_defined_byte);
    /*
     * execute
     */
    rc = requestReceive (session, false, false);
    if (rc == cin01_db_ok) {
        rc = copyFetchBuffer (resultSet);
    }
    return rc;
}

/*----------------------------------------*/

static void
getFetchOutput (
    SQLResultT * result,
    ResultSetT * resultSet,
    int          fetchKind,
    int          rc)
{
    ROUTINE_DBG_MSP00 ("getFetchOutput");
    switch (rc) {
        case cin01_db_ok:
            getOutputParameter2 (resultSet->session,
                resultSet->currentRow,
                &resultSet->fetchInfo [fetchKind],
                result);
            break;
        case cin01_db_row_not_found:
            setResult (result, SL_fromNone ());
            break;
        default:
            result->isOK = false;
            break;
    }
}

/*----------------------------------------*/

static SQLResultT
relativeRow (
    ResultSetT * resultSet,
    int count)
{
    ROUTINE_DBG_MSP00 ("relativeRow");
    int rc = cin01_db_ok;
    SQLResultT result;

    initResult (&result, 0);
    if (resultSet->dataBuf == NULL) {
        /*
         * no result yet requires FETCH
         */
        rc = doFetch (resultSet, relativeFetchE, count);
        if (rc == cin01_db_ok) {
            resultSet->currentRow= resultSet->dataBuf;
            resultSet->currentPos += count;
        }
    }
    else if (moveFetchBufPos (resultSet, count)) {
        resultSet->currentPos += count;
    }
    else {
        /*
         * moving outside of buffer requires FETCH
         */
        rc = doFetch (resultSet, relativeFetchE, count);
        if (rc != cin01_db_ok) {
            if (count > 0) {
                resultSet->currentRow = resultSet->dataBuf;
            }
            else {
                resultSet->currentRow = resultSet->eoRecords
                                        - resultSet->rowSize;
            }
        }
        else {
            resultSet->currentPos += count;
        }
    }
    getFetchOutput (&result, resultSet, relativeFetchE, rc);
    return result;
}

/*----------------------------------------*/

static SQLResultT
absoluteRow (
    ResultSetT * resultSet,
    int count)
{
    ROUTINE_DBG_MSP00 ("absoluteRow");
    int rc = cin01_db_ok;
    SQLResultT result;

    if (count == (resultSet->currentPos + 1)) {
        result = relativeRow(resultSet, 1);
    }
    else {
        initResult (&result, 0);
        rc = doFetch (resultSet, absoluteFetchE, count);
        resultSet->currentRow = resultSet->dataBuf;
        getFetchOutput (&result, resultSet, absoluteFetchE, rc);
        resultSet->currentPos = count;
    }
    return result;
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static PyObject *
absoluteRowBased0 (
    PyObject * pyself,
    int count)
{
    ROUTINE_DBG_MSP00 ("absoluteRowBased0");
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    SQLResultT result;
    PyObject * resultObj;

    result = absoluteRow (nself, count + 1);
    if (!result.isOK) {
        return NULL;
    }
    if (!sqlResultC2Py (self, result, &resultObj)) {
        return NULL;
    }
    if (SL_isNone (resultObj)) {
        PyErr_SetNone (PyExc_IndexError);
        return NULL;
    }
    return resultObj;
}
#endif

/*----------------------------------------*/

static SQLResultT
firstRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("first");
    int rc = cin01_db_ok;
    SQLResultT result;

    initResult (&result, 0);
    rc = doFetch (resultSet, firstFetchE, -1);
    resultSet->currentRow = resultSet->dataBuf;
    getFetchOutput (&result, resultSet, firstFetchE, rc);
    resultSet->currentPos = 1;
    return result;
}

/*----------------------------------------*/

static SQLResultT
lastRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("lastRow");
    int rc;
    SQLResultT result;

    initResult (&result, 0);
    rc = doFetch (resultSet, firstFetchE, -1);
    resultSet->currentRow = resultSet->eoRecords
                            - resultSet->rowSize;
    getFetchOutput (&result, resultSet, lastFetchE, rc);
    resultSet->currentPos = -1;
    return result;
}

/*----------------------------------------*/

static SQLResultT
currentRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("currentRow");
    SQLResultT result;
    int i;

    initResult (&result, 0);
    for (i = 0; i < fetchKindsE; ++i) {
        if (resultSet->fetchInfo [i].paramCount != -1) {
            getOutputParameter2 (resultSet->session,
                resultSet->currentRow,
                &resultSet->fetchInfo [i],
                &result);
            return result;
        }
    }
    result.isOK = false;
    return result;
}

/*----------------------------------------*/

static SQLResultT
nextRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("nextRow");
    return relativeRow (resultSet, 1);
}

/*----------------------------------------*/

static SQLResultT
previousRow (
    ResultSetT * resultSet)
{
    ROUTINE_DBG_MSP00 ("previousRow");
    return relativeRow (resultSet, -1);
}

/*----------------------------------------*/

static int
columnCount (
    ResultSetT * resultSet)
{
    return resultSet->colCount;
}

/*----------------------------------------*/

static const char * sqlTypeNamesC [] = {
    "Fixed",
    "Float",
    "Char",
    "Char EBCDIC",
    "Char Binary",
    "Rowid",
    "Long",
    "Long EBCDIC",
    "Long byte_t",
    "Long Doublebyte",
    "Date",
    "Time",
    "Float",
    "Timestamp",
    "Unknown",
    "Number",
    "Nonumber",
    "Duration",
    "Doublebyte EBCDIC",
    "Long",
    "Long EBCDIC",
    "Long Binary",
    "Long Doublebyte",
    "Boolean",
    "Unicode",
    "filler1",
    "filler2",
    "filler3",
    "filler4",
    "Smallint",
    "Integer",
    "Varchar",
    "Varchar EBCDIC",
    "Varchar Binary",
    "Long Unicode",
    "Long Unicode",
    "Varchar Unicode",
    "UDT",
    "ABAP Table",
};

static SQLResultT
columnInfo (
    ResultSetT * resultSet,
    int colIndex)
{
    ROUTINE_DBG_MSP00 ("columnInfo");
    SQLResultT result;
    const char * namePos;
    tsp1_param_info * paramInfo;
    int i;

    initResult (&result, 0);
    if ((colIndex < 0) || (resultSet->colCount <= colIndex)) {
        result.isOK = false;
        return result;
    }
    allocResult (&result, 4);
    /*
     * get name
     */
    namePos = resultSet->colNames;
    for (i = 0; i < resultSet->colCount; ++i) {
        if (i == colIndex) {
            setResultElement (&result, 0, SL_fromString (namePos + 1, namePos [0]));
            break;
        }
        namePos += namePos [0] + 1;
    }
    /*
     * get type, len and frac
     */
    paramInfo = &resultSet->colInfo [i];
    setResultElement (&result, 1, SL_fromString (sqlTypeNamesC [paramInfo->sp1i_data_type], -1));
    setResultElement (&result, 2, SL_fromInt (paramInfo->sp1i_length));
    setResultElement (&result, 3, SL_fromInt (paramInfo->sp1i_frac));
    return result;
}
/*----------------------------------------*/

static char *
cursorName (
    ResultSetT * nself)
{
    ROUTINE_DBG_MSP00 ("cursorName");
    return nself->cursorName;
}

/*----------------------------------------*/

static void
closeResultSet (
    ResultSetT * nself)
{
    ROUTINE_DBG_MSP00 ("closeResultSet");
    int i;

    for (i = 0; i < fetchKindsE; ++i) {
        clearParseInfo (nself->session, &nself->fetchInfo [i]);
    }
    if (strlen (nself->cursorName) > 0) {
        if (nself->session->is_connected) {
            i28adbsf (nself->session, "CLOSE \"%s\"", STATIC_CAST (char*, nself->cursorName));
        }
    }
    decreaseRef (nself->session->more_data);
    if (nself->colNames != NULL) {
        const void * tmpVoid = nself->colNames;
        free (CONST_CAST (void *, tmpVoid));
        nself->colNames = NULL;
    }
    if (nself->colInfo != NULL) {
        free (nself->colInfo);
        nself->colInfo = NULL;
    }
    nself->colCount = 0;
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static void
destructor_SapDB_ResultSet (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_ResultSet");
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);

    closeResultSet (&self->nself);
    free (pyself);
}
#endif

/*----------------------------------------*/
/* endsnippet SapDB_ResultSet Generic_precode */
/*----------------------------------------*/

static PyObject *
next_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("next_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        NULL};
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = nextRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

/*----------------------------------------*/

static PyObject *
previous_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("previous_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        NULL};
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = previousRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

/*----------------------------------------*/

static PyObject *
relative_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("relative_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        LIT("count"), NULL};
    int count;
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("i"), (char **) kwlist, &count)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = relativeRow (nself, count);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

/*----------------------------------------*/

static PyObject *
absolute_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("absolute_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        LIT("i"), NULL};
    int i;
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("i"), (char **) kwlist, &i)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = absoluteRow (nself, i);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

/*----------------------------------------*/

static PyObject *
first_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("first_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        NULL};
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = firstRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

/*----------------------------------------*/

static PyObject *
last_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("last_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        NULL};
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = lastRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

/*----------------------------------------*/

static PyObject *
current_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("current_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        NULL};
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = currentRow (nself);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

/*----------------------------------------*/

static PyObject *
cursorName_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("cursorName_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        NULL};
    char * result;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = cursorName (nself);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return Py_BuildValue (LIT("s"), result);
}

/*----------------------------------------*/

static PyObject *
columnCount_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("columnCount_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        NULL};
    int result;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = columnCount (nself);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return Py_BuildValue (LIT("i"), result);
}

/*----------------------------------------*/

static PyObject *
columnInfo_SapDB_ResultSet (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("columnInfo_SapDB_ResultSet");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_ResultSetObjectT* self = REINTERPRET_CAST (SapDB_ResultSetObjectT*, pyself);
    ResultSetT* nself = &self->nself;
    static const char * kwlist [] = {
        LIT("i"), NULL};
    int i;
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("i"), (char **) kwlist, &i)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = columnInfo (nself, i);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

static PyMethodDef SapDB_ResultSetClassMethods [] = {
    {LIT("next"), (PyCFunction)next_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "gets the next row.\n\n                    Output values are returned as a list.<BR>\n                    The end of the result set is signaled by an\n                    undefined value (<b>Python</b>: None).\n                    "},
    {LIT("previous"), (PyCFunction)previous_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "gets the previous row.\n\n                    Output values are returned as a list.<BR>\n                    The end of the result set is signaled by an\n                    undefined value (<b>Python</b>: None).\n                    "},
    {LIT("relative"), (PyCFunction)relative_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "moves relative to the current row.\n\n                    A negative value of <i>count</i> will move backwards\n                    in the ResultSet.<BR>\n                    Output values are returned as a list.<BR>\n                    The end of the result set is signaled by an\n                    undefined value (<b>Python</b>: None).\n                    "},
    {LIT("absolute"), (PyCFunction)absolute_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "gets the <i>i</i>th row.\n\n                    Output values are returned as a list.<BR>\n                    The first row has the index 1.<BR>\n                    The end of the result set is signaled by an\n                    undefined value (<b>Python</b>: None).\n                    "},
    {LIT("first"), (PyCFunction)first_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "gets the first row.\n\n                    Output values are returned as a list.<BR>\n                    The end of the result set is signaled by an\n                    undefined value (<b>Python</b>: None).\n                    "},
    {LIT("last"), (PyCFunction)last_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "gets the last row.\n\n                    Output values are returned as a list.<BR>\n                    The end of the result set is signaled by an\n                    undefined value (<b>Python</b>: None).\n                    "},
    {LIT("current"), (PyCFunction)current_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "returns the current row.\n\n                    Output values are returned as a list.\n                    "},
    {LIT("cursorName"), (PyCFunction)cursorName_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "returns the name of the result set."},
    {LIT("columnCount"), (PyCFunction)columnCount_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "returns the number of columns."},
    {LIT("columnInfo"), (PyCFunction)columnInfo_SapDB_ResultSet, METH_VARARGS|METH_KEYWORDS, 
        "returns information about a column.\n\n                    A list of four values is returned:\n                    <OL start=\"0\">\n                    <LI>the column name\n                    <LI>the column type\n                    <LI>then length of the column\n                    <LI>the number of fractional digits\n                    </OL>\n                    The first column has the index 0;\n                    "},
    {NULL, NULL}
};


static PySequenceMethods SapDB_ResultSetSequenceVMT = {
    0,    /* sq_length */
    0,    /* sq_concat */
    0,    /* sq_repeat */
    REINTERPRET_CAST (intargfunc, absoluteRowBased0),    /* sq_item */
    0,    /* sq_slice */
    0,    /* sq_ass_item */
    0,    /* sq_ass_slice */
};

static PyObject*
SapDB_ResultSetGetattr (PyObject* self, char* name)
{
    return Py_FindMethod (SapDB_ResultSetClassMethods, self, name);
}

statichere PyTypeObject SapDB_ResultSetType = {
    PyObject_HEAD_INIT (NULL)
    0,
    LIT("SapDB_ResultSet"),    /* tp_name */
    sizeof (SapDB_ResultSetObjectT),    /* tp_basicsize */
    0,    /* tp_itemsize */
    REINTERPRET_CAST (destructor, destructor_SapDB_ResultSet),    /* tp_dealloc */
    0,    /* tp_print */
    REINTERPRET_CAST (getattrfunc, SapDB_ResultSetGetattr),    /* tp_getattr */
    0,    /* tp_setattr */
    0,    /* tp_compare */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_repr */
    0,    /* tp_as_number */
    &SapDB_ResultSetSequenceVMT,    /* tp_as_sequence */
    0,    /* tp_as_mapping */
    REINTERPRET_CAST (hashfunc, 0),    /* tp_hash */
    REINTERPRET_CAST (ternaryfunc, 0),    /* tp_call */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_str */
    0,    /* tp_getattro */
    0,    /* tp_setattro */
    0,    /* *tp_as_buffer */
    0,    /* tp_xxx4 */
    "allows navigation and retrieval of a result set.\n\n        Objects are created by all sql commands which create a result set\n        (SapDB_Session::sql, SapDB_Prepared::execute).\n        Python: a SapDB_ResultSet can be used as an iterator:\n        <PRE>\n        <b>for</b> row <b>in</b> resultSet:\n            ... process row ...\n        </PRE>\n        Currently, this will iterate over the result set from the beginning,\n        even if some rows have been fetched.\n        ",    /* tp_doc */
};

typedef struct SapDB_PreparedObjectT {
    PyObject_HEAD
    PreparedT nself;
/* no code for key SapDB_Prepared Python_/Generic_ in ['Generic_precode', '::select Generic_nativeCall vars']*/
} SapDB_PreparedObjectT;

staticforward PyTypeObject SapDB_PreparedType;
#define isSapDB_PreparedObject(op) ((op)->ob_type == &SapDB_PreparedType)


static SapDB_PreparedObjectT*
newSapDB_Prepared ()
{
    SapDB_PreparedObjectT* result;

    result = PyObject_NEW (SapDB_PreparedObjectT, &SapDB_PreparedType);
    if (result != NULL) {
        memset (&result->nself, '\0', sizeof (SapDB_PreparedObjectT) - offsetof (SapDB_PreparedObjectT, nself));
    }
    return result;
}

static void
freeSapDB_Prepared (
    SapDB_PreparedObjectT * value)
{
    Py_DECREF (value);
}
/* snippet SapDB_Prepared Generic_precode */

typedef struct LongInputT {
    bool readStarted;
    bool hasData;
    unsigned short colIndex;
    tsp00_LongDescriptor descriptor;
    InputStreamT * stream;
    union {
        StringInputStreamT stringStream;
        CallableInputStreamT callableStream;
    } obj;
} LongInputT;

/*----------------------------------------*/

static void
putLongvals (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    LongInputT * longs,
    bool forPutval)
{
    int    descriptorLen;
    int    descriptorPos;
    byte_t * dataPtr;
#if defined (PUTVAL_REQUIRES_TWO_BYTE_COUNT)
    byte_t * startPtr;
#endif
    int    bufPos;
    int    freeLen;
    int    bytesRead;
    tsp00_Int2 transferCount = 0;
    LongInputT * currentLong;
    InputStreamT * instream;
    tsp1_param_info * paramInfo;
    int i;

    if (forPutval) {
        descriptorLen = sizeof (tsp00_LongDescriptor) + 1;
    }
    else {
        descriptorLen = 0;
    }
    i28getFreePartInfo (session, REINTERPRET_CAST (void**, &dataPtr), &bufPos, &freeLen);
#if defined (PUTVAL_REQUIRES_TWO_BYTE_COUNT)
    startPtr = dataPtr;
    if (forPutval) {
        dataPtr += 2;
        bufPos += 2;
    }
#endif
    for (i = 0; (freeLen > descriptorLen) && (i < parseInfo->inLongCount); ++i) {
        currentLong = &longs [i];
        instream = currentLong->stream;
        paramInfo = &parseInfo->paramInfo [currentLong->colIndex];
        if (currentLong->hasData) {
            bytesRead = (*instream)->read (instream,
                            dataPtr + descriptorLen, freeLen);
            if (bytesRead > 0) {
                if (bytesRead < freeLen) {
                    if (currentLong->readStarted) {
                        BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_lastdata);
                    }
                    else {
                        BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_alldata);
                    }
                    currentLong->hasData = false;
                }
                else {
                    BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_datapart);
                    currentLong->readStarted = true;
                }
                DESCFIELD (currentLong->descriptor, ld_valpos) = bufPos + descriptorLen;
                DESCFIELD (currentLong->descriptor, ld_vallen) = bytesRead;
                if (forPutval) {
                    descriptorPos = bufPos;
                }
                else {
                    descriptorPos = paramInfo->sp1i_bufpos;
                }
                i28parg (session, REINTERPRET_CAST (const char*, &currentLong->descriptor),
                    sizeof (currentLong->descriptor),
                    sizeof (currentLong->descriptor) + 1,
                    descriptorPos, '\0');
                ++transferCount;
                i28extendpart (session, bytesRead);
                dataPtr += bytesRead + descriptorLen;
                bufPos += bytesRead + descriptorLen;
                freeLen -= bytesRead + descriptorLen;
            }
            else {
                currentLong->hasData = false;
            }
        }
    }
    if (forPutval && !(longs [parseInfo->inLongCount - 1].hasData)) {
        /* at end of all longs */
        currentLong = &longs [parseInfo->inLongCount - 1];
        BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_last_putval);
        i28parg (session, REINTERPRET_CAST (const char*, &currentLong->descriptor),
            sizeof (currentLong->descriptor),
            sizeof (currentLong->descriptor) + 1,
            bufPos, '\0');
        /* ++transferCount; */
    }
#if defined (PUTVAL_REQUIRES_TWO_BYTE_COUNT)
    if (forPutval) {
        memcpy (startPtr, &transferCount, sizeof (transferCount));
    }
#endif
    session->part->variant.C_2.sp1p_arg_count_F += transferCount;
}

/*----------------------------------------*/

static int
doPutval (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    LongInputT * longs)
{
    ROUTINE_DBG_MSP00 ("doPutval");
    tsp00_LongDescriptor descriptor;
    int descriptorsReturned;
    byte_t * bufPos;
    int rc;
    int i;

    /*
     *  get descriptors
     */
    i28findpart (session, sp1pk_longdata);
    descriptorsReturned = session->part->variant.C_2.sp1p_arg_count_F;
#if defined (__cplusplus)
    bufPos = (byte_t *) session->part->sp1p_buf ().asCharp ();
#else
    bufPos = session->part->variant.C_1.sp1p_buf_F;
#endif
    for (i = 0; i < descriptorsReturned; ++i) {
        memcpy (&descriptor, bufPos + 1, sizeof (descriptor));
        memcpy (&longs [DESCFIELD (descriptor, ld_valind)].descriptor,
            &descriptor, sizeof (descriptor));
    }
    /*
     * send putval
     */
    i28initspecial (session, sp1m_putval);
    i28newpart (session, sp1pk_longdata);
    putLongvals (session, parseInfo, longs, true);
    rc = requestReceive (session, false, false);
    return rc;
}

/*----------------------------------------*/

static void
    setInputParameter (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    SequenceT    args,
    LongInputT * longs)
{
    ROUTINE_DBG_MSP00 ("setInputParameter");
    tsp1_param_info * paramInfo = parseInfo->paramInfo;
    SL_ValueT  * value;
    int argCount;
    int inputIndex = -1;
    char fillChar = ' ';
    const char * addr;
    int dataLen;
    int i;
    tsp00_Number number;
    char boolBuf [1];
    int longIndex = 0;
    LongInputT * currentLong;

    if (parseInfo->paramCount == 0) {
        return;
    }
    argCount = sequenceLength (&args);
    i28newpart (session, sp1pk_data);
    for (i = 0; i < parseInfo->paramCount; ++i, ++paramInfo) {
        if (paramInfo->sp1i_io_type != sp1io_output) {
            ++inputIndex;
            if (inputIndex < argCount) {
                value = getSequenceElement (&args, inputIndex);
            }
            else {
                value = SL_None;
            }
            if (value == NULL) {
                /* set exception and exit */
            }
            switch (paramInfo->sp1i_data_type) {
                case dfixed:
                case dfloat:
                case dvfloat:
                case dsmallint:
                case dinteger:
                    if (SL_isNone (value)) {
                        i28pnull (session,
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos);
                    }
                    else {
                        SL_getNumber (value, paramInfo, &number);
                        i28parg (session, (tin01_byte *) number,
                            sizeof (number),
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos, '\0');
                    }
                    break;
                case dchb:
                case dvarcharb:
                    fillChar = '\0';
                    /*FALLTHROUGH*/
                case dcha:
                case dche:
                case dvarchara:
                case dvarchare:
                case ddate:
                case dtime:
                case dtimestamp:
                    if (SL_isNone (value)) {
                        i28pnull (session,
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos);
                    }
                    else {
                        SL_getRawString (value, &addr, &dataLen);
                        i28parg (session, addr, dataLen,
                            paramInfo->sp1i_in_out_len,
                            paramInfo->sp1i_bufpos, fillChar);
                    }
                    break;
                case dboolean:
                    if (SL_isTrue (value)) {
                        boolBuf [0] = '\1';
                    }
                    else {
                        boolBuf [0] = '\0';
                    }
                    i28parg (session, boolBuf, sizeof (boolBuf),
                        paramInfo->sp1i_in_out_len,
                        paramInfo->sp1i_bufpos, '\0');
                    break;
                case dstra:
                case dstre:
                case dstrb:
                case dlonga:
                case dlonge:
                case dlongb:
                    /*
                     * put descriptor
                     */
                    currentLong = &longs [longIndex];
                    currentLong->colIndex = i;
                    DESCFIELD (currentLong->descriptor, ld_valind) = longIndex;
                    BECOMES (DESCFIELD (currentLong->descriptor, ld_valmode), vm_nodata);
                    i28parg (session, REINTERPRET_CAST (const char*, &currentLong->descriptor),
                        sizeof (currentLong->descriptor),
                        paramInfo->sp1i_in_out_len,
                        paramInfo->sp1i_bufpos, '\0');
                    /*
                     * convert value to stream
                     */
                    if (SL_isString (value)) {
                        const char * data;
                        int len;
                        SL_getRawString (value, &data, &len);
                        currentLong->stream = initStringInputStream (
                            &currentLong->obj.stringStream, data, len);
                        currentLong->hasData = true;
                    }
                    else if (SL_isCallable (value)) {
                        currentLong->stream = initCallableInputStream (
                            &currentLong->obj.callableStream, value);
                        currentLong->hasData = true;
                    }
                    else {
                        currentLong->hasData = false;
                    }
                    /*
                     *
                     */
                    ++longIndex;
                    break;
            }
        }
    }
    if (longs != NULL) {
        putLongvals (session, parseInfo, longs, false);
    }
}

/*----------------------------------------*/

static bool
fillParamInfos (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    bool initial)
{
    ROUTINE_DBG_MSP00 ("fillParamInfos");
    bool ok = true;
    int infoSize;
    int i;

    i28gparseid (session, parseInfo->pid);
    parseInfo->paramCount = i28paramcount (session);
    parseInfo->paramInfo = NULL;
    if (parseInfo->paramCount > 0) {
        infoSize = parseInfo->paramCount * sizeof (tsp1_param_info);
        if (initial) {
            parseInfo->paramInfo = REINTERPRET_CAST (tsp1_param_info*, malloc (infoSize));
        }
        if (parseInfo->paramInfo == NULL) {
            ok = false;
        }
        else {
            memcpy (parseInfo->paramInfo, i28gparaminfo (session, 0), infoSize);
        }
    }
    parseInfo->outputParamCount = 0;
    for (i = 0; i < parseInfo->paramCount; ++i) {
        tsp1_param_info * paramInfo = &parseInfo->paramInfo [i];
        if (paramInfo->sp1i_io_type != sp1io_input) {
            ++parseInfo->outputParamCount;
        }
        switch (paramInfo->sp1i_data_type) {
            case dstra:
            case dstre:
            case dstrb:
            case dlonga:
            case dlonge:
            case dlongb:
                switch (paramInfo->sp1i_io_type) {
                    case sp1io_input:
                        ++parseInfo->inLongCount;
                        break;
                    case sp1io_output:
                        ++parseInfo->outLongCount;
                        break;
                    case sp1io_inout:
                        ++parseInfo->inLongCount;
                        ++parseInfo->outLongCount;
                        break;
                }
                break;
        }
    }
    return ok;
}

/*----------------------------------------*/

static int
prepare (
    tin01_sql_session * session,
    PreparedT * prepared,
    const char * cmd,
    bool initial)
{
    ROUTINE_DBG_MSP00 ("prepare");
    int rc = 0;
    int cmdLen = strlen (cmd);
    int retryCount = 3;

    prepared->session = session;
    increaseRef (session->more_data);
    prepared->parseInfo.paramCount = -1;
    do {
        i28initparse (prepared->session);
        i28pcmd (prepared->session, cmd, cmdLen);
        if (getAutocommit (session)) {
            session->segment->variant.C_3.sp1c_commit_immediately_F = true;
        }
        rc = requestReceive (prepared->session, true, true);
        --retryCount;
    } while ((rc == cin01_db_parse_again) && (retryCount > 0));
    if (rc == cin01_db_ok) {
        fillParamInfos (session, &prepared->parseInfo, initial);
        i28gparseid (prepared->session, prepared->parseInfo.pid);
        if (initial) {
            prepared->parseInfo.sqlString = (char *) malloc (cmdLen + 1);
            if (prepared->parseInfo.sqlString != NULL) {
                memcpy (prepared->parseInfo.sqlString, cmd, cmdLen + 1);
            }
        }
    }
    return rc;
}

/*----------------------------------------*/

static int
prepareFetch (
    tin01_sql_session * session,
    PreparedT * prepared)
{
    ROUTINE_DBG_MSP00 ("prepareFetch");
    char fetchCmd [100];
    int nameLen;
    int pos;

    strcpy (fetchCmd, "FETCH  ");
    pos = strlen ( fetchCmd);
    nameLen = sizeof (fetchCmd) - pos;
    if (i28gresulttablename (session, &nameLen, fetchCmd + pos)) {
        fetchCmd [pos - 1] = '"';
        fetchCmd [pos + nameLen] = '"';
        fetchCmd [pos + nameLen + 1] = '\0';
    }
    return prepare (session, prepared, fetchCmd, true);
}

/*----------------------------------------*/

static bool
packet2result (
    tin01_sql_session * session,
    ParseInfoT * parseInfo,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("packet2result");

    if (i28findpart (session, sp1pk_data)) {
        ParseInfoT localInfo;
        bool requireDeallocation = false;
        if (parseInfo == NULL) {
            parseInfo = &localInfo;
            if (fillParamInfos (session, parseInfo, true)) {
                requireDeallocation = true;
            }
            else {
                parseInfo->paramInfo = NULL;
            }

        }
        i28findpart (session, sp1pk_data);
        if (parseInfo->paramInfo != NULL) {
            getOutputParameter (session, parseInfo, result);
            if (requireDeallocation) {
                free (localInfo.paramInfo);
            }
        }
        else {
            int partLen = i28partlen (session);
            char * partPtr = i28getpart (session);
            SL_ValueT * resultValue = SL_fromString (partPtr, partLen);
            setResult (result, resultValue);
        }
    }
    else if (i28findpart (session, sp1pk_resulttablename)) {
        setResult (result, createResultSetFromReply (session));
    }
    else {
        setResult (result, NULL);
    }
    return true;
}

/*----------------------------------------*/

static SQLResultT
executePrepared (
    PreparedT * prepared,
    SequenceT   args)
{
    ROUTINE_DBG_MSP00 ("executePrepared");
    SQLResultT   result;
    LongInputT * longs = NULL;
    int          rc = cin01_db_ok;
    ParseInfoT * parseInfo = &prepared->parseInfo;
    int          retryCount = 3;

    initResult (&result, 0);
    if (parseInfo->inLongCount > 0) {
        longs = REINTERPRET_CAST (LongInputT*, calloc (parseInfo->inLongCount,
                    sizeof (LongInputT)));
    }
    do {
        if (rc == cin01_db_parse_again) {
            --retryCount;
            rc = prepare (prepared->session, prepared,
                prepared->parseInfo.sqlString, false);
            if (rc != cin01_db_ok) {
                break;
            }
        }
        i28initexecute (prepared->session, parseInfo->pid);
        prepared->session->segment->variant.C_3.sp1c_with_info_F = true;
        prepared->session->segment->variant.C_3.sp1c_commit_immediately_F
            = getAutocommit (prepared->session);
        addCursorName (prepared->session);
        setInputParameter (prepared->session, parseInfo, args, longs);
        rc = requestReceive (prepared->session, true, true);
    } while ((rc == cin01_db_parse_again) && (retryCount > 0));
    if (rc == cin01_db_ok) {
        packet2result (prepared->session, parseInfo, &result);
        if (longs != NULL) {
            while (longs [parseInfo->inLongCount - 1].hasData
                   && (rc == cin01_db_ok)) {
                rc = doPutval (prepared->session, parseInfo, longs);
            }
        }
    }
    else {
        result.isOK = 0;
    }
    if (longs != NULL) {
        free (longs);
    }
    return result;
}

/*----------------------------------------*/

static SQLResultT
nextPrepared (
    PreparedT * prepared)
{
    ROUTINE_DBG_MSP00 ("nextPrepared");
    SequenceT  args;
    args = EmptySequenceC;
    return executePrepared (prepared, args);
}

/*----------------------------------------*/

static void
closePrepared (
    PreparedT * nself)
{
    ROUTINE_DBG_MSP00 ("closePrepared");
    clearParseInfo (nself->session, &nself->parseInfo);
    decreaseRef (nself->session->more_data);
}

/*----------------------------------------*/

#if defined (PYTHON_GLUE)
static void
destructor_SapDB_Prepared (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_Prepared");
    SapDB_PreparedObjectT* self = REINTERPRET_CAST (SapDB_PreparedObjectT*, pyself);

    closePrepared (&self->nself);
    free (pyself);
}

/*----------------------------------------*/
#endif

/* endsnippet SapDB_Prepared Generic_precode */
/*----------------------------------------*/

static PyObject *
execute_SapDB_Prepared (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("execute_SapDB_Prepared");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_PreparedObjectT* self = REINTERPRET_CAST (SapDB_PreparedObjectT*, pyself);
    PreparedT* nself = &self->nself;
    static const char * kwlist [] = {
        LIT("sqlParms"), NULL};
    SequenceT sqlParms =  EmptySequenceC;
    PyObject * sqlParmsObj = NULL;
    SQLResultT result;
    PyObject * resultObj;

    if (!isSessionConnected (self->nself.session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("|O"), (char **) kwlist, &sqlParmsObj)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!sequencePy2C (self, sqlParmsObj, &sqlParms)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = executePrepared (nself, sqlParms);
    if (!sqlResultOK (&result) && sqlErrOccured (nself->session, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, result, &resultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return resultObj;
}

static PyMethodDef SapDB_PreparedClassMethods [] = {
    {LIT("execute"), (PyCFunction)execute_SapDB_Prepared, METH_VARARGS|METH_KEYWORDS, 
        "execute statement with arguments.\n\n\n                    Output values are returned as a list.\n                    If the command creates a result set, then a\n                    ResultSet is returned.<BR>\n                    <B>Perl:</B> the arguments must be specified as\n                    a Reference to an array <i>[$a, $b, ...]</i>, not\n                    an array itself <i>($a, $b, ...)</i>.\n                    <B>Python:</B> the arguments must be specified\n                    as a list."},
    {NULL, NULL}
};


static PyObject*
SapDB_PreparedGetattr (PyObject* self, char* name)
{
    return Py_FindMethod (SapDB_PreparedClassMethods, self, name);
}

statichere PyTypeObject SapDB_PreparedType = {
    PyObject_HEAD_INIT (NULL)
    0,
    LIT("SapDB_Prepared"),    /* tp_name */
    sizeof (SapDB_PreparedObjectT),    /* tp_basicsize */
    0,    /* tp_itemsize */
    REINTERPRET_CAST (destructor, destructor_SapDB_Prepared),    /* tp_dealloc */
    0,    /* tp_print */
    REINTERPRET_CAST (getattrfunc, SapDB_PreparedGetattr),    /* tp_getattr */
    0,    /* tp_setattr */
    0,    /* tp_compare */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_repr */
    0,    /* tp_as_number */
    0,    /* tp_as_sequence */
    0,    /* tp_as_mapping */
    REINTERPRET_CAST (hashfunc, 0),    /* tp_hash */
    REINTERPRET_CAST (ternaryfunc, 0),    /* tp_call */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_str */
    0,    /* tp_getattro */
    0,    /* tp_setattro */
    0,    /* *tp_as_buffer */
    0,    /* tp_xxx4 */
    "allows execution of SQL command with parameters.\n\n        Objects are created with SapDB_Session::prepare.\n        ",    /* tp_doc */
};

typedef struct SapDB_SessionObjectT {
    PyObject_HEAD
    tin01_sql_session *nself;
/* snippet SapDB_Session Generic_instance vars */
    tin01_sql_session session;
    bool              autocommit;
    int               cursorCount;
    bool              implicitCursorNames;
    bool              lastWasEOT;
/* endsnippet SapDB_Session Generic_instance vars */
} SapDB_SessionObjectT;

staticforward PyTypeObject SapDB_SessionType;
#define isSapDB_SessionObject(op) ((op)->ob_type == &SapDB_SessionType)


static SapDB_SessionObjectT*
newSapDB_Session ()
{
    SapDB_SessionObjectT* result;

    result = PyObject_NEW (SapDB_SessionObjectT, &SapDB_SessionType);
    if (result != NULL) {
        memset (&result->nself, '\0', sizeof (SapDB_SessionObjectT) - offsetof (SapDB_SessionObjectT, nself));
    }
    return result;
}

static void
freeSapDB_Session (
    SapDB_SessionObjectT * value)
{
    Py_DECREF (value);
}
/* snippet SapDB_Session Generic_precode */

static void
setAutocommit (
    tin01_sql_session * session,
    bool autocommit)
{
    ((SapDB_SessionObjectT*)session->more_data)->autocommit = autocommit;
}

/*----------------------------------------*/

static bool
getAutocommit (
    tin01_sql_session * session)
{
    if (((SapDB_SessionObjectT*)session->more_data)->autocommit) {
        return true;
    }
    else {
        return false;
    }
}

/*----------------------------------------*/

static bool
lastWasEOT (
    tin01_sql_session * session)
{
    if (((SapDB_SessionObjectT*)session->more_data)->autocommit) {
        return true;
    }
    if (((SapDB_SessionObjectT*)session->more_data)->lastWasEOT) {
        return true;
    }
    return false;
}

/*----------------------------------------*/

static void
enableCursorNames (
    tin01_sql_session * session,
    bool enable)
{
    ((SapDB_SessionObjectT*)session->more_data)->implicitCursorNames = enable;
}

/*----------------------------------------*/

static int
nextCursorCount (
    tin01_sql_session * session)
{
    SapDB_SessionObjectT * self = (SapDB_SessionObjectT *) session->more_data;
    ++self->cursorCount;
    return self->cursorCount;
}

/*----------------------------------------*/

static void
addCursorName (
    tin01_sql_session * session)
{
    #define NAME_LEN_C 13
    char cursorName [NAME_LEN_C];
    int cursorCount = nextCursorCount (session);
    int index = NAME_LEN_C - 1;

    if (((SapDB_SessionObjectT *) session->more_data)->implicitCursorNames) {
        memcpy (cursorName, "PCURSOR_00000", NAME_LEN_C);
        while (cursorCount > 0) {
            cursorName [index] = (cursorCount % 10) + '0';
            cursorCount /= 10;
            --index;
        }
        i28newpart (session, sp1pk_resulttablename);
        i28addpartdata (session, cursorName, NAME_LEN_C);
    }
    #undef NAME_LEN_C
}

/*----------------------------------------*/

static void
clearEOTFlag (
    tin01_sql_session * session)
{
    ((SapDB_SessionObjectT*)session->more_data)->lastWasEOT = false;
}

/*----------------------------------------*/

static void
setEOTFlag (
    tin01_sql_session * session)
{
    ((SapDB_SessionObjectT*)session->more_data)->lastWasEOT = true;
}

/*----------------------------------------*/



static boolean
initializeRTE ()
{
    ROUTINE_DBG_MSP00 ("initializeRTE");
    tsp_compname compName;

    c2p (compName, sizeof (compName), "R SAP DB Scripting");
    sqlinit (compName, NULL);
    return true;
}

/*----------------------------------------*/

static void
fillXUser (
    tsp4_xuser_record * xuser,
    const char * user,
    const char * pwd,
    const char * dbname,
    const char * node)
{
    ROUTINE_DBG_MSP00 ("fillXUser");
    tsp_name clearPwd;
    int i = 0;

    c2p (xuser->xu_user, sizeof (xuser->xu_user), user);
    c2p (clearPwd, sizeof (clearPwd), pwd);
    for (i = 0; i < sizeof (clearPwd); ++i) {
        clearPwd [i] = toupper (clearPwd [i]);
    }
    s02applencrypt (clearPwd, xuser->xu_password);
    c2p (xuser->xu_serverdb, sizeof (xuser->xu_serverdb), dbname);
    c2p (xuser->xu_servernode, sizeof (xuser->xu_servernode), node);
    c2p (xuser->xu_sqlmode, sizeof (xuser->xu_sqlmode), "ADABAS");
    xuser->xu_cachelimit = -1;
    xuser->xu_timeout = -1;
    xuser->xu_isolation = -1;
    c2p (xuser->xu_dblang, sizeof (xuser->xu_dblang), "");;
}

/*----------------------------------------*/

static void
localRelease (
    tin01_sql_session* nself)
{
    ROUTINE_DBG_MSP00 ("localRelease");

    if ((nself != NULL) && (nself->is_connected)) {
        i28release (nself, 0);
    }
}

/*----------------------------------------*/


#if defined (PYTHON_GLUE)
static void
destructor_SapDB_Session (
    PyObject * pyself)
{
    ROUTINE_DBG_MSP00 ("destructor_SapDB_Session");
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);

    localRelease (self->nself);
    free (pyself);
}

/*----------------------------------------*/
#endif

static int
sql (
    tin01_sql_session * session,
    const char * cmd,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("sql");
    int sqlrc;
    int retryCount = 3;

    initResult (result, 0);
    do {
        i28initadbs (session);
        i28pcmd (session, cmd, UNDEF_SP00);
        if (getAutocommit (session)) {
            session->segment->variant.C_3.sp1c_commit_immediately_F = true;
        }
        addCursorName (session);
        sqlrc = requestReceive (session, true, true);
        --retryCount;
    } while ((sqlrc == cin01_db_parse_again) && (retryCount > 0));
    if (sqlrc == 0) {
        packet2result (session, NULL, result);
    }
    else {
        result->isOK = false;
    }
    return sqlrc;
}

/*----------------------------------------*/

static int
sqlX (
    tin01_sql_session * session,
    const char * cmd,
    SequenceT parms,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("sqlX");
    int sqlrc;
    PreparedT prepared;

    memset (&prepared, '\0', sizeof (prepared));
    sqlrc = prepare (session, &prepared, cmd, true);
    if (sqlrc == 0) {
        *result = executePrepared (&prepared, parms);
        closePrepared (&prepared);
    }
    else {
        result->isOK = false;
    }
    return sqlrc;
}

/*----------------------------------------*/

static int
utility (
    tin01_sql_session * session,
    const char * cmd,
    SQLResultT * result)
{
    ROUTINE_DBG_MSP00 ("utility");
    int sqlrc;

    initResult (result, 1);
    sqlrc = i28utility (session, cmd);
    if ((sqlrc == 0) || (i28findpart (session, sp1pk_data))) {
        packet2result (session, NULL, result);
        if (result->isTuple) {
            setResultElement (result, -1, SL_fromInt (sqlrc));
        }
        sqlrc = 0;
    }
    else {
        result->isOK = false;
    }
    return sqlrc;
}

/*----------------------------------------*/

static int
commit (
    tin01_sql_session * session)
{
    int result = i28adbs_noinfo (session, "COMMIT WORK", false);
    setEOTFlag (session);
    return result;
}

/*----------------------------------------*/

static int
rollback (
    tin01_sql_session * session)
{
    int result = i28adbs_noinfo (session, "ROLLBACK WORK", false);
    setEOTFlag (session);
    return result;
}

/*----------------------------------------*/

#if defined (PERL_GLUE)
static void
increaseRef (
    void * obj)
{
    ROUTINE_DBG_MSP00 ("increaseRef");
    SapDB_SessionObjectT * perlSession =
        REINTERPRET_CAST (SapDB_SessionObjectT*, obj);

    ++perlSession->refCount;
}

/*----------------------------------------*/

static void
decreaseRef (
    void * obj)
{
    ROUTINE_DBG_MSP00 ("decreaseRef");
    SapDB_SessionObjectT * perlSession =
        REINTERPRET_CAST (SapDB_SessionObjectT*, obj);

    --perlSession->refCount;
    if (perlSession->refCount < 0) {
        localRelease (perlSession->nself);
        freeSapDB_Session (perlSession);
    }
}

#endif
/*----------------------------------------*/

/* endsnippet SapDB_Session Generic_precode */
/*----------------------------------------*/

static PyObject *
sql_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("sql_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("cmd"), NULL};
    char * cmd;
    SQLResultT sqlResult;
    PyObject * sqlResultObj;
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("s"), (char **) kwlist, &cmd)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = sql (nself, cmd, &sqlResult);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, sqlResult, &sqlResultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return sqlResultObj;
}

/*----------------------------------------*/

static PyObject *
sqlX_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("sqlX_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("cmd"), LIT("parms"), NULL};
    char * cmd;
    SequenceT parms =  EmptySequenceC;
    PyObject * parmsObj = NULL;
    SQLResultT sqlResult;
    PyObject * sqlResultObj;
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("s|O"), (char **) kwlist, &cmd, &parmsObj)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!sequencePy2C (self, parmsObj, &parms)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = sqlX (nself, cmd, parms, &sqlResult);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, sqlResult, &sqlResultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return sqlResultObj;
}

/*----------------------------------------*/

static PyObject *
prepare_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("prepare_SapDB_Session");  /* generated by Glue */
/* no code for key SapDB_Session::prepare Python_/Generic_ in ['Generic_nativeCall vars']*/
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    SapDB_PreparedObjectT * newObj = NULL;
    static const char * kwlist [] = {
        LIT("cmd"), NULL};
    char * cmd;
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    newObj = newSapDB_Prepared ();
    if (newObj == NULL) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("s"), (char **) kwlist, &cmd)) {
        ok = 0;
        goto cleanup_label;
    }
/* snippet SapDB_Session::prepare Generic_nativeCall */
result = prepare (nself, &newObj->nself, cmd, true);
if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
    ok = 0;
    goto cleanup_label;
}
/* endsnippet SapDB_Session::prepare Generic_nativeCall */
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        Py_DECREF (newObj);
        return NULL;
    }
    return REINTERPRET_CAST (PyObject *, newObj);
}

/*----------------------------------------*/

static PyObject *
release_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("release_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("withCommit"), NULL};
    int withCommit =  0;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("|i"), (char **) kwlist, &withCommit)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    i28release (nself, withCommit);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
commit_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("commit_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        NULL};
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = commit (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
rollback_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("rollback_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        NULL};
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = rollback (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
setAutocommit_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("setAutocommit_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("autocommit"), NULL};
    int autocommit =  1;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("|i"), (char **) kwlist, &autocommit)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    setAutocommit (nself, autocommit);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
getAutocommit_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("getAutocommit_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        NULL};
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = getAutocommit (nself);
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return Py_BuildValue (LIT("i"), result);
}

/*----------------------------------------*/

static PyObject *
utility_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("utility_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("cmd"), NULL};
    char * cmd;
    SQLResultT sqlResult;
    PyObject * sqlResultObj;
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("s"), (char **) kwlist, &cmd)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = utility (nself, cmd, &sqlResult);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
    if (!sqlResultC2Py (self, sqlResult, &sqlResultObj)) {
        ok = 0;
        goto cleanup_label;
    }
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    return sqlResultObj;
}

/*----------------------------------------*/

static PyObject *
switch_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("switch_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("trace"), LIT("test"), NULL};
    char * trace;
    char * test;
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("ss"), (char **) kwlist, &trace, &test)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = i28switch (nself, trace, test);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
switchlimit_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("switchlimit_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("trace"), LIT("test"), LIT("startroutine"), LIT("stoproutine"), LIT("buflimit"), NULL};
    char * trace;
    char * test;
    char * startroutine;
    char * stoproutine;
    int buflimit;
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("ssssi"), (char **) kwlist, &trace, &test, &startroutine, &stoproutine, &buflimit)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = i28switchlimit (nself, trace, test, startroutine, stoproutine, buflimit);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
buflimit_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("buflimit_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        LIT("buflimit"), NULL};
    int buflimit;
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("i"), (char **) kwlist, &buflimit)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = i28buflimit (nself, buflimit);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
minbuf_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("minbuf_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        NULL};
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = i28minbuf (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

/*----------------------------------------*/

static PyObject *
maxbuf_SapDB_Session (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("maxbuf_SapDB_Session");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    SapDB_SessionObjectT* self = REINTERPRET_CAST (SapDB_SessionObjectT*, pyself);
    tin01_sql_session* nself = self->nself;
    static const char * kwlist [] = {
        NULL};
    int result;

    if (!isSessionConnected (self->nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
/* no code for key Python_/Generic_nativeCall */
    result = i28maxbuf (nself);
    if ((result != 0) && sqlErrOccured (nself, croakBuf)) {
        ok = 0;
        goto cleanup_label;
    }
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        return NULL;
    }
    Py_INCREF(Py_None);
    return Py_None;
}

static PyMethodDef SapDB_SessionClassMethods [] = {
    {LIT("sql"), (PyCFunction)sql_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "exeute a SQL command\n\n                Returns data values on SELECT INTO, a ResultSet on SELECT\n                and an empty list otherwise.\n                "},
    {LIT("sqlX"), (PyCFunction)sqlX_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "exeute a SQL command with arguments\n\n                implemented using parse-execute.<BR>\n                Returns data values on SELECT INTO, a ResultSet on SELECT\n                and an empty list otherwise.\n                <B>Perl:</B> the arguments must be given as\n                    a Reference to an array <i>[$a, $b, ...]</i>, not\n                    an array itself <i>($a, $b, ...)</i>."},
    {LIT("prepare"), (PyCFunction)prepare_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "parse a SQL statement, resulting in a SapDB_Prepared object"},
    {LIT("release"), (PyCFunction)release_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "closes the connection\n\n                The default is to end the session without COMMIT"},
    {LIT("commit"), (PyCFunction)commit_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "commits the current transaction."},
    {LIT("rollback"), (PyCFunction)rollback_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "aborts the current transaction."},
    {LIT("setAutocommit"), (PyCFunction)setAutocommit_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "switches autocommit on or off."},
    {LIT("getAutocommit"), (PyCFunction)getAutocommit_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "returns the current autocommit state."},
    {LIT("utility"), (PyCFunction)utility_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "exeute a utility command"},
    {LIT("switch"), (PyCFunction)switch_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "switches traces on and off"},
    {LIT("switchlimit"), (PyCFunction)switchlimit_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "switches kernel traces to start at routine <i>startroutine</I>\n                and end at routine <i>stoproutine</i>"},
    {LIT("buflimit"), (PyCFunction)buflimit_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "limits buffer output in kernel traces to <i>limit</i> bytes"},
    {LIT("minbuf"), (PyCFunction)minbuf_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "sets compact buffer output (hex + char)"},
    {LIT("maxbuf"), (PyCFunction)maxbuf_SapDB_Session, METH_VARARGS|METH_KEYWORDS, 
        "sets full buffer output (hex, decimal, char  [+ EBCDIC])"},
    {NULL, NULL}
};


static PyObject*
SapDB_SessionGetattr (PyObject* self, char* name)
{
    return Py_FindMethod (SapDB_SessionClassMethods, self, name);
}

statichere PyTypeObject SapDB_SessionType = {
    PyObject_HEAD_INIT (NULL)
    0,
    LIT("SapDB_Session"),    /* tp_name */
    sizeof (SapDB_SessionObjectT),    /* tp_basicsize */
    0,    /* tp_itemsize */
    REINTERPRET_CAST (destructor, destructor_SapDB_Session),    /* tp_dealloc */
    0,    /* tp_print */
    REINTERPRET_CAST (getattrfunc, SapDB_SessionGetattr),    /* tp_getattr */
    0,    /* tp_setattr */
    0,    /* tp_compare */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_repr */
    0,    /* tp_as_number */
    0,    /* tp_as_sequence */
    0,    /* tp_as_mapping */
    REINTERPRET_CAST (hashfunc, 0),    /* tp_hash */
    REINTERPRET_CAST (ternaryfunc, 0),    /* tp_call */
    REINTERPRET_CAST (reprfunc, 0),    /* tp_str */
    0,    /* tp_getattro */
    0,    /* tp_setattro */
    0,    /* *tp_as_buffer */
    0,    /* tp_xxx4 */
    "allows execution of sql commands.\n\n        When the last reference to the session is dropped, the database\n        connection is released with an implicit transaction rollback.\n        ",    /* tp_doc */
};

/*----------------------------------------*/

static PyObject *
connect_sapdb (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("connect_sapdb");  /* generated by Glue */
/* no code for key connect_sapdb Python_/Generic_ in ['Generic_nativeCall']*/
    int ok = 1;
    SapDB_SessionObjectT * newObj = NULL;
    static const char * kwlist [] = {
        LIT("user"), LIT("pwd"), LIT("dbname"), LIT("host"), LIT("config"), NULL};
    char * user;
    char * pwd;
    char * dbname;
    char * host = LIT( "");
    char * config = LIT( "");
    tsp_int4 result;

    newObj = newSapDB_Session ();
    if (newObj == NULL) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("sss|ss"), (char **) kwlist, &user, &pwd, &dbname, &host, &config)) {
        ok = 0;
        goto cleanup_label;
    }
/* snippet connect_sapdb Generic_nativeCall */
newObj->nself = &newObj->session;
i28initsession (newObj->nself, NULL);
newObj->session.lasterr.lasterr_on = true;
newObj->autocommit = false;
newObj->cursorCount = 0;
newObj->implicitCursorNames = true;
newObj->lastWasEOT = true;
/* set special senderid */
setStringConfig ((char *) newObj->session.senderid + sizeof (tsp00_C5),
    sizeof (tsp00_C3), "DOM", 3);
fillXUser (&newObj->nself->xuser, user, pwd, dbname, host);
newObj->nself->more_data = newObj;
parseSessionConfig (newObj->nself, config);
result = i28sqlconnect (newObj->nself);
if ((result != 0) && sqlErrOccured (newObj->nself, croakBuf)) {
    ok = 0;
    goto cleanup_label;
}
/* endsnippet connect_sapdb Generic_nativeCall */
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        Py_DECREF (newObj);
        return NULL;
    }
    return REINTERPRET_CAST (PyObject *, newObj);
}

/*----------------------------------------*/

static PyObject *
utilitySession_sapdb (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("utilitySession_sapdb");  /* generated by Glue */
/* no code for key utilitySession_sapdb Python_/Generic_ in ['Generic_nativeCall']*/
    int ok = 1;
    SapDB_SessionObjectT * newObj = NULL;
    static const char * kwlist [] = {
        LIT("user"), LIT("pwd"), LIT("dbname"), LIT("host"), NULL};
    char * user;
    char * pwd;
    char * dbname;
    char * host = LIT( "");
    tsp_int4 result;

    newObj = newSapDB_Session ();
    if (newObj == NULL) {
        ok = 0;
        goto cleanup_label;
    }
    if (!PyArg_ParseTupleAndKeywords (args, keywds, LIT("sss|s"), (char **) kwlist, &user, &pwd, &dbname, &host)) {
        ok = 0;
        goto cleanup_label;
    }
/* snippet utilitySession_sapdb Generic_nativeCall */
newObj->nself = &newObj->session;
i28initsession (newObj->nself, NULL);
newObj->session.lasterr.lasterr_on = true;
fillXUser (&newObj->nself->xuser, user, pwd, dbname, host);
newObj->nself->as_utility = 1;
result = i28sqlconnect (newObj->nself);
if ((result != 0) && sqlErrOccured (newObj->nself, croakBuf)) {
    ok = 0;
    goto cleanup_label;
}
newObj->nself->more_data = newObj;
/* endsnippet utilitySession_sapdb Generic_nativeCall */
    /* outargs */
  cleanup_label:
    if (!ok) {
        /* Cleanup */
        Py_DECREF (newObj);
        return NULL;
    }
    return REINTERPRET_CAST (PyObject *, newObj);
}

/*----------------------------------------*/

static PyObject *
_buildInfo_sapdb (PyObject *pyself, PyObject* args, PyObject* keywds)
{
    ROUTINE_DBG_MSP00 ("_buildInfo_sapdb");  /* generated by Glue */
/* no code for key Python_/Generic_code */
    int ok = 1;
    static const char * kwlist [] = {
        NULL};
    char * result;

/* no code for key Python_/Generic_nativeCall */
    result = buildInfo ();
    /* outargs */
    return Py_BuildValue (LIT("s"), result);
}

static PyMethodDef sapdbModuleMethods [] = {
    {LIT("connect"), (PyCFunction)connect_sapdb, METH_VARARGS|METH_KEYWORDS, 
        "connect to database.\n\n                  The username is case sensitive. <code>config</code>\n                  is a string of the format:<BR>\n                  <i>keyword=value[&amp;keyword=value]...</i><BR>\n                  Allowed keywords are:\n                  <DL>\n                  <DT>version<DD>the assumed version of the client (e.g. '70200')\n                  <DT>component<DD>the component string (ODB, XCI, ...)\n                  <DT>sqlmode<DD>(adabas, oracle, sapr3, ansi, db2)\n                  <DT>isolation<DD>the isolation level as an integer\n                  <DT>timeout<DD>the command timeout in seconds\n                  <DT>cachelimit<DD>\n                  </DL>"},
    {LIT("utilitySession"), (PyCFunction)utilitySession_sapdb, METH_VARARGS|METH_KEYWORDS, 
        "connect as utility user.\n\n                  The username is case sensitive."},
    {LIT("_buildInfo"), (PyCFunction)_buildInfo_sapdb, METH_VARARGS|METH_KEYWORDS, 
        "returns a string containing the release and build number"},
    {NULL, NULL}
};

externC GLUEEXPORT void
initsapdbInternal (const char * moduleName)
{
    ROUTINE_DBG_MSP00 ("initsapdb");  /* generated by Glue */
    PyObject* module;
    PyObject* dict;

    module = Py_InitModule4 ((char *) moduleName, sapdbModuleMethods, "Interface to SAP DB", NULL, PYTHON_API_VERSION);
    if (module == NULL) {
        return;
    }
    dict = PyModule_GetDict (module);
    CommunicationErrorType = createExceptionKind ("sapdb.CommunicationError", CommunicationErrorCodeC);
    PyDict_SetItemString (dict, LIT("CommunicationError"), CommunicationErrorType);
    SQLErrorType = createExceptionKind ("sapdb.SQLError", SQLErrorCodeC);
    PyDict_SetItemString (dict, LIT("SQLError"), SQLErrorType);
    SapDB_LongReaderType.ob_type = &PyType_Type;
    PyDict_SetItemString (dict, LIT("SapDB_LongReader"), REINTERPRET_CAST (PyObject*, &SapDB_LongReaderType));
    SapDB_ResultSetType.ob_type = &PyType_Type;
    PyDict_SetItemString (dict, LIT("SapDB_ResultSet"), REINTERPRET_CAST (PyObject*, &SapDB_ResultSetType));
    SapDB_PreparedType.ob_type = &PyType_Type;
    PyDict_SetItemString (dict, LIT("SapDB_Prepared"), REINTERPRET_CAST (PyObject*, &SapDB_PreparedType));
    SapDB_SessionType.ob_type = &PyType_Type;
    PyDict_SetItemString (dict, LIT("SapDB_Session"), REINTERPRET_CAST (PyObject*, &SapDB_SessionType));
    /* snippet Generic_initroutine */
    IFSLOW_MEO00 (ScriptingTracer::initTrace ());
    initializeRTE ();
    /* endsnippet Generic_initroutine */
    if (PyErr_Occurred ()) {
        Py_FatalError (LIT("can't initialize module sapdb"));
    }
}


externC GLUEEXPORT void
initsapdb ()
{
    initsapdbInternal ("sapdb");
}
        
externC GLUEEXPORT void
initsapdb72 ()
{
    initsapdbInternal ("sapdb72");
}
        