/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#ifndef VPA70SQLDA_H
#define VPA70SQLDA_H
/*!**********************************************************************

  module: vpa70SQLDA.h

  -----------------------------------------------------------------------

  responsible:  

  special area: ODBC-Driver

  description:  ODBC related sqlda functions

  see also:     http://www.microsoft.com/data/odbc/

  -----------------------------------------------------------------------

  copyright:           Copyright by SAP AG, 1998

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "vpa00global.h"

/*!**********************************************************************

  Chapter: exported functions

************************************************************************/

/*!
  Function: pa70AllocSQLDA

  description: Allocate sqlda array.

  arguments:
    phand        [out]  Buffer to return the new array handle
    num_elements [in]   The requested size of the array.
    
  return value: API_OK, API_NOT_OK
  */

API_RETCODE pa70AllocSQLDA( API_HANDLE *phand,
                            UWORD       num_elements );


/*!**********************************************************************
  Function: pa70FreeSQLDA

  description: Free memory for the sqlda array.

  arguments:
      hand  [in]  SQLDA handle
      
  return value: API_OK
  */

API_RETCODE pa70FreeSQLDA( API_HANDLE hand );


/*!**********************************************************************
  Function: pa70NumCol

  description: Return the number of columns in the sqlda structure.

  arguments:
    sqlda_ptr [in]  Pointer to sqlda-structure
    
  return value: The number of columns.
  */

UWORD pa70NumCol( sqldatype *sqlda_ptr );


/*!**********************************************************************
  Function: pa70NumParam

  description: Return the number of parameters in the sqlda structure.

  arguments:
    sqlda_ptr [in]  Pointer to sqlda-structure
      
  return value: The number of parameters.
  */

UWORD pa70NumParam( sqldatype *sqlda_ptr );


/*!**********************************************************************
  Function: pa70Patch

  description: Patch the column type and length for catalog functions.

  arguments:
    sqlda_ptr   [in]  Pointer to sqlda-structure
    number_cols [in]  Number of columns.
    special     [in]  Indicate special case (API_SPEC_...)
    odbcVersion [in]  SQL_OV_ODBC3 or SQL_OV_ODBC2. This is needed because
                      columns for catalog functions are different in
                      ODBC 3.x.
  return value: The number of parameters.
  */

API_RETCODE pa70patch( sqldatype *sqlda_ptr,
                       UWORD      number_cols,
                       UWORD      special,
                       UDWORD     odbcVersion );

/*!**********************************************************************

  EndChapter: exported functions

************************************************************************/

#endif
