/*!**********************************************************************

  module: vpa80String.c

  -----------------------------------------------------------------------

  responsible:  

  special area: ODBC-Driver

  description:  string manipulating functions

  see also:     http://www.microsoft.com/data/odbc/

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "vpa80String.h"

/*!**********************************************************************

  Chapter: exported functions

************************************************************************/


 /* apmstow() - API miscellaneous function, copy C string to ODBC */

API_RETCODE apmstow ( UCHAR *string,
                      UCHAR *wstring,
                      SWORD  wlen,
                      SWORD *size_ptr )
{
  API_RETCODE retcode;
				  /* set the real size/len */
  if (size_ptr != NULL) {
    *size_ptr = (SWORD) API_STRLEN(string);
  }

  if (wstring == NULL || wlen == 0) {
				  /* can't copy any */
    retcode = API_TRUNCATE;
  }
  else if ((SWORD)API_STRLEN(string)+1 <= wlen) {
				  /* it fits, copy it           */
    API_STRCPY(wstring,string);
    retcode = API_OK;
  }
  else {
				  /* it does not fit, copy some */
    API_MEMCPY(wstring,string,(int) (wlen - 1));
    wstring [wlen - 1] = 0;
    retcode = API_TRUNCATE;
  }

  return(retcode);
} /* apmstow */

/* apmwtos() - API miscellaneous function, copy ODBC string to C */

API_RETCODE apmwtos ( UCHAR *wstring,
                      SWORD  wlen,
                      UCHAR *string,
                      SWORD  slen)

{

  API_RETCODE retcode;
  API_TRACE(API_TR_ENTRY,"apmwtos",0);

				  /* is it a NULL terminal      */
				  /* string?                    */
  switch (wlen) {
    case (SQL_NULL_DATA):
      if (slen > 1) {
          API_STRCPY(string,(UCHAR FAR *) "");
          retcode = API_OK;
      }
      else {
          retcode = API_NOT_OK;
      }
      break;
    case (SQL_NTS):
      if (wstring == NULL) {
          API_STRCPY(string,(UCHAR FAR *) "");
          retcode = API_OK;
      }
      else {
          if ((SWORD)API_STRLEN(wstring)+1 <= slen) {
              API_STRCPY(string,wstring);
              retcode = API_OK;
          }
          else {
              retcode = API_NOT_OK;
          }
      }
      break;
    default:                      /* use lengths                */
      if (wstring == NULL) {
          API_STRCPY(string,(UCHAR FAR *) "");
          retcode = API_OK;
      }
      else {
          if (wlen+1 <= slen) {
              API_MEMCPY(string,wstring,(int) wlen);
              string [wlen] = 0;
              retcode = API_OK;
          }
          else {
              retcode = API_NOT_OK;
          }
      }
      break;
  }

  API_TRACE(API_TR_EXIT,"apmwtos",0);
  API_TRACE(API_TR_API_RETCODE,"retcode",&retcode);

  return(retcode);
} /* apmwtos */

/* apmwchs() - API miscellaneous function, check ODBC string */

API_RETCODE apmwchs ( UCHAR FAR * wstring,
  SWORD wlen,
  SWORD *len)

{
   API_RETCODE retcode;
   API_TRACE(API_TR_ENTRY,"apmwchs",0);
   retcode = API_OK;
   switch (wlen) {
      case (SQL_NULL_DATA):
	 *len = 0;
	 break;
      case (SQL_NTS):
	 *len = (SWORD)API_STRLEN(wstring);
	 break;
      default:                      /* use lengths                */
	 if (wstring == NULL)
	    *len = 0;
	 else
	    if (wlen > 0)
	       *len = wlen;
	    else 
	       retcode = API_NOT_OK;
	 break;
  }
  API_TRACE(API_TR_EXIT,"apmwchs",0);
  API_TRACE(API_TR_API_RETCODE,"retcode",&retcode);
  
  return(retcode);
} /* apmwchs */


API_RETCODE pa80AllocString( UCHAR *source, SDWORD len, UCHAR **dest )
{
    SWORD strSize;
    
    if (len == SQL_NTS) {
        /* string length plus null-terminator */
        strSize = (SWORD) API_STRLEN((UCHAR*) source)+1;
        *dest = apdallo(strSize);
        API_STRCPY( *dest, source );
    } else {
        *dest = apdallo(len+1);
        API_MEMCPY( *dest, source, len );
        dest[len] = '\0';
    }; /* else */

    return API_OK;
} /* pa80AllocString */


RETCODE pa80CopyFixedString( UCHAR   *source,
                             SDWORD   sourceLen,
                             SDWORD   maxLen,
                             UCHAR   *dest,
                             SDWORD  *totalLen,
                             UWORD   *sqlStatePtr)
{
    SDWORD strSize;
    RETCODE retcode = SQL_SUCCESS;

    if (API_STRLEN( source ) == 0 ) {
        strSize = 0;
    } else {
        strSize = sourceLen;
        /* ignore blanks at the end */
        while ( (strSize > 0) 
                && (*(source + strSize -1) == ' ') ) {
            strSize--;
        }; /* while */
    }; /* else */
    /* null terminator */
    strSize++;
    if (strSize>maxLen) {
        if (maxLen>0) {
            API_MEMCPY( dest, source, maxLen-1 );
            dest[maxLen-1] = '\0';
        }; /* if */
        *sqlStatePtr = API_ODBC_01004; /* String data right truncated */
        retcode = SQL_SUCCESS_WITH_INFO;
    } else {
        if (strSize>1) {
            API_MEMCPY( dest, source, strSize-1 );
        }; /* if */
        dest[strSize-1] = '\0';
    }; /* else */
    /* string length minus null-terminator */
    if (totalLen != NULL) {
        *totalLen = strSize-1;
    }; /* if */
    
    return retcode;
} /* pa80CopyFixedString */


RETCODE pa80CopyString( UCHAR  *source,
                        SWORD   maxLen,
                        UCHAR  *dest,
                        SWORD  *totalLen,
                        UWORD  *sqlStatePtr)
{
    SWORD strSize;
    RETCODE retcode = SQL_SUCCESS;
    
    /* string length plus null-terminator */
    strSize = (SWORD) API_STRLEN((UCHAR*) source)+1;
    if (strSize>maxLen) {
        if (maxLen>0) {
            API_MEMCPY( dest, source, maxLen-1 );
            dest[maxLen-1] = '\0';
        }; /* if */
        *sqlStatePtr = API_ODBC_01004; /* String data right truncated */
        retcode = SQL_SUCCESS_WITH_INFO;
    } else {
        API_STRCPY( dest, source );
    }; /* else */
    /* string length minus null-terminator */
    if (totalLen != NULL) {
        *totalLen = strSize-1;
    }; /* if */
    
    return retcode;
} /* pa80CopyString */


RETCODE pa80CopyStringD( UCHAR   *source,
                         SDWORD   maxLen,
                         UCHAR   *dest,
                         SDWORD  *totalLen,
                         UWORD   *sqlStatePtr)
{
    SDWORD strSize;
    RETCODE retcode = SQL_SUCCESS;
    
    /* string length plus null-terminator */
    strSize = (SDWORD) API_STRLEN((UCHAR*) source)+1;
    if (strSize>maxLen) {
        if (maxLen>0) {
            API_MEMCPY( dest, source, maxLen-1 );
            dest[maxLen-1] = '\0';
        }; /* if */
        *sqlStatePtr = API_ODBC_01004; /* String data right truncated */
        retcode = SQL_SUCCESS_WITH_INFO;
    } else {
        API_STRCPY( dest, source );
    }; /* else */
    /* string length minus null-terminator */
    if (totalLen != NULL) {
        *totalLen = strSize-1;
    }; /* if */
    
    return retcode;
} /* pa80CopyStringD */

/*!**********************************************************************

  EndChapter: exported functions

************************************************************************/
