/*


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end


*/
#ifndef VPA80STRING_H
#define VPA80STRING_H
/*!**********************************************************************

  module: vpa80String.h

  -----------------------------------------------------------------------

  responsible:  

  special area: ODBC-Driver

  description:  string manipulating functions

  see also:     http://www.microsoft.com/data/odbc/

  -----------------------------------------------------------------------

  copyright:            Copyright by SAP AG, 1998

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "vpa00global.h"

/*!**********************************************************************

  Chapter: exported functions

************************************************************************/

/*!
  Function: pa80AllocString

  description: Allocates and copies a string of given length.

  When an eror occurs (e.g not enough memory) API_NOT_OK is returned.
  If the string was successfully copied the function return API_OK.

  arguments:
    UCHAR   *source [in]   The string that is copied.
    SDWORD   len    [in]   The length of source or SQL_NTS.
    UCHAR  **dest   [out]  Points to the char-pointer that should be
                           allocated.
    
  return value: API_RETCODE (API_OK or API_NOT_OK)
  */

API_RETCODE pa80AllocString( UCHAR *source, SDWORD len, UCHAR **dest );


/*!*********************************************************************
  Function: pa80CopyFixedString

  description: Copies a fixed string into a buffer of given length.
               Length buffers have type SDWORD.

  The source string must not contain a null terminator. Trailing blanks
  of source are ignored.
  If the length of source +1 (null-terminator) is bigger than maxLen then
  only maxLen characters are copied, *sqlStatePtr is set to 01004
  (string data right truncated) and SQL_SUCCESS_WITH_INFO is returned.
  If the dest buffer is big enough for source the function returns
  SQL_SUCCESS.

  arguments:
    UCHAR   *source      [in]   The string that is copied.
    SDWORD   maxLen      [in]   The length of *dest.
    SDWORD   sourceLen   [in]   The length of source.
    UCHAR   *dest        [out]  Points to the target buffer.
    SDWORD  *totalLen    [out]  Points to a variable that holds the total
                                length of source.
    UWORD   *sqlStatePtr [out]  Points a sql-state variable
    
  return value: RETCODE (SQL_SUCCESS or SQL_SUCCESS_WITH_INFO)
  */

RETCODE pa80CopyFixedString( UCHAR   *source,
                             SDWORD   sourceLen,
                             SDWORD   maxLen,
                             UCHAR   *dest,
                             SDWORD  *totalLen,
                             UWORD   *sqlStatePtr );


/*!*********************************************************************
  Function: pa80CopyString

  description: Copies a string into a buffer of given length.
               Length buffers have type SWORD.

  If the length of source +1 (null-terminator) is bigger than maxLen then
  only maxLen characters are copied, *sqlStatePtr is set to 01004
  (string data right truncated) and SQL_SUCCESS_WITH_INFO is returned.
  If the dest buffer is big enough for source the function returns
  SQL_SUCCESS.

  arguments:
    UCHAR   *source      [in]   The string that is copied.
    SWORD    maxLen      [in]   The length of *dest.
    UCHAR   *dest        [out]  Points to the target buffer.
    SWORD   *totalLen    [out]  Points to a variable that holds the total
                                length of source.
    UWORD   *sqlStatePtr [out]  Points a sql-state variable
    
  return value: RETCODE (SQL_SUCCESS or SQL_SUCCESS_WITH_INFO)
  */

RETCODE pa80CopyString( UCHAR  *source,
                        SWORD   maxLen,
                        UCHAR  *dest,
                        SWORD  *totalLen,
                        UWORD  *sqlStatePtr );


/*!*********************************************************************
  Function: pa80CopyStringD

  description: Copies a string into a buffer of given length.
               Length buffers have type SDWORD.

  If the length of source +1 (null-terminator) is bigger than maxLen then
  only maxLen characters are copied, *sqlStatePtr is set to 01004
  (string data right truncated) and SQL_SUCCESS_WITH_INFO is returned.
  If the dest buffer is big enough for source the function returns
  SQL_SUCCESS.

  arguments:
    UCHAR   *source      [in]   The string that is copied.
    SDWORD   maxLen      [in]   The length of *dest.
    UCHAR   *dest        [out]  Points to the target buffer.
    SDWORD  *totalLen    [out]  Points to a variable that holds the total
                                length of source.
    UWORD   *sqlStatePtr [out]  Points a sql-state variable
    
  return value: RETCODE (SQL_SUCCESS or SQL_SUCCESS_WITH_INFO)
  */

RETCODE pa80CopyStringD( UCHAR   *source,
                         SDWORD   maxLen,
                         UCHAR   *dest,
                         SDWORD  *totalLen,
                         UWORD   *sqlStatePtr );

/*!**********************************************************************

  EndChapter: exported functions

************************************************************************/

#endif
