/*!**********************************************************************

  module: vpa82unix.c

  -----------------------------------------------------------------------

  responsible:  

  special area: ODBC-Driver

  description:  Mapping ODBC 2.5 functions to ODBC 3.5 functions

  see also:     http://www.microsoft.com/data/odbc/

  -----------------------------------------------------------------------

                         Copyright by SAP AG, 1998



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

************************************************************************/

/*=======================================================================
 *  INCLUDES
 *=====================================================================*/

#include "vpa00global.h"
#include "vpa40DBC.h"
#include "vpa50Env.h"
#include "vpa60Stmt.h"

#include "hpa83odbcinstext.h" /* definition of HODBCINSTPROPERTY */

/*!**********************************************************************

  Chapter: Setup for unixODBC

************************************************************************/

int ODBCINSTGetProperties( HODBCINSTPROPERTY hLastProperty )
{
	hLastProperty->pNext 				= (HODBCINSTPROPERTY) apdallo( sizeof(ODBCINSTPROPERTY) );
	hLastProperty 						= hLastProperty->pNext;
	memset( hLastProperty, 0, sizeof(ODBCINSTPROPERTY) );
    
    hLastProperty->nPromptType			= ODBCINST_PROMPTTYPE_TEXTEDIT;
	strncpy( hLastProperty->szName, "Host", INI_MAX_PROPERTY_NAME );
	strncpy( hLastProperty->szValue, "", INI_MAX_PROPERTY_VALUE );

	hLastProperty->pNext 				= (HODBCINSTPROPERTY) apdallo( sizeof(ODBCINSTPROPERTY) );
	hLastProperty 						= hLastProperty->pNext;
	memset( hLastProperty, 0, sizeof(ODBCINSTPROPERTY) );
    
	hLastProperty->nPromptType			= ODBCINST_PROMPTTYPE_TEXTEDIT;
	strncpy( hLastProperty->szName, "Database", INI_MAX_PROPERTY_NAME );
	strncpy( hLastProperty->szValue, "", INI_MAX_PROPERTY_VALUE );

	hLastProperty->pNext 				= (HODBCINSTPROPERTY) apdallo( sizeof(ODBCINSTPROPERTY) );
	hLastProperty 						= hLastProperty->pNext;
	memset( hLastProperty, 0, sizeof(ODBCINSTPROPERTY) );
    
	hLastProperty->nPromptType			= ODBCINST_PROMPTTYPE_TEXTEDIT;
	strncpy( hLastProperty->szName, "Tracefilename", INI_MAX_PROPERTY_NAME );
	strncpy( hLastProperty->szValue, "", INI_MAX_PROPERTY_VALUE );

	return 1;
} /* ODBCINSTGetProperties */

/*!**********************************************************************

  EndChapter: Setup for unixODBC

************************************************************************/


/*!**********************************************************************

  Chapter: ODBC functions

************************************************************************/

ODBC_PROC(SQLAllocConnect,( SQLHENV  henv,
                            SQLHDBC *phdbc ),
           ( henv, phdbc ))
{
    return CALL_ODBC( SQLAllocHandle,
                      ( SQL_HANDLE_DBC, henv, phdbc ) );
} /* SQLAllocConnect */


ODBC_PROC( SQLAllocEnv,( SQLHENV  *phenv ),
           ( phenv ))
{
    return CALL_ODBC( SQLAllocHandle,
                      ( SQL_HANDLE_ENV, SQL_NULL_HANDLE, phenv ));
} /* SQLAllocEnv */


ODBC_PROC( SQLAllocStmt,( SQLHDBC   hdbc,
                          SQLHSTMT *phstmt ),
           ( hdbc, phstmt ))
{
    return CALL_ODBC( SQLAllocHandle,
                      ( SQL_HANDLE_STMT, hdbc, phstmt ));
} /* SQLAllocEnv */


ODBC_PROC( SQLBindParam,( SQLHSTMT  hstmt,
                          UWORD     ipar,
                          SWORD     fCType,
                          SWORD     fSqlType,
                          UDWORD    cbColDef,
                          SWORD     ibScale,
                          PTR       rgbValue,
                          SDWORD   *pcbValue),
          (hstmt,
           ipar,
           fCType,
           fSqlType,
           cbColDef,
           ibScale,
           rgbValue,
           pcbValue))
{
    return CALL_ODBC( SQLBindParameter,( hstmt,
                                         ipar,
                                         SQL_PARAM_INPUT,
                                         fCType,
                                         fSqlType,
                                         cbColDef,
                                         ibScale,
                                         rgbValue,
                                         *pcbValue,
                                         pcbValue ));
} /* SQLBindParam */


ODBC_PROC( SQLColAttributes,
           (SQLHSTMT     hstmt,
            UWORD        icol,
            UWORD        fDescType,
            PTR          rgbDesc,
            SWORD        cbDescMax,
            SWORD       *pcbDesc,
            SDWORD      *pfDesc),
           (hstmt, icol, fDescType, rgbDesc, cbDescMax, pcbDesc, pfDesc))
{
    /* date-time data-type conversion for SQL_COLUMN_TYPE ... */
    return CALL_ODBC( SQLColAttribute,
                      (hstmt,
                       icol,
                       fDescType,
                       rgbDesc,
                       cbDescMax,
                       pcbDesc,
                       pfDesc));
} /* SQLColAttribute */


ODBC_PROC(SQLError,
          (SQLHENV    henv,
           SQLHDBC    hdbc,
           SQLHSTMT   hstmt,
           UCHAR     *szSqlState,
           SDWORD    *pfNativeError,
           UCHAR     *szErrorMsg,
           SWORD      cbErrorMsgMax,
           SWORD     *pcbErrorMsg),
          (henv, hdbc, hstmt, szSqlState, pfNativeError,
           szErrorMsg, cbErrorMsgMax, pcbErrorMsg))
{
    SQLRETURN   retcode = SQL_SUCCESS;
    tpa60Stmt  *stmtPtr = NULL;
    tpa40DBC   *dbcPtr  = NULL;
    tpa50Env   *envPtr  = NULL;
    
    if (hstmt != SQL_NULL_HSTMT) {
        if (pa60VerifyStmt( hstmt )) {
            stmtPtr = (tpa60Stmt*) hstmt;
            retcode = CALL_ODBC( SQLGetDiagRec,( SQL_HANDLE_STMT,
                                                 hstmt,
                                                 stmtPtr->errorIndex+1,
                                                 szSqlState,
                                                 pfNativeError,
                                                 szErrorMsg,
                                                 cbErrorMsgMax,
                                                 pcbErrorMsg));
            if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO) {
                stmtPtr->errorIndex++;
            }; /* if */
        } else {
            retcode = SQL_INVALID_HANDLE;
        };
    } else if (hdbc != SQL_NULL_HDBC) {
        if (pa40VerifyDBC( hdbc )) {
            dbcPtr = (tpa40DBC*) hdbc;
            retcode = CALL_ODBC( SQLGetDiagRec,( SQL_HANDLE_DBC,
                                                 hdbc,
                                                 dbcPtr->errorIndex+1,
                                                 szSqlState,
                                                 pfNativeError,
                                                 szErrorMsg,
                                                 cbErrorMsgMax,
                                                 pcbErrorMsg));
            if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO) {
                dbcPtr->errorIndex++;
            }; /* if */
        } else {
            retcode = SQL_INVALID_HANDLE;
        };
    } else if (henv != SQL_NULL_HENV) {
        if (pa50VerifyEnv(henv)) {
            retcode = CALL_ODBC( SQLGetDiagRec,( SQL_HANDLE_ENV,
                                                 henv,
                                                 envPtr->errorIndex+1,
                                                 szSqlState,
                                                 pfNativeError,
                                                 szErrorMsg,
                                                 cbErrorMsgMax,
                                                 pcbErrorMsg));
            if (retcode == SQL_SUCCESS || retcode == SQL_SUCCESS_WITH_INFO) {
                envPtr->errorIndex++;
            }; /* if */
        } else {
            retcode = SQL_INVALID_HANDLE;
        };
    } else {
        retcode = SQL_INVALID_HANDLE;
    }; /* else */

    return retcode;
} /* SQLError */


ODBC_PROC(SQLFreeConnect, 
          (SQLHDBC    hdbc),
          (hdbc))
{
    return CALL_ODBC( SQLFreeHandle, ( SQL_HANDLE_DBC, hdbc ));
} /* SQLFreeConnect */


ODBC_PROC(SQLFreeEnv,
          (SQLHENV    henv),
          (henv))
{
    return CALL_ODBC( SQLFreeHandle, ( SQL_HANDLE_ENV, henv ));
} /* SQLFreeEnv */


ODBC_PROC(SQLGetConnectOption,
          (SQLHDBC    hdbc,
           UWORD      fOption,
           PTR        pvParam),
          (hdbc, fOption, pvParam))
{
    return CALL_ODBC( SQLGetConnectAttr, ( hdbc,
                                           fOption,
                                           pvParam,
                                           SQL_MAX_OPTION_STRING_LENGTH,
                                           NULL ));
} /* SQLGetConnectOption */


ODBC_PROC(SQLGetStmtOption,
	  (SQLHSTMT   hstmt,
	   UWORD      fOption,
	   PTR        pvParam),
	  (hstmt, fOption, pvParam))
{
    return CALL_ODBC( SQLGetStmtAttr, ( hstmt,
                                        fOption,
                                        pvParam,
                                        SQL_MAX_OPTION_STRING_LENGTH,
                                        NULL ));    
} /* SQLGetStmtOption */


ODBC_PROC(SQLParamOptions,
          (SQLHSTMT hstmt,
           UDWORD   crow,
           UDWORD  *pirow),
          (hstmt, crow, pirow))
{
    RETCODE retcode;

    retcode = CALL_ODBC( SQLSetStmtAttr,
                         ( hstmt, SQL_ATTR_PARAMSET_SIZE, (PTR) crow, 0 ));
    if (retcode == SQL_SUCCESS) {
        retcode = CALL_ODBC( SQLSetStmtAttr,
                             ( hstmt, SQL_ATTR_PARAMS_PROCESSED_PTR,
                               pirow, 0 ));
    }; /* if */

    return retcode;
} /* SQLParamOptions */


ODBC_PROC(SQLSetConnectOption,
	  (SQLHDBC    hdbc,
	   UWORD      fOption,
	   UDWORD     vParam),
	  (hdbc, fOption, vParam))
{
    return CALL_ODBC( SQLSetConnectAttr,
                      ( hdbc, fOption, (PTR) vParam, SQL_NTS ));
} /* SQLSetConnectOption */


/* SQLSetParam still exists in vpa10gc */

/* SQLSetScrollOptions -> vpa12pc */

ODBC_PROC( SQLSetStmtOption,
           (SQLHSTMT   hstmt,
            UWORD      fOption,
            UDWORD     vParam),
           (hstmt, fOption, vParam))
{
    return CALL_ODBC( SQLSetStmtAttr,
                      ( hstmt, fOption, (PTR) vParam, SQL_NTS ));
} /* SQlSetStmtOption */


ODBC_PROC(SQLTransact,
          (SQLHENV    henv,
           SQLHDBC    hdbc,
           UWORD      fType),
          (henv, hdbc, fType))
{
    if (hdbc != SQL_NULL_HDBC) {
        return CALL_ODBC( SQLEndTran,( SQL_HANDLE_DBC, hdbc, fType ));
    } else {
        return CALL_ODBC( SQLEndTran,( SQL_HANDLE_ENV, henv, fType ));
    }; /* else */        
} /* SQLTransact */


/*!**********************************************************************

  EndChapter: ODBC functions

************************************************************************/
