/*!================================================================
 module:    vpr01Con.c

 responsible: BurkhardD
 
 special area:   CPCDrv | Client | Connection Verwaltung
 
 description: Connection Verwaltung
 
 see:  
 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "vpr01Con.h"
#include "gmt90.h"
#include "gpr00.h"
#include "gpr01.h"
#include "gpr03.h"
#include "gpr04.h"
#include "gpr08.h"
#include "vpr07Macro.h"
#include "vpr01Env.h"
#include "vpr01Error.h"
#include "vpr01Cursor.h"
#include "vpr01StmtName.h"
#include "vpr01SQL.h"
#include "vpr03Con.h"
#include "vpr02Con.h"
#include "vpr03Segment.h"
#include "vpr06ParseId.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

static tsp00_Int2 SessionCnt = 0;

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

static void pr01ConInitCont(tpr01_ConContainer *Cont, tpr01_EnvDesc *EnvDesc);

static tsp00_Int4 pr01ConOpenCont(tpr01_ConContainer *Cont);

static tsp00_Int4 pr01ConGetCount(tpr01_ConContainer *Cont);

static tpr01_ConDesc *pr01ConEnumDesc(tpr01_ConContainer *Cont);

static tpr01_ConDesc *pr01ConAddDesc(tpr01_ConContainer *Cont, tpr00_DatabaseName szDatabaseName, sqlxatype *sqlxa);

static void pr01ConDeleteDesc(tpr01_ConDesc *ConDesc);

static tsp00_Int4 pr01ConSetIndex(tpr01_ConContainer *Cont, tsp00_Int4 cbIndex);

static  tpr01_ConDesc *pr01ConFindDesc(tpr01_ConContainer *Cont, char *szDatabaseName);

static void pr01ConInitPacket(tpr01_ConDesc *Desc, tsp1_cmd_mess_type MessType);
static void pr01ConConnect(tpr01_ConDesc *Desc, struct tpr01_StmtNameDesc *StmtNameDesc);

static BOOLEAN pr01ConIsConnected(tpr01_ConDesc *Desc);

static void pr01ConAutoConnect (tpr01_ConDesc *ConDesc);

static void pr01ConOpen (tpr01_ConDesc *ConDesc, sqlkaentry *ka);

static void pr01ConRelease (tpr01_ConDesc *ConDesc);

static BOOLEAN pr01ConSetAttr (tpr01_ConDesc *Desc, epr01Con_Attr Attr, void *pAttr, tsp00_Int4 cbLen);

static tpr03_SegmDesc *pr01ConSegmentNew(tpr01_ConDesc *ConDesc);
static void pr01ConSegmentDelete(tpr01_ConDesc *ConDesc);

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

static void pr01ConInitCont(tpr01_ConContainer *Connection, tpr01_EnvDesc *EnvDesc)
{
  Connection->Desc = pr09NewDescriptor(3, sizeof(tpr01_ConDesc));
  Connection->EnvDesc    = EnvDesc;
  Connection->sqlca      = EnvDesc->sqlca;
  Connection->NoLog      = cpr_is_false;
  Connection->InitCont   = pr01ConInitCont;
  Connection->OpenCont   = pr01ConOpenCont;
  Connection->GetCount   = pr01ConGetCount;
  Connection->EnumDesc   = pr01ConEnumDesc;
  Connection->AddDesc    = pr01ConAddDesc;
  Connection->DeleteDesc = pr01ConDeleteDesc;
  Connection->SetIndex   = pr01ConSetIndex;
  Connection->FindDesc   = pr01ConFindDesc;
  Connection->InitPacket = pr01ConInitPacket;
  Connection->Connect    = pr01ConConnect;
  Connection->IsConnected = pr01ConIsConnected;
  Connection->AutoConnect = pr01ConAutoConnect;
  Connection->Open       = pr01ConOpen;
  Connection->Release    = pr01ConRelease;
  Connection->SetAttr    = pr01ConSetAttr;
  Connection->SegmentNew = pr01ConSegmentNew;
  Connection->SegmentDelete = pr01ConSegmentDelete;
}

void pr01ConDeleteCont(tpr01_ConContainer *Connection)
{
  if (Connection) {
    if (pr01ConOpenCont(Connection)) {
      /* liefert <> 0 wenn Module Descriptoren vorhanden sind */
      tpr01_ConDesc *ConDesc;
      while(ConDesc = Connection->EnumDesc(Connection))
	Connection->DeleteDesc(ConDesc);
    }
    pr09DeleteDescriptor(Connection->Desc);
    Connection->Desc = NULL;
    pr03mFree((tsp00_Uint1*)Connection, "Connection");
  }
}

tpr01_ConContainer *pr01ConNewCont(tpr01_EnvDesc *EnvDesc)
{
  tpr01_ConContainer *Connection;
  Connection = pr03mAllocat(sizeof(tpr01_ConContainer), "Connection");
  if (Connection)
    pr01ConInitCont(Connection, EnvDesc);
  return Connection;
}


static tsp00_Int4 pr01ConOpenCont(tpr01_ConContainer *Connection)
{
  if (!Connection->Desc) {
    Connection->Desc = pr09NewDescriptor(10, sizeof(tpr01_ConDesc));
  }
  pr09SetIndex(Connection->Desc, -1);
  return(Connection->GetCount(Connection));
}

static tsp00_Int4 pr01ConGetCount(tpr01_ConContainer *Connection)
{
  return(pr09GetItemCount(Connection->Desc));
}

static tpr01_ConDesc *pr01ConEnumDesc(tpr01_ConContainer *Connection)
{
  tpr01_ConDesc *Desc=pr09GetNextItem(Connection->Desc);
  return(Desc);
}

static tpr01_ConDesc *pr01ConAddDesc(tpr01_ConContainer *Connection, tpr00_DatabaseName szDatabaseName, sqlxatype *sqlxa)
{
  M90TRACE(M90_TR_ENTRY, "pr01ConAddDesc", 0);
  {
    tpr01_ConDesc *ConDesc;      
    ConDesc = pr09AddItem(Connection->Desc);
    ConDesc->DescType = ConDesc_epr01;
    ConDesc->Connection = Connection;
    strcpy(ConDesc->szDatabaseName, szDatabaseName);
    ConDesc->ErrorDesc = pr01NewErrorDesc();
    ConDesc->Status    = Empty_epr01Con;
    ConDesc->Option    = NoSelectDirectFast_bpr01;
    ConDesc->Seg       = NULL;
    ConDesc->sqlca = Connection->sqlca;
    ConDesc->sqlxa = sqlxa;
    ConDesc->ga = pr03mAllocat(sizeof(sqlgaentry), "ConDesc->ga");
    if (!ConDesc->ga)
      pr07CheckAssert(false);
    ConDesc->XUserRecord = &ConDesc->ga->gaxuserrec;
    p03sqlgaentryinit(ConDesc->ga, 1, ConDesc->sqlca->sqlemp);
    if ( ConDesc->sqlca->sqlemp->elzu != sp1ce_ok) {
      p04trint4(ConDesc->sqlca->sqlrap, "No XUSER entry found. Continue", Connection->sqlca->sqlemp->elzu);
      ConDesc->sqlca->sqlemp->elzu = sp1ce_ok;
    }
    ConDesc->ka = pr01cDupKa(NULL);
    if (!ConDesc->ka)
      pr07CheckAssert(false);
    ConDesc->SessionNr = ++SessionCnt;
    ConDesc->ParseIdCache = NULL; 
    /*Drop Parsid Cache activate */
    ConDesc->ParseIdCache = pr06ParseIDNewCache(ConDesc, 20); 
    ConDesc->SegmDesc = NULL;
    M90TRACE(M90_TR_EXIT, "pr01ConAddDesc", 0);
    return(ConDesc);
  }
}

static void pr01ConDeleteDesc(tpr01_ConDesc *ConDesc)
{
  int cbIndex = pr09ItemFindIndex(ConDesc->Connection->Desc, ConDesc);
  if (cbIndex >= 0) {
    pr01DeleteErrorDesc(ConDesc->ErrorDesc);    
    if (ConDesc->ga)
      pr03mFree(ConDesc->ga, "ConDesc->ga");
    if(ConDesc->ka)
      pr01cFreeKa(ConDesc->ka);
    pr09CloseItemEx(ConDesc->Connection->Desc, cbIndex);
    if (ConDesc->SegmDesc)
      pr01ConSegmentDelete(ConDesc);
  }
}

static tsp00_Int4 pr01ConSetIndex(tpr01_ConContainer *Connection, tsp00_Int4 cbIndex)
{
  return(pr09SetIndex(Connection->Desc, cbIndex));
}

static  tpr01_ConDesc *pr01ConFindDesc (tpr01_ConContainer *Connection, tpr00_DatabaseName szDatabaseName)
{
  M90TRACE(M90_TR_ENTRY, "pr01ConFindDesc", 0);
  {
    tsp00_Int4 cbOldIndex = Connection->SetIndex(Connection, -1);
    tpr01_ConDesc *ConDesc=NULL;
    while(ConDesc=Connection->EnumDesc(Connection)) {
      if (!strcmp(szDatabaseName, ConDesc->szDatabaseName)) {
	M90TRACE(M90_TR_STRING, "szDatabaseName", ConDesc->szDatabaseName);
	break;
      }
    }
    if (!ConDesc)
      Connection->SetIndex(Connection, cbOldIndex);
    M90TRACE(M90_TR_EXIT, "pr01ConFindDesc", 0);
    M90TRACE(M90_TR_HANDLE, "ConDesc", &ConDesc);
    return(ConDesc);    
  }
}

static void pr01ConInitPacket(tpr01_ConDesc *ConDesc, tsp1_cmd_mess_type MessType)
{
  PR07_CHECKDESC(ConDesc, ConDesc_epr01);  
}

static void pr01ConConnect(tpr01_ConDesc *ConDesc, tpr01_StmtNameDesc *StmtNameDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01ConConnect", 0);
  {
    PR07_CHECKDESC(ConDesc, ConDesc_epr01);
    {
      sqlgaentry *ga = ConDesc->ga;
      if (ga->gareference == 0) {
	tpr01_ConContainer *Connection = ConDesc->Connection;
	sqlcatype *sqlca = Connection->sqlca;
	sqlgatype *sqlga = sqlca->sqlgap;
	sqlga->gasqlconntyp = cpr_ci_program;
	ga->gareconnect = cpr_is_false;
	pr02ConAnalyseConnect(ConDesc, StmtNameDesc);
      }
      pr01ConAutoConnect(ConDesc);
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01ConConnect", 0);
}

static BOOLEAN pr01ConIsConnected(tpr01_ConDesc *ConDesc) {
  PR07_CHECKDESC(ConDesc, ConDesc_epr01);
  {
    return (ConDesc->Status == Connected_epr01Con);
  }
}

static void pr01ConRelease (tpr01_ConDesc *ConDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01ConRelease", 0);
  {
    tpr01_ConContainer *Connection = ConDesc->Connection;
    sqlcatype *sqlca = Connection->sqlca;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    if (sqlemp->ereturncode >= 0) {
      Connection->DeleteDesc(ConDesc);
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01ConRelease", 0);
}

static void pr01ConAutoConnect (tpr01_ConDesc *ConDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01ConAutoConnect", 0);
  {
    tpr01_ConContainer *Connection = ConDesc->Connection;
    sqlcatype *sqlca = Connection->sqlca;
    sqlxatype *sqlxa = ConDesc->sqlxa;
    sqlgatype *sqlga = sqlca->sqlgap;
    sqlratype *sqlra = sqlca->sqlrap;
    sqlgaentry *ga   = ConDesc->ga;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    M90TRACE (M90_TR_PTR,   "ga", &ga);
    if (sqlemp->ereturncode >= 0 && ga->gareference == 0) {
      if (sqlga->gasqlconntyp == cpr_ci_program || sqlga->gasqlconntyp == cpr_ci_ora_connect) {
	ga->gastatus = cpr_co_ok;
	pr03ConConnect(ConDesc);
	p01pparsidtrace (sqlca, sqlxa, ConDesc->ka, NULL, TraceParseIDSQLStmt_epr00);
	p01xprofinit(sqlca, sqlxa, ConDesc->ga);
	p03returncodeget(sqlca, sqlxa);
	p01xtimetrace(sqlca, sqlxa, ConDesc->ga);
      }
      else {
	if  (sqlga->gasqlconntyp == cpr_ci_explicit) {
	  p01cmd_trace(sqlca, sqlxa, NULL);
	  p08runtimeerror (sqlca, sqlxa, cpr_session_not_connected);
	}
	else {
	  ga->gastatus = cpr_co_implicit;
	  pr03ConConnect(ConDesc);
	  p01pparsidtrace (sqlca, sqlxa, ConDesc->ka, NULL, TraceParseIDSQLStmt_epr00);
	  p01xprofinit(sqlca, sqlxa, ConDesc->ga);
	  p03returncodeget (sqlca, sqlxa);
	  p01xtimetrace (sqlca, sqlxa, ConDesc->ga);
	}
      }
      if (ga->gareference) {
	ConDesc->Status = Connected_epr01Con;
      }
      sqlga->gasqlconntyp = cpr_ci_connect;
      ga->gareconnect = cpr_is_false;
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01ConAutoConnect", 0);
}

static void pr01ConOpen (tpr01_ConDesc *ConDesc, sqlkaentry *ka)
{
  M90TRACE(M90_TR_ENTRY, "pr01ConOpen", 0);
  {
    PR07_CHECKDESC(ConDesc, ConDesc_epr01);
    memcpy(ConDesc->ka, ka, sizeof(sqlkaentry));
    M90TRACE(M90_TR_EXIT, "pr01ConOpen", 0);
  }
}

static void pr01ConC2C(char *cout, char *cin, int clen, int len)
{
  if (len == CPR_NTS) 
    len = (int)strlen(cin);
  len = PR07MIN(len, clen);
  memcpy(cout, cin, len);  
  cout[len] = '\0';
}

static BOOLEAN pr01ConSetAttr (tpr01_ConDesc *ConDesc, epr01Con_Attr Attr, void *pAttr, tsp00_Int4 cbLen)
{
  M90TRACE(M90_TR_ENTRY, "pr01ConSetAttr", 0);
  {
    PR07_CHECKDESC(ConDesc, ConDesc_epr01);
    switch(Attr) {
    case (DatabaseName_epr01Con) : {
      pr01ConC2C(ConDesc->szDatabaseName, pAttr, sizeof(ConDesc->szDatabaseName), cbLen);
      break;
    } 
    case (ServerNode_epr01Con) : {
      pr07C2P(ConDesc->XUserRecord->xu_servernode, pAttr, sizeof(ConDesc->XUserRecord->xu_servernode));
      break;
    } 
    case (ServerName_epr01Con) : {
      pr07C2P(ConDesc->XUserRecord->xu_serverdb, pAttr, sizeof(ConDesc->XUserRecord->xu_serverdb));
      break;
    } 
    case (UserName_epr01Con) : {
      pr07C2P(ConDesc->XUserRecord->xu_user, pAttr, sizeof(ConDesc->XUserRecord->xu_user));
      break;
    } 
    case (CryptPWD_epr01Con) : {
      int len = PR07MIN(cbLen, sizeof(ConDesc->XUserRecord->xu_password));
      memcpy(ConDesc->XUserRecord->xu_password, pAttr, cbLen);
      break;
    } 
    default: {
      return FALSE;
    }
    }
    return TRUE;
  }
  M90TRACE(M90_TR_EXIT, "pr01ConOpen", 0);  
}

static tpr03_SegmDesc *pr01ConSegmentNew(tpr01_ConDesc *ConDesc)
{
  tpr01_ConContainer *Connection = ConDesc->Connection;
  sqlcatype *sqlca = Connection->sqlca;
  sqlratype *sqlra = sqlca->sqlrap;
  sqlgaentry *ga   = ConDesc->ga;
  tsp1_packet    *pck = (tsp1_packet *) ga->gareqptr;
  if (ConDesc->SegmDesc) 
    pr01ConSegmentDelete(ConDesc);
  ConDesc->SegmDesc = pr03SegmentNew(pck, (unsigned char *)sqlra->ratermid);
  return ConDesc->SegmDesc;
}

static void pr01ConSegmentDelete(tpr01_ConDesc *ConDesc)
{
  if (ConDesc->SegmDesc)
    pr03SegmentDelete(ConDesc->SegmDesc);
  ConDesc->SegmDesc = NULL;
}

