/*!================================================================
 module:    vpr01Con.h

 responsible: BurkhardD
 
 special area:   CPCDrv | Client | Connection Verwaltung
 
 description: Connection Verwaltung
 
 see:  
 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

#ifndef __VPR01_CON_H__
#define __VPR01_CON_H__

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "gpr00.h"
#include "gpr01.h"
#include "vpr01StmtName.h"
#include "vpr03Segment.h"
#include "vpr06ParseId.h"
#include "vpr09DynaDesc.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

typedef enum epr01Con_Status {
  Empty_epr01Con, 
  Connected_epr01Con, 
  Rollback_epr01Con, 
  Commited_epr01Con, 
  Subtrans_epr01Con, 
  Closed_epr01Con, 
  Timeout_epr01Con, 
  Crash_epr01Con
} epr01Con_Status;

typedef enum epr01Con_Attr { 
  DatabaseName_epr01Con, 
  ServerNode_epr01Con, 
  ServerName_epr01Con, 
  UserName_epr01Con,
  CryptPWD_epr01Con
} epr01Con_Attr;

typedef struct tpr01_ConDesc {
  tpr01_DescTypeEnum DescType;
  struct tpr01_ConContainer *Connection;
  struct tpr01_SegContainer *Seg;
  tpr00_DatabaseName szDatabaseName;
  tsp4_xuser_record  *XUserRecord;
  epr01Con_Status Status;
  tpr01_ConnectOptionBinary Option;
  struct tpr01_ErrorDesc    *ErrorDesc;
  tpr06_ParseIDCache *ParseIdCache;
  tpr03_SegmDesc  *SegmDesc;
  sqlxatype  *sqlxa;
  sqlcatype  *sqlca;
  sqlgaentry *ga;
  sqlkaentry *ka;
  tsp00_Int2 SessionNr;
} tpr01_ConDesc;

typedef struct tpr01_ConContainer {
  struct tpr01_EnvDesc *EnvDesc;
  tpr09DynamicDescriptor *Desc;
  void           (*InitCont)   (struct tpr01_ConContainer *Cont, struct tpr01_EnvDesc *EnvDesc);
  tsp00_Int4     (*OpenCont)   (struct tpr01_ConContainer *Cont);
  void           (*CloseCont)  (struct tpr01_ConContainer *Cont);
  tpr01_ConDesc *(*AddDesc)    (struct tpr01_ConContainer *Cont, tpr00_DatabaseName szDatabaseName, sqlxatype *sqlxa);
  void           (*DeleteDesc) (tpr01_ConDesc *ConDesc);
  tsp00_Int4     (*GetCount)   (struct tpr01_ConContainer *Cont);
  tpr01_ConDesc *(*EnumDesc)   (struct tpr01_ConContainer *Cont);
  tsp00_Int4     (*SetIndex)   (struct tpr01_ConContainer *Cont, tsp00_Int4 cbIndex);
  tpr01_ConDesc *(*FindDesc)   (struct tpr01_ConContainer *Cont, tpr00_DatabaseName szDatabaseName);
  void           (*InitPacket) (tpr01_ConDesc *Desc, tsp1_cmd_mess_type MessType);
  tpr03_SegmDesc *(*SegmentNew) (tpr01_ConDesc *Desc);
  void           (*SegmentDelete) (tpr01_ConDesc *Desc);
  BOOLEAN        (*SetAttr) (tpr01_ConDesc *Desc, epr01Con_Attr Attr, void *pAttr, tsp00_Int4 cbLen);
  void           *(*GetAttr) (tpr01_ConDesc *Desc, epr01Con_Attr Attr, void **pAttr, tsp00_Int4 *cbLen);
  BOOLEAN        (*IsConnected) (tpr01_ConDesc *Desc);
  void           (*AutoConnect) (tpr01_ConDesc *ConDesc);
  void           (*Connect)     (tpr01_ConDesc *ConDesc, struct tpr01_StmtNameDesc *StmtNameDesc);
  void           (*Open)     (tpr01_ConDesc *ConDesc, sqlkaentry *ka);
  void           (*Release)  (tpr01_ConDesc *ConDesc);
  void           (*Drop)     (tpr01_ConDesc *ConDesc);
  sqlcatype  *sqlca;
  tsp00_Int2 NoLog;
} tpr01_ConContainer;


/*==================================================================*
 *  PROTOTYPES                                                      *
 *==================================================================*/

struct tpr01_ConContainer *pr01ConNewCont(struct tpr01_EnvDesc *EnvDesc);

void pr01ConDeleteCont(struct tpr01_ConContainer *Cont);

#endif
