/*!================================================================
 module:    vpr01Cursor.c

 responsible: BurkhardD
 
 special area:   CPCDrv | Client | Cursorverwaltung
 
 description: Cursor Verwaltung
 
 see:  
 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "vpr07Macro.h"
#include "vpr01Cursor.h"
#include "vpr01Con.h"
#include "vpr01Error.h"
#include "vpr01StmtName.h"
#include "gpr03.h"
#include "gpr03m.h"
#include "gmt90.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

static void pr01CursorInitCont(tpr01_CursorContainer *Cont);

static tsp00_Int4 pr01CursorOpenCont(tpr01_CursorContainer *Cursor);

static tsp00_Int4 pr01CursorGetCount(tpr01_CursorContainer *Cursor);

static void pr01CursorCloseCont(tpr01_CursorContainer *Cursor);

static tpr01_CursorDesc *pr01CursorEnumDesc(tpr01_CursorContainer *Cursor);

static tpr01_CursorDesc *pr01CursorAddDesc(tpr01_CursorContainer *Cursor, char *szCursorName, tpr01_UniqueID UniqueID, tpr01_ConDesc *ConDesc);

static void pr01CursorDeleteDesc(tpr01_CursorDesc *CursorDesc);

static tpr01_CursorDesc *pr01CursorFindDesc(tpr01_CursorContainer *Cursor, char *szCursorName, tpr01_UniqueID UniqueID);

static tsp00_Int4 pr01CursorSetIndex(tpr01_CursorContainer *Cursor, tsp00_Int4 cbIndex);

static void pr01CursorExecute(tpr01_CursorDesc *CursorDesc);

static tpr01_CursorDesc *pr01CursorDeclare (tpr01_CursorContainer *Cursor, tpr01_StmtNameDesc *StmtNameDesc, tpr01_ConDesc *ConDesc, char *szCursorName, tpr01_UniqueID UniqueID);

static void pr01CursorPrepareOpen(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka);
static void pr01CursorPrepareFetch(tpr01_CursorDesc *CursorDesc, struct tpr01_StmtNameDesc *StmtNameDesc, boolean tryexec);
static void pr01CursorPrepareClose(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka);
static void pr01CursorClose(tpr01_CursorDesc *CursorDesc);

static void pr01CursorAlterCursorName (tpr01_CursorDesc *CursorDesc, char *szCursorName);

static tpr01_ConDesc *pr01CursorGetConnection(tpr01_CursorDesc *CursorDesc);

static BOOLEAN pr01CursorSetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
				  void *pAttr, tsp00_Int4 cbLen);

static void pr01CursorGetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
			       void *pAttr, tsp00_Int4 *cbLen);

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

static void pr01CursorInitCont(tpr01_CursorContainer *Cursor)
{
  Cursor->Desc = pr09NewDescriptor(3, sizeof(tpr01_CursorDesc));
  Cursor->ErrorDesc = pr01NewErrorDesc();
  Cursor->InitCont  = pr01CursorInitCont;
  Cursor->OpenCont  = pr01CursorOpenCont;
  Cursor->CloseCont = pr01CursorCloseCont;
  Cursor->GetCount  = pr01CursorGetCount;
  Cursor->EnumDesc  = pr01CursorEnumDesc;
  Cursor->AddDesc   = pr01CursorAddDesc;
  Cursor->DeleteDesc = pr01CursorDeleteDesc;
  Cursor->SetIndex  = pr01CursorSetIndex;
  Cursor->FindDesc  = pr01CursorFindDesc;
  Cursor->Execute   = pr01CursorExecute;
  Cursor->Declare   = pr01CursorDeclare;
  Cursor->Close     = pr01CursorClose;
  Cursor->PrepareOpen = pr01CursorPrepareOpen;
  Cursor->PrepareFetch = pr01CursorPrepareFetch;
  Cursor->PrepareClose = pr01CursorPrepareClose;
  Cursor->AlterCursorName = pr01CursorAlterCursorName;
  Cursor->GetConnection = pr01CursorGetConnection;
}

void pr01CursorDeleteCont(tpr01_CursorContainer *Cursor)
{
  if (Cursor) {
    if (pr01CursorOpenCont(Cursor)) {
      tpr01_CursorDesc *CursorDesc;
      while(CursorDesc = Cursor->EnumDesc(Cursor))
	Cursor->DeleteDesc(CursorDesc);
    }
    pr01DeleteErrorDesc(Cursor->ErrorDesc);    
    pr09DeleteDescriptor(Cursor->Desc);
    Cursor->Desc = NULL;
    pr03mFree((tsp00_Uint1*)Cursor, "Cursor");
  }
}

tpr01_CursorContainer *pr01CursorNewCont()
{
  tpr01_CursorContainer *Cursor;
  Cursor = pr03mAllocat(sizeof(tpr01_CursorContainer), "Cursor");
  if (Cursor)
    pr01CursorInitCont(Cursor);
  return Cursor;
}


static tsp00_Int4 pr01CursorOpenCont(tpr01_CursorContainer *Cursor)
{
  if (!Cursor->Desc) {
    Cursor->Desc = pr09NewDescriptor(10, sizeof(tpr01_CursorDesc));
  }
  pr09SetIndex(Cursor->Desc, -1);
  return(Cursor->GetCount(Cursor));
}

static tsp00_Int4 pr01CursorGetCount(tpr01_CursorContainer *Cursor)
{
  return(pr09GetItemCount(Cursor->Desc));
}

static tpr01_CursorDesc *pr01CursorEnumDesc(tpr01_CursorContainer *Cursor)
{
  tpr01_CursorDesc *CursorDesc=pr09GetNextItem(Cursor->Desc);
  return(CursorDesc);
}

static void pr01CursorCloseCont(tpr01_CursorContainer *Cursor)
{
  M90TRACE(M90_TR_HANDLE, "pr01CursorCloseCont", Cursor);
  pr09CloseAllItems(Cursor->Desc);
}

void pr01CursorInitAttr (tpr01_CursorDesc *CursorDesc)
{
  BOOLEAN keep = FALSE;

  CursorDesc->SetAttr (CursorDesc, openkeep_epr01Cursor, &keep, 1);
}

static tpr01_CursorDesc *pr01CursorAddDesc(tpr01_CursorContainer *Cursor, char *szCursorName, tpr01_UniqueID UniqueID, tpr01_ConDesc *ConDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorAddDesc", 0);
  {
    tpr01_CursorDesc *CursorDesc;      
    CursorDesc = pr09AddItem(Cursor->Desc);
    CursorDesc->DescType = CursorDesc_epr01;
    CursorDesc->Cursor=Cursor;
    CursorDesc->ConDesc      = ConDesc;
    CursorDesc->StmtNameDesc = NULL;
    CursorDesc->OpenKa       = NULL;
    CursorDesc->FetchKa      = NULL;
    CursorDesc->CloseKa      = NULL;
    CursorDesc->ErrorDesc    = pr01NewErrorDesc();
    CursorDesc->cue = pr03mAllocat(sizeof(sqlcuentry), "CursorDesc->cu");
    pr03cCuInit(CursorDesc->cue);
    CursorDesc->UniqueID = UniqueID;
    if (UniqueID != EmptyUniqueID_cpr01) {
      sprintf(CursorDesc->szCursorName, ":%#08lx", UniqueID);  
    }
    else {
      strcpy(CursorDesc->szCursorName, szCursorName);
    }
    CursorDesc->SetAttr = pr01CursorSetAttr;
    CursorDesc->GetAttr = pr01CursorGetAttr;
    pr01CursorInitAttr (CursorDesc);
    M90TRACE(M90_TR_STRING, "szCursorName", CursorDesc->szCursorName);
    pr07C2P(CursorDesc->cue->cuname, szCursorName, sizeof(CursorDesc->cue->cuname));
    M90TRACE(M90_TR_EXIT, "pr01CursorAddDesc", 0);
    return(CursorDesc);
  }
}

static void pr01CursorDeleteDesc(tpr01_CursorDesc *CursorDesc)
{
  int cbIndex = pr09ItemFindIndex(CursorDesc->Cursor->Desc, CursorDesc);
  if (cbIndex >= 0) {
    pr01DeleteErrorDesc(CursorDesc->ErrorDesc);    
    if (CursorDesc->cue) {
      pr03mFree(CursorDesc->cue, "CursorDesc->cue");
    }
    pr09CloseItemEx(CursorDesc->Cursor->Desc, cbIndex);
  }
}

static tsp00_Int4 pr01CursorSetIndex(tpr01_CursorContainer *Cursor, tsp00_Int4 cbIndex)
{
  return(pr09SetIndex(Cursor->Desc, cbIndex));
}

static tpr01_CursorDesc *pr01CursorFindDesc(tpr01_CursorContainer *Cursor, char *szCursorName, tpr01_UniqueID UniqueID)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorFindDesc", 0);
  {
    tsp00_Int4 cbOldIndex = Cursor->SetIndex(Cursor, -1);
    tpr01_CursorDesc *CursorDesc = NULL;
    if ( UniqueID == EmptyUniqueID_cpr01) {
      while(CursorDesc=Cursor->EnumDesc(Cursor)) {
	if (!strcmp(szCursorName, CursorDesc->szCursorName))
	  break;
      }
    }
    else {
      tsp00_Int4 cbOldIndex = Cursor->SetIndex(Cursor, -1);
      tpr01_CursorDesc *CursorDesc = NULL;
      while(CursorDesc=Cursor->EnumDesc(Cursor)) {
	if (UniqueID == CursorDesc->UniqueID)
	  break;
      }
    }
    if(!CursorDesc)
      Cursor->SetIndex(Cursor, cbOldIndex);    
    M90TRACE(M90_TR_EXIT, "pr01CursorFindDesc", 0);
    M90TRACE(M90_TR_HANDLE, "CursorDesc", &CursorDesc);
    return(CursorDesc);    
  }
}

static void pr01CursorExecute(tpr01_CursorDesc *CursorDesc)
{
  CursorDesc->ErrorDesc->Error->SetError(CursorDesc->ErrorDesc, 1, "UNKNOWN", CPR_NTS);
  return;
}

static void pr01CursorClose(tpr01_CursorDesc *CursorDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorClose", 0);
  M90TRACE(M90_TR_HANDLE, "CursorDesc", &CursorDesc);
  {
    M90TRACE(M90_TR_HANDLE, "StmtNameDesc", &CursorDesc->StmtNameDesc);
    if (CursorDesc->StmtNameDesc) 
    {
      tpr01_CursorContainer *Cursor = CursorDesc->Cursor;
      if (CursorDesc->OpenKa) 
      {
	    CursorDesc->OpenKa = NULL;
      }
      if (CursorDesc->FetchKa) 
      {
        /* drop parseid  PTS 1106501 */
        sqlcatype *sqlca = CursorDesc->ConDesc->sqlca;
        sqlxatype *sqlxa = CursorDesc->ConDesc->sqlxa;
        sqlgaentry *ga   = CursorDesc->ConDesc->ga;
        sqlkaentry *ka   = CursorDesc->FetchKa;  
        p01bdrop_parsid (sqlca, sqlxa, ka->kaParseInfo.ParseId);
        if (CursorDesc->ConDesc->ParseIdCache) 
        {
            pr06ParseIdDrop(sqlca, sqlxa, ga, (tpr06_ParseIDCache *)CursorDesc->ConDesc->ParseIdCache, &ka->kaParseInfo.ParseId);
        }
        else 
        {
            p03dropparsid (sqlca, sqlxa, ga, ka->kaParseInfo.ParseId);
        }

	    pr01cFreeKa(CursorDesc->FetchKa);
	    CursorDesc->FetchKa = NULL;
      }
      if (CursorDesc->CloseKa) 
      {
	    CursorDesc->CloseKa = NULL;
      }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorClose", 0);
}

static void pr01CursorPrepareOpen(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka)
{
  BOOLEAN keep;

  M90TRACE(M90_TR_ENTRY, "pr01CursorPrepareOpen", 0);
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  {
    if (!CursorDesc->OpenKa) {
      CursorDesc->OpenKa = ka;
      if (CursorDesc->StmtNameDesc) {
	tpr01_StmtNameDesc *StmtNameDesc = CursorDesc->StmtNameDesc;
	if (StmtNameDesc->ka->kapacount > 0) {
	  /* Command with :hostvars */
	  M90TRACE(M90_TR_WORD, "kapacount", &StmtNameDesc->ka->kapacount);
	  M90TRACE(M90_TR_WORD, "kapaindex", &StmtNameDesc->ka->kapaindex);
	  CursorDesc->OpenKa->kapacount = StmtNameDesc->ka->kapacount;
	  CursorDesc->OpenKa->kapaindex = StmtNameDesc->ka->kapaindex;
	}
	CursorDesc->OpenKa->kafaindex = StmtNameDesc->ka->kafaindex;
	CursorDesc->OpenKa->kafae = StmtNameDesc->ka->kafae;
      }
    }
    M90TRACE(M90_TR_PTR, "OpenKa", &CursorDesc->CloseKa);
    if (CursorDesc->FetchKa) {
      pr01cFreeKa(CursorDesc->FetchKa);
      CursorDesc->FetchKa = NULL;
    }
    if (CursorDesc->CloseKa) {
      CursorDesc->CloseKa = NULL;
    }
    if (ka->katyp == cpr_com_sql_open_keep) {
      keep = TRUE;
      CursorDesc->SetAttr (CursorDesc, openkeep_epr01Cursor, &keep, sizeof (keep));
    } else {
      keep = FALSE;
      CursorDesc->SetAttr (CursorDesc, openkeep_epr01Cursor, &keep, sizeof (keep));
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorPrepareOpen", 0);
}

static void pr01CursorPrepareFetch(tpr01_CursorDesc *CursorDesc, struct tpr01_StmtNameDesc *StmtNameDesc, boolean tryexec)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorPrepareFetch", 0);
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  PR07_CHECKDESC(StmtNameDesc, StmtNameDesc_epr01);
  {
    if (!CursorDesc->FetchKa) {
      CursorDesc->FetchKa = pr01cDupKa(StmtNameDesc->ka);
    }
    CursorDesc->FetchKa->kacuindex = CursorDesc->OpenKa->kacuindex;
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorPrepareFetch", 0);
}
  
static void pr01CursorPrepareClose(tpr01_CursorDesc *CursorDesc, sqlkaentry *ka)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorPrepareClose", 0);
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  {
    CursorDesc->CloseKa = ka;
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorPrepareClose", 0);
}

static tpr01_CursorDesc *pr01CursorDeclare (tpr01_CursorContainer *Cursor, tpr01_StmtNameDesc *StmtNameDesc, tpr01_ConDesc *ConDesc, char *szCursorName, tpr01_UniqueID UniqueID)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorDeclare", 0);
  M90TRACE(M90_TR_HANDLE, "StmtNameDesc", StmtNameDesc);
  M90TRACE(M90_TR_STRING, "szCursorName", szCursorName);
  {
    tpr01_CursorDesc *CursorDesc = Cursor->FindDesc(Cursor, szCursorName, UniqueID);
    if (!CursorDesc) {
      CursorDesc = Cursor->AddDesc(Cursor, szCursorName, UniqueID, ConDesc);
    }
    if (!StmtNameDesc) {
      tsp00_Int4 ErrorCode;
      tpr01_ErrorText szErrorText;
      pr01ErrorGetErrorText(cpr_unknown_statement_name, &ErrorCode, szErrorText);
      Cursor->ErrorDesc->Error->SetError(Cursor->ErrorDesc, ErrorCode, (char*)szErrorText, CPR_NTS);
      CursorDesc = NULL;
    }
    if (CursorDesc) {
      CursorDesc->StmtNameDesc = StmtNameDesc;
      CursorDesc->ConDesc = ConDesc;
    }
    M90TRACE(M90_TR_EXIT, "pr01CursorDeclare", 0);
    M90TRACE(M90_TR_HANDLE, "CursorDesc", &CursorDesc);
    return CursorDesc;
  }
}
static void pr01CursorAlterCursorName (tpr01_CursorDesc *CursorDesc, char *szCursorName)
{
  PR07_CHECKDESC(CursorDesc, CursorDesc_epr01);
  {
    strcpy(CursorDesc->szCursorName, szCursorName);
    CursorDesc->UniqueID = EmptyUniqueID_cpr01;
  }
}

static tpr01_ConDesc *pr01CursorGetConnection(tpr01_CursorDesc *CursorDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorGetConnection", 0);
  M90TRACE(M90_TR_EXIT, "pr01CursorGetConnection", 0);
  M90TRACE(M90_TR_HANDLE, "ConDesc", &CursorDesc->ConDesc);
  return CursorDesc->ConDesc;
}
 
static void pr01CursorB2B(void *cout, void *cin, int clen, int len)
{
  if (len == CPR_NTS) 
    len = (int)strlen((char*)cin);
  len = PR07MIN(len, clen);
  memcpy((char*)cout, (char*)cin, len);  
}

static BOOLEAN pr01CursorSetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
				  void *pAttr, tsp00_Int4 cbLen)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorSetAttr", 0);
  {
    switch (Attr) {
    case openkeep_epr01Cursor: {
      pr01CursorB2B (&CursorDesc->Attribute.openkeep, pAttr,
		     sizeof (CursorDesc->Attribute.openkeep), cbLen);
      break;
    } 
    default: {
      return FALSE;
    }
    }
    return TRUE;
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorSetAttr", 0);  
}

static void pr01CursorGetAttr (tpr01_CursorDesc *CursorDesc, epr01Cursor_Attr Attr, 
			       void *pAttr, tsp00_Int4 *cbLen)
{
  M90TRACE(M90_TR_ENTRY, "pr01CursorGetAttr", 0);
  {
    switch (Attr) {
    case openkeep_epr01Cursor: {
      *cbLen = sizeof (CursorDesc->Attribute.openkeep);
      pr01CursorB2B (pAttr, &CursorDesc->Attribute.openkeep,
		     sizeof (CursorDesc->Attribute.openkeep), *cbLen);
      break;
    }
    default: {
    }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01CursorGetAttr", 0);  
}
