/*!================================================================
 module:    vpr01StmtName.c

 responsible: BurkhardD
 
 special area:   CPCDrv | Client | Statementnamen 
 description: Verwaltung von Statementnamen
 
 see:  
 
    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include <stdio.h>
#include <ctype.h>
#include "vpr07Macro.h"
#include "gpr03.h"
#include "vpr01StmtName.h"
#include "vpr01Con.h"
#include "vpr01SQL.h"
#include "vpr06ParseId.h"
#include "gpr03.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

#define PR01_MAX_KEYWORD_LEN    (18)
typedef struct pr01StmtNameKeywordTabStruct {
  int  entry;
  char keyword [PR01_MAX_KEYWORD_LEN ];
  tpr01_StmtTypeEnum  StmtType;
} pr01StmtNameKeywordTabStruct;

static pr01StmtNameKeywordTabStruct pr01StmtNameKeywordTab [ ] = {
{ KeywordSELECT_epr01,    "SELECT", StmtSelect_epr01},
{ KeywordDECLARE_epr01,   "DECLARE",StmtDeclare_epr01 },
{ KeywordFETCH_epr01,     "FETCH",  StmtFetch_epr01 },
{ KeywordGETVAL_epr01,    "GETVAL",  StmtGetval_epr01 },
{ KeywordPUTVAL_epr01,    "PUTVAL",  StmtPutval_epr01 },
{ KeywordCLOSE_epr01,     "CLOSE",  StmtClose_epr01},
{ KeywordINSERT_epr01,    "INSERT", StmtOther_epr01},
{ KeywordUPDATE_epr01,    "UPDATE", StmtOther_epr01},
{ KeywordDELETE_epr01,    "DELETE", StmtOther_epr01},
{ KeywordDESCRIBE_epr01,  "DESCRIBE", StmtSelect_epr01 }, /* DESCRIBE ? */
{ KeywordCREATE_epr01,    "CREATE", StmtOther_epr01 },
{ KeywordDROP_epr01,      "DROP",   StmtOther_epr01},
{ KeywordOPEN_epr01,      "OPEN",   StmtOther_epr01 },
{ KeywordAPPEND_epr01,    "APPEND", StmtOther_epr01},
{ KeywordPROC_epr01,      "PROC",   StmtOther_epr01 },
{ KeywordSHOW_epr01,      "SHOW",   StmtSelect_epr01 }, /* DESCRIBE ? */
{ KeywordNEXT_epr01,      "NEXT",   StmtOther_epr01 },
{ KeywordNOT_FOUND_epr01, "",       StmtOther_epr01}
};

static pr01StmtNameKeywordTabStruct pr01StmtNameSelectKeywordTab [ ] = {
{ KeywordFIRST_epr01,   "FIRST",     StmtSingleSelect_epr01},
{ KeywordLAST_epr01,    "LAST",      StmtSingleSelect_epr01},
{ KeywordNEXT_epr01,    "NEXT",      StmtSingleSelect_epr01},
{ KeywordPREV_epr01,    "PREV",      StmtSingleSelect_epr01},
{ KeywordPREV_epr01,    "PREV",      StmtSingleSelect_epr01},
{ KeywordDIRECT_epr01, "DIRECT",     StmtSingleSelect_epr01},
{ KeywordNOT_FOUND_epr01, "",        StmtOther_epr01}
};

typedef struct pr01StmtTypeExtraTabStruct {
  tpr01_StmtTypeEnum  StmtType;
  tpr01_StmtTypeExtraBinary StmtTypeExtra;
} pr01StmtTypeExtraTabStruct;

static pr01StmtTypeExtraTabStruct pr01StmtTypeExtraTab [ ] = {
  { StmtSelect_epr01, HasCursor_bpr01 },
  { StmtFetch_epr01,  HasCursor_bpr01 },
  { StmtGetval_epr01, HasCursor_bpr01 },
  { StmtPutval_epr01, HasCursor_bpr01 },
  { StmtClose_epr01,  HasCursor_bpr01 },
  { StmtOther_epr01,  Empty_bpr01 }
};

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

static void pr01StmtNameInitCont(tpr01_StmtNameContainer *Cont);

static tsp00_Int4 pr01StmtNameOpenCont(tpr01_StmtNameContainer *Cont);

static tsp00_Int4 pr01StmtNameGetCount(tpr01_StmtNameContainer *Cont);

static void pr01StmtNameCloseCont(tpr01_StmtNameContainer *Cont);

static tpr01_StmtNameDesc *pr01StmtNameEnumDesc(tpr01_StmtNameContainer *Cont);

static tsp00_Int4 pr01StmtNameSetIndex(tpr01_StmtNameContainer *Cont, tsp00_Int4 cbIndex);

static tpr01_StmtNameDesc *pr01StmtNameAddDesc(tpr01_StmtNameContainer *Cont, char *szStmtName, tpr01_UniqueID UniqueID, sqlkaentry *ka);

static void pr01StmtNameDeleteDesc(tpr01_StmtNameDesc *StmtNameDesc);

static tpr01_StmtNameDesc *pr01StmtNameFindDesc(tpr01_StmtNameContainer *Cont, char *lsnname, tpr01_UniqueID UniqueID);

static char *pr01StmtNameGetSQL(tpr01_StmtNameDesc *StmtNameDesc, char *szSQLStatement, int *cbLen, tpr01_SQLStatementOptionEnum fOption);

static char *pr01StmtNamePutSQL(tpr01_StmtNameDesc *StmtNameDesc, char *szSQLStatement, int *cbLen, tpr01_SQLStatementOptionEnum fOption);

static void pr01StmtNameTraceStmtName (tpr01_StmtNameDesc *StmtNameDesc, sqlcatype * sqlca);

static void pr01StmtNameSetStmtType(struct tpr01_StmtNameDesc *StmtNameDesc);

static void pr01StmtNameSetState(struct tpr01_StmtNameDesc *StmtNameDesc, tpr01_SQLStateBinary State);

static tpr01_StmtTypeEnum pr01StmtNameAnalyseSQL(char *szSQLStatement);

static void pr01StmtNameDropParsId(tpr01_StmtNameDesc *StmtNameDesc, tpr01_SQLDesc *SQLDesc);
/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

static void pr01StmtNameInitCont(tpr01_StmtNameContainer *Cont)
{
  Cont->Desc = pr09NewDescriptor(3, sizeof(tpr01_StmtNameDesc));
  Cont->InitCont   = pr01StmtNameInitCont;
  Cont->OpenCont   = pr01StmtNameOpenCont;
  Cont->CloseCont  = pr01StmtNameCloseCont;
  Cont->GetCount   = pr01StmtNameGetCount;
  Cont->SetIndex   = pr01StmtNameSetIndex;
  Cont->EnumDesc   = pr01StmtNameEnumDesc;
  Cont->DeleteDesc = pr01StmtNameDeleteDesc;
  Cont->AddDesc    = pr01StmtNameAddDesc;
  Cont->FindDesc   = pr01StmtNameFindDesc;
  Cont->PutSQL     = pr01StmtNamePutSQL;
  Cont->GetSQL     = pr01StmtNameGetSQL;
  Cont->TraceStmtName = pr01StmtNameTraceStmtName;
  Cont->DropParsId = pr01StmtNameDropParsId;
  Cont->SetState   = pr01StmtNameSetState;
}

void pr01StmtNameDeleteCont(tpr01_StmtNameContainer *StmtName)
{
  if (StmtName) {
    if (pr01StmtNameOpenCont(StmtName)) {
      /* liefert <> 0 wenn StmtName Descriptoren vorhanden sind */
      tpr01_StmtNameDesc *StmtNameDesc;
      while(StmtNameDesc = StmtName->EnumDesc(StmtName))
	StmtName->DeleteDesc(StmtNameDesc);
    }
    if (StmtName->Desc)
      pr09DeleteDescriptor(StmtName->Desc);
    pr03mFree((tsp00_Uint1*)StmtName, "StmtName");
  }
}

tpr01_StmtNameContainer *pr01StmtNameNewCont()
{
  tpr01_StmtNameContainer *StmtName;
  StmtName = pr03mAllocat(sizeof(tpr01_StmtNameContainer), "StmtName");
  if (StmtName)
    pr01StmtNameInitCont(StmtName);
  return StmtName;
}


static tsp00_Int4 pr01StmtNameOpenCont(tpr01_StmtNameContainer *StmtName)
{
  if (StmtName->Desc) {
    pr09SetIndex(StmtName->Desc, -1);
  }
  return(StmtName->GetCount(StmtName));
}

static tpr01_StmtNameDesc *pr01StmtNameAddDesc(tpr01_StmtNameContainer *StmtName, char *szStmtName, tpr01_UniqueID UniqueID, sqlkaentry *ka)
{
  M90TRACE(M90_TR_ENTRY, "pr01StmtNameAddDesc", 0);
  M90TRACE(M90_TR_STRING, "szStmtName", szStmtName);
  M90TRACE(M90_TR_WORD, "kalineno", &ka->kalineno);
  {
    tpr01_StmtNameDesc *StmtNameDesc;  
    StmtNameDesc = pr09AddItem(StmtName->Desc);
    StmtNameDesc->DescType = StmtNameDesc_epr01;
    StmtNameDesc->StmtName = StmtName;
    StmtNameDesc->sn = pr03mAllocat(sizeof(sqlsnentry), "StmtNameDesc->sn");
    pr01StmtNameSnEntryInit(StmtNameDesc->sn);
    StmtNameDesc->UniqueID = UniqueID;
    if (UniqueID != EmptyUniqueID_cpr01) {
      sprintf(StmtNameDesc->szStmtName, ":%#08lx", UniqueID);  
    }
    else {
      strcpy(StmtNameDesc->szStmtName, szStmtName);
    }
    pr07C2P(StmtNameDesc->sn->snname, szStmtName, sizeof(StmtNameDesc->sn->snname));
    StmtNameDesc->cbAllocLen = 0;
    StmtNameDesc->szSQLStatement = NULL;
    StmtNameDesc->State = Empty_bpr01;
    StmtNameDesc->StmtType = StmtEmpty_epr01;
    StmtNameDesc->ka = pr01cDupKa(ka);
    StmtNameDesc->ore = pr03mAllocat(sizeof(sqlorentry), "StmtNameDesc->ore");
    pr03cOrInit(StmtNameDesc->ore);    
    StmtNameDesc->cue = pr03mAllocat(sizeof(sqlcuentry), "StmtNameDesc->cue");
    pr03cCuInit(StmtNameDesc->cue);    
    M90TRACE(M90_TR_EXIT, "pr01StmtNameAddDesc", 0);    
    M90TRACE(M90_TR_HANDLE, "StmtNameDesc", &StmtNameDesc);    
    return(StmtNameDesc);
  }
}

static void pr01StmtNameDeleteDesc(tpr01_StmtNameDesc *StmtNameDesc)
{
  int cbIndex = pr09ItemFindIndex(StmtNameDesc->StmtName->Desc, StmtNameDesc);
  if (cbIndex >= 0) {
    if (StmtNameDesc->sn)
      pr03mFree(StmtNameDesc->sn, "StmtNameDesc->sn");
    if (StmtNameDesc->ka)
      pr01cFreeKa(StmtNameDesc->ka);
    if (StmtNameDesc->ore)
      pr03mFree(StmtNameDesc->ore, "StmtNameDesc->ore");
    if (StmtNameDesc->cue)
      pr03mFree(StmtNameDesc->cue, "StmtNameDesc->cue");
    if (StmtNameDesc->szSQLStatement && StmtNameDesc->cbAllocLen > 0)
      pr03mFree(StmtNameDesc->szSQLStatement, "StmtNameDesc->szSQLStatement");
    pr09CloseItemEx(StmtNameDesc->StmtName->Desc, cbIndex);
  }
}

static tsp00_Int4 pr01StmtNameGetCount(tpr01_StmtNameContainer *StmtName)
{
  return(pr09GetItemCount(StmtName->Desc));
}

static tpr01_StmtNameDesc *pr01StmtNameEnumDesc(tpr01_StmtNameContainer *StmtName)
{
  tpr01_StmtNameDesc *StmtNameDesc=pr09GetNextItem(StmtName->Desc);
  return(StmtNameDesc);
}

static void pr01StmtNameCloseCont(tpr01_StmtNameContainer *StmtName)
{
  pr09CloseAllItems(StmtName->Desc);
}

static tsp00_Int4 pr01StmtNameSetIndex(tpr01_StmtNameContainer *StmtName, tsp00_Int4 cbIndex)
{
  return(pr09SetIndex(StmtName->Desc, cbIndex));
}

static tpr01_StmtNameDesc *pr01StmtNameFindDesc(tpr01_StmtNameContainer *StmtName, char *szStmtName, tpr01_UniqueID UniqueID)
{
  M90TRACE(M90_TR_ENTRY, "pr01StmtNameFindDesc", 0);
  {
    tsp00_Int4 cbOldIndex = StmtName->SetIndex(StmtName, -1);
    tpr01_StmtNameDesc *StmtNameDesc=NULL;
    if ( UniqueID == EmptyUniqueID_cpr01) {
      while(StmtNameDesc=StmtName->EnumDesc(StmtName)) {
	if (!strcmp(StmtNameDesc->szStmtName, szStmtName)) {
	  break;
	}
      }
    }
    else {
      while(StmtNameDesc=StmtName->EnumDesc(StmtName)) {
	if (UniqueID == StmtNameDesc->UniqueID) {
	  break;
	}
      }
    }
    if (!StmtNameDesc)
      StmtName->SetIndex(StmtName, cbOldIndex);
    M90TRACE(M90_TR_EXIT, "pr01StmtNameFindDesc", 0);
    M90TRACE(M90_TR_HANDLE, "StmtNameDesc", &StmtNameDesc);
    return(StmtNameDesc);    
  }
}


void pr01StmtNameSnEntryInit(sqlsnentry *sn)
{
  memset(sn->snname, 0, sizeof(sn->snname));
  sn->sntype    = cpr_is_statement;
  sn->snstmtp   = NULL;
  sn->sncount   = 0;
  sn->snindex   = 0;
  sn->snbuflen  = 0;
  sn->snstmtlen = 0;
  sn->snsession = ActualSession_cpr00;
  pr06ParseIdCopy(&sn->snParseInfo.ParseId, ParseIdNull_cpr00);
}

static char *pr01StmtNamePutSQL(tpr01_StmtNameDesc *StmtNameDesc, char *szSQLStatement, int *cbLen, tpr01_SQLStatementOptionEnum fOption)
{
  M90TRACE(M90_TR_ENTRY, "pr01StmtNamePutSQL", 0);
  M90TRACE(M90_TR_HANDLE, "StmtNameDesc", &StmtNameDesc);
  M90TRACE(M90_TR_STRING, "szSQLStatement", szSQLStatement);
  M90TRACE(M90_TR_ENUM, "fOption", &fOption);
  PR07_CHECKDESC(StmtNameDesc, StmtNameDesc_epr01);  
  {
    if (szSQLStatement) {
      int Len;
      int Offset;
      if (cbLen == NULL) {
	Len = (int)strlen(szSQLStatement);
      }
      else {
	Len = *cbLen;
      }
      if (fOption == First_epr01 || fOption == Append_epr01) {
	if (fOption == First_epr01) {
	  Offset = 0;
	  StmtNameDesc->StmtType = StmtEmpty_epr01;
	}
	else {
	  Offset = StmtNameDesc->cbLen;
	}
	if (Len+Offset > StmtNameDesc->cbAllocLen) {
	  if (StmtNameDesc->cbAllocLen > 0)
	    /* once for nullterminator */
	    StmtNameDesc->szSQLStatement = pr03mReallocat(StmtNameDesc->szSQLStatement, Len+1+Offset, "StmtNameDesc->szSQLStatement");
	  else {
	    StmtNameDesc->State |= Allocated_bpr01;
	    StmtNameDesc->szSQLStatement = pr03mAllocat(Len+1, "StmtNameDesc->szSQLStatement");
	  }
	  if (!StmtNameDesc->szSQLStatement) {
	    pr07CheckAssert(TRUE);
	    return NULL;
	  }
	  StmtNameDesc->cbAllocLen = Len+Offset;
	}
	StmtNameDesc->cbLen = Len+Offset;
	memcpy(&StmtNameDesc->szSQLStatement[Offset], szSQLStatement, Len);
	StmtNameDesc->szSQLStatement[Len+Offset] = '\0';
      }
      else {
	if (fOption == Reference_epr01) {
	  StmtNameDesc->cbLen = Len;
	  /* if a Statement already allocated free it */
	  /* but don't free it if the new address is same to the old one */
	  if (StmtNameDesc->cbAllocLen > 0 
	      && StmtNameDesc->szSQLStatement != szSQLStatement) {
	    pr03mFree(StmtNameDesc->szSQLStatement, "StmtNameDesc->szSQLStatement");
	    StmtNameDesc->cbAllocLen = 0;
	  }
	  StmtNameDesc->szSQLStatement = szSQLStatement;      
	}
	else
	  pr07CheckAssert(TRUE);
      }
      if (StmtNameDesc->StmtType == StmtEmpty_epr01)
	pr01StmtNameSetStmtType(StmtNameDesc); 
      szSQLStatement = StmtNameDesc->szSQLStatement;
    }
    M90TRACE(M90_TR_EXIT, "pr01StmtNamePutSQL", 0);
    return(szSQLStatement);
  }
}

static char *pr01StmtNameGetSQL(tpr01_StmtNameDesc *StmtNameDesc, char *szSQLStatement, int *cbLen, tpr01_SQLStatementOptionEnum fOption)
{
  M90TRACE(M90_TR_ENTRY, "pr01StmtNameGetSQL", 0);
  M90TRACE(M90_TR_HANDLE, "StmtNameDesc", &StmtNameDesc);
  M90TRACE(M90_TR_INT, "cbLen", cbLen);
  PR07_CHECKDESC(StmtNameDesc, StmtNameDesc_epr01);  
  {
    if (cbLen == NULL || fOption == Reference_epr01) {
      if (cbLen)
	*cbLen = StmtNameDesc->cbLen;
      M90TRACE(M90_TR_EXIT, "pr01StmtNameGetSQL", 0);
      M90TRACE(M90_TR_STRING, "szSQLStatement", StmtNameDesc->szSQLStatement);
      return(StmtNameDesc->szSQLStatement);
    }
    if (*cbLen > 0 && szSQLStatement && *cbLen >= StmtNameDesc->cbLen) {
      /* if enough space -> copy else return NULL */
      *cbLen = StmtNameDesc->cbLen;
      szSQLStatement[*cbLen] = '\0';
      memcpy(szSQLStatement, StmtNameDesc->szSQLStatement, *cbLen);
      M90TRACE(M90_TR_EXIT, "pr01StmtNameGetSQL", 0);
      M90TRACE(M90_TR_STRING, "szSQLStatement", szSQLStatement);
      return(szSQLStatement);
    }
    return (NULL);
  }
}

static void pr01StmtNameTraceStmtName (struct tpr01_StmtNameDesc *StmtNameDesc, sqlcatype *sqlca)
{
  sqlratype *sqlrap = sqlca->sqlrap;
  sqltatype      *tap = sqlrap->rasqltap;  
  short           *ll = &sqlrap->rasqltap->tastr80l;
  struct SQLERROR errmsg;
  strcpy(tap->tastr80, "Statement Name : ");
  *ll = sprintf(tap->tastr80, "Statement Name : %s", StmtNameDesc->szStmtName);
  p03tvfwritetrace (sqlrap, &errmsg);
}

static void pr01StmtNameSetStmtType(struct tpr01_StmtNameDesc *StmtNameDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01StmtNameSetStmtType", 0);
  PR07_CHECKDESC(StmtNameDesc, StmtNameDesc_epr01);  
  {
    int i;
    StmtNameDesc->StmtType = pr01StmtNameAnalyseSQL(StmtNameDesc->szSQLStatement);  
    M90TRACE(M90_TR_ENUM, "StmtType", &StmtNameDesc->StmtType);
    StmtNameDesc->StmtTypeExtra = (tpr01_StmtTypeExtraBinary)Empty_bpr01;
    for(i=0; i < PR07NUMELEM(pr01StmtTypeExtraTab); i++) {
      if (StmtNameDesc->StmtType == pr01StmtTypeExtraTab[i].StmtType) {
	StmtNameDesc->StmtTypeExtra |= pr01StmtTypeExtraTab[i].StmtTypeExtra;
	M90TRACE(M90_TR_ENUM, "StmtTypeExtra", &StmtNameDesc->StmtTypeExtra);
	break;
      }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01StmtNameSetStmtType", 0);
}

static tpr01_StmtTypeEnum pr01StmtNameAnalyse (char *szSQLStatement, pr01StmtNameKeywordTabStruct *KeywordTab, int cbNumElements)
{
  M90TRACE(M90_TR_ENTRY, "pr01StmtNameAnalyse", 0);
  {
    char *begp, *endp;
    int i;
    tsp00_Longint len;
    tpr01_StmtTypeEnum StmtType = StmtOther_epr01;
    char strp [ PR01_MAX_KEYWORD_LEN + 1];
  
    len = (int)strlen(szSQLStatement);
    /* skip blanks and leading '(' */
    begp=szSQLStatement;
    for (i=0; i < len ;i++) {
      if (isspace(*begp) || *begp == '(') {
	*begp++;  
      }    
      else {
	break;  
      }
    }
    endp=begp;
    for (; i < len ;i++) {
      if (!isspace(*endp)) {
	*endp++;  
      }    
      else {
	break;  
      }
    }
    len = (endp - begp);
    if ( len <= PR01_MAX_KEYWORD_LEN ) {
      memcpy(strp, begp, len);
      strp[len] = 0;
      pr07Put2Up(strp);
      /* compare with the keyword table */
      for (i=0; i < cbNumElements; i++ ) {
	len = (int)strlen(KeywordTab[i].keyword);
	if ( memcmp (KeywordTab[i].keyword, strp, len) == 0 ) {
	  StmtType = KeywordTab[i].StmtType;
	  if (StmtType == StmtSelect_epr01) {
	    StmtType = pr01StmtNameAnalyse(begp+len, pr01StmtNameSelectKeywordTab, PR07NUMELEM(pr01StmtNameSelectKeywordTab));
	    if (StmtType == StmtOther_epr01)
	      StmtType = StmtSelect_epr01;
	  }
	  break;
	}
      }
    }
    M90TRACE(M90_TR_EXIT, "pr01StmtNameAnalyse", 0);
    M90TRACE(M90_TR_ENUM, "StmtType", &StmtType);
    return StmtType;
  }
}

static tpr01_StmtTypeEnum pr01StmtNameAnalyseSQL (char *szSQLStatement)
{
  M90TRACE(M90_TR_ENTRY, "pr01StmtNameAnalyseSQL", 0);
  {
    tpr01_StmtTypeEnum StmtType = StmtOther_epr01;
    StmtType = pr01StmtNameAnalyse(szSQLStatement, pr01StmtNameKeywordTab, PR07NUMELEM(pr01StmtNameKeywordTab));
    M90TRACE(M90_TR_EXIT, "pr01StmtNameAnalyseSQL", 0);
    M90TRACE(M90_TR_ENUM, "StmtType", &StmtType);
    return StmtType;
  }
}

static void pr01StmtNameDropParsId(tpr01_StmtNameDesc *StmtNameDesc, tpr01_SQLDesc *SQLDesc)
{
  if (SQLDesc->ConDesc) {
    sqlcatype *sqlca = SQLDesc->ConDesc->sqlca;
    sqlxatype *sqlxa = SQLDesc->ConDesc->sqlxa;
    sqlgaentry *ga   = SQLDesc->ConDesc->ga;
    sqlkaentry *ka = StmtNameDesc->ka;  
    p01bdrop_parsid (sqlca, sqlxa, ka->kaParseInfo.ParseId);
    if (SQLDesc->ConDesc->ParseIdCache) {
      pr06ParseIdDrop(sqlca, sqlxa, ga, (tpr06_ParseIDCache *)SQLDesc->ConDesc->ParseIdCache, &ka->kaParseInfo.ParseId);
    }
    else {
      p03dropparsid (sqlca, sqlxa, ga, ka->kaParseInfo.ParseId);
    }
    StmtNameDesc->ka->kastate = cpr_state_empty;
    StmtNameDesc->ka->kaParseInfo.sp1rFunctionCode = csp1_nil_fc;  
  }
}

static void pr01StmtNameSetState(struct tpr01_StmtNameDesc *StmtNameDesc, tpr01_SQLStateBinary State)
{
  if (State != Empty_bpr01)
    StmtNameDesc->State |= State;
  else
    StmtNameDesc->State = State;
}

