/*!================================================================
 module:    vpr01c.c

 responsible: BurkhardD
 
 special area:   CPCDrv | Client | SQL Commando Abarbeitung
 
 description: SQL Commando Abarbeitung
 
 see:  
 


    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/

/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "hsp26.h"
#define sql_sqldb
#define sql_oracle
#include "gpr00.h"
#include "gpr01.h"
#include "gpr03.h"
#include "gpr04.h"
#include "gpr05.h"
#include "gpr08.h"
#include "vpr07Macro.h"
#include "vpr01Env.h"
#include "vpr01Cursor.h"
#include "vpr01Precom.h"
#include "vpr01StmtName.h"
#include "vpr01Con.h"
#include "vpr01Module.h"
#include "vpr01SQL.h"
#include "vpr03Segment.h"
#include "vpr06ParseId.h"
#include "vpr01Trace.h"
#include "vpr04Long.h"
/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/* externC void p01lpodbcdata (sqlcatype*, sqlxatype*, sqlgaentry*,
			    tpr_longvarcharrec*, tsp_int2*, char*);
externC void p01l_inputlvc (sqlcatype*, sqlxatype*, sqlgaentry*,
tpr_longvarcharrec*); */

/*==================================================================*
 *  EXPORTED FUNCTIONS                                              *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

tsp00_Int4 pr01P2NTS(char *p1, char *p2, tsp00_Int4 len);
tpr01_CursorDesc *pr01cFindCursorEntry(sqlxatype *sqlxa, char *cuname);
static void pr01cExecute(tpr01_SQLDesc *SQLDesc);

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

sqlcupointer 
pr01cGetCursorDesc(sqlxatype *sqlxap, tpr00_CursorName szCursorName)
{
  sqlcupointer pDesc = NULL;
  sqlcupointer cup = sqlxap->sqlcup;
  sqlcupointer cue = sqlxap->sqlcue;
  int cbMaxCu = (int)(cue-cup);
  int i;
  for (i=0;i<cbMaxCu;i++) {
    if(!strcmp((char*)(cup+i)->cuname, szCursorName)) {
      pDesc = cup+i;
      break;
    }
  }
  return(pDesc);
}

int s30klen(char *str, char val, int cnt);

tsp00_Int4 pr01P2NTS(char *p1, char *p2, tsp00_Int4 len)
{
  tsp00_Int4 l1 = s30klen(p2, bsp_c1, len);
  memcpy(p1, p2, l1);
  p1[l1] = '\0';
  return len;
}

boolean pr01cParsePrepare(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cParsePrepare", 0);
  {
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlratype *sqlra = sqlca->sqlrap;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    sqlprentry *pr = SQLDesc->pr;
    sqlorentry *ore = SQLDesc->ore;
    sqlkaentry *ka   = SQLDesc->ka;
    sqlgaentry *ga = SQLDesc->ConDesc->ga;
    sqlsnentry *sn  = SQLDesc->StmtNameDesc->sn;
    tsp1_packet *pck;
    boolean ret = FALSE;
    if (sqlca->sqlcode == 0) {
      pck = p03cpacketinit(sqlra, ga, sp1m_parse);
      if ( !pr06ParseIdIsEqual(&sn->snParseInfo.ParseId, ParseIdNull_cpr00) ) {
	pr06ParseInfoCopy(&ka->kaParseInfo, &sn->snParseInfo);
      }
      ore->oropennew = cpr_is_init;
      pr01cBeg(SQLDesc);
      if (sqlxa->xacmdinit == cpr_is_true) {
	pr01cExecute(SQLDesc);
	/* holt kommando in auftrags puffer */
	p01searchgroupfunc (sqlca, sqlxa, &SQLDesc->Attribute.massopen, &SQLDesc->Attribute.parspre) ;
	/* untersucht kommando , massen , parspre */
	if (!SQLDesc->Attribute.parspre) {
	  tsp00_Int2 cmdfetch;
	  boolean datafound;	  
	  ore->oropennew = cpr_is_true;
	  if(pr01cParse(SQLDesc, &cmdfetch, &datafound)) {
	    if (sqlca->sqlcode == 0) {
	      sn->snsession = ActualSession_cpr00;
	      pr06ParseInfoCopy(&sn->snParseInfo, &ka->kaParseInfo);
	      ret = TRUE;
	    }
	  }
	}
	else {
	  tsp1_part *part_ptr;
	  p03find_part(sqlra, sp1pk_command, &part_ptr);
	  if (part_ptr != NULL ) {
	    /* parse prepare */
	    pr03SegmentSetAttribute(sqlra->rasegptr, Prepare_epr03, TRUE);
	    if (SQLDesc->Attribute.massopen && 
		sqlra->rasqlansi != cpr_kind_ansi 
		&& sqlca->sqldbmode != cpr_kind_db2) {
	      pr03SegmentSetAttribute(sqlra->rasegptr, Mass_Cmd_epr03, TRUE);
	    }
	    pr01cEnd(SQLDesc);
	    if (sqlca->sqlcode == 0) {
	      sn->snsession = ActualSession_cpr00;
	      pr06ParseInfoCopy(&sn->snParseInfo, &ka->kaParseInfo);
	      ret = TRUE;
	    }
	  }
	}
      }
      if (sqlca->sqlcode == 0 && sqlemp->ereturncode == -4000) {
	sqlca->sqlcode = sqlemp->ereturncode;
      }
      if (sqlca->sqlcode == 0 && SQLDesc->Attribute.parspre) {
	/* belege kaprasid, orcolpointer, kastate = execute */
	p01allocsfinfo(sqlca, sqlxa, ka, ore);
	/* belege  orrespointer, */
	p01resallocinfo(sqlca, sqlxa, ka, ore);
	/* belege  orresnampointer,*/
	p01nameallocinfo(sqlca, sqlxa, ka, ore);
      }
    }
    M90TRACE(M90_TR_EXIT, "pr01cParsePrepare", 0);
    M90TRACE(M90_TR_BOOL, "ret", &ret);
    return(ret);
  }
}
  
void pr01cOpen(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cOpen", 0);
  {
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlkaentry *ka = SQLDesc->ka;
    sqlorentry *ore = SQLDesc->ore;
    tpr01_StmtNameDesc *StmtNameDesc = SQLDesc->StmtNameDesc;
    if (StmtNameDesc && (StmtNameDesc->State & Prepared_bpr01)) {
      sqlsnentry *sn  = StmtNameDesc->sn;
      pr06ParseInfoCopy(&ka->kaParseInfo, &sn->snParseInfo);
      ka->kastate = cpr_state_execute;
      SQLDesc->ore = StmtNameDesc->ore;
    }
    pr01cBeg(SQLDesc);
    while (sqlxa->xacmdinit == sqlIsTrue) {
      pr01cExecu(SQLDesc);
      pr01cEnd(SQLDesc);
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cOpen", 0);
}
  
void pr01cBeg(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cBeg", 0);
  {
    tsp00_Int2 cmdfetch;
    BOOLEAN prepars;
    BOOLEAN preopen;
    BOOLEAN open_cuerrd3;
    BOOLEAN execute_at_fetch;
    tsp1_packet_ptr pck;
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlratype *sqlra = sqlca->sqlrap;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    sqlkaentry *ka   = SQLDesc->ka;
    sqltatype  *sqlta = sqlra->rasqltap;
    sqlgaentry *gae  = SQLDesc->ConDesc->ga;
    sqlorentry *ore  = SQLDesc->ore;
    sqlcuentry *cu = SQLDesc->cu;
    prepars = false;
    preopen = false;
    open_cuerrd3 = false;
    execute_at_fetch = false;
    sqlxa->xacmdinit = cpr_is_true;
    sqlra->rakamode  = ka->kamode;
    if (sqlca->sqlcode == 0) {
      if  (gae->gareconnect == cpr_is_true) {
	ka->kastate = cpr_state_empty;
      }
      if ( ka->kastate == cpr_state_execute
	   && sqlca->sqldbmode == cpr_kind_sapr3 
	   && gae->gareference == 0) {
	ka->kastate = cpr_state_empty;
      }
      if (cu && SQLDesc->ComType != cpr_com_sql_open) {
	if (SQLDesc->ComType == cpr_com_sql_fetch || SQLDesc->ComType == cpr_com_mfetch
	    || ka->kapacount == -2) { /* fetch using descr */
	  sqlkaentry *OpenKa;
	  open_cuerrd3 = true;
	  OpenKa = pr01sIsExecuteAtFetch(sqlxa);
	  if (OpenKa != NULL) {
	    pr01sExecuteAtFetch(SQLDesc);
	  }
	}
      }
      if (ka->kastate == cpr_state_empty) {
	/* 20.2.97 describe infos get from prepare */
	if (ka->kapacount == -3) { /* describe */
	  if (SQLDesc->ComType == cpr_com_describe) {
	    p03csqlcaareainit (sqlca);
	    prepars = true;
	    p01colinitsfinfo (sqlca, sqlxa, ka, ore);
	    p01colmovesfinfo (sqlca, sqlxa, ka, ore);
	    p01nammovesfinfo (sqlca, sqlxa, ka, ore, SQLDesc->ComType);
	    sqlxa->xacmdinit = cpr_is_false;
	  }
	  else {
	    if (SQLDesc->ComType == cpr_com_ora_descr_sel) {
	      p03csqlcaareainit (sqlca);
	      prepars = true;
	      p01resinitsfinfo (sqlca, sqlxa, ka, ore, &prepars);
	      if (prepars) {
		p01resmovesfinfo (sqlca, sqlxa, ka, ore, &prepars);
	      }
	      if (prepars) {
		p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDOther_epr00);
		p01nammovesfinfo (sqlca, sqlxa, ka, ore, SQLDesc->ComType);
		sqlxa->xacmdinit = cpr_is_false;
	      }
	    }
	  }
	}
      }
      if ( ka->kastate == cpr_state_execute
	   || ka->kastate == cpr_state_execute_at_fetch
	   || open_cuerrd3) {
	/* describe bei prepare geparsed */
	if (ka->kapacount == -3 && SQLDesc->ComType != cpr_com_ora_descr_sel) {
	  prepars = true;
	  p01colmovesfinfo(sqlca, sqlxa, ka, ore);
	  sqlxa->xacmdinit = cpr_is_false;
	}
	else {
	  /* open cursor ohne using descriptor  b.f 1.8.94 */
	  if (ka->kapacount == 0 && SQLDesc->ComType == cpr_com_sql_open) {
	    preopen = true;
	  }
	  if (ka->kapacount == -1 && SQLDesc->ComType != cpr_com_declare_cursor) {
	    /* open cursor oder execute bei prepare geparsed */
	    preopen = true;
	    p01colmovesfinfo (sqlca, sqlxa, ka, ore);
	    if (sqlca->sqldbmode == cpr_kind_sapr3) {
	      if (pr01sExecuteAtFetchPossible(sqlxa, ka)) {
		execute_at_fetch = true;
		pr01sPrepareExecuteAtFetch(sqlxa->xaSQLDesc);
	      }
	    }
	  }
	  if (open_cuerrd3) {
	    /* fetch, bei open row not found, wird beendet */
	    if ( cu && cu->cuerrd3 == -100 ) {
	      prepars = true;
	      sqlxa->xacmdinit = cpr_is_false;
	      sqlca->sqlerrd[2] = 0;
	      p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDFetch_epr00);
	      sqlemp->ereturncode = 100;
	      sqlemp->eerrorpos  = 0;
	      sqlemp->etextlen = (sqlint2)strlen("ROW NOT FOUND");
	      memcpy(sqlemp->etext, "ROW NOT FOUND", sqlemp->etextlen);
	      p03returncodeget (sqlca, sqlxa);
	      p01xtimetrace (sqlca, sqlxa, gae);
	      p01xcmdclose (sqlca, sqlxa, gae);
	    }
	    if (cu->custate == cpr_is_true && !prepars) {
	      ka->kastate = cpr_state_parse;
	      cu->custate = cpr_is_init;
	    }
	  }
	}
      }
      if (!preopen && !prepars && !execute_at_fetch) {
	if (ka->kastate == cpr_state_empty) {
	  ka->kastate = cpr_state_parse;
	}
	if (ka->kamacro == cpr_is_true) {
	  sqlmatype *sqlma = sqlca->sqlmap;
	  if (sqlma->maversion != ka->kaversion) {
	    ka->kastate = cpr_state_parse;
	  }
	}
	if (ka->kapacount == -2 && ka->kastate == cpr_state_execute) {
	  /* fetch  cursor oder execute  geparsed */
	  if (ka->kaopendata == cpr_is_true)
	    p01resmovesfinfo (sqlca, sqlxa, ka, ore, &prepars);
	  else
	    p01colmovesfinfo (sqlca, sqlxa, ka, ore);
	}
      }
      if (!prepars && !execute_at_fetch) {
	/* abarbeitung kommand */
	if (sqlxa->xaprogn[0] == ' ' || sqlxa->xaprogc == 0) {
	  /* precompiler check-option check */
	  sqlxa->xacmdinit = cpr_is_false;
	  p08runtimeerror (sqlca, sqlxa, cpr_invalid_program);
	  p01xhalt (cpr_result_sqlxa_invalid);
	}
	else {
	  p01xtracefilecheck (sqlca, sqlxa);
	  if (sqlca->sqlcode != 0) {
	    sqlxa->xacmdinit = cpr_is_false;
	  }
	  else {
	    sqlxa->xacmdinit = cpr_is_true;
	    sqlxa->xastopcnt = 0;
	    ka->kaparamcnt = 0;
	    gae->gastatus = cpr_co_ok;
	    if (SQLDesc->ComType == cpr_com_crea_ind) {
	      ka->kastate = cpr_state_command;
	    }
	    switch (ka->kastate) {
	    case (cpr_state_command) : {
	      SQLDesc->MessType = sp1m_dbs;
	      break;
	    }
	    case (cpr_state_parse) : {
	      SQLDesc->MessType = sp1m_parse;
	      break;
	    }
	    case (cpr_state_execute_at_fetch) : {}
	    case (cpr_state_execute) : {
	      SQLDesc->MessType = sp1m_execute;
	      break;
	    }
	    default : {
	      SQLDesc->MessType = sp1m_nil;
	      p08runtimeerror (sqlca, sqlxa, cpr_invalid_commandstate);
	      break;
	    }
	    }
	    if ( sqlemp->elzu != sp1ce_ok) {
	      if ( sqlta->tatrout != cpr_trace_off) {
		/* leerzeile drucken */
		p08puttracename (sqlca, sqlxa, NULL, -1, true);
		p08puttracename (sqlca, sqlxa, NULL, -1, true);
	      }
	      p03returncodeget (sqlca, sqlxa);
	      p01xcmdclose (sqlca, sqlxa, gae);
	      sqlxa->xacmdinit = cpr_is_false;
	    }
	    else {
	      if (sqlca->sqlcode == 0) {
		pck = p03cpacketinit (sqlra, gae, SQLDesc->MessType);
	      }
	      else {
		p03ccmdinit(sqlca, gae, SQLDesc->MessType);
	      }
	      if (ka->kastate == cpr_state_execute 
		  || ka->kastate == cpr_state_execute_at_fetch) {
		p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDSQLStmtBlank_epr00);
		cmdfetch = cpr_is_true;
		p01adbsexecute (sqlca, sqlxa, gae, ka, ore, cu, &cmdfetch, &SQLDesc->MessType, true);
		if (ka->kastate != cpr_state_parse) {
		  /* nur wenn nicht parse again */
		  p01xcmdclose (sqlca, sqlxa, gae);
		}
	      }
	    }
	  }
	}
      }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cBeg", 0);
}

static void pr01cExecute(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cExecute", 0);
  {
    tsp00_Int4 len;
    tsp00_Int4 pos;
    tsp00_Int4 sympos;
    tsp00_Int4 symlen;
    tpr00_Symbol symb;
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlratype *sqlra = sqlca->sqlrap;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    sqlkaentry *ka   = SQLDesc->ka;
    sqlprentry *pr   = SQLDesc->pr;
    sqlgaentry *gae   = SQLDesc->ConDesc->ga;
 
    len = 0;
    if (ka->kapacount == -1 ) {
      if (pr->prDescribe == cpr_is_false ) {
	pr->prDescribe = cpr_is_true;
      }
    }
    if (SQLDesc->ComType == cpr_com_ora_descr_sel) {
      /* oracle describe select list analyse */
      /* erzeugt fetch kommando im sp1p_buf  */
      p01oradescribe (sqlca, sqlxa, gae);
    }
    else {
      tsp1_part *part_ptr;
      char *szSQLStatement;
      int cbLen;
      if (sqlca->sqlcode == 0) {
	p03cpacketinit (sqlra, gae, SQLDesc->MessType);
      }
      else {
	p03ccmdinit(sqlca, gae, SQLDesc->MessType);
      }
      szSQLStatement = SQL->GetSQL(SQLDesc, NULL, &cbLen, Reference_epr01);
      p03cputpart(sqlra, gae, sp1pk_command, szSQLStatement, cbLen, sqlemp);
      p03find_part (sqlra, sp1pk_command, &part_ptr);
      if ( part_ptr != NULL ) {
	if (SQLDesc->ComType == cpr_com_sql ) {
	  len = part_ptr->variant.C_2.sp1p_buf_len_F;
	  if ( len > 0 ) {
	    char *sp1p_buf_F = (char*)part_ptr->variant.C_1.sp1p_buf_F;
	    p05nextsymbol (sp1p_buf_F, len, 1, &sympos, &symlen, &symb);
	    switch (part_ptr->variant.C_1.sp1p_buf_F[sympos-1]) {
	    case ('c'): {}
	    case ('C'): {
	      if ( p05eq((unsigned char*)"CONNECT ", sp1p_buf_F, sympos, 8)) {
		ka->katyp = cpr_com_sql_conn;
		SQLDesc->ComType  = cpr_com_sql_conn;
	      }
	      else {
		if (p05eq((unsigned char*)"CREATE ", sp1p_buf_F, sympos, 7)) {
		  pos = sympos + symlen;
		  p05nextsymbol (sp1p_buf_F, len, pos, &sympos, &symlen, &symb);
		  /* create unique find */
		  if (sp1p_buf_F[sympos-1] == 'u' || sp1p_buf_F[sympos-1] == 'U') {
		    pos = sympos + symlen;
		    p05nextsymbol(sp1p_buf_F, len, pos, &sympos, &symlen, &symb);
		  }
		  if (p05eq((unsigned char*)"INDEX ", sp1p_buf_F, sympos, 6)) {
		    ka->katyp  = cpr_com_crea_ind;
		    SQLDesc->ComType  = cpr_com_crea_ind;
		    ka->kastate = cpr_state_command;
		    pr03SegmentSetMessCode(sqlra->rasegptr, sp1m_dbs);
		  }
		}
	      }
	      break;
	    }
	    case ('f'): {}
	    case ('F'): {
	      if ( p05eq((unsigned char*)"FETCH ", sp1p_buf_F, sympos, 6)) {
		ka->katyp = cpr_com_sql_fetch;
		SQLDesc->ComType  = cpr_com_sql_fetch;
	      }
	      break;
	    }
	    default : {
	      ka->katyp = cpr_com_sql;
	      SQLDesc->ComType  = cpr_com_sql;
	      break;
	    }
	    }
	  }
	}
      } 
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cExecute", 0);
}

void pr01cEnd(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cEnd", 0);
  {
    tsp_knl_identifier ergn;
    BOOLEAN descr;
    int intopos;
    tsp00_Int4 retsqlcode;
    tsp1_packet_ptr pck;
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlratype *sqlra = sqlca->sqlrap;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    sqlkaentry *ka   = SQLDesc->ka;
    sqlprentry *pr   = SQLDesc->pr;
    sqlgaentry *gae   = SQLDesc->ConDesc->ga;
    sqlorentry *ore   = SQLDesc->ore;
    sqlcuentry *cu = SQLDesc->cu;

    sqlra->rakamode = ka->kamode;
    if (sqlca->sqlcode != 0) {
      p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDSQLStmt_epr00);
      p01xtimetrace (sqlca, sqlxa, gae);
    }
    else {
      if (ka->kapacount == -2) {
	/* fetch using descriptor */
	pr08cSearchFetch (sqlca, sqlxa, gae, ergn, &descr, &intopos);
	if (descr) {
	  p01makefetchinto (sqlca, sqlxa, gae, intopos);
	  pr01cDynmfetch(SQLDesc);
	}
      }
      if ( ore->oropennew != cpr_is_describe) {
	/* alle kom. ausser  'describe ergname ' */
	p08finishcmd (sqlca, sqlxa, gae, ka);
	if ( sqlca->sqlcode != 0) {
	  p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDSQLStmt_epr00);
	}
      }
      if ( sqlra->raopprof == cpr_is_true 
	   || sqlra->raopprof == cpr_is_only_pc) {
	/* rette kommando fur sysprofile */
	p03putsysprofbuf (sqlra);
      }
      if (cu && cu->cuerrd3 == -100) {
	sqlxa->xacmdinit = cpr_is_false;
	sqlca->sqlerrd[2] = 0;
	p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDFetch_epr00);
	sqlemp->ereturncode = 100;
	sqlemp->eerrorpos  = 0;
	sqlemp->etextlen = (sqlint2)strlen("ROW NOT FOUND");
	memcpy(sqlemp->etext, "ROW NOT FOUND", sqlemp->etextlen);
	p03returncodeget (sqlca, sqlxa);
	p01xtimetrace (sqlca, sqlxa, gae);
	p01xcmdclose (sqlca, sqlxa, gae);      
      }
      else {
	if (ka->kastate == cpr_state_parse) {

	  /* parse - execute kommando */
	  if (ka->katyp == cpr_com_sql_putval) {
	    /* Special handling for putval command */
	    p01cmd_trace (sqlca, sqlxa, sqlxa->xaSQLDesc->ka);
	    p01pparsidtrace (sqlca, sqlxa, sqlxa->xaSQLDesc->ka, 
			     sqlxa->xaSQLDesc->ka->kaParseInfo.ParseId,
			     TraceParseIDSQLStmt_epr00);	    	    
	    pr04LongPutvalInput (sqlca, sqlxa, gae);
	    p01xtimetrace (sqlca, sqlxa, gae);
	    sqlxa->xacmdinit = cpr_is_false;
	  } else if (ka->katyp == cpr_com_sql_getval) {
	    /* Special handling for getval command */
	    p01cmd_trace (sqlca, sqlxa, sqlxa->xaSQLDesc->ka);
	    p01pparsidtrace (sqlca, sqlxa, sqlxa->xaSQLDesc->ka, 
			     sqlxa->xaSQLDesc->ka->kaParseInfo.ParseId,
			     TraceParseIDSQLStmt_epr00);
	    pr04LongGetvalOutput (sqlca, sqlxa, gae, 0);
	    p01xtimetrace (sqlca, sqlxa, gae);
	  } else
	    p01cmd_pars_execute (sqlca, sqlxa, gae, ka, ore, cu);
	}
	else {
	  p01cmd_execute_immediate (sqlca, sqlxa, gae, ka);
	}
      }
    }
    if ( (sqlxa->xacmdinit == cpr_is_true && sqlca->sqlcode == 0)
	 || sqlca->sqlcode == csp_old_fileversion
	 || sqlca->sqlcode == csp_use_new_pars_info
	 || sqlca->sqlcode == cpr_file_or_table_droped) {

      /* parse again */
      retsqlcode = sqlca->sqlcode;
      /* ADIS 1001501 */
      if (retsqlcode != 0) {
	p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDAgain_epr00);
      }
      p01bdrop_parsid (sqlca, sqlxa, ka->kaParseInfo.ParseId);
      if (SQLDesc->ConDesc->ParseIdCache) {
        pr06ParseIdDrop(sqlca, sqlxa, gae, (tpr06_ParseIDCache *)SQLDesc->ConDesc->ParseIdCache, &ka->kaParseInfo.ParseId);
      }
      else {
	p03dropparsid (sqlca, sqlxa, gae, ka->kaParseInfo.ParseId);
      }
      SQLDesc->MessType = sp1m_parse;
      if (sqlca->sqlcode == 0) {
	pck = p03cpacketinit (sqlra, gae, SQLDesc->MessType);
      }
      else {
	p03ccmdinit(sqlca, gae, SQLDesc->MessType);
      }
      ka->kaparamcnt = 0;
      sqlxa->xacmdinit = cpr_is_true;
      sqlxa->xastopcnt = sqlxa->xastopcnt + 1;
      if (sqlxa->xastopcnt >  Errorloop_mxpr00) {
	sqlxa->xacmdinit = cpr_is_false;
	/* ADIS 1001501 */
	if ( retsqlcode != 0) {
	  sqlca->sqlcode = retsqlcode;
	}
      }
    }
    else {
      p01xcmdclose (sqlca, sqlxa, gae);
      sqlxa->xacmdinit = cpr_is_false;
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cEnd", 0);
}

void pr01cExecu(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cExecu", 0);
  {
    tsp_knl_identifier ergn;
    BOOLEAN descr = false;
    int intopos = 0;
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlratype *sqlra = sqlca->sqlrap;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    sqlkaentry *ka   = SQLDesc->ka;
    sqlprentry *pr   = SQLDesc->pr;
    sqlgaentry *gae   = SQLDesc->ConDesc->ga;
 
    pr01cExecute(SQLDesc);
    if (ka->kapacount == 0 
	&& SQLDesc->ComType != cpr_com_sql_open
	&& (pr && pr->prStmtName[0] == bsp_c1)) {
      /* execute immediate command */
      p01_katyp_get (sqlca, sqlxa);
      if ( sqlca->sqlcode == 0 ) {
	ka->kastate = cpr_state_command;
	pr03SegmentSetMessCode(sqlra->rasegptr, sp1m_dbs);
      }
    }
    else {
      if (SQLDesc->ComType == cpr_com_sql_fetch
	  || SQLDesc->ComType == cpr_com_mfetch) {
	pr08cSearchFetch (sqlca, sqlxa, gae, ergn, &descr, &intopos);
	if (descr)
	  p01makefetchinto (sqlca, sqlxa, gae, intopos);
      } else if (SQLDesc->ComType == cpr_com_sql_putval) {
	pr08cSearchCommand (sqlca, sqlxa, gae, &descr, &intopos, "PUTVAL ");
	if (descr)
	  pr08cmakecommandinto (sqlca, sqlxa, gae, intopos);
      } else if (SQLDesc->ComType == cpr_com_sql_getval) {
	pr08cSearchCommand (sqlca, sqlxa, gae, &descr, &intopos, "GETVAL ");
	if (descr)
	  pr08cmakecommandinto (sqlca, sqlxa, gae, intopos);
      } else if (SQLDesc->ComType == cpr_com_sql_close)
	pr08cSearchClose (SQLDesc);      
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cExecu", 0);
}
 
void pr01cTraceStmtName(tpr01_StmtNameDesc *StmtNameDesc, sqlcatype *sqlca)
{
  tpr01_StmtNameContainer *StmtName = StmtNameDesc->StmtName;
  StmtName->TraceStmtName(StmtNameDesc, sqlca);
}

void pr01cOpenCursor(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cOpenCursor", 0);
  {
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlkaentry *ka   = SQLDesc->ka;
    sqlprentry *pr   = SQLDesc->pr;
    sqlgaentry *gae   = SQLDesc->ConDesc->ga;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    tsp_knl_identifier pnam;
    tsp00_Int2 len;
    int macrono;
    char massopen;
    char selectfound;
    int anf;
    int pos;
    tpr_macroline mline;
    if (ka->kapacount == -1 || SQLDesc->ComType == cpr_com_sql_open) {
      massopen = 0;
      if (ka->kaParseInfo.ParseId[10] == csp1_p_mass_select_found 
	  || ka->kaParseInfo.ParseId[10] == csp1_p_reuse_mass_select_found 
	  || ka->kaParseInfo.ParseId[10] == csp1_p_select_for_update_found 
	  || ka->kaParseInfo.ParseId[10] == csp1_p_reuse_update_sel_found
	  || SQLDesc->ComType == cpr_com_sql_open)	{
	selectfound = true;
      }
      else {
	selectfound = false;
      }
      if (SQLDesc->ComType == cpr_com_sql_open && ka->kapaindex < 0) {
	/* array select */
	massopen = true;
      }
      if (selectfound) {
	if (SQLDesc->CursorDesc) {
	  len = (tsp00_Int2)strlen(SQLDesc->CursorDesc->szCursorName);
	  p03cputpart(sqlca->sqlrap, gae, sp1pk_resulttablename, SQLDesc->CursorDesc->szCursorName, 
				  (tsp_int4)strlen(SQLDesc->CursorDesc->szCursorName), sqlemp);
	  if (ka->kamacro == cpr_is_true && SQLDesc->CursorDesc->szCursorName[0] == CPR_MACROCHR) {
	    anf = 1;
	    pos = anf;
	    p08macrosubstitution(sqlca, sqlxa, gae, &anf , &pos , &macrono , mline);
	  }
	  pr07C2P(pnam, SQLDesc->CursorDesc->szCursorName, sizeof(pnam));
	}
	if (sqlca->sqlrap->rasqltap->tatrout != cpr_trace_off) {
	  char *nam;
	  if (massopen) {
	    nam = "   MASS EXECUTE:  ";
	  }
	  else {
	    nam = "        EXECUTE:  ";
	  }
	  p08puttracename(sqlca, sqlxa, nam, mxsp_name, 0);
	  nam = "OPEN CURSOR       ";
	  p08puttracename(sqlca, sqlxa, nam, mxsp_name, 0);
	  p08puttraceknlid(sqlca, sqlxa, pnam, len , 1);
	}
      }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cOpenCursor", 0);
}

void pr01cDynmfetch(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cDynmfetch", 0);
  {
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlkaentry *ka   = SQLDesc->ka;
    if (ka->kapaindex < 0) {
      tsp1_part_ptr part_ptr;
      sqlratype *sqlra = sqlca->sqlrap;
      sqlparpointer  sqlpa = sqlxa->sqlpap+(-ka->kapaindex)-1;
      if (sqlpa->lo.pakind == SQLPARLO) {
	sqlv1pointer sqlv1p = sqlxa->sqlv1p+sqlpa->lo.pava1index-1;
	tsp00_Int4 *sqlloop = &((sqldatype*)sqlca->sqlcxap->xasqldap)->sqlloop;
	if (sqlpa->lo.pava1index != 0) {
	  sqlv2pointer sqlv2p = sqlxa->sqlv2p+sqlv1p->sc.va1indva2-1;
	  if (sqlv2p->va2typ == cpr_vint2) {
	    *sqlloop = (*(tsp00_Int2*)sqlv1p->sc.va1addr);
	  }
	  if (sqlv2p->va2typ == cpr_vint4) {
	    *sqlloop = (*(tsp00_Int4*)sqlv1p->sc.va1addr);
	  }
	}
	else {
	  *sqlloop = sqlpa->lo.paloopmin;
	}
      }
      else {
	p08trruntimeerror(sqlca, sqlxa, cpr_r_loop_init_wrong);
      }
      p03find_part(sqlra, sp1pk_command, &part_ptr);
      if (part_ptr != NULL) {
	pr03SegmentSetAttribute(sqlra->rasegptr, Mass_Cmd_epr03, TRUE);
      }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cDynmfetch", 0);
}

void pr01cDescribe(tpr01_SQLDesc *SQLDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr01cDescribe", 0);
  {
    tsp_knl_identifier ergn;
    BOOLEAN descr = false;
    int intopos = 0;
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlratype *sqlra = sqlca->sqlrap;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    sqlkaentry *ka   = SQLDesc->ka;
    sqlprentry *pr   = SQLDesc->pr;
    sqlgaentry *gae   = SQLDesc->ConDesc->ga;
    sqlorentry *ore  = SQLDesc->ore;
    tpr00_CursorName szErgn;

    if (sqlca->sqldbmode == cpr_kind_internal) {
      sqldatype *da = sqlca->sqlcxap->xasqldap;
      da->sqlkano = sqlxa->xakano;
      da->sqlprno = ka->kaprindex;
    }
    else {
      if (sqlca->sqldbmode == cpr_kind_oracle || sqlca->sqldbmode == cpr_kind_sapr3) {
	SQLDA *da = sqlca->sqlcxap->xasqldap;
	da->sqlkano = sqlxa->xakano;
	da->sqlprno = ka->kaprindex;
      }
    }
    pr->prDescribe = cpr_is_true;
    if (ka->kastate == cpr_state_parse && ore->oropennew == cpr_is_init) {
      if (SQLDesc->ComType == cpr_com_ora_descr_sel) {      
	p04sfi(sqlca, sqlxa, Describe_Columns_epr00);
      }
      else {
	p04sfi(sqlca, sqlxa, Describe_Params_epr00);      
      }
    }
    else {
      p03ccmdinit(sqlca, gae, sp1m_dbs);
      pr01cExecute(SQLDesc);
      pr08cSearchFetch(sqlca, sqlxa, gae, ergn , &descr , &intopos);
      pr07P2C(szErgn, ergn, sizeof(ergn));
      ka->kapacount = -5;
      ore->oropennew = cpr_is_describe;
      if (descr) {
	sqlratype *sqlra = sqlca->sqlrap;
	sqlgatype *sqlga = sqlca->sqlgap;
	p03ccmdinit(sqlca, gae, sp1m_dbs);
	if (ka->kastate == cpr_state_execute) {
	  ka->katyp = cpr_com_ora_descr_sel;
	  SQLDesc->ComType  = cpr_com_ora_descr_sel;
	}
	ka->kapacount = -4;
	if (SQLDesc->ComType != cpr_com_ora_descr_sel || szErgn[0] != '\0') {
	  char szSQL[sizeof(szErgn)+15];      
	  strcpy(szSQL, "DESCRIBE ");
	  if (ergn[0] != ' ') {
	    strcat(szSQL, "\"");
	    strcat(szSQL, szErgn);
	    strcat(szSQL, "\"");
	  }
	  p03cputpart(sqlra, gae, sp1pk_command, szSQL, (tsp_int4)strlen(szSQL), sqlemp);
	  if (SQLDesc->ComType == cpr_com_ora_descr_sel && szErgn[0] == '\0') {
	    p03cpparsid(sqlra, gae, ka->kaParseInfo.ParseId, sqlca->sqlemp);
	  }
	  pr01cEnd(SQLDesc);
	}
      }
    }
  }
  M90TRACE(M90_TR_EXIT, "pr01cDescribe", 0);
}

boolean pr01cIsComType(tpr01_SQLDesc *SQLDesc, tpr00_ComTypeEnum ComType)
{
  return (SQLDesc->ComType == ComType);
}

void pr01cStmtNameSetState(tpr01_StmtNameDesc *StmtNameDesc, tpr01_SQLStateBinary State)
{
  tpr01_StmtNameContainer *StmtName = StmtNameDesc->StmtName;
  StmtName->SetState(StmtNameDesc, State);
}

/* returns true if execute is possible or false if execute is not possible
   or the command is already executed by parsing */
boolean pr01cParse(tpr01_SQLDesc *SQLDesc, tsp00_Int2 *cmdfetch, boolean *datafound)
{
  tpr01_SQLContainer *SQL = SQLDesc->SQL;
  sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
  sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
  sqlratype *sqlra = sqlca->sqlrap;
  sqlkaentry *ka   = SQLDesc->ka;
  sqlgaentry *gae  = SQLDesc->ConDesc->ga;
  sqlorentry *ore  = SQLDesc->ore;
  sqlcuentry *cue  = SQLDesc->cu;
  struct SQLERROR *sqlemp = sqlca->sqlemp;
  boolean mass_cmd;
  BOOLEAN descr;
  boolean IsParsed = false;
  tsp_knl_identifier ergn;
  int intopos;
  p01bbparsen (sqlca, sqlxa, gae, ka, cmdfetch);
  p01cmd_trace(sqlca, sqlxa, ka);
  if (SQLDesc->ComType == cpr_com_ora_descr_sel) {
    /* describe resultname kommando */
    p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDInput_epr00);
  }    
  pr08cSearchFetch (sqlca, sqlxa, gae, ergn, &descr, &intopos);
  if (descr) {
    SQLDesc->ComType = cpr_com_fetch_describe;
    if (pr01TraceIsTrace(sqlra)) {
	pr01TracePrintf(sqlra, "IGNORE PARSING");
	p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDSQLStmt_epr00);
	p01xtimetrace(sqlca, sqlxa, gae);
    }
    IsParsed = false;
  }
  else {
    /* segment packet absetzen parsen  */
    p01sbparsen (sqlca, sqlxa, gae, ka);
    p03modulnameput (sqlca, sqlxa, gae, ka);
    pr03AbapInfoPutPart(sqlca, sqlra, gae);
    mass_cmd = pr03SegmentGetAttribute(sqlra->rasegptr, Mass_Cmd_epr03);
    p03reqrecpacket (sqlca, gae);
    /* ADIS 1001311 */
    if ( sqlemp->ereturncode == 0) {
      p01resultnget (sqlca, sqlxa, gae);
      ka->kaParseInfo.sp1rFunctionCode = pr01sGetFunctionCode(sqlra);
      p03gparsid (sqlra, sqlemp, ka->kaParseInfo.ParseId);
      ka->kaparserrc = 0;
      p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDSQLStmt_epr00);
      if   (sqlemp->ewarning[0] != bsp_c1) {
	ka->kaparserrc = *((tsp1_packet*)gae->garecptr)->variant.C_2.sp1_segm_F.variant.C_1.sp1s_segm_header_F.variant.C_2.sp1r_extern_warning_F;
      }    
    }
    if ( sqlemp->ereturncode != 0) {
      p03returncodeget (sqlca, sqlxa);
      p01xtimetrace (sqlca, sqlxa, gae);
      sqlxa->xacmdinit = cpr_is_false;
      IsParsed = false;
    }
    else {    
      IsParsed = true;
      p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDOutput_epr00);
      p08privicmd(sqlca, sqlxa, ka, cmdfetch);
      if ( sqlemp->ereturncode != 0) {
	p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDSQLStmt_epr00);
      }
      *datafound = false;
      if (sqlra->raparse_again == FALSE) {
	p01opendata(sqlca, sqlxa, ka, cue, mass_cmd, datafound);
      }
      if (!*datafound) {
	p04sfi (sqlca, sqlxa, Describe_Params_epr00);
      }
      if (!*datafound) {
	p01baparsen (sqlca, sqlxa, ka, cmdfetch);
      }
      p01saparsen (sqlca, sqlxa, gae, ka);
      if (ka->kapacount < -3) {
	sqlxa->xacmdinit = cpr_is_false;
	p01xtimetrace (sqlca, sqlxa, gae);
	IsParsed = false;
      }
      else {
	if (ka->kaParseInfo.ParseId[cpr_p_precom_index] == csp1_p_command_executed) {
	  /* kom. already executed */
	  p03returncodeget (sqlca, sqlxa);
	  p01xtimetrace (sqlca, sqlxa, gae);
	  ka->kastate = cpr_state_command;
	  sqlxa->xacmdinit = cpr_is_false;
	  IsParsed = false;
	}
	else {
	  if (ka->kaParseInfo.ParseId[cpr_p_precom_index] == csp1_p_use_adbs) {
	    /* kom. use execute immed. */
	    ka->kastate = cpr_state_command;
	    p01pparsidtrace(sqlca, sqlxa, ka, ka->kaParseInfo.ParseId, TraceParseIDUseExecute_epr00);
	    IsParsed = false;
	  }
	}
      }
    }
  }
  return IsParsed;
}

sqlorentry *pr01cFromOpenGetOr(tpr01_SQLDesc *SQLDesc)
{
  if (SQLDesc && SQLDesc->CursorDesc)
    return SQLDesc->CursorDesc->StmtNameDesc->ore;
  else
    return NULL;
      
}

sqlkaentry *pr01cDupKa(sqlkaentry *kae)
{
  sqlkaentry *kaedup = pr03mAllocat(sizeof(sqlkaentry), "kae");  
  if (kae) {
    memcpy(kaedup, kae, sizeof(sqlkaentry));
    if (kae->kafaindex > 0) {
      kaedup->kafae = pr03mAllocat(sizeof(sqlfaentry), "kae->kafae");    
    }  
    else {
      kaedup->kafae = NULL;
    }
  }
  return kaedup;
}

void pr01cFreeKa(sqlkaentry *kae)
{
  if (kae->kafae) {
    pr03mFree(kae->kafae, "kae->kafae");
  }  
  pr03mFree(kae, "kae");
}

/* PTS 1108103 */
/* Aendert bei ParseId in allen erforderlichen Descriptoren nachdem der 
Kern das Kommando intern neu geparst hat. Der Kern hat zuvor den Fehlercode
 -9 geliefert und im ReturnSegment einen Parsidpart geliefert.
Geandert werden muessen zusaetzlich die Shortfieldinfos in der or Area */
void pr01cUseNewParseInfo(tpr01_SQLDesc *SQLDesc)
{
  tsp00_Int2 mfindex;

  if (SQLDesc) {
    tpr01_SQLContainer *SQL = SQLDesc->SQL;
    sqlcatype *sqlca = SQL->GetSqlca(SQLDesc);
    sqlxatype *sqlxa = SQL->GetSqlxa(SQLDesc);
    sqlratype *sqlra = sqlca->sqlrap;
    struct SQLERROR *sqlemp = sqlca->sqlemp;
    tpr00_ParseInfo NewParseInfo, OldParseInfo; 
    NewParseInfo.sp1rFunctionCode = pr01sGetFunctionCode(sqlra);
    p03gparsid (sqlra, sqlemp, NewParseInfo.ParseId);
    pr06ParseInfoCopy(&OldParseInfo, &SQLDesc->ka->kaParseInfo);    
    p01pparsidtrace (sqlca, sqlxa, SQLDesc->ka,
          OldParseInfo.ParseId, TraceParseIDChange_epr00);
    p01pparsidtrace (sqlca, sqlxa, SQLDesc->ka, NewParseInfo.ParseId, TraceParseIDNew_epr00);
    mfindex = p01bmfentryget(sqlca, sqlxa, OldParseInfo.ParseId);
    if  (mfindex > 0) {
      memcpy(&sqlca->sqlmfp->mfentry[mfindex-1].mfSelparsid, &NewParseInfo.ParseId, ParseId_mxpr00 );
    }
    pr06ParseInfoCopy(&SQLDesc->ka->kaParseInfo, &NewParseInfo);

    p01allocsfinfo(sqlca, sqlxa, SQLDesc->ka, SQLDesc->ore);
    p01colinitsfinfo (sqlca, sqlxa, SQLDesc->ka, SQLDesc->ore);
    p01colmovesfinfo (sqlca, sqlxa, SQLDesc->ka, SQLDesc->ore);
    p01resallocinfo (sqlca, sqlxa, SQLDesc->ka, SQLDesc->ore);
    p01nameallocinfo (sqlca, sqlxa, SQLDesc->ka, SQLDesc->ore);
    
    /* now change other the ParseId's */
    if (pr01TraceIsTrace(sqlra)) 
      pr01TracePrintf(sqlra, "CHANGE SQLDesc");    

    if (SQLDesc->StmtNameDesc) {
      tpr01_StmtNameContainer *StmtName = SQLDesc->StmtNameDesc->StmtName;
      tpr00_ParseInfo *ParseInfo = &SQLDesc->StmtNameDesc->ka->kaParseInfo;
      if (pr06ParseIdIsEqual(&OldParseInfo.ParseId, &ParseInfo->ParseId)) {
	pr06ParseInfoCopy(ParseInfo, &NewParseInfo);
	if (pr01TraceIsTrace(sqlra)) 
	  pr01TracePrintf(sqlra, "CHANGE SQLDesc->StmtNameDesc");
	StmtName->TraceStmtName(SQLDesc->StmtNameDesc, sqlca);
      }
    }
    if (SQLDesc->StmtNameDesc) {
      tpr01_StmtNameContainer *StmtName = SQLDesc->StmtNameDesc->StmtName;
      tsp00_Int4 cbOldIndex = StmtName->SetIndex(StmtName, -1);
      tpr01_StmtNameDesc *StmtNameDesc=NULL;
      while(StmtNameDesc=StmtName->EnumDesc(StmtName)) {
	sqlsnentry *sn  = StmtNameDesc->sn;
	if (pr06ParseIdIsEqual(&OldParseInfo.ParseId, &sn->snParseInfo.ParseId)) {
	  pr06ParseInfoCopy(&sn->snParseInfo, &NewParseInfo);
	  if (pr01TraceIsTrace(sqlra))
	    pr01TracePrintf(sqlra, "CHANGE SNSTMTNAME");
	  StmtName->TraceStmtName(StmtNameDesc, sqlca);
	}
      }
      StmtName->SetIndex(StmtName, cbOldIndex);
    }
    if (SQLDesc->CursorDesc) {
      tpr00_ParseInfo *ParseInfo = &SQLDesc->CursorDesc->OpenKa->kaParseInfo;
      tpr01_StmtNameContainer *StmtName = SQLDesc->CursorDesc->StmtNameDesc->StmtName;
      if (pr06ParseIdIsEqual(&OldParseInfo.ParseId, &ParseInfo->ParseId)) {
	pr06ParseInfoCopy(ParseInfo, &NewParseInfo);
	StmtName->TraceStmtName(SQLDesc->StmtNameDesc, sqlca);
	if (pr01TraceIsTrace(sqlra)) 
	  pr01TracePrintf(sqlra, "CHANGE SQLDesc->CursorDesc");
      }
    }
  }
}

