/*!================================================================

  module:       vpr02Con.c

 -------------------------------------------------------------------

  responsible:  BurkhardD

  special area: Connect Syntax Analyse

  description:  

  see also:

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/


/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "gmt90.h"
#include "vpr01Con.h"
#include "vpr01SQL.h"
#include "vpr01StmtName.h"
#include "gpr08.h"
#include "vpr02Con.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/
void pr02ConAnalyseConnect(struct tpr01_ConDesc *ConDesc, struct tpr01_StmtNameDesc *StmtNameDesc)
{
  M90TRACE(M90_TR_ENTRY, "pr02ConAnalyseConnect", 0);
  {
    int cbLen;
    tpr01_StmtNameContainer *StmtName = StmtNameDesc->StmtName; 
    char *szSQLStatement = StmtName->GetSQL(StmtNameDesc, NULL, &cbLen, Reference_epr01);
    sqlcatype *sqlca = ConDesc->Connection->sqlca;
    sqlxatype  *sqlxa = ConDesc->sqlxa;
    sqlgaentry     *gaen = ConDesc->ga;
    sqlgatype *sqlga = sqlca->sqlgap;
    p08analyzeconnectstatm (sqlca, sqlxa, sqlga, gaen, StmtNameDesc->ka, cbLen, szSQLStatement);   
  }
  M90TRACE(M90_TR_EXIT, "pr02ConAnalyseConnect", 0);
}

/*!
  Function: pr02ConParseDBString 

  description: Parse einen nach ORACLE Konventionen zusammengesetzten
               Instanznamen.

	       Syntax:   
	              dbstring := [[<networkname>][<servernode>]]<servername>

		      networkname := [<networkstring>]<netsep>
		      servernode  := <nodestring><nodesep>
		      servername  := <dbnamestring>
		      netsep := ':'
		      nodesep := '-'
	       
	       Der Inputstring wird von hinten nach vorne durchsucht um die 
	       Separatoren zu bestimmen. Ist ein <networkname> oder 
	       <servernode> nicht vorhanden oder leer, so zeigen die 
	       Outputwerte auf das Ende des szDBString ('\0').
	       Die Separatoren werden aus szDBString durch '\0' ersetzt.
		      	       
  arguments:
      szDBString   [in/out] Nullterminierter ORACLE Instanzname.
      szNetworName [out] Pointer auf <networkname>.
      szServerNode [out] Pointer auf <servernode>.
      szServerName [out] Pointer auf <servername>.
*/

void pr02ConParseDBString(char *szDBString, char **szNetworkName, char **szServerNode, char **szServerName)
{
  /* points to the end */
  char *szEnd = strrchr(szDBString, '\0');
  char cSep = ':';          /* node separator */
  char *p;
  *szNetworkName = szDBString;
  p = strrchr(szDBString, cSep);
  if (p == NULL) {
    *szServerNode = szDBString;
    *szNetworkName = szEnd;
  }
  else {
    /* drop the separator */
    *p = '\0';
    szDBString = p+1;
  }    
  cSep = '-';
  *szServerNode = szDBString;
  p = strrchr(szDBString, cSep);
  if (p == NULL) {
    cSep = '@';
    p = strrchr(szDBString, cSep);
    if (p == NULL) {
      *szServerName = szDBString;      
      *szServerNode = szEnd;
    }
    else {
      /* drop the separator */
      *p = '\0';
      *szServerName = p+1;
    }
  }
  else {
    *p = '\0';
    *szServerName = p+1;
  }    
}

#ifdef TEST
main()
{
  char *szDBString = "p26235@DB72";
  char *szNetworkName;
  char *szServerNode;
  char *szServerName;
  pr02ConParseDBString(szDBString, &szNetworkName, &szServerNode, &szServerName );
  printf("%s,%s,%s", szNetworkName, szServerNode, szServerName);
}

void p08analyzeconnectstatm ()
{
}
#endif
