
/*!================================================================

  module:       vpr06ParseId.c

 -------------------------------------------------------------------

  responsible:  BurkhardD

  special area: 

  description:  

  see also:

  -------------------------------------------------------------------------

  copyright:    Copyright by SAP AG, 1999



    ========== licence begin LGPL
    Copyright (C) 2000 SAP AG

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
    ========== licence end

 ===================================================================*/


/*==================================================================*
 *  INCLUDES                                                        *
 *==================================================================*/

#include "vpr06ParseId.h"
#include "vpr01Con.h"
#include "gpr03m.h"
#include "gpr03.h"
#include "vpr03Segment.h"
#include "vpr03Part.h"

/*==================================================================*
 *  DECLARATIONS                                                    *
 *==================================================================*/

#define PR06_CMD_DROP_PARSID "DROP PARSEID"

/*==================================================================*
 *  EXTERNAL PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  FUNCTION PROTOTYPES                                             *
 *==================================================================*/

/*==================================================================*
 *  CODE                                                            *
 *==================================================================*/

BOOLEAN pr06ParseIdIsEqual(tpr00_ParseId *p1, tpr00_ParseId *p2)
{
  return ( memcmp(p1, p2, ParseId_mxpr00 ) ? FALSE : TRUE );
}

void pr06ParseIdCopy(tpr00_ParseId *p1, tpr00_ParseId *p2)
{
  memcpy(p1, p2, ParseId_mxpr00);
}

void pr06ParseInfoCopy(tpr00_ParseInfo *p1, tpr00_ParseInfo *p2)
{
  memcpy(p1, p2, sizeof(tpr00_ParseInfo));
}

tpr06_ParseIDCache *pr06ParseIDNewCache(struct tpr01_ConDesc *ConDesc, tsp00_Int4 Size)
{  
  tpr06_ParseIDCache *ParseIDCache = pr03mAllocat(sizeof(tpr06_ParseIDCache), "ParseIDCache");  
  if (ParseIDCache) {
    ParseIDCache->ConDesc = ConDesc;
    ParseIDCache->MaxParseIDs = Size;
    ParseIDCache->ParseIDDesc = pr09NewDescriptor(10, sizeof(tpr00_ParseId));
  }
  return ParseIDCache;
}

externC void pr06ParseIdDrop(sqlcatype *sqlca, sqlxatype *sqlxa, sqlgaentry *ga, tpr06_ParseIDCache *ParseIDCache, tpr00_ParseId *ParseID)
{
  if (ParseIDCache->ConDesc) 
  {
    	tpr01_ConDesc *ConDesc = ParseIDCache->ConDesc;
    	tpr01_ConContainer *Connection = ConDesc->Connection;
    	if (!Connection->IsConnected(ConDesc)) 	
    	/* connection down => delete parseid drop cache */
    	{
      		pr09CloseAllItems(ParseIDCache->ParseIDDesc);
    	}
    	else 
    	{      
    	    if( ! pr06ParseIdIsEqual(ParseID,(tpr00_ParseId*) CPR_PARSIDNULL12)       
    	       && (char)ParseID [CPR_PARS_IDIX] != csp1_p_command_executed 
    	       && (char)ParseID [CPR_PARS_IDIX] != csp1_p_use_adbs)
    	    {
    	        /* put parseid in cache */
      		    tpr00_ParseId *NewParseID = pr09AddItem(ParseIDCache->ParseIDDesc);
      		    pr06ParseIdCopy(NewParseID, ParseID);
      		}
      		if (pr09GetItemCount(ParseIDCache->ParseIDDesc) > ParseIDCache->MaxParseIDs-1) 
                /* cache full => send drop parseid order*/
      	    {
		        tpr03_SegmDesc *SegmDesc;
		        tsp1_part *partPtr; 
		        tpr00_ParseId *ParseID;
		        pr09SetIndex(ParseIDCache->ParseIDDesc, Start_epr09); /* set cache cursor on first element*/
		        SegmDesc = Connection->SegmentNew(ConDesc);			
  
      		   /*for all parseid`s in Cache built order*/
		        while(ParseID = pr09GetNextItem(ParseIDCache->ParseIDDesc)) 
		        {
	  	            pr03SegmentAdd(SegmDesc);		
	  	            partPtr = pr03SegmentAddPart(SegmDesc, sp1pk_command);
	  	            pr03PartPut(partPtr, PR06_CMD_DROP_PARSID, (tsp00_Int4)strlen(PR06_CMD_DROP_PARSID));
		            pr03SegmentFinishPart(SegmDesc);
	  
	  	            partPtr = pr03SegmentAddPart(SegmDesc, sp1pk_parsid);
	  	            pr03PartPut(partPtr, ParseID, sizeof(tpr00_ParseId));
	  	            pr03SegmentFinishPart(SegmDesc);
	            }
		        /* send-recieve part*/	
  		        p03reqrecpacket (sqlca, ga);   /*send dbs message to kernel*/ 
                /* no error handling 
  		        if (sqlca->sqlemp->ereturncode != 0) 
  		        {   
    			    p03returncodeget (sqlca, sqlxa);  
   		        }*/
   				
   		        /* delete parseid drop cache */
                pr09CloseAllItems(ParseIDCache->ParseIDDesc);
      		}
    	} /* end if IsConnected*/
  }/* end if (ParseIDCache->ConDesc) */
  /* set ParseID null */
  pr06ParseIdCopy(ParseID,(tpr00_ParseId*) CPR_PARSIDNULL12);
}
